const semver = require('semver');
const { execSync } = require('child_process');
const { writeFileSync } = require('fs');
const packageJs = require('./package.json');

const REPOS = {
  nrcdeb11100: 11,
};

const DEFAULT_REPO = 'nrcdeb11100';

if (process.argv.length > 3) {
  throw new Error('Too many arguments');
}

const { DEBUG } = process.env;
const currentVersion = packageJs.version;
let [, , releaseVersion] = process.argv;
let releaseBuildNumber = 1;

const parseBuildBranch = (buildTag = '') => {
  if (!buildTag) return '';
  const parsed = buildTag.match(/([0-9]+)\+/);
  if (Array.isArray(parsed) && parsed.length > 1) {
    return Number.parseInt(parsed[1], 10);
  }
};

const parseBuildTag = (buildTag) => {
  if (!buildTag) return '';
  const parsed = buildTag.match(/\+([0-9]+)/);
  if (Array.isArray(parsed) && parsed.length > 1) {
    return Number.parseInt(parsed[1], 10);
  }
};

const splitVersionTags = tag => {
  const [prefix, suffix] = tag.split('%');
  let toSplit = '';
  if (prefix?.includes('@')) {
    toSplit = prefix;
  } else if (suffix?.includes('@')) {
    toSplit = suffix;
  }
  const [version, buildTag, subTag] = toSplit.split('@');
  return {
    version, buildTag, branch: parseBuildBranch(buildTag), subTag, buildNumber: parseBuildTag(buildTag),
  };
};

const tags = execSync('git tag').toString()
  .trim()
  .split('\n')
  .filter(t => !!t.trim())
  .map(splitVersionTags)
  .filter(({ version }) => semver.valid(version));

function findLatestBuildNumber() {
  const builds = tags.filter(({ version, subTag }) => version === releaseVersion && !subTag);
  if (!builds.length) {
    return 0;
  } else {
    return Math.max(...builds.map(({ buildNumber }) => buildNumber));
  }
}

function executeCommand(command) {
  if (DEBUG) {
    console.log(command);
    return '';
  } else {
    return execSync(command).toString();
  }
}

function executeCommands(commands) {
  commands = commands.split('\n').map(c => c.trim()).filter(c => !!c);
  for (const command of commands) {
    executeCommand(command);
  }
}

const highestVersion = tags
  .map(({ version }) => version)
  .reduce((versionA, versionB) => (semver.gt(versionA, versionB) ? versionA : versionB), '0.0.0');

switch (releaseVersion) {
case '--patch':
  releaseVersion = semver.inc(currentVersion, 'patch');
  break;
case '--minor':
  releaseVersion = semver.inc(currentVersion, 'minor');
  break;
case '--major':
  releaseVersion = semver.inc(currentVersion, 'major');
  break;
case undefined:
  releaseVersion = currentVersion;
  break;
default:
}

if (semver.gt(highestVersion, releaseVersion)) {
  throw new Error(`A version with a higher number has already been published: ${highestVersion}`);
}

releaseBuildNumber = findLatestBuildNumber() + 1;
const currentBranch = execSync('git branch')
  .toString()
  .trim()
  .split('\n')
  .find(branch => branch.startsWith('*'))
  .substr(2);

const shouldStash = executeCommand('git diff').trim().length > 0;
if (shouldStash) {
  executeCommand('git stash');
}
executeCommand('git checkout rc');

try {
  if (currentVersion !== releaseVersion) {
    packageJs.version = releaseVersion;
    if (!DEBUG) {
      writeFileSync('./package.json', JSON.stringify(packageJs, null, 2));
    }
    executeCommands(`
git add package.json
git commit --allow-empty -m "bump version: ${releaseVersion}"
git push
`);
  }
  const commands = `
git checkout rc
git pull
git tag 2%${releaseVersion}@nrc${REPOS[DEFAULT_REPO]}+${releaseBuildNumber}
git branch -f debian/${DEFAULT_REPO} 2%${releaseVersion}@nrc${REPOS[DEFAULT_REPO]}+${releaseBuildNumber}
git push --tags -f origin debian/${DEFAULT_REPO}
`;
  executeCommands(commands);
} catch (error) {
  executeCommand(`git checkout ${currentBranch}`);
  if (shouldStash) {
    executeCommand('git stash pop');
  }
}
