/**
  Note: This is the entry-point of the education proxy
  This does the initial bootstrapping
*/

/*
eslint no-console: "off"
*/

import * as path from 'path';
import { ipcMain, desktopCapturer } from 'electron';
import { initialize as initializeRemote } from '@electron/remote/main';

import { ROOT_DIR } from './files';
import { startEduProxy, stopEduProxy } from './services/eduProxy';
import { ensureVersionFile } from './utils/version';

import printHeader from './bootstrap/printHeader';
import createConfigManager from './bootstrap/createConfigManager';
import createEventProxy from './bootstrap/createEventProxy';
import createMessenger from './bootstrap/createMessenger';
import cleanupOldInstance from './bootstrap/cleanupOldInstance';
import setFlags from './bootstrap/setFlags';
import createLogger from './bootstrap/createLogger';
import setNetworkInfo from './bootstrap/setNetworkInfo';
import createPairingServer from './bootstrap/createPairingServer';
import createVersionMonitor from './bootstrap/createVersionMonitor';
import serviceRegistry from './services/serviceRegistry';
import { IPC_GENERAL } from './ipc-constants';

initializeRemote();

// Remove IIFE here when we have support for top-level-await (ES stage 3)
(async () => {
  // eslint-disable-next-line prefer-const
  let logger;

  process.on('unhandledRejection', (error) => {
    if (logger) {
      logger.log('error', 'UNHANDLED EXCEPTION', error);
    }
    console.log('EDU_PROXY ERROR', error);
  });

  const proxyVersion = require(path.join(ROOT_DIR, 'package.json')).version; // eslint-disable-line

  // Set up logging
  logger = createLogger();
  try {
    // As usual for a boostrap procedure, the order of these lines is important

    // Set chrome/electron/node flags
    setFlags(logger);

    // Print start header to the console
    printHeader(logger, proxyVersion);

    // Cleanup possible hanging remainings of a previous run
    cleanupOldInstance(logger);

    // Create the window messenger and set the version info for the identify
    const messenger = await createMessenger(logger, proxyVersion);

    // Loads the configuration
    await createConfigManager();
    createEventProxy();
    ensureVersionFile(logger);

    // Awaits for IP from DHCP and sets the network info to the messenger
    await setNetworkInfo(logger);
    await createPairingServer(logger);

    // Start version monitor which will check for updates and
    // control the start of edu proxy depending on the upgrade status
    const versionMonitor = createVersionMonitor(logger);
    let upgradeMessage;
    const afterUpgrade = () => {
      if (upgradeMessage) {
        upgradeMessage.hide();
        upgradeMessage = null;
      }
      startEduProxy();
      ipcMain.handle(IPC_GENERAL.DESKTOP_CAPTURER_GET_SOURCES, (event, opts) => desktopCapturer.getSources(opts));
    };
    versionMonitor.on('started', async () => {
      if (versionMonitor.isUpgrading) {
        upgradeMessage = messenger.show({ text: 'finishing upgrade' });
      } else {
        try {
          await versionMonitor.checkVersionAndUpgradeIfNeeded();
        } catch (error) {
          // We can ignore this, since Meteor client will keep retrying this anyway
          // This REST-based check is just a safety check in case we ever move from DDP to another protocol
          // So at least continue starting edu proxy
          logger.log('error', 'Could not check/upgrade purple', error);
        }
        startEduProxy();
        ipcMain.handle(IPC_GENERAL.DESKTOP_CAPTURER_GET_SOURCES, (event, opts) => desktopCapturer.getSources(opts));
      }
    });
    versionMonitor.on('upgrade_start', stopEduProxy);
    versionMonitor.on('upgrade_error', afterUpgrade);
    versionMonitor.on('upgrade_done', afterUpgrade);

    const onInterrupt = async (signal) => {
      if (versionMonitor.isUpgrading) {
        logger.log('info', `Received ${signal} during upgrade, ignoring`);
        return;
      }
      logger.log('info', `Received ${signal}, trying to stop gracefully`);
      await stopEduProxy();
      logger.log('info', 'Now exiting');
      // flush buffers
      await serviceRegistry.get('logger').destroy();
      process.exit(0);
    };
    process.once('SIGTERM', onInterrupt.bind(null, 'SIGTERM'));
    process.once('SIGINT', onInterrupt.bind(null, 'SIGINT'));
  } catch (error) {
    logger.log('error', 'FATAL ERROR', error);
  }
})();
