import serviceRegistry from '../services/serviceRegistry';
import VersionMonitor from '../services/upgrades/systemVersionMonitor';
import { waitMaximum, rebootSystem, sleep } from '../utils/utils';

async function start(logger) {
  const versionMonitor = serviceRegistry.get('versionMonitor');
  if (!versionMonitor.checkUrl) {
    logger.info('Version monitor check url not configured, not starting');
    return;
  }

  logger.info('Starting version monitor');
  try {
    await waitMaximum(versionMonitor.start({ retryTimer: 5000 }), 60000, true);
  } catch (error) {
    if (process.env.DISABLE_VERSION_MONITOR_REBOOT) {
      logger.error(`Could not start version monitor because of error:${error}, retrying to start`, {
        serviceLog: true,
      });

      return start(logger);
    } else {
      logger.error(`Could not start version monitor because of error:${error}, rebooting device `, {
        serviceLog: true,
      });
      const messenger = serviceRegistry.get('messenger');
      // Sometimes there's a race condition between network startup and the device startup
      // In that case it can happen that a device gets a wrong ip address via DHCP, but it will not fetch
      // the correct one until the device restarts. This reboot procedure tries to work around this.
      const informUserAboutShutdown = messenger.show({ text: 'Could not start successfully. Rebooting...' });
      await sleep(5000);
      informUserAboutShutdown.hide();
      rebootSystem();
    }
  }
}

export default function createVersionMonitor(logger) {
  logger.info('Creating version monitor');
  const versionMonitor = new VersionMonitor();
  serviceRegistry.set('versionMonitor', versionMonitor);

  const configManager = serviceRegistry.get('configManager');
  const configure = () => {
    logger.info('Version monitor config changed');
    const proxy = configManager.getSetting('device.proxy');
    const restConfig = configManager.getSetting('restRoutes');
    versionMonitor.proxy = proxy;
    versionMonitor.checkUrl = restConfig?.latestVersionInfo;
    versionMonitor.loginUrl = restConfig?.login;
    versionMonitor.upgradeStatusUrl = restConfig?.upgradeStatus;
    if (!versionMonitor.isActive) {
      start(logger);
    }
  };
  configManager.watch('device.proxy', configure);
  configManager.watch('restRoutes', configure);

  return versionMonitor;
}
