import * as pify from 'pify';
import { EventEmitter } from 'events';
import { spawn } from 'child_process';
import * as eventStream from 'event-stream';
import * as constants from 'dbus-native/lib/constants';
import * as stdDbusIfaces from 'dbus-native/lib/stdifaces';
import * as net from 'net';
import * as dbusMessage from 'dbus-native/lib/message';
import * as clientHandshake from 'dbus-native/lib/handshake.js';
import * as serverHandshake from 'dbus-native/lib/server-handshake.js';

import DBusService from './service';
import serviceRegistry from '../services/serviceRegistry';

/**
 * Enum to act as FSM for audio bus objects
 * @type {{UNCONNECTED: number, CONNECTED: number, DESTROYING: number, DESTROYED: number}}
 */
const state = {
  UNCONNECTED: 0,
  CONNECTED: 1,
  DESTROYING: 3,
  DESTROYED: 4,
};

const MESSAGE_TYPE = constants.messageType;

class DBus extends EventEmitter {
  constructor({
    host, direct, port, socket, server, busAddress = process.env.DBUS_SESSION_BUS_ADDRESS,
  }) {
    super();

    const params = {
      host, direct, port, socket, server, busAddress,
    };
    Object.keys(params).forEach((param) => {
      this[`_${param}`] = params[param];
    });

    this._logger = serviceRegistry.get('logger').getInstance('DBus');

    this._methodCallHandlers = {};
    this._init();

    this._state = state.UNCONNECTED;

    this.on('message', this.onMessage.bind(this));
  }

  /**
     * Small utility method to prevent async actions to log after destruction
     */
  _log(...params) {
    if (this._logger) {
      this._logger.log(...params);
    }
  }

  // START FSM getters/setters
  set isConnected(connected) {
    this.state = connected ? state.CONNECTED : state.UNCONNECTED;
  }

  get isConnected() {
    return this._state === state.CONNECTED;
  }

  get isDestroying() {
    return this._state === state.DESTROYING;
  }

  get isDestroyed() {
    return this._state === state.DESTROYED;
  }

  get isActive() {
    return this.state !== state.DESTROYED && this.state !== state.DESTROYING;
  }

  set state(newState) {
    if (newState !== this._state) {
      this._state = newState;
      switch (this._state) {
      case state.CONNECTED:
        this.emit('connected');
        break;
      case state.UNCONNECTED:
        this.emit('disconnected');
        break;
      case state.DESTROYING:
        this.emit('destroyed');
        break;
      default:
      }
    }
  }

  get state() {
    return this._state;
  }
  // END of FSM getters/setters

  _init() {
    this._serial = 1;
    this._methodReturnHandlers = {};
    this._exportedObjects = {};
    this._messages = [];
  }

  async connect() {
    if (this.isConnected) return;

    this._stream = this.createStream();

    this._stream.setNoDelay();

    this._stream.on('error', (error) => {
      // forward network and stream errors
      this._log('verbose', 'Stream error', error);
      this.emit('stream-error', error);
      this._cleanupOnSocketClosed();
    });

    this._stream.once('end', () => {
      this._log('verbose', 'Stream ended');
      this.emit('stream-end');
      this._cleanupOnSocketClosed();
    });

    this._stream.once('connect', () => {
      this._messages.forEach((message) => {
        this._stream.write(dbusMessage.marshall(message));
      });
      this._messages = [];
    });

    try {
      const guid = await pify(this._server
        ? serverHandshake
        : clientHandshake)(this._stream, {});

      this._guid = guid;
      dbusMessage.unmarshalMessages(this._stream, (message) => {
        this.emit('message', message);
      });
      this.isConnected = true;
      return this;
    } catch (error) {
      this._log('error', 'Error while connecting to DBUS', error);
      this.emit('error', error);
      throw error;
    }
  }

  createStream() {
    const parse = function parse(address) {
      const familyParams = address.split(':');
      const family = familyParams[0];

      const params = familyParams[1].split(',').reduce((prev, cur) => {
        const keyVal = cur.split('=');
        prev[keyVal[0]] = keyVal[1];
        return prev;
      }, {});

      return {
        familyParams,
        family,
        params,
      };
    };

    if (this._socket) { return net.createConnection(this._socket); }
    if (this._port) { return net.createConnection(this._port, this._host); }

    if (!this._busAddress) {
      throw new Error('unknown bus address');
    }

    const addresses = this._busAddress.split(';');
    for (const address of addresses) {
      const { family, params } = parse(address);
      try {
        switch (family.toLowerCase()) {
        case 'tcp': {
          const { host, port } = params;
          return net.createConnection(port, host || 'localhost');
        }
        case 'unix':
          if (params.socket) {
            return net.createConnection(params.socket);
          }

          if (params.abstract) {
            const abs = require('abstract-socket'); // eslint-disable-line
            return abs.connect(`\u0000${params.abstract}`);
          }

          if (params.path) {
            return net.createConnection(params.path);
          }
          throw new Error('not enough parameters for \'unix\' connection - you need to specify \'socket\' or \'abstract\' or \'path\' parameter');
        case 'unixexec': {
          const args = [];
          for (let n = 1; params[`arg${n}`]; n++) { args.push(params[`arg${n}`]); }
          const child = spawn(params.path, args);
          return eventStream.duplex(child.stdin, child.stdout);
        }
        default:
          throw new Error(`unknown address type:${family}`);
        }
      } catch (error) {
        if (address === addresses[addresses.length - 1]) {
          throw error;
        }
      }
    }
  }

  end() {
    if (this._stream) {
      this._stream.end();
      this._stream.removeAllListeners();
      delete this._stream;
    }

    Object
      .keys(this._methodReturnHandlers)
      .map(key => this._methodReturnHandlers[key])
      .forEach(handler => handler.reject(new Error('Stream closed')));

    this._init();

    return this;
  }

  /**
     * Destroys all connection-related instance stuff when the connection is closed/lost.
     *
     * Will schedule a reconnection if required.
     * @private
     */
  _cleanupOnSocketClosed() {
    if (this.isConnected || this.isDestroying) {
      const reconnect = this.isActive;
      if (reconnect) {
        this._log('info', `

    =================================================================================
    ==                                                                             ==
    == Connection lost. Cleaning up and triggering reconnect timer                 ==
    ==                                                                             ==
    =================================================================================

            `);
      } else {
        this._log('info', `

    =================================================================================
    ==                                                                             ==
    == Connection lost. Cleaning up.
    ==                                                                             ==
    =================================================================================

            `);
      }

      // Set status variables
      this.isConnected = false;

      // Get rid of the socket
      this.end();

      if (this._reconnectTimer) {
        clearTimeout(this._reconnectTimer);
      }

      if (reconnect) {
        this._reconnectTimer = setTimeout(this.connect.bind(this), 10000);
      }
    }
  }

  message(msg) {
    if (this.isActive) {
      if (!this.isConnected) {
        this._messages.push(msg);
      } else {
        this._stream.write(dbusMessage.marshall(msg));
      }
    }
  }

  invoke(msg) {
    return new Promise((resolve, reject) => {
      if (!this.isConnected) {
        reject(new Error('Not connected'));
        return;
      }

      if (!msg.type) {
        msg.type = MESSAGE_TYPE.methodCall;
      }

      this._serial += 1;
      msg.serial = this._serial;
      this._methodReturnHandlers[msg.serial] = { resolve, reject };
      this.message(msg);
    });
  }

  invokeDbus({ path = '/org/freedesktop/DBus', destination = 'org.freedesktop.DBus', iface = 'org.freedesktop.DBus' }) {
    return this.invoke({ path, destination, interface: iface });
  }

  // bus meta functions
  addMatch(match) {
    if (!this._name) return Promise.resolve();
    return this.invokeDbus({ member: 'AddMatch', signature: 's', body: [match] });
  }

  removeMatch(match) {
    if (!this._name) return Promise.resolve();
    return this.invokeDbus({ member: 'RemoveMatch', signature: 's', body: [match] });
  }

  getId() {
    return this.invokeDbus({ member: 'GetId' });
  }

  requestName(name, flags) {
    return this.invokeDbus({ member: 'RequestName', signature: 'su', body: [name, flags] });
  }

  releaseName(name) {
    return this.invokeDbus({ member: 'ReleaseName', signature: 's', body: [name] });
  }

  listNames() {
    return this.invokeDbus({ member: 'ListNames' });
  }

  listActivatableName() {
    return this.invokeDbus({ member: 'ListActivatableNames' });
  }

  updateActivationEnvironment(env) {
    return this.invokeDbus({ member: 'UpdateActivationEnvironment', signature: 'a{ss}', body: [env] });
  }

  startServiceByName(name, flags) {
    return this.invokeDbus({ member: 'StartServiceByName', signature: 'su', body: [name, flags] });
  }

  getConnectionUnixUser(name) {
    return this.invokeDbus({ member: 'GetConnectionUnixUser', signature: 's', body: [name] });
  }

  getConnectionUnixProcessId(name) {
    return this.invokeDbus({ member: 'GetConnectionUnixProcessID', signature: 's', body: [name] });
  }

  getNameOwner(name) {
    return this.invokeDbus({ member: 'GetNameOwner', signature: 's', body: [name] });
  }

  nameHasOwner(name) {
    return this.invokeDbus({ member: 'GetNameOwner', signature: 's', body: [name] });
  }

  mangle({ path, iface, member }) {
    const obj = {
      path,
      interface: iface,
      member,
    };

    return JSON.stringify(obj);
  }

  sendSignal(path, iface, name, signature, args) {
    this.serial += 1;
    const signalMsg = {
      type: MESSAGE_TYPE.signal,
      serial: this._serial,
      interface: iface,
      path,
      member: name,
    };

    if (signature) {
      signalMsg.signature = signature;
      signalMsg.body = args;
    }

    this.message(signalMsg);
  }

  sendError(msg, errorName, errorText) {
    const reply = {
      type: MESSAGE_TYPE.error,
      replySerial: msg.serial,
      destination: msg.sender,
      errorName,
      signature: 's',
      body: [errorText],
    };

    this.message(reply);
  }

  sendReply(msg, signature, body) {
    const reply = {
      type: MESSAGE_TYPE.methodReturn,
      replySerial: msg.serial,
      destination: msg.sender,
      signature,
      body,
    };
    this.message(reply);
  }

  onMessage(msg) {
    msg = JSON.parse(JSON.stringify(msg));
    msg.iface = msg.interface;

    switch (msg.type) {
    case MESSAGE_TYPE.error:
    case MESSAGE_TYPE.methodReturn:
      this.onMethodReturnOrError(msg);
      break;
    case MESSAGE_TYPE.signal:
      this.onSignal(msg);
      break;
    default:
      this.onMethodCall(msg);
      break;
    }
  }

  onMethodReturnOrError(msg) {
    const handler = this._methodReturnHandlers[msg.replySerial];

    if (handler) {
      delete this._methodReturnHandlers[msg.replySerial];

      const props = {
        bus: this,
        message: msg,
        signature: msg.signature,
      };

      switch (msg.type) {
      case MESSAGE_TYPE.methodReturn:
        handler.resolve(msg.body, props);
        break;
      case MESSAGE_TYPE.error:
        handler.reject(msg.body, props);
        break;
      default:
        handler.reject('Unknown message type');
      }
    }
  }

  onSignal({
    iface, member, path, body, signature,
  }) {
    this.emit(`${iface}.${member}`, path, body, signature);
    this.emit(this.mangle({ iface, path, member }), path, body, signature);
  }

  onMethodCall(msg) {
    // exported interfaces handlers
    const obj = this._exportedObjects[msg.path];

    if (obj) {
      if (stdDbusIfaces(msg, this)) return;
      const iface = obj[msg.interface];
      if (iface) {
        // now we are ready to serve msg.member
        const impl = iface[1];
        const func = impl[msg.member];

        if (!func) {
          // XXX: respond with standard dbus error
          this._log('error', `Method ${msg.member} is not implemented`);
          throw new Error(`Method ${msg.member} is not implemented`);
        }

        try {
          const result = func.apply(impl, msg.body);
          const resultSignature = iface[0].methods[msg.member][1];
          this.sendReply(msg, resultSignature, [result]);
        } catch (error) {
          this._log('error', 'Caught exception while trying to execute handler', error);
          throw error;
        }

        return;
      }
      this._log('error', `Interface ${msg.interface} is not supported`);
      // XXX: respond with standard dbus error
    }

    // setMethodCall handlers

    const handler = this._methodCallHandlers[this.mangle(msg)];

    if (handler) {
      try {
        const result = handler[0].apply(null, msg.body);
        this.sendReply(msg, handler[1], result);
      } catch (error) {
        this._log('error', 'Caught exception while trying to execute handler', error);
        this.sendError(error.message, error.description);
      }
    } else {
      this.sendError(msg, 'org.freedesktop.DBus.Error.UnknownService', 'Uh oh oh');
    }
  }

  setMethodCallHandler(path, iface, member, handler) {
    const key = this.mangle({ path, iface, member });
    this._methodCallHandlers[key] = handler;
  }

  exportInterface(obj, path, iface) {
    if (!this._exportedObjects[path]) {
      this._exportedObjects[path] = {};
    }

    const entry = this._exportedObjects[path];
    entry[iface.name] = [iface, obj];

    // monkey-patch obj.emit()
    if (typeof obj.emit === 'function') {
      const oldEmit = obj.emit;
      obj.emit = (...args) => {
        const signalName = args[0];
        if (!signalName) throw new Error('Trying to emit undefined signal');

        // send signal to bus
        if (iface.signals && iface.signals[signalName]) {
          const signal = iface.signals[signalName];
          this.sendSignal(path, iface.name, signalName, signal[0], args.slice(1));
        }

        // note that local emit is likely to be called
        // before signal arrives to remote subscriber
        oldEmit.apply(obj, args);
      };
    }

    // XXX: emit ObjectManager's InterfaceAdded
  }

  getService(name) {
    return new DBusService(name, this);
  }

  getObject(path, name) {
    return this.getService(path).getObject(name);
  }

  getInterface(path, objName, ifaceName) {
    return this.getService(path).getInterface(objName, ifaceName);
  }

  destroy() {
    if (this.isActive) {
      this.state = state.DESTROYING;

      this.removeAllListeners();

      this._cleanupOnSocketClosed();

      delete this._serial;
      delete this._methodReturnHandlers;
      delete this._exportedObjects;
      delete this._messages;
      delete this._logger;

      this.state = state.DESTROYED;
    }
  }
}

export default DBus;
