/*
eslint no-underscore-dangle: ["error", {
  "allow": ["_path", "_id", "_iface"],
  "allowAfterThis": true,
  "allowAfterSuper": true,
}]
*/

class Fetcher {
  constructor({
    iface,
    schema,
  }) {
    this._interface = iface;
    this._schema = schema;
  }

  async fetch(service, path, {
    includePath = false,
    includeIface = false,
  } = {}) {
    const keys = Object.keys(this._schema);
    // const toFetch = keys
    //   .map(key => this._schema[key].key || key)
    //   .reduce((total, el) => {
    //     if (!total.includes(el)) total.push(el);
    //     return total;
    //   }, []);

    const result = {};
    if (includePath) {
      result._path = path;
    }

    const iface = await service.getInterface(path, this._interface);
    if (!iface) {
      throw new Error(`Could not find interface ${this._interface} for object: ${path}`);
    }

    if (includeIface) {
      result._iface = iface;
    }

    const fetchers = {};

    const promises = keys.map(async (key) => {
      const dbusKey = this._schema[key].key || (key.charAt(0).toUpperCase() + key.slice(1));
      if (!fetchers.hasOwnProperty(dbusKey)) {
        fetchers[dbusKey] = iface[dbusKey];
      }

      try {
        const value = await fetchers[dbusKey];
        result[key] = await this._parseValue(service, value, this._schema[key], { includePath, includeIface });
      } catch (error) {
        if (this._schema[key].optional) {
          result[key] = undefined;
        } else if (this._schema[key].hasOwnProperty('defaultValue')) {
          result[key] = JSON.parse(JSON.stringify(this._schema[key].defaultValue));
        } else {
          throw error;
        }
      }
    });

    await Promise.all(promises);
    return result;
  }

  _parseValue(service, value, definition, options) {
    if (definition.type instanceof Fetcher) {
      return definition.type.fetch(service, value, options);
    }

    switch (definition.type) {
    case Object:
      if (definition.raw) {
        if (definition.element) {
          return value[definition.element];
        }
        return value;
      }
      if (definition.element) {
        return this._parsePropertyListItem(value.find(el => el[0] === definition.element)[1].data);
      }
      return this._parsePropertyList(value);

    case Array:
      if (definition.elementType) {
        return Promise.all(value.map(val => this._parseValue(service, val, {
          type: definition.elementType,
        }, options)));
      }
      return value;

    case Number:
      if (definition.decimal) {
        return Number.parseInt(value, 10);
      }
      return Number.parseFloat(value);

    case String:
      return value.toString();
    case 'StringArray':
      return value.reduce((acc, el) => {
        acc[el[0]] = el[1];
        return acc;
      }, {});
    default:
      return value;
    }
  }

  _parsePropertyList(list) {
    return list.reduce((acc, prop) => {
      acc[prop[0]] = this._parsePropertyListItem(prop[1].data);
      return acc;
    }, {});
  }

  _parsePropertyListItem(item = []) {
    return Buffer.from(item).toString().replace('\u0000', '');
  }
}

export default Fetcher;
