import { EventEmitter } from 'events';

class DBusInterface extends EventEmitter {
  constructor(name, obj) {
    super();
    this.name = name;
    this.obj = obj;
    this.bus = obj.service.bus;
    this.callbacks = [];
    this.sigHandlers = [];

    // In case it's the first listener for a event, start subscribing.
    this.on('newListener', async (sigName, callback) => {
      const signalFullName = this.bus.mangle({ path: this.obj.name, iface: this.name, member: sigName });

      if (!this.obj.service.bus.listenerCount(signalFullName)) {
        // This is the first time, so call addMatch
        const match = this.getMatchRule(sigName);
        await this.bus.addMatch(match);
        this.bus.on(signalFullName, this.getSigHandler(callback));
      } else {
        // The match is already there, just add event listener
        this.bus.on(signalFullName, this.getSigHandler(callback));
      }
    });
  }

  getMatchRule(signame) {
    return `type='signal',path='${this.obj.name}',interface='${this.name}',member='${signame}'`;
  }

  getSigHandler(callback) {
    let handler = this.callbacks.find(c => c === callback);
    if (!handler) {
      const index = this.callbacks.push(callback) - 1;
      handler = function sigHandler(messageBody) {
        callback(...messageBody);
      };
      this.sigHandlers[index] = handler;
    }
    return handler;
  }

  async removeListener(signame, callback) {
    super.removeListener(signame, callback);

    const signalFullName = this.bus.mangle({ path: this.obj.name, iface: this.name, member: signame });

    this.bus.removeListener(signalFullName, this.getSigHandler(callback));

    if (!this.bus.listenerCount(signalFullName)) {
      // There is no event handlers for this match
      const match = this.getMatchRule(this.obj.name, this.name, signame);
      await this.bus.removeMatch(match);
      this.callbacks.length = 0;
      this.sigHandlers.length = 0;
    }
  }

  get(property) {
    return this.bus.invoke({
      destination: this.obj.service.name,
      path: this.obj.name,
      interface: 'org.freedesktop.DBus.Properties',
      member: 'Get',
      signature: 'ss',
      body: [this.name, property],
    });
  }

  set(property, value) {
    return this.bus.invoke({
      destination: this.obj.service.name,
      path: this.obj.name,
      interface: 'org.freedesktop.DBus.Properties',
      member: 'Set',
      signature: 'ssv',
      body: [this.name, property, value],
    });
  }

  getAll() {
    return this.bus.invoke({
      destination: this.obj.service.name,
      path: this.obj.name,
      interface: 'org.freedesktop.DBus.Properties',
      member: 'Get',
      signature: 's',
      body: [this.name],
    });
  }

  destroy() {
    this.removeAllListeners();

    delete this.name;
    delete this.obj;
    delete this.bus;
    delete this.callbacks;
    delete this.sigHandlers;
  }
}

export default DBusInterface;
