import * as os from 'os';
import { debounce, isEqual } from 'lodash';

import MeteorService from './singleDeviceMeteorService';
import * as network from '../utils/network';
import { readCurrentVersion } from '../utils/version';
import { getHardwareType } from '../utils/hardware';
import serviceRegistry from '../services/serviceRegistry';

import MeteorCacheManager from '../services/ddp/meteorCacheManager';
import { deviceStatus } from '../services/ddp/constants';
import bridges from './bridges';

const additionalServicesToWatch = ['versionMonitor'];
const getAdditionalServicesToWatch = () => additionalServicesToWatch.map(serviceName => serviceRegistry.get(serviceName));

class MeteorDeviceAgent extends MeteorService {
  constructor(options) {
    super(options);

    this._setDeviceStatus = debounce(this._setDeviceStatus.bind(this), 3000);
    this.bridges = bridges.map(Bridge => new Bridge({
      meteorClient: this._meteorClient,
      device: this._device,
    }));

    this.activeBridges = [];

    this._serviceHealthMonitors = {
      stateChanged: () => this._setCurrentDeviceStatus(),
      healthStateChanged: () => this._setCurrentDeviceStatus(),
    };

    this.cache = new MeteorCacheManager(this, ['compositions']);
    this._subscriptions = [];

    // set default service log metadata
    this._setServiceLogMetadata(this._device);
  }

  set device(device) {
    super.device = device;
    this.bridges.forEach((bridge) => {
      bridge.device = device;
    });
  }

  get meteorClientListeners() {
    if (!this._meteorListeners) {
      this._meteorListeners = super.meteorClientListeners;
      this._meteorListeners.set('disconnected', () => this._onDisconnected());
    }

    return this._meteorListeners;
  }

  async _trigger(entity, action, record, oldFields, clearedFields, newFields) {
    if (this.isStarted) {
      this._startRequiredServices();
    }

    return super._trigger(entity, action, record, oldFields, clearedFields, newFields);
  }

  _onThisDeviceChanged(device, oldFields, clearedFields, newFields) {
    if (this.isStarted) {
      this._startRequiredServices();
    }

    if (newFields.hasOwnProperty('proxy')) {
      this._log('info', 'Received new proxy configuration. Updating configuration', { serviceLog: true });
      const proxy = {
        url: this._device.proxy.url,
        user: this._device.proxy.user,
        password: this._device.proxy.password,
      };
      serviceRegistry.get('configManager').update('device.proxy', proxy);
    }

    if (newFields.hasOwnProperty('publicIp')) {
      serviceRegistry.get('configManager').update('device.publicIp', this._device.publicIp);
    }

    if (newFields.hasOwnProperty('meetingId')) {
      this._setServiceLogMetadata(this._device);
      this._log('verbose', `Active meeting has changed to ${this._device.meetingId}`, { serviceLog: true });
    }
  }

  async _onStart() {
    const message = this._showMessage({ text: 'synchronizing data' });

    try {
      this._log('verbose', 'Starting');
      await super._onStart();
      await this._updateDeviceData();
      await this._subscribe();
      this._startRequiredServices();
      this._startAdditionalServiceMonitors();
    } finally {
      await message.hide();
      this._messenger.stopProgress();
    }
  }

  async _onStop() {
    this._stopAdditionalServiceMonitors();
    await Promise.all([
      super._onStop(),
      this._stopServices(),
    ]);
    await this._unsubscribe();
  }

  _startRequiredServices() {
    const promises = this.bridges.map((bridge) => {
      if (bridge.isServiceRequired) {
        return this._activateService(bridge);
      }
      return this._deactivateService(bridge);
    });

    return Promise.all(promises);
  }

  _activateService(bridge) {
    if (!this.activeBridges.includes(bridge)) {
      this._log('info', `Activate bridge: ${bridge.serviceName}`);
      this.activeBridges.push(bridge);
      this._startMonitorsForService(bridge);
      if (!bridge.wantsToBeActive) {
        return bridge.start();
      }
    }
    return Promise.resolve();
  }

  _startAdditionalServiceMonitors() {
    this._log('verbose', `Watching additional services: ${additionalServicesToWatch.join(', ')}`);
    getAdditionalServicesToWatch().forEach(service => this._startMonitorsForService(service));
  }

  _startMonitorsForService(service) {
    Object.keys(this._serviceHealthMonitors).forEach((event) => {
      service.on(event, this._serviceHealthMonitors[event]);
    });
  }

  _stopServices() {
    const promises = this.activeBridges.map(bridge => this._deactivateService(bridge));
    return Promise.all(promises);
  }

  async _deactivateService(bridge) {
    if (this.activeBridges.includes(bridge)) {
      this._log('info', `Deactivate bridge: ${bridge.serviceName}`);
      this.activeBridges = this.activeBridges.filter(b => b !== bridge);
      if (bridge.wantsToBeActive) {
        await bridge.stop();
      }
      this._stopMonitorsForService(bridge);
    }
  }

  _stopAdditionalServiceMonitors() {
    getAdditionalServicesToWatch().forEach(service => this._stopMonitorsForService(service));
  }

  _stopMonitorsForService(service) {
    Object.keys(this._serviceHealthMonitors).forEach((event) => {
      service.off(event, this._serviceHealthMonitors[event]);
    });
  }

  async _subscribe() {
    try {
      this._log('info', 'Subscribing to "barco.device.data"');
      this._subscriptions.push(await this._meteorClient.subscribe('barco.device.data', [{
        deviceId: this._device._id,
      }], this));
    } catch (error) {
      this._log('error', 'Error while subscribing to device data', error);
      throw error;
    }

    this._log('info', 'Subscriptions are ready');
    this.emit('deviceDataSubscriptionReady');
  }

  async _unsubscribe() {
    if (this._subscriptions.length) {
      try {
        this._log('info', 'Unsubscribing');
        this._subscriptions.forEach(id => this._meteorClient.unsubscribe(id));
      } catch (error) {
        this._log('error', 'Error while unsubscribing to device data', error);
        // throw error; // no need for an unhandled exception if socket was already closed
      } finally {
        this._subscriptions = [];
      }
    }

    this._log('info', 'Unsubscribed to "barco.device.data"');
    this.emit('deviceDataSubscriptionReady');
  }

  async _updateDeviceData() {
    // after upgrade is completed and reboot, we need to reset upgrading status
    await this._meteorClient.setDeviceUpgradeStatus({ status: false });

    const ip = network.getIpAddress(this._logger);
    const mac = network.getMacAddress(this._logger);
    const hostname = os.hostname();
    const newDeviceData = Object.create(null);

    if (ip && ip !== this._device.privateIp) {
      newDeviceData.privateIp = ip;
    }

    if (mac && mac !== this._device.mac) {
      newDeviceData.mac = mac;
    }

    if (hostname
      && hostname !== this._device.hostname
      && hostname.toUpperCase().includes(mac)) {
      newDeviceData.hostname = hostname;
    }

    const version = readCurrentVersion(this._logger);
    if (version && version !== this._device.version) {
      newDeviceData.version = version;
    }
    this._log('verbose', `Found version: ${version}`);

    const hardwareType = await getHardwareType();
    if (hardwareType && hardwareType !== this._device.hardwareType) {
      newDeviceData.hardwareType = hardwareType;
    }
    this._log('verbose', `Found hardware type: ${hardwareType}`);

    if (Object.keys(newDeviceData).length) {
      newDeviceData._id = this._device._id;
      await this._meteorClient.updateDevice(newDeviceData);
    }
  }

  _setCurrentDeviceStatus() {
    const faultyServices = [
      ...this.activeBridges,
      ...getAdditionalServicesToWatch(),
    ].filter(service => !service.isStarted || service.isFaulty);

    if (faultyServices.length === 0) {
      this._log('verbose', 'All services gone healthy & started');
      this._setDeviceStatus(deviceStatus.ONLINE);
    } else {
      const details = faultyServices.map(b => `${b.serviceName}: ${b.isFaulty ? b.fault : 'Not fully started'}`);
      this._log('verbose', details.join(', \n'));
      this._setDeviceStatus(deviceStatus.WARNING, details);
    }
  }

  _setDeviceStatus(status, statusDetails = []) {
    // This method is debounced and could be called after stop/destroy
    if (!this.isActive) return;

    if (this._device.status !== status || !isEqual(this._device.statusDetails, statusDetails)) {
      this._device.status = status;
      this._device.statusDetails = statusDetails;
      return this._meteorClient.setDeviceStatus({
        devices: [this._device._id],
        status,
        statusDetails,
      });
    }
  }

  _setServiceLogMetadata(device) {
    const loggerService = serviceRegistry.get('logger');

    if (loggerService) {
      loggerService.serviceMetadata = {
        meetingId: device.meetingId,
        instituteId: device.instituteId,
        role: device.role,
      };
    }
  }

  async _onDisconnected() {
    this._log('info', 'Disconnected, activating fallbacks on child bridges...');

    this._stopAdditionalServiceMonitors();
    await Promise.all(this.activeBridges.map((bridge) => {
      this._stopMonitorsForService(bridge);
      return bridge.activateFallback();
    }));

    this._log('info', 'Done activating fallbacks');
  }

  async _onDestroy() {
    try {
      await Promise.all(this.bridges.map(bridge => bridge.destroy()));
    } catch (error) {
      this._log('error', 'Error while destroying child bridge', error);
    }

    this.bridges = [];

    await super._onDestroy();
  }
}

export default MeteorDeviceAgent;
