/*
eslint no-underscore-dangle: ["error", {
  "allow": ["_path", "_id", "_device", "_sevice", "_iface"],
  "allowAfterThis": true,
  "allowAfterSuper": true,
}] no-await-in-loop: "off"
*/

import * as Rx from 'rxjs';
import { debounceTime } from 'rxjs/operators';
import * as path from 'path';
import { keyBy, isEqual } from 'lodash';
import * as isDev from 'electron-is-dev';
import { exec } from 'child_process';
import { serializeError } from 'serialize-error';

import SynchronousBridge from '../synchronousBridge';
import AudioService from './service';
import MeteorCacheManager from './meteorCacheManager';
import PaCacheManager from './paCacheManager';
import AudioStream from '../streaming/audioStream';
import { sleep } from '../../../utils/utils';
import { AUDIO_SAMPLES_DIR } from '../../../files';
import { createBrowserWindow, createBrowserSession } from '../../../utils/browser/window';
import {
  initializeAudioConfig,
  calculateAudioConfig,
  getCurrentRoutingModules,
} from '../utils/audio/calculatePAConfig';

import {
  audioSinkRole,
  audioModuleRole,
  audioSourceRole,
  audioDefaultParameters,
  compositionType,
  sinkType,
  commands,
  audioReferenceSampleType,
  audioCalibrationState,
} from '../../../services/ddp/constants';
import serviceRegistry from '../../../services/serviceRegistry';
import { state as serviceState } from '../../../services/service';
import {
  cleanPort,
  cleanProfile,
  isTheSameRouting,
  getObjectDiff,
} from './utils';

/**
 * The Audio Bridge is the glue code between blue and the pulse audio service.
 *
 */
class AudioBridge extends SynchronousBridge {
  constructor(options) {
    super(options);
    this._audioService = new AudioService();

    this.cache = new MeteorCacheManager(this, ['devices', 'sources', 'sinks', 'compositions', 'layouts']);
    this._paCache = new PaCacheManager(this._audioService);

    this._dynamicSinkAdded = false;

    const schedule = method => (...args) => this.scheduler.schedule(method, ...args);

    this._listeners = {
      service: {
        onPlaybackStreamAdded: schedule('_onPlaybackStreamAdded'),
        onPlaybackStreamRemoved: schedule('_onPlaybackStreamRemoved'),
        sourcesChanged: schedule('_updateInputs'),
        sinksChanged: schedule('_updateOutputs'),
        cardsChanged: schedule('_updateCards'),
      },
      own: {
        updateMutedSources: schedule('_setCorrectVolumeAndMuteAndRoutingForMeteorSources'),
      },
    };

    this._observables = {};
    this._sourcesOnlyOnMutedCompositions = [];
    this._sourcesVisible = [];
    this._subscribed = false;
    this._audioReferenceSampleStreams = [];
    this._syncAVScraper = false;
    this._pacmd_set_profile_command = 'sudo -s -u barco pacmd set-card-profile';
    if (process.env.DEV_WITH_WINDOW_MANAGER) {
      this._pacmd_set_profile_command = 'pacmd set-card-profile';
    }
    // The audioConfig object contains all pulse audio configuration related settings
    this._audioConfig = {};
    initializeAudioConfig(this._audioConfig, this._device);
  }

  get supportedRPCCommands() {
    return [
      commands.PLAY_REFERENCE_SAMPLE,
      commands.STOP_ALL_REFERENCE_SAMPLES,
    ];
  }

  async _onRpcAdded(rpc) {
    this._log('verbose', `Handling rpc ${JSON.stringify(rpc, null, ' ')}`, { serviceLog: true });
    switch (rpc.command) {
    case commands.PLAY_REFERENCE_SAMPLE:
      this._playReferenceSample(rpc.params.sampleType);
      break;
    case commands.STOP_ALL_REFERENCE_SAMPLES:
      this._stopAllReferenceSamples();
      break;
    default:
    }
  }

  get supportedCompositionTypes() {
    return [compositionType.CINERGY];
  }

  get supportedSinkTypes() {
    return [sinkType.RTSP_STREAM];
  }

  /**
   * Disable meteor callbacks while syncing
   */
  get applyActions() {
    this._isDirty = true;
    return super.applyActions && this.isStarted;
  }

  _onThisDeviceChanged(device, oldFields, clearedFields, newFields) {
    const promises = [];
    let triggerSyncOutputConfig = false;
    let includeSyncAVScraper = false;

    // If audio calibration mode has been changed, apply pulse config accordingly
    if (newFields.hasOwnProperty('audio')) {
      if (this._device.audio.hasOwnProperty('calibration')) {
        if (this._device.audio.calibration !== oldFields.audio?.calibration) {
          this._log('verbose', `Audio calibration status changed to ${this._device.audio.calibration}. Applying new settings.`);
          triggerSyncOutputConfig = true;
          includeSyncAVScraper = true;
        }
      }
    }

    // In _requestSyncOutputConfig function we set default source. Therefore we check on both properties
    if (newFields.hasOwnProperty('audioSinkConfig')
    || newFields.hasOwnProperty('audioSourceConfig')
    || newFields.hasOwnProperty('usbCameraConfig') // audio input can be coupled to camera
    || newFields.hasOwnProperty('role') // some parts of audio routing depend on device role
    || (newFields.hasOwnProperty('streamRemoteAudio'))) {
      this._log('verbose', 'Audio configuration changed, about to synchronize output configuration');
      triggerSyncOutputConfig = true;
    }

    // If the active profile has changed for a certain audio card, apply it.
    if (newFields.hasOwnProperty('audioCards')) {
      let cardIndexArray = getObjectDiff(oldFields.audioCards, newFields.audioCards);
      if (!Array.isArray(cardIndexArray)) {
        cardIndexArray = [...cardIndexArray];
      }
      if (cardIndexArray.length) {
        this._log('verbose', 'Active profile of at least one audio card has been changed. Triggering sync active profiles.');
        return this._syncActiveProfilesOnCards();
      }
    }

    if (triggerSyncOutputConfig) {
      // Everything below not needed because this one includes it all
      return this._requestSyncOutputConfig(includeSyncAVScraper);
    }

    // If master audio volume/mute of the device has changed, update all the hardware sinks
    // This happens when changing the volume slider in the teacher UI
    if (newFields.hasOwnProperty('audio')) {
      if ((this._device.audio.volume !== oldFields.audio?.volume)
        || (this._device.audio.mute !== oldFields.audio?.mute)) {
        this._log('verbose', 'Volume/mute settings of device changed. Applying new settings.');
        promises.push(this._applyVolumeAndMuteSettingsToHardwareSinks());
      }
    }

    // If audio content sharing default volume has changed, apply the new volume
    // This happens when changing the slider in the calibration dialog
    if (newFields.hasOwnProperty('contentDefaultVolume')) {
      if (this._device.contentDefaultVolume !== oldFields.contentDefaultVolume) {
        this._log('verbose', `Content sharing default volume has been changed from ${oldFields.contentDefaultVolume}% to ${newFields.contentDefaultVolume}%. Applying new default volume.`);
        promises.push(this._applyContentSharingDefaultVolume());
      }
    }

    // If the default volume of a source has changed, apply it immediately
    // This happens when changing the default volume of an audio input in the administrator UI (devices page)
    if (newFields.hasOwnProperty('audioSources')) {
      let sourceIndexArray = getObjectDiff(oldFields.audioSources, newFields.audioSources);
      if (!Array.isArray(sourceIndexArray)) {
        sourceIndexArray = [...sourceIndexArray];
      }
      sourceIndexArray.forEach((sourceIndex) => {
        this._log('verbose', `Default volume of source with index ${sourceIndex} has been changed`);
        if (sourceIndex < Object.keys(newFields.audioSources).length) {
          promises.push(this._applyVolumeAndMuteSettingsToSource(this._paCache.findSourceById(newFields.audioSources[sourceIndex].id)));
        }
      });
    }

    // If the default volume of a sink has changed, apply it immediately
    // This happens when changing the default volume of an audio output in the administrator UI (devices page)
    if (newFields.hasOwnProperty('audioSinks')) {
      let sinkIndexArray = getObjectDiff(oldFields.audioSinks, newFields.audioSinks);
      if (!Array.isArray(sinkIndexArray)) {
        sinkIndexArray = [...sinkIndexArray];
      }
      sinkIndexArray.forEach((sinkIndex) => {
        this._log('verbose', `Default volume of sink with index ${sinkIndex} has been changed`);
        if (sinkIndex < Object.keys(newFields.audioSinks).length) {
          promises.push(this._applyVolumeAndMuteSettingsToSink(this._paCache.findSinkById(newFields.audioSinks[sinkIndex].id)));
        }
      });
    }

    // Pulse audio's udev detect module doesn't seem to always detect audio from HDMI/DP displays.
    // Therefore we force a rescan when there's a change in the HDMI/DP displays.
    if (newFields.hasOwnProperty('displaysLatest')) {
      const displayFilter = display => display.port.startsWith('HDMI-') || display.port.startsWith('DP-');
      // We need to sort them in case they are listed differently, but there's no change in actual displays
      const getSerials = (displays = []) => displays.filter(displayFilter).map(display => display.serial || display.id).sort();
      const previousDisplaySerials = getSerials(this._device.displaysPrevious);
      const latestDisplaySerials = getSerials(this._device.displaysLatest);
      // If for some reason we don't have a previous set, then consider it a change to be on the safe side.
      if (!this._device.displaysPrevious || !isEqual(previousDisplaySerials, latestDisplaySerials)) {
        this._log('verbose', 'displaysLatest potentially changed w.r.t. HDMI/DP, forcing rescan of hardware audio inputs/outputs');
        promises.push(this._forcePAHardwareDetection());
      }
    }

    return Promise.all(promises);
  }

  _onSinkAdded(sink) {
    if (sink.muteWhenShown && sink.shownOn.length > 0) {
      this._log('verbose', 'Muted sink added');
      this._applyVolumeAndMuteSettingsToHardwareSinks();
    }
  }

  _onSinkChanged(sink, oldFields, clearedFields, newFields) {
    if (newFields.hasOwnProperty('muteWhenShown') || newFields.hasOwnProperty('shownOn')) {
      this._log('verbose', 'Sink mute state changed, re-applying settings');
      this._applyVolumeAndMuteSettingsToHardwareSinks();
    }
  }

  _onSinkRemoved(oldSink) {
    if (oldSink.muteWhenShown && oldSink.shownOn.length > 0) {
      this._log('verbose', 'Muted sink removed, re-applying settings');
      this._applyVolumeAndMuteSettingsToHardwareSinks();
    }
  }

  get muted() {
    this._log('verbose', 'Checking audio mute state of device:');
    if (this._device.audio.mute) {
      this._log('verbose', 'Global device mute is active');
    } else {
      this._log('verbose', 'Device not globally muted');
    }
    const sinkWithMuteWhenShownOnScreen = this.cache.$arrays.sinks
      .filter(sink => sink.muteWhenShown && sink.shownOn.length > 0).length > 0;
    if (sinkWithMuteWhenShownOnScreen) {
      this._log('verbose', 'Muting device because there is a sink shown that requires muting!');
    } else {
      this._log('verbose', 'Currently no sink shown that requires muting');
    }
    return this._forceMute || this._device.audio.mute || sinkWithMuteWhenShownOnScreen;
  }

  _applyVolumeAndMuteSettingsToHardwareSources() {
    const sources = this._paCache.getHardwareSources();
    const promises = sources.map(source => this._applyVolumeAndMuteSettingsToSource(source));
    return Promise.all(promises);
  }

  async _applyVolumeAndMuteSettingsToSource(source) {
    let unscaledVolume = 0;
    let volume = 0;
    let mute = true;

    // Find the default volume of the source, if not defined take 100
    const meteorSource = this.cache.findPaSource(source);
    const defaultVolume = meteorSource ? meteorSource.defaultVolume : 100;
    this._log('verbose', `Default volume for source ${source.id} = ${defaultVolume}%`);
    // Find the role of the source and apply the right volume/mute settings
    const sourceConfig = this.cache.findPaSourceConfig(source);
    const sourceDefaultVolume = Number.isNaN(defaultVolume) ? 100 : defaultVolume;

    // As soon as a source is set for a usb camera we don't mute it as it can be needed for an RTSP or webRTC stream
    const cameraThatUsesSource = this._device.usbCameraConfig.find(cam => (cam.audioDevice === source.id));
    if (cameraThatUsesSource) {
      this._log('verbose', `Source ${source.id} is configured to be streamed with camera '${cameraThatUsesSource.device}' and must not be muted`);
    } else {
      this._log('verbose', `Source ${source.id} is not configured to be streamed with any camera`);
    }
    // Check if the input role is set to 'idle'
    const sourceRoleIsIdle = sourceConfig && (sourceConfig.role === audioSourceRole.IDLE);
    this._log('verbose', `Source ${source.id} is ${sourceRoleIsIdle ? '' : 'not '}configured to role 'Not Used'`);
    // Only mute if source is not used for anything
    const muteSource = sourceRoleIsIdle && !cameraThatUsesSource;
    if (muteSource) {
      this._log('verbose', `Setting volume for idle source ${source.id} to 0%, muted`);
      unscaledVolume = 0;
      volume = 0;
      mute = true;
    } else {
      unscaledVolume = sourceDefaultVolume;
      volume = this._calcAudioVolumeValue(unscaledVolume);
      mute = false;
      this._log('verbose', `Setting volume for source ${source.id} to volume = ${unscaledVolume}%, mute = ${mute}`);
    }
    return this._setSourceProperties(source, { volume, mute });
  }

  _applyVolumeAndMuteSettingsToHardwareSinks() {
    const sinks = this._paCache.getHardwareSinks();
    const promises = sinks.map(sink => this._applyVolumeAndMuteSettingsToSink(sink));
    return Promise.all(promises);
  }

  async _applyVolumeAndMuteSettingsToSink(sink) {
    // First get the master volume and master mute as set in the teacher UI
    const masterVolume = Number.isNaN(this._device.audio.volume) ? 100 : this._device.audio.volume;
    const masterMute = this.muted;

    let unscaledVolume = 0;
    let volume = 0;
    let mute = true;

    // Find the default volume of the sink, if not defined take 100
    const meteorSink = this.cache.findPaSink(sink);
    const defaultVolume = meteorSink ? meteorSink.defaultVolume : 100;
    this._log('verbose', `Default volume for sink ${sink.id} = ${defaultVolume}%`);
    const sinkDefaultVolume = Number.isNaN(defaultVolume) ? 100 : defaultVolume;
    // Find the role of the sink and apply the right volume/mute settings
    const sinkConfig = this.cache.findPaSinkConfig(sink);
    if (sinkConfig) {
      switch (sinkConfig.role) {
      // AEC reference output should always output default volume unmuted irrespective of volume slider and mute in UI
      case audioSinkRole.AEC_REF_MIX:
        this._log('verbose', `Setting volume for AEC Reference Mix sink ${sink.id} to ${sinkDefaultVolume}%, unmuted`);
        unscaledVolume = sinkDefaultVolume;
        volume = this._calcAudioVolumeValue(unscaledVolume);
        mute = false;
        break;
        // Don't touch sinks that are set to role idle
      case audioSinkRole.IDLE:
        this._log('verbose', `Skipping set volume for sink ${sink.id} as it has role 'idle'`);
        return;
      default:
        // Apply the master volume on top of the default volume
        unscaledVolume = Math.trunc((sinkDefaultVolume * masterVolume) / 100);
        volume = this._calcAudioVolumeValue(unscaledVolume);
        mute = masterMute;
      }
      this._log('verbose',
        `Setting volume for sink ${sink.id} with role ${sinkConfig.role}`,
        `to ${unscaledVolume}% (default/master volume = ${sinkDefaultVolume}/${masterVolume}%), mute = ${mute}`);
    } else {
      this._log('verbose', `Not setting volume or mute for sink ${sink.id} because it has no sink config...`);
      return;
    }
    return this._setSinkProperties(sink, { volume, mute });
  }

  async _applyContentSharingDefaultVolume() {
    const contentSharingDefaultVolume = (this._device.hasOwnProperty('contentDefaultVolume')
    && !Number.isNaN(this._device.contentDefaultVolume)) ? this._device.contentDefaultVolume : 100;
    const sink = (this._paCache.findSinkById(this._audioConfig.contentSharingSink));
    if (sink) {
      this._log('verbose', `Applying content sharing default volume ${contentSharingDefaultVolume}% to sink ${sink.id}, unmuted`);
      const volume = this._calcAudioVolumeValue(contentSharingDefaultVolume);
      const mute = false;
      return this._setSinkProperties(sink, { volume, mute });
    }
  }

  _calcAudioVolumeValue(unscaledVolume) {
    return Math.round((unscaledVolume / 100) * audioDefaultParameters.maxVolume);
  }

  _onCompositionAdded() {
    this.emit('updateMutedSources');
  }

  _onCompositionChanged(composition, oldFields, clearedFields, newFields) {
    if (newFields.hasOwnProperty('activeLayoutId') || newFields.hasOwnProperty('muted')) {
      this.emit('updateMutedSources');
    }
  }

  _onCompositionRemoved() {
    this.emit('updateMutedSources');
  }

  _onLayoutAdded() {
    this.emit('updateMutedSources');
  }

  _onLayoutChanged() {
    this.emit('updateMutedSources');
  }

  _onLayoutRemoved() {
    this.emit('updateMutedSources');
  }

  async _onPlaybackStreamAdded([args]) {
    this._log('verbose', `Playback stream added: arguments = ${JSON.stringify(args, null, 2)}`);
    if (!args || (args.length < 2) || (args[1].length < 1)) { return; }
    const [pbsPath] = args[1];
    if (!pbsPath) { return; }
    try {
      const stream = await this._audioService.getPlaybackStream(pbsPath, {
        includePath: true,
        includeIface: true,
      });
      if (this._paCache.handlePlaybackStream(stream)) {
        this._paCache.addPlaybackStreams([stream]);
        await this._syncPaStream(stream);
      }
    } catch (error) {
      this._log('error', `Error while fetching audio stream on path: ${pbsPath} with error: ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`, {
        serviceLog: true,
      });
    }
    this._log('info', '[onPlaybackStreamAdded] Requesting to sync playback streams');
    this._requestSyncPlaybackStreams(); // Pickup playback streams not flagged by DBUS interface
  }

  async _onPlaybackStreamRemoved([args]) {
    this._log('verbose', `Playback stream removed: arguments = ${JSON.stringify(args, null, 2)}`);
    if (!args || (args.length < 2) || (args[1].length < 1)) { return; }
    const [pbsPath] = args[1];
    if (!pbsPath) { return; }
    const stream = this._paCache.findPlaybackStreamByPath(pbsPath);
    if (stream) {
      this._paCache.removePlaybackStreams([stream]);
    }
  }

  _requestSyncPlaybackStreams(timeout = 2000) {
    this._log('info', `Received request to sync playback streams in ${timeout} ms`);
    if (this._syncPlaybackStreamsTimer) {
      clearTimeout(this._syncPlaybackStreamsTimer);
    }
    this._syncPlaybackStreamsTimer = setTimeout(this._syncPlaybackStreams.bind(this), timeout);
  }

  async _syncPlaybackStreams() {
    this._log('verbose', 'SyncPlaybackStreams: Refetching pulse audio cache for playback streams');
    await this._paCache.updatePlaybackStreams(true);
    if (!this.isActive) {
      return;
    }
    this._log('verbose', 'SyncPlaybackStreams: About to put playback streams on correct sinks');
    await Promise.all(this._paCache.$arrays.playbackStreams.map(pbs => this._putStreamOnCorrectSink(pbs)));
  }

  _onSourceAdded(source) {
    return this._syncMeteorSource(source);
  }

  _onSourceChanged(source, oldFields, clearedFields, newFields) {
    const promises = [];
    if (newFields.hasOwnProperty('audio')) {
      promises.push(this._syncMeteorSource(source));
    }
    return Promise.all(promises);
  }

  _onSourceRemoved(source) {
    const stream = this._paCache.findMeteorSource(source);
    if (stream) {
      return this._setStreamProperties(stream, {
        volume: 0,
        mute: true,
      });
    }
  }

  async _printVolumeSettingsOfHardwarePaSources() {
    this._paCache.getHardwareSources()
      .forEach(async (source) => {
        this._log('verbose', `Reading out volume steps for ${source._path}`);
        const baseVolume = await source._iface.BaseVolume;
        this._log('verbose', `Base volume for source ${source._path} = ${baseVolume}`);
        const volumeSteps = await source._iface.VolumeSteps;
        this._log('verbose', `Volume steps for source ${source._path} = ${volumeSteps}`);
        const volume = await source._iface.Volume;
        this._log('verbose', `Volume for source ${source._path} = ${volume}`);
      });
  }

  // async _toggleMuteOnHardwarePaSources(mute) {
  //   const promises = this._paCache.getHardwareSources()
  //     .map(async (source) => {
  //       this._log('verbose', `Setting volume for source '${source.id}' to 100%, mute = ${mute}`);
  //       return this._setSourceProperties(source, { mute, volume: audioDefaultParameters.maxVolume });
  //     });
  //   await Promise.all(promises);
  // }

  async _muteAllHardwarePaSinks() {
    const promises = this._paCache.getHardwareSinks()
      .map(async (sink) => {
        this._log('verbose', `Muting sink '${sink.id}'`);
        return sink._iface.setMute(true);
      });
    await Promise.all(promises);
  }

  _setCorrectVolumeAndMuteAndRoutingForMeteorSources() {
    this._log('verbose', 'About to apply correct mute/volume state and routing for playback streams of sources known in blue');
    this._updateSourcesVisible();
    const promises = this.cache.$arrays.sources.map(source => this._syncMeteorSource(source));
    return Promise.all(promises);
  }

  _updateSourcesVisible() {
    this._log('info', 'Calculating list of sources that are visible on compositions');

    const sourcesOnMutedCompositions = this.cache.$arrays.compositions
      .filter(composition => composition.muted && composition.activeLayoutId)
      .map(composition => this.cache.layouts[composition.activeLayoutId])
      .filter(layout => !!layout)
      .reduce((acc, layout) => [...acc, ...layout.items.map(item => item._id)], []);

    const sourcesOnUnmutedCompositions = this.cache.$arrays.compositions
      .filter(composition => !composition.muted && composition.activeLayoutId)
      .map(composition => this.cache.layouts[composition.activeLayoutId])
      .filter(layout => !!layout)
      .reduce((acc, layout) => [...acc, ...layout.items.map(item => item._id)], []);

    this._sourcesVisible = Array.from(new Set([
      ...sourcesOnMutedCompositions,
      ...sourcesOnUnmutedCompositions,
    ]));

    this._sourcesOnlyOnMutedCompositions = Array.from(new Set(sourcesOnMutedCompositions
      .filter(sourceId => !sourcesOnUnmutedCompositions.includes(sourceId))));

    this._log('info', `Sources that are visible: ${this._sourcesVisible.join(',')}`);
    this._log('info', `Sources only appearing on muted compositions: ${this._sourcesOnlyOnMutedCompositions.join(',')}`);
  }

  _shouldSourceBeMuted(source) {
    return (
      source.audio.muted
      || this._sourcesOnlyOnMutedCompositions.includes(source._id)
      || (!this._sourcesVisible.includes(source._id) && source.audio.autoMute)
    );
  }

  async _setDefaultSource(sourcePath) {
    this._log('verbose', `Setting default source to path '${sourcePath}'`);
    await this._audioService.setFallbackSource(sourcePath);
  }

  async _setDefaultSink(sinkPath) {
    this._log('verbose', `Setting default sink to path '${sinkPath}'`);
    await this._audioService.setFallbackSink(sinkPath);
  }

  async _syncPaStream(stream) {
    const source = this.cache.findPaStream(stream);
    await this._putStreamOnCorrectSink(stream);
    return this._setCorrectVolumeAndMuteForStream(stream, source);
  }

  async _syncMeteorSource(source) {
    this._log('verbose', `About to apply correct mute/volume state and routing for meteor source with id '${source._id}'`);
    const stream = this._paCache.findMeteorSource(source);
    if (!stream) {
      this._log('verbose', `Could not find corresponding stream for meteor source with id '${source._id}'`);
      return;
    }
    await this._putStreamOnCorrectSink(stream);
    return this._setCorrectVolumeAndMuteForStream(stream, source);
  }

  async _setCorrectVolumeAndMuteForStream(stream, source) {
    if (stream) {
      if (source) { // Source is known in blue
        return this._setStreamProperties(stream, {
          mute: this._shouldSourceBeMuted(source),
          // The max allowed volume value is 153%, which is the max value on linux pulse GUI
          volume: this._calcAudioVolumeValue(Math.min(Math.round(source.audio.volume), 153)),
        });
      }
      // Source is not known in blue (eg. reference audio samples, electron stream), set volume to 100%
      return this._setStreamProperties(stream, {
        mute: false,
        volume: this._calcAudioVolumeValue(100),
      });
    }
  }

  async _setStreamProperties(stream, properties) {
    return this._setDeviceProperties(stream, properties);
  }

  async _setSinkProperties(sink, properties) {
    return this._setDeviceProperties(sink, properties);
  }

  async _setSourceProperties(source, properties) {
    return this._setDeviceProperties(source, properties);
  }

  async _setDeviceProperties(device, properties) {
    this._log('verbose', `Setting properties: ${JSON.stringify(properties, null, 2)} on '${device.id}'`);
    if (!properties.hasOwnProperty('mute') && !properties.hasOwnProperty('volume')) {
      this._log('error', `Cannot set invalid properties: ${JSON.stringify(properties, null, 2)} on '${device.id}'`);
      return;
    }

    const channels = await device._iface.Channels;
    const volume = new Array(channels.length).fill(properties.volume);
    await device._iface.setMute(properties.mute);
    // Note: line below also works for playback streams but Cinergy sets
    // the volume of its own sources (overwrites volume set by line below)
    await device._iface.setVolume([volume]);
  }

  async _suspendDevice(device, suspend) {
    suspend = !!suspend;
    if (suspend) {
      this._log('verbose', `Suspending device '${device.id}'`);
    } else {
      this._log('verbose', `Unsuspending device '${device.id}'`);
    }
    await device._iface.Suspend(suspend);
  }

  async _putStreamOnCorrectSink(stream) {
    if (!stream) {
      this._log('error', 'Cannot move undefined audio stream to correct sink. Skipping...');
      return;
    }
    this._log('verbose', `About to put audio stream '${stream.id}' on correct sink...`);
    // Both mirrorOp and webRTC sharing (electron) should go to content
    // sharing mix (will be streamed along in case of eg. overflow room)
    let targetSinkName = this._audioConfig.contentSharingSink;
    let sourceType = 'local audio';
    const source = this.cache.findPaStream(stream);
    if (source) {
      // Source is known in blue...
      if (!this.cache.isLocalSource(source)) {
        // Source (Pa stream) is known in blue and
        // comes from a remote source (eg.slave overflow room)
        targetSinkName = this._audioConfig.remoteSourceSink;
        sourceType = 'remote audio';
      }
    } else {
      // Source is not known in blue, eg. reference audio sample, electron stream, ...
      if ((Object.values(audioReferenceSampleType).includes(stream.id))
        && (this._audioConfig.sinkForReferenceAudioSample.hasOwnProperty(stream.id))) {
        this._log('verbose', `Target sink for reference audio stream of type '${stream.id}' is '${this._audioConfig.sinkForReferenceAudioSample[stream.id]}'`);
        targetSinkName = this._audioConfig.sinkForReferenceAudioSample[stream.id];
        sourceType = 'reference audio sample';
      }
      // Inter tile audio should come on the tile_mix sink if device is a tile, else on default sink.
      if (stream.id.includes('TA_')) {
        this._log('verbose', `Digital Audio Mix: Target sink for inter tile audio stream '${stream.id}' is '${this._audioConfig.interTileMixSink}'`);
        targetSinkName = this._audioConfig.interTileMixSink;
      }
      // The content audio stream should come on the default sink
      if (stream.id.includes('CA_')) {
        this._log('verbose', `Digital Audio Mix: Target sink for content audio stream '${stream.id}' is '${this._audioConfig.defaultSink}'`);
        targetSinkName = this._audioConfig.defaultSink;
      }
      if (stream.id.includes('CMA_')) {
        this._log('verbose', `Digital Audio Mix: Target sink for confidence monitor audio stream '${stream.id}' is '${this._audioConfig.remoteSourceSink}'`);
        targetSinkName = this._audioConfig.remoteSourceSink;
      }
    }
    this._log('verbose', `About to move ${sourceType} stream '${stream.id}' to target sink '${targetSinkName}'`);
    // Get the pulse audio sink with name targetSinkName ...
    const sink = (this._paCache.findSinkById(targetSinkName));
    if (sink) {
      this._log('verbose', `Moving ${sourceType} stream '${stream.id}' to sink '${sink._path}' ('${sink.id}')`);
      try {
        await stream._iface.Move(sink._path);
        stream.sink = sink._path;
      } catch (error) {
        this._log('error', `Error while moving audio stream ${stream.id} to sink ${sink._path}: ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`, {
          serviceLog: true,
        });
      }
    } else {
      // Don't move the stream in this case
      this._log('error', `Cannot put stream ${stream.id} with sourceType ${sourceType} on any sink as no sinks could be found!`, {
        serviceLog: true,
      });
    }
  }

  async _putStreamOnNullSink(stream) {
    const sink = this._paCache.findNullSink();
    if (sink) {
      this._log('verbose', `Putting stream '${stream.id}' on null sink '${sink._path}' ('${sink.id}')`);
      try {
        await stream._iface.Move(sink._path);
        stream.sink = sink._path;
      } catch (error) {
        this._log('error', `Error while moving stream ${stream.id} to null sink: ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`, {
          serviceLog: true,
        });
      }
    }
  }

  async _updateInputs() {
    this._log('verbose', 'Source change detected. Updating PA cache...');
    await this._paCache.updateSources(true);
    await this._syncInputs();
  }

  _cleanInputs() {
    return this._paCache.$arrays.sources
      .filter(source => !source.id.includes('monitor'))
      .map(source => ({
        id: source.id,
        serial: source.serial,
        activePort: cleanPort(source.activePort),
        device: source.device,
        ports: source.ports.map(cleanPort),
        defaultVolume: this._findDefaultVolume('source', source),
      }));
  }

  async _updateOutputs() {
    this._log('verbose', 'Module change detected. Updating PA cache...');
    await this._paCache.updateSinks(true);
    await this._paCache.updateModules(true);
    await this._syncOutputs();
    this._log('debug', 'Sinks and Modules cache updated after module change detected. Now forcing resynchronize output config.');
    this._requestSyncOutputConfig();
  }

  async _updateCards() {
    this._log('verbose', 'Card change detected. Updating PA cache and forcing active profiles...');
    await this._paCache.updateCards(true);
    // Replugging or repowering a device can result in a different active profile
    // than the one set in the admin UI. Solve this by forcing the right profiles.
    await this._syncActiveProfilesOnCards();
  }

  async _syncActiveProfilesOnCards() {
    this._log('verbose', 'About to sync the active profile on each of the audio cards');
    // Make sure the right audio profile is activated on each audio card
    await this._applyCorrectProfilesToCards();
    await this._paCache.warm(true);
    await this._syncAll();
  }

  async _applyCorrectProfilesToCards() {
    this._log('debug', 'About to force correct profile to each audio card.');
    // Sequentially apply the right profile on each of the cards
    for (const card of [...this._paCache.$arrays.cards]) { // clone array because command might overwrite array values
      if (!this.wantsToBeActive) return;
      await this._applyProfileToCard(card);
    }
  }

  async _applyProfileToCard(card) {
    let currentActiveId = card.activeProfile.id;
    // Find the profile to be activated on the card
    const meteorCard = this.cache.findPaCard(card);
    if (!meteorCard
        || !meteorCard.hasOwnProperty('activeProfile')
        || !card
        || !card.hasOwnProperty('activeProfile')
        || !card.hasOwnProperty('profiles')) {
      return;
    }
    if (meteorCard.activeProfile.id !== currentActiveId) {
      for (let i = 0; ((i < 3) && (meteorCard.activeProfile.id !== currentActiveId) && this.wantsToBeActive); i++) {
        if (!card.profiles.find(p => p.id === meteorCard.activeProfile.id)) {
          this._log('error', `Can't force requested profile ${meteorCard.activeProfile.id} as it does not exist on card ${card.id}`);
          return;
        }
        this._log('verbose', `About to force requested profile ${meteorCard.activeProfile.id} on card ${card.id} - iteration ${i}`);
        try {
          await this._forceCardProfile(card.id, meteorCard.activeProfile.id);
        } catch (err) {
          this._log('error', `Command to force requested profile ${meteorCard.activeProfile.id} on card ${card.id} failed. Error = ${JSON.stringify(serializeError(err, { maxDepth: 1 }))}`, { serviceLog: true });
        }
        await sleep(1000); // On some NUCs reading out the card immediately shows the old values, so give it some time
        if (!this.wantsToBeActive) return;
        await this._paCache.updateCards(true);
        currentActiveId = this._paCache.findCardById(card.id).activeProfile.id;
        this._log('verbose', `Current active profile: ${currentActiveId}`);
      }
      if (meteorCard.activeProfile.id !== currentActiveId) {
        this._log('error', `Failed to force requested profile ${meteorCard.activeProfile.id} on card ${card.id}`, { serviceLog: true });
      } else {
        this._log('verbose', `Successfully activated profile ${currentActiveId} on card ${card.id}`);
      }
    } else {
      this._log('verbose', `Profile ${meteorCard.activeProfile.id} is already active on card ${card.id}`);
    }
  }

  _forceCardProfile(cardId, profileId) {
    this._log('verbose', `Activating profile ${profileId} on card ${cardId}`);
    return new Promise(((resolve, reject) => {
      exec(`${this._pacmd_set_profile_command} ${cardId} ${profileId}`, {
        timeout: 2000,
      }, (error, stdout, stderr) => {
        if (error != null) {
          reject(error);
        } else if (stderr && stderr != null) {
          reject(new Error(stderr.toString()));
        } else {
          resolve(stdout.toString());
        }
      });
    }));
  }

  async _setupRouting(routing) {
    // First unload all modules of the old routing
    this._log('verbose', 'About to unload modules of old routing...');
    const modulesToUnload = Object.values(audioModuleRole)
      .map(role => this._paCache.findModulesForWhichSinkNameIncludes(role))
      .reduce((acc, el) => [...acc, ...el], []);
    for (const module of modulesToUnload) {
      this._log('verbose', `Unloading old module with path = ${module._path}`);
      try {
        await module._iface.Unload('foo');
      } catch (error) {
        this._log('error', `Failed to unload module path = ${module._path} , error = ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`, { serviceLog: true });
        // Fix for NET0004-3027: Some tunnelsink might have gone while subscriptions
        // are off, so just continue unloading (don't throw if sink not found).
      }
    }

    // Then load all the modules of the new routing
    this._log('verbose', 'About to load modules of new routing...');
    for (const route of routing) {
      const args = Object.keys(route.arguments).map((arg) => {
        let val = route.arguments[arg];
        // values need to be coma separated
        if (Array.isArray(val)) { val = val.join(','); }
        return [arg, val];
      });
      this._log('verbose', `Loading module ${route.module} with args: ${JSON.stringify(args, null, 2)}`);
      try {
        await this._audioService.loadModule(route.module, args);
      } catch (error) {
        this._log('error', `Failed to load module ${route.module} with args: ${JSON.stringify(args, null, 2)} with error: ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`,
          { serviceLog: true });
        throw error;
      }
    }
  }

  async _requestSyncOutputConfig(syncAVScraper = false) {
    if (this._syncing) {
      this._log('info', 'Request sync output config: Sync in progress => schedule extra sync round');
      this._requestNewSync = true;
      this._requestSyncAVScraper = this._requestSyncAVScraper || syncAVScraper;
    } else {
      this._log('info', 'Request sync output config: Sync not in progress => start sync');
      this._requestNewSync = false;
      await this._syncOutputConfig(syncAVScraper);
      if (this._requestNewSync) {
        await this._requestSyncOutputConfig(this._requestSyncAVScraper);
      }
    }
  }

  async _syncOutputConfig(syncAVScraper = false) {
    this._log('info', 'Syncing output configuration');
    this._syncing = true;

    const wasSubscribed = this._subscribed;

    try {
      await this._doSyncOutputConfig();
      if (syncAVScraper) {
        this._requestSyncAVScraper = false;
        await this._syncAVScraperForAudioCalibration();
      }
    } catch (error) {
      this._log('error', `Failed to synchronize audio output config with error: ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`, {
        serviceLog: true,
      });
      // Make sure to at least re-enable the subscriptions if they were before starting
      if (wasSubscribed && !this._subscribed) {
        this._log('verbose', 'Re-enabling subscriptions...');
        this._subscribe();
      }
    } finally {
      this._syncing = false;
    }
  }

  async _doSyncOutputConfig() {
    calculateAudioConfig(this._audioConfig, this._device, this._logger);
    const currentRouting = getCurrentRoutingModules(this._paCache, audioModuleRole);
    const routingIsTheSame = isTheSameRouting(this._audioConfig.routing, currentRouting);

    this._log('verbose', '================================================================================');
    this._log('verbose', '========================== START APPLYING AUDIO CONFIG =========================');

    if (!routingIsTheSame) {
      this._log('verbose', 'Audio routing needs to be adapted. Preparing for change...');

      // Set the default sink back to the nullsink
      this._log('verbose', 'About to put the default source back to tile_mix.monitor');
      const source = this._paCache.findTileMixSource();
      if (source) { await this._setDefaultSource(source._path); }

      // Set the default sink back to the nullsink
      this._log('verbose', 'About to put the default sink back to nullsink');
      const sink = this._paCache.findNullSink();
      if (sink) { await this._setDefaultSink(sink._path); }

      // Make sure the right combined sinks (mixes) are available in pulseaudio, create them if needed
      this._log('verbose', 'About to apply the new audio routing (changed w.r.t. old routing)');
      // First disable subscriptions to avoid endless loops
      const reSubscribe = this._subscribed;
      if (this._subscribed) {
        this._log('verbose', 'First disabling subscriptions...');
        this._unsubscribe();
      }
      // Apply the new routing...
      this._log('verbose', 'Applying new audio routing. Starting cleanup of old modules and creating new ones');
      await this._setupRouting(this._audioConfig.routing);
      // Update the pulseaudio cache
      this._log('verbose', 'Refetching full pulse audio cache');
      await this._paCache.warm(true);
      // Re-enable subscriptions if needed
      if (reSubscribe) {
        this._log('verbose', 'Re-enabling subscriptions...');
        this._subscribe();
      }
    } else {
      this._log('verbose', 'Audio routing is up-to-date. Keeping current routing.');
    }

    // Set the right default source
    this._log('verbose', `About to set the default audio source to ${this._audioConfig.defaultSource}`);
    const source = this._paCache.findSourceById(this._audioConfig.defaultSource);
    if (source) { await this._setDefaultSource(source._path); }

    // Set the correct default sink
    this._log('verbose', `About to set the default audio sink to ${this._audioConfig.defaultSink}`);
    const sink = this._paCache.findSinkById(this._audioConfig.defaultSink);
    if (sink) { await this._setDefaultSink(sink._path); }

    // Apply correct volume and mute settings to the hardware sinks
    this._log('verbose', 'Applying correct volume and mute settings to all hardware sinks');
    await this._applyVolumeAndMuteSettingsToHardwareSinks();

    if (routingIsTheSame) { // else already updated by this._paCache.warm() above
      this._log('verbose', 'Updating playback stream cache before putting playback streams on correct sinks.');
      await this._paCache.updatePlaybackStreams(true);
    }
    // this._log('verbose', 'About to put streams on correct sinks. List of streams =');
    // this._paCache.$arrays.playbackStreams.forEach((pbs) => { this._log('verbose', this._paCache.clean(pbs)); });
    // Put streams on correct sinks
    this._log('verbose', 'About to put playback streams on correct sinks');
    await Promise.all(this._paCache.$arrays.playbackStreams.map(pbs => this._putStreamOnCorrectSink(pbs)));

    // Make sure all hardware audio inputs of the NRC/NCN are set to the right volume/mute
    this._log('verbose', 'Applying correct volume and mute settings to all hardware sources');
    await this._applyVolumeAndMuteSettingsToHardwareSources();

    // Make sure that sources shown on muted compositions are muted
    this._log('verbose', 'Applying correct mute state for playback streams known in blue...');
    await this._setCorrectVolumeAndMuteAndRoutingForMeteorSources();

    // Apply the correct default volume for content sharing
    this._log('verbose', 'Applying correct default volume for content sharing');
    await this._applyContentSharingDefaultVolume();

    // Scheduling extra round to pickup playback streams not flagged by DBUS interface
    this._log('verbose', '[doSyncOutputConfig] Requesting to sync playback streams');
    this._requestSyncPlaybackStreams(5000);

    // We're done ...
    this._log('verbose', '======================== FINISHED APPLYING AUDIO CONFIG ========================');
    this._log('verbose', '================================================================================');
  }

  _syncAll() {
    this._log('verbose', 'Syncing audio inputs, outputs and cards. Upserting changes if any.');
    const items = {
      audioSources: this._cleanInputs(),
      audioSinks: this._cleanOutputs(),
      audioCards: this._cleanCards(),
    };
    return this._upsert(items);
  }

  _syncInputs() {
    this._log('verbose', 'Syncing audio inputs. Upserting changes if any.');
    return this._upsert({ audioSources: this._cleanInputs() });
  }

  _syncOutputs() {
    this._log('verbose', 'Syncing audio outputs. Upserting changes if any.');
    return this._upsert({ audioSinks: this._cleanOutputs() });
  }

  _cleanOutputs() {
    return this._paCache.$arrays.sinks
      // DBUS reports activePort===null for non hardware sinks
      // eg. {"_path":"/org/pulseaudio/core1/sink3","activePort":null,"id":"nullsink","device":"Null Output","ports":[]}
      .filter(sink => (sink.activePort !== null))
      .map(sink => ({
        id: sink.id,
        activePort: cleanPort(sink.activePort),
        device: sink.device,
        ports: sink.ports.map(cleanPort),
        defaultVolume: this._findDefaultVolume('sink', sink),
      }));
  }

  _syncCards() {
    this._log('verbose', 'Syncing audio cards. Upserting changes if any.');
    return this._upsert({ audioCards: this._cleanCards() });
  }

  _cleanCards() {
    return this._paCache.$arrays.cards
      .map(card => ({
        id: card.id,
        device: card.device,
        activeProfile: cleanProfile(card.activeProfile),
        profiles: card.profiles.map(cleanProfile),
      }));
  }

  _findDefaultVolume(type, item) {
    const meteorItem = (type === 'sink') ? this.cache.findPaSink(item) : this.cache.findPaSource(item);
    if (meteorItem) {
      this._log('verbose', `Found default volume for item ${item.id} = ${meteorItem.defaultVolume}%`);
      const defaultVolume = Number.isNaN(meteorItem.defaultVolume) ? 100 : meteorItem.defaultVolume;
      return defaultVolume;
    }
    this._log('verbose', `Could not find default volume for item ${item.id} . Taking 100 %`);
    return 100;
  }

  _playReferenceSample(sampleType) {
    let sampleDetails = {};
    switch (sampleType) {
    case audioReferenceSampleType.TEACHER:
      sampleDetails = {
        path: path.join(AUDIO_SAMPLES_DIR, 'teacher.mp3'),
        pulseClientName: 'teacher',
      };
      break;
    case audioReferenceSampleType.CONTENT:
      sampleDetails = {
        path: path.join(AUDIO_SAMPLES_DIR, 'content.mp3'),
        pulseClientName: 'content',
      };
      break;
    case audioReferenceSampleType.BACKGROUND:
      sampleDetails = {
        path: path.join(AUDIO_SAMPLES_DIR, 'background.mp3'),
        pulseClientName: 'background',
      };
      break;
    default:
      return true;
    }
    const audioStream = new AudioStream(sampleDetails);
    this._audioReferenceSampleStreams.push(audioStream);
    audioStream.start();
    return true;
  }

  _stopAllReferenceSamples() {
    this._audioReferenceSampleStreams.forEach((stream) => {
      stream.destroy();
    });
    this._audioReferenceSampleStreams = [];
  }

  /**
   * Sometimes pulseaudio doesn't discover that new hardware has been
   * connected, eg. this is the case when a HDMI/DP cable is connected
   * while the device is running, hence the HDMI/DP audio output is not
   * discovered by weConnect. The code below is a trick to force
   * pulseaudio to re-scan for hardware inputs and outputs. We call
   * it whenever we want to force pulseaudio to scan for devices.
   *
   */
  async _forcePAHardwareDetection() {
    const HWDetectionModule = 'module-udev-detect';
    this._log('info', 'Forcing hardware input/output detection by re-loading the detection module.');

    const wasSubscribed = this._subscribed;
    // First disable subscriptions to avoid useless events being handled
    if (this._subscribed) {
      this._log('verbose', 'Unsubscribe before force audio hardware detect');
      this._unsubscribe();
    }

    // First unload the udev detection module if it can be found
    const moduleToUnload = this._paCache.findModulesByName(HWDetectionModule);
    if (moduleToUnload.length > 0) {
      try {
        this._log('verbose', 'Unloading detection module.');
        await moduleToUnload[0]._iface.Unload('foo');
      } catch (error) {
        this._log('error', `Failed to unload module ${HWDetectionModule}, error = ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`,
          { serviceLog: true });
      }
    } else {
      this._log('verbose', `Could not unload module ${HWDetectionModule} as it could not be found.`);
    }
    // Then unload all modules that had been loaded by udev detection module
    this._log('verbose', 'About to unload all modules that were loaded by the detection module.');
    const modulesToUnload = this._paCache.findModulesLoadedByUDevDetectModule();
    for (const module of modulesToUnload) {
      this._log('verbose', `Unloading old module with path = ${module._path}`);
      try {
        await module._iface.Unload('foo');
      } catch (error) {
        this._log('error', `Failed to unload module path = ${module._path} with error: ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`,
          { serviceLog: true });
        throw error;
      }
    }
    // Re-enable subscriptions
    if (wasSubscribed) {
      this._log('verbose', 'Re-enabling subscriptions and applying correct mute state after force audio hardware detect');
      await this._subscribeAndMuteCorrectSources();
    }
    // Re-load the udev detection module. This will scan + reload whatever is needed.
    this._log('verbose', `Re-loading module ${HWDetectionModule}`);
    try {
      await this._audioService.loadModule(HWDetectionModule, {});
    } catch (error) {
      this._log('error', `Failed to re-load module ${HWDetectionModule} with error: ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`,
        { serviceLog: true });
    }
    // Make sure to fetch all interfaces of the modules and publish the new situation.
    this._updateOutputs();
  }

  _upsert(items) {
    const sanitizeForComparison = (list) => {
      list = JSON.parse(JSON.stringify(this._paCache.clean(list)));

      list.forEach((item) => {
        Object.keys(item).forEach((key) => {
          if (item[key] === null) {
            delete item[key];
          }

          if (Array.isArray(item[key])) {
            if (item.hasOwnProperty('ports')) {
              item[key] = sanitizeForComparison(item.ports);
            }
            if (item.hasOwnProperty('profiles')) {
              item[key] = sanitizeForComparison(item.profiles);
            }
          }
        });
      });
      return keyBy(list, 'id');
    };

    const toUpsert = { _id: this._device._id };
    Object.keys(items).forEach((entity) => {
      const deviceKey = `${entity}Latest`;
      const originalItems = sanitizeForComparison(this._device[deviceKey] || []);
      const newItems = sanitizeForComparison(items[entity] || []);
      if (!isEqual(originalItems, newItems)) {
        this._log('verbose', `${entity} changed, preparing to upsert...`);
        toUpsert[deviceKey] = items[entity];
        this._device[deviceKey] = items[entity];
      }
    });
    if (Object.keys(toUpsert).length > 1) {
      this._log('verbose', 'Upserting all changes...');
      return this._meteorClient.updateDevice(toUpsert);
    }
    this._log('verbose', 'No changes detected, not upserting...');
  }

  async _onStart() {
    await super._onStart();
    // Allow 3 attempts to connect to pulseaudio via dbus, if not
    // successful audioBridge will be fully stopped and restarted.
    await this._audioService.start({ retryTimer: 5000, retries: 2, cancelToken: this.getCancelToken(serviceState.STARTED) });
    await this._initializeTillStable();
    await this._subscribeAndMuteCorrectSources();
  }

  async _initializeTillStable() {
    this._log('info', 'Initializing');
    this._initRuns = 0;

    do {
      this._isDirty = false;
      this._initRuns += 1;
      if (this._initRuns > 10) {
        throw new Error('Could not stabilize audio bridge initialization after 10 tries');
      }

      await this._initializeStep();
    } while (this._isDirty);

    this._log('info', 'Initialization done');
  }

  async _initializeStep() {
    this._log('verbose', `Running initialize step ${this._initRuns}`);
    await this._paCache.restart({ cancelToken: this.getCancelToken(serviceState.STARTED) });
    await this._syncActiveProfilesOnCards();
    await this._requestSyncOutputConfig();
  }

  async _onStop() {
    if (this._syncPlaybackStreamsTimer) {
      clearTimeout(this._syncPlaybackStreamsTimer);
    }
    this._stopAllReferenceSamples();
    await this._unsubscribe();
    await this._audioService.stop();
    this._paCache.stop();
    await super._onStop();
  }

  async _subscribeAndMuteCorrectSources() {
    this._subscribe();
    await this._setCorrectVolumeAndMuteAndRoutingForMeteorSources();
  }

  _subscribe() {
    this._log('verbose', 'Enabling subscriptions');
    this._subscribeOnTarget(this._audioService, this._listeners.service);
    this._subscribeOnTarget(this, this._listeners.own);
    this._subscribed = true;
  }

  _subscribeOnTarget(target, listeners) {
    let events = Object.keys(listeners);
    // Remove events that we're already subscribed to
    const existingObservables = Object.keys(this._observables);
    events = events.filter(event => !existingObservables.includes(event));

    events.forEach((event) => {
      // Avoid doing too much work by debouncing events that commonly occur together
      this._observables[event] = Rx.fromEvent(target, event)
        .pipe(debounceTime(100 /* ms */))
        .subscribe((...args) => {
          if (this.isActive) {
            // in case args is undefined we assign an empty array before spreading it to the listener function
            if (!Array.isArray(args)) {
              args = [];
            }
            listeners[event](args);
          }
        });
    });
  }

  _unsubscribe() {
    this._log('verbose', 'Disabling subscriptions');
    const observedEvents = Object.keys(this._observables);
    observedEvents
      .forEach((event) => {
        this._observables[event].unsubscribe();
        delete this._observables[event];
      });
    this._subscribed = false;
  }

  async _onDestroy() {
    this._cleanupAVScraper();
    await this._paCache.destroy();
    await this._audioService.destroy();
    await super._onDestroy();
  }

  async _syncAVScraperForAudioCalibration() {
    this._log('verbose', 'Sync\'in AV Monitor window, calibration state = ', this._device.audio.calibration);
    if (this._device.audio.calibration === audioCalibrationState.OFF) {
      this._cleanupAVScraper();
    } else {
      if (!this._privilegedSession) {
        this._privilegedSession = await createBrowserSession(true, true, 'av_monitor', this._device.proxy, this._logger);
      }
      const serverURL = serviceRegistry.get('configManager').getSetting('meteor.server');
      this._AVScraperURL = `${serverURL}/av_monitor/${this._device.audio.calibrationRealmId}/${this._device._id}`;

      if (!this._AVScraperWindow) {
        this._log('info', 'Creating and launching AV Monitor window');
        const windowData = {};
        windowData.token = serviceRegistry.get('configManager').getSetting('credentials.resume.token');
        windowData.logger = serviceRegistry.get('logger').getInstance('Webview', this._device._id);

        const bounds = {
          width: isDev ? 1280 : 320,
          height: isDev ? 720 : 180,
          x: 10,
          y: 10,
        };
        this._AVScraperWindow = createBrowserWindow(this._privilegedSession, windowData, bounds, this._logger);

        // Cleanup
        this._AVScraperWindow.on('closed', () => {
          delete this._AVScraperWindow;
        });
      }
      this._log('verbose', `Refreshing AV Monitor window, calibration state = ${this._device.audio.calibration}, URL = ${this._AVScraperURL}`);
      this._AVScraperWindow.loadURL(this._AVScraperURL);
    }
  }

  _cleanupAVScraper() {
    if (this._AVScraperWindow) {
      this._log('info', 'Deleting AV Monitor window...');
      this._AVScraperWindow.close();
      delete this._AVScraperWindow;
    }
  }
}

export default AudioBridge;
