import * as dbus from 'dbus-native';
import * as pify from 'pify';
import { serializeError } from 'serialize-error';

import { execParallel, waitMaximum } from '../../../utils/utils';
import DBus from '../../../dbus/bus';
import ProxiedService from '../../../services/proxiedService';
import Fetcher from '../../../dbus/fetcher';

import { detectDbusSessionAddress } from '../../../dbus/detect';

const Profile = new Fetcher({
  iface: 'org.PulseAudio.Core1.CardProfile',
  schema: {
    id: {
      key: 'Name',
      type: String,
    },
    title: {
      key: 'Description',
      type: String,
    },
    // Uncomment lines below if the "unplugged" status of
    // each of the profiles is to be known. Till now not used,
    // because we do not get any notification if the unplugged
    // status changes, so we cannot reliably show the status.
    // available: {
    //   key: 'Available',
    //   type: Boolean,
    // },
  },
});

const Port = new Fetcher({
  iface: 'org.PulseAudio.Core1.DevicePort',
  schema: {
    id: {
      key: 'Name',
      type: String,
    },
    title: {
      key: 'Description',
      type: String,
    },
    available: {
      type: Number,
    },
  },
});

const Client = new Fetcher({
  iface: 'org.PulseAudio.Core1.Client',
  schema: {
    id: {
      type: Object,
      key: 'PropertyList',
    },
    // id: {
    //   type: Object,
    //   key: 'PropertyList',
    //   element: 'application.name'
    // },
    // process: {
    //   type: Object,
    //   key: 'PropertyList',
    //   element: 'application.process.binary'
    // },
  },
});

const Clients = new Fetcher({
  iface: 'org.PulseAudio.Core1',
  schema: {
    clients: {
      type: Array,
      elementType: Client,
    },
  },
});

// Full PropertyList of a stream coming from a tunnel sink:
// { 'application.name': 'PulseAudio',
// 'application.id': 'org.PulseAudio.PulseAudio',
// 'application.version': '11.1',
// 'application.process.id': '871',
// 'application.process.user': 'barco',
// 'application.process.host': 'nrc-94c691124beb',
// 'application.process.binary': 'pulseaudio',
// 'application.language': 'en_US.UTF-8',
// 'window.x11.display': ':0',
// 'application.process.machine_id': 'd1718ca74cd948c99dab8d4b1c12f319',
// 'media.name': 'Tunnel for barco@nrc-94c691124beb',
// 'native-protocol.peer': 'TCP/IP client from 10.200.20.186:49022',
// 'native-protocol.version': '32' },

const Stream = new Fetcher({
  iface: 'org.PulseAudio.Core1.Stream',
  schema: {
    id: {
      type: Object,
      key: 'PropertyList',
      defaultValue: null,
      element: 'application.name',
    },
    process: {
      type: Object,
      key: 'PropertyList',
      defaultValue: null,
      element: 'application.process.binary',
    },
    host: {
      type: Object,
      key: 'PropertyList',
      defaultValue: null,
      element: 'application.process.host',
    },
    name: {
      type: Object,
      key: 'PropertyList',
      defaultValue: null,
      element: 'media.name',
    },
    peer: {
      type: Object,
      key: 'PropertyList',
      defaultValue: null,
      element: 'native-protocol.peer',
    },
    sink: {
      key: 'Device',
      type: String,
    },
    driver: {
      key: 'Driver',
      type: String,
    },
    // Uncomment lines below if you need PA index of card.
    // (Needed by some pacmd commands. Not used till now.)
    // index: {
    //   key: 'Index',
    //   type: Number,
    // }
  },
});

const Sink = new Fetcher({
  iface: 'org.PulseAudio.Core1.Device',
  schema: {
    id: {
      type: String,
      key: 'Name',
    },
    device: {
      type: Object,
      key: 'PropertyList',
      element: 'device.description',
    },
    ports: {
      type: Array,
      elementType: Port,
      defaultValue: [],
    },
    activePort: {
      type: Port,
      defaultValue: null,
    },
  },
});

const Source = new Fetcher({
  iface: 'org.PulseAudio.Core1.Device',
  schema: {
    id: {
      type: String,
      key: 'Name',
    },
    device: {
      type: Object,
      key: 'PropertyList',
      element: 'device.description',
    },
    serial: {
      type: Object,
      key: 'PropertyList',
      element: 'device.serial',
      optional: true,
    },
    ports: {
      type: Array,
      elementType: Port,
      defaultValue: [],
    },
    activePort: {
      type: Port,
      defaultValue: null,
    },
  },
});

const Module = new Fetcher({
  iface: 'org.PulseAudio.Core1.Module',
  schema: {
    id: {
      type: String,
      key: 'Index',
    },
    name: {
      type: String,
      key: 'Name',
    },
    arguments: {
      type: 'StringArray',
      key: 'Arguments',
    },
    // ,
    // properties: {
    //   type: Object,
    //   key: 'PropertyList'
    // }
  },
});

const Card = new Fetcher({
  iface: 'org.PulseAudio.Core1.Card',
  schema: {
    id: {
      type: String,
      key: 'Name',
    },
    device: {
      type: Object,
      key: 'PropertyList',
      element: 'device.description',
    },
    profiles: {
      type: Array,
      elementType: Profile,
      defaultValue: [],
    },
    activeProfile: {
      type: Profile,
      defaultValue: null,
    },
  },
});

const Sinks = new Fetcher({
  iface: 'org.PulseAudio.Core1',
  schema: {
    sinks: {
      type: Array,
      elementType: Sink,
    },
  },
});

const Sources = new Fetcher({
  iface: 'org.PulseAudio.Core1',
  schema: {
    sources: {
      type: Array,
      elementType: Source,
    },
  },
});

const PlaybackStreams = new Fetcher({
  iface: 'org.PulseAudio.Core1',
  schema: {
    playbackStreams: {
      type: Array,
      elementType: Stream,
    },
  },
});

const Modules = new Fetcher({
  iface: 'org.PulseAudio.Core1',
  schema: {
    modules: {
      type: Array,
      elementType: Module,
    },
  },
});

const Cards = new Fetcher({
  iface: 'org.PulseAudio.Core1',
  schema: {
    cards: {
      type: Array,
      elementType: Card,
    },
  },
});

/**
 * The Audio Service is a client for PulseAudio
 * It connects to pulse audio via DBUS.
 * It acts as an event proxy and provide utilities like fetching the list of sinks or muting the audio
 */
class PulseAudioService extends ProxiedService {
  constructor(options) {
    super(options);

    this.on('sinkRemoved', () => {
      this.emit('sinksChanged');
    });

    this.on('sinkAdded', () => {
      this.emit('sinksChanged');
    });

    this.on('sourceRemoved', () => {
      this.emit('sourcesChanged');
    });

    this.on('sourceAdded', () => {
      this.emit('sourcesChanged');
    });

    this.on('cardRemoved', () => {
      this.emit('cardsChanged');
    });

    this.on('cardAdded', () => {
      this.emit('cardsChanged');
    });

    this.on('playbackStreamAdded', (args) => {
      this.emit('onPlaybackStreamAdded', args);
    });

    this.on('playbackStreamRemoved', (args) => {
      this.emit('onPlaybackStreamRemoved', args);
    });

    // Not used yet at this moment...
    // this.on('moduleAdded', () => {
    //   this.emit('modulesChanged');
    // });

    // Not used yet at this moment...
    // this.on('moduleRemoved', () => {
    //   this.emit('modulesChanged');
    // });

    // Code below is still commented at this time. It detects whether
    // the active port of an input or output changes. You can trigger the
    // event by using pavucontrol and eg. for the built-in Audio Analog
    // Stereo sink change the port, eg. from Line Out -> Speaker. However
    // as it is not yet handled at server side and not strictly needed for
    // the product, it is still in comment for now.
    // this.on('activePortChanged', (device) => {
    //   if (device[0].includes('/org/pulseaudio/core1/sink')) {
    //     this.emit('sinksChanged');
    //   } else if (device[0].includes('/org/pulseaudio/core1/source')) {
    //     this.emit('sourcesChanged');
    //   }
    // });
  }

  get readyToSubscribeEvent() {
    return 'readyToSubscribe';
  }

  get readyToSubscribe() {
    return this.isStarted && this._dbus && this._dbus.isConnected;
  }

  get proxyMap() {
    return {
      sinkAdded: 'org.PulseAudio.Core1.NewSink',
      sinkRemoved: 'org.PulseAudio.Core1.SinkRemoved',
      sourceAdded: 'org.PulseAudio.Core1.NewSource',
      sourceRemoved: 'org.PulseAudio.Core1.SourceRemoved',
      cardAdded: 'org.PulseAudio.Core1.NewCard',
      cardRemoved: 'org.PulseAudio.Core1.CardRemoved',
      activePortChanged: 'org.PulseAudio.Core1.Device.ActivePortUpdated',
      playbackStreamRemoved: 'org.PulseAudio.Core1.PlaybackStreamRemoved',
      playbackStreamAdded: 'org.PulseAudio.Core1.NewPlaybackStream',
      clientRemoved: 'org.PulseAudio.Core1.ClientRemoved',
      clientAdded: 'org.PulseAudio.Core1.NewClient',
      moduleAdded: 'org.Pulseaudio.Core1.NewModule',
      moduleRemoved: 'org.Pulseaudio.Core1.ModuleRemoved',
    };
  }

  getClients(options) {
    return Clients.fetch(this._pa, '/org/pulseaudio/core1', options);
  }

  getSinks(options) {
    return Sinks.fetch(this._pa, '/org/pulseaudio/core1', options);
  }

  getSources(options) {
    return Sources.fetch(this._pa, '/org/pulseaudio/core1', options);
  }

  getCards(options) {
    return Cards.fetch(this._pa, '/org/pulseaudio/core1', options);
  }

  getPlaybackStreams(options) {
    return PlaybackStreams.fetch(this._pa, '/org/pulseaudio/core1', options);
  }

  getModules(options) {
    return Modules.fetch(this._pa, '/org/pulseaudio/core1', options);
  }

  getCard(path, options) {
    return Card.fetch(this._pa, path, options);
  }

  getModule(path, options) {
    return Module.fetch(this._pa, path, options);
  }

  getSource(path, options) {
    return Source.fetch(this._pa, path, options);
  }

  getSink(path, options) {
    return Sink.fetch(this._pa, path, options);
  }

  getPlaybackStream(path, options) {
    return Stream.fetch(this._pa, path, options);
  }

  _subscribeForNotification(event, mappedEvent) {
    // this._log('verbose', `===>>> Starting to listen for PA events of type ${event}`);
    this._paCore.ListenForSignal(event, []);
    this._dbus.on(event, (...args) => {
      // this._log('verbose', `===>>> Received event:
      // ${JSON.stringify(event, null, 2)} with arguments ${JSON.stringify(args, null, 2)}`);
      // this._log('verbose', `===>>> Emitting mapped event:
      // ${JSON.stringify(mappedEvent, null, 2)} with arguments ${JSON.stringify(args, null, 2)}`);
      this.emit(mappedEvent, args);
    });
  }

  _unsubscribeForNotification(event) {
    this._paCore.StopListeningForSignal(event, []);
    this._dbus.removeAllListeners(event);
  }

  async _detectPulseAddress() {
    const address = await detectDbusSessionAddress();

    this._log('info', `DBUS session address: ${address}`);

    const service = dbus.sessionBus({
      busAddress: address,
    }).getService('org.PulseAudio1');

    const PASL = await pify(service.getInterface.bind(service))('/org/pulseaudio/server_lookup1', 'org.PulseAudio.ServerLookup1');
    const paDbusAddress = await pify(PASL.Address.bind(PASL))();
    return paDbusAddress;
  }

  /**
   * Start DBUS client and cache the audio devices
   */
  async _onStart() {
    this._log('info', 'Starting pulse audio DBUS client');
    let address;
    try {
      // If not finished within 1 second, stop the attempt and retry (avoid hanging audioBridge)
      address = await waitMaximum(this._detectPulseAddress(), 1000);
    } catch (error) {
      this._log('error', `Failed to detect PA session bus address, about to retry. Error = ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`, { serviceLog: true });
      throw error;
    }

    this._log('info', `Connecting with PulseAudio on sessionBus address: ${address}`);

    this._dbus = new DBus({
      direct: true,
      busAddress: address,
    });

    this._dbus.on('connected', () => {
      this._log('info', 'Connected with PulseAudio');

      if (this.readyToSubscribe) {
        this.emit('readyToSubscribe');
      }
    });

    this._pa = this._dbus.getService('org.PulseAudio');
    await this._dbus.connect();
    const paCore = await this._pa.getInterface('/org/pulseaudio/core1', 'org.PulseAudio.Core1');
    this._paCore = paCore;
    this.emit('readyToSubscribe');
    await super._onStart();
  }

  loadModule(name, args) {
    return this._paCore.LoadModule(name, args);
  }

  async setFallbackSink(path) {
    let objPath;
    try {
      objPath = await this._paCore.FallbackSink;
    } catch (err) {
      this._log('error', `Reading FallbackSink value ERROR: ${err}`);
    }
    if (objPath !== path) {
      this._log('verbose', `Changing fallback sink from ${objPath} to ${path}`);
      try {
        await this._paCore.setFallbackSink(path);
      } catch (err) {
        this._log('error', `Setting default sink to path: ${path} ERROR: ${err}`);
      }
    } else {
      this._log('verbose', `Fallback sink is already set to ${path}, skipping`);
    }
  }

  async setFallbackSource(path) {
    let objPath;
    try {
      objPath = await this._paCore.FallbackSource;
    } catch (err) {
      this._log('error', `Reading FallbackSource value ERROR: ${err}`);
    }
    if (objPath !== path) {
      this._log('verbose', `Changing fallback source from ${objPath} to ${path}`);
      try {
        await this._paCore.setFallbackSource(path);
      } catch (err) {
        this._log('error', `Setting default source to path: ${path} ERROR: ${err}`);
      }
    } else {
      this._log('verbose', `Fallback source is already set to ${path}, skipping`);
    }
  }

  async startPaService() {
    this._log('info', 'Starting PulseAudio Services');

    try {
      await execParallel([
        'sudo service pulseaudio start',
      ]);
    } catch (error) {
      this._log('error', 'Error while stopping PA services', error);
    }
  }

  async stopPaService() {
    this._log('info', 'Stopping PulseAudio Services');

    try {
      await execParallel([
        'sudo service pulseaudio stop',
      ]);
    } catch (error) {
      this._log('error', 'Error while stopping PA services', error);
    }
  }

  async restartPaService() {
    this._log('info', 'Restarting PA Services');
    await this.stopPaService();
    await this.startPaService();
  }

  /**
   * Stop the DBUS client and delete cached data
   */
  async _stopDBusClient() {
    if (this._dbus) { await this._dbus.destroy(); }

    delete this._dbus;
    delete this._pa;
    delete this._paCore;
  }

  async _onStop() {
    await this._stopDBusClient();
    await super._onStop();
  }
}

export default PulseAudioService;
