import { isEqual } from 'lodash';

export const cleanPort = (port) => {
  if (port) {
    return {
      id: port.id,
      title: port.title,
    };
  }
};

export const cleanProfile = (profile) => {
  if (profile) {
    return {
      id: profile.id,
      title: profile.title,
    };
  }
};

/**
 * This function throws to stop forEach loop if route configuration is not found in active pulsaudio modules
 * */
export const paRouteShouldExistInModules = (routeToCompare, modules) => {
  const found = modules
  // first filter modules of the same type
    .filter(route => route.module === routeToCompare.module)
    // then check arguments
    .filter((route) => {
      let match = true;
      Object.keys(route.arguments).forEach((key) => {
        let moduleArgVal = route.arguments[key];
        let valToCompare = routeToCompare.arguments[key];
        // If an array is expected, convert if necessary before comparing:
        // Example below should result in Required routing === Current routing
        //
        // Required routing = [
        //   {
        //     "module": "module-combine-sink",
        //     "arguments": {
        //       "sink_name": "main_mix",
        //       "slaves": [                                      <-- "slaves" expects array to load module
        //         "alsa_output.pci-0000_00_1b.0.analog-stereo"
        //       ]
        //     }
        //   }
        // ]
        // Current routing = [
        //   {
        //     "module": "module-combine-sink",
        //     "arguments": {
        //       "sink_name": "main_mix",
        //       "slaves": "alsa_output.pci-0000_00_1b.0.analog-stereo"  <-- Reading out pulse gives no array
        //     }
        //   }
        // ]
        if (Array.isArray(valToCompare) || Array.isArray(moduleArgVal)) {
          if (!Array.isArray(valToCompare)) {
            valToCompare = valToCompare.split(',');
          }
          if (!Array.isArray(moduleArgVal)) {
            moduleArgVal = moduleArgVal.split(',');
          }
        }
        match = match && isEqual(moduleArgVal, valToCompare);
      });
      return match;
    });
  if (!found.length) {
    throw new Error("Didn't match any known module");
  }
};

export const isTheSameRouting = (currentRouting, requestedRouting) => {
  // return false if number of requested and current routing modules is different
  if (currentRouting.length !== requestedRouting.length) return false;
  try {
    requestedRouting.forEach(route => paRouteShouldExistInModules(route, currentRouting));
  } catch (e) {
    // exception was thrown, it means this route didn't match any existing route
    return false;
  }
  // if got there, it means requested routing is the same as current
  return true;
};

// Returns an array of indices of elements for which the defaultVolume has changed.
// object 1 :  [ { id: 'alsa_output.pci-0000_00_1b.0.analog-stereo',
//   activePort: { id: 'analog-output-lineout', title: 'Line Out' },
//   device: 'Built-in Audio Analog Stereo',
//   ports: [ [Object], [Object] ],
//   defaultVolume: 60 },
//   { id: 'alsa_output.usb-Sennheiser_Communications_Sennheiser_USB_headset-00.analog-stereo',
//   activePort: { id: 'analog-output-headphones', title: 'Headphones' },
//   device: 'Sennheiser USB headset Analog Stereo',
//   ports: [ [Object] ],
//   defaultVolume: 15 } ]
// object 2 :  [ { id: 'alsa_output.pci-0000_00_1b.0.analog-stereo',
//   activePort: { id: 'analog-output-lineout', title: 'Line Out' },
//   device: 'Built-in Audio Analog Stereo',
//   ports: [ [Object], [Object] ],
//   defaultVolume: 45 },
//   { id: 'alsa_output.usb-Sennheiser_Communications_Sennheiser_USB_headset-00.analog-stereo',
//   activePort: { id: 'analog-output-headphones', title: 'Headphones' },
//   device: 'Sennheiser USB headset Analog Stereo',
//   ports: [ [Object] ],
//   defaultVolume: 15 } ]
// result     :  [ 0 ]
export const getObjectDiff = (obj1, obj2) => {
  const result = [];
  for (let i = 0; i < obj2.length; i++) {
    const ele2 = obj1.filter(a => (a.id === obj2[i].id));
    if ((ele2.length > 0) && ele2[0].hasOwnProperty('defaultVolume') && obj2[i].hasOwnProperty('defaultVolume')
      && (ele2[0].defaultVolume !== obj2[i].defaultVolume)) {
      result.push(i);
    }
    if ((ele2.length > 0) && ele2[0].hasOwnProperty('activeProfile') && obj2[i].hasOwnProperty('activeProfile')
      && (ele2[0].activeProfile.id !== obj2[i].activeProfile.id)) {
      result.push(i);
    }
  }
  return result;
};
