import { URL } from 'url';
import {
  isRegExp, isString, isFunction, isEqual,
} from 'lodash';
import * as os from 'os';
import { screen } from 'electron';

import Bridge from '../bridge';
import MeteorCacheManager from '../../../services/ddp/meteorCacheManager';
import {
  sinkType,
  compositionType,
} from '../../../services/ddp/constants';
import serviceRegistry from '../../../services/serviceRegistry';
import { translateDisplayBounds } from '../../../utils/display';
import { createBrowserWindow, createBrowserSession, getProxyConfig } from '../../../utils/browser/window';
import { sleep } from '../../../utils/utils';

import '../../../utils/speech-to-text/ipcMain';

const PDD_ENV = JSON.parse((process.env.BROWSER_ENV || '{}'));
const ALWAYS_BROWSER_DEBUG = !!process.env.BROWSER_DEBUG;
const DISABLE_BROWSER_ALWAYS_ON_TOP = !!process.env.DISABLE_BROWSER_ALWAYS_ON_TOP;

/**
 * The BrowserBridge is the glue code between blue and the browser
 * It leverages the electron browser window API
 *
 * It will manage the browser windows for the browser-based compositions
 *
 * There are 2 types of browser windows: privileged once and unprivileged ones
 * Privileged ones are the ones served by our cloud application, unprivleged ones are all the others.
 * Privileged ones run in a different session and have access to extra variables defined in the preload script
 * We also automatically grant any permission request (e.g. webcam access).
 */
class BrowserBridge extends Bridge {
  constructor(options) {
    super(options);

    this.cache = new MeteorCacheManager(this, ['sinks', 'compositions']);
    this.windows = {};

    // Users can enter any url in the admin UI, however, this could expose confidential files on the device.
    // So disallow any
    this._forbiddenUrls = [
      'localhost',
      '127.0.0.1',
      ...Object.values(os.networkInterfaces()).reduce((acc, el) => [...acc, ...el], []).map(el => el.address),
    ];
    this._eventProxy = serviceRegistry.get('eventProxy');

    this._updateAllWindows = this._updateAllWindows.bind(this);
  }

  async createSessions() {
    // Create the privileged and public sessions
    this.sessions = {
      privileged: await createBrowserSession(true, true, 'privileged', this._device.proxy, this._logger),
      public: await createBrowserSession(false, true, 'public', this._device.proxy, this._logger),
    };
  }

  get supportedCompositionTypes() {
    return [compositionType.BROWSER];
  }

  get supportedSinkTypes() {
    return [sinkType.SCREEN];
  }

  get isServiceRequired() {
    return (
      super.isServiceRequired
      && this._device.configured
      && this._meteorClient._ddpClient.collections.compositions && //eslint-disable-line
      Object.values(this._meteorClient._ddpClient.collections.compositions) //eslint-disable-line
        .filter(comp => comp.type === compositionType.BROWSER).length > 0
    );
  }

  async _onStart() {
    if (!this.sessions) {
      await this.createSessions();
    }

    screen.on('display-metrics-changed', this._updateAllWindows);
    await super._onStart();
  }

  async _onStop() {
    await super._onStop();
    this._log('verbose', 'Destroying all browser windows');
    screen.off('display-metrics-changed', this._updateAllWindows);
    Object.values(this.windows).forEach(win => win.destroy());
  }

  _onThisDeviceChanged(device, oldFields, clearedFields, newFields) {
    const promises = [];
    if (newFields.hasOwnProperty('proxy')) {
      promises.push(this._setProxy());
    }

    if (newFields.hasOwnProperty('displaysLatest')) {
      promises.push(this._updateAllWindows());
    }

    return Promise.all(promises);
  }

  _onSinkAdded(sink) {
    return this._createOrUpdateBrowserWindowForSink(sink);
  }

  _onSinkChanged(sink, oldFields, clearedFields, newFields) {
    if (newFields.hasOwnProperty('activeCompositionId')) {
      return this._createOrUpdateBrowserWindowForSink(sink);
    }
    return true;
  }

  _onSinkRemoved(sink) {
    return this._destroyBrowserWindowForSink(sink);
  }

  _onCompositionAdded(composition) {
    this._log('verbose', 'Composition added, activating url on sinks');
    return this._createOrUpdateBrowserWindowsForComposition(composition);
  }

  _onCompositionChanged(composition, oldFields, clearedFields, newFields) {
    let needsUpdate = false;

    if (newFields.hasOwnProperty('url')) {
      this._log('verbose', `URL on composition ${composition._id} changed to: ${composition.url}, activating on sinks`);
      needsUpdate = true;
    }

    if (newFields.hasOwnProperty('muted')) {
      this._log('verbose', `Composition ${composition._id} ${composition.muted ? '' : 'un'}muted`);
      needsUpdate = true;
    }

    if (newFields.hasOwnProperty('enableDevtools')) {
      this._log('verbose', `Devtools enabled for composition ${composition._id}`);
      needsUpdate = true;
    }

    if (needsUpdate) {
      return this._createOrUpdateBrowserWindowsForComposition(composition);
    }

    return true;
  }

  _updateAllWindows() {
    const promises = this.cache.$arrays.sinks
      .map(sink => this._createOrUpdateBrowserWindowForSink(sink));
    return Promise.all(promises);
  }

  _destroyBrowserWindowForSink(sink) {
    if (this.windows.hasOwnProperty(sink._id)) {
      this._log('verbose', `Destroying browser window for sink: ${sink._id}`);
      this.windows[sink._id].destroy();
      delete this.windows[sink._id];
    }
  }

  async _createOrUpdateBrowserWindowForSink(sink) {
    const composition = this.cache.findCompositionOnSink(sink);
    return this._createOrUpdateBrowserWindow(sink, composition);
  }

  _createOrUpdateBrowserWindowsForComposition(composition) {
    const promises = this.cache.findActiveSinksForComposition(composition)
      .map(sink => this._createOrUpdateBrowserWindow(sink, composition));

    return Promise.all(promises);
  }

  async _createOrUpdateBrowserWindow(sink, composition) {
    if (sink && composition && composition.url) {
      const url = this._sanitizeUrl(composition.url);
      const privileged = this._isPrivilegedUrl(composition.url);
      const login = composition.isGui;

      this._log('info', `Updating window for sink: ${sink._id}`);

      // We can't swap a browser window from session.
      // So if the privileged status changes, we need to re-create the window.
      let currentWindow = this.windows[sink._id];
      if (currentWindow && (
        currentWindow.privileged !== privileged
        || currentWindow.login !== login
      )) {
        this._destroyBrowserWindowForSink(sink);
        currentWindow = null;
      }

      let bounds;
      if (sink.matchDisplay) {
        bounds = translateDisplayBounds(sink);
      } else {
        bounds = {
          width: sink.width,
          height: sink.height,
          x: sink.x,
          y: sink.y,
        };
      }

      // Apparently, electron misses one pixel, workaround:
      bounds.width += 1;
      bounds.height += 1;

      const enableBrowserDebug = ALWAYS_BROWSER_DEBUG || composition.enableDevtools;
      if (enableBrowserDebug) {
        sink.hidden = false;
        bounds.x = 0;
        bounds.y = 0;
        bounds.width = 1680;
        bounds.height = 1050;

        if (currentWindow) {
          currentWindow.openDevtools();
        }
      }

      // Create window if none exist
      if (!currentWindow) {
        this._log('info', `Creating ${privileged ? '' : 'un'}privileged browser window for sink: ${sink._id}, at: ${JSON.stringify(bounds)}`);

        const windowData = {};

        if (privileged) {
          windowData.sinkId = sink._id;
          windowData.logger = serviceRegistry.get('logger').getInstance('Webview', sink._id);
          windowData.env = PDD_ENV;
          windowData.token = serviceRegistry.get('configManager').getSetting('credentials.resume.token');
        }

        // Due to the asynchronisity in the above code, another window could have been created in the meantime.
        // That's probably a newer one, so abort
        if (this.windows[sink._id]) {
          return;
        }

        // Else create the window
        currentWindow = createBrowserWindow(
          privileged ? this.sessions.privileged : this.sessions.public,
          windowData, bounds, this._logger, enableBrowserDebug, !sink.hidden, url,
        );

        if (!sink.hidden && !DISABLE_BROWSER_ALWAYS_ON_TOP) {
          // if sink should be visible, move it to front after CinergyBridge has been initialized
          // Note: this sets the "always on top" flag of the windows to true
          this._eventProxy.on('CinergyBridge:initialized', currentWindow.bringToFront);

          currentWindow.on('closed', () => {
            this._eventProxy.removeListener('CinergyBridge:initialized', currentWindow.bringToFront);
          });
        }

        currentWindow.on('closed', () => {
          delete this.windows[sink._id];
        });

        currentWindow._connectionMonitor.on('timeout', () => {
          this._log('error', 'Browser window is not responsive. Heartbeat reached the timeout');
          if (process.env.DISABLE_BROWSER_MONITOR || currentWindow.isDestroyed()) {
            return;
          }
          this._destroyBrowserWindowForSink(sink);
          if (currentWindow._connectionMonitor) {
            this._log('error', 'expected window connection monitor to be cleaned up, but apparently closed event was not fired');
            currentWindow._connectionMonitor.destroy();
            delete currentWindow._connectionMonitor;
          }
          this._createOrUpdateBrowserWindowForSink(sink).catch(err => this._log('error', err.message));
        });

        currentWindow.privileged = privileged;
        currentWindow.login = login;

        this.windows[sink._id] = currentWindow;
      }

      const currentBounds = currentWindow.getBounds();
      currentBounds.width += 1;
      currentBounds.height += 1;
      this._log('info', `Checking if window is truly : ${JSON.stringify(bounds, null, 2)}`);

      if (!isEqual(currentBounds, bounds)) {
        this._log('info', `The current window bounds do not match the requested ones. Current: ${JSON.stringify(currentBounds, null, 2)}, requested: ${JSON.stringify(bounds, null, 2)}`);

        currentWindow.setBounds(bounds);
      }

      // Prevent users from performing actions to leave kiosk-like mode
      this._setSecurityHandlers(sink._id);

      // Update audio state
      this._setWebContentsMuted(sink._id, composition.muted);

      // Load url if changed
      if (currentWindow.url !== url) {
        this._log('info', `Loading ${privileged ? '' : 'un'}privileged url: ${url} on sink: ${sink._id}`);
        await this._tryAndLoadUrl(sink, url);
        // We could query the url on the window,
        // but on interactive windows (e.g. touchscreens) this might have changed and we don't need to reload
        currentWindow.url = url;
      }
    }
  }

  async _tryAndLoadUrl(sink, url) {
    try {
      const composition = this.cache.findCompositionOnSink(sink);
      const window = this.windows[sink._id];
      if (window && !window.isDestroyed() && sink && composition && composition.url === url) {
        window.loadURL(url);
      }
    } catch (error) {
      this._log('error', 'Could not load url', error);
      await sleep(2000);
      return this._tryAndLoadUrl(sink, url);
    }
  }

  _setSecurityHandlers(sinkId) {
    if (this.windows.hasOwnProperty(sinkId)) {
      const window = this.windows[sinkId];
      if (window) {
        window.webContents.setWindowOpenHandler(() => ({ action: 'deny' }));
      }
    }
  }

  _setWebContentsMuted(sinkId, muted) {
    if (this.windows.hasOwnProperty(sinkId)) {
      const window = this.windows[sinkId];
      if (window && window.webContents.audioMuted !== muted) {
        this.windows[sinkId].webContents.audioMuted = muted;
      }
    }
  }

  /**
   * Checks if the url is not in the forbidden list
   */
  _isForbiddenUrl(url) {
    return !!this._forbiddenUrls.find((forbidden) => {
      if (isRegExp(forbidden)) {
        return forbidden.test(url);
      } else if (isString(forbidden)) {
        return forbidden === url;
      } else if (isFunction(forbidden)) {
        return forbidden(url);
      }
      return false;
    });
  }

  /**
   * Url is user input, so do some sanitation
   *
   * @param {*} url
   */
  _sanitizeUrl(url) {
    this._log('debug', `Sanitizing url: ${url}`);

    url = url.trim();

    // This is for when people forget to enter the protocol, but
    // Note that this is also a safety measure to avoid loading files with file protocol etc.
    if (!url.startsWith('http')) {
      url = `http://${url}`;
    }

    if (!this._validateUrl(url)) {
      url = 'about:blank';
    }

    return url;
  }

  /**
   * Validate if this is a valid url.
   * Valid means correct & allowed
   * @param {*} url
   */
  _validateUrl(url) {
    this._log('debug', `Validating url: ${url}`);

    try {
      const parsedUrl = new URL(url);

      if (!parsedUrl.hostname) {
        throw new Error(`Invalid url: ${url}`);
      }

      if (this._isForbiddenUrl(parsedUrl.hostname)) {
        throw new Error(`Forbidden url: ${url}`);
      }
    } catch (error) {
      this._log('error', `Error while parsing url ${url}`, error);
      return false;
    }

    return true;
  }

  /**
   * Check wether this url should run in the privileged session
   * i.e. if this url is served by our cloud application
   * @param {*} url
   */
  _isPrivilegedUrl(url) {
    try {
      const cloudUrl = new URL(serviceRegistry.get('configManager').getSetting('meteor.server'));
      const parsedUrl = new URL(url);
      return (parsedUrl.hostname.endsWith(cloudUrl.hostname));
    } catch (error) {
      return false;
    }
  }

  _setProxy() {
    this._log('info', 'Setting proxy on browser sessions');
    return Promise.all(Object
      .values(this.sessions)
      .map(session => session.setProxy(getProxyConfig(this._device.proxy))));
  }

  async _onDestroy() {
    await super._onDestroy();
    delete this._forbidden;
    delete this.windows;
  }
}

export default BrowserBridge;
