/*
eslint no-underscore-dangle: ["error", {
  "allow": ["_cameraService", "_id", "_device"],
  "allowAfterThis": true,
  "allowAfterSuper": true,
}]
*/
import * as _ from 'lodash';

import Bridge from '../bridge';
import CameraService from './service';

import {
  transformCamerasToMeteor,
  transformCameraFormatToV4L,
} from './utils';
import { usbCameraSourceInfo } from '../utils/audio/calculatePAConfig';
import V4LCacheManager from './v4lCacheManager';
import MeteorCacheManager from '../../../services/ddp/meteorCacheManager';
import { state as serviceState } from '../../../services/service';
import {
  sinkType, compositionType, usbCameraRole,
} from '../../../services/ddp/constants';

const isSWFramegrabber = camera => camera && CameraService.isSWCamera(camera);
const isActiveSWFramegrabber = camera => isSWFramegrabber(camera) && camera.role !== usbCameraRole.IDLE;
const parsePort = ({ port }) => Number.parseInt(port, 10);
const getFirstAvailableRTSPPort = (ports) => {
  for (let port = 8700; port < 8800; port += 2) {
    if (!ports.includes(port)) {
      return port;
    }
  }

  const error = 'No more RTSP port available....';
  throw new Error(error);
};
const isSWFramegrabberSink = sink => sink.type === sinkType.SWFG_STREAM;
// Avoid race conditions when overtaking lectures by using sink id as part of the stream id.
const generateSWFGStreamId = (cameraId, sinkId) => `${cameraId}_${sinkId}`;

/**
 * The USB Camera Bridge is a bridge between Blue and the Camera service.
 * From Blue->Device it will manage the required RTSP streams and sync image formats & image settings
 * From Device->Blue it will update the list of attached devices
 */
class UsbCameraBridge extends Bridge {
  constructor(options) {
    super(options);
    this._cameraService = new CameraService();
    this._initializeSWFrameGrabber();

    this._streams = {};

    this._cameraService.on('usbCamerasChanged', ({
      cameras,
    }) => {
      this._cameraCache.setCameras(transformCamerasToMeteor(cameras));
      this._syncCameras();
    });

    this._cameraCache = new V4LCacheManager();
    this.cache = new MeteorCacheManager(this, ['compositions', 'sinks']);
    this._usbCameraAudioInfo = [];
  }

  get supportedCompositionTypes() {
    return [compositionType.CINERGY, compositionType.BROWSER];
  }

  get supportedSinkTypes() {
    return [sinkType.SWFG_STREAM];
  }

  async _onStart() {
    await super._onStart();
    await this._cameraService.start({ cancelToken: this.getCancelToken(serviceState.STARTED) });
    await this._cameraCache.start({ cancelToken: this.getCancelToken(serviceState.STARTED) });
    await this._updateAndSyncCameras();
    await this._setupStreams();
  }

  async _onStop() {
    await super._onStop();
    await this._cameraService.stop();
    await this._cameraCache.stop();
    await Object.values(this._streams).map(stream => stream.destroy());
  }

  _onThisDeviceChanged(device, oldFields, clearedFields, newFields) {
    let restart = false;
    let forceRestart = false;

    // The camera configuration (based on device role)
    // determines if we need to stream or not, so set-up streams if this changes.
    if (newFields.hasOwnProperty('usbCameraConfig') || newFields.hasOwnProperty('rtspAuthentication')) {
      this._log('verbose', 'USB Camera stream config changed on device. About to recreate all USB camera streams if needed.');
      restart = true;
    }

    if ((newFields.hasOwnProperty('audio'))) {
      if (newFields.audio.calibration !== oldFields.audio?.calibration) {
        this._log('verbose', `Audio calibration status changed to ${newFields.audio.calibration}. About to recreate all USB camera streams.`);
        restart = true;
        forceRestart = true;
      }
    }

    if (restart && this.isActive) {
      this._setupStreams(forceRestart);
    }

    // This contains the camera settings not part of the device role
    // these are the image format & parameters.
    if (newFields.hasOwnProperty('usbCameras')) {
      this._log('verbose', 'USB cameras on device changed, syncing camera settings');
      this._syncCameraConfiguration();
    }
  }

  _isSWFrameGrabberId(cameraId) {
    return cameraId === this._cameraService.SWFrameGrabberId;
  }

  _getCameraIdFromStreamId(streamId) {
    if (streamId.includes(this._cameraService.SWFrameGrabberId)) {
      return this._cameraService.SWFrameGrabberId;
    }
    return streamId;
  }

  _getStreamIdFromCameraId(cameraId) {
    if (this._isSWFrameGrabberId(cameraId)) {
      const sink = this._getSWFrameGrabberSink();
      return sink ? generateSWFGStreamId(cameraId, sink._id) : cameraId;
    }
    return cameraId;
  }

  _getSWFrameGrabberSink() {
    return this.cache.$arrays.sinks.find(sink => sink.type === sinkType.SWFG_STREAM);
  }

  _shouldHaveStreamFor(camera) {
    return !!camera && ((camera.stream || camera.useLocally) // regular usb cam case
      || this._shouldHaveContentStreamFor(camera)); // virtual camera case
  }

  _shouldHaveContentStreamFor(camera) {
    return isActiveSWFramegrabber(camera) && this.cache.$arrays.sinks.some(sink => sink.type === sinkType.SWFG_STREAM);
  }

  _onSinkAdded(sink) {
    if (!isSWFramegrabberSink(sink)) {
      this._log('verbose', 'New sink is not a SW framegrabber sink, skipping');
      return Promise.resolve();
    }

    const streamId = generateSWFGStreamId(this._cameraService.SWFrameGrabberId, sink._id);
    if (this._streams.hasOwnProperty(streamId)) {
      this._log('verbose', 'V4L stream already exists, skipping');
      return Promise.resolve();
    }
    return this._createAndStartStreamFor(this._cameraService.SWFrameGrabberId, streamId);
  }

  _onSinkRemoved(sink) {
    if (!isSWFramegrabberSink(sink)) {
      this._log('verbose', 'Removed sink is not a SW framegrabber sink, skipping');
      return Promise.resolve();
    }

    const streamId = generateSWFGStreamId(this._cameraService.SWFrameGrabberId, sink._id);
    if (this._streams.hasOwnProperty(streamId)) {
      this._log('info', `Destroying V4L stream with id ${streamId} for sink ${sink._id}`);
      return this._streams[streamId].destroy();
    }
    return Promise.resolve();
  }

  async _onSinkChanged(sink, oldFields, clearedFields, newFields) {
    if (isSWFramegrabberSink(sink)
      && this._anyChanged(['width', 'height', 'framerate'], newFields)) {
      await this._onSinkRemoved(sink);
      await this._onSinkAdded(sink);
    }
  }

  getSWFramegrabberConfig() {
    return this._device.hasOwnProperty('usbCameraConfig') && this._device.usbCameraConfig.find(isSWFramegrabber);
  }

  /**
   * Creates the V4L stream that feeds video into the SW framegrabber
   */
  _createContentStream() {
    this._log('verbose', 'About to create V4L content stream');

    // Find the sink (determines resolution)
    const sink = this.cache.$arrays.sinks.find(snk => isSWFramegrabberSink(snk));
    if (!sink) {
      this._log('error', 'No SW framegrabber sink found, cannot create a V4L content stream');
      return;
    }

    // Find the composition (determines the shm to take content from)
    const composition = this.cache.findCompositionOnSink(sink);
    if (!composition) {
      this._log('verbose', 'No composition found for this sink, skipping');
      return;
    }

    // Find the SW framegrabber (determines path to stream to)
    const camera = this._cameraCache.$arrays.cameras.find(cam => cam.id === this._cameraService.SWFrameGrabberId);
    if (!camera) {
      this._log('verbose', 'No SW framegrabber found, skipping V4L content stream creation');
      return;
    }

    // Find the camera config (determines stream resolution)
    const cameraConfig = this.getSWFramegrabberConfig();
    if (!cameraConfig) {
      this._log('verbose', 'No configuration found for SW framegrabber, skipping V4L content stream creation');
      return;
    }

    // Create the stream
    this._log('verbose', `Creating V4L content stream for sink ${sink._id}`);
    switch (composition.type) {
    case compositionType.CINERGY:
      return this._cameraService.createV4LShmStream({
        path: `/tmp/cinergy/screen${composition.index ? composition.index : ''}`,
        resolution: { width: sink.width, height: sink.height },
        framerate: { numerator: 1, denominator: 30 }, // consistent opposite language
        v4lPath: camera.path,
        v4lResolution: cameraConfig.activeResolution,
      });
    default:
      return this._cameraService.createV4LScreenStream({
        origin: { x: sink.x, y: sink.y },
        resolution: { width: sink.width, height: sink.height },
        framerate: { numerator: 1, denominator: 30 }, // consistent opposite language
        v4lPath: camera.path,
        v4lResolution: cameraConfig.activeResolution,
      });
    }
  }

  /**
   * This sets up SW framegrabber
   */
  _initializeSWFrameGrabber() {
    if (this._device.hasOwnProperty('usbCameraConfig')) {
      // If a SW framegrabber is present in the configuration, immediately launch it with right format
      const camera = this.getSWFramegrabberConfig();
      if (camera) {
        this._cameraService.installAndConfigureSWFrameGrabber(camera.activeResolution, camera.activeFramerate);
      } else {
        // Take defaults ...
        this._cameraService.installAndConfigureSWFrameGrabber();
      }
    }
  }

  /**
   * This sets up the streams as requested by the camera config (from the device role configuration)
   */
  _setupStreams(forceRestart = false) {
    this._log('info', `Setting up streams, forceRestart = ${forceRestart}`);
    this._usbCameraAudioInfo = usbCameraSourceInfo(this._device);

    if (this._device.hasOwnProperty('usbCameraConfig')) {
      // removed streams
      Object.keys(this._streams)
        .filter((id) => {
          const cameraId = this._getCameraIdFromStreamId(id);
          const camera = this._device.usbCameraConfig.find(cam => cam.id === cameraId);
          return !this._shouldHaveStreamFor(camera);
        })
        .forEach(this._destroyStreamWithId.bind(this));

      // (Re)start camera streams if needed.
      this._device.usbCameraConfig
        .filter(camera => (
          // We need to stream
          this._shouldHaveStreamFor(camera)
          // The camera still exists (think about the config locked case)
          && !!this._cameraCache.$arrays.cameras.find(cam => cam.id === camera.id)
          // Create stream if it doesn't yet exist
          && (!this._streams.hasOwnProperty(camera.id)
            // Or re-create if the config has changed.
            || (!this._streams[camera.id].equals(this._createStreamFor(camera.id)))
            || forceRestart
          )))
        .forEach(camera => this._restartStreamFor(camera.id));
    }
  }

  async _restartStreamFor(cameraId) {
    const streamId = this._getStreamIdFromCameraId(cameraId);
    this._log('info', `About to restart stream with id ${streamId} for camera with id ${cameraId}`);
    await this._destroyStreamWithId(streamId);
    this._createAndStartStreamFor(cameraId, streamId);
  }

  /**
   * Destroys a stream for a certain camera (if it exists)
   * @param {*} camera
   */
  _destroyStreamWithId(streamId) {
    if (this._streams[streamId]) {
      this._log('info', `About to cleanup stream with id ${streamId}`);
      return this._streams[streamId].destroy();
    }
    this._log('info', `No stream with id ${streamId} found. Nothing to cleanup.`);
    return true;
  }

  /**
   * Creates a stream for a certain camera
   * @param {*} camera
   */
  _createAndStartStreamFor(cameraId, streamId) {
    this._log('info', `Setting up stream for usb camera: ${cameraId}`);
    const stream = this._createStreamFor(cameraId);
    if (stream) {
      this._log('verbose', `Created stream with id ${streamId} for usb camera ${cameraId}, installing callbacks and starting stream.`);
      stream.on('destroyed', () => {
        delete this._streams[streamId];
        this._log('verbose', `Successfully cleaned up camera stream with id ${streamId}`);
      });

      this._streams[streamId] = stream;
      return stream.start();
    }

    return Promise.resolve();
  }

  _createStreamFor(cameraId) {
    const cameraConfig = this._device.usbCameraConfig.find(cam => cam.id === cameraId);
    const camera = this._cameraCache.$arrays.cameras.find(cam => cam.id === cameraId);

    if (cameraConfig && camera) {
      // SW Framegrabber case
      if (isActiveSWFramegrabber(cameraConfig)) {
        this._log('info', `Creating V4L stream for usb camera: ${camera.id}`);
        return this._createContentStream();
      }

      // There may be an AGC node in between the HW audio input and the RTSP server
      const cameraSourceInfo = this._usbCameraAudioInfo.find(cam => cam.id === cameraId);

      // Regular USB camera case
      this._log('info', `Creating RTSP stream for usb camera: ${camera.id}`);
      const stream = this._cameraService.createStream({
        port: cameraConfig.port,
        path: camera.path,
        format: cameraConfig.activeFormat.name,
        resolution: cameraConfig.activeResolution,
        framerate: cameraConfig.activeFramerate,
        audioDevice: (cameraSourceInfo && cameraSourceInfo.rtspAudioDevice) ? cameraSourceInfo.rtspAudioDevice : cameraConfig.audioDevice,
        ...this._device.rtspAuthentication,
      });
      return stream;
    }
    return false;
  }

  /**
   * Update the camera cache.
   * This fetches the list of cameras from the camera services
   * and updates the cache
   */
  async _updateCameras() {
    this._log('verbose', 'Update cameras');
    let cameras = await this._cameraService.getUsbCameras();
    cameras = transformCamerasToMeteor(cameras);
    this._cameraCache.setCameras(cameras);
  }

  /**
   * First updates the camera cache
   * Then it syncs the real cameras with what is in the cloud
   */
  async _updateAndSyncCameras() {
    this._log('verbose', 'Update and syncing cameras');
    await this._updateCameras();
    await this._syncCameras();
  }

  _isMarshallCV503(camera) {
    return (camera && (camera.vendor === '04b4') && (camera.model === '00d1'));
  }

  /**
   * This syncs the camera configuration
   * consists of 2 parts:
   * - syncing the image settings
   * - syncing the image format
   */
  _syncCameraConfiguration() {
    if (this._device.usbCameras) {
      const { cameras } = this._cameraCache.$arrays;
      cameras
        .map(camera => ({
          camera,
          shadow: this._device.usbCameras.find(c => camera.id === c.id),
        }))
        .filter(({
          shadow,
        }) => !!shadow)
        .forEach((item) => {
          this._syncSettings(item);
          this._syncFormat(item);
        });
    }
  }

  /**
   * This syncs the camera image settings
   * This will find the differences between the current and requested settings
   * It will ask the camera to change the settings that are different from what's requested
   * @param {*} param0
   */
  _syncSettings({
    camera,
    shadow,
  }) {
    if (this._isMarshallCV503(camera)) {
      this._log('verbose', `Not applying camera settings on camera ${camera.device} (${camera.path}) as it is a Marshall CV503`);
      return;
    }
    this._log('verbose', `Setting camera settings on camera ${camera.device} (${camera.path})`);
    camera
      .settings
      .map(setting => ({
        setting,
        shadow: shadow.settings.find(s => s.id === setting.id),
      }))
      .filter(({
        setting,
        shadow, //eslint-disable-line
      }) => !!shadow && shadow.current !== setting.current)
      .forEach(({
        setting,
        shadow, //eslint-disable-line
      }) => {
        // The above whill filter out settings that are already correct.
        // So if we come here, it means we need to change the setting.
        this._log('verbose', `Setting control ${setting.name}(${setting.id}) to value ${shadow.current} on device ${camera.device} (${camera.path})`);
        try {
          this._cameraService.setControl(camera.path, setting.id, shadow.current);
        } catch (error) {
          // For some settings we get an input/output error.
          // There's not much we can do about this, so we just log and continue.
          this._log('error', `Error while setting control ${setting.name}(${setting.id}) to value ${shadow.current} on device ${camera.device}`, error);
        }
        setting.current = shadow.current;
      });
  }

  /**
   * Syncs the image format
   * The format is set by the RTSP server and can not be changed when the device is in use
   * So this will stop the stream and restart the RTSP server with new parameters.
   * @param {*} param0
   */
  async _syncFormat({
    camera,
    shadow,
  }) {
    const realFormat = transformCameraFormatToV4L(camera);
    const shadowFormat = transformCameraFormatToV4L(shadow);

    if (realFormat && shadowFormat && !_.isEqual(realFormat, shadowFormat)) {
      // Format differs from what's requested in the cloud
      this._log('verbose', `Setting camera format on camera ${camera.device} (${camera.path})`);
      this._log('debug', 'Shadow format', JSON.stringify(shadowFormat, null, 2));
      this._log('debug', 'Real format', JSON.stringify(realFormat, null, 2));

      Object.assign(camera, _.pick(shadow, [
        'activeFormat',
        'activeFramerate',
        'activeResolution',
      ]));

      if (CameraService.isSWCamera(camera)) {
        this._log('verbose', 'Reconfiguring SW framegrabber as requested format does not match real format');
        // CHECK CHECK CHECK: If changing the SW framegrabber resolution (admin ui) is to be supported
        // while a lecture is running, then in the browser bridge the browser window that streams out the
        // SW framegrabber content must first be stopped and may only be restarted after the two lines here
        // below have been executed (eg. after a timeout of 5 seconds or so (as services must be independent...)).
        // This is currently not implemented, you can only change resolution after stopping the lecture.
        await this._destroyStreamWithId(this._getStreamIdFromCameraId(camera.id));
        this._cameraService.installAndConfigureSWFrameGrabber(shadow.activeResolution, shadow.activeFramerate);
      } else if (this._streams.hasOwnProperty(camera.id)) {
        // Restart stream (the new gstreamer pipeline that is started will apply the right video format to the camera)
        this._log('verbose', `Camera ${camera.label}: Restarting stream as requested format does not match real format`);
        this._restartStreamFor(camera.id);
      } else {
        // In case there is no stream we need to apply the video format to the camera ourselves
        this._log('verbose', `Setting video format to ${JSON.stringify(shadowFormat, null, 2)} on device ${camera.device} (${camera.path})`);
        try {
          this._cameraService.setFormat(camera.path, shadowFormat);
        } catch (error) {
          // If it fails there's not much we can do about this, so we just log and continue.
          this._log('error', `Error while setting video format to value ${JSON.stringify(shadowFormat, null, 2)} on device ${camera.device}`);
          this._log('error', error);
        }
      }
    }
  }

  /**
   * Sync the cameras with what's in the cloud.
   * This will first sync the camera configuration (image format & settings)
   * and then upsert changes in the cameras (new or removed ones) to the cloud.
   */
  _syncCameras() {
    this._log('info', 'Syncing cameras');
    this._syncCameraConfiguration();

    const { cameras } = this._cameraCache.$arrays;

    const hasChanged = !Array.isArray(this._device.usbCamerasLatest)
      || (cameras.length !== this._device.usbCamerasLatest.length)
      || (_.difference(cameras.map(c => c.id), this._device.usbCamerasLatest.map(c => c.id)).length > 0);

    if (
      // If there is a change ...
      hasChanged
      // Or if multiple cameras claim the same port or there are cameras without ports
      // XXX: This check is a temporary workaround for a bug that existed where multiple cameras got the same port
      || new Set(this._device.usbCamerasLatest.map(parsePort)).size
        !== this._device.usbCamerasLatest.length
    ) {
      const assignedPorts = [];
      const oldAssignedPorts = this._device.usbCameras.map(parsePort);
      // Ensure consistency of camera ports
      // make every camera consistent
      cameras.forEach((camera) => {
        // try to see if we already know this camera
        const shadow = this._device.usbCameras.find(cam => cam.id === camera.id);
        if (shadow && !assignedPorts.includes(shadow.port)) {
          camera.port = shadow.port;
        } else {
          // if we cannot keep a previous port, assign the next available one
          // don't use any port that was already in use in the past or will be in use in the future
          camera.port = getFirstAvailableRTSPPort([
            ...assignedPorts,
            ...oldAssignedPorts,
          ]);
        }
        // make sure we remember the port as unavailable for other cameras
        assignedPorts.push(camera.port);
      });

      this._log('info', `Change in USB cameras with server. New cameras: ${JSON.stringify(cameras, null, 2)}. Old cameras: ${JSON.stringify(this._device.usbCameras, null, 2)}.`, { serviceLog: true });
      this._device.usbCamerasLatest = cameras;
      this._setupStreams();

      return this._meteorClient.updateDevice({
        _id: this._device._id,
        usbCamerasLatest: cameras,
      });
    }

    return Promise.resolve();
  }

  _onDestroy() {
    this._cameraService.destroy();
    this._cameraCache.destroy();
    delete this._cameraCache;
    delete this._cameraService;
    delete this._streams;
    return super._onDestroy();
  }
}

export default UsbCameraBridge;
