import * as Rx from 'rxjs';
import { throttle } from 'rxjs/operators';
import * as Path from 'path';
import { exec, execSync } from 'child_process';
import * as v4l2camera from '@barco/v4l2camera';
import * as usb from 'usb';
import { promises as fs } from 'fs';

import { Service } from '../../../services/service';
import RTSPStream from '../streaming/rtspStream';
import V4LShmStream from '../streaming/v4lShmStream';
import V4LScreenStream from '../streaming/v4lScreenStream';
import SynchronousScheduler from '../../../services/schedulers/synchronousScheduler';
import { SCRIPTS_DIR } from '../../../files';
import { waitMaximum } from '../../../utils/utils';

// SW Framegrabber definitions
const SW_FRMGRBR_VENDOR = 'Barco';
const SW_FRMGRBR_MODEL = '001';
// Allow to specify name with and without required vender/model id
const nameSanitizeRegex = new RegExp(`"| \\(${SW_FRMGRBR_VENDOR}:${SW_FRMGRBR_MODEL}\\)"$`, 'ig');
const sanitizeSoftwareCameraName = name => `"${name.replace(nameSanitizeRegex, '')} (${SW_FRMGRBR_VENDOR}:${SW_FRMGRBR_MODEL})"`;

const SW_FRMGRBR_NAME = 'Teacher Content';
const SW_FRMGRBR_ID = '00ABCDEF';
const SW_FRMGRBR_DEVICE_NUMBER = 99; // use /dev/video99
const SW_FRMGRBR_DEFAULT_RES = { width: 960, height: 540 };
const SW_FRMGRBR_DEFAULT_FPS = { numerator: 1, denominator: 30 };
// Formats below can be chosen in the admin UI for the SW framegrabber.
// They should all have 30fps (unless we change the v4l pipeline) but
// as the SW framegrabber is about fast moving content, keep 30fps.
// List can be extended if necessary... All formats here are 16/9 AR.
const SW_FRMGRBR_FORMATS = [
  {
    formatName: 'YU12',
    format: 842093913,
    width: 320,
    height: 180,
    interval: { numerator: 1, denominator: 30 },
  },
  {
    formatName: 'YU12',
    format: 842093913,
    width: 640,
    height: 360,
    interval: { numerator: 1, denominator: 30 },
  },
  {
    formatName: 'YU12',
    format: 842093913,
    width: 960,
    height: 540,
    interval: { numerator: 1, denominator: 30 },
  },
  {
    formatName: 'YU12',
    format: 842093913,
    width: 1200,
    height: 896,
    interval: { numerator: 1, denominator: 30 },
  },
  {
    formatName: 'YU12',
    format: 842093913,
    width: 1280,
    height: 720,
    interval: { numerator: 1, denominator: 30 },
  },
];

const isCaptureDevice = (path) => {
  try {
    const cam = new v4l2camera.Camera(path);

    if (!Array.isArray(cam.deviceCapabilities)) {
      // Support for older versions (< debian 10)
      return true;
    }

    const requiredCaps = [
      ['Streaming'], // &&
      [
        'Video Capture', // ||
        'Video Capture Multiplanar',
      ],
    ];

    const success = requiredCaps.every(caps => caps.some(cap => cam.deviceCapabilities.includes(cap)));
    return success;
  } catch (error) {
    return false;
  }
};

/**
 * This service allows to query the list of attached usb webcams
 * Furthermore, when started, it can watch and notify changes in attached webcams
 */
class UsbCamClient extends Service {
  constructor() {
    super({
      id: false,
    });
  }

  static isSWCamera(camera) {
    return ((camera.vendor || camera.vendor_id) === SW_FRMGRBR_VENDOR && (camera.model || camera.model_id) === SW_FRMGRBR_MODEL);
  }

  static async listDevices() {
    const files = await fs.readdir('/dev');
    return files.filter(file => file.startsWith('video')).map(file => `/dev/${file}`);
  }

  static async fetchDevices(extended) {
    let files = await UsbCamClient.listDevices();
    if (extended) {
      // We don't want to filter them out for non-extended since the check triggers camera activation (e.g. logitech rally)
      // Not very nice, but alas.
      files = files.filter(isCaptureDevice);
    }
    return Promise.all(files.map(device => UsbCamClient.fetchDeviceInfo(device, extended)));
  }

  static fetchDeviceInfo(device, extended) {
    const getCameraPromise = new Promise(((resolve, reject) => {
      const cmd = `${Path.join(SCRIPTS_DIR, 'video.sh')} -d ${device}${extended ? ' -e' : ''}`;
      exec(cmd, {
        timeout: 9000,
      }, (error, stdout, stderr) => {
        if (error != null) {
          reject(error);
        } else if (stderr && stderr != null) {
          reject(new Error(stderr));
        } else {
          resolve(JSON.parse(stdout.toString()));
        }
      });
    }));
    return waitMaximum(getCameraPromise, 10000);
  }

  get mixins() {
    return [
      ...super.mixins,
      SynchronousScheduler,
    ];
  }

  _onStart() {
    this._startListeners();
    return super._onStart();
  }

  _onStop() {
    this._stopListeners();
    delete this._usbCameras;
    delete this._usbCamerasExtended;
    return super._onStop();
  }

  _startListeners() {
    this._log('verbose', 'Start listening to usb device changes');

    const usbAttach = Rx.fromEvent(usb, 'attach');
    const usbDetach = Rx.fromEvent(usb, 'detach');
    const interval = Rx.interval(15000);

    this.listener = Rx.merge(usbAttach, usbDetach, interval)
      .pipe(throttle(() => Rx.interval(1000)))
      .subscribe(() => {
        this.scheduler.schedule(() => this._checkForChanges());
      });
  }

  _stopListeners() {
    this._log('verbose', 'Stop listening to usb device changes');

    this.listener.unsubscribe();
    delete this.listener;
  }

  /**
   * Gets a specific control (image setting)
   * @param {*} cameraId
   * @param {*} controlId
   */
  getControl(cameraId, controlId) {
    const camera = new v4l2camera.Camera(cameraId);
    if (camera) {
      return camera.controlGet(controlId);
    }
    throw new Error('Camera not found');
  }

  /**
   * Sets a certain control (image setting)
   * @param {*} cameraId
   * @param {*} controlId
   * @param {*} value
   */
  setControl(cameraId, controlId, value) {
    const camera = new v4l2camera.Camera(cameraId);
    if (camera) {
      camera.controlSet(controlId, value);
    }
  }

  /**
   * Gets the current image format for the given camera
   * @param {*} cameraId
   */
  getFormat(cameraId) {
    const camera = new v4l2camera.Camera(cameraId);
    if (camera) {
      return camera.configGet();
    }
    throw new Error('Camera not found');
  }

  /**
   * Sets the current image format for the given camera
   */
  setFormat(cameraId, format) {
    const camera = new v4l2camera.Camera(cameraId);
    if (camera) {
      camera.configSet(format);
    }
  }

  /**
   * Returns the list of usb cameras.
   * This returns a cached version if available
   */
  getUsbCameras() {
    if (!this._usbCamerasExtended) {
      return this._fetchUsbCameras();
    }
    return Promise.resolve(this._usbCamerasExtended);
  }

  get SWFrameGrabberId() {
    return SW_FRMGRBR_ID;
  }

  /**
   * Create a stream for a camera.
   * @param {*} param0
   */
  createStream({
    path,
    format,
    resolution,
    framerate,
    url = 'cam',
    port,
    audioDevice,
    username,
    password,
  }) {
    if (!path || !format || !resolution || !framerate) {
      this._log('info', `Camera ${path} doesn't have a valid configuration, ignoring...`);
      return undefined;
    }

    // Some raw formats have multiple names and are identified different between video4linux and gstreamer.
    // This maps the v4l format with the gstreamer format.
    const formatMapping = {
      YUYV: 'YUY2',
    };

    format = formatMapping[format] || format;

    // List of supported raw formats by gstreamer
    const rawFormats = ['RGB16', 'BGR', 'RGB', 'GRAY8', 'GRAY16_LE', 'GRAY16_BE', 'YVU9', 'YV12', 'YUY12', 'YUY2', 'YVYU', 'UYVY', 'Y42B', 'Y41B', 'YUV9', 'NV12_64Z32', 'NV24', 'NV61', 'NV16', 'NV21', 'NV12', 'I420', 'BGRA', 'BGRx', 'ARGB', 'xRGB', 'BGR15', 'RGB1'];

    let mimeType = null;

    // Capabilities of the stream to pass to the rtsp server
    const capabilities = {
      framerate: `${framerate.denominator}/${framerate.numerator}`,
      width: resolution.width,
      height: resolution.height,
    };

    // We have to pass a mimetype.
    // This list maps the format we get from v4l to a gstreamer type.
    switch (format) {
    case 'MJPG':
      mimeType = 'image/jpeg';
      break;
    case 'H263':
      mimeType = 'video/x-h263';
      break;
    case 'WVC1':
    case 'WMV3':
      mimeType = 'video/x-wmv';
      break;
    case 'H264':
      mimeType = 'video/x-h264';
      break;
    default:
      if (rawFormats.includes(format)) {
        mimeType = 'video/x-raw';
        capabilities.format = format;
      }
      break;
    }

    // The above list is non-exhaustive
    // Because we don't know the entire mapping and the rtsp server doesn't support all formats
    if (!mimeType) {
      this._log('info', `Camera ${path} doesn't have a supported configuration, ignoring...`);
      return undefined;
    }

    const stream = new RTSPStream({
      port,
      url,
      path,
      params: `-v "${mimeType},${Object.keys(capabilities).map(cap => `${cap}=${capabilities[cap]}`).join(',')}"`,
      paSrcDevice: audioDevice,
      username,
      password,
    });

    return stream;
  }

  /**
   * Create a stream for a camera from shared memory.
   * @param {*} param0
   */
  createV4LShmStream({
    path,
    resolution,
    framerate,
    v4lPath,
    v4lResolution,
  }) {
    if (!path || !resolution || !framerate || !v4lPath || !v4lResolution) {
      this._log('info', `Camera ${v4lPath} doesn't have a valid configuration, ignoring...`);
      return undefined;
    }

    const stream = new V4LShmStream({
      path,
      resolution,
      framerate,
      v4lPath,
      v4lResolution,
    });

    return stream;
  }

  /**
   * Create a stream for a camera from x screen.
   * @param {*} param0
   */
  createV4LScreenStream({
    origin,
    resolution,
    framerate,
    v4lPath,
    v4lResolution,
  }) {
    if (!origin || !resolution || !framerate || !v4lPath || !v4lResolution) {
      this._log('info', `Camera ${v4lPath} doesn't have a valid configuration, ignoring...`);
      return undefined;
    }

    const stream = new V4LScreenStream({
      origin,
      resolution,
      framerate,
      v4lPath,
      v4lResolution,
    });

    return stream;
  }

  /**
   * Return the device id for a camera
   * This is a concatenation of the video device path, vendor id and model id.
   * @param {*} camera
   */
  _deviceIdForUsbCamera(camera) {
    // old id: return `${camera.dev}_${camera.vendor_id}_${camera.model_id}`
    return camera.serial;
  }

  /**
   * The label we get from the getVideoDevices module replaces spaces with underscores.
   * This replaces the underscore back to spaces
   * Returns 'Camera' as default label if there is none
   * @param {*} label
   */
  _parseLabel(label) {
    return (label ? label.replace(/_/g, ' ') : 'Camera');
  }

  async _checkForChanges() {
    // Fetching the extended camera details triggers camera activation on some cameras (e.g. logitech rally)
    // If these are PTZ cameras which make a lot of noise (also looking at you logitech)
    // then our customers tell us this is experienced as rather annoying
    // So we first do a simpler check which doesn't activate the cameras but still gives us enough info
    // to know if something changed.
    let usbCameras = await UsbCamClient.fetchDevices();
    // since all info in here doesn't change if usb cameras don't change, we can just stringify this whole thing to check for changes
    usbCameras = usbCameras.join(';');
    if (this._usbCameras !== usbCameras) {
      this._usbCameras = usbCameras;
      this._usbDevicesChanged();
    }
  }

  /**
   * Fetch the list of usb cameras (ignoring cache)
   * This fetches the list of all video devices and adds information from video 4 linux
   */
  async _fetchUsbCameras() {
    this._log('silly', 'Fetching list of usb cameras with extended info');
    const cameras = await UsbCamClient.fetchDevices(true);

    return cameras.map((camera) => {
      // Check whether it is a real USB camera or the SW framegrabber...
      const isRealCam = !UsbCamClient.isSWCamera(camera);
      let currentFormat;
      let settings;
      let formats;
      if (isRealCam) {
        // Get the formats and controls info via lib
        const cameraInfo = new v4l2camera.Camera(camera.dev);
        currentFormat = cameraInfo.configGet();
        // GREY formats are usually additional infrared eyes for AI purposes.
        // We filter these out, since they are not useful for us.
        formats = cameraInfo.formats.filter(({ formatName: name }) => name !== '' && name !== 'GREY');
        settings = cameraInfo.controls;
        settings.forEach((setting) => {
          setting.current = cameraInfo.controlGet(setting.id);
        });
      } else {
        // SW Framegrabber: Hardcoded formats, get current resolution
        // from the data parsed from the output of the video.sh script.
        // You cannot use the lib for the SW framegrabber.
        const regex = new RegExp('\\b([0-9]{1,4})\\s\\x\\s([0-9]{1,4})');
        const resolution = regex.exec(camera.resolution);
        currentFormat = {
          formatName: 'YU12',
          format: 842093913,
          width: parseInt(resolution[1], 10),
          height: parseInt(resolution[2], 10),
          interval: { numerator: 1, denominator: 30 },
        };
        formats = SW_FRMGRBR_FORMATS;
        settings = [];
      }

      // Delete the resolution info - resolution info will be part of formats
      delete camera.resolution;

      return {
        id: this._deviceIdForUsbCamera(camera),
        serial: camera.serial,
        path: camera.dev,
        vendor: camera.vendor_id,
        model: camera.model_id,
        formats,
        settings,
        device: this._parseLabel(camera.label),
        currentFormat,
      };
    }).filter(camera => camera.formats.length > 0);
  }

  /**
   * We listen to changes on usb devices, however, this can include non-camera devices.
   * Therefore we have to fetch the list of cameras and only emit events when
   * there is an actual change in usb camera devices
   */
  async _usbDevicesChanged() {
    const isSameDevice = (a, b) => (a.dev === b.dev && a.serial === b.serial);

    let cameras;
    try {
      cameras = await this._fetchUsbCameras();
    } catch (error) {
      this._log('error', 'Error while fetching USB cameras', error);
      throw error;
    }

    const oldCameras = this._usbCamerasExtended ? JSON.parse(JSON.stringify(this._usbCamerasExtended)) : [];
    this._usbCamerasExtended = cameras;

    const newCameras = this._usbCamerasExtended.filter(camera => !oldCameras
      .find(oldCamera => isSameDevice(camera, oldCamera)));
    const removedCameras = oldCameras.filter(oldCamera => !this._usbCamerasExtended
      .find(camera => isSameDevice(camera, oldCamera)));

    if (newCameras.length) {
      this.emit('usbCamerasAdded', newCameras);
    }

    if (removedCameras.length) {
      this.emit('usbCamerasRemoved', removedCameras);
    }

    if (newCameras.length || removedCameras.length) {
      this.emit('usbCamerasChanged', {
        cameras: this._usbCamerasExtended,
        oldCameras,
        newCameras,
        removedCameras,
      });
    }
  }

  installAndConfigureSWFrameGrabber(
    res = SW_FRMGRBR_DEFAULT_RES,
    fps = SW_FRMGRBR_DEFAULT_FPS,
    name = SW_FRMGRBR_NAME,
    nbr = SW_FRMGRBR_DEVICE_NUMBER,
  ) {
    name = sanitizeSoftwareCameraName(name);
    this._log('verbose', `Installing and configuring SW framegrabber with resolution ${res.width}x${res.height} @ ${fps.denominator}/${fps.numerator} fps`);
    this._unloadSWFrameGrabber();
    this._loadSWFrameGrabber(res, fps, name, nbr);
  }

  _loadSWFrameGrabber(res, fps, name, nbr) {
    this._log('verbose', `Loading and configuring SW framegrabber with resolution ${res.width}x${res.height} @ ${fps.denominator}/${fps.numerator} fps`);
    try {
      execSync(`sudo modprobe v4l2loopback devices=1 exclusive_caps=1 card_label=${name} video_nr=${nbr}`);
      execSync(`v4l2loopback-ctl set-caps "video/x-raw, framerate=${fps.denominator}/${fps.numerator}, width=${res.width}, height=${res.height}, format=I420" /dev/video${nbr}`);
    } catch (error) {
      this._log('error', 'Could not load SW framegrabber', error);
    }
  }

  _unloadSWFrameGrabber() {
    this._log('verbose', 'Unloading SW framegrabber (if any)');
    try {
      execSync('sudo modprobe -r v4l2loopback');
    } catch (error) {
      this._log('error', 'Could not unload sw framegrabber', error);
    }
  }

  _onDestroy() {
    this._unloadSWFrameGrabber();
    return super._onDestroy();
  }
}

export default UsbCamClient;
