import * as chai from 'chai';
import * as sinon from 'sinon';
import * as sinonChai from 'sinon-chai';
import * as usb from 'usb';
import * as v4l2camera from '@barco/v4l2camera';
import {
  registerAll,
  unregisterAll,
} from '../../../utils/test-helpers/fillServiceRegistry';
import UsbCamClient from './service';

const { expect } = chai;
chai.use(sinonChai);

describe('Cameras service', () => {
  before(() => {
    registerAll();
  });

  after(() => {
    unregisterAll();
  });

  describe('Camera change detection', () => {
    let clock;
    let usbAddListenerStub;
    let listDevicesStub;
    let fetchDeviceInfoStub;
    let constructorStub;
    const usbOnMethods = {};
    let usbCamClient;
    let usbCamerasChangedSpy;
    const camera1 = {
      currentFormat: {},
      device: 'Camera',
      formats: ['format1', 'format2'],
      path: '/dev/video1',
      id: 'TEST_SERIAL_/dev/video1',
      serial: 'TEST_SERIAL_/dev/video1',
      settings: [],
      vendor: 'TEST_VENDOR',
      model: 'TEST_MODEL',
    };
    const camera2 = {
      currentFormat: {},
      device: 'Camera',
      formats: ['format1', 'format2'],
      path: '/dev/video2',
      id: 'TEST_SERIAL_/dev/video2',
      serial: 'TEST_SERIAL_/dev/video2',
      settings: [],
      vendor: 'TEST_VENDOR',
      model: 'TEST_MODEL',
    };

    beforeEach(() => {
      clock = sinon.useFakeTimers(Date.now());
      usbAddListenerStub = sinon.stub(usb, 'addListener').callsFake((name, fn) => {
        usbOnMethods[name] = fn;
      });
      listDevicesStub = sinon.stub(UsbCamClient, 'listDevices');
      fetchDeviceInfoStub = sinon.stub(UsbCamClient, 'fetchDeviceInfo')
        .callsFake((device) => new Promise((resolve) => {
          resolve({
            dev: device,
            resolution: '',
            serial: `TEST_SERIAL_${device}`,
            vendor_id: 'TEST_VENDOR',
            model_id: 'TEST_MODEL',
            path: [undefined],
          });
        }));
      const stubbedCamera = sinon.createStubInstance(v4l2camera.Camera);

      stubbedCamera.deviceCapabilities = ['Streaming', 'Video Capture'];
      stubbedCamera.formats = ['format1', 'format2'];
      stubbedCamera.controls = [];
      stubbedCamera.configGet.returns({});

      constructorStub = sinon.stub(v4l2camera, 'Camera').returns(stubbedCamera);

      usbCamClient = new UsbCamClient();

      usbCamerasChangedSpy = sinon.spy();
      usbCamClient.on('usbCamerasChanged', usbCamerasChangedSpy);
    });

    afterEach(() => {
      clock.restore();
      usbAddListenerStub.restore();
      listDevicesStub.restore();
      fetchDeviceInfoStub.restore();
      constructorStub.restore();
    });

    it('should find the cameras', async () => {
      await usbCamClient.start();

      listDevicesStub.returns(['/dev/video1', '/dev/video2']);
      usbOnMethods.attach();

      await clock.tickAsync(1);
      await usbCamClient.stop();

      expect(usbCamerasChangedSpy).to.be.calledOnce;
      expect(usbCamerasChangedSpy.firstCall).to.be.calledWithExactly(
        {
          cameras: [camera1, camera2],
          oldCameras: [],
          newCameras: [camera1, camera2],
          removedCameras: [],
        },
      );
    });

    it('should find the cameras after a quick disconnect and reconnect', async () => {
      await usbCamClient.start();

      // list devices is called 2 times in the cam service when the cameras change
      listDevicesStub.onFirstCall().returns(['/dev/video1', '/dev/video2']);
      listDevicesStub.onSecondCall().returns(['/dev/video1', '/dev/video2']);
      listDevicesStub.onThirdCall().returns(['/dev/video1']);
      listDevicesStub.onCall(3).returns(['/dev/video1']);
      listDevicesStub.onCall(4).returns(['/dev/video1', '/dev/video2']);
      listDevicesStub.onCall(5).returns(['/dev/video1', '/dev/video2']);

      usbOnMethods.attach();

      await clock.tickAsync(1010);

      usbOnMethods.detach();
      usbOnMethods.attach();

      await clock.tickAsync(15010);

      await usbCamClient.stop();

      expect(usbCamerasChangedSpy).to.be.calledThrice;
      expect(usbCamerasChangedSpy.firstCall).to.be.calledWithExactly(
        {
          cameras: [camera1, camera2],
          oldCameras: [],
          newCameras: [camera1, camera2],
          removedCameras: [],
        },
      );
      expect(usbCamerasChangedSpy.secondCall).to.be.calledWithExactly(
        {
          cameras: [camera1],
          oldCameras: [camera1, camera2],
          newCameras: [],
          removedCameras: [camera2],
        },
      );
      expect(usbCamerasChangedSpy.thirdCall).to.be.calledWithExactly(
        {
          cameras: [camera1, camera2],
          oldCameras: [camera1],
          newCameras: [camera2],
          removedCameras: [],
        },
      );
    });
  });
});
