import { usbCameraSettingType } from '../../../services/ddp/constants';

// Some utility methods for mapping between the camera service output and the cloud data model

/**
 * Transforms the list of video formats between v4l service and meteor data model.
 * v4l returns a list like:
 * [
 *  {
 *    "format":123,
 *    "formatName":"MJPG",
 *    "height":1080
 *    "width":1920,
 *    "interval":{
 *      "numerator":1,
 *      "denominator":30
 *    }
 *  },
 *  {
 *    "format":123,
 *    "formatName":"MJPG",
 *    "height":1080
 *    "width":1920,
 *    "interval":{
 *      "numerator":1,
 *      "denominator":15
 *    }
 *  }
 * ]
 *
 * This method de-duplicates this and returns
 * [
 *  {
 *    "id":123
 *    "name":"MJPG",
 *    "resolutions":[
 *      {
 *        "width":1920,
 *        "height":1080,
 *        "framerates":[
 *          {
 *            "numerator":1,
 *            "denominator":30
 *          },
 *          {
 *            "numerator":1,
 *            "denominator":15
 *          }
 *        ]
 *      }
 *    ]
 *  }
 * ]
 * @param {*} formats
 */
function transformFormatsToMeteor(formats) {
  const transformedFormats = [];
  formats.forEach((element) => {
    // If a transformed format is already existing with the same format, extend that one
    const formatWithSameId = transformedFormats.find(transformedFormat => transformedFormat.id === element.format);
    if (formatWithSameId) {
      // If a transformed format is already present with the same resolution, extend that one
      const formatWithSameIdAndRes = formatWithSameId.resolutions
        .find(resolution => (
          (resolution.width === element.width)
          && (resolution.height === element.height)
        ));
      if (formatWithSameIdAndRes) {
        // If no transformed format is already present with the same resolution and framerate extend
        const hasFramerate = formatWithSameIdAndRes.framerates
          .find(framerate => (
            (framerate.numerator === element.interval.numerator)
            && (framerate.denominator === element.interval.denominator)
          ));

        if (!hasFramerate) {
          const framerate = {
            numerator: element.interval.numerator,
            denominator: element.interval.denominator,
          };
          formatWithSameIdAndRes.framerates.push(framerate);
        }
      } else {
        // Else add a new resolution to the list
        const resolution = {
          width: element.width,
          height: element.height,
          framerates: [{
            numerator: element.interval.numerator,
            denominator: element.interval.denominator,
          }],
        };
        formatWithSameId.resolutions.push(resolution);
      }
    } else {
      // Else add a new format
      const transformedFormat = {};
      transformedFormat.id = element.format;
      transformedFormat.name = element.formatName;
      transformedFormat.resolutions = [{
        width: element.width,
        height: element.height,
        framerates: [{
          numerator: element.interval.numerator,
          denominator: element.interval.denominator,
        }],
      }];
      transformedFormats.push(transformedFormat);
    }
  });
  return (transformedFormats);
}

/**
 * In meteor we only want writable & enabled controls
 * This strips the flags and filters out readonly & disabled controls
 * @param {*} settings
 * @param {*} cameraInfo
 */
function transformSettingsToMeteor(settings) {
  // Cameras have following possible setting types, but we do not support all of them (due to lack of examples):
  // "invalid","int","bool","menu","int64","class""string","bitmask","int_menu",
  const allowedSettingTypes = Object.values(usbCameraSettingType);

  // First filter out the controls that are disabled or that are read only
  const transformedSettings = settings.filter(setting => (
    allowedSettingTypes.includes(setting.type)
    && setting.flags.disabled === false
    && setting.flags.readOnly === false));

  // Remove the flags from the objects in the list
  transformedSettings.forEach((setting) => {
    delete setting.flags;
  });

  return (transformedSettings);
}

/**
 * Transforms a video 4 linux format to a separate format/resolution/framerate object
 * Inverse of transformCameraFormatToV4L
 * @param {*} format
 */
function transformCameraFormatToMeteor(format) {
  const transformed = {};

  if (format) {
    transformed.activeFormat = {
      id: format.format,
      name: format.formatName,
    };
    transformed.activeResolution = {
      width: format.width,
      height: format.height,
    };
    transformed.activeFramerate = format.interval;
  }

  return transformed;
}

/**
 * Transforms a meteor camera's active format/resolution/framerate to a video 4 linux format.
 * Inverse of transformCameraFormatToMeteor
 * @param {*} camera
 */
function transformCameraFormatToV4L(camera) {
  if (camera.activeResolution && camera.activeFormat && camera.activeFramerate) {
    return {
      format: parseInt(camera.activeFormat.id, 10),
      formatName: camera.activeFormat.name,
      width: camera.activeResolution.width,
      height: camera.activeResolution.height,
      interval: camera.activeFramerate,
    };
  }
  return null;
}

/**
 * Transforms camera list from the v4l service to blue data model
 * @param {*} cameras
 */
function transformCamerasToMeteor(cameras) {
  return cameras.map((camera) => {
    camera.formats = transformFormatsToMeteor(camera.formats);
    camera.settings = transformSettingsToMeteor(camera.settings);
    Object.assign(camera, transformCameraFormatToMeteor(camera.currentFormat));
    delete camera.currentFormat;
    return camera;
  });
}

export {
  transformCameraFormatToMeteor,
  transformCameraFormatToV4L,
  transformFormatsToMeteor,
  transformSettingsToMeteor,
  transformCamerasToMeteor,
};
