/**
 * (C) 2015 Copyright BARCO NV
 * All rights reserved. Reproduction in whole or in part is
 * prohibited without the written consent of the copyright owner.
 *
 * Part of Barco WeConnect
 */
import * as Rx from 'rxjs';
import { debounceTime } from 'rxjs/operators';
import * as path from 'path';
import * as URL from 'url';
import * as _ from 'lodash';
import * as isDev from 'electron-is-dev';
import { execSync } from 'child_process';

import Random from '../../../utils/random';
import { rebootSystem } from '../../../utils/utils';
import { RESOURCES_DIR, SCRIPTS_DIR } from '../../../files';
import {
  sourceAddedBy,
  sourceType,
  sinkType,
  compositionType,
  commands,
} from '../../../services/ddp/constants';
import SynchronousBridge from '../synchronousBridge';

import bridgeUtils from './utils';
import MeteorCacheManager from './meteorCacheManager';
import CngCacheManager from './cngCacheManager';
import PreviewBridge from './previewBridge';

import AuthBridge from './authBridge';
import CinergyClient from './service';
import serviceRegistry from '../../../services/serviceRegistry';
import { state as serviceState } from '../../../services/service';

const cngShowTitlePolicy = ['never', 'timedAfterFadeIn', 'always'];

const state = {
  UNINITIALIZED: 0,
  INITIALIZING: 1,
  INITIALIZED: 2,
};

const enableBrowserDebug = !!process.env.BROWSER_DEBUG;

function compareVersion(a, b) {
  let aNumber = Number.parseInt(a, 10);
  let bNumber = Number.parseInt(b, 10);
  let index = 0;

  while ((aNumber !== undefined) && (bNumber !== undefined)) {
    if (bNumber > aNumber) { return -1; }
    if (bNumber < aNumber) { return 1; }

    index = a.indexOf('.');
    if (index >= 0) {
      a = a.substring(index + 1);
      aNumber = Number.parseInt(a, 10);
    } else { aNumber = undefined; }

    index = b.indexOf('.');
    if (index >= 0) {
      b = b.substring(index + 1);
      bNumber = parseInt(b, 10);
    } else { bNumber = undefined; }
  }
  return 0;
}

/**
 * CinergyBridge implements the Bridge pattern to act as a translator between the NRC API and the Meteor API.
 * An instance of CinergyBridge manages exactly one device/NRC.
 * It caches both the data from the NRC and Meteor related to this device and only this device.
 *
 * Because of the asynchronous nature of both API's we need to change the cached data optimistically
 * and believe the cache is the source of truth. For this to work, we need to believe all NRC api calls are successful
 * If we do not apply changes to the cache optimistically, we can (and will) run into race conditions.
 *
 *  The flow of the bridge is as follows:
 * - The bridge creates an NRC API client (cngClient) instance and subscribes for device data from the NRC
 * - When all initial data is received AND we are connected to the NRC, we start an initialization phase.
 * - The initialization phase will force the initial set of meteor data (sinks, compositions, layouts, ...)
 *   onto the NRC and merge the sources
 * - If new data arrives (or data is changed) during the initialization phase,
 *    we re-run the initialization phase until stable.
 * - Once initialized, we listen to incoming events from Meteor and the NRC (only for sources)
 *    and apply the changes accordingly
 *
 * The class implements an FSM to know whether it is uninitialized, initializing, initialized.
 *
 * If the connection with the CNG gets lost, the instance goes back into uninitialized state.
 * As soon as the connection returns, we assume the entire state of the device can be changed
 * (since this can in fact be the case)
 * and therefore we restart from the initialization procedure.
 *
 * Note that Meteor works with Mongo Identifiers and CNG with a range of identifiers depending on the type of object.
 * (e.g. compositorId for compositions, channelId/deviceId for sources, ...)
 * And not all of these are in our control (e.g. compositorId 'screen' is fixed)
 *
 * Therefore depending on the object we save the mongo id in a field that's available
 *  and that way we can map the objects.
 * For layouts this is the title, for sources this is the deviceId
 *
 *
 * Also, the datastructure of the Meteor and CNG objects do not always map 1-on-1.
 *  Therefore we have some translation methods
 */
class CinergyBridge extends SynchronousBridge {
  /*= =====================================
     * Object construction - destruction
     *====================================== */
  /**
     * Construct a new meteor bridge instance
     * @param device - The meteor device object for this bridge to manage
     * @param meteorClient - The meteor client (to attach listeners to)
     * @param previewServer - A reference to the preview server to add new preview providers
     *
     * This will:
     * - Set up initial datastructures
     * - Create a preview provider manager for this device
     * - Create an Cng Client instance
     * - Bootstrap the data caches
     * - Install listeners to both the Cng Client API and Meteor Client API
     * - Subscribe to the device-related data
     */
  constructor(options) {
    super(options);

    const params = {};
    Object.keys(params).forEach((param) => {
      if (params[param] === undefined) throw new Error(`Please provide a valid ${param}`);
      this[`_${param}`] = params[param];
    });

    this._state = state.UNINITIALIZED;

    // Create a new cng client instance for this device
    this._cngClient = new CinergyClient({
      ipAddress: '127.0.0.1', // XXX: as long as purple runs on same device, connect to localhost this._device.privateIp,
      port: 5010,
      useTLS: false,
      autoReconnect: true,
    });

    // Set up a new preview provider manager
    this._previewMgr = new PreviewBridge(Object.assign(options, {
      cngBridge: this,
    }));

    this._authBridge = new AuthBridge(options);

    this.cache = new MeteorCacheManager(this);

    // Create the CNG cache
    this._cngData = new CngCacheManager(this._cngClient);

    // Install listeners to CNG API
    this._installCngListeners();

    this._cinergyNotReadyCounter = 0; // XXX: dirty ronny hack

    this._eventProxy = serviceRegistry.get('eventProxy');
    this._eventProxy.register(this, ['initialized']);
  }

  get supportedCompositionTypes() {
    return [compositionType.CINERGY];
  }

  get supportedRPCCommands() {
    return [commands.DISCONNECT_BYOD_CLIENTS];
  }

  get isServiceRequired() {
    return (
      super.isServiceRequired
            && this._device.configured
            && this._meteorClient._ddpClient.collections.compositions //eslint-disable-line
            && Object.values(this._meteorClient._ddpClient.collections.compositions) //eslint-disable-line
              .filter(comp => comp.type === compositionType.CINERGY).length > 0
    );
  }

  set device(device) {
    super.device = device;
    this._previewMgr.device = device;
    this._authBridge.device = device;
  }

  _runCleanupScript() {
    this._log('verbose', 'Starting unused resources cleanup script');
    try {
      execSync(`sudo ${path.join(SCRIPTS_DIR, 'cleanUnusedResources.sh')}`);
    } catch (error) {
      this._log('error', `Failed to execute cleanup script, error = ${error.message}`);
    }
  }

  async _onStart() {
    await super._onStart();
    // Before starting Cinergy, cleanup unused resources (if any) from previous runs
    this._runCleanupScript();
    try {
      await this._cngClient.startServices();
    } catch (e) {
      this._log('error', 'Could not start cinergy services', e);
    }

    return Promise.all([
      this._cngClient.start({ cancelToken: this.getCancelToken(serviceState.STARTED) }),
      this._previewMgr.start({ cancelToken: this.getCancelToken(serviceState.STARTED) }),
      this._authBridge.start({ cancelToken: this.getCancelToken(serviceState.STARTED) }),
    ]);
  }

  async _onStop() {
    this._clearAllTimers();

    await Promise.all([
      this._cngClient.stop(),
      this._cngData.stop(),
      this._previewMgr.stop(),
      this._authBridge.stop(),
    ]);

    if (!this.isServiceRequired) {
      // We might want to skip this in the case we restart the bridge after a short cloud disconnect
      // In that case the agent will restart the bridges, but we likely still need cinergy
      // with an unchanged sink/comp configuration.
      // So we don't want to stop it, since it takes some seconds to start.
      try {
        await this._cngClient.stopServices();
      } catch (e) {
        this._log('error', 'Could not stop cinergy services', e);
      }
    }

    await super._onStop();
  }

  _clearAllTimers() {
    this._clearInitializationTimer();
    this._clearShowTimer();
  }

  _clearInitializationTimer() {
    if (this._initializationTimer) {
      clearTimeout(this._initializationTimer);
      delete this._initializationTimer;
    }
  }

  _clearShowTimer() {
    if (this._showTimer) {
      clearTimeout(this._showTimer);
      delete this._showTimer;
    }
  }

  _installCngListeners() {
    this._log('info', 'Installing CNG listeners');

    // If the connection with the CNG is lost, we go back to 'uninitialized' state.
    this._cngClient.on('disconnected', () => {
      this.state = state.UNINITIALIZED;
      this.cngData.reset();
    });

    // Then, if the connection comes back, we start again with the initialization procedure
    this._cngClient.on('connected', this._initialize.bind(this));

    // This deals with the case if only the compositor crashes.
    // This can lead to a ton of weird behavior, therefore we just close the connection
    // and re-initialize everything if the compositor comes back up
    this._cngClient.on('systemReadyStateChanged', (result) => {
      if (result.isReady) {
        this._initialize();
      } else {
        this._cngClient._cleanupOnSocketClosed(); //eslint-disable-line
      }
    });

    // The cng api vomits a huge amount of sourcechanged notifications.
    // Since dealing with these notifications means merging the source caches and this is quite expensive,
    // we debounce the events until the sources stabilize
    // That way we can handle multiple events at once at the cost of a small delay (60ms after last event).
    Rx.fromEvent(this._cngClient, 'sourcesChanged')
      .pipe(debounceTime(60 /* ms */))
      .subscribe(() => {
        this._log('verbose', 'Now scheduling sourceschanged event');

        this.scheduler.schedule(async () => {
          this._log('verbose', 'Handling sourcesChanged event');
          const sources = await this._cngClient.getAllSources();
          return this._mergeSources(this.cngData.$arrays.sources, sources);
        });
      });
  }

  /*= =====================================
     * Caching
     *====================================== */

  get cngData() {
    return this._cngData;
  }

  set cngData(data) {
    if (this._cngData) {
      this._cngData.destroy();
    }
    this._cngData = data;
  }

  /*= =====================================
     * Utilities
     *====================================== */

  /*
     * START FSM utilities
     */

  get initializing() {
    return this._state === state.INITIALIZING;
  }

  get initialized() {
    return this._state === state.INITIALIZED;
  }

  get uninitialized() {
    return this._state === state.UNINITIALIZED;
  }

  set state(newState) {
    this._state = newState;

    if (this.initialized || !this.isActive) {
      if (this._message) {
        this._message.hide();
        delete this._message;
      }
    } else if (this.initializing || (this.uninitialized && this.isActive)) {
      if (!this._message) {
        this._message = this._showMessage({ text: 'loading screen content ...' });
      }
    }

    this._clearInitializationTimer();
  }

  /*
     * END FSM utilities
     */

  /*= =====================================
     * Initialization procedure
     *====================================== */
  /**
     * Initializes the device.
     * To be on the safe-side of things, we assume the NRC can be in whatever-state after we connect to it.
     * Instead of comparing all data, which is expensive and hard to get right,
     *  we do a complete reset of all data that may have changed.
     * This means:
     * - Warming the CNG cache
     * - Comparing the Composition and Sink configurations with the data in Meteor, reconfigure if necessary
     * - Delete and restore all layouts, if possible
     * - Set all BYOD related settings (moderation, locks, code visibility, ...)
     * - Merge old and new sources
     * - Start RTSP and MJPEG servers
     * - Kick-off the preview providers (= initialize preview provider manager)
     *
     * @returns {*|Promise.<T>|Promise<T>}
     * @private
     */
  async _initialize() {
    this._log('info', 'Trying to initialize');
    if (this.isActive && this._meteorClient.isConnected && this._cngClient.isConnected && this.uninitialized) {
      this._log('info', 'Starting initialization');

      this.state = state.INITIALIZING;

      await this.cngData.stop();

      this._clearAllTimers();

      try {
        await this._authenticateWithCng();

        try {
          await this._ensureCngReady();
        } catch (error) {
          this._cinergyNotReadyCounter += 1;
          this._log('info', `Cinergy not ready after ${this._cinergyNotReadyCounter} attempts...`);
          throw error;
        }
        this._cinergyNotReadyCounter = 0;

        await this.cngData.start();
        await this._synchronizeTillStable();
        await this._tryRestoreLayouts();
        // XXX: not yet implemented
        // The showCompositor command below starts the renderloop of Cinergy
        await this._cngClient.showCompositor();
        if (enableBrowserDebug) {
          // Make sure Cinergy is not in front of any browser window. After the renderloop
          // of Cinergy has been started using the showCompositor command, we minimize
          // the output window of Cinergy so that it is active but not visible on screen.
          await this._cngClient.hideCompositor();
        } else {
          // Fix for itrack https://itrack.barco.com/browse/NET0004-3728
          // On slow machines Cinergy is sometimes shown on wrong position/size.
          // This can be solved by simply sending one more showCompositor command.
          this._showTimer = setTimeout(this._cngClient.showCompositor.bind(this._cngClient), 2000);
        }
        this._log('info', 'Initialization done');

        this.state = state.INITIALIZED;

        this.emit('initialized');
        return true;
      } catch (error) {
        this._log('error', 'Error during initialization phase', error);

        this.state = state.UNINITIALIZED;

        if ((this._cinergyNotReadyCounter + 1) % 21 === 0) {
          this._log('error', 'Cinergy still not ready after 20 attempts, restarting compositor.');
          try {
            this._cngClient.restartServices();
          } catch (e) {
            this._log('error', 'Error while restarting cinergy services', e);
          }
        }

        // If for some reason cinergy doesn't start after 10 tries
        // reboot the machine
        if (this._cinergyNotReadyCounter > 61) {
          this._log('error', 'Cinergy still not ready after 60 attempts, rebooting device', { serviceLog: true });
          try {
            rebootSystem();
          } catch (e) {
            this._log('error', 'Error while rebooting', e);
          }
          this._cinergyNotReadyCounter = 0;
        }

        if (this.isActive) {
          this._log('info', 'Retrying initialization in 1000 ms');
          this._initializationTimer = setTimeout(() => {
            this._initialize();
          }, 1000);
        } else {
          this._log('info', 'Service not active, not retrying');
        }
      }
    } else {
      this._log('info', 'Not yet ready to initialize');
    }
  }

  /**
     * Kicks of the initialization loop
     * @returns {*}
     * @private
     */
  async _synchronizeTillStable() {
    this._synchronizationRuns = 0;
    this._isDirty = this._device.configured;

    if (!this._isDirty) {
      return true;
    }
    return this._synchronizeStep();
  }

  /**
     * This step gets repeated as long as data changes in the NRC or from Meteor.
     * It:
     * - Deactivates all layouts
     * - Removes all layouts
     * - Cleans all sources
     * - Sets correct compositions
     * - Sets BYOD locks
     * - Sets moderation
     * - Sets BYOD access codes
     * - Sets BYOD code visibilities
     * - Checks if data has changed during previous steps and re-runs till stable.
     * @returns {*}
     * @private
     */
  async _synchronizeStep() {
    this._log('info', `Running synchronization step ${this._synchronizationRuns}`);

    this._isDirty = false;
    this._synchronizationRuns += 1;

    if (this._synchronizationRuns > 20) {
      throw new Error('Device initialization did not stabilize after 20 iterations');
    }

    await this._syncConfig();
    await this._setBYODLocks();
    // Set moderation needs to happen before deactivateAllLayouts
    // Otherwise, if we deactivate layouts of a non-moderated system
    // Cinergy will stop the mirrorop share.
    await this._setModeration();
    await this._deactivateAllLayouts();
    await this._cleanLayouts();
    await this._cleanSources();
    await this._setBYODComposition();
    await this._setSinkOverrides();
    await this._cngClient.setBYODAccessCode(this._device.byodAccessCode, this._device.dongleAccessCode);
    await this._setBYODCodeVisibilities();
    const isStable = await this._isStable();
    return isStable ? true : this._synchronizeStep();
  }

  _authenticateWithCng() {
    return this._cngClient.authenticate('eD2pR0xYl0g1n4Ap1', 'b4cKd0OrP4sSw0rD,4eDuNt4B4rco');
  }

  /**
     * Checks if data has changed during initialization. If not => re-run the initialization loop
     * @returns {Promise}
     * @private
     */
  async _isStable() {
    this._log('verbose', 'Checking if stable');
    const cache = await this.cngData.fetch();
    this._isDirty = this._isDirty || this.cngData.equals(cache);
    if (this._isDirty) {
      this._log('verbose', 'Data changed while synchronizing => not stable');
      this.cngData.set(cache);
    }
    return !this._isDirty;
  }

  /**
     * After first connection, we need to clean the sources.
     * This means removing all sources that are neither BYOD or already present in Meteor.
     *
     * @returns {Promise}
     * @private
     */
  _cleanSources() {
    this._log('verbose', 'Cleaning sources (merge with meteor data)');
    return this._mergeSources(this.cache.$arrays.sources
      .filter(src => !this._isUnnecessaryWallpaper(src))
      .map(source => this._translateSourceFromMeteorToCng(source)), this.cngData.$arrays.sources);
  }

  /**
     * Remove all layouts that are currently in the NRC, since we want a fresh-start after we connected.
     *
     * @returns {Promise}
     * @private
     */
  _cleanLayouts() {
    this._log('verbose', 'Cleaning layouts');

    const promises = this.cngData.$arrays.layouts
      .filter(layout => layout.presentationId === 'customLayout')
      .map(layout => this._removeCngLayout(layout.title));

    return Promise.all(promises);
  }

  /**
     * Try to restore the old layout.
     * We only restore a layout if all sources are present in the NRC.
     * In case not all sources are present, the initialization loop will make sure these sources are removed from Meteor
     * That way, a subsequent iteration loop will have a new layout which we can apply.
     *
     * @returns {Promise}
     * @private
     */
  _tryRestoreLayouts() {
    this._log('verbose', 'Trying to restore old layouts');

    const promises = this.cache.$arrays.layouts
      .map(layout => ({
        layout,
        compositions: this.cache.$arrays.compositions.filter(comp => comp.activeLayoutId === layout._id),
      }))
      .filter(layoutInfo => (
        layoutInfo.compositions.length > 0
                && layoutInfo.layout.hasOwnProperty('items')
                && layoutInfo.layout.items.length > 0
                && layoutInfo.layout.items
                  .reduce((acc, item) => (acc && !!this.cngData.findMeteorSource(this.cache.sources[item._id])), true)
      ))
      .map((layoutInfo) => {
        this._log('verbose', `We can restore layout: ${layoutInfo.layout._id}`);
        const cngLayout = this._translateLayoutFromMeteorToCng(layoutInfo.layout);
        return this._addCngLayout(cngLayout)
          .then(() => Promise.all(layoutInfo.compositions
            .map(comp => this.cngData.findMeteorComposition(comp))
            .filter(cngComp => !!cngComp)
            .map(cngComp => this._activateCngLayout(cngLayout.title, cngComp))));
      });

    return Promise.all(promises);
  }

  /*= =====================================
     * Meteor => Cng configuration
     *====================================== */

  /**
     * Configure the NRC, but only if there are compositions.
     * This avoids getting errors in some edge-cases where a device data is cleaned in Meteor
     * (e.g. after it gets removed from a room)
     *
     * @param config
     * @returns {*}
     * @private
     */
  async _configureNRC(config) {
    if (!this._device.configured) {
      this._log('error', 'Can not write config, since the device is not yet fully configured');
      throw new Error('Can not write config, since the device is not yet fully configured');
    }

    if (!config) {
      config = this._generateConfigFromCache();
    }

    if (!config.compositions.length
            || !!config.compositions.find(comp => !comp.sinks.length)
            || !config.compositions.find(comp => comp.deviceId === 'screen')
            || !config.compositions.reduce((acc, comp) => [...acc, ...comp.sinks], []).find(sink => sink.renderToScreen)
    ) {
      this._log('error', 'Config invalid (no valid sink/composition combination) => disable cinergy service');
      return this._cngClient.stopServices();
    }
    this._log('info', 'Configuring NRC');

    if (this._cngClient.isStarted) {
      try {
        // await new Promise(resolve => { setTimeout(resolve, 100000); });
        await this._cngClient.configureNRC(config, false);
        this._cngClient.restartServices();
      } catch (error) {
        this._log('error', 'Error while configuring NRC', error);
      }
    }

    this.state = state.UNINITIALIZED;
  }

  get applyActions() {
    this.isDirty = true;
    return this.isActive && this.initialized && this._device.configured;
  }

  _isUnnecessaryWallpaper(source) {
    if (source.type !== sourceType.WALLPAPER) {
      return false;
    }
    return !(this.cache.$arrays.compositions.map(cmp => cmp.wallpaperSourceId).includes(source._id));
  }

  /**
     * enerates a CNG configuration from the Meteor cache in order to validate and configure the NRC.
     * @returns {
     *  {
     *    streamDestinationIP: *,
     *    streamDestinationHost: (string|*|zoneHost|RegExp|string),
     *    connectionName: (*|string|zoneHost|RegExp|string),
     *    compositionParameters: {enableCompositorMinimizeButton: boolean,
     *    moderateClickShareSources: boolean,
     *    targetFPS: string,
     *    dongleCode: (*|string),
     *    enableAccessCodeForSharing: boolean,
     *    targetFPSEnablePreciseControl: boolean,
     *    launchClickShareSourcesToSelectedComposition: boolean},
     *    compositions: Array, sources: Array
     *  }
     * }
     * @private
     */
  _generateConfigFromCache() {
    const config = {
      streamDestinationIP: this._device.publicIp,
      streamDestinationHost: this._device.hostname,
      connectionName: this._device.dns || this._device.hostname,
      compositionParameters: {
        // these settings don't really matter since we set it per API call anyway
        enableCompositorMinimizeButton: false,
        moderateClickShareSources: !!this._device.moderated,
        targetFPS: '-1',
        dongleCode: this._device.dongleAccessCode || '',
        enableAccessCodeForSharing: true, // we control it per composition
        enableAccessCodeForAirplay: this._device.hasOwnProperty('enableAccessCodeForAirplay') ? this._device.enableAccessCodeForAirplay : true,
        targetFPSEnablePreciseControl: false,
        launchClickShareSourcesToSelectedComposition: true,
        showCompositorOnStartup: false,
      },
      compositions: [],
      sources: [],
    };

    this.cache.$arrays.compositions
      .forEach((composition) => {
        config.compositions.push(this._translateCompositionFromMeteorToCng(composition));
      });

    this.cache.$arrays.sources
      .filter(src => !this._isUnnecessaryWallpaper(src))
      .forEach((source) => {
        config.sources.push(this._translateSourceFromMeteorToCng(source));
      });
    return config;
  }

  /**
     * Generates a CNG configuration from the CNG cache in order to validate and configure the NRC.
     * @returns {
     *  {
     *    streamDestinationIP: String,
     *    streamDestinationHost: string,
     *    connectionName: string,
     *    compositionParameters: {
     *    enableCompositorMinimizeButton: boolean,
     *    moderateClickShareSources: boolean,
     *    targetFPS: string,
     *    enableAccessCodeForSharing: boolean,
     *    targetFPSEnablePreciseControl: boolean,
     *    launchClickShareSourcesToSelectedComposition: boolean
     *  },
     *  compositions: Array,
     *  sources: Array
     *  }
     * }
     * @private
     */
  _generateConfigFromCngData() {
    const config = {
      streamDestinationIP: this._device.publicIp,
      streamDestinationHost: this._device.hostname,
      connectionName: this._device.dns || this._device.hostname,
      compositionParameters: {
        enableCompositorMinimizeButton: false,
        moderateClickShareSources: this.cngData.compositionParameters.moderateClickShareSources,
        targetFPS: '-1',
        dongleCode: this._device.dongleAccessCode,
        enableAccessCodeForSharing: true,
        targetFPSEnablePreciseControl: false,
        launchClickShareSourcesToSelectedComposition: this.cngData.compositionParameters.launchClickShareSourcesToSelectedCompositor, // eslint-disable-line
        showOnStartup: false,
      },
      compositions: [],
      sources: [],
    };

    this.cngData.$arrays.compositions.forEach((composition) => {
      composition.deviceId = composition.compositorId; // XXX: remove on https://itrack.barco.com/browse/NET0005-1050
      config.compositions.push(composition);
    });

    this.cngData.$arrays.sources.forEach((composition) => {
      config.sources.push(composition);
    });
    return config;
  }

  /**
     * This generates a CNG configuration from both the Meteor cache and the CNG cache and compares them
     * If there is a discrepancy, we reconfigure the NRC
     *
     * @returns {*}
     * @private
     */
  _syncConfig() {
    const meteorConfig = this._generateConfigFromCache();
    const cngConfig = this._generateConfigFromCngData();

    // Some properties can be set dynamically and therefore do not require a reconfiguration.
    // Therefore, we can filter them out to avoid unnecessary reconfigurations
    const sanitizeForComparison = (config) => {
      config = _.cloneDeep(_.omit(config, ['sources', 'compositionParameters']));
      config.compositions.forEach((comp) => {
        delete comp.hideBYODSharingCode;
        delete comp.fadeOutDurationMs;
        delete comp.flyOverDurationMs;
        delete comp.fadeInDurationMs;
        delete comp.showSourceNameDurationMs;
        delete comp.generateAlternativesForCustomLayouts;
        delete comp.lastTakenSnapshotID;

        comp.sinks = _.keyBy(comp.sinks, 'deviceId');
      });
      config.compositions = _.keyBy(config.compositions, 'deviceId');
      return config;
    };

    // Do a deep check between the two configurations
    // Not that we only expect all fields from the Meteor config to be equal to the fields in the CNG config
    // but not the other way around. This is to avoid configuration loops if CNG would add fields to its objects
    const configEqual = (a, b) => {
      const props = Object.keys(a);
      for (const prop of props) {
        if (!b.hasOwnProperty(prop)) return false;
        if (a[prop] !== null && typeof a[prop] === 'object') {
          if (!configEqual(a[prop], b[prop])) {
            return false;
          }
        } else if (Array.isArray(a[prop]) && Array.isArray(b[prop]) && a[prop].length === b[prop].length) {
          for (let i = 0, len = a[prop].length; i < len; i++) {
            if (!configEqual(a[prop][i], b[prop][i])) return false;
          }
        } else if (a[prop] !== b[prop]) {
          this._log('debug', `Difference in config property: ${prop}`);
          return false;
        }
      }
      return true;
    };

    this._log('verbose', `Comparing configurations. Meteor config: ${JSON.stringify(meteorConfig, null, ' ')}, device config:${JSON.stringify(cngConfig, null, ' ')}`);

    if (!configEqual(sanitizeForComparison(meteorConfig), sanitizeForComparison(cngConfig))) {
      this._log('info', 'Configurations do not match => reconfigure');
      return this._configureNRC(meteorConfig);
    }
    this._log('info', 'Configurations match');
    return true;
  }

  /*= =====================================
     * Meteor <=> Cng data translations
     * Meteor data structure does not map 1-on-1 with cng data structures
     * These method do the necessary translation
     *====================================== */

  _translateCompositionFromMeteorToCng(comp) {
    const wallpaperSource = this.cache.sources[comp.wallpaperSourceId];

    if (!wallpaperSource) {
      throw new Error(`No wallpaper source for comp ${comp._id}`);
    }

    const sinks = this.cache.findDefaultSinksForComposition(comp).filter(sink => sink.type === sinkType.SCREEN);
    const displaysBB = bridgeUtils.calculateDisplaysBB(this._device.displayConfig);

    if (sinks.length === 0) {
      // render offscreen
      sinks.push({
        _id: comp._id,
        x: displaysBB.width,
        y: displaysBB.height,
        width: comp.width,
        height: comp.height,
        type: sinkType.SCREEN,
        defaultCompositionId: comp._id,
        activeCompositionId: comp._id,
      });
    }

    sinks.forEach((sink) => {
      sink.renderToScreen = (
        displaysBB.x <= sink.x
                && displaysBB.y <= sink.y
                && displaysBB.width >= (sink.x + sink.width)
                && displaysBB.height >= (sink.y + sink.height)
      );
    });

    const {
      channelId,
      deviceId,
    } = bridgeUtils.keyToSource(bridgeUtils.sourceToKey(wallpaperSource));

    const cngComp = {
      index: comp.index,
      icon: 'largeDisplay',
      deviceId: bridgeUtils.compositorIdForComposition(comp),
      title: comp._id,
      wallpaper: '',
      hideBYODSharingCode: false,
      forceLaunchClickShareSourcesToThisCompositor: (
        comp._id === bridgeUtils.getFallbackSettings(this._device).byodCompositionId
      ),
      wallpaperSourceDeviceId: deviceId,
      wallpaperSourceChannelId: channelId,
      fadeOutDurationMs: 300,
      flyOverDurationMs: 400,
      fadeInDurationMs: 300,
      showSourceNameDurationMs: 4000,
      generateAlternativesForCustomLayouts: false,
      sinks: sinks.map(this._translateSinkFromMeteorToCng.bind(this)),
    };

    return cngComp;
  }

  _translateSinkFromMeteorToCng(sink) {
    // const display = this._device.displayConfig.find(d => d.id === sink.display);
    // const touchScreenConnected = display && display.isTouchScreen;
    const cngSink = {
      x: sink.x,
      y: sink.y,
      width: sink.width,
      height: sink.height,
      type: 'GST',
      title: sink._id,
      channelId: '0',
      deviceId: sink._id,
      renderToScreen: sink.renderToScreen,
      registerRTSPWorker: false,
      // touchScreenConnected
    };

    const composition = this.cache.compositions[sink.defaultCompositionId];
    const auxiliarySinks = this.cache.findDefaultSinksForComposition(composition);

    const mjpegSink = auxiliarySinks.find(s => s.type === sinkType.MJPEG_STREAM);
    const rtspSink = auxiliarySinks.find(s => s.type === sinkType.RTSP_STREAM);
    const swfgSink = auxiliarySinks.find(s => s.type === sinkType.SWFG_STREAM);

    const rtspFps = (rtspSink && rtspSink.framerate)
      ? (rtspSink.framerate.denominator / rtspSink.framerate.numerator)
      : 1;

    const swfgFps = (swfgSink && swfgSink.framerate)
      ? (swfgSink.framerate.denominator / swfgSink.framerate.numerator)
      : 1;

    // There are problems if we write just as fast to shared memory as we read
    // Therefore we increase the mjpeg framerate by 1.
    const mjpegReadFps = (mjpegSink && mjpegSink.framerate)
      ? (mjpegSink.framerate.denominator / mjpegSink.framerate.numerator)
      : 1;
    const mjpegWriteFps = mjpegReadFps + 1;

    // Stream fps tells cinergy how fast to write to shared memory
    // since both the rtsp server and mjpeg pipeline read from this, we need to pick the maximum fps
    const streamFps = Math.max(rtspFps, mjpegWriteFps, swfgFps);
    Object.assign(cngSink, {
      renderToStream: !!rtspSink || !!swfgSink,
      streamFps,
      renderThumbnails: !!mjpegSink,
      thumbnailFps: mjpegReadFps,
    });

    // Since rtsp server and mjpeg pipeline read from the same shared memory, they should have equal resolution...
    if (mjpegSink) {
      Object.assign(cngSink, {
        streamThumbnailWidth: mjpegSink.width || 1920,
        streamThumbnailHeight: mjpegSink.height || 1080,
      });
    }

    if (rtspSink) {
      Object.assign(cngSink, {
        streamEncoding: rtspSink.encoding,
        streamBitrate: rtspSink.bitrate || 5000000,
        streamCpuLimit: rtspSink.cpuLimit || 100,
        streamMemLimit: rtspSink.memLimit || 512,
      });

      Object.assign(cngSink, {
        streamThumbnailWidth: rtspSink.width || 1920,
        streamThumbnailHeight: rtspSink.height || 1080,
      });
    }

    return cngSink;
  }

  _translateLayoutFromMeteorToCng(layout) {
    if (!layout || !layout.items.length) {
      return {};
    }

    const oldCngLayout = this.cngData.findMeteorLayout(layout);
    let index = 0;
    if (oldCngLayout) {
      const parsed = bridgeUtils.parseLayoutTitle(oldCngLayout.title);
      index = parsed.index + 1;
    }

    const cngLayout = {
      title: `${layout._id}_${index}`,
      elements: [],
    };

    for (const item of layout.items) {
      const cngSource = this.cngData.findMeteorSource(this.cache.sources[item._id]);
      if (cngSource) {
        cngLayout.elements.push({
          type: 'Insertion',
          x: item.area[0],
          y: item.area[1],
          width: item.area[2],
          height: item.area[3],
          SourceRef: {
            deviceId: cngSource.deviceId,
            channelId: cngSource.channelId,
          },
        });
      }
    }

    return cngLayout;
  }

  _translateSourceFromMeteorToCng(source) {
    if (!source) {
      return {};
    }

    const signalType = (s) => {
      switch (s.type) {
      case sourceType.WALLPAPER:
      case sourceType.ACTIVITY:
        return 'HTML';
      case sourceType.BYOD: {
        switch (s.addedBy[0].trigger) {
        case sourceAddedBy.DEVICE:
          return 'mirrorOp';
        default:
          return 'HTML';
        }
      }
      case sourceType.FILE: {
        const extension = path.extname(s.url);
        switch (extension.toLowerCase()) {
        case '.jpg':
        case '.png':
        case '.jpeg':
        case '.tiff':
        case '.bmp':
        case '.gif':
        case '.webp':
          return 'localImage';
        default:
          return 'localMedia';
        }
      }
      case sourceType.STREAM: {
        const url = URL.parse(s.url);
        switch (url.protocol.substr(0, url.protocol.indexOf(':'))) {
        case 'udp':
          return 'UDP';
        case 'v2d':
        case 'rtsp':
        default:
          return 'RTSP';
        }
      }
      case sourceType.USB:
      case sourceType.OTHER:
      default:
        return 'CUSTOM';
      }
    };

    const {
      channelId,
      deviceId,
    } = bridgeUtils.keyToSource(bridgeUtils.sourceToKey(source));

    source.showTitlePolicy = Number.isInteger(source.showTitlePolicy) ? source.showTitlePolicy : 1;
    source.showTitlePolicy = Math.max(0, Math.min(cngShowTitlePolicy.length, source.showTitlePolicy));

    const cngSource = {
      title: source.title.substr(0, Math.min(32, source.title.length)),
      exposeInUI: !!source.exposeInUi,
      signalType: signalType(source),
      addToConfiguration: true,
      url: source.url,
      latencyInMSec: source.latencyInMSec || 0,
      deviceId,
      channelId,
      audioEnabled: !!source.audio.enabled,
      nativeWidth: source.width,
      nativeHeight: source.height,
      image: 'NGSD220',

    };

    if (cngSource.title !== 'wallpaper') {
      cngSource.showTitlePolicy = (cngShowTitlePolicy[source.showTitlePolicy]);
    }

    if (cngSource.signalType.toLowerCase() === 'html') {
      cngSource.width = source.width;
      cngSource.height = source.height;
      cngSource.fps = source.framerate ? Math.round(source.framerate.denominator / source.framerate.numerator) : 15;
      cngSource.proxy = (this._device.proxy && this._device.proxy.url) ? this._device.proxy.url : '';
      cngSource.proxyUser = (this._device.proxy && this._device.proxy.user) ? this._device.proxy.user : '';
      cngSource.proxyPassword = (this._device.proxy && this._device.proxy.password) ? this._device.proxy.password : '';
    }

    if (!source.audio.enabled) {
      cngSource.audioMuted = true;
      cngSource.audioVolume = 0;
      cngSource.autoMute = false;
    } else {
      // We let pulse audio deal with source audio, so enable always.
      cngSource.audioMuted = false; //! !source.audio.muted;
      cngSource.autoMute = !!source.audio.autoMute;
      cngSource.audioVolume = 100; // source.audio.volume * 100 || 0;
    }

    return cngSource;
  }

  _isLocalBYOD(signalType) {
    signalType = signalType.toLowerCase();
    return (signalType.includes('clickshare') || signalType.includes('mirrorop'));
  }

  _translateSourceFromCngToMeteor(cngSource) {
    if (!cngSource) {
      return {};
    }

    const translateType = (signalType) => {
      signalType = signalType.toLowerCase();
      switch (signalType) {
      case 'mirroroppc':
      case 'clickshare':
      case 'clicksharedongle':
      case 'mirrorop':
      case 'clickshareios':
      case 'clicksharemirrorop':
        return sourceType.BYOD;
      case 'v2d':
      case 'webrtc':
      case 'rtsp':
      case 'udp':
        return sourceType.STREAM;
      case 'localmedia':
      case 'localimage':
        return sourceType.FILE;
      case 'http':
        return sourceType.ACTIVITY;
      case 'custom':
      default:
        // Make sure we catch any possible new types of BYOD signaltypes
        if (this._isLocalBYOD(signalType)) {
          return sourceType.BYOD;
        } else if (cngSource.url.includes('v4l2src')) {
          return sourceType.USB;
        }
        return sourceType.OTHER;
      }
    };

    const type = translateType(cngSource.signalType);

    // XXX : Temporary fix for bug in Cinergy getSources command.
    if (type === sourceType.BYOD) {
      cngSource.hasAudio = true;
    }

    const cinergyId = bridgeUtils.sourceToKey(cngSource);
    const meteorSource = {
      _id: Random.id(),
      title: cngSource.title,
      type,
      exposeInUi: cngSource.addable,
      latencyInMSec: cngSource.latencyInMSec,
      width: cngSource.nativeWidth,
      height: cngSource.nativeHeight,
      addedBy: [{
        trigger: sourceAddedBy.DEVICE,
      }],
      origin: {
        cinergyId,
      },
      // We let pulseaudio deal with audio...
      audio: {
        autoMute: cngSource.autoMute,
        enabled: cngSource.hasAudio,
        // Next line: in case of automute, purple decides the muting in audiobridge
        muted: !cngSource.hasAudio || (cngSource.audioMuted && (cngSource.hasOwnProperty('autoMute') && !cngSource.autoMute)),
        volume: (cngSource.hasAudio ? 100 : 0),
      },
      url: cngSource.url,
      device: {
        _id: this._device._id,
      },
      showTitlePolicy: cngShowTitlePolicy.indexOf(cngSource.showTitlePolicy),
    };

    if (cngSource.fps) {
      meteorSource.framerate = {
        denominator: cngSource.fps,
        numerator: 1,
      };
    }

    return meteorSource;
  }

  /*= =====================================
     * Meteor event listeners - specific
     *====================================== */

  /*-------------------------------------
     * Meteor composition changes
     *-------------------------------------*/
  /**
     * Compositions can not be added on-the-fly, so we need to reconfig CNG
     * @param comp
     * @returns {*}
     * @private
     */
  _onCompositionAdded() {
    return this._configureNRC();
  }

  /**
     * Composition changed.
     * Not all composition properties can be changed on-the-fly, for these we need to reconfig CNG
     *
     * @param comp
     * @param oldFields
     * @param clearedFields
     * @param newFields
     * @returns {*}
     * @private
     */
  _onCompositionChanged(comp, oldFields, clearedFields, newFields) {
    if (this._anyChanged(['order', 'type', 'width', 'height', 'AR', 'wallpaperSourceId'], newFields)) {
      // Not all composition properties can be changed on-the-fly, for these we need to reconfig CNG
      return this._configureNRC();
    }

    const promises = [];

    if (newFields.hasOwnProperty('showByodSharingPincode')) {
      promises.push(this._setBYODCodeVisibility(comp));
    }

    if (clearedFields.indexOf('activeLayoutId') > -1) {
      this._log('verbose', `No layout on composition ${comp._id} -> deactivate`);
      promises.push(this._deactivateCngLayout(this.cngData.findMeteorComposition(comp)));
    }

    if (newFields.hasOwnProperty('activeLayoutId')) {
      this._log('verbose', `Active layout changed on composition ${comp._id} -> update`);
      const cngComp = this.cngData.findMeteorComposition(comp);

      if (cngComp) {
        const meteorLayout = this.cache.layouts[comp.activeLayoutId];
        const cngLayout = this.cngData.findMeteorLayout(meteorLayout);
        if (cngLayout && meteorLayout.items.length) {
          this._log('debug', `Active layout changed on composition ${comp._id} -> activate`);
          promises.push(this._activateCngLayout(cngLayout.title, cngComp));
        } else {
          this._log('debug', `Active layout changed on composition ${comp._id}, but empty or not found -> deactivate`);
          promises.push(this._deactivateCngLayout(cngComp));
        }
      } else {
        this._log('error', `Composition ${comp._id} not found in cinergy cache`);
      }
    }

    return Promise.all(promises);
  }

  /**
     * Composition removed
     * Compositions can not be removed on-the-fly, so we need to reconfig CNG
     *
     * @param comp
     * @returns {*}
     * @private
     */
  _onCompositionRemoved() {
    return this._configureNRC();
  }

  /*-------------------------------------
     * Meteor sink changes
     *-------------------------------------*/

  /**
     * Sink added
     * Sinks can not be added on-the-fly, so we need to reconfig CNG
     *
     * @param sink
     * @returns {*}
     * @private
     */
  _onSinkAdded() {
    return this._configureNRC();
  }

  /**
     * Sink Changed
     * Not all sink properties can be changed on-the-fly, for these we need to reconfig CNG
     *
     * @param sink
     * @param oldFields
     * @param clearedFields
     * @param newFields
     * @returns {*}
     * @private
     */
  _onSinkChanged(sink, oldFields, clearedFields, newFields) {
    if (this._anyChanged(['x', 'y', 'width', 'height'], newFields)) {
      // Not all sink properties can be changed on-the-fly, for these we need to reconfig CNG
      return this._configureNRC();
    }

    const promises = [];
    if (newFields.hasOwnProperty('activeCompositionId')) {
      promises.push(this._setSinkOverride(sink));
    }
    return Promise.all(promises);
  }

  /**
     * Sink removed
     * Sinks can not be removed on-the-fly, so we need to reconfig CNG
     *
     * @param sink
     * @returns {*}
     * @private
     */
  _onSinkRemoved() {
    return this._configureNRC();
  }

  /*-------------------------------------
     * Meteor source changes
     *-------------------------------------*/

  /**
     * Source added => add the source to CNG and create a new preview provider
     * @param source
     * @private
     */
  _onSourceAdded(source) {
    const cngSource = this.cngData.findMeteorSource(source);
    const promises = [];
    if (!cngSource && (source.addedBy[0].trigger !== sourceAddedBy.DEVICE) && !this._isUnnecessaryWallpaper(source)) {
      if (source.type === sourceType.BYOD) {
        this._log('verbose', 'Sharing source has been added', { ...source, serviceLog: true });
      }
      this._log('debug', `About to add new source pushed by blue to Cinergy: ${JSON.stringify(source, null, 2)}`);
      const translatedSource = this._translateSourceFromMeteorToCng(source);
      promises.push(this._addCngSource(translatedSource));
    }
    return Promise.all(promises);
  }

  /**
     * Source changed => update properties controlled by CNG
     *
     * @param source
     * @param oldFields
     * @param clearedFields
     * @param newFields
     * @returns {Promise}
     * @private
     */
  _onSourceChanged(source, oldFields, clearedFields, newFields) {
    const promises = [];
    const cngSource = this._cngData.findMeteorSource(source);

    // Audio is dealt with in pulseaudio

    if (newFields.hasOwnProperty('title')) {
      this._log('verbose', `Updating source title of source ${source._id}`);
      promises.push(this._cngClient.updateSourceTitle(cngSource.deviceId, cngSource.channelId, source.title));
    }

    if (cngSource.signalType === 'HTML') {
      if (newFields.hasOwnProperty('url')
                || newFields.hasOwnProperty('framerate')
                || newFields.hasOwnProperty('width')
                || newFields.hasOwnProperty('height')
      ) {
        this._log('verbose', `Updating source url of source ${source._id}`);
        promises.push(this._cngClient.updateSourceURL(
          cngSource.deviceId,
          cngSource.channelId,
          this._translateSourceFromMeteorToCng(source),
        ));
      }
    }

    return Promise.all(promises);
  }

  /**
     * Source removed => remove it from CNG if it exists
     *
     * @param source
     * @returns {Promise}
     * @private
     */
  _onSourceRemoved(source) {
    const cngSource = this.cngData.findMeteorSource(source);

    const promises = [];

    // We currently do not allow removing sources that have been added by the device.
    if (cngSource && source.addedBy[0].trigger !== sourceAddedBy.DEVICE) {
      if (source.type === sourceType.BYOD) {
        this._log('verbose', 'Sharing source has been removed', { ...source, serviceLog: true });
      }

      promises.push(this._cngClient.removeSource(cngSource.deviceId, cngSource.channelId));
    }
    return Promise.all(promises);
  }

  /*-------------------------------------
     * Meteor layout changes
     *-------------------------------------*/
  /**
     * New layout, add it to CNG
     *
     * @param layout
     * @returns {*|{executed: boolean}}
     * @private
     */
  _onLayoutAdded(layout) {
    if (layout.items.length) {
      return this._addCngLayout(this._translateLayoutFromMeteorToCng(layout));
    }
    return true;
  }

  /**
     * Layout changes.
     * Because layouts can not be changed/updated in CNG, we use the following workaround:
     * - Add a new version to CNG
     * - Activate the new version
     * - Remove the old version
     *
     * Note that a layout without items = Wallpaper in Meteor, but black screen in CNG.
     * Therefore, we need to deactivate the layout if the item list is empty
     *
     * @param layout
     * @param oldFields
     * @param clearedFields
     * @param newFields
     * @returns {*}
     * @private
     */
  _onLayoutChanged(layout, oldFields, clearedFields, newFields) {
    let needsUpdate = false;
    if (newFields.items) {
      if (oldFields.items.length !== newFields.items.length) {
        needsUpdate = true;
      } else {
        const newItems = newFields.items.reduce((obj, item) => {
          obj[item._id] = item;
          return obj;
        }, {});
        for (const oldItem of oldFields.items) {
          const newItem = newItems[oldItem._id];
          if (!newItem) {
            needsUpdate = true;
            break;
          } else if (!_.isEqual(oldItem.area, newItem.area)) {
            needsUpdate = true;
            break;
          }
        }
      }
    }

    const promises = [];

    if (needsUpdate) {
      // Layout items have changed -> update CNG
      this._log('verbose', 'Layout has changed and needs update');

      if (!layout.items.length) {
        // Layout empty -> deactivate in CNG to go back to wallpaper, then remove the stored layout
        this._log('verbose', 'Layout empty => deactivate and remove');

        this.cache.$arrays.compositions
          .filter(comp => comp.activeLayoutId === layout._id)
          .map(comp => this.cngData.findMeteorComposition(comp))
          .filter(cngComp => !!cngComp)
          .forEach((cngComp) => {
            promises.push(this._deactivateCngLayout(cngComp));
          });

        promises.push(this._removeCngLayouts(l => bridgeUtils.parseLayoutTitle(l.title)._id === layout._id));
      } else {
        // Layout items changed -> Add new version, activate and remove old versions
        this._log('verbose', 'Layout not empty => overwrite (= add, activate and remove old versions)');
        const layoutToAdd = this._translateLayoutFromMeteorToCng(layout);
        promises.push(this._addCngLayout(layoutToAdd));

        const cngComps = this.cache.$arrays.compositions
          .filter(comp => comp.activeLayoutId === layout._id)
          .map(comp => this.cngData.findMeteorComposition(comp))
          .filter(cngComp => !!cngComp);

        this._log('verbose', `Activate layout on compositions ${cngComps.map(comp => comp.compositorId).join(', ')}`);
        promises.push(this._activateCngLayout(layoutToAdd.title, ...cngComps));

        promises.push(this._removeCngLayouts(l => (bridgeUtils.parseLayoutTitle(l.title)._id === layout._id
          && l.title !== layoutToAdd.title)));
      }
    }

    return Promise.all(promises);
  }

  /**
     * Act on layout removed.
     * This will deactivate the layouts on the compositions it runs, and then remove all stored versions
     *
     * @param removedLayout
     * @returns {Promise.<T>|Promise<TResult>|Promise<R>}
     * @private
     */
  _onLayoutRemoved(removedLayout) {
    const promises = [];
    this.cache.$arrays.compositions
      .filter(comp => comp.activeLayoutId === removedLayout._id)
      .forEach((comp) => {
        const cngComp = this.cngData.findMeteorComposition(comp);
        if (cngComp) {
          this._log('verbose', `Deactivating layout on composition ${comp._id}`);
          promises.push(this._deactivateCngLayout(cngComp));
        }
      });

    promises.push(this._removeCngLayouts(layout => bridgeUtils.parseLayoutTitle(layout.title)._id === removedLayout._id));

    return Promise.all(promises);
  }

  /*-------------------------------------
     * Meteor device changes
     *-------------------------------------*/
  /**
     * Act on device change
     * @param device
     * @param oldFields
     * @param clearedFields
     * @param newFields
     * @returns {*}
     * @private
     */
  _onThisDeviceChanged(device, oldFields, clearedFields, newFields) {
    if (newFields.hasOwnProperty('configured') || newFields.hasOwnProperty('dns') || newFields.hasOwnProperty('enableAccessCodeForAirplay')) {
      return this._configureNRC();
    }

    const promises = [];

    if (newFields.hasOwnProperty('byodAccessCode') || newFields.hasOwnProperty('dongleAccessCode')) {
      this._log('verbose', 'Setting BYOD access codes');
      promises.push(this._cngClient.setBYODAccessCode(device.byodAccessCode, device.dongleAccessCode));
    }

    if (newFields.hasOwnProperty('moderated')) {
      promises.push(this._setModeration());
    }

    if (newFields.hasOwnProperty('byodEnabled')) {
      promises.push(this._setBYODLocks());
    }

    if (newFields.hasOwnProperty('byodCompositionId')) {
      const cngComp = this.cngData.findMeteorComposition(this.cache.compositions[newFields.byodCompositionId]);
      promises.push(this._cngClient.setSelectedComposition(cngComp.compositorId));
    }

    return Promise.all(promises);
  }

  /**
     * Act on incoming client RPCs
     * RPC's are useful for one-time actions that are not part of the application state.
     *
     * @param rpc
     * @returns {*}
     * @private
     */
  _onRpcAdded(rpc) {
    this._log('verbose', `Handling rpc ${JSON.stringify(rpc, null, ' ')}`, { serviceLog: true });
    switch (rpc.command) {
    case 'refreshSource': {
      const cngSource = this.cngData.findMeteorSource(this.cache.sources[rpc.params.sourceId]);
      this._cngClient.refreshSource(cngSource.deviceId, cngSource.channelId);
      break;
    }
    case commands.DISCONNECT_BYOD_CLIENTS:
      this._cngClient.disconnectAllBYODevices();
      break;
    default:
    }
  }

  /*= =====================================
     * Cng Interaction
     *====================================== */
  /**
     * Activate all fallback settings, this will:
     * - Set the fallback sink overrides
     * - Set the fallback audio configuration
     * - Set the fallback compositions
     * - Set the fallback moderation
     * - Set the fallback BYOD code visibility
     * - Set the fallback BYOD lock
     * @returns {Promise.<T>|*|Promise<TResult>|Promise<R>}
     * @private
     */
  async activateFallback() {
    if (this.initialized) {
      this._log('info', 'Activating fallback');

      const fallbackSettings = bridgeUtils.getFallbackSettings(this._device);
      if (fallbackSettings) {
        Object.assign(this._device, fallbackSettings);
      }

      this.cache.$arrays.sinks.forEach((sink) => {
        const sinkFallbackSettings = bridgeUtils.getFallbackSettings(sink);
        if (sinkFallbackSettings) {
          Object.assign(sink, sinkFallbackSettings);
        }
      });

      this.cache.$arrays.compositions.forEach((comp) => {
        const compFallbackSettings = bridgeUtils.getFallbackSettings(comp);
        if (compFallbackSettings) {
          Object.assign(comp, compFallbackSettings);
        }
      });

      await Promise.all([
        this._setBYODLocks(),
        this._setModeration(),
        this._setBYODCodeVisibilities(true),
        this._setBYODComposition(),
        this._setSinkOverrides(),
      ]);

      await this.stop();
      await super.activateFallback();

      this._log('info', 'Done setting fallback');
    }
  }

  _setFallbackWallpapers() {
    this._log('debug', 'Setting fallback wallpapers');
    const promises = this.cache.$arrays.compositions
      .map((comp) => {
        const cngComp = this.cngData.findMeteorComposition(comp);
        if (cngComp) {
          return this._cngClient.updateSourceURL(cngComp.wallpaperSourceDeviceId, cngComp.wallpaperSourceChannelId, {
            url: path.join('file://', RESOURCES_DIR, 'fallbackWallpaper', `${comp.wallpaper}.html?fallback=true&hostname=${encodeURIComponent(this._device.dns || this._device.hostname)}`),
          });
        }
        return Promise.resolve(true);
      });

    return Promise.all(promises);
  }

  /**
     * Remove all layouts where the title matches the given filter
     * @param filter
     * @returns {Promise}
     * @private
     */
  _removeCngLayouts(filter = () => true) {
    const promises = this.cngData.$arrays.layouts
      .filter(filter)
      .map(layout => this._removeCngLayout(layout.title));

    return Promise.all(promises);
  }

  /**
     * Add the layout to CNG, optimistically caching the result
     * @param layout
     * @returns {*|{executed: boolean}}
     * @private
     */
  async _addCngLayout(layout) {
    this._log('info', 'Adding layout');
    await this._cngClient.addLayout(layout);
    this._cngData.add('layouts', layout.title, layout);
  }

  /**
     * Add the source to CNG, optimistically caching the result
     * @param source
     * @returns {*|{executed: boolean}}
     * @private
     */
  async _addCngSource(source) {
    await this._cngClient.addSource(source);
    this._cngData.add('sources', bridgeUtils.sourceToKey(source), source);
  }

  /**
     * Add the source to CNG, optimistically caching the result
     * @param source
     * @returns {*|{executed: boolean}}
     * @private
     */
  async _removeCngSource(source) {
    const key = bridgeUtils.sourceToKey(source);

    const {
      channelId,
      deviceId,
    } = bridgeUtils.keyToSource(key);

    await this._cngClient.removeSource(deviceId, channelId);
    this._cngData.remove('sources', key);
  }

  /**
     * Remove the layout with given title from the CNG, optimistically caching the result
     * @param title layout title to remove
     * @returns {*}
     * @private
     */
  async _removeCngLayout(title) {
    this._log('verbose', `Removing layout ${title}`);

    await this._cngClient.removeLayout(title);
    this._cngData.remove('layouts', title);
  }

  /**
     * Activate a layout on a composition, optimistically caching the result
     * @param layout    layout of layout to activate
     * @param cngComps  Compositions or composition ids to activate the layout on
     * @returns {*}
     * @private
     */
  async _activateCngLayout(layout, ...cngComps) {
    cngComps = cngComps
      .map(cngComp => (typeof cngComp === 'string'
        ? this._cngData.findCompositionByCompositorId(cngComp)
        : cngComp))
      .filter(cngComp => !!cngComp);

    if (cngComps.length > 0) {
      await this._cngClient.activateLayout(layout, ...cngComps.map(cngComp => cngComp.compositorId));
      cngComps.forEach((cngComp) => {
        cngComp.activeLayout = layout;
      });
    }
  }

  /**
     * Deactivate a layout on the given composition, optimistically caching the result
     * @param cngComp
     * @param hideSources
     * @returns {*}
     * @private
     */
  async _deactivateCngLayout(cngComp, hideSources = true) {
    if (typeof cngComp === 'string') {
      cngComp = this._cngData.findCompositionByCompositorId(cngComp);
    }

    if (!cngComp) {
      this._log('error', 'No layout found to deactivate');
    } else {
      await this._cngClient.deactivateLayout(cngComp.compositorId, hideSources);
      delete cngComp.activeLayout;
    }
  }

  /**
     * Set the active BYOD lock state
     * @returns {*}
     * @private
     */
  _setBYODLocks() {
    if (!this._device.byodEnabled) {
      this._log('verbose', 'Locking BYOD device content sharing');
      return this._cngClient.lockBYODeviceContentSharing();
    }
    this._log('verbose', 'Unlocking BYOD device content sharing');
    return this._cngClient.unlockBYODeviceContentSharing();
  }

  /**
     * Sets the active moderation situation
     * @returns {*}
     * @private
     */
  _setModeration() {
    if (this._device.moderated) {
      this._log('verbose', 'Enabling moderation on device');
      return this._cngClient.enableBYODModeration();
    }
    this._log('verbose', 'Disabling moderation on device');
    return this._cngClient.disableBYODModeration();
  }

  /**
     * Deactivates the layouts on each composition
     * @returns {Promise}
     * @private
     */
  _deactivateAllLayouts() {
    this._log('verbose', 'Deactivating all layouts');
    const promises = this.cngData.$arrays.compositions.map(comp => this._deactivateCngLayout(comp));
    return Promise.all(promises);
  }

  /**
     * Sets the BYOD code visibility on all compositions
     * @param force
     * @returns {Promise}
     * @private
     */
  _setBYODCodeVisibilities(force) {
    this._log('verbose', 'Setting BYOD code visibilities on all compositions');
    const promises = this.cache.$arrays.compositions.map(comp => this._setBYODCodeVisibility(comp, force));
    return Promise.all(promises);
  }

  /**
     * Sets the BYOD code visibility on the given composition
     * @param comp
     * @param force
     * @returns {*}
     * @private
     */
  _setBYODCodeVisibility(comp, force) {
    this._log('verbose', `Setting BYOD code visibility on composition ${comp._id}`);
    const cngComp = this.cngData.findMeteorComposition(comp);
    if (cngComp && (force || cngComp.enableAccessCodeForSharing !== comp.showByodSharingPincode)) {
      if (comp.showByodSharingPincode) {
        return this._cngClient.showBYODSharingPincode(cngComp.compositorId);
      }
      return this._cngClient.hideBYODSharingPincode(cngComp.compositorId);
    }
  }

  /**
     * Sets the correct composition on each sink (by setting the correct overrides)
     * @returns {*}
     * @private
     */
  _setBYODComposition() {
    this._log('verbose', 'Setting correct compositions and overrides');
    const composition = this.cngData.findMeteorComposition(this.cache.compositions[this._device.byodCompositionId]);

    if (composition) {
      return this._cngClient.setSelectedComposition(composition.compositorId);
    }

    return Promise.resolve(true);
  }

  /**
     * Loops over all sinks and sets the correct overrides
     * @returns {Promise}
     * @private
     */
  _setSinkOverrides() {
    const promises = [];

    // XXX: ATM, we can only set 1 sink override, because Cinergy implementation is [enter four-letter word here]
    const sinkToOverride = this.cache.$arrays.sinks
      .find(sink => sink.type === sinkType.SCREEN && sink.defaultCompositionId !== sink.activeCompositionId);
    if (sinkToOverride) {
      promises.push(this._setSinkOverride(sinkToOverride));
    } else {
      promises.push(this._cngClient.disableSinkOverride());
    }
    return Promise.all(promises);
  }

  /**
     * Sets the override on the sink to the currently active composition
     * @param sink
     * @returns {*|{executed}|{executed: boolean}}
     * @private
     */
  _setSinkOverride(sink) {
    this._log('verbose', `Setting sink override on sink ${sink._id}`);
    const cngTargetSink = this.cngData.findMeteorSink(sink);

    let cngSourceSink = this.cngData.findMeteorSink(this.cache.$arrays.sinks
      .find(s => s.type === sinkType.SCREEN && s.defaultCompositionId === sink.activeCompositionId));

    if (!cngSourceSink) {
      cngSourceSink = this.cngData.sinks[sink.activeCompositionId] || cngTargetSink;
    }
    return this._cngClient.overrideSink({
      srcDeviceId: cngSourceSink.deviceId,
      srcChannelId: cngSourceSink.channelId,
      trgDeviceId: cngTargetSink.deviceId,
      trgChannelId: cngTargetSink.channelId,
    });
  }

  /**
     * Fetches and verifies the CNG version with this version of purple
     * @returns {*}
     * @private
     */
  async _checkCngVersion() {
    this._log('verbose', 'Checking CNG version');

    let version;
    try {
      version = await this._cngClient.getVersionInfo();
    } catch (error) {
      if (isDev) {
        return true; // E.g. development machines fail on this, so ignore
      }
      throw error;
    }

    if (compareVersion(version, CinergyClient.minimumSupportedVersion) === -1) {
      const error = `Unsupported Cinergy firmware version detected! Version = ${version}, minimum supported version = ${CinergyClient.minimumSupportedVersion}`;
      this._log('error', error);
      throw new Error(error);
    } else {
      this._cngVersion = version;
      return version;
    }
  }

  /**
     * Ensures cinergy is ready
     * @returns {*}
     * @private
     */
  async _ensureCngReady() {
    const ready = await this._cngClient.getSystemReadyStatus();
    if (ready) {
      return true;
    }
    throw new Error('Cinergy not yet ready');
  }

  /**
     * Sources are the only entitites that should be synchronized in two-ways.
     * This is so because the origin of the source can be the device (as for BYOD)
     *  and Meteor (wallpapers, activities, ...)
     * Therefore we need to be able to merge sources between CNG and Meteor.
     * This is pretty hard because of a couple of reasons:
     * - We don't always get the latest state from CNG (runtime not yet updated)
     * - Some changes can still be queued in the CinergyClient
     * - Some properties are controlled by Meteor even though the origin of the source is CnG (E.g. volume)
     * - Some properties are controlled by CnG even though the origin of the source is Meteor (E.g. width/height)
     *
     * Therefore the merge code is quite elaborate and
     *  it will optimistically update the CnG cache to do away with stale reads from CNG.
     * I know, this ain't nice, but a man's got to do what a man's got to do.
     *
     * @param oldSources
     * @param newSources
     * @returns {Promise}
     * @private
     */
  _mergeSources(oldSources, newSources) {
    this._log('verbose', 'Merging old and new sources');

    // Filter out BYOD sources which are not actively sharing, since we can't do anything useful with them anyway
    newSources = newSources.filter(source => !this._isLocalBYOD(source.signalType)
      || ((source.nativeWidth > -1 && source.nativeHeight > -1) && (source.addable || source.onScreen)));

    const promises = [];

    // We index the sources by key, so we can do quick look-ups in the merge code below
    const oldSourcesByKey = oldSources.reduce((obj, source) => {
      obj[bridgeUtils.sourceToKey(source)] = source;
      return obj;
    }, {});
    const newSourcesByKey = newSources.reduce((obj, source) => {
      obj[bridgeUtils.sourceToKey(source)] = source;
      return obj;
    }, {});

    // Handle changed or removed sources
    for (const oldSource of oldSources) {
      const newSource = newSourcesByKey[bridgeUtils.sourceToKey(oldSource)];
      const meteorSource = this.cache.findCngSource(oldSource);

      // we don't care for changes that are not in Meteor
      if (meteorSource) {
        if (!newSource) {
          // source removed by CNG
          if (meteorSource.addedBy[0].trigger === sourceAddedBy.DEVICE) {
            // Mirrorop sources are controlled on CNG, so sync with blue
            this._log('verbose', `Removed BYOD source ${meteorSource._id}, removing from BLUE`);
            promises.push(this._meteorClient.removeSource({
              sourceId: meteorSource._id,
            }));
            this.cache.remove('sources', meteorSource);
          } else {
            // Source controlled by blue => may not be removed by purple, so revert (= add back)
            this._log('verbose', 'Removed source controlled by Meteor, adding it back to CNG');
            promises.push(this._cngClient.addSource(oldSource));
            newSources.push(oldSource);
          }
        } else {
          // Source changed by CNG

          // source url changed, we currently can only change the URL of HTML sources
          // see: https://itrack.barco.com/browse/NET0005-1140 and https://itrack.barco.com/browse/NET0005-1141
          if (oldSource.signalType === 'HTML') {
            if (!this._device.proxy) {
              this._device.proxy = {
                url: '',
                user: '',
                password: '',
              };
            }

            if (
              (oldSource.url !== newSource.url && meteorSource.url !== newSource.url)
                            || (oldSource.fps !== newSource.fps
                              && (meteorSource.framerate.denominator / meteorSource.framerate.numerator)
                              !== newSource.fps)
                            || (oldSource.width !== newSource.width && meteorSource.width !== newSource.width)
                            || (oldSource.height !== newSource.height && meteorSource.height !== newSource.height)
                            || (oldSource.proxy !== newSource.proxy && this._device.proxy.url !== newSource.proxy)
                            || (oldSource.proxyUser !== newSource.proxyUser
                              && this._device.proxy.user !== newSource.proxyUser)
                            || (oldSource.proxyPassword !== newSource.proxyPassword
                              && this._device.proxy.password !== newSource.proxyPassword)
            ) {
              if ((meteorSource.addedBy[0].trigger === sourceAddedBy.DEVICE) && (meteorSource.url !== newSource.url)) {
                // Source controlled by cng => update to blue
                this._log('verbose', `URL of source ${meteorSource._id} changed from ${oldSource.url} to ${newSource.url} for source controlled by CNG, pushing to BLUE`);
                promises.push(this._meteorClient.updateSource({
                  _id: meteorSource._id,
                  url: newSource.url,
                }));
              } else if (meteorSource.addedBy[0].trigger !== sourceAddedBy.DEVICE) {
                // Source controlled by blue => revert
                this._log('verbose', `URL of source ${meteorSource._id} changed from ${oldSource.url} @ ${oldSource.fps}fps to ${newSource.url} @ ${newSource.fps}fps for source controlled by Meteor, setting to CNG`);

                const newInfo = _.pick(meteorSource, ['url', 'width', 'height']);
                Object.assign(newInfo, {
                  fps: (meteorSource.framerate.denominator / meteorSource.framerate.numerator),
                  proxy: this._device.proxy.url,
                  proxyUser: this._device.proxy.user,
                  proxyPassword: this._device.proxy.password,
                });
                promises.push(this._cngClient.updateSourceURL(oldSource.deviceId, oldSource.channelId, newInfo));
                Object.assign(newSource, newInfo);
              }
            }
          }

          if (oldSource.title.substr(0, Math.min(32, oldSource.title.length))
            !== newSource.title.substr(0, Math.min(32, newSource.title.length))
            && meteorSource.title.substr(0, Math.min(32, meteorSource.title.length))
            !== newSource.title.substr(0, Math.min(32, newSource.title.length))) {
            if (meteorSource.addedBy[0].trigger === sourceAddedBy.DEVICE && !meteorSource.origin.userId) {
              // controlled by cng => update to blue
              this._log('verbose', `Title of source ${meteorSource._id} changed for source controlled by CNG, pushing to BLUE`);
              promises.push(this._meteorClient.updateSource({
                _id: meteorSource._id,
                title: newSource.title,
              }));
              meteorSource.title = newSource.title;
            } else {
              // controlled by blue => revert
              this._log('verbose', `Title of source ${meteorSource._id} changed for source controlled by Meteor, setting to CNG`);
              promises.push(this._cngClient
                .updateSourceTitle(oldSource.deviceId, oldSource.channelId, meteorSource
                  .title.substr(0, Math.min(32, meteorSource.title.length))));

              newSource.title = meteorSource.title;
            }
          }

          // resolution is always controlled by CNG and therefore must be pushed to blue,
          // ignore when width and height is unknown
          if ((newSource.nativeWidth !== oldSource.nativeWidth || newSource.nativeHeight !== oldSource.nativeHeight)
            && (newSource.nativeWidth > -1 && newSource.nativeHeight > -1)) {
            if (meteorSource.width !== newSource.nativeWidth || meteorSource.height !== newSource.nativeHeight) {
              meteorSource.width = newSource.nativeWidth;
              meteorSource.height = newSource.nativeHeight;

              this._log('verbose', `Resolution of source ${meteorSource._id} changed, push to BLUE`);
              promises.push(this._meteorClient.updateSource({
                _id: meteorSource._id,
                width: meteorSource.width,
                height: meteorSource.height,
              }));
            }
          }

          // Audio is dealt with by the PA service, so ignore in the cinergy bridge
        }
      }
    }

    // Handle new sources
    for (const newSource of newSources) {
      const oldSource = oldSourcesByKey[bridgeUtils.sourceToKey(newSource)];
      if (!oldSource) {
        // Source added
        const meteorSource = this.cache.findCngSource(newSource);
        if (!meteorSource) {
          if (this._isLocalBYOD(newSource.signalType)) {
            if (!meteorSource) {
              this._log('verbose', 'Source added in CNG, push to BLUE');
              const newMeteorSource = this._translateSourceFromCngToMeteor(newSource);
              promises.push(this._meteorClient.addSource(newMeteorSource));
            }
          } else {
            // Source added which shouldn't be added
            this._log('error', 'Non BYOD-Source added in CNG, but not coming from BLUE => remove');
            promises.push(this._cngClient.removeSource(newSource.deviceId, newSource.channelId));
          }
        }
      }
    }
    this.cngData.cacheSources(newSources);

    this._log('verbose', 'Done merging sources');
    return Promise.all(promises);
  }

  /**
     * Destroy and clean-up. This will:
     *
     * - Remove all outstanding timers
     * - Remove all incoming and outgoing listeners
     * - Set the device in a fallback state
     * - Clean-up all data
     */
  async _onDestroy() {
    this._cngClient.off();
    const destroy = [this._previewMgr, this._cngClient, this._authBridge];
    this._log('info', 'Destroying child services');
    try {
      await Promise.all(destroy.map(toDestroy => toDestroy.destroy()));
    } catch (error) {
      this._log('error', 'Error while destroying child service', error);
    }
    delete this._previewMgr;
    delete this._cngClient;
    delete this._authBridge;

    this._log('info', 'Destroying cinergy cache');
    this.cngData.destroy();

    this._log('info', 'Removing references');
    delete this._cngData;
    this._eventProxy.unregister(this);
    await super._onDestroy();
  }
}

export default CinergyBridge;
