import CacheManager from '../../../services/cacheManager';
import bridgeUtils from './utils';

class CngCacheManager extends CacheManager {
  constructor(cngClient) {
    super(['sinks', 'compositions', 'sources', 'layouts'], {
      compositionParameters: {},
      systemInfo: {},
    });
    this._cngClient = cngClient;
  }

  /**
   * Fetch + Fill the cache
   */
  async warm() {
    const result = await this.fetch();
    this.set(result);
  }

  /**
   * Fetches all CNG info: compositions, system info, layouts, active layouts and sources.
   * @returns {*}
   */
  async fetch() {
    this._log('verbose', 'Fetching cng cache');

    this._log('verbose', 'Fetching all CNG compositions and sinks');
    const cache = await this._cngClient.getCompositions();

    this._log('verbose', 'Fetching all active CNG layouts');
    (await this._fetchAllActiveLayouts(cache.compositions.compositions)).forEach((activeLayout) => {
      if (!activeLayout) return;
      const comp = cache.compositions.find(c => c.compositorId === activeLayout.compositorId);
      comp.activeLayout = activeLayout.title;
    });

    this._log('verbose', 'Fetching all CNG sources');
    cache.sources = await this._cngClient.getAllSources();

    this._log('verbose', 'Fetching all CNG layouts');
    cache.layouts = await this._cngClient.getCustomLayouts();

    this._log('verbose', 'Fetching sytem info');
    cache.systemInfo = await this._cngClient.getSystemInfo();

    return cache;
  }

  /**
   * Fetch all active layouts from CNG.
   * In CNG we can only request the layouts for one compositor at a time, so this means iterating over compositions.
   *
   * @param compositions
   * @returns {Promise}
   * @private
   */
  _fetchAllActiveLayouts(compositions) {
    this._log('verbose', 'Fetching all active CNG layouts');

    if (!compositions) {
      compositions = this._data.$arrays.compositions;
    }

    return Promise.all(compositions.map(comp => this._cngClient.getActiveLayout(comp.compositorId)));
  }

  /**
   * Cache the given compositions and sinks
   * Compositions and sinks can only be requested together in the Cng API, but in Meteor they are decoupled.
   * Therefore, we split them in this method to separate collections
   *
   * @param compositions and sinks
   * @param add if true, it will add the compositions and sinks to the existing cache, otherwise it will replace them
   */
  cacheCompositionsAndSinks(compositions, add) {
    if (!add) {
      this._data.compositions = {};
      this._data.$arrays.compositions = [];
      this._data.sinks = {};
      this._data.$arrays.sinks = [];
    }

    for (const composition of compositions) {
      this.add('compositions', composition.title, composition);
      for (const sink of composition.sinks) {
        this.add('sinks', sink.deviceId, sink);
      }
    }
  }

  /**
   * Cache the given sources
   * @param sources
   * @param add if true, it will add the sources to the existing cache, otherwise it will replace them
   */
  cacheSources(sources, add) {
    if (!add) {
      this._data.sources = {};
      this._data.$arrays.sources = [];
    }
    for (const source of sources) {
      this.add('sources', bridgeUtils.sourceToKey(source), source);
    }
  }

  /**
   * Cache the given layouts
   * @param layouts
   * @param add if true, it will add the layouts to the existing cache, otherwise it will replace them
   */
  cacheLayouts(layouts, add) {
    if (!add) {
      this._data.layouts = {};
      this._data.$arrays.layouts = [];
    }
    for (const layout of layouts) {
      this.add('layouts', layout.title, layout);
    }
  }

  /**
   * Fill the cache with the given info (that comes out of the fetch method)
   * @param cache
   */
  set(cache) {
    this.reset();
    this.cacheCompositionsAndSinks(cache.compositions);
    this.cacheSources(cache.sources);
    this.cacheLayouts(cache.layouts);
    this._data.compositionParameters = cache.compositionParameters;
    this._data.systemInfo = cache.systemInfo;
    return Promise.resolve(true);
  }

  /**
   * Given a Meteor composition object, find the corresponding Cng composition object.
   * Composition Mongo Ids are saved as composition titles in Cng.
   * @param comp
   * @returns {*}
   */
  findMeteorComposition(comp) {
    if (!comp || !comp._id) {
      return undefined;
    }
    return this._data.compositions[comp._id];
  }

  /**
   * Given a cng compositor id, find the corresponding Cng sink object.
   * @param id
   * @returns {*}
   */
  findCompositionByCompositorId(id) {
    return this._data.$arrays.compositions.find(comp => comp.compositorId === id);
  }

  /**
   * Given a Meteor sink object, find the corresponding Cng sink object.
   * Sink Mongo Ids are saved as sink device Ids in Cng.
   *
   * @param sink
   * @returns {*}
   */
  findMeteorSink(sink) {
    if (!sink || !sink._id) {
      return undefined;
    }
    return this._data.sinks[sink._id];
  }

  findSource(source) {
    return this._data.sources[bridgeUtils.sourceToKey(source)];
  }

  /**
   * Given a Meteor source object, find the corresponding Cng source object.
   * Source objects are mapped using a key, which equals channelId_deviceId
   * @param source
   * @returns {*}
   */
  findMeteorSource(source) {
    if (!source || !source._id) {
      return undefined;
    }
    return this._data.sources[bridgeUtils.sourceToKey(source)];
  }

  /**
   * Given a Meteor layout object, find the corresponding Cng layout object.
   * The Mongo Id is mapped to the layout title (+ version number)
   * @param layout
   * @returns {*}
   */
  findMeteorLayout(layout) {
    if (!layout) {
      return undefined;
    }

    let maxIndex = -1;
    let newest;

    // A meteor layout can be modified, a CNG layout can't.
    // Therefore, we the layout title does not only contain a mongo id, but also a version number.
    // e.g. E41Sdsa584Ws_3 as the third version.
    // If we want the Cng Layout from a Meteor layout, we have to find the latest version still stored in Cng...
    for (const l of this._data.$arrays.layouts) {
      const parsed = bridgeUtils.parseLayoutTitle(l.title);
      if (parsed._id === layout._id && parsed.index > maxIndex) {
        newest = l;
        maxIndex = parsed.index;
      }
    }
    return newest;
  }

  /**
   * Clean-up
   */
  _onDestroy() {
    delete this._cngClient;
    return super._onDestroy();
  }
}

export default CngCacheManager;
