/*
eslint no-underscore-dangle: ["error", {
  "allow": ["_parse"],
  "allowAfterThis": true,
  "allowAfterSuper": true,
}]
*/

import * as getValue from 'get-value';

import CinergyError from './cinergyError';

/**
 * Parses a string or num to boolean
 * @param val
 * @returns {boolean}
 */
function parseBool(val) {
  if (typeof val === 'undefined') {
    return false;
  }

  const num = +val;
  return !Number.isNaN(num) ? !!num : !!String(val).toLowerCase().replace(!!0, '');
}

/**
 * Parses responses from Cinergy API
 */
const CommandResponseProcessor = {
  /**
     * Parses a response from the API and delegates the response to the specific response processor
     * @param host
     * @param data
     * @returns {*}
     */
  process: (host, data) => {
    const action = data.Notify[0].Action[0];
    const result = data.Notify[0].Result[0];

    // Handle error
    if (!parseBool(result.$.actionExecuted)) {
      if (Array.isArray(result.Error)) {
        throw new CinergyError(result.Error[0].$.errorReason, result.Error[0].$.errorDescription);
      } else {
        throw new CinergyError(-1, 'Unknown');
      }
    }

    // Handle response that has been received
    const cmdStr = action.$.command;
    if (CommandResponseProcessor.hasOwnProperty(cmdStr)) {
      return CommandResponseProcessor[cmdStr].call(host, action, result);
    }
    return {};
  },

  /**
     * Parses/casts values from an object
     * @param input the input object
     * @param transform, Object with arrays of properties to cast.
     *  e.g. {
     *   bool:['foo','bar'],
     *   float:['baz']
     * }
     * Will cast the properties `foo` and `bar` on the input object to booleans, and `baz` to float.
     * @returns The input object with casted values.
     * @private
     */
  _parse(input, transform) {
    const transformers = {
      bool: parseBool,
      float: Number.parseFloat,
      int: Number.parseInt,
    };
    transform = {
      bool: [],
      float: [],
      int: [],
      ...transform,
    };
    Object.keys(transform).forEach((transformer) => {
      transform[transformer].forEach((i) => {
        if (input.hasOwnProperty(i)) {
          input[i] = transformers[transformer](input[i]);
        }
      });
    });
    return input;
  },

  /**
     * Below methods parse the response of CNG API methods with the same name.
     */

  /**
     * processes response to getCompositions
     * @param action
     * @param compositions
     * @returns {
     *  {
     *  compositionParameters:
     *    {
     *    launchClickShareSourcesToSelectedCompositor: boolean,
     *    moderateClickShareSources: boolean
     *    },
     *    compositions: *[
     *      {
     *      wallpaperSourceDeviceId: string,
     *      wallpaper: string,
     *      wallpaperSourceChannelId: string,
     *      index: number,
     *      compositorId: string,
     *      icon: string, title: string,
     *      forceLaunchClickShareSourcesToThisCompositor: boolean,
     *      lastTakenSnapshotID: string,
     *      sinks: [
     *        {
     *        type: string,
     *        renderThumbnails: boolean,
     *        renderToStream: boolean,
     *        streamFps: number,
     *        thumbnailFps: number,
     *        height: number,
     *        renderToScreen: boolean,
     *        streamThumbnailWidth: number,
     *        streamThumbnailHeight: number,
     *        width: number,
     *        channelId: string,
     *        x: number,
     *        registerRTSPWorker: boolean,
     *        deviceId: string,
     *        y: number,
     *        title: string
     *        }
     *        ]
     *      }
     *    }
     *  ]
     * }}
     * e.g.
     * {
          "compositionParameters": {
            "launchClickShareSourcesToSelectedCompositor": true,
            "moderateClickShareSources": true
          },
          "compositions": [
            {
              "wallpaperSourceDeviceId": "main",
              "wallpaper": "",
              "wallpaperSourceChannelId": "wallpapers",
              "index": 0,
              "compositorId": "screen",
              "icon": "largeDisplay",
              "title": "main",
              "forceLaunchClickShareSourcesToThisCompositor": true,
              "lastTakenSnapshotID": "-1",
              "sinks": [
                {
                  "type": "GST",
                  "renderThumbnails": true,
                  "renderToStream": false,
                  "streamFps": 1,
                  "thumbnailFps": 2,
                  "height": 1050,
                  "renderToScreen": true,
                  "streamThumbnailWidth": 1920,
                  "streamThumbnailHeight": 600,
                  "width": 3360,
                  "channelId": "0",
                  "x": 0,
                  "registerRTSPWorker": false,
                  "deviceId": "GjxesGT53TBCPi5os",
                  "y": 0,
                  "title": "main"
                }
              ]
            },
            {
              "wallpaperSourceDeviceId": "clone",
              "wallpaper": "",
              "wallpaperSourceChannelId": "wallpapers",
              "index": 1,
              "compositorId": "screen1",
              "icon": "largeDisplay",
              "title": "clone",
              "forceLaunchClickShareSourcesToThisCompositor": false,
              "lastTakenSnapshotID": "-1",
              "sinks": [
                {
                  "streamThumbnailHeight": 1080,
                  "renderToScreen": false,
                  "renderToStream": true,
                  "renderThumbnails": true,
                  "streamMemLimit": 512,
                  "registerRTSPWorker": false,
                  "streamThumbnailWidth": 1920,
                  "deviceId": "eoqAojuEAn9CJwjQf",
                  "height": 1080,
                  "type": "GST",
                  "title": "clone",
                  "streamCpuLimit": 100,
                  "channelId": "0",
                  "streamFps": 30,
                  "thumbnailFps": 2,
                  "x": 3360,
                  "streamEncoding": "H264",
                  "y": 0,
                  "width": 1920,
                  "streamBitrate": 5000000
                }
              ]
            }
          ]
        }
     */
  getCompositions(action, compositions) {
    const result = {
      compositionParameters: {},
      compositions: [],
    };

    if (compositions.Compositions) {
      compositions = compositions.Compositions[0];

      result.compositionParameters = CommandResponseProcessor._parse(compositions.$, {
        bool: ['moderateClickShareSources', 'launchClickShareSourcesToSelectedCompositor'],
      });

      if (compositions.Composition) {
        compositions.Composition.forEach((composition) => {
          const newComp = CommandResponseProcessor._parse(composition.$, {
            bool: ['forceLaunchClickShareSourcesToThisCompositor', 'generateAlternativesForCustomLayouts'],
            int: ['index', 'fadeOutDurationMs', 'flyOverDurationMs', 'fadeInDurationMs', 'showSourceNameDurationMs'],
          });

          Object.assign(newComp, {
            lastTakenSnapshotID: '-1',
            sinks: [],
          });

          for (const sink of composition.Sink) {
            const newSink = CommandResponseProcessor._parse(sink.$, {
              bool: ['registerRTSPWorker', 'renderThumbnails', 'renderToScreen', 'renderToStream'],
              float: ['x', 'y', 'width', 'height', 'streamBitrate', 'streamCpuLimit',
                'streamMemLimit', 'streamFps', 'thumbnailFps',
                'streamThumbnailWidth', 'streamThumbnailHeight'],
            });
            newComp.sinks.push(newSink);
          }

          result.compositions.push(newComp);
        });
      }
    }
    return result;
  },

  /**
     * processes response to getSinkOverrideStatus
     * @param action
     * @param overrides
     * @returns {[
     * {
     *  overrideDeviceId: string,
     *  overrideChannelId: string,
     *  originalChannelId: string,
     *  originalDeviceId: string
     * }]}
     * e.g.
     * [
     *  {
     *    overrideDeviceId:"VGA-0",
     *    overrideChannelId:"0",
     *    originalChannelId:"0",
     *    originalDeviceId:"LVDS-0
     *  }
     * ]
     */
  getSinkOverrideStatus(action, overrides) {
    const result = [];
    const sinks = getValue(overrides, 'GLWindows.0.GLWindow.0.SinkOverrides.0.Sink') || [];

    sinks.forEach((sink) => {
      result.push(sink.$);
    });

    return result;
  },

  /**
     * processes response to getDisplayInformation
     * @param action
     * @param displayInfo
     * @returns {*}
     * e.g.
     * {
          "serverId": "mgsserver001-78d7449628ef3b93832aa5c256caddcf",
          "displayName": "NRC-200-0060e05bdf17",
          "hostName": "NRC-200-0060e05bdf17",
          "location": "Location",
          "workgroup": "all"
        }
     */
  getDisplayInformation(action, displayInfo) {
    return getValue(displayInfo, 'DisplayInformation.0.$');
  },

  /**
     * processes response to getSystemInfo
     * @param action
     * @param systemInfo
     * @returns {
     *  licenseInfo:
     *  {
     *    networkStreamOut: boolean,
     *    clickshareBaseunitEmbedded: boolean,
     *    maximumCompositions: number,
     *    maximumOnScreenSources: number
     *  },
     *  displayInfo:
     *  {
     *    displayName: string,
     *    location: string
     *  }
     * }
     * e.g.
     * {
          "licenseInfo": {
            "networkStreamOut": false,
            "clickshareBaseunitEmbedded": false,
            "maximumCompositions": 3,
            "maximumOnScreenSources": 16
          },
          "displayInfo": {
            "displayName": "NRC-200-0060e05bdf17",
            "location": "Location"
          }
        }
     */
  getSystemInfo(action, systemInfo) {
    return {
      licenseInfo: CommandResponseProcessor._parse(getValue(systemInfo, 'LicenseInfo.0.$'), {
        int: ['maximumOnScreenSources', 'maximumCompositions'],
        bool: ['clickshareBaseunitEmbedded', 'networkStreamOut'],
      }),
      displayInfo: getValue(systemInfo, 'DisplayInformation.0.$'),
    };
  },

  /**
     * processes response to getVersionInfo
     * @param action
     * @param versionInfo
     * @returns String
     * e.g. 1.1.0-b18
     */
  getVersionInfo(action, versionInfo) {
    return getValue(versionInfo, 'Version.0.$.softwareVersion');
  },

  /**
     * processes response to getActiveLayout
     * @param action
     * @param result
     * @returns {
     *  {
     *    compositorId: *,
     *    title: *,
     *    layoutId: *,
     *    presentationId: *,
     *    activeLayout: string,
     *    isNotificationResult: boolean
     *  }
     * }
     */
  getActiveLayout(action, result) {
    if (result.Layout) {
      const layout = result.Layout[0];
      return {
        compositorId: action.$.compositorId,
        title: layout.$.title,
        layoutId: layout.$.layoutId,
        presentationId: layout.$.presentationId,
        activeLayout: 'true',
        isNotificationResult: false,
      };
    }
  },

  /**
     * processes response to getLayouts
     * @param action
     * @param layouts
     * @returns {Array}
     */
  getLayouts(action, layouts) {
    const result = [];
    layouts = getValue(layouts, 'Layouts.0.Layout');

    for (let layout of layouts) {
      layout = CommandResponseProcessor._parse(layout.$, {
        bool: ['exposeInUI'],
      });
      result.push(layout);
    }

    return result;
  },

  /**
     * Parses a layout and returns all insertions and regions in one big list.
     *
     * Note that we currently flatten out nested layouts and solely work on order in the list to achieve Z-positioning.
     *
     * @param action
     * @param result
     * @returns {
     *    "title": "Foo",
     *    "items": [
     *     {x,y,width,height},
     *     {x,y,width,height},
     *    ]
     *  }
     */
  getLayoutRuntimeData(action, result) {
    const collectChildInsertions = function collectChildInsertions(insertions, insertion) {
      const inserts = insertion.Insertion;
      if (inserts) {
        for (const insert of inserts) {
          insertions.push(insert);
          collectChildInsertions(insertions, insert);
        }
      }

      if (insertion.Region) {
        const regions = insertion.Region;
        for (const region of regions) {
          collectChildInsertions(insertions, region);
        }
      } else {
        return insertions;
      }
    };

    if (result.Layouts[0]) {
      const layouts = result.Layouts[0].Layout;
      const json = [];
      for (const layout of layouts) {
        const parentInsertion = layout.Insertion[0];

        // const parent = CommandResponseProcessor._parse(parentInsertion.$, {
        //   float: ['x', 'y', 'width', 'height', 'absoluteX', 'absoluteY', 'absoluteWidth', 'absoluteHeight'],
        // });

        const insertions = [];
        const itemList = [];
        collectChildInsertions(insertions, parentInsertion);

        for (const insertion of insertions) {
          itemList.push(CommandResponseProcessor._parse(insertion.SourceRef[0].$, {
            float: ['x', 'y', 'width', 'height'],
          }));
        }

        json.push({
          title: layout.$.title,
          items: itemList,
        });
      }
      return json;
    }
  },

  /**
     * processes response to getSources
     * @param action
     * @param sources
     * @returns {[{
     *  playingState: string,
     *  onScreen: boolean,
     *  showTitlePolicy: string,
     *  autoMute: boolean,
     *  hasAudio: boolean,
     *  urlPassword: string,
     *  nativeWidth: number,
     *  latencyInMSec: string,
     *  audioEnabled: boolean,
     *  image: string,
     *  urlUser: string,
     *  proxyPassword: string,
     *  deviceId: string,
     *  audioMuted: boolean,
     *  exposeInUI: boolean,
     *  nativeHeight: number,
     *  height: number,
     *  fps: number,
     *  addable: boolean,
     *  extraTags: string[],
     *  title: string,
     *  url: string,
     *  audioVolume: number,
     *  proxy: string,
     *  channelId: string,
     *  signalType: string,
     *  proxyUser: string,
     *  width: number
     * }]}
     * e.g.
     * [
     {
       "playingState": "",
       "onScreen": true,
       "hasAudio": false,
       "urlPassword": "",
       "nativeWidth": 3360,
       "latencyInMSec": "0",
       "audioEnabled": false,
       "showTitlePolicy":"always",
       "autoMute": false,
       "image": "NGSD220",
       "urlUser": "",
       "proxyPassword": "",
       "deviceId": "main",
       "audioMuted": true,
       "exposeInUI": false,
       "nativeHeight": 1050,
       "height": 1050,
       "fps": 15,
       "addable": false,
       "extraTags": [
         ""
       ],
       "title": "wallpaper",
       "url": "http://192.168.12.131:3000/wallpaper/main.html?hostname=NRC-200-0060e05bdf17",
       "audioVolume": 0,
       "proxy": "",
       "channelId": "wallpapers",
       "signalType": "HTML",
       "proxyUser": "",
       "width": 3360
     },
     {
       "extraTags": [
         ""
       ],
       "image": "TV",
       "onScreen": false,
       "audioVolume": 70,
       "signalType": "localMedia",
       "hasAudio": true,
       "playingState": "",
       "channelId": "0",
       "nativeHeight": 688,
       "audioEnabled": true,
       "showTitlePolicy": true,
       "nativeWidth": 1280,
       "deviceId": "3000",
       "url": "/usr/share/data/mp4/ForTheBirdsTrailer.mp4",
       "audioMuted": false,
       "exposeInUI": true,
       "addable": true,
       "title": "Birds"
     }
     ]
     */
  getSources(action, sources) {
    const result = [];
    sources = getValue(sources, 'Sources.0.Source');

    for (let source of sources) {
      source = CommandResponseProcessor._parse(source.$, {
        float: ['fps', 'width', 'height', 'nativeWidth', 'nativeHeight', 'audioVolume'],
        int: ['audioPort', 'videoPort', 'controlPort'],
        bool: ['hasAudio', 'audioMuted', 'autoMute', 'audioEnabled', 'addable', 'onScreen', 'useMulticast', 'exposeInUI'],
      });

      source.nativeWidth = Number.isNaN(source.nativeWidth) ? -1 : source.nativeWidth;
      source.nativeHeight = Number.isNaN(source.nativeHeight) ? -1 : source.nativeHeight;
      source.audioVolume = Number.isNaN(source.audioVolume) ? 70.0 : source.audioVolume;
      source.extraTags = (source.hasOwnProperty('extraTags')) ? source.extraTags.split(';') : [];

      result.push(source);
    }

    return result;
  },

  /**
     * processes response to configureNRC
     * @param action
     * @param result
     * @returns {{executed: Boolean}}
     */
  configureNRC(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to setHiresThumbnailFactor
     * @param action
     * @param result
     * @returns {{executed: Boolean}}
     */
  setHiresThumbnailFactor(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to getSourceThumbnail
     * @param action
     * @param result
     * @returns {{deviceId: String, channelId: String, src: String}}
     */
  getSourceThumbnail(action, result) {
    return {
      deviceId: action.$.deviceId,
      channelId: action.$.channelId,
      src: getValue(result, 'Image.0.$.src'),
    };
  },

  /**
     * processes response to takeSnapshot
     * @param action
     * @param result
     * @returns {{compositorId: String, snapshotId: String}}
     */
  takeSnapshot(action, result) {
    return {
      compositorId: action.$.compositorId,
      snapshotId: result.Snapshot[0].$.snapshotId,
    };
  },

  /**
     * processes response to downloadSnapshot
     * @param action
     * @param result
     * @returns {{snapshotId: String, src: String}}
     */
  downloadSnapshot(action, result) {
    return {
      snapshotId: result.Snapshot[0].$.snapshotId,
      src: result.Image[0].$.src,
    };
  },

  /**
     * processes response to deleteSnapshot
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  deleteSnapshot(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to showSource
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  showSource(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to hideSource
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  hideSource(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to updateSource
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  updateSource(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to setSourceTitle
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  setSourceTitle(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to refreshSource
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  refreshSource(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to setVolumePercentageForSource
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  setVolumePercentageForSource(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to muteSource
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  muteSource(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to unmuteSource
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  unmuteSource(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to showCompositorVisible
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  showCompositorVisible(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to getCustomLayoutData
     * @param action
     * @param layouts
     * @returns {[{"title":String,presentationId":(customLayout|autoLayout)}]}
     * e.g.
     * [
     *    {
     *        "title":"foo",
     *        "presentationId":"customLayout",
     *
     *    }
     * ]
     */
  getCustomLayoutData(action, layouts) {
    layouts = getValue(layouts, 'Layouts.0.Layout');
    const result = [];
    if (layouts) {
      for (const layout of layouts) {
        result.push({
          title: layout.$.title,
          presentationId: layout.$.presentationId,
          layoutId: layout.$.layoutId,
          items: layout.Insertion,
        });
      }
    }

    return result;
  },

  /**
     * processes response to addLayout
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  addLayout(action, result) {
    return result;
  },

  /**
     * processes response to addSource
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  addSource(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to removeSource
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  removeSource(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to lockBYODeviceContentSharing
     * @param action
     * @param result
     * @returns {{lock:boolean,executed: boolean}}
     */
  lockBYODeviceContentSharing(action, result) {
    return {
      lock: action.$.lock,
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to hideBYODSharingPincode
     * @param action
     * @param result
     * @returns {{hidden:boolean,executed: boolean}}
     */
  hideBYODSharingPincode(action, result) {
    return {
      hidden: action.$.value,
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to disconnectAllBYODevices
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  disconnectAllBYODevices(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to setBYODeviceAccessCode
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  setBYODeviceAccessCode(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to cloneLayout
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  cloneLayout(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to overrideSink
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  overrideSink(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to getSystemReadyStatus
     * @param action
     * @param result
     * @returns {boolean}
     */
  getSystemReadyStatus(action, result) {
    return parseBool(getValue(result, 'SystemState.0.$.systemIsReady'));
  },

  /**
     * processes response to selectCompositor
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  selectCompositor(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to unsubscribe
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  unsubscribe(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },

  /**
     * processes response to subscribe
     * @param action
     * @param result
     * @returns {{executed: boolean}}
     */
  subscribe(action, result) {
    return {
      executed: result.$.actionExecuted,
    };
  },
};

export default CommandResponseProcessor;
