/**
 * (C) 2015 Copyright BARCO NV
 * All rights reserved. Reproduction in whole or in part is
 * prohibited without the written consent of the copyright owner.
 *
 * Part of Barco WeConnect
 */
import { URL } from 'url';

import serviceRegistry from '../../../services/serviceRegistry';
import {
  StaticPreviewProvider,
} from '../../../services/previews/provider';
import SnapshotProvider from '../../../services/snapshots/provider';
import {
  sinkType,
  compositionType,
  commands,
} from '../../../services/ddp/constants';
import MeteorCacheManager from '../../../services/ddp/meteorCacheManager';
import Bridge from '../bridge';

/**
 * A preview provider manager is a friend-class of the MeteorBridge,
 *  and relives the meteor cngBridge of the management of the preview providers related to one NRC.
 */
class PreviewBridge extends Bridge {
  constructor(options) {
    super(options);

    this._cngBridge = options.cngBridge;

    this._previewProviders = {
      sources: new Map(),
      sinks: new Map(),
      snapshots: new Map(),
    };

    this.cache = new MeteorCacheManager(this, ['compositions', 'sinks', 'sources']);

    this._log('verbose', 'New PreviewBridge created.');

    this.validatePreviewRequest = this._validatePreviewRequest.bind(this);
  }

  get server() {
    return serviceRegistry.get('previewServer');
  }

  get supportedCompositionTypes() {
    return [compositionType.CINERGY];
  }

  get supportedSinkTypes() {
    return [sinkType.MJPEG_STREAM];
  }

  get serviceName() {
    return 'Cinergy/PreviewBridge';
  }

  get supportedRPCCommands() {
    return [commands.DISCONNECT_PREVIEW_CLIENTS];
  }

  _updateServerHostname() {
    this.server.hostname = this._device.dns || this._device.publicIp || this._device.privateIp;
  }

  /**
   * Incoming preview requests may only be served if the user knows the meeting id.
   *
   * @param req
   * @param res
   * @returns {boolean}
   * @private
   */
  _validatePreviewRequest(req) {
    this._log('silly', `Validating incoming request. Device meeting id: ${this._device.meetingId}, request meeting id: ${req.query.accessToken}`);
    if (this._device.meetingId) {
      return req.query && (req.query.accessToken === this._device.meetingId);
    }
    return true;
  }

  /**
   * Called as part of the meteor cngBridges initalization procedure.
   * This will set up the initial set of static and dynamic (source and sink) preview providers
   */
  async _onStart() {
    this._updateServerHostname();
    await super._onStart();
    this._subscription = this._onPreviewReceived.bind(this);
    this._cngBridge._cngClient.subscribeForPreviewThumbnailNotification(this._subscription, 3000);  // eslint-disable-line
  }

  /**
   * Stop this manager
   */
  async _onStop() {
    this._cngBridge._cngClient.off('previewThumbnailChanged', this._subscription); // eslint-disable-line

    this._log('verbose', `Destroying ${this._previewProviders.sources.size} source providers`);
    this._previewProviders.sources.forEach((provider, sourceId) => {
      let meteorSource = this.cache.sources[sourceId];
      if (!meteorSource) {
        this._log('error', `Could not find meteor source for source id: ${sourceId}`);
        meteorSource = {
          _id: sourceId,
        };
      }
      this._destroySourcePreviewProvider(meteorSource);
    });

    this._log('verbose', `Destroying ${this._previewProviders.sinks.size} sink providers`);
    this._previewProviders.sinks.forEach((provider, sinkId) => {
      let meteorSink = this.cache.sinks[sinkId];
      if (!meteorSink) {
        this._log('error', `Could not find meteor sink for sink id: ${sinkId}`);
        meteorSink = {
          _id: sinkId,
        };
      }
      this._destroySinkPreviewProvider(meteorSink);
      this._destroySnapshotProvider(meteorSink);
    });

    await super._onStop();
  }

  /**
   * Act on incoming client RPCs
   * RPC's are useful for one-time actions that are not part of the application state.
   *
   * @param rpc
   * @returns {*}
   * @private
   */
  _onRpcAdded(rpc) {
    this._log('verbose', `Handling rpc ${JSON.stringify(rpc, null, ' ')}`, { serviceLog: true });
    switch (rpc.command) {
    case commands.DISCONNECT_PREVIEW_CLIENTS:
      this.resetAllConnections();
      break;
    default:
    }
  }

  /*-------------------------------------
   * Meteor device changes
   *-------------------------------------*/
  /**
   * Act on device change
   * @param device
   * @param oldFields
   * @param clearedFields
   * @param newFields
   * @returns {*}
   * @private
   */
  _onThisDeviceChanged(device, oldFields, clearedFields, newFields) {
    const promises = [];
    if (newFields.hasOwnProperty('dns') || newFields.hasOwnProperty('publicIp') || newFields.hasOwnProperty('mjpegServer')) {
      this._log('info', 'Change mjpeg server, restarting');
      promises.push(this.restart());
    }
    return Promise.all(promises);
  }

  /**
   * Source added => add the source to CNG and create a new preview provider
   * @param source
   * @private
   */
  _onSourceAdded(source) {
    return this._createSourcePreviewProviderIfNeeded(source);
  }

  /**
   * Source removed => remove it from CNG if it exists
   *
   * @param source
   * @returns {Promise}
   * @private
   */
  _onSourceRemoved(source) {
    return this._destroySourcePreviewProvider(source);
  }

  //  _onCompositionAdded(composition){
  //     const promises = this.cache.findActiveSinksForComposition(composition)
  //     .map(sink => this._onSinkAdded(sink));

  //   return Promise.all(promises);
  // }

  // _onCompositionRemoved(composition){
  //   const promises = this.cache.findActiveSinksForComposition(composition)
  //     .filter(sink => this._previewProviders.sinks.has(sink._id))
  //     .map(sink => this._onSinkRemoved(sink));

  //   return Promise.all(promises);
  // }

  _onSinkAdded(sink) {
    return this._createSinkPreviewProviderIfNeeded(sink);
  }

  _onSinkChanged(sink, oldFields, clearedFields, newFields) {
    const promises = [];
    if (this._previewProviders.sinks.has(sink._id)
      && this._anyChanged(['activeCompositionId', 'port', 'url', 'width', 'height', 'framerate'], newFields)) {
      promises.push(this._onSinkRemoved(sink).then(() => this._onSinkAdded(sink)));
    }
    return Promise.all(promises);
  }

  _onSinkRemoved(sink) {
    this._destroySinkPreviewProvider(sink);
    this._destroySnapshotProvider(sink);
  }

  /**
   * Routes the NRC preview notification to the correct preview provider
   * @param result
   * @private
   */
  _onPreviewReceived(result) {
    const source = this._cngBridge.cache.findCngSource(result);
    if (source && this._previewProviders.sources.has(source._id)) {
      this._log('silly', `New preview received for source: ${source._id}`);
      this._previewProviders.sources.get(source._id).previewDataPNG = result.src;
    }
  }

  /**
   * Closes all connections on the MJPEG providers
   */
  resetAllConnections() {
    this._previewProviders.sinks.forEach(provider => provider.closeConnections());
  }

  /**
   * Create a static preview provider for the given source.
   * @param source
   */
  _createSourcePreviewProviderIfNeeded(source) {
    // Only if source can be shown in some UI (not for internal sources like wallpapers)
    if (!this._previewProviders.sources.has(source._id) && source.exposeInUi) {
      this._log('verbose', `Creating new source preview provider for source ${source._id}`);
      const path = this._pathForSource(source);

      const provider = new StaticPreviewProvider({
        id: source._id,
        validateRequest: this.validatePreviewRequest,
      });
      provider.start();

      this._previewProviders.sources.set(source._id, provider);

      const url = this.server.addPreviewProviderIfNeeded(path, provider);

      if (!source.url || source.url !== url) {
        this._meteorClient.updateSourcePreviewUrl({
          sourceId: source._id,
          url,
        });
      }
    }
  }

  /**
   * Creates a dynamic preview provider for the given sink
   *
   * This will start the corresponding MJPEG server
   * @param sink
   */
  _createSinkPreviewProviderIfNeeded(sink) {
    if (sink.type !== sinkType.MJPEG_STREAM) {
      this._log('verbose', 'Not a streaming sink, skipping');
      return;
    }

    // Note: only JPEG used since chrome 86 (MJPEG part removed)
    if (!this._previewProviders.sinks.has(sink._id)) {
      this._log('verbose', `Creating new sink preview provider for sink ${sink._id}`);
      if (this.cache.findCompositionOnSink(sink)) {
        this._createSnapshotProvider(sink);
      }
    }
  }

  _createSnapshotProvider(sink) {
    this._log('verbose', `Creating new snapshot provider for sink ${sink._id}`);
    const path = this._pathForSnapshot(sink);

    const composition = this.cache.findCompositionOnSink(sink);

    const snapshotProvider = new SnapshotProvider({
      id: sink._id,
      validateRequest: this.validatePreviewRequest,
      composition: `screen${composition && composition.index > 0 ? composition.index : ''}`,
    });
    snapshotProvider.start();

    this._previewProviders.snapshots.set(sink._id, snapshotProvider);

    this.server.addPreviewProviderIfNeeded(path, snapshotProvider);
  }

  /**
   * Destroys the sink preview provider for the given sink
   *
   * This will stop the corresponding MJPEG server
   * @param sink
   */
  _destroySinkPreviewProvider(sink) {
    this._log('debug', `Deleting sink preview provider: ${sink._id}`);
    if (this._previewProviders.sinks.has(sink._id)) {
      this._previewProviders.sinks.get(sink._id).destroy();
      this._previewProviders.sinks.delete(sink._id);
    }
  }

  /**
   * Destroys the source preview provider for the given source
   *
   * @param source
   */
  _destroySourcePreviewProvider(source) {
    this._log('debug', `Deleting source preview provider: ${source._id}`);
    if (this._previewProviders.sources.has(source._id)) {
      this._previewProviders.sources.get(source._id).destroy();
      this._previewProviders.sources.delete(source._id);
    } else {
      this._log('error', `Deleting source preview provider failed (not found): ${source._id}`);
    }
  }

  _destroySnapshotProvider(sink) {
    this._log('debug', `Deleting snapshot preview provider: ${sink._id}`);
    if (this._previewProviders.snapshots.has(sink._id)) {
      this._previewProviders.snapshots.get(sink._id).destroy();
      this._previewProviders.snapshots.delete(sink._id);
    }
  }

  _pathForSource(source) {
    return `${source._id}.png`;
  }

  _pathForSink(sink) {
    try {
      const url = new URL(sink.url);
      return url.pathname;
    } catch (error) {}
  }

  _pathForSnapshot(sink) {
    try {
      const url = new URL(sink.url);
      let [file, ext] = url.pathname.split('.');
      file += 'snp';
      return [file, ext].join('.');
    } catch (error) {}
  }

  /**
   * Destroys and cleans up this instance.
   * Will stop the NRC subscription for preview changes and tear down all providers and corresponding MJPEG servers
   */
  _onDestroy() {
    this._log('verbose', 'Destroying');

    delete this._previewProviders;
    delete this._cngBridge;
    delete this._subscription;
    delete this.validatePreviewRequest;

    return super._onDestroy();
  }
}

export default PreviewBridge;
