/**
 * (C) 2015 Copyright BARCO NV
 * All rights reserved. Reproduction in whole or in part is
 * prohibited without the written consent of the copyright owner.
 *
 * Part of Barco WeConnect
 */
/*
eslint no-underscore-dangle: ["error", {
  "allow": ["_attr"],
  "allowAfterThis": true,
  "allowAfterSuper": true,
}]
*/

import { Socket } from 'net';
import { parseString as parseXML } from 'xml2js';
import * as buildXML from 'xml';
import * as tls from 'tls';
import * as os from 'os';
import * as _ from 'lodash';

import CommandResponseProcessor from './commandResponseProcessor';
import ProxiedService from '../../../services/proxiedService';
import ProcessManager from '../../../services/processManager';
import CinergyError from './cinergyError';

const barcoCerts = {
  rejectUnauthorized: false,
  // key: fs.readFileSync('../cert/ICBarcoXDSKey.pem'),
  // cert: fs.readFileSync('../cert/ICBarcoXDS.pem'),
  // ca: [fs.readFileSync('../cert/ICBarcoXDSCaCert.pem')], checkServerIdentity: (host, cert) => true;}
};

/**
 * Some helper methods
 */

/**
 * Parses a string or num to boolean
 * @param val
 * @returns {boolean}
 */
function parseBool(val) {
  if (typeof val === 'undefined') {
    return false;
  }

  const num = +val;
  return !Number.isNaN(num)
    ? !!num
    : !!String(val)
      .toLowerCase()
      .replace(!!0, '');
}

/**
 * Cng API Wrapper, this class is a JS interface to the XML api of Cingergy.
 *
 * This class must be stateless (e.g. not cache CNG info) since it does not know about it's consumers.
 *
 * In fact, if it would be decided to take purple out of the NRC, this could still run on the NRC's and be opened
 * by a JSON-RPC api (see jrpc-schema), so nothing else needs to change in purple.
 *
 * Also, in the other direction, this class could act as an abstraction layer above the NRC API as that one changes.
 *
 * This wrapper deals with the NRC API network connection and offers a promise-based interface for all NRC API calls.
 * It implements a queue commands to an NRC and process them one by one.
 *
 * There is some special behavior towards subscriptions. Subscriptions are handled by emitting events,
 * since it is a more javascript/node way of handling events.
 *
 * Furthermore, subscriptions are actively bundled.
 * E.g. you can have 20 subscribers to an event, but only one subscription to the NRC api event.
 * This reduces traffic and processing time.
 *
 * This class implements a finite state machine to handle states when it is not connected,
 * connecting, connected and authenticated to the CNG API.
 *
 * Note, that ideally we'd want purple to be able to manage multiple NRC's.
 * Therefore all functionality should be remote available.
 *
 * Currently this is not the case, since the PulseAudio, RTSP and MJPEG servers
 * are not part of the API and executes code that must run on the device itself.
 */
class CinergyClient extends ProxiedService {
  /**
   * Create a new instance of CinergyClient
   * @param ipAddress ipAddress of CNG to connect to (default: "127.0.0.1")
   * @param hostname hostname of CNG to connect to
   * @param port port on CNG to connect to (default: 5010)
   * @param useTLS whether to use a secure connection (default: true)
   * @param autoReconnect whether to auto reconnect if connection fails or is lost (default:false)
   * @param autoReconnectTimer time between reconnection attempts (default: 10s).
   */
  constructor({
    ipAddress = '127.0.0.1',
    hostname,
    port = 5010,
    useTLS = true,
    autoReconnect = false,
    autoReconnectTimer = 3000,
    username = '',
    password = '',
  }) {
    if (!hostname) {
      if (ipAddress === '127.0.0.1') {
        hostname = os.hostname();
      } else {
        hostname = ipAddress;
      }
    }

    super({
      id: `${hostname}:${port}`,
    });

    const params = {
      ipAddress,
      hostname,
      port,
      useTLS,
      autoReconnect,
      autoReconnectTimer,
      username,
      password,
    };
    Object.keys(params).forEach((param) => {
      if (params[param] === undefined) { throw new Error(`Please provide a valid ${param}`); }
      this[`_${param}`] = params[param];
    });

    // Queue of commands waiting to be send to Cinergy API
    this._cmdQueue = [];

    // Last executed command
    this._currentCommand = undefined;

    this.isAuthenticated = false;
    this.isCinergyReady = false;
    this.isConnected = false;
    this.isConfiguring = false;

    this._startRootTag = '<Root>';
    this._endRootTag = '</Root>';

    this._processManager = new ProcessManager({
      id: `Cinergy ${this._serviceId}`,
      startCommands: [
        'sudo service barco-cinergy-iconnectserver start',
        'sudo service barco-cinergy-icdeviceserver start',
        'sudo service barco-cinergy-iccontrolsysteminterface start',
        'sudo service barco-cinergy-icgpucompositor start',
      ],
      stopCommands: [
        'sudo service barco-cinergy-icgpucompositor stop',
        'sudo service barco-cinergy-iccontrolsysteminterface stop',
        'sudo service barco-cinergy-icdeviceserver stop',
        'sudo service barco-cinergy-iconnectserver stop',
      ],
      timeoutMs: 10000,
      maxRetries: 1,
    });

    this.retries = 0;
    this.MAX_RETRIES = 20;
  }

  get cinergyAuthenticated() {
    return this.isAuthenticated;
  }

  get readyToSubscribe() {
    return this.isCinergyReady;
  }

  get readyToSubscribeEvent() {
    return 'cinergyReady';
  }

  set isCinergyReady(value) {
    if (value === this._isCinergyReady) {
      return;
    }

    this._isCinergyReady = value;
    if (this._isCinergyReady) {
      this._log('verbose', 'Cinergy is ready. Emitting event to activate proxied subscriptions.');
      this.emit('cinergyReady');
    }
  }

  get isCinergyReady() {
    return this._isCinergyReady;
  }

  get proxyMap() {
    return {
      activeLayoutChanged: 'activeLayoutChanged',
      sourcesChanged: 'sourcesChanged',
      layoutsChanged: 'layoutsChanged',
      layoutRuntimeDataChanged: 'layoutRuntimeDataChanged',
      previewThumbnailChanged: 'previewThumbnailChanged',
      selectedCompositionChanged: 'selectedCompositionChanged',
      sourceAudioChanged: 'sourceAudioChanged',
      displayPowerStateChanged: 'displayPowerStateChanged',
      systemReadyStateChanged: 'systemReadyStateChanged',
    };
  }

  get minimumSupportedVersion() {
    return CinergyClient.minimumSupportedVersion;
  }

  get hostname() {
    return this._hostname;
  }

  /**
   * Destroys all connection-related instance stuff when the connection is closed/lost.
   * Destroys mjpeg servers, rtsp servers, rejects active listeners & method calls etc.
   *
   * Will schedule a reconnection if required.
   * @private
   */
  _cleanupOnSocketClosed() {
    if (this.isConnected) {
      const tryReconnect = this._autoReconnect && this.isActive;
      this._log(
        'info',
        `

    =================================================================================
    ==                                                                             ==
    == Connection lost. Cleaning up ${
  tryReconnect
    ? 'and triggering reconnect timer'
    : '                               '
}                 ==
    ==                                                                             ==
    =================================================================================

            `,
      );

      // Set status variables
      this.isConnected = false;
      this.isAuthenticated = false;
      this.isCinergyReady = false;

      // Get rid of the socket
      if (this._clientSock) {
        // this._clientSock.end();
        this._clientSock.removeAllListeners();
        this._clientSock.destroy();
        delete this._clientSock;
      }
      this._clientSock = undefined;

      // Reject all pending promises
      try {
        this._rejectAll();
      } catch (error) {
        this._log('error', 'Error while rejecting commands', error);
      }

      this.emit('disconnected');
      // Clean the receive buffer
      delete this._buffer;

      if (tryReconnect) {
        this._scheduleReconnect();
      }
    }
  }

  _scheduleReconnect() {
    this._log('info', 'Scheduling re-initialization');
    this._clearReconnectTimer();
    this._reconnectTimer = setTimeout(() => {
      this._init().catch((err) => {
        this._log('warn', `Reconnection failed with error ${err}`);
      });
    }, this._autoReconnectTimer);
  }

  _clearReconnectTimer() {
    if (this._reconnectTimer) {
      clearTimeout(this._reconnectTimer);
      delete this._reconnectTimer;
    }
  }

  get _isQEmpty() {
    return this._cmdQueue.length === 0;
  }

  /**
   * Builds NRC API XML command and sends it to the NRC iConnect interface
   * @param cmdObj - JSOn representation of the command to send
   * @private
   */
  _executeCommand(cmdObj) {
    const commandStr = buildXML([
      {
        Barco: [cmdObj],
      },
    ]);

    if (cmdObj.Action[0]._attr.command === 'authenticate') {
      // hide backdoor login authentication details from log files, which are visible in CNG Web UI
      const censoredCmdObj = JSON.parse(JSON.stringify(cmdObj));
      censoredCmdObj.Action[0]._attr.clientUserName = 'HIDDEN';
      censoredCmdObj.Action[0]._attr.clientPassword = 'HIDDEN';

      this._log(
        'verbose',
        `EXECUTING COMMAND -> ${buildXML([{ Barco: [censoredCmdObj] }])}`,
      );
    } else {
      this._log('verbose', `EXECUTING COMMAND -> ${commandStr}`);
    }

    // Send the command over the socket
    if (this._clientSock && this.isConnected && !this.isConfiguring) {
      this._clientSock.write(commandStr);
    } else {
      this._log('error', 'Cannot write command to disconnected socket!');
    }
  }

  /**
   * NRC can only handle 1 command at once
   * therefore we store commands in a queue. This method will take the first command
   * from the queue and execute it. After execution, this method will be invoked again.
   * @private
   */
  _processQueue() {
    if (!this._isQEmpty) {
      this._currentCommand = this._cmdQueue[0];
      this._log(
        'debug',
        `PROCESS QUEUE FOR COMMAND >>>>> ${this._currentCommand.name}`,
      );
      this._executeCommand(this._currentCommand.command);
    } else {
      this._log('debug', 'Command queue empty !!');
    }
  }

  /**
   * Adds a command to the processing queue. It will kick-off the processing if the queue was empty
   * @param command - JSON version of the command to add
   * @returns {Promise}
   * @private
   */
  _addCmdToQueue(command) {
    return new Promise((resolve, reject) => {
      if (!this.isConnected) {
        reject(new Error('Not connected'));
        return;
      }

      if (
        !this.isAuthenticated
        && command.Action[0]._attr.command !== 'authenticate'
      ) {
        reject(new Error('Not authenticated'));
        return;
      }

      const queueObj = {
        name: command.Action[0]._attr.command,
        command,
        resolve,
        reject,
      };

      const startProcessing = this._isQEmpty;

      this._cmdQueue.push(queueObj);
      this._log(
        'debug',
        `Added cmd ${queueObj.name} to queue and the queue length now is: ${
          this._cmdQueue.length
        }`,
      );

      if (startProcessing) {
        // Q was empty, need to start processing again.
        this._processQueue();
      }
    });
  }

  async _onStart() {
    await super._onStart();
    await this._init();
  }

  startServices() {
    this._processManager.start();
  }

  stopServices() {
    return this._processManager.stop();
  }

  restartServices() {
    return this._processManager.restart();
  }

  async _init() {
    await this.connect();
    if (this._username && this._password) {
      await this.authenticate(this._username, this._password);
    }
  }

  /**
   * Connect to the NRC API socket
   * @returns {Promise}
   */
  connect() {
    if (!this.isActive) {
      return Promise.reject(new Error('Service inactive, not connecting'));
    }

    // Already connected
    if (this.isConnected === true) {
      this._log(
        'warn',
        `Socket already connected to Cinergy at ${this._ipAddress}:${
          this._port
        }. Returning.`,
      );
      return Promise.resolve('already connected');
    }

    this.isConfiguring = false;

    // Connect
    return new Promise((resolve, reject) => {
      if (this._reconnectTimer) {
        clearTimeout(this._reconnectTimer);
      }

      const onConnected = () => {
        if (this.isDestroyed) {
          reject(new Error('Service destroyed during connect'));
        }

        if (this._useTLS && !this._clientSock.encrypted) {
          // Something may be wrong with your certificates
          this._log('warn', 'Failed to establish TLS connection');
          this._log(
            'warn',
            `TLS connection error : ${this._clientSock.authorizationError}`,
          );
          this._reconnectTimer = setTimeout(() => {
            this.connect().catch((err) => {
              this._log('warn', `Reconnection failed with error ${err}`);
            });
          }, this._autoReconnectTimer);
          reject(new Error('Failed to establish TLS connection'));
        }

        this._log('info', 'Connected to Cinergy');
        this.isConnected = true;
        this.emit('connected');
        resolve(new Error('Connected to Cinergy'));
      };

      if (!this._useTLS) {
        this._clientSock = new Socket().connect(
          this._port,
          this._ipAddress,
          onConnected.bind(this),
        );
      } else {
        this._clientSock = tls.connect(
          this._port,
          this._ipAddress,
          barcoCerts,
          onConnected.bind(this),
        );
      }

      this._clientSock.on('error', (error) => {
        if (this.isConnected) {
          // Also log as verbose because Winston does not respect logging order
          this._log('verbose', 'This socket is ended. Cleaning up.', error);
          this._log('error', 'This socket is ended. Cleaning up', error);
          this._cleanupOnSocketClosed();
        } else {
          this._log('warn', 'Could not connect to Cinergy', error);
          this._reconnectTimer = setTimeout(() => {
            this.connect().catch((err) => {
              this._log('warn', `Reconnection failed with error ${err}`);
            });
          }, this._autoReconnectTimer);
          reject(new Error(error));
        }
      });

      // this._clientSock.on('end', data => {
      //    this._log('warn', `This socket is ended. Cleaning up.`);
      //    this._cleanupOnSocketClosed(true);
      // });

      this._clientSock.on('close', () => {
        this._log('info', 'The socket has been closed. Cleaning up.');
        this._cleanupOnSocketClosed();
      });

      this._clientSock.on('data', (data) => {
        if (this.isActive) {
          this._processReceivedData(data);
        }
      });
    });
  }

  disconnect() {
    this._rejectAll();
    if (this.isConnected) {
      const promise = new Promise((resolve) => {
        this.on('disconnected', resolve);
      });
      this._clientSock.end();
      return promise;
    }
    return Promise.resolve(true);
  }

  // START NRC API Commands
  // Below methods are a more-or-less 1-on-1 mapping of the NRC API
  // For documentation, we refer to the official NRC API manual.

  async authenticate(username, password) {
    if (this.isAuthenticated) {
      return true;
    }

    const authObj = {
      Action: [
        {
          _attr: {
            command: 'authenticate',
            clientUserName: username,
            clientPassword: password,
          },
        },
      ],
    };
    await this._addCmdToQueue(authObj);
    // Previous line throws on failure so won't run next lines
    this.isAuthenticated = true;
    return true;
  }

  getCompositions() {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'getCompositions',
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  getSinkOverrideStatus() {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'getSinkOverrideStatus',
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  getAllSources() {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'getSources',
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  getLayoutsForCompositor(compositorId) {
    if (!compositorId) {
      compositorId = 'screen';
    }
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'getLayouts',
            compositorId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  getLayouts() {
    return this.getLayoutsForCompositor();
  }

  getSourceThumbnail(deviceId, channelId) {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'getSourceThumbnail',
            channelId,
            deviceId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  getSourcesForCompositor(compositorId = 'screen') {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'getSources',
            compositorId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  getSources(compositorId) {
    if (compositorId) {
      return this.getSourcesForCompositor(compositorId);
    }
    return this.getAllSources();
  }

  getDisplayInformation() {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'getDisplayInformation',
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  getVersionInfo() {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'getVersionInfo',
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  getSystemInfo() {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'getSystemInfo',
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  deactivateLayout(compositorId = 'screen', hideAllSources = true) {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'deactivateLayout',
            compositorId,
            hideAllSources,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  getActiveLayout(compositorId = 'screen') {
    this._log(
      'verbose',
      `Getting active layout for compositorId = ${compositorId}`,
    );
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'getActiveLayout',
            compositorId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  getLayoutRuntimeData(compositorId = 'screen') {
    this._log(
      'verbose',
      `Getting layout runtime data for compositorId = ${compositorId}`,
    );
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'getLayoutRuntimeData',
            compositorId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  getCustomLayouts(compositorId = 'screen') {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'getCustomLayoutData',
            compositorId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  updateSourceURL(deviceId, channelId, attrs) {
    attrs = _.pick(attrs, [
      'width',
      'height',
      'fps',
      'proxy',
      'proxyUser',
      'proxyPassword',
      'url',
      'urlUser',
      'urlPassword',
    ]);
    this._log(
      'verbose',
      `Setting CEF source with deviceId = ${deviceId}, channelId = ${channelId} to ${JSON.stringify(attrs)}`,
    );
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'updateSource',
            saveInConfiguration: true,
            deviceId,
            channelId,
          },
        },
        {
          Source: [
            {
              _attr: attrs,
            },
          ],
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  updateSourceTitle(deviceId, channelId, title) {
    this._log(
      'verbose',
      `Updating source title with deviceId = ${deviceId}, channelId = ${channelId} to ${title}`,
    );
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'setSourceTitle',
            deviceId,
            channelId,
            title,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  refreshSource(deviceId, channelId) {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'refreshSource',
            deviceId,
            channelId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  takeSnapshot(compositorId) {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'takeSnapshot',
            compositorId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  lockBYODeviceContentSharing() {
    this._log('verbose', 'Locking BYOD sharing functionality.');
    return this._lockBYODeviceContentSharing(true);
  }

  unlockBYODeviceContentSharing() {
    this._log('verbose', 'Unlocking BYOD sharing functionality.');
    return this._lockBYODeviceContentSharing(false);
  }

  _lockBYODeviceContentSharing(lock) {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'lockBYODeviceContentSharing',
            lock,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  enableBYODModeration() {
    this._log('verbose', 'Enabling BYOD moderation functionality.');
    return this._enableBYODModeration(true);
  }

  disableBYODModeration() {
    this._log('verbose', 'Disabling BYOD moderation functionality.');
    return this._enableBYODModeration(false);
  }

  _enableBYODModeration(value) {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'enableBYODModeration',
            value,
            saveInConfiguration: true,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  hideBYODSharingPincode(compositorId) {
    this._log('verbose', 'Hiding BYOD sharing functionality.');
    return this._hideBYODSharingPincode(compositorId, true);
  }

  showBYODSharingPincode(compositorId) {
    this._log('verbose', 'Showing BYOD sharing functionality.');
    return this._hideBYODSharingPincode(compositorId, false);
  }

  _hideBYODSharingPincode(compositorId, value) {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'hideBYODSharingPincode',
            value,
            compositorId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  disconnectAllBYODevices() {
    this._log('verbose', 'Sending command to disconnect all BYOD devices');
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'disconnectAllBYODevices',
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  showIdentificationMessage(timeoutMs) {
    this._log('verbose', `Showing identification message for ${timeoutMs} ms`);
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'showIdentificationMessage',
            timeoutMs,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  setBYODAccessCode(code = '', dongleCode = '') {
    let cmdObj;
    if (code && dongleCode) {
      cmdObj = {
        Action: [
          {
            _attr: {
              command: 'setBYODeviceAccessCode',
              code,
              dongleCode,
            },
          },
        ],
      };
    } else if (code) {
      cmdObj = {
        Action: [
          {
            _attr: {
              command: 'setBYODeviceAccessCode',
              code,
            },
          },
        ],
      };
    } else {
      cmdObj = {
        Action: [
          {
            _attr: {
              command: 'setBYODeviceAccessCode',
              dongleCode,
            },
          },
        ],
      };
    }

    return this._addCmdToQueue(cmdObj);
  }

  subscribeForPreviewThumbnailNotification(callback, intervalMs = 5000) {
    this._thumbnailUpdateIntervalMs = intervalMs;
    this.on('previewThumbnailChanged', callback);
  }

  _subscribeForNotification(subEvent) {
    let cmdObj;
    switch (subEvent) {
    case 'previewThumbnailChanged':
      cmdObj = {
        Action: [
          {
            _attr: {
              command: 'subscribe',
              notificationType: 'previewThumbnailChanged',
              updateInterval: this._thumbnailUpdateIntervalMs,
            },
          },
        ],
      };
      break;
    default:
      cmdObj = {
        Action: [
          {
            _attr: {
              command: 'subscribe',
              notificationType: subEvent,
            },
          },
        ],
      };
    }
    return this._addCmdToQueue(cmdObj);
  }

  _unsubscribeForNotification(subEvent) {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'unsubscribe',
            notificationType: subEvent,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  showSource({
    deviceId, channelId, compositorId, showFullScreen,
  }) {
    this._log(
      'verbose',
      `========> Launching source with device id = "${deviceId}", channel id = "${channelId}", fullscreen = ${showFullScreen}`,
    );

    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'showSource',
            compositorId,
            channelId,
            deviceId,
            showFullScreen,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  hideSource(deviceId, channelId, compositorId) {
    this._log(
      'verbose',
      `========> Hiding source with device id = "${deviceId}", channel id = "${channelId}"`,
    );

    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'hideSource',
            compositorId,
            channelId,
            deviceId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  activateLayout(layout, ...compositorId) {
    compositorId = compositorId.join(';');
    this._log(
      'verbose',
      `Activating layout with layout title = "${layout}" on compositors with ids = "${compositorId}"`,
    );
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'activateLayout',
            compositorId,
            title: layout,
          },
        },
      ],
    };
    this._addCmdToQueue(cmdObj);
  }

  removeLayout(layoutTitle) {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'removeLayout',
            title: layoutTitle,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  setVolumePercentageForSource(deviceId, channelId, value) {
    return this.setSourceVolume(deviceId, channelId, value);
  }

  setSourceVolume(deviceId, channelId, value) {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'setVolumePercentageForSource',
            channelId,
            deviceId,
            value,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  muteSource(deviceId, channelId) {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'muteSource',
            channelId,
            deviceId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  unmuteSource(deviceId, channelId) {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'unmuteSource',
            channelId,
            deviceId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  // Mute all the sources which are onscreen

  muteAllSources() {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'muteSource',
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  // Brings back the onscreen sources to their previous audio status

  showCompositor() {
    this._log('info', 'Showing compositor');
    return this.toggleCompositor(true);
  }

  hideCompositor() {
    this._log('info', 'Hiding compositor');
    return this.toggleCompositor(false);
  }

  toggleCompositor(show) {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'setCompositorVisible',
            value: show,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  unmuteAllSources() {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'unmuteSource',
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  addLayout(layout) {
    const formattedLayout = [
      {
        _attr: _.omit(layout, ['elements']),
      },
    ];

    for (const element of layout.elements) {
      switch (element.type) {
      case 'Insertion':
        formattedLayout.push({
          Insertion: [
            {
              _attr: _.omit(element, ['type', 'SourceRef']),
            },
            {
              SourceRef: [
                {
                  _attr: element.SourceRef,
                },
              ],
            },
          ],
        });
        break;
      case 'Region': // XXX: needs work
        formattedLayout.push({
          Region: [
            {
              _attr: _.omit(element, ['type']),
            },
          ],
        });
        break;
      default:
        throw new Error('Invalid element in layout');
      }
    }

    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'addLayout',
          },
        },
        {
          Layout: formattedLayout,
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  addSource(source) {
    source.command = 'addSource';
    const cmdObj = {
      Action: [
        {
          _attr: source,
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  removeSource(deviceId, channelId) {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'removeSource',
            channelId,
            deviceId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  setThumbnailResolutionFactor(factor) {
    this._log('verbose', `Setting thumbnail resolution factor to = ${factor}`);
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'setHiresThumbnailFactor',
            value: factor,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  cloneLayout(srcCompositorId, dstCompositorId) {
    this._log(
      'verbose',
      `Cloning composition "${srcCompositorId}" to composition "${dstCompositorId}"`,
    );
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'cloneLayout',
            srcCompositorId,
            dstCompositorId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  async configureNRC(config, reboot = true) {
    return this._addCmdToQueue(this._deviceConfigToJSONXML(config, reboot));
  }

  _deviceConfigToJSONXML(config, autoReboot) {
    const xmljson = {
      Action: [
        {
          _attr: {
            command: 'configureNRC',
            autoReboot,
          },
        },
      ],
    };

    // add action attrs
    const actionAttrs = ['streamDestinationIP', 'streamDestinationHost'];
    actionAttrs.forEach((attr) => {
      if (!config[attr]) return;
      xmljson.Action[0]._attr[attr] = config[attr];
    });

    // add general settings (room name)
    if (config.connectionName) {
      xmljson.Action.push({
        GeneralSettings: [
          {
            Room: [
              {
                _attr: {
                  roomName: config.connectionName,
                },
              },
            ],
          },
        ],
      });
    }
    // add compositions
    const compositions = {
      Compositions: [
        {
          _attr: config.compositionParameters,
        },
      ],
    };
    config.compositions.forEach((composition) => {
      const sink = composition.sinks;
      delete composition.sinks;
      compositions.Compositions.push({
        Composition: [
          {
            _attr: composition,
          },
          {
            Sink: [
              {
                _attr: sink[0],
              },
            ],
          },
        ],
      });
    });
    xmljson.Action.push(compositions);

    // add sources
    const sources = {
      Sources: [],
    };
    config.sources.forEach((source) => {
      sources.Sources.push({
        Source: [
          {
            _attr: source,
          },
        ],
      });
    });
    xmljson.Action.push(sources);

    return xmljson;
  }

  async getSystemReadyStatus() {
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'getSystemReadyStatus',
          },
        },
      ],
    };
    const result = await this._addCmdToQueue(cmdObj);
    this.isCinergyReady = result;
    return result;
  }

  disableSinkOverride() {
    return this.overrideSink({
      srcDeviceId: '0',
      srcChannelId: '0',
      trgDeviceId: '0',
      trgChannelId: '0',
    });
  }

  enableSinkOverride(params) {
    return this.overrideSink(params);
  }

  overrideSink({
    srcDeviceId, srcChannelId, trgDeviceId, trgChannelId,
  }) {
    const enableOverride = srcDeviceId !== trgDeviceId || srcChannelId !== trgChannelId;

    if (!enableOverride) {
      this._log('verbose', 'Disabling sink override.');
    } else {
      this._log(
        'verbose',
        `Overriding sink with deviceId = ${trgDeviceId}, channelId = ${trgChannelId} with sink with deviceId = ${srcDeviceId}, channelId = ${srcChannelId}`,
      );
    }

    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'overrideSink',
            enabled: enableOverride,
          },
        },
        {
          SourceSink: [
            {
              _attr: {
                deviceId: srcDeviceId,
                channelId: srcChannelId,
              },
            },
          ],
        },
        {
          TargetSink: [
            {
              _attr: {
                deviceId: trgDeviceId,
                channelId: trgChannelId,
              },
            },
          ],
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }

  setSelectedComposition(compositorId) {
    this._log(
      'verbose',
      `Setting currently selected composition to "${compositorId}"`,
    );
    const cmdObj = {
      Action: [
        {
          _attr: {
            command: 'selectCompositor',
            compositorId,
          },
        },
      ],
    };
    return this._addCmdToQueue(cmdObj);
  }
  // END OF NRC API Commands

  /**
   * Process incoming notification data (parse to JSON).
   * @param data
   * @private
   */
  _processNotificationData(data) {
    let json = {};
    const action = data.Notify[0].Action[0];
    const { notificationType } = action.$;
    const skipNotification = false;

    this._log(
      'verbose',
      `Processing notification of "${notificationType}" subscription`,
    );

    switch (notificationType) {
    case 'activeLayoutChanged':
      this._log(
        'verbose',
        '=========>  Received Active Layout Changed notification',
      );
      json = {
        compositorId: action.$.compositorId,
        title: action.$.title,
        layoutId: action.$.layoutId,
        presentationId: action.$.presentationId,
        activeLayout: action.$.activeLayout,
        isNotificationResult: true,
      };
      break;
    case 'sourcesChanged':
      this._log('debug', 'Calling sources changed callback');
      json.compositorId = action.$.compositorId;
      json.numberSourcesAdded = action.$.numberSourcesAdded;
      json.numberSourcesRemoved = action.$.numberSourcesRemoved;
      break;
    case 'layoutsChanged':
      json.compositorId = action.$.compositorId;
      break;
    case 'systemReadyStateChanged':
      json.isReady = parseBool(action.$.systemIsReady);
      break;
    case 'layoutRuntimeDataChanged':
      json.compositorId = action.$.compositorId;
      break;
    default:
      this._log(
        'info',
        `Received notification for unknown event: ${notificationType}`,
      );
    }

    if (!skipNotification) {
      this.emit(notificationType, json);
    }
  }

  /**
   * Send the command response to the command response processor and resolve or reject the relevant promise.
   * @param data
   * @private
   */
  _processCommandResponse(data) {
    const action = data.Notify[0].Action[0];
    const cmdStr = action.$.command;
    this._log('silly', `Processing response to "${cmdStr}" command`, data);

    try {
      const result = CommandResponseProcessor.process(this, data);

      if (cmdStr === 'configureNRC') {
        this.isConfiguring = true;
      }

      if (cmdStr !== 'subscribe') {
        this._currentCommand.resolve(result);
      }
    } catch (error) {
      this._log(
        'error',
        `Error while processing response to "${cmdStr}". Command send: ${JSON.stringify(this._currentCommand.command)}`,
        error,
      );
      this._currentCommand.reject(error);
    }
  }

  /**
   * Try processing a thumbnail without parsing the XML.
   * This is an optimization because the XML parsing eats up a lot of memory on larger XML strings.
   *
   * @param command
   * @returns {boolean}
   * @private
   */
  _tryProcessThumbnailWithoutParsing(command) {
    const indexOfPreviewThumbnailChanged = command.indexOf('previewThumbnailChanged');
    if (indexOfPreviewThumbnailChanged === -1) {
      return false;
    }

    const startIndexOfDeviceId = command.indexOf('deviceId="');
    const startIndexOfChannelId = command.indexOf('channelId="');
    if (startIndexOfDeviceId === -1 || startIndexOfChannelId === -1) {
      return false;
    }

    const endIndexOfDeviceId = command.indexOf('"', startIndexOfDeviceId + 10);
    const endIndexOfChannelId = command.indexOf('"', startIndexOfChannelId + 11);
    if (
      endIndexOfDeviceId === -1
      || endIndexOfDeviceId <= startIndexOfDeviceId
      || endIndexOfChannelId === -1
      || endIndexOfChannelId <= startIndexOfChannelId
    ) {
      this._log(
        'error',
        `Received invalid thumbnail, device Id or channeld Id missing. Message = ${command}`,
      );
      return false; // it's not a new incoming thumbnail
    }

    const deviceId = command.slice(
      startIndexOfDeviceId + 10,
      endIndexOfDeviceId,
    );
    const channelId = command.slice(
      startIndexOfChannelId + 11,
      endIndexOfChannelId,
    );

    this._log(
      'silly',
      `Received thumbnail update for deviceId ${deviceId} and channelId ${channelId}`,
    );
    if (deviceId && channelId) {
      const startIndexOfImage = command.indexOf('previewData=');
      const endIndexOfImage = command.lastIndexOf('"/>');
      // If the image content is more than 100 chars, consider it as a valid thumbnail.

      const json = {
        deviceId,
        channelId,
      };

      if (
        endIndexOfImage - (startIndexOfImage + 13) > 100
        && startIndexOfImage !== -1
        && endIndexOfImage !== -1
      ) {
        json.src = command.substring(startIndexOfImage + 13, endIndexOfImage);
      } else {
        this._log('silly', 'Received empty thumbnail.');
      }
      this.emit('previewThumbnailChanged', json);
    }
    return true;
  }

  /**
   * Process raw incoming socket data and sends it to the notification or command processor accordingly.
   * Since it turns out that the XML parsing eats up a lot of memory, we bypass this for thumbnail updates.
   *
   * @private
   */
  _processReceivedData(data) {
    this._buffer += data.toString();
    let nbBarco = this._buffer.indexOf('</Barco>');

    while (nbBarco !== -1 && this._buffer) {
      const command = this._startRootTag
        + this._buffer.slice(0, nbBarco + 8)
        + this._endRootTag;
      const newBuffer = {
        data: this._buffer.slice(nbBarco + 8),
      };

      delete this._buffer;
      this._buffer = newBuffer.data;
      delete newBuffer.data;

      // Since it turns out that the XML parsing eats up a lot of memory, we bypass this for thumbnail updates.
      if (!this._tryProcessThumbnailWithoutParsing(command)) {
        parseXML(command, (err, received) => {
          for (const dataNode of received.Root.Barco) {
            if (dataNode.Acknowledge) {
              this.retries = 0;
              this._log(
                'debug',
                `5) Received acknowledgement for command : ${
                  this._currentCommand.name
                }`,
              );
            } else if (dataNode.Notify) {
              if (dataNode.Notify[0].$.commandNumber === '-1') {
                // Received notification, send to notification processor
                this._log('debug', '5) Received Result for a notification');
                this._processNotificationData(dataNode);
              } else {
                // Received command response, send to command processor
                this._log(
                  'debug',
                  `5) Received Result for a command : ${
                    dataNode.Notify[0].Action[0].$.command
                  }`,
                );
                this._processCommandResponse(dataNode);
                this._cmdQueue.shift();
                this._processQueue();
              }
            } else if (dataNode.Busy) {
              // The NRC API designer chose not to implement a queue on the NRC side.
              // Therefore it is possible the NRC returns 'BUSY'.
              // In this case, try again until a max number of retries has been reached.

              this._log(
                'warn',
                `5) Received Busy (nb:${this.retries}) for a command : ${
                  dataNode.Busy[0].$.currentCommandNumber
                }`,
              );
              if (this.retries >= this.MAX_RETRIES) {
                this._rejectAll();
                this._log(
                  'warn',
                  `5) Maximum amount of retries (${this.MAX_RETRIES}) exceeded`,
                );
              } else {
                setTimeout(() => {
                  this.retries += 1;
                  this._executeCommand(this._currentCommand.command);
                }, this.retries * 10);
              }
            }
          }
        });
      }

      if (this._buffer) {
        // since parseXML is async, this might have been deleted in the meantime
        nbBarco = this._buffer.indexOf('</Barco>');
      } else {
        nbBarco = -1;
      }
    }
  }

  /**
   * Reject all outstanding command promises
   * @private
   */
  _rejectAll(filter = () => true, includeCurrent = true) {
    this._log('verbose', 'Rejecting all outstanding commands');
    // Break all ongoing promises for commands
    if (
      includeCurrent
      && this._currentCommand
      && filter(this._currentCommand)
    ) {
      this._currentCommand.reject(new CinergyError(0, 'Rejected by command'));
    }

    this._cmdQueue.filter(filter).forEach((cmd) => {
      cmd.reject(new CinergyError(0, 'Rejected by command'));
    });

    this._cmdQueue = [];
  }

  async _onStop() {
    await super._onStop();
    await this.disconnect();
  }

  /**
   * Destroyer, will tear down the connection and clean up the memory.
   */
  async _onDestroy() {
    delete this._display;
    await this._processManager.destroy();
    await super._onDestroy();
  }
}
CinergyClient.minimumSupportedVersion = '1.1.1.14';

export default CinergyClient;
