/**
 * (C) 2015 Copyright BARCO NV
 * All rights reserved. Reproduction in whole or in part is
 * prohibited without the written consent of the copyright owner.
 *
 * Part of Barco WeConnect
 */
import * as _ from 'lodash';
import * as Rx from 'rxjs';
import { debounceTime } from 'rxjs/operators';

import Bridge from '../bridge';
import DisplayService from './x11Service';
import { commands } from '../../../services/ddp/constants';
import { translateDisplayBounds } from '../../../utils/display';
import { state as serviceState } from '../../../services/service';

/**
 * The Display Bridge glues blue and the display service.
 * From blue->device it will set the display configuration as required by blue
 * From device->blue it updates the attached displays
 */
class DisplayBridge extends Bridge {
  constructor(options) {
    super(options);
    this._displayService = new DisplayService();
    this._observable = Rx.fromEvent(this._displayService, 'displaysChanged')
      .pipe(debounceTime(1000 /* ms */))
      .subscribe(() => {
        this._updateDisplays();
      });
  }

  get supportedRPCCommands() {
    return [commands.IDENTIFY_SCREENS];
  }

  async _onStart() {
    await super._onStart();
    this._log('debug', 'Initialization started');
    await this._displayService.start({ cancelToken: this.getCancelToken(serviceState.STARTED) });
    this._log('debug', 'Triggering initial _updateDisplays');
    await this._updateDisplays(true);
    this._log('debug', 'Initial _updateDisplays completed');
    await this._configureTouchScreens();
    this._log('debug', 'Completed initialization');
  }

  async _onStop() {
    await super._onStop();
    await this._displayService.stop();
  }

  _onThisDeviceChanged(device, oldFields, clearedFields, newFields) {
    const promises = [];

    if (newFields.hasOwnProperty('displayConfig')) {
      promises.push(this._configureTouchScreens());
    }

    if (newFields.hasOwnProperty('displays')) {
      promises.push(this._configureDisplays(this._device.displays));
    }

    return Promise.all(promises);
  }

  /**
   * Act on incoming client RPCs
   * RPC's are useful for one-time actions that are not part of the application state.
   *
   * @param rpc
   * @returns {*}
   * @private
   */
  _onRpcAdded(rpc) {
    this._log('verbose', `Handling rpc ${JSON.stringify(rpc, null, ' ')}`, { serviceLog: true });
    switch (rpc.command) {
    case commands.IDENTIFY_SCREENS:
      this.identify();
      break;
    default:
    }
  }

  async identify() {
    const displays = await this._getDisplays();

    return Promise.all(displays.map(async display => this._showMessage({
      text: display.port,
      bounds: await translateDisplayBounds({
        x: display.x,
        y: display.y,
        width: display.activeMode.width,
        height: display.activeMode.height,
      }),
      time: 25,
      flash: true,
    })));
  }

  _configureTouchScreens() {
    // To get correct input coordinates on a touch screen
    // we should tell xinput to which display the touch screen panel belongs.
    this._log('verbose', 'Checking if we need to set a touchscreen input');
    const touchScreen = this._device.displayConfig.find(d => d.isTouchScreen);
    if (touchScreen) {
      this._log('info', `We need to set a touch screen on port: ${touchScreen.port}`);
      return this._displayService.setTouchScreen(touchScreen.port);
    }
    return true;
  }

  async _getDisplays() {
    this._log('silly', 'Fetching display information');
    let displays = (await this._displayService.getDisplayInformation({
      includeEdid: true,
    })).filter(display => display.connected);

    const extractMode = m => ({
      height: m.height,
      width: m.width,
      native: m.native,
      rate: m.rate,
    });

    const extractCurrentMode = (modes) => {
      const current = modes.find(m => m.current);
      return current ? extractMode(current) : undefined;
    };

    displays = displays.map(display => ({
      id: display.port,
      port: display.port,
      x: display.left,
      y: display.top,
      vendor: display.edid.vendor || display.edid.eisaId,
      model: display.edid.modelName,
      productCode: display.edid.productCode,
      serial: display.edid.serialNumber,
      activeMode: extractCurrentMode(display.modes),
      modes: display.modes.map(extractMode),
    }));

    return displays;
  }

  _configureDisplays(displays) {
    return this._displayService.configureDisplays(displays.map(d => ({
      port: d.port,
      x: d.x,
      y: d.y,
      width: d.activeMode.width,
      height: d.activeMode.height,
      refreshRate: d.activeMode.rate,
    })));
  }

  /**
   * This will try to find a new best placement for the screens.
   * Say the displays are locked in the admin UI and there are 3 screens configured in this order:
   * HDMI2, DP1, HDMI3
   *
   * In case the DP1 scren gets unplugged and re-plugged, it will regain its old configuration.
   *
   * Also if a screen moves from one slot to another, which was previously empty, it will keep its configuration
   *
   * @param {*} newDisplays
   */
  _enhancePlacement(newDisplays) {
    const hasMode = (modes, mode) => modes.find(m => m.width === mode.width && m.height === mode.height && m.rate === mode.rate);

    // const findExactMatch = (displays, display) => displays.find(d => d.serial === display.serial);
    // const findProductMatch = (displays, display) => displays
    //  .find(d => d.productCode === display.productCode && d.vendor === display.vendor);

    const getNativeMode = display => display.modes.find(m => m.native) || {
      width: 1024,
      height: 768,
      rate: 59.99,
    };

    const sanitizeDisplayPositions = (displays) => {
      displays.sort((a, b) => a.order - b.order);

      let x = 0;
      displays.forEach((d) => {
        d.x = x;
        d.y = 0;
        x += d.activeMode.width;
        delete d.order;
      });

      return displays;
    };

    const currentDisplays = [...this._device.displays];
    currentDisplays.sort((a, b) => a.x - b.x);

    newDisplays.forEach((display, index) => {
      const current = currentDisplays.find(d => d.id === display.id);
      if (current) {
        // Port was already utilized => try to keep the same mode and position
        if (hasMode(display.modes, current.activeMode)) {
          display.activeMode = current.activeMode;
        }
        display.order = currentDisplays.indexOf(current);
      } else if (display.serial) {
        // Screen moved to empty slot => keep mode and possibly position
        const previous = currentDisplays.find(d => d.serial === display.serial);
        if (previous) {
          // Try to keep the same mode
          // (should be possible since it's the same serial, but you never know with these EDIDs)
          if (hasMode(display.modes, previous.activeMode)) {
            display.activeMode = previous.activeMode;
          }

          // If no other display has taken the old one's port,
          // re-use the same position
          if (!newDisplays.find(d => d.id === previous.id)) {
            display.order = currentDisplays.indexOf(current);
          }
        }
      }

      // If it doesn't get an order according to the previous display configuraiton,
      // move it to the right
      if (!display.hasOwnProperty('order')) {
        display.order = Math.max(newDisplays.length, currentDisplays.length) + index;
      }

      // Make sure we always force a display mode
      if (!display.activeMode) {
        display.activeMode = getNativeMode(display);
      }
    });

    // The above algorithm works with relative positions (order), now apply the absolute positions
    return sanitizeDisplayPositions(newDisplays);
  }

  async _updateDisplays(force) {
    const startTimestamp = new Date();
    const isDifferent = (a, b) => {
      if (!a || !b) {
        return true;
      }

      const sanitizeForComparision = (screens) => {
        screens = JSON.parse(JSON.stringify(screens));
        screens.forEach((screen) => {
          if (screen.activeMode) {
            delete screen.activeMode.native;
          }
        });
        return _.keyBy(screens, 'id');
      };

      a = sanitizeForComparision(a);
      b = sanitizeForComparision(b);
      return !_.isEqual(a, b);
    };

    const displays = await this._getDisplays();
    if (!this.isActive) { return; }
    if (isDifferent(displays, this._device.displaysLatest) || force) {
      // Sometimes we need to force it, e.g. on bridge startup since the display settings might have changed
      // This will not be reflected in the displaysLatest untill we force set the new enhancements
      this._log('verbose', `Change in connected displays, old displays: ${JSON.stringify(this._device.displaysLatest)}`, { serviceLog: true });
      this._log('verbose', `Change in connected displays, new displays: ${JSON.stringify(displays)}`, { serviceLog: true });

      // Choose placement for screens, according to previous settings (e.g. locked screen config)
      let enhancedDisplays = JSON.parse(JSON.stringify(displays));
      enhancedDisplays = this._enhancePlacement(enhancedDisplays);

      // Apply new placement
      if (isDifferent(enhancedDisplays, displays)) {
        this._log('verbose', `Setting displays after enhancements: ${JSON.stringify(enhancedDisplays, null, 2)}`);

        await this._configureDisplays(enhancedDisplays);
        if (!this.isActive) { return; }
      } else {
        this._log('verbose', 'No better display placement found');
      }

      // Might not be different anymore after placing screens in prefered positions
      if (isDifferent(enhancedDisplays, this._device.displaysLatest)) {
        await this._upsertDisplays(enhancedDisplays);
      } else {
        this._log('verbose', 'Displays do not differ anymore after enhancing placement => not upserting');
      }
    } else {
      this._log(
        'silly',
        `No change detected in connected displays, milliseconds since start of call ${new Date() - startTimestamp}`,
      );
    }
  }

  async _upsertDisplays(displays) {
    this._logger.log('info', 'Upserting displays', displays);
    // We save the displays latest here as a form of latency compensation
    // so that in the time between our upsert call and the server sending us the updated info
    // We don't send multiple updates in the time between this upsert call and
    // the server sending us the update.
    // However, we set the displaysLatest to displaysPrevious if we want to still check what
    // has changed when we get an update from the server.
    this._device.displaysPrevious = this._device.displaysLatest;
    this._device.displaysLatest = displays;
    return this._meteorClient.updateDevice({
      _id: this._device._id,
      displaysLatest: displays,
    });
  }

  async _onDestroy() {
    this._displayService.off();
    await this._displayService.destroy();
    delete this._displayService;
    this._observable.unsubscribe();
    delete this._observable;
    await super._onDestroy();
  }
}

export default DisplayBridge;
