/*
eslint no-await-in-loop: "off", no-cond-assign: "off"
*/

/**
 * (C) 2015 Copyright BARCO NV
 * All rights reserved. Reproduction in whole or in part is
 * prohibited without the written consent of the copyright owner.
 *
 * Part of Barco WeConnect
 */
import * as childProcess from 'child_process';
import { screen } from 'electron';
import * as pify from 'pify';
import * as path from 'path';
import * as EdidReader from 'edid-reader';
import { serializeError } from 'serialize-error';

import Service from '../../../services/service';
import { sleep, waitMaximum } from '../../../utils/utils';
import { SCRIPTS_DIR } from '../../../files';

const edidReader = new EdidReader();

// It takes a while for electron to get updated screen coordinates
// If electron is not yet updated, but we already changed the sinks
// then the mapping in browserbridge is going to be wrong
// So wait for electron to get updated
// However, electron screen events are a bit flaky, so put in a maximum wait of 1s.
function metricsUpdated() {
  const updatePromise = new Promise((resolve) => {
    screen.once('display-metrics-changed', () => {
      resolve();
    });
  });
  return waitMaximum(updatePromise, 1000, false);
}

/**
 * Abstract Display Management service
 * Implements the commonalities between the different types
 * of display management API's like x11 or the electron one.
 */
class DisplayService extends Service {
  // Make it easy to stub the exec function
  static exec = pify(childProcess.exec);

  async getDisplayInformation(options) {
    const xrandrOutput = await this._runXrandrTillStable();
    return this._transformXrandrOutput(xrandrOutput, options);
  }

  // On some systems xrandr sometimes shows a connected screen as disconnected.
  // Usually on the next check it is shown correctly again.
  // So for safety we run xrandr till we get the same response 3 times, with a maximum of 10 attempts
  async _runXrandrTillStable() {
    let previous;
    let identicalResponses = 0;
    let i = 0;

    try {
      for (; i < 10 && this.isActive && identicalResponses < 2; i++) {
        const xrandr = await this._runXrandr();
        if (previous && previous === xrandr) {
          identicalResponses += 1;
        } else {
          identicalResponses = 0;
        }

        // Wait a bit between xrandr attempts
        await sleep(300);
        previous = xrandr;
      }
    } catch (error) {
      this._log('error', 'Xrandr error', error);
    }

    if (identicalResponses === 2) {
      return previous;
    }
    if (i === 10) {
      this._log('error', 'Xrandr did not stabilize after 10 attempts');
    }
    throw new Error('Could not determine xrandr output');
  }

  _runXrandr() {
    return DisplayService.exec('DISPLAY=:0.0 xrandr -q --verbose');
  }

  /**
   * Maps a touch input device to a camera port.
   * Uses an external script that uses xinput to achieve this.
   * Note that it's kinda sketchy since we just find the first touch device
   * So it's not guaranteed to be correct when there's more than 1 touch device, but it's 6PM and fuck it.
   */
  async setTouchScreen(port) {
    try {
      this._log('info', `Mapping input device to display port ${port}`);
      const scriptPath = path.join(SCRIPTS_DIR, 'mapTouchInputToDisplayOutput.sh');
      await DisplayService.exec(`DISPLAY=:0.0 ${scriptPath} ${port}`);
    } catch (error) {
      if (this._logger) {
        this._log('error', `Error while mapping touch input on display port ${port}`, error);
      }
    }
  }

  _buildDisplaySetting({
    port,
    x,
    y,
    width,
    height,
    refreshRate,
    rotation = 'normal',
  }) {
    // Important: we explicitely disable panning as it does not respect the requested display position
    return `--output ${port} --mode ${width}x${height} --pos ${x}x${y} --panning ${width}x${height}+${x}+${y} --rate ${refreshRate} --rotate ${rotation}`;
  }

  static calculateRequiredFramebufferSize(displays) {
    return displays.reduce((acc, dp) => ({
      width: Math.max(acc.width, dp.x + dp.width),
      height: Math.max(acc.height, dp.y + dp.height),
    }), { width: 0, height: 0 });
  }

  async configureDisplays(displays) {
    let retryApplyFBSize = false;

    // Calculate and apply the required framebuffer size
    const { width: fbWidth, height: fbHeight } = DisplayService.calculateRequiredFramebufferSize(displays);
    const setFrameBufferSize = `DISPLAY=:0.0 xrandr --fb ${fbWidth}x${fbHeight}`;
    this._log('info', `Setting framebuffer size to ${fbWidth}x${fbHeight}. Executing: ${setFrameBufferSize}`);
    try {
      await DisplayService.exec(setFrameBufferSize);
    } catch (error) {
      // The command can fail in case the requested framebuffer size is smaller
      // then the current active display size. In that case we execute the command
      // once more AFTER changing (=reducing) the display sizes.
      retryApplyFBSize = true;
      this._log('error', `Error while applying framebuffer size. Will try again after setting displays sizes. Error = ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`);
    }

    // Apply resolutions and refresh rates of individual displays
    const command = `DISPLAY=:0.0 xrandr ${displays.map(this._buildDisplaySetting).join(' ')}`;
    this._log('info', `Setting display resolutions: 
    ${displays.map(({
    port, x, y, width, height, refreshRate,
  }) => `${port} => Resolution: ${width}x${height}, Position: ${x}x${y}, RefreshRate: ${refreshRate}`).join('\n')}
    Executing: ${command}`);
    await DisplayService.exec(command);

    // If needed, re-apply the requested framebuffer size
    if (retryApplyFBSize) {
      this._log('info', `Retrying to set framebuffer size to ${fbWidth}x${fbHeight}. Executing: ${setFrameBufferSize}`);
      await DisplayService.exec(setFrameBufferSize);
    }
    await metricsUpdated();
  }

  /** This function parses an xrandr output with the following format
   * Screen 0: minimum 320 x 200, current 1080 x 1920, maximum 2048 x 2048
   * HDMI-1 connected 1080x1920+0+0 (0x44) right (normal left inverted right x axis y axis) 477mm x 268mm
   * 	Identifier: 0x42
   * 	Timestamp:  96798461
   * 	Subpixel:   unknown
   * 	Gamma:      1.0:1.0:1.0
   * 	Brightness: 1.0
   * 	Clones:
   * 	CRTC:       2
   * 	CRTCs:      2
   * 	Transform:  1.000000 0.000000 0.000000
   * 	            0.000000 1.000000 0.000000
   * 	            0.000000 0.000000 1.000000
   * 	           filter:
   * 	EDID:
   * 		00ffffffffffff001e6dd75801010101
   *   1920x1080 (0x44) 148.500MHz +HSync +VSync *current +preferred
   *         h: width  1920 start 2008 end 2052 total 2200 skew    0 clock  67.50KHz
   *         v: height 1080 start 1084 end 1089 total 1125           clock  60.00Hz
   *   1920x1080 (0x45) 148.500MHz +HSync +VSync
   *         h: width  1920 start 2448 end 2492 total 2640 skew    0 clock  56.25KHz
   *         v: height 1080 start 1084 end 1089 total 1125           clock  50.00Hz
   *
   * @param {*} data
   */
  _transformXrandrOutput(src, {
    includeVirtual = false,
    includeDisconnected = false,
    includeEdid = false,
    includeInterlaced = false,
    includeDimensions = false,
  } = {}) {
    const type = {
      connected: /^(\S+) connected(?: primary)? (?:(\d+)x(\d+)\+(\d+)\+(\d))*\s*(\(\w+\))*\s*(\w*)/,
      edid: /\tEDID:/,
      edid_block: /^\t\t([\da-f]{32})$/i,
      disconnected: /^(\S+) disconnected/,
      mode: /^\s+(\d+)x([0-9i]+)\s+(\(\w+\))\s+([0-9]+\.[0-9]+)MHz\s+((\s*[+]*[-]*\w*\s)*)/,
      dimension_horizontal: /^\s+h:\s+width\s+(\d+)\s+start\s+(\d+)\s+end\s+(\d+)\s+total\s+(\d+)\s*skew\s*(\w)\s*clock\s*([0-9]*.[0-9]*)/,
      dimension_vertical: /^\s+v:\s+height\s+(\d+)\s+start\s+(\d+)\s+end\s+(\d+)\s+total\s+(\d+)\s*clock\s*([0-9]*.[0-9]*)/,
    };

    const lines = src.split('\n');
    let result = [];
    let currentConnection = null;
    let currentMode = null;
    let index = 0;

    let processingEdid = false;
    let edidBlock = '';

    const compareModes = (a, b) => a.width === b.width && a.height === b.height && a.rate === b.rate;

    lines.forEach((line) => {
      let tmp;
      if ((tmp = type.connected.exec(line))) {
        let [_1, port, width, height, left, top, _7, orientation] = tmp;
        width = width || 0;
        height = height || 0;
        left = left || 0;
        top = top || 0;
        orientation = orientation || 'normal';
        index += 1;
        const connection = {
          port,
          connected: true,
          orientation,
          modes: [],
          index,
        };

        connection.width = Number.parseInt(width, 10);
        connection.height = Number.parseInt(height, 10);
        connection.left = Number.parseInt(left, 10);
        connection.top = Number.parseInt(top, 10);

        currentConnection = connection;
        result.push(currentConnection);
      } else if ((tmp = type.disconnected.exec(line))) {
        const [_0, port] = tmp;
        index += 1;
        currentConnection = {
          port,
          connected: false,
          modes: [],
          index,
        };

        result.push(currentConnection);
      } else if ((tmp = type.mode.exec(line))) {
        const [_0, width, height, _3, rate, optionals] = tmp;

        const mode = {
          name: `${width}x${height}`,
          width: Number.parseInt(width, 10),
          height: Number.parseInt(height, 10),
          rate: parseFloat(rate),
          optionals,
          current: !!line.includes('current'),
          native: !!line.includes('preferred'),
        };

        if (includeDimensions) {
          mode.dimensions = {
            vertical: null,
            horizontal: null,
          };
        }

        mode.interlaced = mode.name.includes('i');

        // Some crappy screens, E.g. Barco Screens have duplicates modes in their EDID
        // If you have 2 duplicate modes, one being 'native', the other one not being 'native'
        // you can get configuration loops. Filter them out.
        const duplicateModeId = currentConnection.modes.findIndex(m => compareModes(m, mode));
        if (duplicateModeId >= 0) {
          if (mode.current || mode.native) {
            currentConnection.modes[duplicateModeId] = mode;
          }
        } else {
          currentConnection.modes.push(mode);
        }

        currentMode = mode;
      } else if (includeDimensions && (tmp = type.dimension_horizontal.exec(line))) {
        const [_, width, start, end, total, skew, clock] = tmp;
        const dimensionH = {
          width: Number.parseInt(width, 10),
          start: Number.parseInt(start, 10),
          end: Number.parseInt(end, 10),
          total: Number.parseInt(total, 10),
          skew: Number.parseInt(skew, 10),
          clock: parseFloat(clock),
        };
        currentMode.dimensions.horizontal = dimensionH;
      } else if ((tmp = type.dimension_vertical.exec(line))) {
        const [_, width, start, end, total, clock] = tmp;
        const dimensionV = {
          width: Number.parseInt(width, 10),
          start: Number.parseInt(start, 10),
          end: Number.parseInt(end, 10),
          total: Number.parseInt(total, 10),
          clock: parseFloat(clock),
        };

        currentMode.rate = dimensionV.clock;

        if (includeDimensions) {
          currentMode.dimensions.vertical = dimensionV;
        }
      } else if (includeEdid && (tmp = type.edid.exec(line))) {
        processingEdid = true;
        edidBlock = '';
      } else if (processingEdid) {
        if ((tmp = type.edid_block.exec(line))) {
          edidBlock += tmp[0].trim();
        } else {
          processingEdid = false;
          currentConnection.edid = EdidReader.parse(edidReader.formatEdid(edidBlock));
          edidBlock = '';
        }
      }
    });

    if (!includeDisconnected) {
      result = result.filter(d => d.connected);
    }

    if (!includeVirtual) {
      result = result.filter(d => !d.port.includes('VIRTUAL'));
    }

    if (!includeInterlaced) {
      result.forEach((d) => {
        d.modes = d.modes.filter(m => !m.interlaced);
      });
    }

    return result;
  }
}

export default DisplayService;
