import * as chai from 'chai';
import * as sinon from 'sinon';
import * as sinonChai from 'sinon-chai';
import DisplayService from './service';
import { registerAll, unregisterAll } from '../../../utils/test-helpers/fillServiceRegistry';

const { expect, assert } = chai;
chai.use(sinonChai);

describe('Base Display Service', () => {
  before(() => {
    registerAll();
  });

  after(() => {
    unregisterAll();
  });

  describe('Static function to calculate frame buffer size', () => {
    it('Should calculate the right framebuffer size', () => {
      const displays = [
        { port: 'DP-0', x: 0, y: 1080, width: 1920, height: 1080, refreshRate: 60 },
        { port: 'DP-1', x: 1920, y: 0, width: 1920, height: 1080, refreshRate: 60 },
      ];
      assert.deepEqual(DisplayService.calculateRequiredFramebufferSize(displays), { width: 3840, height: 2160 });
    });
  });

  describe('Configure displays', () => {
    let service;
    let execStub;

    before(() => {
      service = new DisplayService();
      execStub = sinon.stub(DisplayService, 'exec');
    });

    after(() => {
      execStub.restore();
    });

    beforeEach(() => {
      execStub.resetHistory();
    });

    it('Should configure the displays in case of expanding framebuffer', async () => {
      execStub.resolves();
      const displays = [
        { port: 'DP-0', x: 0, y: 0, width: 1920, height: 1080, refreshRate: 60 },
        { port: 'DP-1', x: 1920, y: 0, width: 1920, height: 1080, refreshRate: 60 },
      ];
      await service.configureDisplays(displays);
      expect(DisplayService.exec).to.be.calledTwice;
      const firstCall = DisplayService.exec.getCall(0);
      expect(firstCall).to.be.calledWithExactly('DISPLAY=:0.0 xrandr --fb 3840x1080');
      const secondCall = DisplayService.exec.getCall(1);
      expect(secondCall).to.be.calledWithExactly('DISPLAY=:0.0 xrandr --output DP-0 --mode 1920x1080 --pos 0x0 --panning 1920x1080+0+0 --rate 60 --rotate normal --output DP-1 --mode 1920x1080 --pos 1920x0 --panning 1920x1080+1920+0 --rate 60 --rotate normal');
    });

    it('Should configure the displays in case of contracting framebuffer', async () => {
      execStub.onFirstCall().rejects(); // Fails if you try to reduce buffer size to size smaller than current display size.
      execStub.onSecondCall().resolves();
      execStub.onThirdCall().resolves();
      const displays = [
        { port: 'DP-0', x: 0, y: 0, width: 1680, height: 1050, refreshRate: 60 },
        { port: 'DP-1', x: 1680, y: 0, width: 1680, height: 1050, refreshRate: 60 },
      ];
      await service.configureDisplays(displays);
      expect(DisplayService.exec).to.have.callCount(3);
      const firstCall = DisplayService.exec.getCall(0);
      expect(firstCall).to.be.calledWithExactly('DISPLAY=:0.0 xrandr --fb 3360x1050');
      const secondCall = DisplayService.exec.getCall(1);
      expect(secondCall).to.be.calledWithExactly('DISPLAY=:0.0 xrandr --output DP-0 --mode 1680x1050 --pos 0x0 --panning 1680x1050+0+0 --rate 60 --rotate normal --output DP-1 --mode 1680x1050 --pos 1680x0 --panning 1680x1050+1680+0 --rate 60 --rotate normal');
      const thirdCall = DisplayService.exec.getCall(2);
      expect(thirdCall).to.be.calledWithExactly('DISPLAY=:0.0 xrandr --fb 3360x1050');
    });
  });
});

// import * as sinon from 'sinon';
// import { rewire, unwire } from '../../../utils/test-helpers/rewire';
// import { registerAll } from '../../../utils/test-helpers/fillServiceRegistry';

// describe('Base Display Service', () => {
//   let service;
//   before(() => {
//     registerAll();
//     service = new DisplayService();
//   });

//   after(() => {});

//   describe('Test subscriptions', () => {
//     it('Should parse xrandr output correctly', async () => {
//       const xrandrOutput1 = `Screen 0: minimum 320 x 200, current 3926 x 1440, maximum 8192 x 8192
//       XWAYLAND1 connected 1366x768+0+567 (0x25) normal (normal left inverted right x axis y axis) 280mm x 160mm
//         Identifier: 0x23
//         Timestamp:  167562
//         Subpixel:   unknown
//         Gamma:      1.0:1.0:1.0
//         Brightness: 0.0
//         Clones:
//         CRTC:       0
//         CRTCs:      0
//         Transform:  1.000000 0.000000 0.000000
//                     0.000000 1.000000 0.000000
//                     0.000000 0.000000 1.000000
//                    filter:
//         1366x768 (0x25) 84.750MHz -HSync +VSync *current +preferred
//               h: width  1366 start 1431 end 1567 total 1776 skew    0 clock  47.72KHz
//               v: height  768 start  771 end  781 total  798           clock  59.80Hz
//       XWAYLAND46 connected 2560x1440+1366+0 (0x24) normal (normal left inverted right x axis y axis) 620mm x 340mm
//         Identifier: 0x2fa
//         Timestamp:  167562
//         Subpixel:   unknown
//         Gamma:      1.0:1.0:1.0
//         Brightness: 0.0
//         Clones:
//         CRTC:       1
//         CRTCs:      1
//         Transform:  1.000000 0.000000 0.000000
//                     0.000000 1.000000 0.000000
//                     0.000000 0.000000 1.000000
//                    filter:
//         2560x1440 (0x24) 312.000MHz -HSync +VSync *current +preferred
//               h: width  2560 start 2752 end 3024 total 3488 skew    0 clock  89.45KHz
//               v: height 1440 start 1443 end 1448 total 1493           clock  59.91Hz`;
//       const parsedXrandrOutput = [
//         {
//           "port": "XWAYLAND1",
//           "connected": true,
//           "orientation": "67",
//           "modes": [
//             {
//               "name": "1366x768",
//               "width": 1366,
//               "height": 768,
//               "rate": 59.8,
//               "optionals": "-HSync +VSync ",
//               "current": true,
//               "native": true,
//               "interlaced": false
//             }
//           ],
//           "index": 1,
//           "width": 1366,
//           "height": 768,
//           "left": 0,
//           "top": 5
//         },
//         {
//           "port": "XWAYLAND46",
//           "connected": true,
//           "orientation": "normal",
//           "modes": [
//             {
//               "name": "2560x1440",
//               "width": 2560,
//               "height": 1440,
//               "rate": 59.91,
//               "optionals": "-HSync +VSync ",
//               "current": true,
//               "native": true,
//               "interlaced": false
//             }
//           ],
//           "index": 2,
//           "width": 2560,
//           "height": 1440,
//           "left": 1366,
//           "top": 0
//         }
//       ];
//       const parsedXrandrOutputWithDimensions = {

//       };
//       const parsedXrandrOutputWithEdid = {

//       };
//       const parsedXrandrOutputWithVirtual = {};
//       const parsedXrandrOutputWithInterlaced = {};
//       assert.strictEqual(service._transformXrandrOutput(xrandrOutput1), parsedXrandrOutput);
//       assert.strictEqual(service._transformXrandrOutput(xrandrOutput1, {includeVirtual: true}), parsedXrandrOutputWithVirtual);
//       assert.strictEqual(service._transformXrandrOutput(xrandrOutput1, {includeEdid: true}), parsedXrandrOutputWithEdid);
//       assert.strictEqual(service._transformXrandrOutput(xrandrOutput1, {includeInterlaced: true}), parsedXrandrOutputWithInterlaced);
//       assert.strictEqual(service._transformXrandrOutput(xrandrOutput1, {includeDimensions: true}), parsedXrandrOutputWithDimensions);
//       const xrandrOutput2 = `Screen 0: minimum 320 x 200, current 1920 x 1080, maximum 8192 x 8192
//       HDMI-1 disconnected primary (normal left inverted right x axis y axis)
//         Identifier: 0x42
//         Timestamp:  364903915
//         Subpixel:   unknown
//         Clones:
//         CRTCs:      0 1 2
//         Transform:  1.000000 0.000000 0.000000
//                     0.000000 1.000000 0.000000
//                     0.000000 0.000000 1.000000
//                    filter:
//         aspect ratio: Automatic
//           supported: Automatic, 4:3, 16:9
//         Broadcast RGB: Automatic
//           supported: Automatic, Full, Limited 16:235
//         audio: auto
//           supported: force-dvi, off, auto, on
//       DP-1 connected 1920x1080+0+0 (0x46) normal (normal left inverted right x axis y axis) 509mm x 286mm
//         Identifier: 0x43
//         Timestamp:  364903915
//         Subpixel:   unknown
//         Gamma:      1.0:1.0:1.0
//         Brightness: 1.0
//         Clones:     HDMI-2
//         CRTC:       0
//         CRTCs:      0 1 2
//         Transform:  1.000000 0.000000 0.000000
//                     0.000000 1.000000 0.000000
//                     0.000000 0.000000 1.000000
//                    filter:
//         EDID:
//           00ffffffffffff0010ac9f404c303932
//           1f170104a5331d783ae595a656529d27
//           105054a54b00714f8180a9c0d1c00101
//           010101010101023a801871382d40582c
//           4500fd1e1100001e000000ff004b3559
//           43353338323239304c0a000000fc0044
//           454c4c205032333134540a20000000fd
//           00384c1e5311010a2020202020200130
//           02031cf14f9005040302071601061112
//           1513141f2309070783010000023a8018
//           71382d40582c4500fd1e1100001e011d
//           8018711c1620582c2500fd1e1100009e
//           011d007251d01e206e285500fd1e1100
//           001e8c0ad08a20e02d10103e9600fd1e
//           11000018000000000000000000000000
//           0000000000000000000000000000003f
//         Broadcast RGB: Automatic
//           supported: Automatic, Full, Limited 16:235
//         audio: auto
//           supported: force-dvi, off, auto, on
//         1920x1080 (0x46) 148.500MHz +HSync +VSync *current +preferred
//               h: width  1920 start 2008 end 2052 total 2200 skew    0 clock  67.50KHz
//               v: height 1080 start 1084 end 1089 total 1125           clock  60.00Hz
//         1920x1080 (0x47) 148.500MHz +HSync +VSync
//               h: width  1920 start 2448 end 2492 total 2640 skew    0 clock  56.25KHz
//               v: height 1080 start 1084 end 1089 total 1125           clock  50.00Hz
//         1920x1080 (0x48) 148.352MHz +HSync +VSync
//               h: width  1920 start 2008 end 2052 total 2200 skew    0 clock  67.43KHz
//               v: height 1080 start 1084 end 1089 total 1125           clock  59.94Hz
//         1920x1080i (0x49) 74.250MHz +HSync +VSync Interlace
//               h: width  1920 start 2008 end 2052 total 2200 skew    0 clock  33.75KHz
//               v: height 1080 start 1084 end 1094 total 1125           clock  60.00Hz
//         1920x1080i (0x4a) 74.250MHz +HSync +VSync Interlace
//               h: width  1920 start 2448 end 2492 total 2640 skew    0 clock  28.12KHz
//               v: height 1080 start 1084 end 1094 total 1125           clock  50.00Hz
//         1920x1080i (0x4b) 74.176MHz +HSync +VSync Interlace
//               h: width  1920 start 2008 end 2052 total 2200 skew    0 clock  33.72KHz
//               v: height 1080 start 1084 end 1094 total 1125           clock  59.94Hz
//         1600x900 (0x4c) 108.000MHz +HSync +VSync
//               h: width  1600 start 1624 end 1704 total 1800 skew    0 clock  60.00KHz
//               v: height  900 start  901 end  904 total 1000           clock  60.00Hz
//         1280x1024 (0x4d) 135.000MHz +HSync +VSync
//               h: width  1280 start 1296 end 1440 total 1688 skew    0 clock  79.98KHz
//               v: height 1024 start 1025 end 1028 total 1066           clock  75.02Hz
//         1280x1024 (0x4e) 108.000MHz +HSync +VSync
//               h: width  1280 start 1328 end 1440 total 1688 skew    0 clock  63.98KHz
//               v: height 1024 start 1025 end 1028 total 1066           clock  60.02Hz
//         1152x864 (0x4f) 108.000MHz +HSync +VSync
//               h: width  1152 start 1216 end 1344 total 1600 skew    0 clock  67.50KHz
//               v: height  864 start  865 end  868 total  900           clock  75.00Hz
//         1280x720 (0x50) 74.250MHz +HSync +VSync
//               h: width  1280 start 1390 end 1430 total 1650 skew    0 clock  45.00KHz
//               v: height  720 start  725 end  730 total  750           clock  60.00Hz
//         1280x720 (0x51) 74.250MHz +HSync +VSync
//               h: width  1280 start 1720 end 1760 total 1980 skew    0 clock  37.50KHz
//               v: height  720 start  725 end  730 total  750           clock  50.00Hz
//         1280x720 (0x52) 74.176MHz +HSync +VSync
//               h: width  1280 start 1390 end 1430 total 1650 skew    0 clock  44.96KHz
//               v: height  720 start  725 end  730 total  750           clock  59.94Hz
//         1024x768 (0x53) 78.750MHz +HSync +VSync
//               h: width  1024 start 1040 end 1136 total 1312 skew    0 clock  60.02KHz
//               v: height  768 start  769 end  772 total  800           clock  75.03Hz
//         1024x768 (0x54) 65.000MHz -HSync -VSync
//               h: width  1024 start 1048 end 1184 total 1344 skew    0 clock  48.36KHz
//               v: height  768 start  771 end  777 total  806           clock  60.00Hz
//         800x600 (0x55) 49.500MHz +HSync +VSync
//               h: width   800 start  816 end  896 total 1056 skew    0 clock  46.88KHz
//               v: height  600 start  601 end  604 total  625           clock  75.00Hz
//         800x600 (0x56) 40.000MHz +HSync +VSync
//               h: width   800 start  840 end  968 total 1056 skew    0 clock  37.88KHz
//               v: height  600 start  601 end  605 total  628           clock  60.32Hz
//         720x576 (0x57) 27.000MHz -HSync -VSync
//               h: width   720 start  732 end  796 total  864 skew    0 clock  31.25KHz
//               v: height  576 start  581 end  586 total  625           clock  50.00Hz
//         720x480 (0x58) 27.027MHz -HSync -VSync
//               h: width   720 start  736 end  798 total  858 skew    0 clock  31.50KHz
//               v: height  480 start  489 end  495 total  525           clock  60.00Hz
//         720x480 (0x59) 27.000MHz -HSync -VSync
//               h: width   720 start  736 end  798 total  858 skew    0 clock  31.47KHz
//               v: height  480 start  489 end  495 total  525           clock  59.94Hz
//         640x480 (0x5a) 31.500MHz -HSync -VSync
//               h: width   640 start  656 end  720 total  840 skew    0 clock  37.50KHz
//               v: height  480 start  481 end  484 total  500           clock  75.00Hz
//         640x480 (0x5b) 25.200MHz -HSync -VSync
//               h: width   640 start  656 end  752 total  800 skew    0 clock  31.50KHz
//               v: height  480 start  490 end  492 total  525           clock  60.00Hz
//         640x480 (0x5c) 25.175MHz -HSync -VSync
//               h: width   640 start  656 end  752 total  800 skew    0 clock  31.47KHz
//               v: height  480 start  490 end  492 total  525           clock  59.94Hz
//         720x400 (0x5d) 28.320MHz -HSync +VSync
//               h: width   720 start  738 end  846 total  900 skew    0 clock  31.47KHz
//               v: height  400 start  412 end  414 total  449           clock  70.08Hz
//       HDMI-2 disconnected (normal left inverted right x axis y axis)
//         Identifier: 0x44
//         Timestamp:  364903915
//         Subpixel:   unknown
//         Clones:     DP-1
//         CRTCs:      0 1 2
//         Transform:  1.000000 0.000000 0.000000
//                     0.000000 1.000000 0.000000
//                     0.000000 0.000000 1.000000
//                    filter:
//         aspect ratio: Automatic
//           supported: Automatic, 4:3, 16:9
//         Broadcast RGB: Automatic
//           supported: Automatic, Full, Limited 16:235
//         audio: auto
//           supported: force-dvi, off, auto, on`;

//           assert.strictEqual(service._transformXrandrOutput(xrandrOutput2), parsedXrandrOutput2);

//     });

//     it('Should throw when the socket is not connected', async () => {
//       client._ddpClient.socket.readyState = Websocket.CLOSING;
//       try {
//         await client.subscribe()
//         assert.fail('did not throw');
//       } catch(error) {
//         assert(error.message === 'Not connected');
//       }
//     });

//     it('Should throw when the socket gets disconnected', (done) => {
//       client._isConnected = true;
//       client._ddpClient.socket.readyState = Websocket.OPEN;
//       client.subscribe('',[100], {}, '_sub', 200)
//       .then(assert.fail)
//       .catch((err) => {
//         assert.strictEqual(err, DDPClient.ERRORS.DISCONNECTED);
//         done();
//       });
//       client._isConnected = false;
//       client._ddpClient.emit('socket-close');
//     });

//     it('Should throw if the server does not respond in time', () => {
//       client._isConnected = true;
//       client._ddpClient.socket.readyState = Websocket.OPEN;
//       client.subscribe('',[300], {}, '_sub', 100).then(assert.fail)
//       .catch((err) => {
//         assert.strictEquals(err, DDPClient.ERRORS.DISCONNECTED)
//       });
//     });
//   });
// });
