import * as fs from 'fs-extra';
import * as path from 'path';
import { exec, execSync } from 'child_process';
import { promisify } from 'util';
import * as _ from 'lodash';
import Bridge from '../bridge';
import { deviceRoleLabels, commands } from '../../../services/ddp/constants';

const FB_BASE_DIR = '/etc/barco-nrc-td-agent-bit';
const FB_CONFIG_DIR = path.join(FB_BASE_DIR, 'config');
const FB_CONFIG_FILE = 'service.conf';

/**
 * The FluentbitBridge is responsible for writing the config file
 * of fluentbit. It always stops and restarts the service whenever
 * the config file is to be changed. Fluentbit is used to log device
 * health status to the TMX cloud service that can generate alerts.
 */
class FluentbitBridge extends Bridge {
  constructor(options) {
    super(options);
    this._fbConfig = {};
    this._fbConfigfullPath = path.join(FB_CONFIG_DIR, FB_CONFIG_FILE);
    this._updateFBConfigOngoing = false;
  }

  get supportedRPCCommands() {
    return [commands.UNPAIR];
  }

  _onRpcAdded(rpc) {
    this._log('verbose', `Handling rpc ${JSON.stringify(rpc, null, ' ')}`, { serviceLog: true });
    switch (rpc.command) {
    case commands.UNPAIR:
      this._log('info', 'Device is being unpaired. Disabling TMX if it was active.', { serviceLog: true });
      this._stopFBService();
      break;
    default:
    }
  }

  _ensureConfigDirExists() {
    [FB_BASE_DIR, FB_CONFIG_DIR].forEach((dir) => {
      try {
        const stats = fs.lstatSync(dir);
        if ((stats.mode & 0o777).toString(8) !== '755') {
          this._log('verbose', `Setting permissions for directory ${dir}`);
          execSync(`sudo chmod 755 ${dir}`);
        }
      } catch (error) {
        this._log('verbose', `Directory ${dir} does not exist. Creating it.`, error);
        try {
          execSync(`sudo mkdir ${dir}; sudo chmod 755 ${dir}; sudo chown barco ${dir}; sudo chgrp barco ${dir}`);
        } catch (error2) {
          this._log('error', `Directory ${dir} does not exist and could not be created. Error = ${error2.message}`, { serviceLog: true });
        }
      }
    });
  }

  _readFBConfig() {
    this._fbConfig = {};
    if (fs.existsSync(this._fbConfigfullPath)) {
      try {
        const config = fs.readFileSync(this._fbConfigfullPath, 'utf8');
        const lines = config.split('\n');
        lines.forEach((line) => {
          const match = line.match(/^(\S+)='((\S*\s*)*)'/);
          if (match) {
            this._fbConfig[match[1]] = (match.length > 2) ? match[2] : '';
          }
        });
      } catch (error) {
        this._log('error', `Could not parse fluentbit config file, falling back to empty config, error = ${error.message}`, { serviceLog: true });
      }
    } else {
      this._log('warn', 'Config file for fluentbit doesn\'t exist, falling back to empty config.');
    }
  }

  async _removeFBConfig() {
    this._log('verbose', 'Removing fluentbit config file (if existing)');
    await promisify(exec)(`sudo rm -f ${this._fbConfigfullPath}`);
  }

  _JSONtoFB(config) {
    let FBconfig = '';
    for (const key of Object.keys(config)) {
      FBconfig += `${key}='${config[key]}'\n`;
    }
    return FBconfig;
  }

  async _writeFBConfig(config) {
    this._log('verbose', 'Creating new fluentbit config file');
    await promisify(fs.writeFile)(this._fbConfigfullPath, this._JSONtoFB(config),
      {
        encoding: 'utf8',
        flag: 'w+',
        mode: 0o666,
      });
  }

  async _startFBService() {
    this._log('verbose', 'Starting fluentbit service and enabling auto startup');
    await promisify(exec)('sudo service barco-nrc-td-agent-bit start; sudo systemctl enable barco-nrc-td-agent-bit.service');
  }

  async _stopFBService() {
    this._log('verbose', 'Stopping fluentbit service (if running) and disabling auto startup');
    await promisify(exec)('sudo service barco-nrc-td-agent-bit stop; sudo systemctl disable barco-nrc-td-agent-bit.service');
  }

  async _updateFBConfigIfNeeded() {
    this._updateFBConfigOngoing = true;
    if (!this._device.tmx?.active) {
      try {
        this._log('verbose', 'TMX feature disabled: Removing config and not restarting fluentbit service');
        await this._stopFBService();
        await this._removeFBConfig();
        this._fbConfig = {};
      } catch (error) {
        this._log('error', `Some error occured while stopping and disabling fluentbit service, error = ${error.message}`, { serviceLog: true });
      }
    } else {
      const institute = this._meteorClient._ddpClient.collections.institutes
        && Object.values(this._meteorClient._ddpClient.collections.institutes).find(inst => inst._id === this._device.instituteId);
      const location = this._meteorClient._ddpClient.collections.locations
        && Object.values(this._meteorClient._ddpClient.collections.locations).find(loc => loc._id === this._device.locationId);
      const requiredConfig = {
        URL: this._device.tmx?.url || 'unkown', // backward compatibility
        DEVICE_ID: this._device._id,
        DEVICE_TYPE: this._device.type,
        DEVICE_TITLE: this._device.title,
        DEVICE_ROLE: deviceRoleLabels[this._device.role]?.label ?? 'unknown',
        DEVICE_ROLE_EXTENDED: deviceRoleLabels[this._device.role]?.extendedLabel ?? 'unknown',
        HARDWARE_TYPE: this._device.hardwareType,
        INSTITUTE_ID: this._device.instituteId,
        INSTITUTE_NAME: institute?.name || 'unkown',
        MEETING_ID: this._device.meetingId || 'none',
        LOCATION_ID: this._device.locationId || 'unkown',
        LOCATION_NAME: location?.name || 'unkown',
        LOCATION_PATH_NAME: location?.pathName || 'unkown',
        VERSION: this._device.version,
      };

      if (!(_.isEqual(requiredConfig, this._fbConfig))) {
        try {
          await this._stopFBService();
          await this._removeFBConfig();
          this._log('verbose', `Updating fluentbit config file to ${JSON.stringify(requiredConfig, null, 2)}`);
          await this._writeFBConfig(requiredConfig);
          this._fbConfig = requiredConfig;
          await this._startFBService();
        } catch (error) {
          this._log('error', `Some error occured while updating fluentbit config and restarting service, error = ${error.message}`, { serviceLog: true });
        }
      }
    }
    this._updateFBConfigOngoing = false;
  }

  async _onStart() {
    this._ensureConfigDirExists();
    this._readFBConfig();
    this._requestUpdateFBConfig();
    return super._onStart();
  }

  async _onStop() {
    return super._onStop();
  }

  _onThisDeviceChanged(_device, _oldFields, _clearedFields, newFields) {
    if (newFields.hasOwnProperty('tmx')
    || newFields.hasOwnProperty('meetingId')
    || newFields.hasOwnProperty('instituteId')
    || newFields.hasOwnProperty('locationId')) {
      this._requestUpdateFBConfig();
    }
  }

  async _requestUpdateFBConfig() {
    if (this._updateFBConfigOngoing) {
      this._log('debug', 'A fluentbit config file update cycle is already ongoing, scheduling extra round');
      this._requestNewUpdate = true;
    } else {
      this._requestNewUpdate = false;
      await this._updateFBConfigIfNeeded();
      if (this._requestNewUpdate) {
        await this._requestUpdateFBConfig();
      }
    }
  }

  async _onDestroy() {
    return super._onDestroy();
  }
}

export default FluentbitBridge;
