import * as later from 'later';
import * as childProcess from 'child_process';
import * as pify from 'pify';

import { rebootSystem, wait } from '../../../utils/utils';
import Service from '../../../services/service';
import serviceRegistry from '../../../services/serviceRegistry';

const exec = pify(childProcess.exec);

/**
 * The health service allows setting a health check based on a cron schedule
 * When started, it will perform a first health check at start and according to the said schedule.
 * The first health check at startup will never reboot to avoid endless boot loops.
 */
class HealthService extends Service {
  constructor({
    id,
    meetingActive,
    schedule,
    checkOnStartSec,
  }) {
    super({ id });
    later.date.localTime();
    this._schedule = schedule;
    this._checkOnStartSec = checkOnStartSec;
    this.meetingActive = meetingActive;
    this._checkHealth = this._checkHealth.bind(this);
    this._checkHealthButDontReboot = this._checkHealthButDontReboot.bind(this);
  }

  set schedule(schedule) {
    this._schedule = schedule;
    if (schedule) {
      this._parsedSchedule = later.parse.cron(schedule);
    }
    if (this.isStarted) {
      this._scheduleHealthChecks();
    }
  }

  get schedule() {
    return this._schedule;
  }

  async _onStart() {
    this._scheduleHealthChecks();
    await super._onStart();
    if (this._checkOnStartSec) {
      this._log('debug', `Scheduling a first health check after ${this._checkOnStartSec} seconds`);
      this._firstCheckTimer = setTimeout(this._checkHealthButDontReboot, this._checkOnStartSec * 1000);
    }
  }

  async _onStop() {
    this._clearFirstCheckTimer();
    this._clearHealthChecksSchedule();
    await super._onStop();
  }

  async _checkHealth(dontReboot = false) {
    this._log('debug', `Executing ${dontReboot ? 'startup' : 'scheduled'} health check${dontReboot ? ' without reboot' : ''}`);
    this._clearFirstCheckTimer();
    const vm = serviceRegistry.get('versionMonitor');
    if (vm?.isUpgrading) {
      this._log('debug', 'Skipping health check since we\'re upgrading');
    } else {
      // Only check for corrupt usb devices for time being (more checks can be added later)
      const problemDetected = await this._checkUsbHealth();
      if (problemDetected) {
        if (dontReboot) {
          this._log('error', 'Found health problem but not rebooting because of debug safety (avoid reboot loop)', { serviceLog: true });
          return;
        }
        if (this.meetingActive) {
          this._log('error', 'Found health problem but not rebooting because lecture ongoing', { serviceLog: true });
          return;
        }
        this._log('error', 'Found health problem, launching reboot cycle (rtcwake) to resolve situation', { serviceLog: true });
        await wait(5000); // allow logger to flush
        rebootSystem({
          shutdownDelay: 0,
          startupDelay: 90, // use rtcwake in order to reset usb devices
        });
      } else {
        this._log('debug', 'Performed health check - system is healthy', { serviceLog: true });
      }
    }
  }

  _checkHealthButDontReboot() {
    this._checkHealth(true);
  }

  _runJournalCtl() {
    try {
      // return all kernel related messages of priority error or higher from this boot
      const result = exec('sudo journalctl -b -k -p err');
      return result;
    } catch (error) {
      this._log('error', 'Error in retrieving kernel error logs since last boot, error =', error);
    }
    return '';
  }

  _transformJournalCtlOutput(src) {
    const lines = src.split('\n');
    const result = [];
    lines.forEach((line) => {
      // search for lines like: 'Sep 02 13:39:24 nrc-0060e05bdf17 kernel: usb usb1-port2: unable to enumerate USB device'
      const tmp = line.match(/(?<date>((\S*\s)*)\S*)\s\S*\skernel:\susb\s(?<device>\S*): unable to enumerate USB device/);
      if (tmp) {
        result.push(
          {
            device: tmp[4],
            date: tmp[1],
          },
        );
      }
    });
    return result;
  }

  async _checkUsbHealth() {
    const journalCtlOutput = await this._runJournalCtl();
    const result = this._transformJournalCtlOutput(journalCtlOutput);
    result.forEach((item) => {
      this._log('error', `Corrupt usb device ${item.device} detected by OS at ${item.date}`, { serviceLog: true });
    });
    return (result.length > 0); // true => problem detected
  }

  /**
   * Schedules health checks according to a given schema
   */
  _scheduleHealthChecks() {
    this._clearHealthChecksSchedule();
    if (this.schedule) {
      try {
        this._log('verbose', `Setting health check schedule to: ${this.schedule}`);
        this._healthCheckTimer = later.setInterval(this._checkHealth, this._parsedSchedule);
      } catch (error) {
        this._log('error', 'Error while setting health check cron', error);
      }
    } else {
      this._log('verbose', 'Empty health check schedule, ignoring...');
    }
  }

  _clearHealthChecksSchedule() {
    if (this._healthCheckTimer) {
      this._log('verbose', 'Clear health check schedule timer');
      this._healthCheckTimer.clear();
      delete this._healthCheckTimer;
    }
  }

  _clearFirstCheckTimer() {
    if (this._firstCheckTimer) {
      clearTimeout(this._firstCheckTimer);
      delete this._firstCheckTimer;
    }
  }
}

export default HealthService;
