import Bridge from '../bridge';
import Rebooter from './service';
import { commands } from '../../../services/ddp/constants';
import { state as serviceState } from '../../../services/service';

/**
 * The reboot bridge is the glue between blue and the device reboot service.
 */
class RebootBridge extends Bridge {
  constructor(options) {
    super(options);
    this._rebooter = new Rebooter({ id: this._device._id, schedule: this._device.rebootSchedule });
  }

  get supportedRPCCommands() {
    return [commands.REBOOT];
  }

  async _onStart() {
    await super._onStart();
    this._rebooter.schedule = this._device.rebootSchedule;
    return this._rebooter.start({ cancelToken: this.getCancelToken(serviceState.STARTED) });
  }

  async _onStop() {
    await super._onStop();
    return this._rebooter.stop();
  }

  _onThisDeviceChanged(device, oldFields, clearedFields, newFields) {
    if (newFields.hasOwnProperty('rebootSchedule')) {
      this._log('verbose', 'Setting new reboot schedule', { serviceLog: true });
      this._rebooter.schedule = newFields.rebootSchedule;
    }
  }

  /**
   * Act on incoming client RPCs
   * RPC's are useful for one-time actions that are not part of the application state.
   *
   * @param rpc
   * @returns {*}
   * @private
   */
  _onRpcAdded(rpc) {
    this._log('verbose', `Handling rpc ${JSON.stringify(rpc, null, ' ')}`, { serviceLog: true });
    switch (rpc.command) {
    case commands.REBOOT:
      // We want less downtime when we reboot using the command
      // then during the nightly/scheduled reboot.
      // Because this might happen during a lecture to fix a problem
      // And we want to be up & running asap
      // setTimeout to ensure the markRPCDone is called before actually rebooting
      setTimeout(() => this._rebooter.reboot({ shutdownDelay: 0 }), 5 * 1000);
      break;
    default:
    }
  }

  async _onDestroy() {
    await this._rebooter.destroy();
    delete this._rebooter;
    await super._onDestroy();
  }
}

export default RebootBridge;
