import * as later from 'later';

import { rebootSystem } from '../../../utils/utils';
import Service from '../../../services/service';
import serviceRegistry from '../../../services/serviceRegistry';

export const REBOOT_TIMEOUT = Number.parseInt(process.env.REBOOT_TIMEOUT, 10) || 0;
/**
 * The reboot service allows setting a reboot cron schedule
 * When started, it will restart the device according to the said schedule.
 */
class RebootService extends Service {
  constructor({ id, schedule }) {
    super({ id });
    later.date.localTime();
    this._schedule = schedule;
  }

  set schedule(schedule) {
    this._schedule = schedule;
    if (schedule) {
      this._parsedSchedule = later.parse.cron(schedule);
    }

    if (this.isStarted) {
      this._scheduleReboots();
    }
  }

  get schedule() {
    return this._schedule;
  }

  async _onStart() {
    this._scheduleReboots();
    await super._onStart();
  }

  async _onStop() {
    this._clearRebootSchedule();
    await super._onStop();
  }

  // We may want to wait a period so that connected USB devices are powered off
  // for a sufficient amount of time to reset them
  // This solves some issues with freezing USB devices on 24/7 use.
  reboot({ shutdownDelay = Math.ceil(Math.random() * 10), startupDelay = REBOOT_TIMEOUT } = {}) {
    rebootSystem({ shutdownDelay, startupDelay });
  }

  /**
   * Schedules reboots according to a given schema
   */
  _scheduleReboots() {
    this._clearRebootSchedule();

    if (this.schedule) {
      try {
        this._log('verbose', `Setting reboot schedule: ${this.schedule}`);

        this._rebootTimer = later.setInterval(() => {
          this._log('info', 'Executing scheduled restart');
          const vm = serviceRegistry.get('versionMonitor');
          if (vm && vm.isUpgrading) {
            this._log('info', 'Skipping reboot since we\'re upgrading');
          } else {
            this.reboot();
          }
        }, this._parsedSchedule);
      } catch (error) {
        this._log('error', 'Error while setting reboot cron', error);
      }
    } else {
      this._log('verbose', 'Empty reboot schedule, ignoring...');
    }
  }

  _clearRebootSchedule() {
    if (this._rebootTimer) {
      this._log('verbose', 'Clear reboot schedule timer');
      this._rebootTimer.clear();
      delete this._rebootTimer;
    }
  }
}

export default RebootService;
