import * as fs from 'fs-extra';
import { serializeError } from 'serialize-error';

import Bridge from '../bridge';
import { serviceRegistry } from '../../../services/serviceRegistry';
import { commands, recordingState } from '../../../services/ddp/constants';
import { RtspRecordingService } from './rtspRecordingService';
import { createUploader, validateUploaderOptions } from './uploaders/mapping';

const streamRegistry = serviceRegistry.get('RTSPRegistry');

/**
 * The recording bridge is orchestrating RTSP stream capturing.
 */

export default class RecordingBridge extends Bridge {
  constructor(...args) {
    super(...args);
    this.recordings = new Map();
    this.uploaders = new Map();
    this._onStreamAdded = this._onStreamAdded.bind(this);
  }

  async _onStart() {
    if (this.shouldStartRecording) {
      this._startRecording();
    }
    await super._onStart();
  }

  get supportedRPCCommands() {
    return [commands.VERIFY_RECORDING_OPTIONS];
  }

  async _onRpcAdded(rpc) {
    this._log('verbose', `Handling rpc ${JSON.stringify(rpc, null, ' ')}`, { serviceLog: true });
    let verified = true;
    switch (rpc.command) {
    case commands.VERIFY_RECORDING_OPTIONS:
      try {
        await validateUploaderOptions(this.device[rpc.params.type]);
      } catch (err) {
        this._log('verbose', JSON.stringify(err, null, 4));
        verified = false;
      }

      return { verified };
    default:
    }
  }

  get isServiceRequired() {
    return (
      super.isServiceRequired
      && this._meteorClient._ddpClient.collections.institutes
      && Object.values(this._meteorClient._ddpClient.collections.institutes)
        .some(institute => institute.enabledFeatures.includes('recording'))
    );
  }

  get uploaderIsSet() {
    return this.device.recordingOptions.uploadType !== 'none';
  }

  get shouldStartRecording() {
    return this.uploaderIsSet && !!this.device.meetingId && this.device.recording === recordingState.RECORDING;
  }

  async validateRecordingOptions() {
    let mainValid = false;
    let backupValid = false;

    if (this.uploaderIsSet) {
      try {
        mainValid = await validateUploaderOptions(this.device.recordingOptions);
      } catch (error) {
        this._log('error', `Recording uploading options are invalid. ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`, {
          serviceLog: true,
        });
      }
    } else {
      mainValid = true;
    }

    if (this.device.backupRecordingOptions.uploadType !== 'none') {
      try {
        backupValid = await validateUploaderOptions(this.device.backupRecordingOptions);
      } catch (e) {
        this._log('error', `Recording uploading options are invalid with error: ${JSON.stringify(serializeError(e, { maxDepth: 1 }))}`, {
          serviceLog: true,
        });
      }
    }

    if (mainValid || backupValid) {
      this.markHealthy();
    } else {
      this.markUnhealthy('Invalid upload options');
    }
  }

  async _onStop() {
    await super._onStop();
    await this._stopRecording();
  }

  async _onDestroy() {
    await super._onDestroy();
    const recordings = Array.from(this.recordings.values());
    recordings.forEach((recording) => {
      recording.destroy();
    });
    this.recordings.clear();
  }

  async _onThisDeviceChanged(device, oldFields, clearedFields, newFields) {
    if (newFields.hasOwnProperty('recordingOptions')) {
      this.restart().catch(error => this._log('error', `Error on devices restart: ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`, {
        serviceLog: true,
      }));
      return;
    }

    if (newFields.hasOwnProperty('recording') || newFields.hasOwnProperty('meetingId')) {
      if (this.shouldStartRecording) {
        await this.validateRecordingOptions();
        this._startRecording();
      } else {
        this._stopRecording();
      }
    }
  }

  _startRecording() {
    if (this.isUnhealthy) {
      this._log('error', 'Cannot start recording. Bridge is in unhealthy state.', {
        serviceLog: true,
      });
      return;
    }
    this._stopRecording();
    this._log('info', 'Starting Recording', { serviceLog: true });
    const streams = Array.from(streamRegistry);
    streams.forEach(([id, stream]) => this._onStreamAdded(id, stream));
    streamRegistry.on('registered', this._onStreamAdded);
  }

  _onStreamAdded(id, stream) {
    let recording = this.recordings.get(stream);
    if (!recording) {
      recording = new RtspRecordingService({ stream, id });
      this.recordings.set(stream, recording);
      recording.on('file:new', this._onFileNew.bind(this));
      recording.on('file:ready', this._onFileReady.bind(this));
      stream.once('destroying', () => this._onStreamDestroying(stream));
    }
    recording.start();
  }

  _getUploader(filePath, recording) {
    let uploaderPromise = this.uploaders.get(filePath);
    if (!uploaderPromise) {
      const getUploader = async () => {
        const metadata = await this._meteorClient.getRecordingInformation({ meetingId: this.device.meetingId });
        if (!metadata) {
          throw new Error('Metadata not found');
        }
        metadata.deviceId = this.device._id;
        metadata.sourceTitle = recording.title;
        metadata.sourceType = recording.sourceType;
        const uploader = createUploader(
          this.device.recordingOptions.uploadType,
          {
            filePath,
            metadata,
            settings: this.device.recordingOptions,
            backupSettings: this.device.backupRecordingOptions,
          },
        );
        uploader.once('destroyed', () => this.uploaders.delete(filePath));

        return uploader;
      };

      uploaderPromise = getUploader();
      this.uploaders.set(filePath, uploaderPromise);
    }
    return uploaderPromise;
  }

  async _onFileNew(filePath, recording) {
    try {
      const uploader = await this._getUploader(filePath, recording);
      await uploader.start();
    } catch (error) {
      this._log('error', `Failed to offload file with error: ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`, {
        serviceLog: true,
      });
    }
  }

  async _onFileReady(filePath) {
    let uploader;

    try {
      const uploaderPromise = this.uploaders.get(filePath);
      if (uploaderPromise) {
        uploader = await uploaderPromise;
      }
    } catch (error) {
      this._log('error', `Error while getting uploader for file ${filePath}`, {
        serviceLog: true,
      });
    }

    if (uploader) {
      try {
        await uploader.stop();
      } catch (error) {
        this._log('error', `Failed to offload file with error: ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`, {
          serviceLog: true,
        });
      }
    } else {
      this._log('error', `Uploader for file ${filePath} was not found`, {
        serviceLog: true,
      });
      this._log('warn', `Removing file ${filePath}`, {
        serviceLog: true,
      });
      try {
        await fs.unlink(filePath);
      } catch (error) {
        this._log('error', `Unable to remove recorded file. ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`, {
          serviceLog: true,
        });
      }
    }
  }

  _onStreamDestroying(stream) {
    const recording = this.recordings.get(stream);
    if (recording) {
      recording.destroy();
      this.recordings.delete(stream);
    }
  }

  _stopRecording() {
    this._log('info', 'Stopping recording', { serviceLog: true });
    streamRegistry.off('registered', this._onStreamAdded);
    const recordings = Array.from(this.recordings.values());
    recordings.forEach((recording) => {
      recording.stop();
    });
  }
}

export { RecordingBridge };
