import * as chai from 'chai';
import * as chaiAsPromised from 'chai-as-promised';
import * as sinon from 'sinon';
import * as sinonChai from 'sinon-chai';
import { EventEmitter2 } from 'eventemitter2';
import * as FsModule from 'fs-extra';
import { recordingState } from '../../../services/ddp/constants';
import {
  registerAll,
  unregisterAll,
} from '../../../utils/test-helpers/fillServiceRegistry';
import { sleep } from '../../../utils/utils';
import { serviceRegistry } from '../../../services/serviceRegistry';
import * as RtspRecordingServiceModule from './rtspRecordingService';
import * as UploaderMappingModule from './uploaders/mapping';
import { RecordingBridge } from './bridge';

const { expect } = chai;
chai.use(chaiAsPromised);
chai.use(sinonChai);

describe('RecordingBridge', () => {
  let recordingBridge;
  let stream;
  let meteorClient;
  let streamRegistry;
  let uploader;
  let unlinkStub;
  let createUploaderStub;
  const device = {
    id: 'DEVICE_ID',
    meetingId: 'MEETING_ID',
    recordingOptions: {
      uploadType: 'scp',
    },
    backupRecordingOptions: {
      uploadType: 'none',
    },
  };

  class RtspRecordingService extends EventEmitter2 {
    start() {
      this.emit('file:new', 'test-file.mp4', this);
    }

    stop() {
      this.emit('file:ready', 'test-file.mp4', this);
    }
  }

  before(() => {
    registerAll();
    streamRegistry = serviceRegistry.get('RTSPRegistry');
  });

  after(() => {
    unregisterAll();
  });

  beforeEach(() => {
    streamRegistry.clear();
    sinon.stub(RtspRecordingServiceModule, 'RtspRecordingService').callsFake((args) => new RtspRecordingService(args));
    createUploaderStub = sinon.stub(UploaderMappingModule, 'createUploader');
    createUploaderStub.callsFake(() => uploader);
    unlinkStub = sinon.stub(FsModule, 'unlink');
    meteorClient = new EventEmitter2();
    meteorClient.getRecordingInformation = sinon.stub();
    meteorClient.getRecordingInformation.resolves({});
    uploader = new EventEmitter2();
    uploader.start = sinon.stub();
    uploader.stop = sinon.stub().callsFake(() => {
      uploader.emit('destroyed');
    });
    stream = new EventEmitter2();
    recordingBridge = new RecordingBridge({ device, meteorClient });
  });

  afterEach(async () => {
    await recordingBridge.stop();
    sinon.restore();
  });

  it('should create the bridge', () => {
    expect(recordingBridge).to.exist;
  });

  it('should start the recording', async () => {
    device.recording = recordingState.RECORDING;
    streamRegistry.set('STREAM_ID', stream);
    await recordingBridge.start();

    expect(recordingBridge.isStarted).to.be.true;

    while (uploader.start.callCount === 0) {
      // eslint-disable-next-line no-await-in-loop
      await sleep(10);
    }

    expect(uploader.start).to.be.calledOnce;
  });

  it('should stop the uploader when the recording is finished', async () => {
    device.recording = recordingState.RECORDING;
    streamRegistry.set('STREAM_ID', stream);
    await recordingBridge.start();

    while (uploader.start.callCount === 0) {
      // eslint-disable-next-line no-await-in-loop
      await sleep(10);
    }

    await recordingBridge.stop();

    while (uploader.stop.callCount === 0) {
      // eslint-disable-next-line no-await-in-loop
      await sleep(10);
    }

    expect(uploader.stop).to.be.calledOnce;
  });

  it('should delete the file when there is no uploader', async () => {
    device.recording = recordingState.RECORDING;
    streamRegistry.set('STREAM_ID', stream);

    createUploaderStub.callsFake(() => null);
    await recordingBridge.start();

    await recordingBridge.stop();

    while (unlinkStub.callCount === 0) {
      // eslint-disable-next-line no-await-in-loop
      await sleep(10);
    }

    expect(unlinkStub).to.be.calledOnce;
  });

  it('should stop the uploader after starting if the recording immediately stops', async () => {
    class ImmediatelyStopsRtspRecordingService extends EventEmitter2 {
      start() {
        this.emit('file:new', 'test-file.mp4', this);
        this.emit('file:ready', 'test-file.mp4', this);
      }

      stop() {
      }
    }
    RtspRecordingServiceModule.RtspRecordingService.restore();
    sinon.stub(RtspRecordingServiceModule, 'RtspRecordingService').callsFake((args) => new ImmediatelyStopsRtspRecordingService(args));

    device.recording = recordingState.RECORDING;
    streamRegistry.set('STREAM_ID', stream);

    recordingBridge.start();

    while (uploader.stop.callCount === 0) {
      // eslint-disable-next-line no-await-in-loop
      await sleep(10);
    }

    expect(uploader.start).to.be.calledOnce;
    expect(uploader.stop).to.be.calledOnce;
    expect(uploader.start).to.be.calledBefore(uploader.stop);
  });
});
