import * as fs from 'fs-extra';

import { Service } from '../../../services/service';

/**
 * This service is meant to monitor recording files
 * It will announce whenever file has been finished writing
 * and is ready to be offloaded.
 *
 * It takes filePrefix as a parameter and it will check files which
 * starts with a given prefix, i.e.
 *
 * filePrefix === 'video-abc123'
 * it will check for files video-abc123-001, video-abc123-002 and so on.
 *
 * This service should be destroyed only when the recording process is stopped.
 * While being destroyed it will emit the last file:ready event.
 */
class FileMonitor extends Service {
  constructor(cwd, filePrefix) {
    super({ id: filePrefix });

    this._cwd = cwd;
    this._fileRgx = new RegExp(filePrefix);
    this.currentFile = null;
    this._finishedFiles = [];
  }

  _onStart() {
    this.check();
    this._startCheckInterval();
    return super._onStart();
  }

  /**
   * Start the interval check
   * @private
   */
  _startCheckInterval() {
    this._checkIntervalHandle = setInterval(
      this.check.bind(this),
      5000,
    );
  }

  _clearCheckIntervalHandle() {
    if (this._checkIntervalHandle) {
      clearInterval(this._checkIntervalHandle);
      this._checkIntervalHandle = null;
    }
  }

  async check() {
    try {
      let files = await fs.readdir(this._cwd);
      files = files.filter(file => this._fileRgx.test(file)).sort();

      const latestFile = files.pop();
      // if there is a new file detected, it means that previous ones are finished writing
      // and ready to be offloaded.
      if (latestFile && this.currentFile !== latestFile) {
        this.currentFile = latestFile;
        this._emitNew(latestFile);
        const finishedFiles = files.filter(f => !this._finishedFiles.includes(f));
        finishedFiles.forEach(file => this._emitReady(file));
        this._finishedFiles.push(...finishedFiles);
      }
    } catch (err) {
      this._log('error', err.message);
    }
  }

  _emitReady(file) {
    this.emit('file:ready', file);
  }

  _emitNew(file) {
    this.emit('file:new', file);
  }

  async _onStop() {
    this._clearCheckIntervalHandle();
    // perform the last check.
    await this.check();
    return super._onStop();
  }

  async _onDestroy() {
    // we assume that this service is destroyed only when recording is finished
    // and recording process is no longer writing to a file.
    // onStop method will check if there's a new file generated in a mean time
    // and finally before we get destroyed, we emit ready event about the last known file.
    this._emitReady(this.currentFile);
    await super._onDestroy();
    this.currentFile = null;
    this._finishedFiles = [];
  }
}

export default FileMonitor;
export { FileMonitor };
