import getGstDebugEnv from '../utils/getGstDebugEnv';
import { SpawningService } from '../../../services/spawningService';
import { FileMonitor } from './fileMonitor';

function pad(n) {
  return `00${n}`.substr(-3);
}

const MINUTE = 60000000000;
const GIGABYTE = 1024 * 1024 * 1024;
/**
 * This service will start GST pipeline to record a RTSP stream.
 *
 * The pipeline uses splitmuxsink and splits recording into 30 minutes long files.
 * Filenames are incremented every time we start a pipeline to avoid overwriting.
 *
 * This service watches recording files and emits file:ready events, when they are
 * ready to be offloaded.
 */
class RtspRecordingService extends SpawningService {
  constructor({ stream, id }) {
    super({ id, killSignal: 'SIGINT' });
    this._stream = stream;
    this.title = stream.recordSink.title;
    this.sourceType = stream.recordSink.sourceType;
    this._counter = 0;
    this._cwd = '/tmp';
    this._fileWatchers = [];
    this._incrementCounter = false;
  }

  get command() {
    return 'gst-launch-1.0';
  }

  get args() {
    const args = [
      '-e',
      'rtspsrc name=elem1',
    ];
    const uri = `rtsp://127.0.0.1:${this._stream.port}/${this._stream.url}`;
    args.push(`location="${uri}"`);
    if (this._stream.username) {
      args.push(`user-id=${this._stream.username}`);
    }
    if (this._stream.password) {
      args.push(`user-pw=${this._stream.password}`);
    }
    // (only) if file is existing, increment counter in filename to avoid overwriting after pipeline restart.
    if (this._incrementCounter) {
      this._counter += 1;
    }
    const fileName = `video-${this._serviceId}-${pad(this._counter)}`;
    this._watchFiles(fileName);
    args.push('elem1. ! rtph264depay ! h264parse ! queue ! muxer.video');
    if (this._stream.paSrcDevice) {
      args.push('elem1. ! rtpL16depay ! audioconvert ! voaacenc bitrate=128000 ! aacparse ! queue ! muxer.audio_0');
    }
    args.push(`splitmuxsink name=muxer location=${this._cwd}/${fileName}-%03d.mp4  max-size-time=${120 * MINUTE} max-size-bytes=${4 * GIGABYTE}`);
    return args.join(' ');
  }

  get env() {
    return getGstDebugEnv('RTSP_RECORDING');
  }

  /**
   * Create a file monitor to observe when the file is ready to be offloaded.
   * @param fileName - filename prefix
   * @private
   */
  _watchFiles(fileName) {
    const fileMonitor = new FileMonitor(this._cwd, fileName);
    this._fileWatchers.push(fileMonitor);
    fileMonitor.on('file:ready', file => this._emitFileReady(file));
    fileMonitor.on('file:new', file => this._emitFileNew(file));
    fileMonitor.once('destroyed', () => {
      this._fileWatchers = this._fileWatchers.filter(watcher => fileMonitor !== watcher);
    });
    fileMonitor.start();
    // Recording service becomes unhealthy when pipeline process exits unexpectedly.
    // That means that all files are closed and ready to offload.
    // While destroying a file monitor, it will emit ready event for the last known file
    // and that event will be emitted to the bridge.
    this.once('unhealthy', () => fileMonitor.destroy());
  }

  async _onStop() {
    await super._onStop();
    await Promise.all(this._fileWatchers.map(fileWatcher => fileWatcher.destroy()));
  }

  _emitFileReady(file) {
    if (file) {
      this._log('debug', `Recording file ${this._cwd}/${file} is ready`);
      this.emit('file:ready', `${this._cwd}/${file}`, this);
    }
  }

  _emitFileNew(file) {
    if (file) {
      this._log('debug', `A new recording file ${this._cwd}/${file} has been created`);
      this._incrementCounter = true;
      this.emit('file:new', `${this._cwd}/${file}`, this);
    }
  }
}

export default RtspRecordingService;
export { RtspRecordingService };
