/**
 * (C) 2015 Copyright BARCO NV
 * All rights reserved. Reproduction in whole or in part is
 * prohibited without the written consent of the copyright owner.
 *
 * Part of Barco WeConnect
 */

import { Socket } from 'net';
import * as os from 'os';
import * as _ from 'lodash';
import * as semver from 'semver';

import { Service } from '../../../services/service';
import UbicastError from './ubicastError';
/**
 * Enum to act as FSM for cngClient objects
 * @type {{DISCONNECTED: number, CONNECTED: number, DESTROYING: number, DESTROYED: number}}
 */
const state = {
  DISCONNECTED: 0,
  CONNECTED: 1 << 1,
};

/**
 * Some helper methods
 */

/**
 * Parses a string or num to boolean
 * @param val
 * @returns {boolean}
 */
function parseBool(val) {
  if (typeof val === 'undefined') {
    return false;
  }

  const num = +val;
  return !Number.isNaN(num) ? !!num : !!String(val).toLowerCase().replace(!!0, '');
}

const filterResponses = {
  record: /^start_recording/i,
  stop: /^stop_recording/i,
};

/**
 * Ubicast TCP API adapter
 */
class UbicastService extends Service {
  /**
     * Create a new instance of UbicastService
     * @param ipAddress ipAddress of Recorder to connect to (default: "127.0.0.1")
     * @param hostname hostname of Recorder to connect to
     * @param port port on Recorder to connect to (default: 5010)
     * @param autoReconnect whether to auto reconnect if connection fails or is lost (default:false)
     * @param autoReconnectTimer time between reconnection attempts (default: 10s).
     * @param optimizeQueue remove deprecated messages from the command queue if enabled. (default: false).
     */
  constructor({
    ipAddress = '127.0.0.1',
    hostname,
    secret = '',
    port = 5010,
    autoReconnect = false,
    autoReconnectTimer = 10000,
    optimizeQueue = false,
  }) {
    super({ id: `${ipAddress}:${port}` });

    if (!hostname) {
      if (ipAddress === '127.0.0.1') {
        hostname = os.hostname();
      } else {
        hostname = ipAddress;
      }
    }

    const params = {
      ipAddress,
      hostname,
      port,
      secret,
      autoReconnect,
      autoReconnectTimer,
      optimizeQueue,
    };
    Object.keys(params).forEach((param) => {
      if (params[param] === undefined) throw new Error(`Please provide a valid ${param}`);
      this[`_${param}`] = params[param];
    });

    this._isActive = true;

    // Queue of commands waiting to be send to Cinergy API
    this._cmdQueue = [];

    // Last executed command
    this._lastCommand = undefined;

    this.retries = 0;
    this.MAX_RETRIES = 20;

    this.on('disconnected', () => {
      this.state = state.DISCONNECTED;
    });
  }

  get minimumSupportedVersion() {
    return UbicastService.minimumSupportedVersion;
  }

  set isConnected(connected) {
    this.state = connected ? state.CONNECTED : state.DISCONNECTED;
  }

  get isConnected() {
    return (this._state === state.CONNECTED);
  }

  set state(newState) {
    if (newState !== this._state) {
      this._state = newState;
      switch (this._state) {
      case state.CONNECTED:
        this.emit('connected');
        break;
      case state.DISCONNECTED:
        this.emit('disconnected');
        break;
      default:
      }
    }
  }

  get secret() {
    return this._secret;
  }

  set secret(secret) {
    this._secret = secret;
  }

  get state() {
    return this._state;
  }
  // END of FSM getters/setters

  get hostname() {
    return this._hostname;
  }

  /**
     * Destroys all connection-related instance stuff when the connection is closed/lost.
     * rejects active listeners & method calls etc.
     *
     * Will schedule a reconnection if required.
     * @private
     */
  _cleanupOnSocketClosed() {
    if (this.isConnected) {
      const tryReconnect = this._autoReconnect && this._isActive;
      if (tryReconnect) {
        this._log('info', `

    =================================================================================
    ==                                                                             ==
    == Connection lost. Cleaning up and triggering reconnect timer                 ==
    ==                                                                             ==
    =================================================================================

            `);
      } else {
        this._log('info', `

    =================================================================================
    ==                                                                             ==
    == Connection lost. Cleaning up.
    ==                                                                             ==
    =================================================================================

            `);
      }

      // Set status variables
      this.isConnected = false;

      // Get rid of the socket
      if (this._clientSock) {
        // this._clientSock.end()
        this._clientSock.removeAllListeners();
        this._clientSock.destroy();
        delete this._clientSock;
      }
      this._clientSock = undefined;

      // Reject all pending promises
      this._rejectAll();

      // Clean the receive buffer
      delete this._buffer;

      if (this._reconnectTimer) {
        clearTimeout(this._reconnectTimer);
      }

      if (tryReconnect) {
        this._reconnectTimer = setTimeout(this.connect.bind(this), this._autoReconnectTimer);
      }
    }
  }

  /**
     * Builds NRC API XML command and sends it to the NRC iConnect interface
     * @param cmdObj - JSOn representation of the command to send
     * @private
     */
  _executeCommand(cmdObj) {
    const commandStr = this._buildCommandString(cmdObj);

    cmdObj.command = commandStr;

    this._log('verbose', `EXECUTING COMMAND -> ${commandStr}`);

    // Send the command over the socket
    if (this._clientSock && this.isConnected) {
      this._clientSock.write(commandStr);
    } else {
      this._log('error', 'Cannot write command to disconnected socket!');
    }
  }

  _buildCommandString(cmdObj) {
    if (!Array.isArray(cmdObj.params)) {
      cmdObj.params = [cmdObj.params];
    }

    const parts = ['*', cmdObj.name, ...cmdObj.params];
    if (this.secret) {
      parts.push({
        secret: this.secret,
      });
    }

    const formatParamValue = (value) => {
      value = value.toString();
      if (/\s/g.test(value)) {
        value = `"${value}"`;
      }
      return value;
    };

    return parts
      .map((param) => {
        if (_.isObject(param)) {
          return Object.keys(param).map(key => `${key.toUpperCase()}=${formatParamValue(param[key])}`).join(' ');
        } else if (_.isString(param)) {
          return param.toUpperCase();
        }
        return false;
      })
      .filter(part => !!part)
      .join(' ');
  }

  get _isQEmpty() {
    return this._cmdQueue.length === 0;
  }

  /**
     * NRC can only handle 1 command at once,
     * therefore we store commands in a queue. This method will take the first command
     * from the queue and execute it. After execution, this method will be invoked again.
     * @private
     */
  _processQueue() {
    if (this._cmdQueue.length > 0) {
      this._lastCommand = this._cmdQueue.shift();
      this._log('debug', `PROCESS QUEUE FOR COMMAND >>>>> ${this._lastCommand.name}`);
      this._executeCommand(this._lastCommand);
    } else {
      this._log('debug', 'Command queue empty !!');
    }
  }

  /**
     * Adds a command to the processing queue. It will kick-off the processing if the queue was empty
     * @param command - JSON version of the command to add
     * @returns {Promise}
     * @private
     */
  _addCmdToQueue(command, params = []) {
    return new Promise((resolve, reject) => {
      if (!this.isConnected) {
        reject(new Error('Not connected'));
        return;
      }

      const queueObj = {
        name: command.toLowerCase(),
        params,
        resolve,
        reject,
      };

      if (this._optimizeQueue) {
        if (!this._optimizeQueueFor(queueObj)) {
          reject(new Error('Rejected because of duplicate'));
          return;
        }
      }

      const startProcessing = this._isQEmpty;

      this._cmdQueue.push(queueObj);
      this._log('debug', `Added cmd ${queueObj.name} to queue and the queue length now is: ${this._cmdQueue.length}`);

      if (startProcessing) {
        // Q was empty, need to start processing again.
        this._processQueue();
      }
    });
  }

  /**
     * Sometimes we can optimize the queue
     * @returns {Promise}
     */
  _optimizeQueueFor(queueObj) {
    const rejectFor = {
      record: ['stop'],
      stop: ['start', 'pause', 'resume'],
      pause: ['resume'],
      resume: ['pause'],
      reboot: ['*'],
      shutdown: ['*'],
    };

    if (rejectFor.hasOwnProperty(queueObj.name)) {
      const toReject = rejectFor[queueObj.name];
      if (toReject.includes('*')) {
        this._rejectAll();
      } else {
        toReject.push(queueObj.name);
        this._rejectAll(cmd => toReject.includes(cmd.name));
      }
    }
    return true;
  }

  /**
     * Connect to the Ubicast API socket
     * @returns {Promise}
     */
  async connect() {
    // Already connected
    if (this.isConnected === true) {
      this._log('warn', `Socket already connected to recorder at ${this._ipAddress}:${this._port}. Returning.`);
      return true;
    }

    if (this._reconnectTimer) {
      clearTimeout(this._reconnectTimer);
    }

    this._clientSock = new Socket();

    this._clientSock.on('error', (error) => {
      if (this.isConnected) {
        // Also log as verbose because Winston does not respect logging order
        this._log('verbose', 'This socket is ended. Cleaning up', error);
        this._log('error', 'This socket is ended. Cleaning up', error);
        this._cleanupOnSocketClosed();
      }
    });

    this._clientSock.on('close', () => {
      this._log('info', 'The socket has been closed. Cleaning up.');
      this._cleanupOnSocketClosed();
    });

    this._clientSock.on('data', (data) => {
      if (this._isActive) {
        this._processReceivedData(data);
      }
    });

    try {
      await this._clientSock.connect(this._port, this._ipAddress);
      this._log('info', 'Connected to recorder');
      this.isConnected = true;
      return true;
    } catch (error) {
      this._log('warn', 'Could not connect to recorder', error);
      this._reconnectTimer = setTimeout(this.connect.bind(this), this._autoReconnectTimer);
      return false;
    }
  }

  disconnect() {
    this._rejectAll();
    if (this.isConnected) {
      this._clientSock.end();
    }
  }

  _parseResponse(response, signature) {
    return response
      .split(' ')
      .map(part => part.split('='))
      .filter(part => part.length === 2)
      .map(part => [part[0].toLowerCase(), part[1]])
      .filter(part => signature.hasOwnProperty(part[0]))
      .map(part => ({
        key: part[0],
        value: this._parse(part[1], signature[part[0]]),
      }))
      .reduce((accumulator, part) => {
        accumulator[part.key] = part.value;
        return accumulator;
      }, {});
  }

  _parse(input, type) {
    const parseArray = (arrString, elType) => {
      let result = arrString.split(',');
      if (elType) {
        result = result.map(item => this._parse(item, elType));
      }
      return result;
    };

    const transformers = {
      bool: parseBool,
      float: parseFloat,
      int: parseInt,
      array: parseArray,
      string: i => i,
      array_int: i => parseArray(i, 'int'),
      array_bool: i => parseArray(i, 'bool'),
      array_float: i => parseArray(i, 'float'),
    };

    if (transformers.hasOwnProperty(type)) {
      return transformers[type](input);
    }
    return input;
  }

  _handleResponse(response, signature = false) {
    if (response.includes('ERROR TRUE')) {
      throw new Error('Unspecified error received from Ubicast');
    } else if (signature === false) {
      return response;
    } else if (_.isString(signature)) {
      if (response === signature) {
        return true;
      }
      throw new Error('Invalid response');
    } else {
      response = this._parseResponse(response, signature);
      return response;
    }
  }

  // START RECORDER API commands
  async record(params) {
    params.profile = params.profile || 'Nrc';
    // workaround for bug in ubicast which doesn't allow some characters
    // don't know exactly which, so just strip every non alphabetic character
    params.profile = params.profile.replace(/[^a-zA-Z]/g, '');
    const response = await this._addCmdToQueue('record', params);
    return this._handleResponse(response, 'STATUS RECBUSY');
  }

  startRecording(params) {
    return this.record(params);
  }

  async stopRecording() {
    const response = await this._addCmdToQueue('stop');
    return this._handleResponse(response, 'STATUS RECBUSY');
  }

  async pause() {
    const response = await this._addCmdToQueue('pause');
    return this._handleResponse(response, 'PAUSE_RECORDING DONE');
  }

  async resume() {
    const response = await this._addCmdToQueue('resume');
    return this._handleResponse(response, 'RESUME_RECORDING DONE');
  }

  async getValues(values) {
    const response = await this._addCmdToQueue('get_value', values);
    return this._handleResponse(response, {
      status: 'string',
      profile: 'string',
      audio_status: 'int',
      audio_peak: 'int',
      video_status: 'int',
      video_signal: 'string',
      vga_status: 'int',
      vga_signal: 'string',
      timecode: 'string',
      free_space: 'float',
      custom_meta: 'string',
      profiles: 'array_string',
      version: 'string',
    });
  }

  async getValue(value) {
    const values = await this.getValues([value]);
    return values[value];
  }

  getStatus() {
    return this.getValue('status');
  }

  isPaused() {
    return this.is(UbicastService.status.PAUSED);
  }

  isRecording() {
    return this.is(UbicastService.status.RECORDING);
  }

  async is(status) {
    const actualStatus = await this.getStatus();
    return (status === actualStatus);
  }

  async isSupportedVersion() {
    const version = await this.getValue('version');
    return semver.gte(version, UbicastService.minimumSupportedVersion);
  }

  async reboot() {
    const response = await this._addCmdToQueue('reboot');
    return this._handleResponse(response, 'REBOOT DONE');
  }

  async shutdown() {
    const response = await this._addCmdToQueue('shutdown');
    return this._handleResponse(response, 'SHUTDOWN DONE');
  }

  async getErrorCode() {
    const response = await this._addCmdToQueue('shutdown');
    return this._handleResponse(response).replace('EMESSAGE ', '');
  }

  async addChapter() {
    const response = await this._addCmdToQueue('add_chapter');
    return this._handleResponse(response);
  }

  async _ptzmove(params) {
    const response = await this._addCmdToQueue('ptzmove', params);
    return this._handleResponse(response, 'PTZMOVE DONE');
  }

  moveCamera(direction) {
    return this._ptzmove({
      direction,
    });
  }

  loadCameraPosition(id) {
    return this._ptzmove({
      presetcall: id,
    });
  }

  saveCameraPosition(id) {
    return this._ptzmove({
      presetsave: id,
    });
  }

  moveCameraLeft() {
    return this.moveCamera(UbicastService.cameraDirection.LEFT);
  }

  moveCameraRight() {
    return this.moveCamera(UbicastService.cameraDirection.RIGHT);
  }

  moveCameraUp() {
    return this.moveCamera(UbicastService.cameraDirection.UP);
  }

  moveCameraDown() {
    return this.moveCamera(UbicastService.cameraDirection.DOWN);
  }

  zoomCameraIn() {
    return this.moveCamera(UbicastService.cameraDirection.IN);
  }

  zoomCameraOut() {
    return this.moveCamera(UbicastService.cameraDirection.OUT);
  }

  stopCameraMovement() {
    return this.moveCamera(UbicastService.cameraDirection.STOP);
  }
  // END OF SRECORDER API COMMANDS

  /**
     * Process raw incoming socket data and sends it to the notification or command processor accordingly.
     * Since it turns out that the XML parsing eats up a lot of memory, we bypass this for thumbnail updates.
     *
     * @private
     */
  _processReceivedData(data) {
    this._buffer += data.toString();

    const responses = this._buffer.split('\r');
    this._buffer = responses.pop();

    responses.forEach((response) => {
      // clean response
      response = response.trim().replace(/^(undefined)?(\* )?/, '');

      if (response === 'BUSY') {
        this._log('debug', `Received Busy (nb:${this.retries}) for command: ${this._lastCommand.command}`);
        if (this.retries >= this.MAX_RETRIES) {
          this._rejectAll();
          this._log('error', 'Maximum amount of retries exceeded');
        } else {
          setTimeout(() => {
            this.retries += 1;
            this._executeCommand(this._lastCommand);
          }, this.retries * 10);
        }
      } else {
        this._log('verbose', `Received Result for a command : ${this._lastCommand.command}: ${response}`);

        // Filter out intermediary updates
        if (filterResponses.hasOwnProperty(this._lastCommand.name)
                    && filterResponses[this._lastCommand.name].test(response)) {
          return;
        }

        this._lastCommand.resolve(response);
        this._processQueue();
      }
    });
  }

  /**
     * Reject all outstanding command promises
     * @private
     */
  _rejectAll(filter = () => true, includeCurrent = true) {
    this._log('verbose', 'Rejecting all outstanding commands');
    // Break all ongoing promises for commands
    if (includeCurrent && this._lastCommand && filter(this._lastCommand)) {
      this._lastCommand.reject(new UbicastError({
        msg: 'Rejected by command',
        code: '0',
      }));
    }

    this._cmdQueue
      .filter(filter)
      .forEach((cmd) => {
        cmd.reject(new UbicastError({
          msg: 'Rejected by command',
          code: '0',
        }));
      });

    this._cmdQueue = [];
  }

  async _onStart() {
    await this.connect();
    await super._onStart();
  }

  _onStop() {
    this._isActive = false;
    this.disconnect();
    return super._onStop();
  }
}
UbicastService.minimumSupportedVersion = '2.8.5';

UbicastService.cameraDirection = {
  LEFT: 'LEFT',
  RIGHT: 'RIGHT',
  UP: 'UP',
  DOWN: 'DOWN',
  IN: 'IN',
  OUT: 'OUT',
  STOP: 'STOP',
};

UbicastService.status = {
  HOME: 'HOME',
  RECORDER_IDLE: 'RECORDER_IDLE',
  RECORDING: 'RECORDING',
  STREAMING: 'STREAMING',
  PAUSED: 'PAUSED',
  RECORDER_ERROR: 'RECORDER_ERROR',
  DETECTION: 'DETECTION',
  LOADING: 'LOADING',
  REVIEW_IDLE: 'REVIEW_IDLE',
};

export default UbicastService;
