import * as path from 'path';
import { Client } from 'ssh2';
import * as ssh2Streams from 'ssh2-streams';
import * as fs from 'fs-extra';

import { sleep } from '../../../../utils/utils';
import { BaseUploader } from './base-uploader';
import serviceRegistry from '../../../../services/serviceRegistry';

const MAX_RETRIES = 5;
const RETRY_INTERVAL = 5 * 60 * 1000;
const DESTROY_TOKEN = Symbol('destroy-token');
const SUPPORTED_SERVER_HOST_KEYS = ssh2Streams.constants.ALGORITHMS.SUPPORTED_SERVER_HOST_KEY;
let staticLogger;

function sanitizeFileServerSettings({
  url,
  uploadLocation,
  user,
  password,
}) {
  if ([url, user, password].some(arg => !arg)) {
    throw new Error('Incorrect server configuration');
  }
  const rgx = /^(\w){2,5}:\/\//;
  const hasProtocol = rgx.test(url);
  if (!hasProtocol) {
    url = `scp://${url}`;
  }

  url = new URL(url);

  return {
    host: url.host,
    username: user,
    password,
    // Backwards compatibility. Also ensures uploadLocation is '/' if path is not set
    uploadLocation: (uploadLocation || url.pathname).replace(/\/$/, '') || '/',
  };
}

export default class NASUploader extends BaseUploader {
  constructor(args) {
    super(args);
    const { meetingId, deviceId } = this.metadata;
    const source = path.parse(this._filePath);
    this._fileName = source.base.replace(/video-(.+)(-\d{3}-\d{3}.+)/, `${meetingId}-${deviceId}-$1$2`);
    this._retryCounter = 0;
  }

  get fileName() {
    return this._fileName;
  }

  static getStaticLogger() {
    if (!staticLogger) {
      staticLogger = serviceRegistry.get('logger').getInstance('NASUploader');
    }
    return staticLogger.log.bind(staticLogger);
  }

  /**
   * Initiates a sftp connection using ssh2
   * It will also ensure the propper ssh2 shutdown hooks
   *  are called when sftp.end() is invoked.
   * @param {} settings
   */
  static async createSftpConnection(settings) {
    const {
      host,
      username,
      password,
      debug = true,
    } = settings;

    const log = this._log ? this._log.bind(this) : NASUploader.getStaticLogger();
    log('debug', JSON.stringify({ ...settings, password: '********' }, null, 4));
    const ssh = new Client();
    try {
      const connection = await new Promise((resolve, reject) => {
        ssh.once('error', (err) => {
          reject(err);
        }).once('ready', () => {
          ssh.sftp((err, conn) => {
            // Ensure we end both ssh and sftp
            const endSftp = conn.end.bind(conn);
            conn.end = function endConnection() {
              endSftp();
              ssh.end();
            };

            (err ? reject(err) : resolve(conn));
          });
        }).once('close', () => {
          ssh.destroy();
          ssh.removeAllListeners();
        }).connect({
          host,
          username,
          password,
          /*
          * To be able to connect to a sftp server it needs to support at least one of the algorithms
          * listed here.
          * We want to support both default and optional server host key algorithms which is why we need
          * to explicitely add it here.
          * See https://github.com/mscdex/ssh2-streams/blob/master/lib/constants.js#L230-L245 where this
          * comes from.
          *
          * See https://superuser.com/questions/1010050/how-does-openssh-decide-which-host-key-to-use
          * for more info on SSH handshaking algorithms.
          */
          algorithms: {
            serverHostKey: SUPPORTED_SERVER_HOST_KEYS,
          },
          ...debug && { debug: data => log('verbose', data) },
        });
      });
      return connection;
    } catch (err) {
      log('error', 'Failed to create SFTP connection');
      ssh.end();

      throw err;
    }
  }

  /**
   * Verifies if the uploader can connect to the server
   * Needs to be static as it is being called by ../mapping.js
   */
  static async validate(settings) {
    const log = this._log ? this._log.bind(this) : NASUploader.getStaticLogger();

    const {
      host,
      username,
      password,
      uploadLocation,
    } = sanitizeFileServerSettings(settings);

    let sftp;
    try {
      sftp = await NASUploader.createSftpConnection({ username, password, host });

      const result = await new Promise((resolve, reject) => {
        sftp.stat(
          // default to '/' to make stat work if uploadLocation is empty (root path)
          uploadLocation,
          (err, stat) => {
            log('debug', `sftp.stat(${uploadLocation}) OUTPUT =`);
            log('debug', JSON.stringify(stat, null, 4));
            log('debug', JSON.stringify(err, null, 4));
            (err ? reject(err) : resolve(stat));
          },
        );
      });
      return !!result;
    } finally {
      if (sftp) {
        sftp.end();
      }
    }
  }

  _checkDestroyToken(token) {
    return token === DESTROY_TOKEN;
  }

  async _onStop() {
    this._log('info', `Starting upload of ${this._filePath}`);
    await this._safeCopy();
    await this._uploadMetadata();

    await super._onStop();
    // no await here! It should first stop before it could be destroyed
    this.destroy(DESTROY_TOKEN);
  }

  async _uploadMetadata() {
    const {
      title,
      meetingId,
      locations,
      description,
      hosts,
      hostIdentifiers,
      date,
      configuration,
      deviceId,
      sourceTitle,
      sourceType,
    } = this.metadata;

    const metadata = {
      title,
      meetingId,
      locations,
      description,
      hosts,
      hostIdentifiers,
      date,
      configuration,
      deviceId,
      sourceTitle,
      sourceType,
    };

    try {
      const { name } = path.parse(this._fileName);
      const data = JSON.stringify(metadata, null, 2);
      const filename = `${name}.json`;
      const filePath = `/tmp/${filename}`;
      await fs.writeFile(filePath, data);
      await this._safeCopy(filePath, filename);
    } catch (err) {
      this._log('error', `Failed to upload metadata: ${err}`);
    }
  }

  async validate(settings) {
    return NASUploader.validate.bind(this)(settings);
  }

  /**
   * Uploads a file via SCP using the ssh2 fastPut method
   *
   * See https://github.com/mscdex/ssh2-streams/blob/master/SFTPStream.md

   * @param {*} localPath  The file path on the local file system
   */
  async _safeCopy(localPath = this._filePath, fileName = this._fileName) {
    const {
      username, password, host, uploadLocation,
    } = sanitizeFileServerSettings(this._settings);

    const remotePath = path.join(uploadLocation, fileName);

    let sftp;
    try {
      sftp = await NASUploader.createSftpConnection.bind(this)({ username, password, host });

      await new Promise((resolve, reject) => {
        sftp.fastPut(localPath, remotePath, { mode: 0o664 }, (err) => {
          this._log('debug', `sftp.fastPut(localPath= ${localPath}, remotePath = ${remotePath}) ${err ? 'Failed' : ''}`);

          (err ? reject(err) : resolve());
        });
      });
    } catch (err) {
      this._log('error', `Failed to upload. ${err}`);
      if (this._retryCounter < MAX_RETRIES) {
        this._retryCounter += 1;
        const timeout = RETRY_INTERVAL * this._retryCounter;
        this._log('info', `Will retry in ${timeout} ms`);
        await sleep(timeout);
        return this._safeCopy();
      } else {
        this._log('error', 'Exceeded max retries. Will terminate.');
        this.markUnhealthy(`Failed to upload: ${err}`);
      }
    } finally {
      if (sftp) {
        sftp.end();
      }
    }
  }
}

export { NASUploader, sanitizeFileServerSettings };
