import * as chai from 'chai';
import * as sinon from 'sinon';
import * as sinonChai from 'sinon-chai';

import { NASUploader, sanitizeFileServerSettings } from './NASUploader';
import serviceRegistry from '../../../../services/serviceRegistry';

const { expect } = chai;
chai.use(sinonChai);

describe('device-agent/services/recording/uploaders/NASUploader/sanitizeFileServerSettings', () => {
  const loggerServiceMock = { getInstance: sinon.stub().returns({ log() {} }), on() {} };

  it('should check inputs', () => {
    expect(() => sanitizeFileServerSettings()).to.throw();
    expect(() => sanitizeFileServerSettings({})).to.throw();
  });

  it('should prefix the host correctly', () => {
    const payload = { user: 'user', password: 'password', url: 'so.me/hodor/', uploadLocation: '/hador' };

    const output = sanitizeFileServerSettings(payload);

    expect(output).to.have.a.property('username').that.eql(payload.user);
    expect(output).to.have.a.property('password').that.eql(payload.password);
    expect(output).to.have.a.property('host').that.eql('so.me');
    // prefixed with /and removes ending /
    expect(output).to.have.a.property('uploadLocation').that.eql('/hador');
  });

  describe('device-agent/services/recording/uploaders/NASUploader/getStaticLogger', () => {
    before(() => {
      sinon.stub(serviceRegistry, 'get').withArgs('logger').returns(loggerServiceMock);
    });

    after(() => {
      sinon.restore();
    });

    it('should fetch the same logger instance', () => {
      NASUploader.getStaticLogger();
      NASUploader.getStaticLogger();

      expect(loggerServiceMock.getInstance).to.be.calledOnceWith('NASUploader');
    });
  });

  describe('file name generation for backup files', () => {
    const data = {
      filePath: '/tmp/video-s1-000-001.mp4',
      metadata: {
        meetingId: 'm1',
        deviceId: 'd1',
      },
    };

    before(() => {
      sinon.stub(serviceRegistry, 'get').returns(loggerServiceMock);
    });

    after(() => {
      sinon.restore();
    });

    it('should create the right target filename to avoid overwriting files', () => {
      const uploader = new NASUploader(data);
      expect(uploader.fileName).to.equal('m1-d1-s1-000-001.mp4');
    });
  });
});
