import * as fs from 'fs-extra';
import * as path from 'path';
import { serializeError } from 'serialize-error';

import { Service } from '../../../../services/service';
import serviceRegistry from '../../../../services/serviceRegistry';

const uploadStatusMap = {
  OVERALL: 0,
  MAIN: 5,
  BACKUP: 10,
  MASK: 0b111111111111111,
  UPLOADING: 0b001,
  DONE: 0b010,
  ERROR: 0b100,
};

export const {
  BACKUP,
  MAIN,
  OVERALL,
  UPLOADING,
  DONE,
  ERROR,
} = uploadStatusMap;

/*
The uploadStatus is expressed in bit flags, there are 3 groups of bits:
bits [0-5]   indicate the overall status (combination of main and backup)
bits [5-10]  indicate the the main upload status
bits [10-15] indicate the backup upload status

00000   00111   00001
  |       |       |
BACKUP   MAIN   OVERALL

For each group:
bit at position 1 -> uploading status: 0 = not uploading / 1 = uploading
bit at position 2 -> done status: 0 = uncomplete / 1 = done
bit at position 3 -> error status: 0 = no errors / 1 = errors

   0       0       0       0       0
   |       |       |       |       |
UNUSED  UNUSED UPLOADING DONE    ERROR

Examples:

INACTIVE          00000 00000 00000 (decimal 0)
MAIN UPLOADING    00000 00001 00001 (decimal 33)
MAIN SUCCESSFUL   00000 00011 00001 (decimal 97)
MAIN FAILED       00000 00111 00001 (decimal 225)
BACKUP UPLOADING  00001 00?11 00001 (decimal 1121/1249)
BACKUP SUCCESSFUL 00011 00?11 00001 (decimal 3169/3297)
BACKUP FAILED     00111 00?11 00001 (decimal 7265/7393)
FINISHED          00011 00011 00011 (decimal 3171)
PROCESSING_ERROR  00011 00011 00111 (decimal 3175)
*/

export class BaseUploader extends Service {
  constructor({
    filePath,
    settings,
    metadata,
    backupUploader,
    backup,
    ...args
  }) {
    super(args);
    this._filePath = filePath;
    this._settings = settings;
    this.metadata = metadata;
    this.backupUploader = backupUploader;
    this.isBackup = !!backup;
    this.uploadStatus = UPLOADING << OVERALL;

    this.meteorClient = serviceRegistry.get('meteorClient');

    // if upload fails, report the upload status immediately
    this.on('unhealthy', () => {
      this.uploadStatus |= ERROR << (this.isBackup ? BACKUP : MAIN);
      this.reportUploadState({
        uploadError: this.fault,
      }).catch(() => {
        this._log('error', 'Failed to report error');
      });
    });
  }

  _checkDestroyToken() {
    return false;
  }

  async _onStart() {
    // if we start a backup uploader, we already added a recording file
    // and now we just report a current state.
    if (this.isBackup) {
      this.uploadStatus |= UPLOADING << BACKUP;
      await this.reportUploadState();
    } else {
      this.uploadStatus |= (UPLOADING << MAIN);
      await this.meteorClient.addRecordingFile({
        recordingId: this.metadata._id,
        file: {
          ...this.getFileDescription(),
        },
      });
    }
    return super._onStart();
  }

  setUploadStatus(status) {
    this.uploadStatus |= status;
  }

  getFileDescription() {
    const filename = path.basename(this._filePath);
    return {
      filename,
      deviceId: this.metadata.deviceId,
      title: this.metadata.sourceTitle,
      type: this.metadata.sourceType,
      status: this.uploadStatus,
    };
  }

  /**
   * Reports file uploading progress towards cloud
   * @param {{ uploadError?: string, uploadedBytes?: number, uploadFinished?: boolean, backup?: boolean }}params
   * @return {Promise<void>}
   */
  async reportUploadState(params = {}) {
    const fileDescription = this.getFileDescription();

    if (params.uploadError) {
      this._log('error', 'Recording Error: failed during upload', {
        ...fileDescription,
        error: params.uploadError,
        serviceLog: true,
      });
    }

    await this.meteorClient.updateRecordingFile({
      recordingId: this.metadata._id,
      file: {
        ...fileDescription,
        ...params,
      },
    });
  }

  destroy(token) {
    if (this._checkDestroyToken(token)) {
      return super.destroy();
    } else {
      this._log('error', 'Service can only be destroyed internally.');
    }
  }

  async _onDestroy() {
    let mask = DONE;
    if (this.isUnhealthy) {
      mask |= ERROR;
    }
    const position = this.isBackup ? BACKUP : MAIN;
    mask <<= position;
    if (!this.backupUploader) {
      mask |= (DONE << OVERALL);
    }
    this.uploadStatus |= mask;

    // if upload failed, then we reset size to 0
    let size = 0;
    if (this.isHealthy) {
      const stats = await fs.stat(this._filePath);
      size = stats.size;
    }

    await this.reportUploadState({
      uploadedBytes: size,
    });

    if (this.backupUploader) {
      // If the uploader is unhealthy that could mean the upload failed
      // for safety, try to store the file on the backup to prevent data loss
      try {
        this._log('info', 'Using backup uploader to store recording');
        this.backupUploader.setUploadStatus(this.uploadStatus);
        await this.backupUploader.start();
        await this.backupUploader.stop();
      } catch (err) {
        this._log('error', `Backup uploader failed to store recording with error: ${JSON.stringify(serializeError(err, { maxDepth: 1 }))}`, {
          serviceLog: true,
        });
      }
    } else {
      // file will be removed if it's a main uploader without a backup settings, or a backup uploader
      try {
        await fs.unlink(this._filePath);
      } catch (err) {
        this._log('error', `Failed to remove tmp file from path: ${this._filePath} with error: ${JSON.stringify(serializeError(err, { maxDepth: 1 }))}`, {
          serviceLog: true,
        });
      }
    }
    return super._onDestroy();
  }
}

export default BaseUploader;
