import { Kaltura, KalturaMediaType, KalturaSessionType } from '@barco/util-kaltura';
import { serializeError } from 'serialize-error';
import MultipartUploader from './multipartUploader';
import serviceRegistry from '../../../../services/serviceRegistry';
import { proxyConfigToUrl } from '../../../../utils/network';

const unhealthyMessage = 'Kaltura Recording service is not available';
const KALTURA_REQUEST_TIMEOUT = Number.parseInt(process.env.KALTURA_REQUEST_TIMEOUT, 10) || 180 * 1000;
const UNCHANGING_HEADER_BYTES = 32;
const CHUNK_SIZE = 2 * 1024 * 1024;

function createKalturaClient(settings) {
  const proxy = serviceRegistry.get('configManager').getSetting('device.proxy');
  return new Kaltura({
    proxy: proxyConfigToUrl(proxy),
    serviceUrl: settings.url,
    timeout: settings.hasOwnProperty('timeout') ? settings.timeout : KALTURA_REQUEST_TIMEOUT,
    hooks: settings.hooks,
    retryLimit: settings.retryLimit,
  });
}
export default class KalturaUploader extends MultipartUploader {
  // The uploading strategy is as follows (needed for Kaltura upload API):
  //
  // We first upload the first chunk. This is a chunk of 32 bytes as this part in the MP4 file does not
  // change anymore when the file is being completely written. Uploading of the second chunk is skipped
  // then as in this 'header' chunk changes are being written when the file is finalized. Uploading of
  // the chunks then proceeds as normal until the recording is finished.
  //
  // When the recording is finished the second chunk is uploaded first as the changes to the header are
  // now written. We wait until the upload of the second chunk is finished and only then we upload the
  // final chunk with the 'isFinal' flag.
  async initializeUpload() {
    try {
      const log = this._log.bind(this);
      this._client = createKalturaClient({
        ...this._settings,
        retryLimit: 10,
        hooks: {
          error: ({ error, url }) => {
            log('error', `Failed to execute kaltura url ${url} with error: ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`, {
              serviceLog: true,
            });
          },
          response: ({
            url, statusCode, statusMessage, headers,
          }) => {
            log('verbose', `Executed Kaltura url ${url} executed with response ${statusCode} (${statusMessage}), headers: ${JSON.stringify(headers)}`);
          },
          retry: ({ error, url, retryCount }) => {
            log('info', `Retrying failed kaltura url ${url} for error (# ${retryCount}): ${error ? JSON.stringify(serializeError(error, { maxDepth: 1 })) : ''}`);
          },
        },
      });

      await this._client.startSession({
        secret: this._settings.password,
        userId: this._settings.user,
        type: this._settings.hasOwnProperty('accountType') ? this._settings.accountType : KalturaSessionType.USER,
        partnerId: this._settings.partnerId,
      });
      const id = await this._client.addUploadToken();
      this._uploadTokenId = id;
      this._log('debug', `Started Kaltura upload for partner ID ${this._settings.partnerId} with uploadToken: ${this._uploadTokenId}`);
    } catch (error) {
      this.markUnhealthy(unhealthyMessage);
      throw error;
    }

    // create media entry
    await this._createMediaEntry();
  }

  async finalizeUpload() {
    try {
      await this._client.endSession();
    } catch (error) {
      this.markUnhealthy(unhealthyMessage);
      throw error;
    }
  }

  /**
  * @param {number} part
  * @return {ChunkDetails}
  */
  getChunkDetails(part) {
    const partSizes = [UNCHANGING_HEADER_BYTES, CHUNK_SIZE - UNCHANGING_HEADER_BYTES];

    return {
      size: partSizes[part - 1] || CHUNK_SIZE,
      awaitUpload: part === 1,
    };
  }

  /**
   * Should return the part number of the changing header chunk.
   * @return {number}
   */
  get changingHeaderChunkPartNumber() {
    return 2;
  }

  async uploadChunk(chunk, { position, isInitial = false, isFinal = false }) {
    try {
      this._log('verbose', `uploadChunk position:${position} isInitial:${isInitial} isFinal:${isFinal} length:${chunk.length}`);
      return await this._client.uploadChunk(this._uploadTokenId, {
        fileData: chunk,
        resume: !isInitial,
        finalChunk: isFinal,
        resumeAt: position,
      });
    } catch (error) {
      this.markUnhealthy(unhealthyMessage);
      throw error;
    }
  }

  async _createMediaEntry() {
    try {
      const {
        meetingId,
        title,
        hosts,
        hostIdentifiers,
        locations,
        configuration,
        description,
      } = this.metadata;

      const mediaEntryId = await this._client.addMediaEntry({
        name: title,
        description,
        mediaType: KalturaMediaType.VIDEO,
        tags: [meetingId, hosts, hostIdentifiers, locations, configuration]
          .filter(tag => !!tag)
          .join(','),
      });

      return await this._client.addMediaEntryTokenResource(mediaEntryId, {
        token: this._uploadTokenId,
      });
    } catch (error) {
      this.markUnhealthy(unhealthyMessage);
      throw error;
    }
  }
}

KalturaUploader.validate = async function validate(settings) {
  const client = createKalturaClient({ ...settings, retryLimit: 0 });
  await client.startSession({
    secret: settings.password,
    userId: settings.user,
    type: settings.hasOwnProperty('accountType') ? settings.accountType : KalturaSessionType.USER,
    partnerId: settings.partnerId,
  });
  await client.endSession();

  return true;
};

export { KalturaUploader };
