import * as chai from 'chai';
import * as chaiAsPromised from 'chai-as-promised';
import * as nock from 'nock';
import * as urljoin from 'url-join';
import * as sinon from 'sinon';
import * as sinonChai from 'sinon-chai';
import * as express from 'express';
import * as http from 'http';
import * as proxy from 'proxy';
import serviceRegistry from '../../../../services/serviceRegistry';
import {
  registerAll,
  unregisterAll,
} from '../../../../utils/test-helpers/fillServiceRegistry';
import KalturaUploader from './kalturaUploader';

const { expect } = chai;
chai.use(chaiAsPromised);
chai.use(sinonChai);

describe('KalturaUploader', () => {
  const url = 'https://test.kaltura.com';
  const serviceBase = '/api_v3/service';
  const KalturaAPIException = 'KalturaAPIException';
  const settingsMap = {};
  const metadata = {
    meetingId: 'MEETING_ID',
    title: 'TITLE',
    hosts: 'HOSTS',
    hostIdentifiers: 'HOST_IDENTIFIERS',
    locations: 'LOCATIONS',
    configuration: 'CONFIGURATION',
    description: 'DESCRIPTION',
  };
  const kalturaSettings = {
    url,
    userId: 'TEST_USER',
    secret: 'TEST_SECRET',
    accountType: 0,
    partnerId: 'TEST_PARTNER_ID',
  };
  let clock;
  let server;
  let proxyServer;

  const createAndInitializeKaltura = async (args) => {
    const scope = nock(url)
      .post(urljoin(serviceBase, '/session/action/start'))
      .reply(200, '"SESSION_TOKEN"')
      .post(urljoin(serviceBase, '/uploadtoken/action/add'))
      .reply(200, { id: 'UPLOAD_TOKEN' })
      .post(
        urljoin(serviceBase, 'media/action/add'),
        {
          ks: 'SESSION_TOKEN',
          format: 1,
          entry: {
            name: 'TITLE',
            description: 'DESCRIPTION',
            mediaType: 1,
            tags: 'MEETING_ID,HOSTS,HOST_IDENTIFIERS,LOCATIONS,CONFIGURATION',
          },
        },
      )
      .reply(200, { id: 'MEDIA_ID' })
      .post(
        urljoin(serviceBase, 'media/action/addContent'),
        {
          entryId: 'MEDIA_ID',
          ks: 'SESSION_TOKEN',
          format: 1,
          resource: {
            token: 'UPLOAD_TOKEN',
            objectType: 'KalturaUploadedFileTokenResource',
          },
        },
      )
      .reply(200, '""');

    const kalturaUploader = new KalturaUploader({ ...args, metadata });

    await expect(kalturaUploader.initializeUpload()).not.to.be.rejected;

    await clock.runAllAsync();

    expect(scope.isDone()).to.be.true;

    return kalturaUploader;
  };

  before(() => {
    registerAll();
    serviceRegistry.set('configManager', {
      getSetting: (value) => settingsMap[value],
      on: () => { },
      removeListener: () => { },
    });
    nock.restore();
    nock.activate();
  });

  after(() => {
    unregisterAll();
    serviceRegistry.delete('configManager');
    nock.restore();
  });

  beforeEach(() => {
    settingsMap['device.proxy'] = null;
    clock = sinon.useFakeTimers({
      now: Date.now(),
    });
  });

  afterEach(() => {
    nock.cleanAll();
    clock.restore();

    if (server) {
      server.close();
      server = null;
    }

    if (proxyServer) {
      proxyServer.close();
      proxyServer = null;
    }
  });

  it('should validate the correct kaltura configuration', async () => {
    const scope = nock(url)
      .post(urljoin(serviceBase, '/session/action/start'))
      .reply(200, '"SESSION_TOKEN"')
      .post(urljoin(serviceBase, '/session/action/end'))
      .reply(200);

    expect(await KalturaUploader.validate(kalturaSettings)).to.be.true;

    expect(scope.isDone()).to.be.true;
  });

  it('should not validate the incorrect kaltura configuration', async () => {
    const scope = nock(url)
      .post(urljoin(serviceBase, '/session/action/start'))
      .reply(200, {
        objectType: KalturaAPIException,
        message: 'START_SESSION_ERROR',
      });

    await expect(KalturaUploader.validate(kalturaSettings)).to.be.rejected;

    expect(scope.isDone()).to.be.true;
  });

  it('should not retry failing requests when validating the kaltura configuration', async () => {
    // the mocked upload post query will fail the first time and succeed the second time
    const failingStartScope = nock(url)
      .post(urljoin(serviceBase, '/session/action/start'))
      .reply(500);
    const succeedingStartScope = nock(url)
      .post(urljoin(serviceBase, '/session/action/start'))
      .reply(200, '"SESSION_TOKEN"');

    const promise = KalturaUploader.validate(kalturaSettings);

    await clock.runAllAsync();

    expect(promise).to.be.rejected;
    expect(failingStartScope.isDone()).to.be.true;
    expect(succeedingStartScope.isDone()).to.be.false;
  });

  it('should initialize the upload', async () => {
    await createAndInitializeKaltura({
      settings: {
        ...kalturaSettings,
        timeout: undefined,
      },
    });
  });

  it('should upload the chunk', async () => {
    const kalturaUploader = await createAndInitializeKaltura({
      settings: kalturaSettings,
    });

    const scope = nock(url)
      .post(urljoin(serviceBase, '/uploadtoken/action/upload'))
      .query({
        ks: 'SESSION_TOKEN',
        format: 1,
        uploadTokenId: 'UPLOAD_TOKEN',
        resume: 'true',
        resumeAt: '100',
        finalChunk: 'false',
      })
      .reply(200, '""');

    await expect(kalturaUploader.uploadChunk(
      Buffer.alloc(100),
      { isInitial: false, position: 100, isFinal: false },
    )).not.to.be.rejected;

    expect(scope.isDone()).to.be.true;
  });

  it('should retry uploading the chunk 10 times', async () => {
    const kalturaUploader = await createAndInitializeKaltura({
      settings: {
        ...kalturaSettings,
        timeout: undefined,
      },
    });

    const uploadQuery = {
      ks: 'SESSION_TOKEN',
      format: 1,
      uploadTokenId: 'UPLOAD_TOKEN',
      resume: 'true',
      resumeAt: '100',
      finalChunk: 'false',
    };

    // the mocked upload post query will fail 10 times and succeed the 11th time
    const scope = nock(url)
      .post(urljoin(serviceBase, '/uploadtoken/action/upload'))
      .query(uploadQuery)
      .times(10)
      .reply(429)
      .post(urljoin(serviceBase, '/uploadtoken/action/upload'))
      .query(uploadQuery)
      .reply(200, '""');

    let isResolved = false;

    const uploadChunkPromise = kalturaUploader.uploadChunk(
      Buffer.alloc(100),
      { isInitial: false, position: 100, isFinal: false },
    ).then((x) => {
      isResolved = true;
      return x;
    });

    while (!isResolved) {
      // eslint-disable-next-line no-await-in-loop
      await clock.runAllAsync();
    }

    await expect(uploadChunkPromise).not.to.be.rejected;

    expect(scope.isDone()).to.be.true;
  });

  it('should mark the service as unhealthy if the upload fails', async () => {
    const kalturaUploader = await createAndInitializeKaltura({
      settings: kalturaSettings,
    });

    const uploadQuery = {
      ks: 'SESSION_TOKEN',
      format: 1,
      uploadTokenId: 'UPLOAD_TOKEN',
      resume: 'true',
      resumeAt: '100',
      finalChunk: 'false',
    };

    const scope = nock(url)
      .post(urljoin(serviceBase, '/uploadtoken/action/upload'))
      .query(uploadQuery)
      .times(11)
      .reply(429);

    let isErrored = false;

    const uploadchunkPromise = kalturaUploader.uploadChunk(
      Buffer.alloc(100),
      { isInitial: false, position: 100, isFinal: false },
    ).catch((error) => {
      isErrored = true;
      throw error;
    });

    // let all the timers run until the promise is resolved
    while (!isErrored) {
      // eslint-disable-next-line no-await-in-loop
      await clock.runAllAsync();
    }

    await expect(uploadchunkPromise).to.be.rejected;
    expect(kalturaUploader.isHealthy).to.be.false;
    expect(scope.isDone()).to.be.true;
  });

  it('should finalize the upload', async () => {
    const kalturaUploader = await createAndInitializeKaltura({
      settings: kalturaSettings,
    });

    const scope = nock(url)
      .post(urljoin(serviceBase, '/session/action/end'))
      .reply(200, '""');

    await expect(kalturaUploader.finalizeUpload()).not.to.be.rejected;

    expect(scope.isDone()).to.be.true;
  });

  it('should work using a proxy', async () => {
    // proxies cannot be mocked with nock
    nock.restore();

    settingsMap['device.proxy'] = {
      url: 'http://localhost:3334',
    };

    const viaHeaders = [];

    let resolveStart;
    const startPromise = new Promise((resolve) => { resolveStart = resolve; });
    let resolveAdd;
    const addPromise = new Promise((resolve) => { resolveAdd = resolve; });
    let resolveAddMedia;
    const addMediaPromise = new Promise((resolve) => { resolveAddMedia = resolve; });
    let resolveAddMediaContent;
    const addMediaContentPromise = new Promise((resolve) => { resolveAddMediaContent = resolve; });

    // start a kalture api server
    const app = express();
    app.post(`${serviceBase}/session/action/start`, (req, res) => {
      if (req.headers.via) {
        viaHeaders.push(req.headers.via);
      }
      res.json('"SESSION_TOKEN"');
      resolveStart();
    });
    app.post(`${serviceBase}/uploadtoken/action/add`, (req, res) => {
      if (req.headers.via) {
        viaHeaders.push(req.headers.via);
      }
      res.json({ id: 'UPLOAD_TOKEN' });
      resolveAdd();
    });
    app.post(`${serviceBase}/media/action/add`, (req, res) => {
      if (req.headers.via) {
        viaHeaders.push(req.headers.via);
      }
      res.json({ id: 'MEDIA_ID' });
      resolveAddMedia();
    });
    app.post(`${serviceBase}/media/action/addContent`, (req, res) => {
      if (req.headers.via) {
        viaHeaders.push(req.headers.via);
      }
      res.json('');
      resolveAddMediaContent();
    });
    server = app.listen('3333');

    // start a local proxy server
    proxyServer = proxy(http.createServer());
    proxyServer.listen(3334);

    const kalturaUploader = new KalturaUploader({
      settings: {
        ...kalturaSettings,
        url: 'http://localhost:3333',
      },
      metadata,
    });

    await expect(kalturaUploader.initializeUpload()).not.to.be.rejected;

    await Promise.all([
      startPromise,
      addPromise,
      addMediaPromise,
      addMediaContentPromise,
    ]);

    // if the request is proxied, the via header should be set
    expect(viaHeaders.length).to.equal(4);
  });
});
