import * as fs from 'fs-extra';
import { sleep } from '../../../../utils/utils';
import BaseUploader from './base-uploader';

const CHUNK_READ_INTERVAL = 1000;
const DESTROY_TOKEN = Symbol('destroy-token');

/**
 * Metadata for a chunk which has to be uploaded
 * @typedef {Object} ChunkDetails
 * @property {boolean} awaitUpload - Should the chunk should be uploaded.
 * @property {number} size - The chunk size.
 */

class MultipartUploader extends BaseUploader {
  constructor(args) {
    super(args);
    this._totalBytesRead = 0;
    this._finished = false;
    this._uploads = [];
    this._part = 1;
    this._remainingHeaderChunkDetails = undefined;

    // if upload fails in one of the last steps, we want to cleanup and remove file
    this.on('unhealthy', () => {
      if (this._finished) {
        this.destroy(DESTROY_TOKEN);
      }
    });
  }

  /**
   * Should implement how upload is initialized
   * Creating session, setting upload tokens etc...
   *
   * @return {Promise<void>}
   */
  async initializeUpload() {
    throw new Error('Should be implemented in child class');
  }

  /**
   * Should implement upload logic
   *
   * @param {buffer} chunk
   * @param {*} options
   * @return {Promise<*>}
   */
  // eslint-disable-next-line no-unused-vars
  async uploadChunk(chunk, options) {
    throw new Error('Should be implemented in child class');
  }

  /**
   * Should implement upload closing logic.
   * @param {array} uploads
   * @return {Promise<void>}
   */
  // eslint-disable-next-line no-unused-vars
  async finalizeUpload(uploads) {
    throw new Error('Should be implemented in child class');
  }

  /**
   * Should return the chunk detail for the part.
   * @param {number} part
   * @return {ChunkDetails}
   */
  // eslint-disable-next-line no-unused-vars
  getChunkDetails(part) {
    throw new Error('Should be implemented in child class');
  }

  /**
   * Should return the part number of the changing header chunk.
   * @return {number}
   */
  get changingHeaderChunkPartNumber() {
    throw new Error('Should be implemented in child class');
  }

  // Destroy allowed only from within the service with a token
  _checkDestroyToken(token) {
    return token === DESTROY_TOKEN;
  }

  async _onStart() {
    try {
      await super._onStart();
      this._fd = await fs.open(this._filePath, 'r');
      await this.initializeUpload();
      this._processChunk();
    } catch (err) {
      this._log('error', `failed to start uploading: ${err.message}`);
      this.markUnhealthy(`Failed to start uploading: ${err.message}`);
    }
  }

  async _onStop() {
    this._finished = true;
    await super._onStop();
    if (this.isUnhealthy) {
      // no await here! It should first stop before it could be destroyed
      this.destroy(DESTROY_TOKEN);
    }
  }

  async _processChunk() {
    const chunkDetails = this.getChunkDetails(this._part);
    if (this._part === this.changingHeaderChunkPartNumber) {
      this._remainingHeaderChunkDetails = {
        size: chunkDetails.size,
        part: this._part,
        position: this._totalBytesRead,
      };
    }

    if (this.isUnhealthy) {
      return;
    }

    const chunk = await this._readChunk(chunkDetails.size);
    if (!chunk) {
      // reading failed
      return;
    }

    // in case file ends in an equal chunk size, it will send one more chunk of 0 bytes, which is acceptable
    const isFinalChunk = this._finished && chunk.length < chunkDetails.size;
    if (isFinalChunk) {
      if (this._remainingHeaderChunkDetails) {
        const remainingHeaderChunkPromise = this._uploadRemainingHeaderChunk(this._remainingHeaderChunkDetails);
        this._uploads.splice(this._remainingHeaderChunkDetails.part - 1, 0, remainingHeaderChunkPromise);
      }
      // first send all non final chunks and wait for them to upload before doing the final one
      await Promise.all(this._uploads);
    }

    // we skip the header chunk, because the header will change when file gets closed
    if (this._part !== this.changingHeaderChunkPartNumber) {
      const uploadPromise = this._uploadChunk(chunk, isFinalChunk);
      if (chunkDetails.awaitUpload) {
        await uploadPromise;
      }
    }

    this._part += 1;
    this._totalBytesRead += chunk.length;

    if (isFinalChunk) {
      this._finalizeUpload();
    } else {
      this._processChunk();
    }
  }

  async _uploadChunk(chunk, isFinalChunk) {
    const uploadedBytes = this._totalBytesRead;
    const chunkOptions = {
      position: uploadedBytes,
      part: this._part,
      isInitial: this._part === 1,
      isFinal: isFinalChunk,
    };
    const uploadPromise = this.uploadChunk(
      chunk,
      chunkOptions,
    );
    this._uploads.push(uploadPromise);
    uploadPromise
      .then(() => {
        this.reportUploadState({
          uploadedBytes,
        });
      })
      .catch((err) => {
        this._log('error', `Failed to upload chunk: ${err.message}, chunk details ${JSON.stringify(chunkOptions)}`);
        this._log('error', err);
        this.markUnhealthy(`Failed to upload recording to the configured recording service: ${err.message}`);
      });

    return uploadPromise;
  }

  async _readChunk(chunkSize) {
    try {
      const buffer = Buffer.alloc(chunkSize);
      let chunkBytesRead = 0;
      let continueReadingChunk = true;

      while (continueReadingChunk) {
        // eslint-disable-next-line no-await-in-loop
        const { bytesRead } = await fs.read(this._fd, buffer, chunkBytesRead, chunkSize - chunkBytesRead);
        chunkBytesRead += bytesRead;
        continueReadingChunk = chunkBytesRead < chunkSize && !this._finished;

        if (continueReadingChunk) {
          // eslint-disable-next-line no-await-in-loop
          await sleep(CHUNK_READ_INTERVAL);
        }
      }

      const chunk = chunkBytesRead < chunkSize ? buffer.subarray(0, chunkBytesRead) : buffer;
      return chunk;
    } catch (err) {
      const errMessage = `Could not read next chunk ${err.message}`;
      this._log('error', errMessage);
      this.markUnhealthy(errMessage);
      return false;
    }
  }

  async _finalizeUpload() {
    try {
      const uploadResults = await Promise.all(this._uploads);
      await this.finalizeUpload(uploadResults);
      await this.reportUploadState({ uploadedBytes: this._totalBytesRead });
    } catch (err) {
      const errorMessage = `Failed to finalize upload: ${err.message}`;
      this.markUnhealthy(errorMessage);
      this._log('error', errorMessage);
    } finally {
      this.destroy(DESTROY_TOKEN);
    }
  }

  async _uploadRemainingHeaderChunk(changingHeaderChunkDetails) {
    const buffer = Buffer.alloc(changingHeaderChunkDetails.size);
    const { bytesRead } = await fs.read(
      this._fd, buffer, 0, changingHeaderChunkDetails.size, changingHeaderChunkDetails.position,
    );
    return this.uploadChunk(
      buffer.subarray(0, bytesRead),
      {
        position: changingHeaderChunkDetails.position,
        part: changingHeaderChunkDetails.part,
        isInitial: false,
        // if the recording is very short this part could be the last one
        isFinal: this._part === changingHeaderChunkDetails.part,
      },
    );
  }

  destroy(token) {
    if (this._fd) {
      fs.close(this._fd);
    }
    super.destroy(token);
  }
}

export default MultipartUploader;
