import * as chai from 'chai';
import * as chaiAsPromised from 'chai-as-promised';
import * as sinon from 'sinon';
import * as sinonChai from 'sinon-chai';
import * as fs from 'fs-extra';
import serviceRegistry from '../../../../services/serviceRegistry';
import {
  registerAll,
  unregisterAll,
} from '../../../../utils/test-helpers/fillServiceRegistry';
import MultipartUploader from './multipartUploader';

const { expect } = chai;
chai.use(chaiAsPromised);
chai.use(sinonChai);

const UNCHANGING_HEADER_BYTES = 64;
const CHUNK_SIZE = 1024;

class TestUploader extends MultipartUploader {
  constructor(args) {
    super(args);
    this._getChunkDetails = args.getChunkDetails
      || (() => ({
        size: CHUNK_SIZE,
        awaitUpload: false,
      }));
    this._changingHeaderChunkPartNumber = args.changingHeaderChunkPartNumber || -1;
  }

  async initializeUpload() {
    return undefined;
  }

  async uploadChunk() {
    return undefined;
  }

  async finalizeUpload() {
    return undefined;
  }

  getChunkDetails(part) {
    return this._getChunkDetails(part);
  }

  get changingHeaderChunkPartNumber() {
    return this._changingHeaderChunkPartNumber;
  }
}

const dummyFilePath = './MultipartUploader_test_file.mp4';

const createDummyFile = async (size) => {
  const chunkSize = 1024;
  const stream = fs.createWriteStream(dummyFilePath);

  const chunks = Math.floor(size / chunkSize);
  const lastChunkSize = size % chunkSize;

  for (let i = 0; i < chunks; i++) {
    // eslint-disable-next-line no-await-in-loop
    await stream.write(Buffer.alloc(chunkSize));
  }

  await stream.end(Buffer.alloc(lastChunkSize));
};

const deleteDummyFile = async () => {
  try {
    await fs.unlink(dummyFilePath);
  } catch { }
};

describe('MultipartUploader', () => {
  const sandbox = sinon.createSandbox();
  const settingsMap = {};
  let clock;

  before(() => {
    registerAll();
    serviceRegistry.set('configManager', {
      getSetting: (value) => settingsMap[value],
      on: () => { },
      removeListener: () => { },
    });
    serviceRegistry.set('meteorClient', {
      addRecordingFile: async () => { },
      updateRecordingFile: async () => { },
      on: () => { },
      removeListener: () => { },
    });
  });

  after(() => {
    unregisterAll();
    serviceRegistry.delete('configManager');
    serviceRegistry.delete('meteorClient');
  });

  beforeEach(() => {
    clock = sandbox.useFakeTimers({
      now: Date.now(),
    });
  });

  afterEach(async () => {
    sandbox.restore();
    await deleteDummyFile();
  });

  describe('basic functionality', () => {
    // The number of parts + 1 part of zero bytes if the file exactly matches the total chunk size
    const getNumberOfChunks = (fileSize, chunkSize = CHUNK_SIZE) => (
      Math.floor(fileSize / chunkSize) + 1
    );

    let testUploader;

    beforeEach(() => {
      testUploader = new TestUploader({
        filePath: dummyFilePath,
        metadata: {},
      });
      sandbox.spy(testUploader);
    });

    it('should upload all the parts correctly', async () => {
      const fileSize = CHUNK_SIZE * 20;
      await createDummyFile(fileSize);

      testUploader
        .start()
        .then(() => testUploader.stop());

      while (!testUploader.isDestroyed) {
        // eslint-disable-next-line no-await-in-loop
        await clock.runAllAsync();
      }

      expect(testUploader.uploadChunk).to.have.callCount(getNumberOfChunks(fileSize));
    });

    it('should call initializeUpload before uploading the first chunk', async () => {
      const fileSize = CHUNK_SIZE * 2;
      await createDummyFile(fileSize);

      testUploader
        .start()
        .then(() => testUploader.stop());

      while (!testUploader.isDestroyed) {
        // eslint-disable-next-line no-await-in-loop
        await clock.runAllAsync();
      }

      expect(testUploader.initializeUpload).to.be.calledOnce;
      expect(testUploader.initializeUpload.getCall(0)).to.be.calledBefore(testUploader.uploadChunk.getCall(0));
    });

    it('should call finalizeUpload after uploading the last chunk', async () => {
      const fileSize = CHUNK_SIZE * 3;
      await createDummyFile(fileSize);

      testUploader
        .start()
        .then(() => testUploader.stop());

      while (!testUploader.isDestroyed) {
        // eslint-disable-next-line no-await-in-loop
        await clock.runAllAsync();
      }

      const lastUploadCall = testUploader.uploadChunk.getCall(getNumberOfChunks(fileSize) - 1);
      expect(testUploader.finalizeUpload).to.be.calledOnce;
      expect(testUploader.finalizeUpload.getCall(0)).to.be.calledAfter(lastUploadCall);
    });

    it('should upload all the parts correctly for a custom chunk size', async () => {
      const chunkSize = 377;
      const fileSize = chunkSize * 20;
      await createDummyFile(fileSize);

      const testUploaderCustomChunkSize = new TestUploader({
        filePath: dummyFilePath,
        metadata: {},
        getChunkDetails: (() => ({
          size: chunkSize,
          awaitUpload: false,
        })),
      });
      sandbox.spy(testUploaderCustomChunkSize);

      testUploaderCustomChunkSize
        .start()
        .then(() => testUploaderCustomChunkSize.stop());

      while (!testUploaderCustomChunkSize.isDestroyed) {
        // eslint-disable-next-line no-await-in-loop
        await clock.runAllAsync();
      }

      expect(testUploaderCustomChunkSize.uploadChunk).to.have.callCount(getNumberOfChunks(fileSize, chunkSize));
    });
  });

  describe('changing header chunks', () => {
    // The unchanging header chunk + the other parts + 1 part of zero bytes
    // if the file exactly matches the total chunk size
    const getNumberOfChunks = (fileSize, chunkSize = CHUNK_SIZE) => (
      1 + Math.floor((fileSize - UNCHANGING_HEADER_BYTES) / chunkSize) + 1
    );

    let testUploader;

    beforeEach(() => {
      testUploader = new TestUploader({
        filePath: dummyFilePath,
        metadata: {},
        getChunkDetails: ((part) => ({
          size: part === 1 ? UNCHANGING_HEADER_BYTES : CHUNK_SIZE,
          awaitUpload: part === 1,
        })),
        changingHeaderChunkPartNumber: 2,
      });

      sandbox.spy(testUploader);
    });

    it('should first upload the unchanging header part', async () => {
      const fileSize = CHUNK_SIZE * 20;
      await createDummyFile(fileSize);

      testUploader
        .start()
        .then(() => testUploader.stop());

      while (!testUploader.isDestroyed) {
        // eslint-disable-next-line no-await-in-loop
        await clock.runAllAsync();
      }

      expect(testUploader.uploadChunk.getCall(0)).to.be.calledWith(
        Buffer.alloc(UNCHANGING_HEADER_BYTES),
        {
          position: 0,
          part: 1,
          isInitial: true,
          isFinal: false,
        },
      );
    });

    it('should upload the remaining header chunk part as the second last', async () => {
      const fileSize = CHUNK_SIZE * 5;
      await createDummyFile(fileSize);

      testUploader
        .start()
        .then(() => testUploader.stop());

      while (!testUploader.isDestroyed) {
        // eslint-disable-next-line no-await-in-loop
        await clock.runAllAsync();
      }

      const secondLastUploadCall = testUploader.uploadChunk.getCall(getNumberOfChunks(fileSize) - 2);
      expect(secondLastUploadCall).to.be.calledWith(
        Buffer.alloc(CHUNK_SIZE),
        {
          position: UNCHANGING_HEADER_BYTES,
          part: 2,
          isInitial: false,
          isFinal: false,
        },
      );
    });

    it('should upload final part as the last', async () => {
      const fileSize = CHUNK_SIZE * 6;
      await createDummyFile(fileSize);

      testUploader
        .start()
        .then(() => testUploader.stop());

      while (!testUploader.isDestroyed) {
        // eslint-disable-next-line no-await-in-loop
        await clock.runAllAsync();
      }

      const lastUploadCall = testUploader.uploadChunk.getCall(getNumberOfChunks(fileSize) - 1);
      expect(lastUploadCall).to.be.calledWith(
        Buffer.alloc(CHUNK_SIZE - UNCHANGING_HEADER_BYTES),
        {
          position: UNCHANGING_HEADER_BYTES + 5 * CHUNK_SIZE,
          part: 7,
          isInitial: false,
          isFinal: true,
        },
      );
    });

    it('should upload each part with the correct part number', async () => {
      const fileSize = CHUNK_SIZE * 15;
      await createDummyFile(fileSize);

      testUploader
        .start()
        .then(() => testUploader.stop());

      while (!testUploader.isDestroyed) {
        // eslint-disable-next-line no-await-in-loop
        await clock.runAllAsync();
      }

      const calls = testUploader.uploadChunk.getCalls();

      for (let i = 0; i < calls.length; i++) {
        const call = calls[i];
        const { part } = call.args[1];

        if (i === 0 || i === calls.length - 1) {
          expect(part).to.equal(i + 1);
        } else if (i === calls.length - 2) {
          expect(part).to.equal(2);
        } else {
          expect(part).to.equal(i + 2);
        }
      }
    });

    it('should upload each part with the correct position', async () => {
      const fileSize = CHUNK_SIZE * 33;
      await createDummyFile(fileSize);

      testUploader
        .start()
        .then(() => testUploader.stop());

      while (!testUploader.isDestroyed) {
        // eslint-disable-next-line no-await-in-loop
        await clock.runAllAsync();
      }

      const calls = testUploader.uploadChunk.getCalls();

      for (let i = 0; i < calls.length; i++) {
        const call = calls[i];
        const { position } = call.args[1];

        if (i === 0) {
          expect(position).to.equal(0);
        } else if (i === calls.length - 2) {
          expect(position).to.equal(UNCHANGING_HEADER_BYTES);
        } else if (i === calls.length - 1) {
          expect(position).to.equal(UNCHANGING_HEADER_BYTES + (i - 1) * CHUNK_SIZE);
        } else {
          expect(position).to.equal(UNCHANGING_HEADER_BYTES + i * CHUNK_SIZE);
        }
      }
    });

    it('should upload a final part of 0 bytes when the file size exactly matches a multiple of the chunk sizes', async () => {
      const fileSize = CHUNK_SIZE * 7 + UNCHANGING_HEADER_BYTES;
      await createDummyFile(fileSize);

      testUploader
        .start()
        .then(() => testUploader.stop());

      while (!testUploader.isDestroyed) {
        // eslint-disable-next-line no-await-in-loop
        await clock.runAllAsync();
      }

      const lastUploadCall = testUploader.uploadChunk.getCall(getNumberOfChunks(fileSize) - 1);
      expect(lastUploadCall).to.be.calledWith(
        Buffer.alloc(0),
        {
          position: UNCHANGING_HEADER_BYTES + 7 * CHUNK_SIZE,
          part: 9,
          isInitial: false,
          isFinal: true,
        },
      );
    });
  });
});
