import * as S3 from 'aws-sdk/clients/s3';
import * as path from 'path';
import * as proxy from 'proxy-agent';

import MultipartUploader from './multipartUploader';
import serviceRegistry from '../../../../services/serviceRegistry';
import { proxyConfigToUrl } from '../../../../utils/network';

const unhealthyMessage = 'Panopto Recording service is not available';
// Panopto (AWS S3) expects a minimum chunk size of 5 mb
const CHUNK_SIZE = 5 * 1024 * 1024;

function parseS3Params(session, filename) {
  const uploadTarget = new URL(session.UploadTarget);
  const pathElements = uploadTarget.pathname.replace(/^\//, '').split('/');
  const endpoint = `${uploadTarget.origin}/${pathElements[0]}`;
  const bucket = pathElements[1];
  const key = `${pathElements[2]}/${filename}`;
  return { endpoint, Bucket: bucket, Key: key };
}

export default class PanoptoUploader extends MultipartUploader {
  // The uploading strategy is as follows (needed for Kaltura upload API):
  //
  // Uploading of the first chunk is skipped as in this 'header' chunk
  // changes are being written when the file is finalized. Uploading of
  // the chunks then proceeds as normal until the recording is finished.
  //
  // When the recording is finished the first chunk is uploaded as the changes to the header are
  // now written.

  async initializeUpload() {
    try {
      const filename = path.basename(this._filePath);
      const { panoptoSession } = this.metadata;
      const { endpoint, ...s3Params } = parseS3Params(panoptoSession, filename);
      const options = {
        endpoint,
        accessKeyId: 'dummy',
        secretAccessKey: 'dummy',
        maxRetries: 10,
        // used for unit testing
        s3BucketEndpoint: panoptoSession.s3BucketEndpoint,
      };
      const proxySettings = serviceRegistry.get('configManager').getSetting('device.proxy');
      const proxyUrl = proxyConfigToUrl(proxySettings);
      if (proxyUrl) {
        options.httpOptions = { agent: proxy(proxyUrl) };
      }

      this._client = new S3(options);
      const result = await this._client.createMultipartUpload({ ...s3Params }).promise();
      s3Params.UploadId = result.UploadId;
      this._s3Params = s3Params;
      this._log('verbose', `Created upload to panopto, bucket:${this._s3Params.Bucket} key:${this._s3Params.Key} uploadId:${this._s3Params.UploadId}`);
    } catch (error) {
      this.markUnhealthy(unhealthyMessage);
      throw error;
    }
  }

  async uploadChunk(chunk, { part }) {
    try {
      const uploadResult = await this._client.uploadPart({
        Body: chunk,
        PartNumber: part,
        ...this._s3Params,
      }).promise();

      this._log('verbose', `Uploaded part #${part} to panopto, bucket:${this._s3Params.Bucket} key:${this._s3Params.Key} uploadId:${this._s3Params.UploadId}`);

      return {
        ...uploadResult,
        PartNumber: part,
      };
    } catch (error) {
      this.markUnhealthy(unhealthyMessage);
      throw error;
    }
  }

  async finalizeUpload(uploads) {
    try {
      await this._client.completeMultipartUpload({
        MultipartUpload: { Parts: uploads },
        ...this._s3Params,
      }).promise();

      this._log('verbose', `Finalized upload to panopto, bucket:${this._s3Params.Bucket} key:${this._s3Params.Key} uploadId:${this._s3Params.UploadId}`);
    } catch (error) {
      this.markUnhealthy(unhealthyMessage);
      throw error;
    }
  }

  /**
  * @return {ChunkDetails}
  */
  getChunkDetails() {
    return {
      size: CHUNK_SIZE,
      awaitUpload: false,
    };
  }

  /**
   * Should return the part number of the changing header chunk.
   * @return {number}
   */
  get changingHeaderChunkPartNumber() {
    return 1;
  }
}

// Validation already happens in Blue.
PanoptoUploader.validate = () => Promise.resolve(true);
