import * as chai from 'chai';
import * as chaiAsPromised from 'chai-as-promised';
import * as nock from 'nock';
import * as xml from 'xml';
import * as sinon from 'sinon';
import * as sinonChai from 'sinon-chai';
import * as express from 'express';
import * as http from 'http';
import * as proxy from 'proxy';
import serviceRegistry from '../../../../services/serviceRegistry';
import {
  registerAll,
  unregisterAll,
} from '../../../../utils/test-helpers/fillServiceRegistry';
import PanoptoUploader from './panoptoUploader';

const { expect } = chai;
chai.use(chaiAsPromised);
chai.use(sinonChai);

describe('PanoptoUploader', () => {
  const url = 'http://mybucket.test.panopto.com';
  const settingsMap = { };
  const panoptoArgs = {
    filePath: '/test/video.mp4',
    metadata: {
      panoptoSession: {
        UploadTarget: 'http://test.panopto.com/upload/mybucket/key',
      },
    },
  };
  const xlmResponseHeaders = {
    'Content-Type': 'application/xml',
  };
  let clock;
  let server;
  let proxyServer;

  const createAndInitializePanopto = async (args) => {
    const body = xml(
      {
        InitiateMultipartUploadResult: [
          { Bucket: 'mybucket' },
          { Key: 'key/video.mp4' },
          { UploadId: 'UPLOAD_ID' },
        ],
      },
    );
    const scope = nock(url)
      .post('/upload/key/video.mp4?uploads')
      .reply(200, body, xlmResponseHeaders);

    const panoptoUploader = new PanoptoUploader(args);

    let isResolved = false;

    panoptoUploader.initializeUpload()
      .then((x) => {
        isResolved = true;
        return x;
      });

    // let all the timers run until the promise is resolved
    while (!isResolved) {
      // eslint-disable-next-line no-await-in-loop
      await clock.runAllAsync();
    }
    expect(scope.isDone()).to.be.true;

    return panoptoUploader;
  };

  before(() => {
    registerAll();
    serviceRegistry.set('configManager', {
      getSetting: (value) => settingsMap[value],
      on: () => {},
      removeListener: () => {},
    });
    nock.restore();
    nock.activate();
  });

  after(() => {
    unregisterAll();
    serviceRegistry.delete('configManager');
    nock.restore();
  });

  beforeEach(() => {
    settingsMap['device.proxy'] = null;
    clock = sinon.useFakeTimers({
      now: Date.now(),
    });
  });

  afterEach(() => {
    nock.cleanAll();
    clock.restore();

    if (server) {
      server.close();
      server = null;
    }

    if (proxyServer) {
      proxyServer.close();
      proxyServer = null;
    }
  });

  it('should initialize the upload', async () => {
    await createAndInitializePanopto(panoptoArgs);
  });

  it('should upload the chunk', async () => {
    const panoptoUploader = await createAndInitializePanopto(panoptoArgs);

    const scope = nock(url)
      .put('/upload/key/video.mp4?partNumber=5&uploadId=UPLOAD_ID')
      .reply(200);

    let isResolved = false;

    const uploadPromise = panoptoUploader.uploadChunk(
      Buffer.alloc(100),
      { part: 5 },
    ).then((x) => {
      isResolved = true;
      return x;
    });

    // let all the timers run until the promise is resolved
    while (!isResolved) {
      // eslint-disable-next-line no-await-in-loop
      await clock.runAllAsync();
    }

    expect(await uploadPromise).to.eql({
      PartNumber: 5,
    });

    expect(scope.isDone()).to.be.true;
  });

  it('should retry uploading the chunk 10 times', async () => {
    const panoptoUploader = await createAndInitializePanopto(panoptoArgs);

    // the mocked upload post query will fail 10 times and succeed the 11th time
    const scope = nock(url)
      .put('/upload/key/video.mp4?partNumber=6&uploadId=UPLOAD_ID')
      .times(10)
      .reply(429)
      .put('/upload/key/video.mp4?partNumber=6&uploadId=UPLOAD_ID')
      .reply(200);

    let isResolved = false;

    const uploadPromise = panoptoUploader.uploadChunk(
      Buffer.alloc(100),
      { part: 6 },
    ).then((x) => {
      isResolved = true;
      return x;
    });

    while (!isResolved) {
      // eslint-disable-next-line no-await-in-loop
      await clock.runAllAsync();
    }

    expect(await uploadPromise).to.eql({
      PartNumber: 6,
    });

    expect(scope.isDone()).to.be.true;
  });

  it('should finalize the upload', async () => {
    const panoptoUploader = await createAndInitializePanopto(panoptoArgs);

    const body = xml(
      {
        CompleteMultipartUploadResult: [
          { Location: 'location' },
          { Bucket: 'mybucket' },
          { Key: 'key/video.mp4' },
          { ETag: 'ETAG' },
        ],
      },
    );
    const scope = nock(url)
      .post('/upload/key/video.mp4?uploadId=UPLOAD_ID')
      .reply(200, body, xlmResponseHeaders);

    let isResolved = false;

    const finalizePromise = panoptoUploader.finalizeUpload([
      { PartNumber: 1 },
      { PartNumber: 2 },
      { PartNumber: 3 },
    ]).then((x) => {
      isResolved = true;
      return x;
    });

    // let all the timers run until the promise is resolved
    while (!isResolved) {
      // eslint-disable-next-line no-await-in-loop
      await clock.runAllAsync();
    }

    await expect(finalizePromise).not.to.be.rejected;

    expect(scope.isDone()).to.be.true;
  });

  it('should mark the service as unhealthy if the upload fails', async () => {
    const panoptoUploader = await createAndInitializePanopto(panoptoArgs);

    // the mocked upload post query will fail 10 times and succeed the 11th time
    const scope = nock(url)
      .put('/upload/key/video.mp4?partNumber=6&uploadId=UPLOAD_ID')
      .times(11)
      .reply(429);

    let isErrored = false;

    const uploadPromise = panoptoUploader.uploadChunk(
      Buffer.alloc(100),
      { part: 6 },
    ).catch((error) => {
      isErrored = true;
      throw error;
    });

    // let all the timers run until the promise is resolved
    while (!isErrored) {
      // eslint-disable-next-line no-await-in-loop
      await clock.runAllAsync();
    }

    await expect(uploadPromise).to.be.rejected;

    expect(panoptoUploader.isHealthy).to.be.false;
    expect(scope.isDone()).to.be.true;
  });

  it('should work using a proxy', async () => {
    // proxies cannot be mocked with nock
    nock.restore();

    settingsMap['device.proxy'] = {
      url: 'http://localhost:3334',
    };

    const body = xml(
      {
        InitiateMultipartUploadResult: [
          { Bucket: 'mybucket' },
          { Key: 'key/video.mp4' },
          { UploadId: 'UPLOAD_ID' },
        ],
      },
    );

    let viaHeader;

    let resolveStart;
    const startPromise = new Promise((resolve) => { resolveStart = resolve; });

    // start an AWS api server
    const app = express();
    app.post('/upload/key/video.mp4', (req, res) => {
      viaHeader = req.headers.via;
      res.set('Content-Type', 'text/xml');
      res.send(body);
      resolveStart();
    });
    server = app.listen('3333');

    // start a local proxy server
    proxyServer = proxy(http.createServer());
    proxyServer.listen(3334);

    const panoptoUploader = new PanoptoUploader({
      ...panoptoArgs,
      metadata: {
        panoptoSession: {
          s3BucketEndpoint: true,
          UploadTarget: 'http://localhost:3333/upload/mybucket/key',
        },
      },
    });

    await expect(panoptoUploader.initializeUpload()).not.to.be.rejected;

    await startPromise;

    // if the request is proxied, the via header should be set
    expect(viaHeader).to.have.string('proxy');
  });
});
