import { serviceRegistry } from '../../../services/serviceRegistry';
import Stream from './stream';
import { Random } from '../../../utils/random';

const streamRegistry = serviceRegistry.get('RTSPRegistry');

/**
 * RTSP stream
 */
class RTSPStream extends Stream {
  constructor({
    port,
    path,
    url,
    params,
    paSrcDevice,
    username,
    password,
  }) {
    super({ id: Random.id() });

    this.port = port;
    this.path = path;
    this.url = url;
    this.params = params;
    this.paSrcDevice = paSrcDevice;
    this.username = username;
    this.password = password;
    this.recordSink = null;
  }

  /**
   * register stream to be recorded
   * @param {{ title:string, sourceType:string}} recordSink
   */
  registerForRecording(recordSink) {
    this.recordSink = recordSink;
    // Registry will cleanup itself when stream will get destroyed.
    // 2 things can trigger registering a stream for recording:
    // 1) the record attribute of an usb camera configuration being true
    // 2) a RECORD_STREAM sink being present for a certain composition (can be Cinergy or Browser composition)
    this._log('info', `Registering stream ${this.path} at 0.0.0.0:${this.port}/${this.url} for recording`);
    streamRegistry.set(this._serviceId, this);
  }

  get command() {
    return '/usr/lib/barco-edu-rtspserver/barco-edu-rtspserver';
  }

  get streamType() {
    return 'RTSP';
  }

  get args() {
    // We want to set a max size on the stream size, by setting the vaapih264enc rate-control
    // However, due to current limitations, we can't.
    // cqp (1) – Constant QP
    //    => Default, current best until vbr_constrained is possible
    //    In CQP mode, bitrate parameter should be a maximum bitrate,
    //    but this doesn't seem to work (it's ignored)
    //    However, 5000 would be a good maximum for full hd 30fps
    //    We do set init-qp on 30 (lower quality than the default of 26)
    //    This saves about 20% bandwith, while still being very good quality
    //    Higher than 30 degrades video quality too much
    //
    // cbr (2) – Constant bitrate
    //    => Bad for idle images like powerpoints
    //
    // vbr (4) – Variable bitrate
    //    => Unfortunately, this seems to act like CBR in our tests.
    //        Where bitrate is respected, but also matched on still images.
    //        Supported on i965 va driver 2.3.0, but not on the one in giles (1.7.2)
    //
    // vbr_constrained (5) – Variable bitrate - Constrained
    //      => This is what we want, but is not supported on driver version 2.3.0
    //
    let args = ` -e "quality-level=5 init-qp=30 rate-control=1 bitrate=5000" -p ${this.port} -u /${this.url} -s "${this.path}" -d `; // debian buster

    if (this.paSrcDevice) {
      args += this.paSrcDevice;
    } else {
      args += '-1 -n';
    }

    if (this.username) {
      args += ` --username ${this.username}`;
    }

    if (this.password) {
      args += ` --password ${this.password}`;
    }

    if (this.params) {
      args += ` ${this.params}`;
    }
    return args;
  }

  equals(stream) {
    return stream instanceof RTSPStream
      && (this.port === stream.port)
      && (this.path === stream.path)
      && (this.url === stream.url)
      && (this.params === stream.params)
      && (this.paSrcDevice === stream.paSrcDevice)
      && (this.username === stream.username)
      && (this.password === stream.password);
  }

  _onStart() {
    this._log('info', `Streaming ${this.path} at 0.0.0.0:${this.port}/${this.url}`);
    return super._onStart();
  }
}

export default RTSPStream;
