/**
 * Created by karolis on 6/26/17.
 */
import { URL } from 'url';
import * as _ from 'lodash';

import ServiceBridge from '../bridge';
import MeteorCacheManager from '../../../services/ddp/meteorCacheManager';
import {
  deviceRole,
  displayRole,
  compositionType,
  sinkType,
  deviceStatus,
  userRoleInMeeting,
} from '../../../services/ddp/constants';
import {
  findRtspContentStreamSource,
  deviceIsTile,
} from '../utils/audio/calculatePAConfig';

import IncomingAudioStream from './incomingAudioStream';
import MjpegStream from './mjpegStream';
import RtspAudioStream from './rtspAudioStream';
import RtspShmStream from './rtspShmStream';
import RtspScreenStream from './rtspScreenStream';
import cinergyUtils from '../cinergy/utils';

const deviceIsConnected = device => [deviceStatus.ONLINE, deviceStatus.WARNING].includes(device.status);
const hasPresenterOrRoomCameraWithAudio = device => device.hasOwnProperty('usbCameraConfig')
  && device.usbCameraConfig.some(conf => (conf.role === 'presenter' || conf.role === 'room') && conf.hasOwnProperty('audioDevice')
  && conf.audioDevice);
const deviceIsHostTile = device => deviceIsTile(device) && device.displayConfig
  && device.displayConfig.some(display => display.parameters.useForRole === userRoleInMeeting.HOST);
const deviceIsMaster = device => device.hasOwnProperty('displayConfig') && device.displayConfig.some(cfg => cfg.role === displayRole.MAIN_OUT);
const itsMe = (device, thisDevice) => (device._id === thisDevice._id);

// Allow at most one audio RTSP server per device (for this reason sink.device._id
// concatenated with meetingId is used i.o. sink._id) as stream id for audio RTSP servers.
const streamIdForSink = sink => (sink.type !== sinkType.RTSP_AUDIO_STREAM ? sink._id : `${sink.device._id}_${sink.meetingId}`);

// Avoid orphaned incoming audio rtsp streams when lectures are taken over by using meeting id as part of the stream id.
const rtspAudioPlaybackStreamId = (device, thisDevice) => `${device._id}_${thisDevice.meetingId}`;

const audioMixPullFrom = {};

const forceDisableDAM = !!process.env.FORCE_DISABLE_DAM;

// REMARK: in the future we should try to move as much as possible of this logic to blue i.o. purple.
//         In a similar way like we have video compositions and video sinks, we should introduce
//         audio compositions and audio sinks in blue, then the logic below can move to blue.

// Part 1: Virtual/Hybrid classroom
// Read as: a virtual camera manager needs to pull from at most one virtual_master
// and from any virtual tile, on the condition that virtual camera manager itself
// has a presenter or room camera with audio.
audioMixPullFrom[deviceRole.VIRTUAL_CAMERA_MANAGER] = [
  {
    deviceRoles: [deviceRole.VIRTUAL_MASTER],
    displayRoles: [displayRole.MAIN_OUT],
    thisDeviceCondDesc: 'The pulling device must have a Presenter or Room cam with audio',
    thisDeviceCondition: device => hasPresenterOrRoomCameraWithAudio(device),
    onlyPullOnce: true,
  },
  {
    deviceRoles: [deviceRole.VIRTUAL_TILE, deviceRole.VIRTUAL_OFFWALL_TILE],
    displayRoles: [displayRole.HTML_TILE_OUT, displayRole.HTML_OFFWALL_TILE_OUT],
    thisDeviceCondDesc: 'The pulling device must have a Presenter or Room cam with audio',
    thisDeviceCondition: device => hasPresenterOrRoomCameraWithAudio(device),
    onlyPullOnce: false,
  },
];
audioMixPullFrom[deviceRole.VIRTUAL_TILE] = [
  {
    deviceRoles: [deviceRole.VIRTUAL_TILE, deviceRole.VIRTUAL_OFFWALL_TILE],
    displayRoles: [displayRole.HTML_TILE_OUT, displayRole.HTML_OFFWALL_TILE_OUT],
    deviceCondDesc: 'The tested device may not be a host tile',
    deviceCondition: device => !deviceIsHostTile(device),
    onlyPullOnce: false,
  },
];
audioMixPullFrom[deviceRole.VIRTUAL_OFFWALL_TILE] = [
  {
    deviceRoles: [deviceRole.VIRTUAL_TILE, deviceRole.VIRTUAL_OFFWALL_TILE],
    displayRoles: [displayRole.HTML_TILE_OUT, displayRole.HTML_OFFWALL_TILE_OUT],
    deviceCondDesc: 'The tested device may not be a host tile',
    deviceCondition: device => !deviceIsHostTile(device),
    onlyPullOnce: false,
  },
];
// Part 2: Overflow and Collaboration with breakout rooms
audioMixPullFrom[deviceRole.CAMERA_MANAGER_MASTER] = [
  {
    deviceRoles: [deviceRole.MASTER, deviceRole.OVERFLOW_MASTER],
    displayRoles: [displayRole.MAIN_OUT],
    thisDeviceCondDesc: 'The pulling device must have a Presenter or Room cam with audio',
    thisDeviceCondition: device => hasPresenterOrRoomCameraWithAudio(device),
    onlyPullOnce: true,
  },
  {
    deviceRoles: [deviceRole.COLLABORATION_MASTER, deviceRole.COLLABORATION_OVERFLOW_MASTER],
    displayRoles: [displayRole.COLLABORATION_MASTER_MAIN_OUT],
    thisDeviceCondDesc: 'The pulling device must have a Presenter or Room cam with audio',
    thisDeviceCondition: device => hasPresenterOrRoomCameraWithAudio(device),
    onlyPullOnce: true,
  },
  {
    deviceRoles: [deviceRole.CONFIDENCE_OVERFLOW_MASTER, deviceRole.COLLABORATION_CONFIDENCE_MASTER],
    displayRoles: [displayRole.OVERFLOW_REMOTE],
    thisDeviceCondDesc: 'The pulling device must have a Presenter or Room cam with audio',
    thisDeviceCondition: device => hasPresenterOrRoomCameraWithAudio(device),
    onlyPullOnce: true,
  },
  {
    deviceRoles: [deviceRole.AUXILIARY],
    displayRoles: [displayRole.AUXILIARY_MAIN],
    thisDeviceCondDesc: 'The pulling device must have a Presenter or Room cam with audio',
    thisDeviceCondition: device => hasPresenterOrRoomCameraWithAudio(device),
    onlyPullOnce: false,
  },
];
audioMixPullFrom[deviceRole.CAMERA_MANAGER_SLAVE] = [
  {
    deviceRoles: [deviceRole.OVERFLOW_SLAVE, deviceRole.CONFIDENCE_OVERFLOW_SLAVE],
    displayRoles: [
      displayRole.SLAVE_COMBINED_FROM_MASTER,
      displayRole.SLAVE_CONTENT_FROM_MASTER,
      displayRole.SLAVE_TEACHER_FROM_MASTER,
      displayRole.OVERFLOW_REMOTE,
    ],
    thisDeviceCondDesc: 'The pulling device must have a Presenter or Room cam with audio',
    thisDeviceCondition: device => hasPresenterOrRoomCameraWithAudio(device),
    onlyPullOnce: true,
  },
];
audioMixPullFrom[deviceRole.COLLABORATION_OVERFLOW_MASTER] = [
  {
    deviceRoles: [deviceRole.COLLABORATION_CONFIDENCE_MASTER],
    displayRoles: [displayRole.OVERFLOW_REMOTE],
    thisDeviceCondDesc: 'The pulling device must have a Presenter or Room cam with audio',
    thisDeviceCondition: device => hasPresenterOrRoomCameraWithAudio(device),
    onlyPullOnce: true,
  },
];
audioMixPullFrom[deviceRole.OVERFLOW_MASTER] = [
  {
    deviceRoles: [deviceRole.CONFIDENCE_OVERFLOW_MASTER],
    displayRoles: [displayRole.OVERFLOW_REMOTE],
    thisDeviceCondDesc: 'The pulling device must have a Presenter or Room cam with audio',
    thisDeviceCondition: device => hasPresenterOrRoomCameraWithAudio(device),
    onlyPullOnce: true,
  },
];
audioMixPullFrom[deviceRole.OVERFLOW_SLAVE] = [
  {
    deviceRoles: [deviceRole.CONFIDENCE_OVERFLOW_SLAVE],
    displayRoles: [
      displayRole.SLAVE_COMBINED_FROM_MASTER,
      displayRole.SLAVE_CONTENT_FROM_MASTER,
      displayRole.SLAVE_TEACHER_FROM_MASTER,
      displayRole.OVERFLOW_REMOTE,
    ],
    thisDeviceCondDesc: 'The pulling device must have a Presenter or Room cam with audio',
    thisDeviceCondition: device => hasPresenterOrRoomCameraWithAudio(device),
    onlyPullOnce: true,
  },
];
audioMixPullFrom[deviceRole.BROADCAST_COMPOSITOR] = [
  {
    deviceRoles: [deviceRole.VIRTUAL_TILE, deviceRole.VIRTUAL_OFFWALL_TILE],
    displayRoles: [displayRole.HTML_TILE_OUT, displayRole.HTML_OFFWALL_TILE_OUT],
    deviceCondDesc: 'The tested device may not be a host tile',
    deviceCondition: device => !deviceIsHostTile(device),
    onlyPullOnce: false,
  },
];

const usecaseNeedsPulling = (candidatesToPullFrom, device, devices, thisDevice, logger) => {
  const result = candidatesToPullFrom.find(cand => cand.deviceRoles.includes(device.role)
    && (device.displayConfig.some(conf => cand.displayRoles.includes(conf.role)))
    && (cand.hasOwnProperty('thisDeviceCondition') ? cand.thisDeviceCondition(thisDevice) : true)
    && (cand.hasOwnProperty('deviceCondition') ? cand.deviceCondition(device) : true)
    && (cand.hasOwnProperty('devicesCondition') ? cand.devicesCondition(devices) : true));
  // If found, remove the matched device and display roles from the candidate
  // to avoid pulling audio more than once from this type of device
  if (result) {
    const displayConf = device.displayConfig.find(conf => result.displayRoles.includes(conf.role));
    if (logger) { logger.log('verbose', `Digital Audio Mix: device '${device.privateIp}' with device role '${device.role}' and display role, ${displayConf.role}] matches candidate ${JSON.stringify(result, null, 2)}`); }
    if (result.onlyPullOnce) {
      result.displayRoles = result.displayRoles.filter(role => role !== displayConf.role);
      if (logger) { logger.log('verbose', `Digital Audio Mix: Removed candidate from list as we can only pull its audio once, new list = ${JSON.stringify(candidatesToPullFrom, null, 2)}`); }
    }
  }
  return !!result;
};

// Function that tells whether or not we should pull audio from a certain device
const shouldStreamFromDevice = (candidatesToPullFrom, device, devices, thisDevice, logger) => {
  if (device.meetingId !== thisDevice.meetingId) {
    if (logger) { logger.log('verbose', `Digital Audio Mix: device '${device.privateIp}' is not in the lecture, not adding to list of devices to pull audio from`); }
    return false;
  }
  if (itsMe(device, thisDevice)) {
    if (logger) { logger.log('verbose', `Digital Audio Mix: device '${device.privateIp}' is 'this' (=device that is pulling) device, not adding to list of devices to pull audio from`); }
    return false;
  }
  if (!deviceIsConnected(device)) {
    if (logger) { logger.log('verbose', `Digital Audio Mix: device '${device.privateIp}' is offline, not adding to list of devices to pull audio from`); }
    return false;
  }
  if (!usecaseNeedsPulling(candidatesToPullFrom, device, devices, thisDevice, logger)) {
    if (logger) { logger.log('verbose', `Digital Audio Mix: device '${device.privateIp}' does not match criteria for pulling, not adding to list of devices to pull audio from`); }
    return false;
  }
  return true;
};

const generateListOfDevicesToPullAudioFrom = (devices, thisDevice, authenticationDisabled, logger) => {
  const candidates = [];
  if (!audioMixPullFrom.hasOwnProperty(thisDevice.role)) { return []; }
  audioMixPullFrom[thisDevice.role].forEach(ele => candidates.push({ ...ele }));
  if (logger) { logger.log('verbose', 'Digital Audio Mix: Checking following candidates ', JSON.stringify(candidates, null, 2)); }
  return devices
    .filter(d => shouldStreamFromDevice(candidates, d, devices, thisDevice, logger))
    .map((d) => {
      // Create a name so that streams can be recognized in pavucontrol
      // Prefix shows type of stream (content, tile or confidence monitor)
      // and last two numbers are last part of IP address of device from where audio is pulled.
      let prefix = deviceIsTile(d) ? 'TA' : 'CMA'; // TA: Tile Audio - CMA: Confidence Monitor Audio
      prefix = deviceIsMaster(d) ? 'CA' : prefix; // CA: Content Audio
      const addressArray = d.privateIp.split('.');
      const sinkname = `${prefix}_${addressArray[2]}_${addressArray[3]}`;
      const authenticationActive = !authenticationDisabled && d.rtspAuthentication && d.rtspAuthentication.username;
      return {
        _id: d._id,
        role: d.role,
        privateIp: d.privateIp,
        hostname: d.hostname,
        meetingId: d.meetingId,
        username: authenticationActive ? d.rtspAuthentication.username : undefined,
        password: authenticationActive ? d.rtspAuthentication.password : undefined,
        sinkname,
      };
    });
};

class StreamingBridge extends ServiceBridge {
  constructor(options) {
    super(options);
    this.cache = new MeteorCacheManager(this, ['devices', 'compositions', 'sinks']);
    this.streams = {};
    this.incomingAudioStreams = {};
    this._paSrcDevice = 'nullsink.monitor';
    this._syncing = false;
    this._requestNewSync = false;
    this._forceRestart = false;
    this._onlyIncomingAudioStreams = true;
    this._siblingSubscriptionEnabled = false;
  }

  get supportedSinkTypes() {
    return [sinkType.SCREEN, sinkType.MJPEG_STREAM, sinkType.RTSP_STREAM, sinkType.RTSP_AUDIO_STREAM, sinkType.RECORD_STREAM];
  }

  get _rtspAuthenticationDisabled() {
    return this._meteorClient._ddpClient.collections.institutes
      && Object.values(this._meteorClient._ddpClient.collections.institutes)
        .some(institute => (institute._id === this._device.instituteId
          && institute.enabledFeatures.includes('disable_rtsp_authentication')));
  }

  get _digitalAudioMixEnabled() {
    return !!this._device.audio.digitalAudioMix && !forceDisableDAM;
  }

  _onThisDeviceChanged(device, oldFields, clearedFields, newFields) {
    let restart = false;
    let incomingAudioStreamsOnly = false;
    let forceRestart = false;

    if (newFields.hasOwnProperty('audioSinkConfig')
    || newFields.hasOwnProperty('audioSourceConfig')
    || newFields.hasOwnProperty('usbCameraConfig') // audio input can be coupled to camera
    || newFields.hasOwnProperty('role') // some parts of audio routing depend on device role
    || (newFields.hasOwnProperty('streamRemoteAudio'))) {
      this._logger.log('verbose', 'Audio source configuration changed. Checking to recreate all non-camera RTSP streams if needed.');
      restart = true;
    }

    if ((newFields.hasOwnProperty('audio'))) {
      if (newFields.audio.calibration !== oldFields.audio?.calibration) {
        this._log('verbose', `Audio calibration status changed to ${newFields.audio.calibration}. About to recreate all non-camera RTSP streams.`);
        restart = true;
        forceRestart = true;
      }
      if (newFields.audio.digitalAudioMix !== oldFields.audio?.digitalAudioMix) {
        this._log('verbose', `Digital Audio Mix status changed to ${newFields.audio.digitalAudioMix ? 'enabled' : 'disabled'}. About to recreate all non-camera RTSP streams.`);
        restart = true;
        forceRestart = true;
      }
    }

    if (newFields.meetingId !== oldFields.meetingId) {
      this._log('verbose', `${newFields.meetingId ? 'New session started. About to recreate all incoming audio RTSP streams.'
        : 'Session stopped. About to cleanup all incoming audio RTSP streams.'}`);
      this._checkSiblingSubscription(); // Make sure changes to DAM setting are applied without rebooting
      restart = true;
      incomingAudioStreamsOnly = true; // RTSP servers are handled by '_onSinkAdded' and '_onSinkRemoved'
      forceRestart = true;
    }

    if (restart) {
      return this._requestRecreateRTSPStreams(incomingAudioStreamsOnly, forceRestart);
    }
  }

  _onDeviceAdded(device) {
    if (this._digitalAudioMixEnabled
      && !itsMe(device, this._device)
      && deviceIsConnected(device)) {
      return this._requestRecreateRTSPStreams(true);
    }
  }

  _onDeviceChanged(device, oldFields, clearedFields, newFields) {
    if (this._digitalAudioMixEnabled
      && deviceIsConnected(device)
      && (newFields.hasOwnProperty('role')
      || newFields.hasOwnProperty('status')
      || newFields.hasOwnProperty('meetingId')
      || newFields.hasOwnProperty('rtspAuthentication'))) {
      return this._requestRecreateRTSPStreams(true);
    }
  }

  // For an arbitrary sink of supported type (sinkType.SCREEN, sinkType.RTSP_STREAM, sinkType.RTSP_AUDIO_STREAM, sinkType.RECORD_STREAM)
  // find all the related sinks that are linked (and hence supposed to render) to the same composition:
  // - is there a screen sink defined for this composition (means we render the composition on screen)
  // - is there an RTSP sink defined for this composition (means this composition has to be available as an RTSP stream)
  // - is there an RTSP audio sink defined for this composition (means the audio of this composition has to be available as an RTSP stream)
  // - is there a recording sink defined for this compositions (means this composition is to be recorded (uses RTSP stream!))
  _findCorrespondingRTSPStreamForSink(sink) {
    let screenSink;
    let rtspSink;
    let rtspAudioSink;
    let recordSink;
    const composition = this.cache.findCompositionOnSink(sink);
    if (composition) {
      screenSink = this.cache.findDefaultSinksForComposition(composition).find(s => s.type === sinkType.SCREEN);
      rtspSink = this.cache.findDefaultSinksForComposition(composition).find(s => s.type === sinkType.RTSP_STREAM);
      rtspAudioSink = this.cache.findDefaultSinksForComposition(composition).find(s => s.type === sinkType.RTSP_AUDIO_STREAM);
      recordSink = this.cache.findDefaultSinksForComposition(composition).find(s => s.type === sinkType.RECORD_STREAM);
    }
    return {
      composition,
      screenSink,
      rtspSink,
      rtspAudioSink,
      recordSink,
    };
  }

  _onSinkAdded(sink) {
    this._log('verbose', `onSinkAdded: Adding sink with id ${sink._id} and type ${sink.type} ...`);
    if (!this.isSupportedSink(sink)) {
      this._log('verbose', `Sink with id ${sink._id} and type ${sink.type} is not a streaming sink, skipping.`);
      return;
    }

    // First find all sinks that render the same composition as the given sink.
    const {
      composition,
      screenSink,
      rtspSink,
      recordSink,
    } = this._findCorrespondingRTSPStreamForSink(sink);

    // If there is no composition, skip...
    if (!composition) {
      this._log('verbose', `No composition found for sink with id ${sink._id} and type ${sink.type}, not adding stream server for this sink.`);
      return;
    }

    const compositionIsBrowser = composition.type === compositionType.BROWSER;
    const bothScreenAndRTSPSinksAreDefined = (screenSink && rtspSink);
    const sinkIsScreen = sink.type === sinkType.SCREEN;
    const sinkIsMJPEGStream = sink.type === sinkType.MJPEG_STREAM;
    const sinkIsRtspAudioStream = sink.type === sinkType.RTSP_AUDIO_STREAM;

    // If the added sink is a screen sink, no RTSP stream has to be created unless the composition type is browser
    if (sinkIsScreen && !compositionIsBrowser) {
      this._log('verbose', `Screen sink with id ${sink._id} does not refer to a browser composition, not adding stream server for this sink.`);
      return;
    }

    // Only start an RTSP stream for a browser composition if both screen sink and RTSP sink and defined.
    // RTSP sink for browser comps have to screenscrape (whereas for cinergy it can use the shared memory)
    // That means we need the matching screen sink to know the area to scrape. However, it can happen that
    // the screen sink is added later than the rtsp one, which this block of code deals with.
    if (compositionIsBrowser && !bothScreenAndRTSPSinksAreDefined && !sinkIsRtspAudioStream) {
      this._log('verbose', `Sink with id ${sink._id} and type ${sink.type} refers to a browser composition but the corresponding rtsp or screen sink is not defined, not adding stream server for this sink.`);
      return;
    }

    // There are 3 kinds of stream: MJPEG or RTSP or RTSP_AUDIO.
    // For everything that is not MJPEG or RTSP_AUDIO, pass the RTSP sink to create the stream.
    if (!sinkIsMJPEGStream && !sinkIsRtspAudioStream) {
      sink = rtspSink;
    }
    if (sink) {
      // Creates MJPEG stream or RTSP (audio) stream. Won't do anything if the stream already exists
      this._log('verbose', `About to create stream server for sink with id ${sink._id} and type ${sink.type} and port ${sink.port}.`);
      this._createStreamFor(sink, composition, screenSink);
    }
    // If an RTSP stream exists/has been created and it must be recorded, let it register itself for recording.
    if (recordSink
      && rtspSink
      && this.streams[rtspSink._id]
      && this.streams[rtspSink._id].isActive) {
      this._log('verbose', `About to register stream for RTSP sink with id ${rtspSink._id} for recording.`);
      this.streams[rtspSink._id].registerForRecording(recordSink);
    }
  }

  _onSinkRemoved(sink) {
    const streamId = streamIdForSink(sink);
    if (this.streams.hasOwnProperty(streamId)) {
      this._log('info', `Destroying stream server with id ${streamId} for sink of type ${sink.type} and sink id ${sink._id}`);
      const sinkToDestroy = this.streams[streamId];
      delete this.streams[streamId]; // Avoid any further usage eg. in _onSinkAdded (destroy is async and takes some time)
      return sinkToDestroy.destroy();
    }
    return Promise.resolve();
  }

  _onSinkChanged(sink, oldFields, clearedFields, newFields) {
    if (this.isSupportedSink(sink)
      && this._anyChanged(['activeCompositionId', 'port', 'url', 'width', 'height', 'framerate'], newFields)) {
      return this._onSinkRemoved(sink).then(() => this._onSinkAdded(sink));
    }
  }

  async _requestRecreateRTSPStreams(onlyIncomingAudioStreams, forceRestart = false) {
    if (!this.isActive) return;
    if (this._syncing) {
      this._log('info', `Request recreate all non-camera RTSP servers and streams (restrict to incoming audio streams only = ${onlyIncomingAudioStreams}, forceRestart = ${forceRestart} ): Sync in progress => schedule extra sync round`);
      this._requestNewSync = true;
      this._forceRestart = this._forceRestart || forceRestart;
      this._onlyIncomingAudioStreams = this._onlyIncomingAudioStreams && onlyIncomingAudioStreams;
    } else {
      this._log('info', `Request recreate all non-camera RTSP servers and streams (restrict to incoming audio streams only = ${onlyIncomingAudioStreams}, forceRestart = ${forceRestart} ): Sync not in progress => start sync`);
      this._requestNewSync = false;
      this._syncing = true;
      await this._recreateRTSPStreams(onlyIncomingAudioStreams, forceRestart);
      this._syncing = false;
      if (this._requestNewSync) {
        await this._requestRecreateRTSPStreams(this._onlyIncomingAudioStreams, this._forceRestart);
      }
      this._onlyIncomingAudioStreams = true;
      this._forceRestart = false;
    }
  }

  async _recreateRTSPStreams(onlyIncomingAudioStreams, forceRestart) {
    this._log('verbose', `About to recreate all non-camera RTSP servers and streams (restrict to incoming audio streams only = ${onlyIncomingAudioStreams}, forceRestart = ${forceRestart} )`);

    // Part 1: Launch RTSP servers (can be for audio+video or audio only)
    if (!onlyIncomingAudioStreams) {
      this._log('verbose', 'About to recreate all non-camera RTSP (audio video) stream servers if needed...');
      const oldSource = this._paSrcDevice;
      this._paSrcDevice = findRtspContentStreamSource(this._device);
      if ((oldSource !== this._paSrcDevice) || forceRestart) {
        this._logger.log('verbose', `${forceRestart ? 'Forcefully' : 'Audio configuration changed,'} recreating all non-camera RTSP (audio video) stream servers.`);
        await Promise.all(this.cache.findSinkByTypes([sinkType.RTSP_STREAM, sinkType.RTSP_AUDIO_STREAM]).map(async (sink) => {
          if (this._device._id === sink.device._id) {
            this._logger.log('verbose', `About to recreate RTSP server for sink of type ${sink.type} with id ${sink._id}`);
            await this._onSinkRemoved(sink);
            if (this.isActive) {
              this._onSinkAdded(sink);
            }
          }
        }));
      } else {
        this._log('verbose', 'No need to recreate any non-camera RTSP (audio video) stream servers');
      }
    }

    // Part 2: RTSP receivers: pull (=receive) audio from other devices to create audio mixes without the need for external mixers/processors.
    let devicesToPullAudioFrom = [];

    if (!!this._device.meetingId && audioMixPullFrom.hasOwnProperty(this._device.role) && this._digitalAudioMixEnabled) {
      this._log('verbose', `Digital Audio Mix: About to recreate audio playback streams, forceRestart = ${forceRestart}  ...`);
      this._log('verbose', `Digital Audio Mix: About to check all devices for this session: ${JSON.stringify(this.cache.$arrays.devices.map(d => ({ ip: d.privateIp, role: d.role, status: d.status })), null, 2)}`);
      // Create a list of all the devices that this device should pull audio from
      devicesToPullAudioFrom = generateListOfDevicesToPullAudioFrom(this.cache.$arrays.devices, this._device, this._rtspAuthenticationDisabled, this._logger);
      this._log('verbose', `Digital Audio Mix: Devices to pull audio from: ${JSON.stringify(devicesToPullAudioFrom, null, 2)}`);
      await this._adaptRtspAudioPlaybackStreamsIfNeeded(this._devicesToPullAudioFrom, devicesToPullAudioFrom, forceRestart);
      this._devicesToPullAudioFrom = devicesToPullAudioFrom;
    } else {
      this._log('verbose', `Digital Audio Mix: ${this._digitalAudioMixEnabled ? 'Current device configuration does not require pulling any audio streams' : 'Mixing functionality switched off'}. About to cleanup streams (if any).`);
      await this._cleanupRtspAudioPlaybackStreams();
      this._devicesToPullAudioFrom = [];
    }
  }

  async _adaptRtspAudioPlaybackStreamsIfNeeded(oldList, newList, forceRestart) {
    this._log('verbose', `Digital Audio Mix: Adapt audio playback streams: forceRestart = ${forceRestart}. First cleaning up unnecessary streams.`);
    if (forceRestart) { // Cleanup ALL audio streams.
      await this._cleanupRtspAudioPlaybackStreams();
    } else {
      // Cleanup audio streams pulled from other devices that are not needed anymore.
      // _.isEqual deep compares objects in lists, so things like rtsp username and password are compared.
      const streamIdsToRemove = _.differenceWith(oldList, newList, _.isEqual).map(device => rtspAudioPlaybackStreamId(device, this._device));
      await this._cleanupRtspAudioPlaybackStreams(streamIdsToRemove);
    }

    // Add streams to pull audio from other devices where needed
    // _.isEqual deep compares objects in lists, so things like rtsp username and password are compared.
    // (Re)launch ALL streams in case forceRestart is true.
    const streamsToAdd = forceRestart ? newList : _.differenceWith(newList, oldList, _.isEqual);
    if (streamsToAdd.length) {
      streamsToAdd.forEach(((device) => {
        const streamId = rtspAudioPlaybackStreamId(device, this._device);
        this._log('verbose', `Digital Audio Mix: Creating stream with id = ${streamId} to pull audio from device ${device._id} with IP address ${device.privateIp}`);
        const stream = new IncomingAudioStream({
          path: `rtsp://${device.privateIp}:8550/audio`,
          pulseSinkName: device.sinkname,
          username: device.username,
          password: device.password,
        });
        this.incomingAudioStreams[streamId] = stream;
        stream.once('destroyed', () => {
          delete this.incomingAudioStreams[streamId];
          this._log('verbose', `Digital Audio Mix: Successfully cleaned up stream with id = ${streamId}`);
        });
        stream.start();
      }));
    } else {
      this._log('verbose', 'Digital Audio Mix: No streams to be added');
    }
  }

  async _cleanupRtspAudioPlaybackStreams(streamIdsToRemove = Object.keys(this.incomingAudioStreams)) {
    if (streamIdsToRemove.length) {
      this._log('verbose', `Digital Audio Mix: About to cleanup streams with following stream id's: ${JSON.stringify(streamIdsToRemove, null, 2)}`);
      const promises = streamIdsToRemove.map(async (streamId) => {
        this._log('verbose', `Digital Audio Mix: About to destroy stream with id = ${streamId}`);
        await this.incomingAudioStreams[streamId].destroy();
        delete this.incomingAudioStreams[streamId];
      });
      await Promise.all(promises);
    } else {
      this._log('verbose', 'Digital Audio Mix: No streams to be cleaned up');
    }
  }

  /**
   * Creates a stream for a certain sink
   * @param {*} sink
   */
  _createStreamFor(sink, composition, screenSink) {
    const streamId = streamIdForSink(sink);
    this._log('info', `CreateStreamFor ${sink.type}, list has following ids: ${JSON.stringify(Object.keys(this.streams), null, 2)}`);
    if (this.streams.hasOwnProperty(streamId)) {
      this._log('verbose', `Stream server with id ${streamId} for sink with id ${sink._id} and type ${sink.type} already exists, skipping`);
      return;
    }

    let stream;

    if (sink.type === sinkType.MJPEG_STREAM) {
      if (composition.type === compositionType.CINERGY) {
        this._log('verbose', `Starting MJPEG server for sink ${sink._id}`);

        stream = new MjpegStream({
          path: `/tmp/cinergy/${cinergyUtils.compositorIdForComposition(composition)}`,
          port: 8085 + composition.index,
          width: sink.width,
          height: sink.height,
          framerate: sink.framerate,
        });
      } else {
        throw new Error('Mjpeg streams are only supported for cinergy shared memory');
      }
    }

    // Make sure we use the right audio source for the rtsp server
    this._paSrcDevice = findRtspContentStreamSource(this._device);

    if (sink.type === sinkType.RTSP_STREAM) {
      this._log('verbose', `Starting RTSP server for sink ${sink._id}`);

      const sinkUrl = new URL(sink.url);

      const sharedOptions = {
        url: sinkUrl.pathname.replace(/^\/|\/$/g, ''),
        port: sink.port,
        width: sink.width,
        height: sink.height,
        framerate: sink.framerate,
        paSrcDevice: this._paSrcDevice,
        username: sinkUrl.username,
        password: sinkUrl.password,
      };

      if (composition.type === compositionType.CINERGY) {
        // RTSP server gets video from shm in Cinergy case (eg. Master, RTSP Camera Manager)
        stream = new RtspShmStream({
          path: `/tmp/cinergy/${cinergyUtils.compositorIdForComposition(composition)}`,
          ...sharedOptions,
        });
      }

      if (composition.type === compositionType.BROWSER) {
        // RTSP server gets video from screenscraping in case of Browser (eg. Virtual Camera Manager)
        stream = new RtspScreenStream({
          source: screenSink,
          ...sharedOptions,
          // If device has a camera with audio, use recording mix (virtual camera manager case)
          paSrcDevice: hasPresenterOrRoomCameraWithAudio(this._device) ? this._paSrcDevice : null,
        });
      }
    }

    if (sink.type === sinkType.RTSP_AUDIO_STREAM) {
      if (this._digitalAudioMixEnabled) {
        this._log('verbose', `Digital Audio Mix: Starting audio RTSP server for device ${sink.device._id} (sink id = ${sink._id}, meeting id = ${sink.meetingId})`);

        const sinkUrl = new URL(sink.url);

        stream = new RtspAudioStream({
          url: sinkUrl.pathname.replace(/^\/|\/$/g, ''),
          port: sink.port,
          paSrcDevice: 'main_mix.monitor', // stream out what's sent to the speakers
          // params: '...',
          username: sinkUrl.username,
          password: sinkUrl.password,
        });
      } else {
        this._log('verbose', `Digital Audio Mix feature is disabled: Not creating audio RTSP server for device ${sink.device._id} (sink id = ${sink._id}, meeting id = ${sink.meetingId})`);
      }
    }

    if (stream) {
      this.streams[streamId] = stream;
      stream.once('destroyed', () => {
        delete this.streams[streamId];
        this._log('verbose', `Successfully cleaned up stream server with id ${streamId}`);
      });
      stream.start();
    }
  }

  _checkSiblingSubscription() {
    if (this._digitalAudioMixEnabled) {
      this._enableSiblingSubscription();
    } else {
      this._disableSiblingSubscription();
    }
  }

  _enableSiblingSubscription() {
    if (!this._siblingSubscriptionEnabled && this._digitalAudioMixEnabled) {
      this._log('info', 'Enabling sibling device info subscription');
      this._siblingSubscriptionEnabled = true;
      this._devicesToPullAudioFrom = [];
      this._meteorClient.subscribe('barco.devices.siblings', [{
        deviceId: this._device._id,
      }], this, '_tileSubscriptionId');
    }
  }

  _disableSiblingSubscription() {
    if (this._siblingSubscriptionEnabled) {
      this._log('info', 'Disabling sibling device info subscription');
      this._siblingSubscriptionEnabled = false;
      this._devicesToPullAudioFrom = [];
      try {
        this._meteorClient.unsubscribe(this._tileSubscriptionId);
      } catch (error) {
        // We might already be disconnected from the cloud, so ignore if this throws
      }
      delete this._tileSubscriptionId;
    }
  }

  _onStart() {
    this.streams = {};
    this.incomingAudioStreams = {};
    this._paSrcDevice = findRtspContentStreamSource(this._device);
    this._enableSiblingSubscription();
    return super._onStart();
  }

  async _onStop() {
    this._disableSiblingSubscription();
    await super._onStop();
    await this._cleanupRtspAudioPlaybackStreams();
    await Promise.all(Object.values(this.streams).map(stream => stream.destroy()));
  }

  _onDestroy() {
    delete this.streams;
    delete this.incomingAudioStreams;
    delete this._paSrcDevice;
    return super._onDestroy();
  }
}

export default StreamingBridge;

export {
  generateListOfDevicesToPullAudioFrom,
};
