import * as fetch from 'node-fetch';
import { exec, execSync } from 'child_process';
import { promisify } from 'util';
import * as Path from 'path';

import Bridge from '../bridge';
import { commands } from '../../../services/ddp/constants';
import { SCRIPTS_DIR } from '../../../files';

const port = 8091;

const teleportStatus = {
  UNKNOWN: 0,
  DISABLED: 1,
  ENABLED: 2,
};

class TeleportBridge extends Bridge {
  get supportedRPCCommands() {
    return [commands.SET_TELEPORT, commands.CLEAR_TELEPORT_CACHE];
  }

  /**
   * Act on incoming client RPCs
   * RPC's are useful for one-time actions that are not part of the application state.
   *
   * @param rpc
   * @returns {*}
   * @private
   */
  async _onRpcAdded(rpc) {
    this._log('verbose', `Handling rpc ${JSON.stringify(rpc, null, ' ')}`, { serviceLog: true });
    switch (rpc.command) {
    case commands.CLEAR_TELEPORT_CACHE:
      this.clearCache();
      break;
    case commands.SET_TELEPORT:
      await this.setStatus(rpc.params.status, rpc.params.token);
      // Cleanup old log archives and create a new one in case teleport is enabled
      this._cleanupLogArchives();
      if (rpc.params.status === teleportStatus.ENABLED) {
        await this._createLogArchive();
      }
      break;
    default:
    }
  }

  clearCache() {
    if (this._device.teleportStatus === teleportStatus.DISABLED) {
      this._log('info', 'Clearing teleport cache');
      try {
        execSync('sudo rm -rf /var/lib/teleport/*');
        this._log('verbose', 'Cleared teleport cache');
      } catch (error) {
        this._log('error', 'Error while clearing teleport cache', error);
      }
    } else {
      this._log('info', 'Cannot clear cache while service is enabled');
    }
  }

  async configure(token) {
    const res = await fetch(TeleportBridge.url('configure'), {
      method: 'POST',
      headers: {
        Authorization: `Bearer ${token}`,
      },
    });
    return res;
  }

  async setStatus(status, token) {
    switch (status) {
    case teleportStatus.ENABLED:
      this._log('info', 'Enable teleport service', { serviceLog: true });
      // XXX: do this until we're on teleport 3.1
      // Currently we create a new cluster name every time
      // therefore we need to clear the cache...
      // https://github.com/gravitational/teleport/issues/1941
      // https://github.com/gravitational/teleport/issues/2179
      this.clearCache();
      await this._setStatus('enable', token);
      break;
    case teleportStatus.DISABLED:
      this._log('info', 'Disable teleport service', { serviceLog: true });
      await this._setStatus('disable', token);
      break;
    default:
      this._log('error', `Received unknown status: ${status}`);
    }
    await this._checkAndUpdateStatus();
  }

  async _setStatus(path, token) {
    try {
      const response = await fetch(TeleportBridge.url(path), {
        method: 'POST',
        headers: {
          Authorization: `Bearer ${token}`,
        },
      });

      if (response.status === 200) {
        this._log('verbose', 'Status succesfully updated');
      } else {
        throw new Error(`Failed to change status: ${response.status}`);
      }
    } catch (error) {
      this._log('error', 'error', error);
    }
  }

  static url(path) {
    return `http://localhost:${port}/${path}`;
  }

  async _checkAndUpdateStatus() {
    try {
      const response = await fetch(TeleportBridge.url('public/check'));

      if (response.status === 200) {
        const isActive = await response.json();
        const status = isActive ? teleportStatus.ENABLED : teleportStatus.DISABLED;
        await this._updateStatus(status);
      } else {
        throw new Error(`Unexpected code from teleport proxy: ${response.status}`);
      }
    } catch (error) {
      await this._updateStatus(teleportStatus.UNKNOWN);
      this._log('error', 'error', error);
    }
  }

  async _updateStatus(status) {
    if (status !== this._device.teleportStatus) {
      this._log('info', `Updating teleport status from ${this._device.teleportStatus} to ${status}`);
      this._device.teleportStatus = status;
      await this._meteorClient.updateDevice({
        _id: this._device._id,
        teleportStatus: this._device.teleportStatus,
      });
    } else {
      this._log('silly', 'Teleport status does not need updating...');
    }
  }

  _startCheckInterval() {
    this._log('verbose', 'Installing check interval');
    this._checkInterval = this._checkInterval || setInterval(this._checkAndUpdateStatus.bind(this), 5000);
  }

  _stopCheckInterval() {
    this._log('verbose', 'Uninstalling check interval');
    if (this._checkInterval) {
      clearInterval(this._checkInterval);
      delete this._checkInterval;
    }
  }

  async _createLogArchive() {
    this._log('verbose', 'Generating log archive in /tmp folder');
    try {
      await promisify(exec)(`sudo ${Path.join(SCRIPTS_DIR, 'generateLogsArchive.sh')}; sudo chmod 666 /tmp/*.tar.gz`);
    } catch (error) {
      this._log('error', `Error generating logs archive. Error = ${error.message}`);
    }
  }

  _cleanupLogArchives() {
    this._log('verbose', 'Removing old log archives (if any)');
    try {
      execSync('sudo rm -f /tmp/*.tar.gz');
    } catch (error) {
      this._log('error', `Error while deleting old log archives, error = ${error.message}`);
    }
  }

  async _onStart() {
    await this._checkAndUpdateStatus();
    await this._startCheckInterval();
    this._startCheckInterval();
    return super._onStart();
  }

  async _onStop() {
    await this._stopCheckInterval();
    return super._onStop();
  }

  async _onDestroy() {
    delete this.request;
    await super._onDestroy();
  }
}

export default TeleportBridge;
