import * as fs from 'fs-extra';
import { execSync } from 'child_process';
import TimeBridge, { defaultServers } from './bridge';

const CONFIG_FILE = '/etc/ntp.conf';
const serverRegex = /^\s*((?:server|pool)\s+.+)/i;

/**
 * The time brige sets the ntp servers and timezone
 */
class NTPTimeBridge extends TimeBridge {
  _getRequiredServers() {
    let servers = super._getRequiredServers();
    if (servers === defaultServers) {
      servers = defaultServers.map(server => `pool ${server} iburst`);
    } else {
      servers = servers.map(server => `server ${server}`);
    }
    return servers;
  }

  async _getCurrentServers() {
    const currentConfig = (await fs.readFile(CONFIG_FILE, 'utf-8')).split('\n');

    const servers = currentConfig
      .map(line => serverRegex.exec(line))
      .filter(match => !!match && match.length === 2)
      .map(match => match[1]);

    return servers;
  }

  async _setNtpServers(servers) {
    this._log('info', 'Setting timezone servers', servers);

    const config = `
# /etc/ntp.conf, configuration for ntpd; see ntp.conf(5) for help

driftfile /var/lib/ntp/ntp.drift

# Enable this if you want statistics to be logged.
#statsdir /var/log/ntpstats/

statistics loopstats peerstats clockstats
filegen loopstats file loopstats type day enable
filegen peerstats file peerstats type day enable
filegen clockstats file clockstats type day enable


# You do need to talk to an NTP server or two (or three).
#server ntp.your-provider.example

# pool.ntp.org maps to about 1000 low-stratum NTP servers.  Your server will
# pick a different set every time it starts up.  Please consider joining the
# pool: <http://www.pool.ntp.org/join.html>
${servers.join('\n')}


# Access control configuration; see /usr/share/doc/ntp-doc/html/accopt.html for
# details.  The web page <http://support.ntp.org/bin/view/Support/AccessRestrictions>
# might also be helpful.
#
# Note that "restrict" applies to both servers and clients, so a configuration
# that might be intended to block requests from certain clients could also end
# up blocking replies from your own upstream servers.

# By default, exchange time with everybody, but don't allow configuration.
restrict -4 default kod notrap nomodify nopeer noquery limited
restrict -6 default kod notrap nomodify nopeer noquery limited

# Local users may interrogate the ntp server more closely.
restrict 127.0.0.1
restrict ::1pool

# Needed for adding pool entries
restrict source notrap nomodify noquery

# Clients from this (example!) subnet have unlimited access, but only if
# cryptographically authenticated.
#restrict 192.168.123.0 mask 255.255.255.0 notrust


# If you want to provide time to your local subnet, change the next line.
# (Again, the address is an example only.)
#broadcast 192.168.123.255

# If you want to listen to time broadcasts on your local subnet, de-comment the
# next lines.  Please do this only if you trust everybody on the network!
#disable auth
#broadcastclient
    `;
    await fs.writeFile(CONFIG_FILE, config);
    await this._restartNtpDaemon();
  }

  _restartNtpDaemon() {
    this._log('verbose', 'Restarting NTP daemon');
    try {
      execSync('sudo service ntp restart');
    } catch (error) {
      this._log('error', 'Could not restart NTP daemon', error);
    }
  }
}

export default NTPTimeBridge;
