import * as fs from 'fs-extra';
import { execSync } from 'child_process';

import TimeBridge from './bridge';

const CONFIG_FILE = '/etc/systemd/timesyncd.conf.d/edu.conf';

/**
 * The time brige sets the ntp servers and timezone
 */
class TimesyncdTimeBridge extends TimeBridge {
  _isSynchronized() {
    try {
      this._log('info', 'checking timesyncd status');
      const status = execSync('timedatectl status', { encoding: 'utf-8' });
      this._log('verbose', 'timesyncd status:', status);
      // Different messages depending on timedatectl version
      const isSynchronized = status.includes('System clock synchronized: yes') || status.includes('NTP synchronized: yes');
      if (isSynchronized) {
        this._log('debug', 'Time is synchronized with server');
      } else {
        this._log('debug', 'Time is not synchronized with server');
      }
      return isSynchronized;
    } catch (error) {
      this._log('error', 'error while checking timesyncd status', error);
    }
  }

  _ensureNtpIsActive() {
    this._log('verbose', 'Making sure Timesync daemon runs');
    try {
      execSync('sudo timedatectl set-ntp on');
    } catch (error) {
      this._log('error', 'Could not start Timesync daemon', error);
    }
  }

  async _getCurrentServers() {
    const currentConfig = (await fs.readFile(CONFIG_FILE, 'utf-8')).split('\n').find(line => line.startsWith('NTP='));
    if (currentConfig) {
      return currentConfig.substr(4).split(' ');
    }
    return [];
  }

  async _setNtpServers(servers) {
    this._log('info', 'Setting timezone servers', servers);

    const config = `
[Time]
NTP=${servers.join(' ')}
    `;

    await fs.writeFile(CONFIG_FILE, config);
    await this._restartNtpClient();
  }

  _restartNtpClient() {
    this._log('verbose', 'Restarting Timesync daemon');
    try {
      execSync('sudo timedatectl set-ntp off');
      execSync('sudo timedatectl set-ntp on');
    } catch (error) {
      this._log('error', 'Could not restart Timesync daemon', error);
    }
    this._syncTimeStatus();
  }
}

export default TimesyncdTimeBridge;
