import * as fs from 'fs-extra';
import { execSync } from 'child_process';

import platform from '../../../utils/platform';
import Bridge from '../bridge';
import IntervalRunner from '../../../services/intervalRunner';
import { commands } from '../../../services/ddp/constants';
import { rebootSystem } from '../../../utils/utils';

const TIMEZONE_FILE = '/etc/timezone';

export const defaultServers = [
  '0.debian.pool.ntp.org',
  '1.debian.pool.ntp.org',
  '2.debian.pool.ntp.org',
  '3.debian.pool.ntp.org',
];

const validIpRegex = /^(([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])\.){3}([0-9]|[1-9][0-9]|1[0-9]{2}|2[0-4][0-9]|25[0-5])$/;
const validHostnameRegex = /^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9-]*[a-zA-Z0-9])\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9-]*[A-Za-z0-9])$/;
function isValidServer(server) {
  return validIpRegex.test(server) || validHostnameRegex.test(server);
}

const TIMECHECK_INTERVAL_IN_S = 60 * 60; // 1 hour in seconds

/**
 * The time brige sets the ntp servers and timezone
 */
class TimeBridge extends Bridge {
  constructor(...args) {
    super(...args);
    this._syncTimeStatus = this._syncTimeStatus.bind(this);
    this._intervalRunner = new IntervalRunner(this._syncTimeStatus, (1 / TIMECHECK_INTERVAL_IN_S));
  }

  get supportedRPCCommands() {
    return [commands.RESET_HW_CLOCK];
  }

  /**
   * Act on incoming client RPCs
   * RPC's are useful for one-time actions that are not part of the application state.
   *
   * @param rpc
   * @returns {*}
   * @private
   */
  _onRpcAdded(rpc) {
    this._log('verbose', `Handling rpc ${JSON.stringify(rpc, null, ' ')}`, { serviceLog: true });
    switch (rpc.command) {
    case commands.RESET_HW_CLOCK:
      setTimeout(() => this._resetHWClock(rpc.params.date), 5000);
      break;
    default:
    }
  }

  async _onStart() {
    if (!this._device.time) {
      // compatability with older clouds
      this._log('info', 'No time settings on device, skipping... (probably older cloud version)');
      return;
    }
    this._ensureNtpIsActive();
    await this._ensureNtpServers();
    await this._ensureTimeZone();
    this._intervalRunner.start();
    this._syncTimeStatus();
    await super._onStart();
  }

  _onStop() {
    this._intervalRunner.stop();
    return super._onStop();
  }

  async _onThisDeviceChanged(device, oldFields, clearedFields, newFields) {
    if (newFields.hasOwnProperty('time')) {
      this._log('verbose', 'Time settings have been changed for device', { serviceLog: true });

      await this._ensureNtpServers();
      await this._ensureTimeZone();
    }
  }

  _resetHWClock(date) {
    // Since we're going to reboot, we don't bother syncing the system clock
    // only the hardware clock
    platform.setHWClock(date, this._log.bind(this));
    this._log('info', 'Hardware clock set, rebooting device', { serviceLog: true });
    try {
      // Since setting the hardware and/or system clock does not affect running timers
      // and this action is only to be performed in very rare situations
      // We don't deal with changing all running timers and just reboot
      rebootSystem();
    } catch (error) {
      this._log('error', 'Could not reboot', error);
    }
  }

  _syncTimeStatus() {
    if (!this._device.meetingId) { // Only when no lecture ongoing
      const isSynchronized = this._isSynchronized();
      if (this._device.hasOwnProperty('timeSynchronized')
        && this._device.timeSynchronized !== isSynchronized) {
        this._log('info', 'Updating ntp synchronized status');
        this._meteorClient.updateTimeStatus({
          deviceId: this._device._id,
          synchronized: isSynchronized,
        });
      }
    }
  }

  _getRequiredServers() {
    const servers = this._device.time.ntpServers;
    const validServers = servers
      .map(server => server.trim())
      .filter(isValidServer);

    if (validServers.length !== servers.length) {
      this._log('error', 'Not all servers are valid', servers, validServers);
    }

    if (!validServers.length) {
      this._log('warn', 'No valid servers found, falling back to default servers');
      return defaultServers;
    }

    return validServers;
  }

  _getCurrentServers() {
    throw new Error('not implemented');
  }

  _isSynchronized() {
    throw new Error('not implemented');
  }

  _ensureNtpIsActive() {
    throw new Error('not implemented');
  }

  async _setNtpServers() {
    throw new Error('not implemented');
  }

  async _ensureNtpServers() {
    if (!this._device.time) {
      // compatability with older clouds
      return;
    }
    this._log('info', 'Checking NTP servers');
    const requiredServers = this._getRequiredServers();
    const currentServers = await this._getCurrentServers();

    if (currentServers.length !== requiredServers.length || requiredServers.some(server => !currentServers.includes(server))) {
      this._log('verbose', 'Difference in ntp servers detected', currentServers, requiredServers);
      await this._setNtpServers(requiredServers);
    } else {
      this._log('debug', 'NTP servers matching');
    }
  }

  async _ensureTimeZone() {
    if (!this._device.time) {
      // compatibility with older clouds
      return;
    }
    this._log('info', 'Checking timezone');
    const requestedTimezone = this._device.time.zone;
    const currentTimezone = await this._getTimeZone();
    if (currentTimezone !== requestedTimezone) {
      this._log('verbose', `Change in timezone detected, requested: ${requestedTimezone}, current: ${currentTimezone}`);
      this._setTimeZone(requestedTimezone);
    } else {
      this._log('debug', 'Configured timezone matches requested timezone');
    }
  }

  async _getTimeZone() {
    let currentTimeZone;

    try {
      currentTimeZone = (await execSync('timedatectl | grep "Time zone" | awk \'{print $3}\''))
        .toString()
        .trim();
    } catch (err) {
      this._log('error', `Error while checking timezone :${err.message}`);
    }

    if (!currentTimeZone) {
      // fallback to TIMEZONE_FILE if above fails since TIMEZONE_FILE can be out of sync with actual timedatectl
      currentTimeZone = (await fs.readFile(TIMEZONE_FILE, 'utf-8')).trim();
    }

    this._log('info', `Detected timezone is ${currentTimeZone}`);
    return currentTimeZone;
  }

  _setTimeZone(timezone = 'UTC') {
    this._log('info', `Setting timezone to: ${timezone}`);
    try {
      execSync(`sudo timedatectl set-timezone ${timezone}`);
    } catch (error) {
      this._log('error', 'Could not set timezone', error);
    }
  }

  _onDestroy() {
    this._intervalRunner.destroy();
    delete this._intervalRunner;
    super._onDestroy();
  }
}

export default TimeBridge;
