import Bridge from '../bridge';
import serviceRegistry from '../../../services/serviceRegistry';
import { commands } from '../../../services/ddp/constants';

/**
 * The upgrade bridge handles the forced upgrade actions
 * It glues blue with the systemVersionMonitor service
 */
class UpgradeBridge extends Bridge {
  get supportedRPCCommands() {
    return [commands.FORCE_UPGRADE];
  }

  /**
   * Act on incoming client RPCs
   * RPC's are useful for one-time actions that are not part of the application state.
   *
   * @param rpc
   * @returns {*}
   * @private
   */
  _onRpcAdded(rpc) {
    this._log('verbose', `Handling rpc ${JSON.stringify(rpc, null, ' ')}`, { serviceLog: true });
    switch (rpc.command) {
    case commands.FORCE_UPGRADE:
      // By returning here, the rpc will only be marked as complete when the upgrade is complete (or fails)
      // This does mean we can't handle any other RPC in the meantime, since we only process rpc's sequentially
      // But that's ok during an upgrade.
      // This also means that if the watchdog timer decides to reboot, the upgrade will be retried (as it won't have been marked as completed)
      // In some scenarios (usually when upgrading kernels), we need 2 attempts.
      // So this helps to make sure the upgrade succeeds when doing a forced upgrade
      //
      // When the upgrade fails for another reason, like the health check, we will mark the RPC as completed, so we avoid endless loops.
      // When the upgrade succeeds, purple may have been killed. In that case we will retry, but it will be a no-op.
      return this.forceUpgrade();
    default:
    }
  }

  forceUpgrade() {
    const versionMonitor = serviceRegistry.get('versionMonitor');
    return versionMonitor.checkVersionAndUpgrade();
  }
}

export default UpgradeBridge;
