import * as _ from 'lodash';

import {
  deviceRole,
  audioSinkRole,
  audioModuleRole,
  audioSourceRole,
  pulseaudioModules,
  audioReferenceSampleType,
  audioCalibrationState,
  usbCameraRole,
} from '../../../../services/ddp/constants';
import serviceRegistry from '../../../../services/serviceRegistry';

// Info on pulse modules:
// https://www.freedesktop.org/wiki/Software/PulseAudio/Documentation/User/Modules/#index64h3

const cleanObject = (object) => {
  if (Array.isArray(object)) {
    return object.map(el => cleanObject(el));
  } else if (_.isObject(object)) {
    return Object.keys(object).reduce((acc, el) => {
      if ((el.substr(0, 1) !== '_') || (el.substr(0, 5) === '_path')) {
        acc[el] = cleanObject(object[el]);
      }
      return acc;
    }, {});
  }
  return object;
};

const deviceIsTile = device => device.role === deviceRole.VIRTUAL_TILE || device.role === deviceRole.VIRTUAL_OFFWALL_TILE;
const deviceIsBroadcastCompositor = device => device.role === deviceRole.BROADCAST_COMPOSITOR;
const calibrationState = device => (device.audio.calibration || audioCalibrationState.OFF);
const hasPresenterCameraToBeRecordedWithAudio = device => device.hasOwnProperty('usbCameraConfig')
  && Array.isArray(device.usbCameraConfig)
  && device.usbCameraConfig.some(conf => (conf.role === usbCameraRole.PRESENTER && conf.record && conf.hasOwnProperty('audioDevice') && conf.audioDevice));

// In case of an overflow scenario we need to create a mix to connect the local sources to.
// Also in audio calibration mode, create a local mix because we need to be able to compare
// the volume of content sharing (sent to local mix) with the volume of a reference sample
// (sent to main mix) and adjust the gain of the local mix while listening on main mix (both
// reference and content sharing end up on main mix)
const localMixNeeded = device => (
  (calibrationState(device) !== audioCalibrationState.OFF) || (!device.streamRemoteAudio)
);

// We need a remote mix as soon as there is a local mix and calibration state is off
const remoteMixNeeded = device => (
  localMixNeeded(device) && (calibrationState(device) === audioCalibrationState.OFF)
);

// We need a main mix as soon as one output is configured for it or when local/remote mix exist
const mainMixNeeded = device => (
  // Always generate a main mix, because we sometimes need to stream out (digital audio mix)
  // even if we do not send the audio to an output of the device (eg. a tile in a Virtual Classroom).
  // In such case the combine sink is connected to the null output which acts as a dummy slave.
  !!device // = return true for each non-null device
);

// A recording mix mixes the teachers voice with the other audio (content, remote participants, ...)
const recordingMixNeeded = device => (
  hasPresenterCameraToBeRecordedWithAudio(device)
  && (calibrationState(device) === audioCalibrationState.OFF)
);

const groupHardwareSourcesPerRole = device => (
  // Note: audioSourcesLatest contains upserted hardware sources
  Object
    .values(audioSourceRole)
    .reduce((acc, role) => {
      acc[role] = device.audioSourcesLatest.filter((source) => {
        const sourceConfig = device.audioSourceConfig.find(s => s.id === source.id);
        // If source is used for a camera, consider it as IDLE (camera setting has priority on input setting).
        const sourceIsUsedByUsbCamera = device.usbCameraConfig.some(conf => conf.audioDevice === source.id);
        return (
          ((sourceConfig && sourceConfig.role === role) && !sourceIsUsedByUsbCamera)
          || ((!sourceConfig || sourceIsUsedByUsbCamera) && role === audioSourceRole.IDLE)
        );
      });
      return acc;
    }, {})
);

const groupHardwareSinksPerRole = device => (
  // Note: audioSinksLatest contains upserted hardware sinks
  Object
    .values(audioSinkRole)
    .reduce((acc, role) => {
      acc[role] = device.audioSinksLatest.filter((sink) => {
        const sinkConfig = device.audioSinkConfig.find(s => s.id === sink.id);
        return (
          (sinkConfig && sinkConfig.role === role)
          || (!sinkConfig && role === audioSinkRole.IDLE)
        );
      });
      return acc;
    }, {})
);

const findRtspContentStreamSource = (device) => {
  if (recordingMixNeeded(device)) {
    return `${audioModuleRole.RECORDING_MIX}.monitor`;
  }
  if (localMixNeeded(device)) {
    return `${audioModuleRole.LOCAL_MIX}.monitor`;
  }
  if (mainMixNeeded(device)) {
    return `${audioModuleRole.MAIN_MIX}.monitor`;
  }
  return null;
};

// Find info on audio source to be streamed along each of the USB cameras. Info is
// used in the camera bridge, where the rtsp streams for the USB cameras are launched
// (= presenter or room camera in overflow scenario and auto-launched source cases)
// ATTENTION: in the streamingBridge the rtsp streams for confidence monitors and
//            recording are launched and those use a different audio source (defined
//            by the findRtspContentStreamSource function here above)
const usbCameraSourceInfo = device => (
  device.usbCameraConfig.map((conf, index) => {
    const micCalibrationActive = calibrationState(device) === audioCalibrationState.MICROPHONE;
    const audioDevice = conf.audioDevice ? conf.audioDevice : null;
    const cameraNeedsRTSPStreaming = !!conf.stream && !micCalibrationActive;

    // First set values for the default case ... (eg. room camera, auto launched source)
    // If the usb device has an audio source configured, use it for the RTSP stream
    // No compression, use HW input directly for streaming out
    //
    // Hardware input X --> room camera or auto launched source RTSP stream (UsbCameraBridge)
    //
    const result = {};
    result.id = conf.id;
    result.audioDevice = audioDevice;
    result.loopbackNode = null;
    result.sinkForLoopbackNode = null;
    result.latencyForLoopbackNode = null;
    result.agcNode = null;
    result.micMixNode = null;
    result.recLoopbackNode = null;
    result.latencyForRecLoopbackNode = null;
    result.rtspAudioDevice = cameraNeedsRTSPStreaming ? audioDevice : null;
    result.setRtspAudioDeviceAsDefaultSource = false;

    // In case of microphone calibration (SPMAS), send all used HW inputs to a single mix
    // that is to be streamed out using webRTC to the administrator UI
    //
    // Hardware input X --> loopbackNode X --> micMix (= default source, picked up by SPMAS webRTC (AudioBridge))
    //
    if (audioDevice && micCalibrationActive) {
      result.loopbackNode = `${audioModuleRole.LOOPBACK}_${index}`;
      result.sinkForLoopbackNode = audioModuleRole.MIC_MIX;
      result.micMixNode = audioModuleRole.MIC_MIX;
      result.rtspAudioDevice = `${audioModuleRole.MIC_MIX}.monitor`;
      result.setRtspAudioDeviceAsDefaultSource = true;
      return result;
    }

    // Presenter cam: Send HW input first through compressor before streaming out
    if (audioDevice && (conf.role === usbCameraRole.PRESENTER)) {
      if (cameraNeedsRTSPStreaming) { // RTSP case (Camera mgr. or Master) (eg. master overflow room <--> overflow room)
        //
        //                                                                      |--> camera or auto launched source RTSP stream (UsbCameraBridge)
        // Hardware input X --> loopbackNode X --> compressor X --> micMix X -->|
        //                                                                      |-->|
        //                                                                          |--> recLoopbackNode (delay 350ms) --> recordingMix (if existing) --> recording RTSP stream (StreamingBridge)
        // main_mix (contains a.o. audio from Digital Audio Mix streams)----------->|
        //
        result.loopbackNode = `${audioModuleRole.LOOPBACK}_${index}`;
        result.sinkForLoopbackNode = `${audioModuleRole.AGC_SINK}_${index}`;
        result.agcNode = `${audioModuleRole.AGC_SINK}_${index}`;
        result.micMixNode = `${audioModuleRole.MIC_MIX}_${index}`;
        result.recLoopbackNode = recordingMixNeeded(device) ? `${audioModuleRole.LOOPBACK}_rec` : null;
        result.latencyForRecLoopbackNode = recordingMixNeeded(device) ? '350' : null;
        result.rtspAudioDevice = `${audioModuleRole.MIC_MIX}_${index}.monitor`;
      } else if (recordingMixNeeded(device)) { // WebRTC case (Camera mgr.) (eg. virtual classroom with recording)
        //
        //                     |--> picked up by chromium browser for webRTC streaming (BrowserBridge)
        // Hardware input X -->|
        //                     |--> loopbackNode X ----------------------->|
        //                                                                 |--> recordingMix --> recording RTSP stream (StreamingBridge)
        // main_mix (contains a.o. audio from Digital Audio Mix streams)-->|
        //
        result.loopbackNode = `${audioModuleRole.LOOPBACK}_${index}`;
        result.sinkForLoopbackNode = audioModuleRole.RECORDING_MIX;
      }
    }

    return result;
  })
);

// Find the audio source to be used as intertile source.
const auxInSourceInfo = (device) => {
  const sourcesWithRole = groupHardwareSourcesPerRole(device);
  return sourcesWithRole[audioSourceRole.AUX_IN].map((source, index) => {
    const micCalibrationActive = calibrationState(device) === audioCalibrationState.MICROPHONE;
    const contentCalibrationActive = calibrationState(device) === audioCalibrationState.CONTENT_SHARING;
    // Output of external DSP is connected to input with role AUX_IN, use it for intertile audio
    let loopbackNode = null;
    let sinkForLoopbackNode = null;
    const latencyForLoopbackNode = null;
    const agcNode = null;
    let micMixNode = null;
    const recLoopbackNode = null;
    const latencyForRecLoopbackNode = null;
    const audioDevice = source.id;
    let rtspAudioDevice = source.id;
    let setRtspAudioDeviceAsDefaultSource = !contentCalibrationActive; // Set as fallback source if no content sharing calibration
    const sourceIsUsedByUsbCamera = device.usbCameraConfig.some(conf => conf.audioDevice === source.id);

    // In case of microphone calibration (SPMAS), send aux inputs to a single mix
    // that is to be streamed out using webRTC to the administrator UI
    // Aux input X --> loopbackNode X --> micMix (=fallback source picked up by SPMAS webRTC)
    if (micCalibrationActive && !sourceIsUsedByUsbCamera) {
      loopbackNode = `${audioModuleRole.LOOPBACK}_aux_${index}`;
      sinkForLoopbackNode = audioModuleRole.MIC_MIX;
      micMixNode = audioModuleRole.MIC_MIX;
      rtspAudioDevice = `${audioModuleRole.MIC_MIX}.monitor`;
      setRtspAudioDeviceAsDefaultSource = true;
    }

    // In case of a broadcast compositor, add AUX_IN to the main mix
    if (deviceIsBroadcastCompositor(device)) {
      loopbackNode = `${audioModuleRole.LOOPBACK}_aux_${index}`;
      sinkForLoopbackNode = audioModuleRole.MAIN_MIX;
      rtspAudioDevice = `${audioModuleRole.MAIN_MIX}.monitor`;
      setRtspAudioDeviceAsDefaultSource = true;
    }

    return {
      id: source.id,
      audioDevice,
      loopbackNode,
      sinkForLoopbackNode,
      latencyForLoopbackNode,
      agcNode,
      micMixNode,
      recLoopbackNode,
      latencyForRecLoopbackNode,
      rtspAudioDevice,
      setRtspAudioDeviceAsDefaultSource,
    };
  });
};

// Find the audio source to be sent out with the main mix.
const loopthroughSourceInfo = (device) => {
  const sourcesWithRole = groupHardwareSourcesPerRole(device);
  return sourcesWithRole[audioSourceRole.AEC_REF_IN].map((source, index) => {
    const micCalibrationActive = calibrationState(device) === audioCalibrationState.MICROPHONE;
    const loopbackNode = `${audioModuleRole.LOOPBACK}_loopthrough_${index}`;
    let sinkForLoopbackNode = audioModuleRole.MAIN_MIX;
    let latencyForLoopbackNode = '20'; // We want low latency connection from input to output
    const agcNode = null;
    let micMixNode = null;
    const recLoopbackNode = null;
    const latencyForRecLoopbackNode = null;
    const audioDevice = source.id;
    let rtspAudioDevice = null;
    let setRtspAudioDeviceAsDefaultSource = false;

    // In case of microphone calibration (SPMAS), send aec ref in inputs to a single mix
    // that is to be streamed out using webRTC to the administrator UI
    // AEC input X --> loopbackNode X --> micMix (=fallback source picked up by SPMAS webRTC)
    if (micCalibrationActive) {
      sinkForLoopbackNode = audioModuleRole.MIC_MIX;
      micMixNode = audioModuleRole.MIC_MIX;
      rtspAudioDevice = `${audioModuleRole.MIC_MIX}.monitor`;
      setRtspAudioDeviceAsDefaultSource = true;
      latencyForLoopbackNode = null;
    }

    return {
      id: source.id,
      audioDevice,
      loopbackNode,
      sinkForLoopbackNode,
      latencyForLoopbackNode,
      agcNode,
      micMixNode,
      recLoopbackNode,
      latencyForRecLoopbackNode,
      rtspAudioDevice,
      setRtspAudioDeviceAsDefaultSource,
    };
  });
};

const logAudioConfig = (logger, config) => {
  logger.log('verbose', '================================================================================');
  logger.log('verbose', '======================= START AUDIO CONFIGURATION SUMMARY ======================');
  logger.log('verbose', `Current meeting id = ${config.meetingId}`);
  logger.log('verbose', `Current device role = '${config.currentDeviceRole}'`);
  logger.log('verbose', `Stream remote audio = ${config.streamRemoteAudio}`);
  logger.log('verbose', `Audio calibration state = '${config.calibrationState}'`);
  if (config.agcUsed) {
    logger.log('verbose', `Compressor parameters = ${config.compressorParms}`);
  }
  logger.log('verbose', `Local mix needed: '${config.localMixNeeded}'`);
  logger.log('verbose', `Remote mix needed: '${config.remoteMixNeeded}'`);
  logger.log('verbose', `Main mix needed: '${config.mainMixNeeded}'`);
  logger.log('verbose', `Recording mix needed: '${config.recordingMixNeeded}'`);
  logger.log('verbose', `Default (=fallback) source is set to '${config.defaultSource}'`);
  logger.log('verbose', `Default (=fallback) sink is set to '${config.defaultSink}'`);
  logger.log('verbose', `Content sharing default volume will be applied on sink '${config.contentSharingSink}'`);
  logger.log('verbose', `Content sharing stream source is set to '${config.rtspContentStreamSource}'`);
  logger.log('verbose', `Local sources will be routed to sink '${config.contentSharingSink}'`);
  logger.log('verbose', `Remote sources (if any) will be routed to sink '${config.remoteSourceSink}'`);
  logger.log('verbose', `Inter tile mix sources (if any) will be routed to sink '${config.interTileMixSink}'`);
  if (config.calibrationState === audioCalibrationState.MICROPHONE) {
    logger.log('verbose', `Reference samples (SPMAS) of type '${audioReferenceSampleType.TEACHER}' will be routed to sink '${config.sinkForReferenceAudioSample[audioReferenceSampleType.TEACHER]}'`);
  }
  if (config.calibrationState === audioCalibrationState.CONTENT_SHARING) {
    logger.log('verbose', `Reference samples (SPMAS) of type '${audioReferenceSampleType.CONTENT}' will be routed to sink '${config.sinkForReferenceAudioSample[audioReferenceSampleType.CONTENT]}'`);
  }
  logger.log('verbose', 'Audio sources grouped per role:', JSON.stringify(cleanObject(config.sourcesWithRole), null, 2));
  logger.log('verbose', 'Audio sinks grouped per role:', JSON.stringify(cleanObject(config.sinksWithRole), null, 2));
  logger.log('verbose', 'USB camera sources:', JSON.stringify(config.usbCameraSourceInfo, null, 2));
  logger.log('verbose', 'AUX IN sources:', JSON.stringify(config.auxInSourceInfo, null, 2));
  logger.log('verbose', 'AEC REF IN sources:', JSON.stringify(config.loopthroughSourceInfo, null, 2));
  logger.log('verbose', 'Required routing:', JSON.stringify(config.routing, null, 2));
  logger.log('verbose', '======================== END AUDIO CONFIGURATION SUMMARY =======================');
  logger.log('verbose', '================================================================================');
};

const initializeAudioConfig = (config, device) => {
  config.routing = [];
  config.meetingId = device.meetingId;
  config.currentDeviceRole = device.role;
  config.streamRemoteAudio = device.streamRemoteAudio;
  config.localMixNeeded = localMixNeeded(device);
  config.remoteMixNeeded = remoteMixNeeded(device);
  config.mainMixNeeded = mainMixNeeded(device);
  config.recordingMixNeeded = recordingMixNeeded(device);
  config.calibrationState = calibrationState(device);
  config.defaultSource = 'tile_mix.monitor'; // exists always, defined in /etc/pulse/default.pa
  config.defaultSink = 'nullsink'; // exists always, defined in /etc/pulse/default.pa
  config.contentSharingSink = 'nullsink';
  config.remoteSourceSink = 'nullsink';
  config.interTileMixSink = 'tile_mix';
  config.sinkForReferenceAudioSample = {};
  config.rtspContentStreamSource = findRtspContentStreamSource(device);
  config.sourcesWithRole = groupHardwareSourcesPerRole(device);
  config.sinksWithRole = groupHardwareSinksPerRole(device);
  config.usbCameraSourceInfo = usbCameraSourceInfo(device);
  config.auxInSourceInfo = auxInSourceInfo(device);
  config.loopthroughSourceInfo = loopthroughSourceInfo(device);
  config.compressorParms = process.env.EDU_COMPRESSOR_PARMS || '0,10,150,-18,5,5,3';
  config.agcUsed = false;
  config.micMixExists = false;
  Object.values(audioReferenceSampleType).forEach((val) => {
    config.sinkForReferenceAudioSample[val] = 'nullsink';
  });
};

const calculateAudioConfig = (config, device, logger, logRouting = true) => {
  // First initialize everything
  initializeAudioConfig(config, device);

  const recordingMixAsSlave = [];
  if (config.recordingMixNeeded) {
    const recordingMixNode = {
      module: pulseaudioModules.NULL_SINK,
      arguments: {
        sink_name: audioModuleRole.RECORDING_MIX,
      },
    };
    config.routing.push(recordingMixNode);
    recordingMixAsSlave.push(audioModuleRole.RECORDING_MIX);
  }

  if (config.mainMixNeeded) {
    // Check what needs to be connected to the output of the main mix combine sink
    const mainMixSlaves = [
      ...recordingMixAsSlave,
      ...config.sinksWithRole[audioSinkRole.MAIN_MIX].map(sink => sink.id),
      ...config.sinksWithRole[audioSinkRole.AEC_REF_MIX].map(sink => sink.id),
    ];

    // Never use a nullsink for the main mix as it causes audio hiccups (inter tile audio)
    // Connect to a dummy slave if there is nothing else to connect to
    if (!mainMixSlaves.length) {
      mainMixSlaves.push('nullsink');
    }

    const mainMix = {
      module: pulseaudioModules.COMBINED_SINK,
      arguments: {
        sink_name: audioModuleRole.MAIN_MIX,
        slaves: mainMixSlaves,
      },
    };

    // For time being set the fallback sink to MAIN_MIX
    config.defaultSink = audioModuleRole.MAIN_MIX;
    // For time being set the content sharing sink to MAIN_MIX,
    // and use it to control the default volume of content sharing.
    config.contentSharingSink = audioModuleRole.MAIN_MIX;
    // Same for remote sources, intialize but may be overwritten later
    config.remoteSourceSink = audioModuleRole.MAIN_MIX;
    // Send reference audio samples used during calibration to the main mix
    config.sinkForReferenceAudioSample[audioReferenceSampleType.CONTENT] = audioModuleRole.MAIN_MIX;
    // If audio calibration session for content sharing is ongoing, we must
    // set MAIN_MIX.monitor as default source because that is what we want to
    // hear in the admin UI (we are adjusting the volume of content sharing
    // by adjusting the gain of the LOCAL_MIX while comparing the volume of
    // content sharing with a reference sample that is sent to MAIN_MIX).
    // Also, if the device is not a tile (which uses tile_mix.monitor as default
    // source (set at start of function)), use the main mix as default source.
    if (config.calibrationState === audioCalibrationState.CONTENT_SHARING) {
      config.defaultSource = `${audioModuleRole.MAIN_MIX}.monitor`;
    } else if (!deviceIsTile(device)) {
      // Realtime caption generation on camera manager needs recording mix as input
      config.defaultSource = config.recordingMixNeeded ? `${audioModuleRole.RECORDING_MIX}.monitor` : `${audioModuleRole.MAIN_MIX}.monitor`;
    }

    config.routing.push(mainMix);
  }

  // Create a mix for local sources
  if (config.localMixNeeded) {
    // Set it as default sink for incoming playbackstreams
    config.defaultSink = audioModuleRole.LOCAL_MIX;
    // Add everything that needs to receive the local mix to the slave list
    const localMixSlaves = config.sinksWithRole[audioSinkRole.LOCAL_MIX].map(sink => sink.id);
    localMixSlaves.push(audioModuleRole.MAIN_MIX);

    const localMix = {
      module: pulseaudioModules.COMBINED_SINK,
      arguments: {
        sink_name: audioModuleRole.LOCAL_MIX,
        slaves: localMixSlaves,
      },
    };

    // If a LOCAL_MIX exists, set that one as content sharing sink,
    // to control the default volume of content sharing.
    config.contentSharingSink = audioModuleRole.LOCAL_MIX;

    config.routing.push(localMix);
  }

  // Create mix for remote sources
  if (config.remoteMixNeeded) {
    // Add everything that needs to receive the remote mix to the slave list
    const remoteMixSlaves = config.sinksWithRole[audioSinkRole.REMOTE_MIX].map(sink => sink.id);
    remoteMixSlaves.push(audioModuleRole.MAIN_MIX);

    const remoteMix = {
      module: pulseaudioModules.COMBINED_SINK,
      arguments: {
        sink_name: audioModuleRole.REMOTE_MIX,
        slaves: remoteMixSlaves,
      },
    };

    // If a REMOTE_MIX exists, set that one as sink for remote sources
    config.remoteSourceSink = audioModuleRole.REMOTE_MIX;

    config.routing.push(remoteMix);
  }

  // Create modules for USB cameras that need to be streamed out using RTSP with compressed audio
  // and for AEC_REF_IN and AUX_IN inputs
  [...config.usbCameraSourceInfo, ...config.loopthroughSourceInfo, ...config.auxInSourceInfo].forEach((cam) => {
    // Create at most one 'micMix' without index (would be eg. 'micMix_1') mixing node.
    // It can be used by several inputs if they refer to it.
    const createNewMicMixNode = (cam.micMixNode && ((cam.micMixNode !== audioModuleRole.MIC_MIX) || (!config.micMixExists)));
    if (createNewMicMixNode) {
      const micMixNode = {
        module: pulseaudioModules.NULL_SINK,
        arguments: {
          sink_name: cam.micMixNode,
        },
      };
      config.routing.push(micMixNode);
      if (cam.micMixNode === audioModuleRole.MIC_MIX) {
        config.micMixExists = true;
      }
    }
    // Add Automatice Gain Control (AGC) node (= 'compressor/expander') if necessary
    if (cam.agcNode) {
      const agcNode = {
        module: pulseaudioModules.LADSPA_SINK,
        arguments: {
          sink_name: cam.agcNode,
          sink_master: cam.micMixNode,
          plugin: 'sc4m_1916',
          label: 'sc4m',
          // control: '1,1.5,401,-30,20,5,12',  <-- these settings config in noise when nobody speaks
          // See also: http://plugin.org.uk/ladspa-swh/docs/ladspa-swh.html#tth_sEc2.92
          // Control parameters of the compressor/expander and what they do:
          // 1) RMS/PEAK: 0 to 1, default 0. If threshold should apply to peaks (machine-like) or RMS values (human-ear-like).
          //    The balance between the RMS and peak envelope followers. RMS is generally better for subtle,
          //    musical compression and peak is better for heavier, fast compression and percussion.
          // 2) ATTACK: 1.5 to 400, default 101.125. How quick to react. The attack time in milliseconds.
          // 3) RELEASE: 2 to 800, default 401. How slow to release. The release time in milliseconds.
          // 4) THRESHOLD: -30 to 0, default 0. When to react. Threshold level (dB): The point at which the compressor will start to kick in.
          // 5) RATIO: 1 to 20, default 1. How much to compress levels above the treshold.
          //    The gain reduction ratio used when the signal level exceeds the threshold.
          // 6) KNEE RADIUS: 1 to 10, default 3.25. How soft the compressor should kick in.
          //    Knee radius (dB): The distance from the threshold where the knee curve starts.
          //    Zero will apply ratio immediately when level hits threshold.
          //    Higher values leads to softer compression closer to the threshold.
          // 7) MAKEUP GAIN: 0 to 24, default 0. Fixed audio boost. Makeup gain (dB): Controls the gain of the makeup input signal in dB's.
          // Next line: settings based on recommendations found at
          // https://www.geekality.net/2013/11/01/settings-for-vlc-dynamic-range-compression/
          // const compressorParms = process.env.EDU_COMPRESSOR_PARMS || '0,10,150,-18,5,5,3';
          control: config.compressorParms,
        },
      };
      config.routing.push(agcNode);
      config.agcUsed = true;
    }
    if (cam.loopbackNode) {
      const loopbackArguments = {
        source: cam.audioDevice,
        sink: cam.sinkForLoopbackNode,
      };
      if (cam.latencyForLoopbackNode) {
        loopbackArguments.latency_msec = cam.latencyForLoopbackNode;
      }
      const loopbackNode = {
        module: pulseaudioModules.LOOPBACK,
        arguments: loopbackArguments,
      };
      config.routing.push(loopbackNode);
    }
    if (cam.recLoopbackNode) { // Add to delay the audio (recording lipsync)
      const recLoopbackArguments = {
        source: cam.rtspAudioDevice,
        sink: audioModuleRole.RECORDING_MIX,
      };
      if (cam.latencyForRecLoopbackNode) {
        recLoopbackArguments.latency_msec = cam.latencyForRecLoopbackNode;
      }
      const recLoopbackNode = {
        module: pulseaudioModules.LOOPBACK,
        arguments: recLoopbackArguments,
      };
      config.routing.push(recLoopbackNode);
    }
    if (cam.setRtspAudioDeviceAsDefaultSource) { // AUX IN (tile) or SPMAS microphone calibration case
      config.defaultSource = cam.rtspAudioDevice;
    }
    if (config.calibrationState === audioCalibrationState.MICROPHONE) {
      // Send reference audio samples used during calibration to the microphone mix
      config.sinkForReferenceAudioSample[audioReferenceSampleType.TEACHER] = cam.micMixNode;
      config.defaultSource = `${audioModuleRole.MIC_MIX}.monitor`; // Already done but for safety...
    }
  });

  config.interTileMixSink = deviceIsTile(device) ? 'tile_mix' : config.defaultSink;

  if (logRouting) {
    logAudioConfig(logger, config);
  }
};

const getCurrentRoutingModules = (paCache, roles, logRouting = true) => {
  const sanitizeArguments = (args) => {
    Object.entries(args)
      .filter(([, val]) => val.includes(','))
      .forEach(([key, val]) => { args[key] = val.split(','); });
    return args;
  };
  const result = Object.values(roles)
    .reduce(
      (acc, role) => ([
        ...acc,
        ...paCache.findModulesForWhichSinkNameIncludes(role),
      ]),
      [],
    )
    .map(module => ({
      module: module.name,
      arguments: sanitizeArguments(module.arguments),
    }));

  if (logRouting) {
    const logger = serviceRegistry.get('logger').getInstance('CalcAudioConfig');
    logger.log('verbose', `Current routing = ${JSON.stringify(result, null, 2)}`);
  }

  return result;
};

export {
  deviceIsTile,
  findRtspContentStreamSource,
  usbCameraSourceInfo,
  initializeAudioConfig,
  calculateAudioConfig,
  logAudioConfig,
  getCurrentRoutingModules,
};
