/* eslint-disable */
import * as  EJSON from 'ddp-ejson';
import * as _ from 'lodash';

import DiffSequence from './diff-sequence';
import MongoID from './mongo-id';
import IdMap from './id-map';
import Tracker from './tracker';
import OrderedDict from './ordered-dict';

const CollectionHelpers = {};

CollectionHelpers._observeChangesCallbacksAreOrdered = function (callbacks) {
  if (callbacks.added && callbacks.addedBefore)
    throw new Error('Please specify only one of added() and addedBefore()')
  return !!(callbacks.addedBefore || callbacks.movedBefore)
}

CollectionHelpers._observeCallbacksAreOrdered = function (callbacks) {
  if (callbacks.addedAt && callbacks.added)
    throw new Error('Please specify only one of added() and addedAt()')
  if (callbacks.changedAt && callbacks.changed)
    throw new Error('Please specify only one of changed() and changedAt()')
  if (callbacks.removed && callbacks.removedAt)
    throw new Error('Please specify only one of removed() and removedAt()')

  return !!(callbacks.addedAt || callbacks.movedTo || callbacks.changedAt ||
    callbacks.removedAt)
}

// the handle that comes back from observe.
CollectionHelpers.ObserveHandle = function () {}


// XXX the sorted-query logic below is laughably inefficient. we'll
// need to come up with a better datastructure for this.
//
// XXX the logic for observing with a skip or a limit is even more
// laughably inefficient. we recompute the whole results every time!

CollectionHelpers._insertInResults = function (query, doc) {
  var fields = EJSON.clone(doc)
  delete fields._id
  if (query.ordered) {
    if (!query.sorter) {
      query.addedBefore(doc._id, query.projectionFn(fields), null)
      query.results.push(doc)
    } else {
      var i = CollectionHelpers._insertInSortedList(
        query.sorter.getComparator({
          distances: query.distances
        }),
        query.results, doc)
      var next = query.results[i + 1]
      if (next)
        next = next._id
      else
        next = null
      query.addedBefore(doc._id, query.projectionFn(fields), next)
    }
    query.added(doc._id, query.projectionFn(fields))
  } else {
    query.added(doc._id, query.projectionFn(fields))
    query.results.set(doc._id, doc)
  }
}

CollectionHelpers._removeFromResults = function (query, doc) {
  if (query.ordered) {
    var i = CollectionHelpers._findInOrderedResults(query, doc)
    query.removed(doc._id)
    query.results.splice(i, 1)
  } else {
    var id = doc._id // in case callback mutates doc
    query.removed(doc._id)
    query.results.remove(id)
  }
}

CollectionHelpers._updateInResults = function (query, doc, old_doc) {
  if (!EJSON.equals(doc._id, old_doc._id))
    throw new Error("Can't change a doc's _id while updating")
  var projectionFn = query.projectionFn
  var changedFields = DiffSequence.makeChangedFields(
    projectionFn(doc), projectionFn(old_doc))

  if (!query.ordered) {
    if (!_.isEmpty(changedFields)) {
      query.changed(doc._id, changedFields)
      query.results.set(doc._id, doc)
    }
    return
  }

  var orig_idx = CollectionHelpers._findInOrderedResults(query, doc)

  if (!_.isEmpty(changedFields))
    query.changed(doc._id, changedFields)
  if (!query.sorter)
    return

  // just take it out and put it back in again, and see if the index
  // changes
  query.results.splice(orig_idx, 1)
  var new_idx = CollectionHelpers._insertInSortedList(
    query.sorter.getComparator({
      distances: query.distances
    }),
    query.results, doc)
  if (orig_idx !== new_idx) {
    var next = query.results[new_idx + 1]
    if (next)
      next = next._id
    else
      next = null
    query.movedBefore && query.movedBefore(doc._id, next)
  }
}


CollectionHelpers._findInOrderedResults = function (query, doc) {
  if (!query.ordered)
    throw new Error("Can't call _findInOrderedResults on unordered query")
  for (var i = 0; i < query.results.length; i++)
    if (query.results[i] === doc)
      return i
  throw Error('object missing from query')
}

// This binary search puts a value between any equal values, and the first
// lesser value.
CollectionHelpers._binarySearch = function (cmp, array, value) {
  var first = 0,
    rangeLength = array.length

  while (rangeLength > 0) {
    var halfRange = Math.floor(rangeLength / 2)
    if (cmp(value, array[first + halfRange]) >= 0) {
      first += halfRange + 1
      rangeLength -= halfRange + 1
    } else {
      rangeLength = halfRange
    }
  }
  return first
}

CollectionHelpers._insertInSortedList = function (cmp, array, value) {
  if (array.length === 0) {
    array.push(value)
    return 0
  }

  var idx = CollectionHelpers._binarySearch(cmp, array, value)
  array.splice(idx, 0, value)
  return idx
}


// ordered: bool.
// old_results and new_results: collections of documents.
//    if ordered, they are arrays.
//    if unordered, they are IdMaps
CollectionHelpers._diffQueryChanges = function (ordered, oldResults, newResults, observer, options) {
  return DiffSequence.diffQueryChanges(ordered, oldResults, newResults, observer, options)
}

CollectionHelpers._diffQueryUnorderedChanges = function (oldResults, newResults, observer, options) {
  return DiffSequence.diffQueryUnorderedChanges(oldResults, newResults, observer, options)
}

CollectionHelpers._diffQueryOrderedChanges =
  function (oldResults, newResults, observer, options) {
    return DiffSequence.diffQueryOrderedChanges(oldResults, newResults, observer, options)
  }

CollectionHelpers._diffObjects = function (left, right, callbacks) {
  return DiffSequence.diffObjects(left, right, callbacks)
}

const isPlainObject = CollectionHelpers._isPlainObject = function (x) {
  return x && CollectionHelpers._f._type(x) === 3
}

// Is this selector just shorthand for lookup by _id?
CollectionHelpers._selectorIsId = function (selector) {
  return (typeof selector === 'string') ||
    (typeof selector === 'number') ||
    selector instanceof MongoID.ObjectID
}

// Is the selector just lookup by _id (shorthand or not)?
CollectionHelpers._selectorIsIdPerhapsAsObject = function (selector) {
  return CollectionHelpers._selectorIsId(selector) ||
    (selector && typeof selector === 'object' &&
      selector._id && CollectionHelpers._selectorIsId(selector._id) &&
      _.size(selector) === 1)
}

// If this is a selector which explicitly constrains the match by ID to a finite
// number of documents, returns a list of their IDs.  Otherwise returns
// null. Note that the selector may have other restrictions so it may not even
// match those document!  We care about $in and $and since those are generated
// access-controlled update and remove.
CollectionHelpers._idsMatchedBySelector = function (selector) {
  // Is the selector just an ID?
  if (CollectionHelpers._selectorIsId(selector))
    return [selector]
  if (!selector)
    return null

  // Do we have an _id clause?
  if (_.has(selector, '_id')) {
    // Is the _id clause just an ID?
    if (CollectionHelpers._selectorIsId(selector._id))
      return [selector._id]
    // Is the _id clause {_id: {$in: ["x", "y", "z"]}}?
    if (selector._id && selector._id.$in &&
      _.isArray(selector._id.$in) &&
      !_.isEmpty(selector._id.$in) &&
      _.every(selector._id.$in, CollectionHelpers._selectorIsId)) {
      return selector._id.$in
    }
    return null
  }

  // If this is a top-level $and, and any of the clauses constrain their
  // documents, then the whole selector is constrained by any one clause's
  // constraint. (Well, by their intersection, but that seems unlikely.)
  if (selector.$and && _.isArray(selector.$and)) {
    for (var i = 0; i < selector.$and.length; ++i) {
      var subIds = CollectionHelpers._idsMatchedBySelector(selector.$and[i])
      if (subIds)
        return subIds
    }
  }

  return null
}

// Wrap a transform function to return objects that have the _id field
// of the untransformed document. This ensures that subsystems such as
// the observe-sequence package that call `observe` can keep track of
// the documents identities.
//
// - Require that it returns objects
// - If the return value has an _id field, verify that it matches the
//   original _id field
// - If the return value doesn't have an _id field, add it back.
CollectionHelpers.wrapTransform = function (transform) {
  if (!transform)
    return null

  // No need to doubly-wrap transforms.
  if (transform.__wrappedTransform__)
    return transform

  var wrapped = function (doc) {
    if (!_.has(doc, '_id')) {
      // XXX do we ever have a transform on the oplog's collection? because that
      // collection has no _id.
      throw new Error('can only transform documents with _id')
    }

    var id = doc._id
    // XXX consider making tracker a weak dependency and checking Package.tracker here
    var transformed = Tracker.nonreactive(function () {
      return transform(doc)
    })

    if (!isPlainObject(transformed)) {
      throw new Error('transform must return object')
    }

    if (_.has(transformed, '_id')) {
      if (!EJSON.equals(transformed._id, id)) {
        throw new Error("transformed document can't have different _id")
      }
    } else {
      transformed._id = id
    }
    return transformed
  }
  wrapped.__wrappedTransform__ = true
  return wrapped
}

// Oddball function used by upsert.
CollectionHelpers._removeDollarOperators = function (selector) {
  var selectorDoc = {};
  for (var k in selector)
    if (k.substr(0, 1) !== '$')
      selectorDoc[k] = selector[k];
  return selectorDoc;
};


// helpers used by compiled selector code
 CollectionHelpers._f = {
  // XXX for _all and _in, consider building 'inquery' at compile time..

  _type(v) {
    if (typeof v === 'number')
      return 1
    if (typeof v === 'string')
      return 2
    if (typeof v === 'boolean')
      return 8
    if (Array.isArray(v) && !EJSON.isBinary(v))
      return 4
    if (v === null)
      return 10
    if (v instanceof RegExp)
      // note that typeof(/x/) === "object"
      return 11
    if (typeof v === 'function')
      return 13
    if (v instanceof Date)
      return 9
    if (EJSON.isBinary(v))
      return 5
    if (v instanceof MongoID.ObjectID)
      return 7
    return 3 // object

  // XXX support some/all of these:
  // 14, symbol
  // 15, javascript code with scope
  // 16, 18: 32-bit/64-bit integer
  // 17, timestamp
  // 255, minkey
  // 127, maxkey
  },

  // deep equality test: use for literal document and array matches
  _equal(a, b) {
    return EJSON.equals(a, b, {keyOrderSensitive: true})
  },

  // maps a type code to a value that can be used to sort values of
  // different types
  _typeorder(t) {
    // http://www.mongodb.org/display/DOCS/What+is+the+Compare+Order+for+BSON+Types
    // XXX what is the correct sort position for Javascript code?
    // ('100' in the matrix below)
    // XXX minkey/maxkey
    return [-1, // (not a type)
      1, // number
      2, // string
      3, // object
      4, // array
      5, // binary
      -1, // deprecated
      6, // ObjectID
      7, // bool
      8, // Date
      0, // null
      9, // RegExp
      -1, // deprecated
      100, // JS code
      2, // deprecated (symbol)
      100, // JS code
      1, // 32-bit int
      8, // Mongo timestamp
      1 // 64-bit int
    ][t]
  },

  // compare two values of unknown type according to BSON ordering
  // semantics. (as an extension, consider 'undefined' to be less than
  // any other value.) return negative if a is less, positive if b is
  // less, or 0 if equal
  _cmp(a, b) {
    if (a === undefined)
      return b === undefined ? 0 : -1
    if (b === undefined)
      return 1
    var ta =  CollectionHelpers._f._type(a)
    var tb =  CollectionHelpers._f._type(b)
    var oa =  CollectionHelpers._f._typeorder(ta)
    var ob =  CollectionHelpers._f._typeorder(tb)
    if (oa !== ob)
      return oa < ob ? -1 : 1
    if (ta !== tb)
      // XXX need to implement this if we implement Symbol or integers, or
      // Timestamp
      throw Error('Missing type coercion logic in _cmp')
    if (ta === 7) { // ObjectID
      // Convert to string.
      ta = tb = 2
      a = a.toHexString()
      b = b.toHexString()
    }
    if (ta === 9) { // Date
      // Convert to millis.
      ta = tb = 1
      a = a.getTime()
      b = b.getTime()
    }

    if (ta === 1) // double
      return a - b
    if (tb === 2) // string
      return a < b ? -1 : (a === b ? 0 : 1)
    if (ta === 3) { // Object
      // this could be much more efficient in the expected case ...
      var to_array = function (obj) {
        var ret = []
        for (var key in obj) {
          ret.push(key)
          ret.push(obj[key])
        }
        return ret
      }
      return  CollectionHelpers._f._cmp(to_array(a), to_array(b))
    }
    if (ta === 4) { // Array
      for (var i = 0;; i++) {
        if (i === a.length)
          return (i === b.length) ? 0 : -1
        if (i === b.length)
          return 1
        var s =  CollectionHelpers._f._cmp(a[i], b[i])
        if (s !== 0)
          return s
      }
    }
    if (ta === 5) { // binary
      // Surprisingly, a small binary blob is always less than a large one in
      // Mongo.
      if (a.length !== b.length)
        return a.length - b.length
      for (i = 0; i < a.length; i++) {
        if (a[i] < b[i])
          return -1
        if (a[i] > b[i])
          return 1
      }
      return 0
    }
    if (ta === 8) { // boolean
      if (a) return b ? 0 : 1
      return b ? -1 : 0
    }
    if (ta === 10) // null
      return 0
    if (ta === 11) // regexp
      throw Error('Sorting not supported on regular expression') // XXX
    // 13: javascript code
    // 14: symbol
    // 15: javascript code with scope
    // 16: 32-bit integer
    // 17: timestamp
    // 18: 64-bit integer
    // 255: minkey
    // 127: maxkey
    if (ta === 13) // javascript code
      throw Error('Sorting not supported on Javascript code') // XXX
    throw Error('Unknown type to sort')
  }
}


// Knows how to compile a fields projection to a predicate function.
// @returns - Function: a closure that filters out an object according to the
//            fields projection rules:
//            @param obj - Object: MongoDB-styled document
//            @returns - Object: a document with the fields filtered out
//                       according to projection rules. Doesn't retain subfields
//                       of passed argument.
 CollectionHelpers._compileProjection = function (fields) {
   CollectionHelpers._checkSupportedProjection(fields);

  var _idProjection = _.isUndefined(fields._id) ? true : fields._id;
  var details = projectionDetails(fields);

  // returns transformed doc according to ruleTree
  var transform = function (doc, ruleTree) {
    // Special case for "sets"
    if (_.isArray(doc))
      return _.map(doc, function (subdoc) { return transform(subdoc, ruleTree); });

    var res = details.including ? {} : EJSON.clone(doc);
    _.each(ruleTree, function (rule, key) {
      if (!_.has(doc, key))
        return;
      if (_.isObject(rule)) {
        // For sub-objects/subsets we branch
        if (_.isObject(doc[key]))
          res[key] = transform(doc[key], rule);
        // Otherwise we don't even touch this subfield
      } else if (details.including)
        res[key] = EJSON.clone(doc[key]);
      else
        delete res[key];
    });

    return res;
  };

  return function (obj) {
    var res = transform(obj, details.tree);

    if (_idProjection && _.has(obj, '_id'))
      res._id = obj._id;
    if (!_idProjection && _.has(res, '_id'))
      delete res._id;
    return res;
  };
};

// Traverses the keys of passed projection and constructs a tree where all
// leaves are either all True or all False
// @returns Object:
//  - tree - Object - tree representation of keys involved in projection
//  (exception for '_id' as it is a special case handled separately)
//  - including - Boolean - "take only certain fields" type of projection
const projectionDetails = function (fields) {
  // Find the non-_id keys (_id is handled specially because it is included unless
  // explicitly excluded). Sort the keys, so that our code to detect overlaps
  // like 'foo' and 'foo.bar' can assume that 'foo' comes first.
  var fieldsKeys = _.keys(fields).sort();

  // If _id is the only field in the projection, do not remove it, since it is
  // required to determine if this is an exclusion or exclusion. Also keep an
  // inclusive _id, since inclusive _id follows the normal rules about mixing
  // inclusive and exclusive fields. If _id is not the only field in the
  // projection and is exclusive, remove it so it can be handled later by a
  // special case, since exclusive _id is always allowed.
  if (fieldsKeys.length > 0 &&
      !(fieldsKeys.length === 1 && fieldsKeys[0] === '_id') &&
      !(_.includes(fieldsKeys, '_id') && fields['_id']))
    fieldsKeys = _.reject(fieldsKeys, function (key) { return key === '_id'; });

  var including = null; // Unknown

  _.each(fieldsKeys, function (keyPath) {
    var rule = !!fields[keyPath];
    if (including === null)
      including = rule;
    if (including !== rule)
      // This error message is copied from MongoDB shell
      throw MinimongoError("You cannot currently mix including and excluding fields.");
  });


  var projectionRulesTree = pathsToTree(
    fieldsKeys,
    function (path) { return including; },
    function (node, path, fullPath) {
      // Check passed projection fields' keys: If you have two rules such as
      // 'foo.bar' and 'foo.bar.baz', then the result becomes ambiguous. If
      // that happens, there is a probability you are doing something wrong,
      // framework should notify you about such mistake earlier on cursor
      // compilation step than later during runtime.  Note, that real mongo
      // doesn't do anything about it and the later rule appears in projection
      // project, more priority it takes.
      //
      // Example, assume following in mongo shell:
      // > db.coll.insert({ a: { b: 23, c: 44 } })
      // > db.coll.find({}, { 'a': 1, 'a.b': 1 })
      // { "_id" : ObjectId("520bfe456024608e8ef24af3"), "a" : { "b" : 23 } }
      // > db.coll.find({}, { 'a.b': 1, 'a': 1 })
      // { "_id" : ObjectId("520bfe456024608e8ef24af3"), "a" : { "b" : 23, "c" : 44 } }
      //
      // Note, how second time the return set of keys is different.

      var currentPath = fullPath;
      var anotherPath = path;
      throw MinimongoError("both " + currentPath + " and " + anotherPath +
                           " found in fields option, using both of them may trigger " +
                           "unexpected behavior. Did you mean to use only one of them?");
    });

  return {
    tree: projectionRulesTree,
    including: including
  };
};

// paths - Array: list of mongo style paths
// newLeafFn - Function: of form function(path) should return a scalar value to
//                       put into list created for that path
// conflictFn - Function: of form function(node, path, fullPath) is called
//                        when building a tree path for 'fullPath' node on
//                        'path' was already a leaf with a value. Must return a
//                        conflict resolution.
// initial tree - Optional Object: starting tree.
// @returns - Object: tree represented as a set of nested objects
const pathsToTree = function (paths, newLeafFn, conflictFn, tree) {
  tree = tree || {};
  _.each(paths, function (keyPath) {
    var treePos = tree;
    var pathArr = keyPath.split('.');

    // use _.every just for iteration with break
    var success = _.every(pathArr.slice(0, -1), function (key, idx) {
      if (!_.has(treePos, key))
        treePos[key] = {};
      else if (!_.isObject(treePos[key])) {
        treePos[key] = conflictFn(treePos[key],
                                  pathArr.slice(0, idx + 1).join('.'),
                                  keyPath);
        // break out of loop if we are failing for this path
        if (!_.isObject(treePos[key]))
          return false;
      }

      treePos = treePos[key];
      return true;
    });

    if (success) {
      var lastKey = _.last(pathArr);
      if (!_.has(treePos, lastKey))
        treePos[lastKey] = newLeafFn(keyPath);
      else
        treePos[lastKey] = conflictFn(treePos[lastKey], keyPath, keyPath);
    }
  });

  return tree;
};

 CollectionHelpers._checkSupportedProjection = function (fields) {
  if (!_.isObject(fields) || _.isArray(fields))
    throw MinimongoError("fields option must be an object");

  _.each(fields, function (val, keyPath) {
    if (_.includes(keyPath.split('.'), '$'))
      throw MinimongoError("Minimongo doesn't support $ operator in projections yet.");
    if (typeof val === 'object' && _.intersection(['$elemMatch', '$meta', '$slice'], _.keys(val)).length > 0)
      throw MinimongoError("Minimongo doesn't support operators in projections yet.");
    if (_.indexOf([1, 0, true, false], val) === -1)
      throw MinimongoError("Projection values should be one of 1, 0, true, or false");
  });
};


// XXX maybe move these into another ObserveHelpers package or something

// _CachingChangeObserver is an object which receives observeChanges callbacks
// and keeps a cache of the current cursor state up to date in self.docs. Users
// of this class should read the docs field but not modify it. You should pass
// the "applyChange" field as the callbacks to the underlying observeChanges
// call. Optionally, you can specify your own observeChanges callbacks which are
// invoked immediately before the docs field is updated; this object is made
// available as `this` to those callbacks.
CollectionHelpers._CachingChangeObserver = function (options) {
  var self = this;
  options = options || {};

  var orderedFromCallbacks = options.callbacks &&
        CollectionHelpers._observeChangesCallbacksAreOrdered(options.callbacks);
  if (_.has(options, 'ordered')) {
    self.ordered = options.ordered;
    if (options.callbacks && options.ordered !== orderedFromCallbacks)
      throw Error("ordered option doesn't match callbacks");
  } else if (options.callbacks) {
    self.ordered = orderedFromCallbacks;
  } else {
    throw Error("must provide ordered or callbacks");
  }
  var callbacks = options.callbacks || {};

  if (self.ordered) {
    self.docs = new OrderedDict(MongoID.idStringify);
    self.applyChange = {
      addedBefore: function (id, fields, before) {
        var doc = EJSON.clone(fields);
        doc._id = id;
        callbacks.addedBefore && callbacks.addedBefore.call(
          self, id, fields, before);
        // This line triggers if we provide added with movedBefore.
        callbacks.added && callbacks.added.call(self, id, fields);
        // XXX could `before` be a falsy ID?  Technically
        // idStringify seems to allow for them -- though
        // OrderedDict won't call stringify on a falsy arg.
        self.docs.putBefore(id, doc, before || null);
      },
      movedBefore: function (id, before) {
        var doc = self.docs.get(id);
        callbacks.movedBefore && callbacks.movedBefore.call(self, id, before);
        self.docs.moveBefore(id, before || null);
      }
    };
  } else {
    self.docs = new IdMap();
    self.applyChange = {
      added: function (id, fields) {
        var doc = EJSON.clone(fields);
        callbacks.added && callbacks.added.call(self, id, fields);
        doc._id = id;
        self.docs.set(id,  doc);
      }
    };
  }

  // The methods in _IdMap and OrderedDict used by these callbacks are
  // identical.
  self.applyChange.changed = function (id, fields) {
    var doc = self.docs.get(id);
    if (!doc)
      throw new Error("Unknown id for changed: " + id);
    callbacks.changed && callbacks.changed.call(
      self, id, EJSON.clone(fields));
    DiffSequence.applyChanges(doc, fields);
  };
  self.applyChange.removed = function (id) {
    callbacks.removed && callbacks.removed.call(self, id);
    self.docs.remove(id);
  };
};

CollectionHelpers._observeFromObserveChanges = function (cursor, observeCallbacks) {
  var transform = cursor.getTransform() || function (doc) {return doc;};
  var suppressed = !!observeCallbacks._suppress_initial;

  var observeChangesCallbacks;
  if (CollectionHelpers._observeCallbacksAreOrdered(observeCallbacks)) {
    // The "_no_indices" option sets all index arguments to -1 and skips the
    // linear scans required to generate them.  This lets observers that don't
    // need absolute indices benefit from the other features of this API --
    // relative order, transforms, and applyChanges -- without the speed hit.
    var indices = !observeCallbacks._no_indices;
    observeChangesCallbacks = {
      addedBefore: function (id, fields, before) {
        var self = this;
        if (suppressed || !(observeCallbacks.addedAt || observeCallbacks.added))
          return;
        var doc = transform(_.extend(fields, {_id: id}));
        if (observeCallbacks.addedAt) {
          var index = indices
                ? (before ? self.docs.indexOf(before) : self.docs.size()) : -1;
          observeCallbacks.addedAt(doc, index, before);
        } else {
          observeCallbacks.added(doc);
        }
      },
      changed: function (id, fields) {
        var self = this;
        if (!(observeCallbacks.changedAt || observeCallbacks.changed))
          return;
        var doc = EJSON.clone(self.docs.get(id));
        if (!doc)
          throw new Error("Unknown id for changed: " + id);
        var oldDoc = transform(EJSON.clone(doc));
        DiffSequence.applyChanges(doc, fields);
        doc = transform(doc);
        if (observeCallbacks.changedAt) {
          var index = indices ? self.docs.indexOf(id) : -1;
          observeCallbacks.changedAt(doc, oldDoc, index);
        } else {
          observeCallbacks.changed(doc, oldDoc);
        }
      },
      movedBefore: function (id, before) {
        var self = this;
        if (!observeCallbacks.movedTo)
          return;
        var from = indices ? self.docs.indexOf(id) : -1;

        var to = indices
              ? (before ? self.docs.indexOf(before) : self.docs.size()) : -1;
        // When not moving backwards, adjust for the fact that removing the
        // document slides everything back one slot.
        if (to > from)
          --to;
        observeCallbacks.movedTo(transform(EJSON.clone(self.docs.get(id))),
                                 from, to, before || null);
      },
      removed: function (id) {
        var self = this;
        if (!(observeCallbacks.removedAt || observeCallbacks.removed))
          return;
        // technically maybe there should be an EJSON.clone here, but it's about
        // to be removed from self.docs!
        var doc = transform(self.docs.get(id));
        if (observeCallbacks.removedAt) {
          var index = indices ? self.docs.indexOf(id) : -1;
          observeCallbacks.removedAt(doc, index);
        } else {
          observeCallbacks.removed(doc);
        }
      }
    };
  } else {
    observeChangesCallbacks = {
      added: function (id, fields) {
        if (!suppressed && observeCallbacks.added) {
          var doc = _.extend(fields, {_id:  id});
          observeCallbacks.added(transform(doc));
        }
      },
      changed: function (id, fields) {
        var self = this;
        if (observeCallbacks.changed) {
          var oldDoc = self.docs.get(id);
          var doc = EJSON.clone(oldDoc);
          DiffSequence.applyChanges(doc, fields);
          observeCallbacks.changed(transform(doc),
                                   transform(EJSON.clone(oldDoc)));
        }
      },
      removed: function (id) {
        var self = this;
        if (observeCallbacks.removed) {
          observeCallbacks.removed(transform(self.docs.get(id)));
        }
      }
    };
  }

  var changeObserver = new CollectionHelpers._CachingChangeObserver(
    {callbacks: observeChangesCallbacks});
  var handle = cursor.observeChanges(changeObserver.applyChange);
  suppressed = false;

  return handle;
};



export default CollectionHelpers;