/* eslint-disable */
import * as _ from 'lodash';
import * as EJSON from 'ddp-ejson';

import Tracker from './tracker'
import IdMap from './id-map'
import SynchronousQueue from './synchronous-queue'
import MongoId from './mongo-id'
import CollectionHelpers from './collection-helpers';
import { MinimongoError } from './minimongo';
import Sorter from './sorter';
import Matcher from './matcher';

// don't call this ctor directly.  use LocalCollection.find().
class Cursor {
  constructor(collection, selector, options = {}) {
    this.collection = collection
    this.sorter = null
    this.matcher = new Matcher(selector)

    if (CollectionHelpers._selectorIsId(selector)) {
      // stash for fast path
      this._selectorId = selector
    } else if (CollectionHelpers._selectorIsIdPerhapsAsObject(selector)) {
      // also do the fast path for { _id: idString }
      this._selectorId = selector._id
    } else {
      this._selectorId = undefined
      if (this.matcher.hasGeoQuery() || options.sort) {
        this.sorter = new Sorter(options.sort || [], {
          matcher: this.matcher
        })
      }
    }

    this.skip = options.skip
    this.limit = options.limit
    this.fields = options.fields

    this._projectionFn = CollectionHelpers._compileProjection(this.fields || {})

    this._transform = CollectionHelpers.wrapTransform(options.transform)

    // by default, queries register w/ Tracker when it is available.
    if (typeof Tracker !== 'undefined')
      this.reactive = (options.reactive === undefined) ? true : options.reactive
  }




  /**
   * @callback IterationCallback
   * @param {Object} doc
   * @param {Number} index
   */
  /**
   * @summary Call `callback` once for each matching document, sequentially and synchronously.
   * @locus Anywhere
   * @method  forEach
   * @instance
   * @memberOf Mongo.Cursor
   * @param {IterationCallback} callback Function to call. It will be called with three arguments: the document, a 0-based index, and <em>cursor</em> itthis.
   * @param {Any} [thisArg] An object which will be the value of `this` inside `callback`.
   */
  forEach(callback, thisArg) {
    var objects = this._getRawObjects({
      ordered: true
    })

    if (this.reactive) {
      this._depend({
        addedBefore: true,
        removed: true,
        changed: true,
        movedBefore: true
      })
    }

    _.each(objects, (elt, i) => {
      // This doubles as a clone operation.
      elt = this._projectionFn(elt)

      if (this._transform)
        elt = this._transform(elt)
      callback.call(thisArg, elt, i, this)
    })
  }

  getTransform() {
    return this._transform
  }

  /**
   * @summary Map callback over all matching documents.  Returns an Array.
   * @locus Anywhere
   * @method map
   * @instance
   * @memberOf Mongo.Cursor
   * @param {IterationCallback} callback Function to call. It will be called with three arguments: the document, a 0-based index, and <em>cursor</em> itthis.
   * @param {Any} [thisArg] An object which will be the value of `this` inside `callback`.
   */
  map(callback, thisArg) {
    var res = []
    this.forEach((doc, index) => {
      res.push(callback.call(thisArg, doc, index, this))
    })
    return res
  }

  /**
   * @summary Return all matching documents as an Array.
   * @memberOf Mongo.Cursor
   * @method  fetch
   * @instance
   * @locus Anywhere
   * @returns {Object[]}
   */
  fetch() {
    var res = []
    this.forEach(function (doc) {
      res.push(doc)
    })
    return res
  }

  /**
   * @summary Returns the number of documents that match a query.
   * @memberOf Mongo.Cursor
   * @method  count
   * @instance
   * @locus Anywhere
   * @returns {Number}
   */
  count() {
    if (this.reactive)
      this._depend({
          added: true,
          removed: true
        },
        true /* allow the observe to be unordered */ )

    return this._getRawObjects({
      ordered: true
    }).length
  }



  _getCollectionName() {
    return this.collection.name
  }

  /**
   * @summary Watch a query.  Receive callbacks as the result set changes.
   * @locus Anywhere
   * @memberOf Mongo.Cursor
   * @instance
   * @param {Object} callbacks Functions to call to deliver the result set as it changes
   */
  observe(options) {
    return CollectionHelpers._observeFromObserveChanges(this, options)
  }

  /**
   * @summary Watch a query.  Receive callbacks as the result set changes.  Only the differences between the old and new documents are passed to the callbacks.
   * @locus Anywhere
   * @memberOf Mongo.Cursor
   * @instance
   * @param {Object} callbacks Functions to call to deliver the result set as it changes
   */
  observeChanges(options) {
    var ordered = CollectionHelpers._observeChangesCallbacksAreOrdered(options)

    // there are several places that assume you aren't combining skip/limit with
    // unordered observe.  eg, update's EJSON.clone, and the "there are several"
    // comment in _modifyAndNotify
    // XXX allow skip/limit with unordered observe
    if (!options._allow_unordered && !ordered && (this.skip || this.limit))
      throw new Error("must use ordered observe (ie, 'addedBefore' instead of 'added') with skip or limit")

    if (this.fields && (this.fields._id === 0 || this.fields._id === false))
      throw Error('You may not observe a cursor with {fields: {_id: 0}}')

    var query = {
      dirty: false,
      matcher: this.matcher, // not fast pathed
      sorter: ordered && this.sorter,
      distances: (
        this.matcher.hasGeoQuery() && ordered && new IdMap()),
      resultsSnapshot: null,
      ordered: ordered,
      cursor: this,
      projectionFn: this._projectionFn
    }
    var qid

    // Non-reactive queries call added[Before] and then never call anything
    // else.
    if (this.reactive) {
      qid = this.collection.next_qid++
        this.collection.queries[qid] = query
    }
    query.results = this._getRawObjects({
      ordered: ordered,
      distances: query.distances
    })
    if (this.collection.paused)
      query.resultsSnapshot = (ordered ? [] : new IdMap())

 
    // wrap callbacks we were passed. callbacks only fire when not paused and
    // are never undefined
    // Filters out blacklisted fields according to cursor's projection.
    // XXX wrong place for this?

    // furthermore, callbacks enqueue until the operation we're working on is
    // done.
    var wrapCallback = (f) => {
      var self = this;
      if (!f)
        return function () {}
      return function ( /*args*/ ) {
        var context = this
        var args = arguments

        if (self.collection.paused)
          return

        self.collection._observeQueue.queueTask(function () {
          f.apply(context, args)
        })
      }
    }
    query.added = wrapCallback(options.added)
    query.changed = wrapCallback(options.changed)
    query.removed = wrapCallback(options.removed)
    if (ordered) {
      query.addedBefore = wrapCallback(options.addedBefore)
      query.movedBefore = wrapCallback(options.movedBefore)
    }

    if (!options._suppress_initial && !this.collection.paused) {
      // XXX unify ordered and unordered interface
      var each = ordered ?
        _.bind(_.each, null, query.results) :
        _.bind(query.results.forEach, query.results)
      each( (doc) => {
        var fields = EJSON.clone(doc)

        delete fields._id
        if (ordered)
          query.addedBefore(doc._id, this._projectionFn(fields), null)
        query.added(doc._id, this._projectionFn(fields))
      })
    }

    var handle = new CollectionHelpers.ObserveHandle
    _.extend(handle, {
      collection: this.collection,
      stop: () => {
        if (this.reactive)
          delete this.collection.queries[qid]
      }
    })

    if (this.reactive && Tracker.active) {
      // XXX in many cases, the same observe will be recreated when
      // the current autorun is rerun.  we could save work by
      // letting it linger across rerun and potentially get
      // repurposed if the same observe is performed, using logic
      // similar to that of subscribe.
      Tracker.onInvalidate(function () {
        handle.stop()
      })
    }
    // run the observe callbacks resulting from the initial contents
    // before we leave the observe.
    this.collection._observeQueue.drain()

    return handle
  }

  // options to contain:
  //  * callbacks for observe():
  //    - addedAt (document, atIndex)
  //    - added (document)
  //    - changedAt (newDocument, oldDocument, atIndex)
  //    - changed (newDocument, oldDocument)
  //    - removedAt (document, atIndex)
  //    - removed (document)
  //    - movedTo (document, oldIndex, newIndex)
  //
  // attributes available on returned query handle:
  //  * stop(): end updates
  //  * collection: the collection this query is querying
  //
  // iff x is a returned query handle, (x instanceof
  // CollectionHelpers.ObserveHandle) is true
  //
  // initial results delivered through added callback
  // XXX maybe callbacks should take a list of objects, to expose transactions?
  // XXX maybe support field limiting (to limit what you're notified on)




  // Returns a collection of matching objects, but doesn't deep copy them.
  //
  // If ordered is set, returns a sorted array, respecting sorter, skip, and limit
  // properties of the query.  if sorter is falsey, no sort -- you get the natural
  // order.
  //
  // If ordered is not set, returns an object mapping from ID to doc (sorter, skip
  // and limit should not be set).
  //
  // If ordered is set and this cursor is a $near geoquery, then this function
  // will use an _IdMap to track each distance from the $near argument point in
  // order to use it as a sort key. If an _IdMap is passed in the 'distances'
  // argument, this function will clear it and use it for this purpose (otherwise
  // it will just create its own _IdMap). The observeChanges implementation uses
  // this to remember the distances after this function returns.
  _getRawObjects(options = {}) {


    // XXX use OrderedDict instead of array, and make IdMap and OrderedDict
    // compatible
    var results = options.ordered ? [] : new IdMap()

    // fast path for single ID value
    if (this._selectorId !== undefined) {
      // If you have non-zero skip and ask for a single id, you get
      // nothing. This is so it matches the behavior of the '{_id: foo}'
      // path.
      if (this.skip)
        return results

      var selectedDoc = this.collection._docs.get(this._selectorId)
      if (selectedDoc) {
        if (options.ordered)
          results.push(selectedDoc)
        else
          results.set(this._selectorId, selectedDoc)
      }
      return results
    }

    // slow path for arbitrary selector, sort, skip, limit

    // in the observeChanges case, distances is actually part of the "query" (ie,
    // live results set) object.  in other cases, distances is only used inside
    // this function.
    var distances
    if (this.matcher.hasGeoQuery() && options.ordered) {
      if (options.distances) {
        distances = options.distances
        distances.clear()
      } else {
        distances = new IdMap()
      }
    }

    this.collection._docs.forEach( (doc, id) => {

      var matchResult = this.matcher.documentMatches(doc)
      if (matchResult.result) {
        if (options.ordered) {
          results.push(doc)
          if (distances && matchResult.distance !== undefined)
            distances.set(id, matchResult.distance)
        } else {
          results.set(id, doc)
        }
      }
      // Fast path for limited unsorted queries.
      // XXX 'length' check here seems wrong for ordered
      if (this.limit && !this.skip && !this.sorter &&
        results.length === this.limit)
        return false // break
      return true // continue
    })

    if (!options.ordered)
      return results

    if (this.sorter) {
      var comparator = this.sorter.getComparator({
        distances: distances
      })
      results.sort(comparator)
    }

    var idx_start = this.skip || 0
    var idx_end = this.limit ? (this.limit + idx_start) : results.length
    return results.slice(idx_start, idx_end)
  }

  // XXX Maybe we need a version of observe that just calls a callback if
  // anything changed.
  _depend(changers, _allow_unordered) {
    if (Tracker.active) {
      var v = new Tracker.Dependency
      v.depend()
      var notifyChange = _.bind(v.changed, v)

      var options = {
        _suppress_initial: true,
        _allow_unordered: _allow_unordered
      }
      _.each(['added', 'changed', 'removed', 'addedBefore', 'movedBefore'],
        function (fnName) {
          if (changers[fnName])
            options[fnName] = notifyChange
        })

      // observeChanges will stop() when this computation is invalidated
      this.observeChanges(options)
    }
  }
}

export default Cursor;