/* eslint-disable */
import * as EJSON from 'ddp-ejson';
import * as _ from 'lodash';

import Tracker from './tracker'
import IdMap from './id-map'
import SynchronousQueue from './synchronous-queue'
import MongoID from './mongo-id'
import CollectionHelpers from './collection-helpers';
import Cursor from './cursor';
import { MinimongoError } from './minimongo';
import Matcher from './matcher';
import Random from './random';
import _modify from './modify';

// XXX type checking on selectors (graceful error if malformed)

// LocalCollection: a set of documents that supports queries and modifiers.

// Cursor: a specification for a particular subset of documents, w/
// a defined order, limit, and offset.  creating a Cursor with LocalCollection.find(),

// ObserveHandle: the return value of a live query.

// XXX type checking on selectors (graceful error if malformed)

// LocalCollection: a set of documents that supports queries and modifiers.

// Cursor: a specification for a particular subset of documents, w/
// a defined order, limit, and offset.  creating a Cursor with LocalCollection.find(),

// ObserveHandle: the return value of a live query.

class LocalCollection {
  constructor(name) {
    this.name = name
    this._docs = new IdMap()
    this._observeQueue = new SynchronousQueue()
    this.next_qid = 1 // live query id generator

    // qid -> live query object. keys:
    //  ordered: bool. ordered queries have addedBefore/movedBefore callbacks.
    //  results: array (ordered) or object (unordered) of current results
    //    (aliased with this._docs!)
    //  resultsSnapshot: snapshot of results. null if not paused.
    //  cursor: Cursor object for the query.
    //  selector, sorter, (callbacks): functions
    this.queries = {}

    // null if not saving originals; an IdMap from id to original document value if
    // saving originals. See comments before saveOriginals().
    this._savedOriginals = null

    // True when observers are paused and we should not send callbacks.
    this.paused = false
  }

  // options may include sort, skip, limit, reactive
  // sort may be any of these forms:
  //     {a: 1, b: -1}
  //     [["a", "asc"], ["b", "desc"]]
  //     ["a", ["b", "desc"]]
  //   (in the first form you're beholden to key enumeration order in
  //   your javascript VM)
  //
  // reactive: if given, and false, don't register with Tracker (default
  // is true)
  //
  // XXX possibly should support retrieving a subset of fields? and
  // have it be a hint (ignored on the client, when not copying the
  // doc?)
  //
  // XXX sort does not yet support subkeys ('a.b') .. fix that!
  // XXX add one more sort form: "key"
  // XXX tests
  find(selector, options) {
    // default syntax for everything is to omit the selector argument.
    // but if selector is explicitly passed in as false or undefined, we
    // want a selector that matches nothing.
    if (arguments.length === 0)
      selector = {}

    return new Cursor(this, selector, options)
  }

  findOne (selector, options) {
    if (arguments.length === 0)
      selector = {}

    // NOTE: by setting limit 1 here, we end up using very inefficient
    // code that recomputes the whole query on each update. The upside is
    // that when you reactively depend on a findOne you only get
    // invalidated when the found object changes, not any object in the
    // collection. Most findOne will be by id, which has a fast path, so
    // this might not be a big deal. In most cases, invalidation causes
    // the called to re-query anyway, so this should be a net performance
    // improvement.
    options = options || {};
    options.limit = 1

    return this.find(selector, options).fetch()[0]
  }


  // XXX enforce rule that field names can't start with '$' or contain '.'
  // (real mongodb does in fact enforce this)
  // XXX possibly enforce that 'undefined' does not appear (we assume
  // this in our handling of null and $exists)
  insert(doc, callback) {
    doc = EJSON.clone(doc)

    // Make sure field names do not contain Mongo restricted
    // characters ('.', '$', '\0').
    // https://docs.mongodb.com/manual/reference/limits/#Restrictions-on-Field-Names
    if (doc) {
      const invalidCharMsg = {
        '.': "contain '.'",
        '$': "start with '$'",
        '\0': 'contain null bytes'
      }
      JSON.stringify(doc, (key, value) => {
        let match
        if (_.isString(key) && (match = key.match(/^\$|\.|\0/))) {
          throw MinimongoError(`Key ${key} must not ${invalidCharMsg[match[0]]}`)
        }
        return value
      })
    }

    if (!_.has(doc, '_id')) {
      // if you really want to use ObjectIDs, set this global.
      // Mongo.Collection specifies its own ids and does not use this code.
      doc._id = CollectionHelpers._useOID ? new MongoID.ObjectID() :
        Random.id()
    }
    var id = doc._id

    if (this._docs.has(id))
      throw MinimongoError("Duplicate _id '" + id + "'")

    this._saveOriginal(id, undefined)
    this._docs.set(id, doc)

    var queriesToRecompute = []
    // trigger live queries that match
    for (var qid in this.queries) {
      var query = this.queries[qid]
      if (query.dirty) continue
      var matchResult = query.matcher.documentMatches(doc)
      if (matchResult.result) {
        if (query.distances && matchResult.distance !== undefined)
          query.distances.set(id, matchResult.distance)
        if (query.cursor.skip || query.cursor.limit)
          queriesToRecompute.push(qid)
        else
          CollectionHelpers._insertInResults(query, doc)
      }
    }

    _.each(queriesToRecompute, (qid) => {
      if (this.queries[qid])
        this._recomputeResults(this.queries[qid])
    })
    this._observeQueue.drain()

    // Defer because the caller likely doesn't expect the callback to be run
    // immediately.
    if (callback)
      Meteor.defer(function () {
        callback(null, id)
      })
    return id
  }

  // Iterates over a subset of documents that could match selector; calls
  // f(doc, id) on each of them.  Specifically, if selector specifies
  // specific _id's, it only looks at those.  doc is *not* cloned: it is the
  // same object that is in _docs.
  _eachPossiblyMatchingDoc(selector, f) {
    var specificIds = CollectionHelpers._idsMatchedBySelector(selector)
    if (specificIds) {
      for (var i = 0; i < specificIds.length; ++i) {
        var id = specificIds[i]
        var doc = this._docs.get(id)
        if (doc) {
          var breakIfFalse = f(doc, id)
          if (breakIfFalse === false)
            break
        }
      }
    } else {
      this._docs.forEach(f)
    }
  }

  remove(selector, callback) {

    // Easy special case: if we're not calling observeChanges callbacks and we're
    // not saving originals and we got asked to remove everything, then just empty
    // everything directly.
    if (this.paused && !this._savedOriginals && EJSON.equals(selector, {})) {
      var result = this._docs.size
      this._docs.clear()
      _.each(this.queries, function (query) {
        if (query.ordered) {
          query.results = []
        } else {
          query.results.clear()
        }
      })
      if (callback) {
        Meteor.defer(function () {
          callback(null, result)
        })
      }
      return result
    }

    var matcher = new Matcher(selector)
    var remove = []
    this._eachPossiblyMatchingDoc(selector, function (doc, id) {
      if (matcher.documentMatches(doc).result)
        remove.push(id)
    })

    var queriesToRecompute = []
    var queryRemove = []
    for (var i = 0; i < remove.length; i++) {
      var removeId = remove[i]
      var removeDoc = this._docs.get(removeId)
      _.each(this.queries, function (query, qid) {
        if (query.dirty) return

        if (query.matcher.documentMatches(removeDoc).result) {
          if (query.cursor.skip || query.cursor.limit)
            queriesToRecompute.push(qid)
          else
            queryRemove.push({
              qid: qid,
              doc: removeDoc
            })
        }
      })
      this._saveOriginal(removeId, removeDoc)
      this._docs.remove(removeId)
    }

    // run live query callbacks _after_ we've removed the documents.
    _.each(queryRemove, (remove) => {
      var query = this.queries[remove.qid]
      if (query) {
        query.distances && query.distances.remove(remove.doc._id)
        CollectionHelpers._removeFromResults(query, remove.doc)
      }
    })
    _.each(queriesToRecompute, (qid) => {
      var query = this.queries[qid]
      if (query)
        this._recomputeResults(query)
    })
    this._observeQueue.drain()
    result = remove.length
    if (callback)
      Meteor.defer(function () {
        callback(null, result)
      })
    return result
  }

  // XXX atomicity: if multi is true, and one modification fails, do
  // we rollback the whole operation, or what?
  update(selector, mod, options, callback) {
    if (!callback && options instanceof Function) {
      callback = options
      options = null
    }
    if (!options) options = {}

    var matcher = new Matcher(selector)

    // Save the original results of any query that we might need to
    // _recomputeResults on, because _modifyAndNotify will mutate the objects in
    // it. (We don't need to save the original results of paused queries because
    // they already have a resultsSnapshot and we won't be diffing in
    // _recomputeResults.)
    var qidToOriginalResults = {}
    // We should only clone each document once, even if it appears in multiple queries
    var docMap = new IdMap()
    var idsMatchedBySelector = CollectionHelpers._idsMatchedBySelector(selector)

    _.each(this.queries, (query, qid) => {
      if ((query.cursor.skip || query.cursor.limit) && !this.paused) {
        // Catch the case of a reactive `count()` on a cursor with skip
        // or limit, which registers an unordered observe. This is a
        // pretty rare case, so we just clone the entire result set with
        // no optimizations for documents that appear in these result
        // sets and other queries.
        if (query.results instanceof IdMap) {
          qidToOriginalResults[qid] = query.results.clone()
          return
        }

        if (!(query.results instanceof Array)) {
          throw new Error('Assertion failed: query.results not an array')
        }

        // Clones a document to be stored in `qidToOriginalResults`
        // because it may be modified before the new and old result sets
        // are diffed. But if we know exactly which document IDs we're
        // going to modify, then we only need to clone those.
        var memoizedCloneIfNeeded = function (doc) {
          if (docMap.has(doc._id)) {
            return docMap.get(doc._id)
          } else {
            var docToMemoize

            if (idsMatchedBySelector && !_.some(idsMatchedBySelector, function (id) {
                return EJSON.equals(id, doc._id)
              })) {
              docToMemoize = doc
            } else {
              docToMemoize = EJSON.clone(doc)
            }

            docMap.set(doc._id, docToMemoize)
            return docToMemoize
          }
        }

        qidToOriginalResults[qid] = query.results.map(memoizedCloneIfNeeded)
      }
    })
    var recomputeQids = {}

    var updateCount = 0

    this._eachPossiblyMatchingDoc(selector, (doc, id) => {
      var queryResult = matcher.documentMatches(doc)
      if (queryResult.result) {
        // XXX Should we save the original even if mod ends up being a no-op?
        this._saveOriginal(id, doc)
        this._modifyAndNotify(doc, mod, recomputeQids, queryResult.arrayIndices)
          ++updateCount
        if (!options.multi)
          return false // break
      }
      return true
    })

    _.each(recomputeQids, (dummy, qid) => {
      var query = this.queries[qid]
      if (query)
        this._recomputeResults(query, qidToOriginalResults[qid])
    })
    this._observeQueue.drain()

    // If we are doing an upsert, and we didn't modify any documents yet, then
    // it's time to do an insert. Figure out what document we are inserting, and
    // generate an id for it.
    var insertedId
    if (updateCount === 0 && options.upsert) {
      var newDoc = CollectionHelpers._removeDollarOperators(selector)
      _modify(newDoc, mod, {
        isInsert: true
      })
      if (!newDoc._id && options.insertedId)
        newDoc._id = options.insertedId
      insertedId = this.insert(newDoc)
      updateCount = 1
    }

    // Return the number of affected documents, or in the upsert case, an object
    // containing the number of affected docs and the id of the doc that was
    // inserted, if any.
    var result
    if (options._returnObject) {
      result = {
        numberAffected: updateCount
      }
      if (insertedId !== undefined)
        result.insertedId = insertedId
    } else {
      result = updateCount
    }

    if (callback)
      Meteor.defer(function () {
        callback(null, result)
      })
    return result
  }

  // A convenience wrapper on update. LocalCollection.upsert(sel, mod) is
  // equivalent to LocalCollection.update(sel, mod, { upsert: true, _returnObject:
  // true }).
  upsert(selector, mod, options, callback) {
    if (!callback && typeof options === 'function') {
      callback = options
      options = {}
    }
    return this.update(selector, mod, _.extend({}, options, {
      upsert: true,
      _returnObject: true
    }), callback)
  }

  _modifyAndNotify(doc, mod, recomputeQids, arrayIndices) {
    var matched_before = {}
    for (var qid in this.queries) {
      var query = this.queries[qid]
      if (query.dirty) continue

      if (query.ordered) {
        matched_before[qid] = query.matcher.documentMatches(doc).result
      } else {
        // Because we don't support skip or limit (yet) in unordered queries, we
        // can just do a direct lookup.
        matched_before[qid] = query.results.has(doc._id)
      }
    }

    var old_doc = EJSON.clone(doc)

    _modify(doc, mod, {
      arrayIndices: arrayIndices
    })

    for (qid in this.queries) {
      query = this.queries[qid]
      if (query.dirty) continue

      var before = matched_before[qid]
      var afterMatch = query.matcher.documentMatches(doc)
      var after = afterMatch.result
      if (after && query.distances && afterMatch.distance !== undefined)
        query.distances.set(doc._id, afterMatch.distance)

      if (query.cursor.skip || query.cursor.limit) {
        // We need to recompute any query where the doc may have been in the
        // cursor's window either before or after the update. (Note that if skip
        // or limit is set, "before" and "after" being true do not necessarily
        // mean that the document is in the cursor's output after skip/limit is
        // applied... but if they are false, then the document definitely is NOT
        // in the output. So it's safe to skip recompute if neither before or
        // after are true.)
        if (before || after)
          recomputeQids[qid] = true
      } else if (before && !after) {
        CollectionHelpers._removeFromResults(query, doc)
      } else if (!before && after) {
        CollectionHelpers._insertInResults(query, doc)
      } else if (before && after) {
        CollectionHelpers._updateInResults(query, doc, old_doc)
      }
    }
  }

  // Recomputes the results of a query and runs observe callbacks for the
  // difference between the previous results and the current results (unless
  // paused). Used for skip/limit queries.
  //
  // When this is used by insert or remove, it can just use query.results for the
  // old results (and there's no need to pass in oldResults), because these
  // operations don't mutate the documents in the collection. Update needs to pass
  // in an oldResults which was deep-copied before the modifier was applied.
  //
  // oldResults is guaranteed to be ignored if the query is not paused.
  _recomputeResults (query, oldResults) {
    if (this.paused) {
      // There's no reason to recompute the results now as we're still paused.
      // By flagging the query as "dirty", the recompute will be performed
      // when resumeObservers is called.
      query.dirty = true
      return
    }

    if (!this.paused && !oldResults)
      oldResults = query.results
    if (query.distances)
      query.distances.clear()
    query.results = query.cursor._getRawObjects({
      ordered: query.ordered,
      distances: query.distances
    })

    if (!this.paused) {
      CollectionHelpers._diffQueryChanges(
        query.ordered, oldResults, query.results, query, {
          projectionFn: query.projectionFn
        })
    }
  }

  // To track what documents are affected by a piece of code, call saveOriginals()
  // before it and retrieveOriginals() after it. retrieveOriginals returns an
  // object whose keys are the ids of the documents that were affected since the
  // call to saveOriginals(), and the values are equal to the document's contents
  // at the time of saveOriginals. (In the case of an inserted document, undefined
  // is the value.) You must alternate between calls to saveOriginals() and
  // retrieveOriginals().
  saveOriginals() {
    if (this._savedOriginals)
      throw new Error('Called saveOriginals twice without retrieveOriginals')
    this._savedOriginals = new IdMap()
  }

  retrieveOriginals() {
    if (!this._savedOriginals)
      throw new Error('Called retrieveOriginals without saveOriginals')

    var originals = this._savedOriginals
    this._savedOriginals = null
    return originals
  }

  _saveOriginal(id, doc) {

    // Are we even trying to save originals?
    if (!this._savedOriginals)
      return
    // Have we previously mutated the original (and so 'doc' is not actually
    // original)?  (Note the 'has' check rather than truth: we store undefined
    // here for inserted docs!)
    if (this._savedOriginals.has(id))
      return
    this._savedOriginals.set(id, EJSON.clone(doc))
  }

  // Pause the observers. No callbacks from observers will fire until
  // 'resumeObservers' is called.
  pauseObservers() {
    // No-op if already paused.
    if (this.paused)
      return

    // Set the 'paused' flag such that new observer messages don't fire.
    this.paused = true

    // Take a snapshot of the query results for each query.
    for (var qid in this.queries) {
      var query = this.queries[qid]

      query.resultsSnapshot = EJSON.clone(query.results)
    }
  }

  // Resume the observers. Observers immediately receive change
  // notifications to bring them to the current state of the
  // database. Note that this is not just replaying all the changes that
  // happened during the pause, it is a smarter 'coalesced' diff.
  resumeObservers() {

    // No-op if not paused.
    if (!this.paused)
      return

    // Unset the 'paused' flag. Make sure to do this first, otherwise
    // observer methods won't actually fire when we trigger them.
    this.paused = false

    for (var qid in this.queries) {
      var query = this.queries[qid]
      if (query.dirty) {
        query.dirty = false
        // re-compute results will perform `CollectionHelpers._diffQueryChanges` automatically.
        this._recomputeResults(query, query.resultsSnapshot)
      } else {
        // Diff the current results against the snapshot and send to observers.
        // pass the query object for its observer callbacks.
        CollectionHelpers._diffQueryChanges(
          query.ordered, query.resultsSnapshot, query.results, query, {
            projectionFn: query.projectionFn
          })
      }
      query.resultsSnapshot = null
    }
    this._observeQueue.drain()
  }

}

export default LocalCollection;