import * as _ from 'lodash';
import * as EJSON from 'ddp-ejson';

import { Service } from './service';
/**
 * Abstract cache class for Meteor and Cng cache to extend from.
 *
 * Note that Meteor works with Mongo Identifiers and CNG with a range of identifiers depending on the type of object.
 * (e.g. compositorId for compositions, channelId/deviceId for sources, ...)
 * And not all of these are in our control (e.g. compositorId 'screen' is fixed)
 *
 * Therefore depending on the object we save the mongo id in a field that's available
 * and that way we can map the objects.
 * For layouts this is the title, for sources this is the deviceId
 *
 * The cache stores its data in the `_data` property.
 * However, for ease-of-use it automatically defines getters and setters on all collections
 * Therefore, one can work with the cache as a regular object.
 * You can get the raw `_data` property using the `raw` getter (e.g. `myCache.raw`)
 */
class CacheManager extends Service {
  constructor(collections, base = {}) {
    super();
    this._base = { $arrays: {}, ...base };
    this._collections = collections;
    this.reset();

    // Define getters and setters on all collections for ease-of-use
    this._collections.concat(Object.keys(this._base)).forEach((key) => {
      Object.defineProperty(this, key, {
        get: () => this._data[key],
        set: (val) => { this._data[key] = val; },
      });
    });
  }

  /**
   * The cache stores its data in the `_data` property.
   * However, for ease-of-use it automatically defines getters and setters on all collections
   * @returns {*}
   */
  get data() {
    return this._data;
  }

  set data(data) {
    this._data = data;
  }

  async _onStart() {
    await this.warm();
    await super._onStart();
  }

  _onStop() {
    this.reset();
    return super._onStop();
  }

  warm() {}

  /**
   * Add an item to a certain collection by the given key
   * @param collection
   * @param key
   * @param item
   */
  add(collection, key, item) {
    if (this._data[collection][key]) {
      this.remove(collection, key);
    }

    this._data[collection][key] = item;
    this._data.$arrays[collection].push(item);
  }

  /**
   * Remove the object with given key from the requested collection
   * @param collection
   * @param key
   */
  remove(collection, key) {
    const exists = this._data[collection][key];
    if (exists) {
      const index = this._data.$arrays[collection].indexOf(exists);
      this._data.$arrays[collection].splice(index, 1);
      delete this._data[collection][key];
    }
  }

  /**
   * Check whether this cache matches with all given caches
   * @param caches list of caches to compare equality with
   * @returns {*}
   */
  equals(...caches) {
    const a = _.omit(this._data, '$arrays');
    return caches.reduce((prevEqual, cache) => prevEqual && _.isEqual(a, _.omit(cache.hasOwnProperty('data') ? cache.data : cache, '$arrays')), true);
  }

  /**
   * Reset (empty) all collections
   */
  reset(collections) {
    if (!collections) {
      collections = this._collections;
      this._data = EJSON.clone(this._base);
    }

    collections.forEach((collection) => {
      this._data[collection] = {};
      this._data.$arrays[collection] = [];
    });
  }

  /**
   * Clean-up
   */
  _onDestroy() {
    delete this._data;
    delete this._base;
    delete this._collections;

    return super._onDestroy();
  }
}

export default CacheManager;
