import EventEmitter from '../utils/eventEmitter';
import CancelError from './cancelError';

/**
 * Cancellation token
 * Inspired by C# dotnet 4.
 *
 * This is a way to stop/cancel an asynchronous task by passing a token that has a 'request to cancel' status
 * The asynchronous task may still finish what it was doing.
 * Therefore it's more a 'request to cancel' then a 'command to cancel immediately'
 *
 * Usually a task will check at certain points in its flow (or on every loop iteration) if its requested to be canceled
 *
 * The cancellation token can be created by a task creator who then has the control to stop the task.
 * Since it's the responsability of the task creator to cancel the task, the token itself has no cancel function.
 *
 * To create a token you'd usually call the static 'CancelToken.source' method,
 * which returns both a token and a cancel function.
 *
 * The token also leverages promises so you can combine tokens and race/wait for all tokens.
 * This is useful when you got parent-child tasks.
 * Usually you'd want to cancel the child task when cancelling the parent task, but not the other way around.
 *
 */
class CancelToken extends EventEmitter {
  constructor(executor) {
    super();

    if (typeof executor !== 'function') {
      throw new TypeError('must provide an executor function');
    }

    this._cancelled = false;

    // leverage promises to race / wait for all
    this._promise = new Promise((resolve) => {
      this.resolve = resolve;
    });

    executor(reason => this._cancel(reason));
  }

  _cancel(reason) {
    // Actually cancel with a certain reason
    if (this.requested) return;
    this._cancelled = true;
    this._cancelError = new CancelError(reason);
    this.resolve(this._cancelError);
    this.emit('cancelled');
    delete this.resolve;
    this.removeAllListeners('cancelled');
  }

  /**
   * Usual way to create a token/cancel action
   */
  static source() {
    let cancel;
    const token = new CancelToken((c) => { cancel = c; });

    return { token, cancel };
  }

  get requested() {
    return this._cancelled;
  }

  get promise() {
    return this._promise;
  }

  /**
   * Throw the cancel error with the cancel reason if we're cancelled
   */
  throwIfRequested() {
    if (this._cancelError) {
      throw this._cancelError;
    }
  }

  /**
   * Fork this token into a child token that can be cancelled individually or together with the parent token
   */
  fork() {
    const { token, cancel } = CancelToken.source();
    const forkedToken = this.concat(token);
    return { token: forkedToken, cancel };
  }

  /**
   * Race this token with the given token
   */
  concat(token) {
    return CancelToken.race([this, token]);
  }

  static all(tokens) {
    return new CancelToken((cancel) => {
      Promise.all(tokens.map(token => token.promise)).then(cancel);
    });
  }

  /**
   * Race the given array of tokens
   */
  static race(tokens) {
    return new CancelToken((cancel) => {
      Promise.race(tokens.map(token => token.promise)).then(cancel);
    });
  }

  static from(cancelTokenlike) {
    if (cancelTokenlike == null) return null;
    /* istanbul ignore else */
    if (cancelTokenlike instanceof CancelToken) return cancelTokenlike;
    throw new TypeError('not a CancelToken');
  }

  /**
   * Cleanup
   */
  destroy() {
    super.destroy();
    delete this._promise;
    delete this._resolve;
    delete this._cancelError;
  }
}
export default CancelToken;
