/**
 * (C) 2015 Copyright BARCO NV
 * All rights reserved. Reproduction in whole or in part is
 * prohibited without the written consent of the copyright owner.
 *
 * Part of Barco WeConnect
 */

import * as fs from 'fs-extra';
import * as path from 'path';
import { isFunction, isEqual } from 'lodash';
import * as getValue from 'get-value';
import * as isDev from 'electron-is-dev';

import { ROOT_DIR } from '../../files';
import { createDirectory } from '../../utils/utils';
import { removeAll, ensureLatest, writeSafely } from '../../utils/fs';
import ConfigUpgrader from './upgrader';
import LocalCollection from '../../minimongo/local-collection';
import ProxiedService from '../proxiedService';

// Build in a version number
const packageInfo = JSON.parse(fs.readFileSync(path.join(ROOT_DIR, 'package.json'), 'utf8'));
let proxyVersion = packageInfo.version;

// In dev mode we often have invalid semver versions (e.g. 1.2.1-0). Strip the -0
if (isDev) {
  proxyVersion = proxyVersion.replace(/-[\d]+/, '');
}

/**
 * The configuration manager does what you'd expect it to do: it manages configuration
 * Basically it reads and writes the configuration file in a secure way.
 *
 * Next to that it allows to watch changes on certain configuration settings.
 * This is done by leveraging minimongo for reactivity
 *
 * This also takes care of using the configuration upgrader to get the configuration
 * in the correct version.
 *
 * You can subscribe to updates using the `watch` function
 */
class ConfigurationManager extends ProxiedService {
  /**
   * Takes the location of the config file & the default config in case there's no config file
   * @param {*} param0
   */
  constructor({
    dir,
    file,
    defaultConfig = {},
  }) {
    super();

    if (dir) {
      dir = createDirectory(dir);
    }

    if (!dir) {
      dir = path.join(process.cwd(), '..');
    }

    this._configFile = path.join(dir, file);

    this._defaultConfig = defaultConfig;

    this._configCollection = new LocalCollection('config');
    this._configUpgrader = new ConfigUpgrader(path.join(__dirname, './migrations'));

    // Watchers of certain configuration keys
    this._observers = {};

    // Uses the observer functionality to decide when to write the configuration
    this._writeObserver = this._configCollection.find('config')
      .observe({
        added: () => {
          if (!this.initialAdded && !this._isUpgraded && !this._isDefaultConfig) {
            // Don't write on initial insertion, since this overwrites the same content
            // Except if we've upgraded the cloud
            // But not when we've upgraded the default configuration.
            this._log('info', 'Initial insertion of the config, not re-writing');
            this.initialAdded = true;
          } else {
            this._write();
          }
        },
        changed: () => this._write(),
        remove: () => this._write(),
      });
  }

  /**
   * React on changes in certain configuration keys.
   * The last parameter is the callback to call when a change occurs
   *
   * @param {*} events
   */
  watch(...events) {
    if (!this.isStarted) {
      throw new Error('Can not watch the configuration before the service is started');
    }

    if (events.length < 2 || !isFunction(events[events.length - 1])) {
      return;
    }
    const cb = events.pop();
    const fields = this._eventsToFields(events);
    const newFields = this._configCollection.findOne('config', {
      fields,
    });

    cb(this._parseFields(newFields, fields), true);
    events.sort();
    this.on(events.join(','), cb);
  }

  /**
   * Get the entire configuration object
   */
  get config() {
    const config = this._configCollection.findOne('config');
    if (config) {
      delete config._id;
    }
    return config;
  }

  /**
   * Get one key of the configuration
   * Can use nested-object notation, e.g. obj.subobj.key
   *
   * @param {*} setting
   */
  getSetting(setting) {
    return getValue(this.config, setting);
  }

  /**
   * Overwrite the entire configuration.
   * This is asynchronous since an asynchronous upgrade might be executed
   * in case the configuration is not in the correct version
   *
   * @param {*} config
   */
  async setConfig(config) {
    this._log('info', 'Setting config to:', JSON.stringify(config, null, 2));

    this._isDefaultConfig = false;
    // Use the default configuration
    if (!config && this._defaultConfig) {
      config = this._defaultConfig;
      this._isDefaultConfig = true;
    }

    if (config && config.version) {
      const upgradedConfig = await this._upgrade(config, proxyVersion);
      if (!isEqual(config, upgradedConfig)) {
        // This makes sure we don't ignore the initial insertion in the write observer
        // i.e. this makes sure we write the config again in case of an upgraded configuration
        this._isUpgraded = true;
        config = upgradedConfig;
      }
    }

    if (!config) {
      // clear configuration
      this._configCollection.remove('config');
    } else if (!this.config) {
      // No existing configuration object => insert
      config._id = 'config';
      this._log('info', 'Inserting new config');
      this._configCollection.insert(config);
    } else {
      // Existing configuration object => update
      delete config._id;
      this._log('info', 'Updating existing config');
      this._configCollection.update('config', config);
    }
  }

  /**
   * Update one key in the configuration
   * @param {*} key
   * @param {*} value
   */
  update(key, value) {
    const toSet = {};
    this._log('verbose', `Updating ${key} to ${JSON.stringify(value, null, 2)}`);
    if (typeof value === 'undefined') {
      toSet.$unset = {};
      toSet.$unset[key] = 1;
    } else {
      toSet.$set = {};
      toSet.$set[key] = value;
    }

    this._configCollection.update('config', toSet);
  }

  /**
   * Remove one key in the configuration
   * @param {*} key
   */
  remove(key) {
    this.update(key);
  }

  /**
   * Clears the entire configuration
   */
  async clear() {
    await this.setConfig(null);
  }

  get proxyMap() {
    return {
      '*': '*',
    };
  }

  /**
   * Maps the file observers to events.
   *
   * @param {*} mappedEvent
   * @param {*} event
   */
  _subscribeForNotification(mappedEvent, event) {
    if (!this._observers.hasOwnProperty(event)) {
      const fields = this._eventsToFields(event.split(','));

      this._log('debug', 'Subscribing on fields', fields);
      this._observers[event] = this._configCollection.find('config', {
        fields,
      }).observeChanges({
        changed: (id, newFields) => {
          this._log('debug', `Change in config, triggering: ${event}`);
          this.emit(event, this._parseFields(newFields, fields), false);
        },
      });
    }
  }

  /**
   * Utility to help group duplicate watchers
   * Just an optimization
   *
   * @param {*} newFields
   * @param {*} fields
   */
  _parseFields(newFields = {}, fields = {}) {
    fields = Object.keys(fields);
    if (fields.length === 1) {
      return getValue(newFields, fields[0]);
    }
    delete newFields._id;

    return fields.reduce((acc, field) => {
      acc[field] = getValue(newFields, field);
      return acc;
    }, {});
  }

  /**
   * Utility to help group duplicate watchers
   * Just an optimization
   *
   * @param {*} newFields
   * @param {*} fields
   */
  _eventsToFields(events) {
    if (events.includes('*')) {
      return {};
    }
    return events.reduce((acc, el) => {
      acc[el] = 1;
      return acc;
    }, {});
  }

  _unsubscribeForNotification(mappedEvent, event) {
    if (this._observers.hasOwnProperty(event)) {
      this._observers[event].stop();
      delete this._observers[event];
    }
  }

  /**
   * Starts the manager
   * This will start the upgrader & read the configuation
   */
  async _onStart() {
    this._log('info', 'Populating config upgrader');
    await this._configUpgrader.start();
    await this._read();
    await super._onStart();
  }

  async _onStop() {
    Object.values(this._observers).forEach(observer => observer.stop());
    this._observers = [];
    await this._configUpgrader.stop();
    await super._onStop();
  }

  /**
   * Tries to read the config file
   * @param {*} file
   */
  _tryToRead(file) {
    try {
      this._log('info', `Trying to read ${file}`);
      return JSON.parse(fs.readFileSync(file, 'utf-8'));
    } catch (error) {
      this._log('error', 'Error while reading config', error);
      throw error;
    }
  }

  /**
   * Reads the config file or makes sure the default settings are set when no config is given
   */
  async _read() {
    try {
      ensureLatest(this._configFile);
      await this.setConfig(this._tryToRead(this._configFile));
    } catch (error) {
      await this.setConfig(null);
    }
  }

  /**
   * Upgrade a configuration object to a given version format
   *
   * @param {*} config
   * @param {*} to
   */
  async _upgrade(config, to) {
    try {
      this._log('info', 'Trying to upgrade');

      if (config) {
        this._log('info', `Current version: ${config.version}, target version: ${to}`);
      } else {
        this._log('info', 'No config given to upgrader => no upgrade');
        return config;
      }

      if (this._configUpgrader.needsUpgrade(config, to)) {
        this._log('info', 'Needs config upgrade... upgrading');
        config = await this._configUpgrader.upgrade(config, to);
        this._log('info', `Config after upgrading: ${JSON.stringify(config, null, 2)}`);
      } else {
        this._log('info', 'Config does not need upgrading');
      }
    } catch (error) {
      this._log('error', 'Config upgrade error', error);
    }

    return config;
  }

  /**
   * Safely writes the configuration file
   * Triggered by changes in the configuration object
   */
  _write() {
    this._log('info', 'Writing config file');
    this._log('silly', 'Config to write', JSON.stringify(this.config, null, 2));
    if (this.config) {
      try {
        this._log('verbose', 'Writing new config file');
        writeSafely(this._configFile, JSON.stringify(this.config, null, 2), this._logger);
      } catch (error) {
        this._log('error', 'Error while writing config file', error);
      }
    } else {
      this._log('verbose', 'Empty config => Removing config files');
      removeAll(this._configFile, this._logger);
    }
  }

  /**
   * Cleanup
   */
  async _onDestroy() {
    this._writeObserver.stop();
    delete this._writeObserver;
    await this._configUpgrader.destroy();
    delete this._configUpgrader;
    this._configCollection.remove({});
    delete this._configCollection;
    delete this._observers;

    await super._onDestroy();
  }
}

export default ConfigurationManager;
