/*
eslint-disable no-await-in-loop
*/

import * as semver from 'semver';
import * as fs from 'fs-extra';
import * as path from 'path';

import { Service } from '../service';

/**
 * Class to upgrade a configuration.
 *
 * An upgrader basically consists of a target version and a transformation method.
 * The transformation method operates on a given config.
 * Then tranformations can be asynchronous, since they may need to fetch information asynchronous.
 *
 * Given a configuration with a certain version, the configuration upgrader applies all
 * transformations sequentially until it reaches the target state.
 * Thus an upgrader can assume it gets the result of the version exactly before its defined version
 */
class ConfigUpgrader extends Service {
  /**
   * Takes the migration path, that is the directory with transformation functions
   * @param {*} migrationsPath
   */
  constructor(migrationsPath) {
    super();
    this._migrationsPath = migrationsPath;
    this._reset();
  }

  /**
   * Populates the transformations
   * E.g. read the directory of upgraders
   */
  _onStart() {
    this._populate();
    return super._onStart();
  }

  /**
   * Cleans the set of upgraders
   */
  _onStop() {
    this._reset();
    return super._onStop();
  }

  /**
   * Clear list of upgraders
   */
  _reset() {
    this.upgraders = new Map();
  }

  /**
   * Reads all the files from the transformations directory
   * And adds the transformations
   */
  _populate() {
    const re = /\.js$/;

    this._log('info', `Adding files from ${this._migrationsPath}`);

    fs.readdirSync(this._migrationsPath)
      .filter(file => re.test(file))
      .forEach((file) => {
        const filePath = path.join(this._migrationsPath, file);
        this._log('info', `Adding ${file} to configupgrader (at ${filePath})`);
        this._addUpgrader(require(filePath).default); // eslint-disable-line
      });
  }

  /**
   * Add an upgrader for a certain version to the set of tranformations
   *
   * @param {*} param0
   */
  _addUpgrader({ version, upgrade }) {
    const wrappedUpgrade = function wrappedUpgrade(config) {
      config = upgrade(config);
      config.version = version;
      return config;
    };
    this.upgraders.set(version, wrappedUpgrade);
  }

  /**
   * Gets the list of upgraders (transformations) to apply between 2 versions
   *
   * @param {*} from
   * @param {*} to
   */
  _getUpgradersBetween(from, to) {
    try {
      return [...this.upgraders.keys()]
        .filter(v => (!from || semver.gt(v, from)) && (!to || semver.lte(v, to)))
        .sort(semver.compare)
        .map(v => this.upgraders.get(v));
    } catch (error) {
      return [];
    }
  }

  /**
   * Checks if we actually need to upgrade given a config and a target version
   *
   * @param {*} config
   * @param {*} to
   */
  needsUpgrade(config, to) {
    return !!this._getUpgradersBetween(config.version, to).length;
  }

  /**
   * This performs the actual upgrade to the requested version
   *
   * First calculates a list of upgraders to apply
   * Then applies them step by step.
   *
   * In case of an error, returns the latest successful transformation
   *
   * @param {*} config
   * @param {*} to
   */
  async upgrade(config, to) {
    try {
      // make sure we don't modify the original
      config = JSON.parse(JSON.stringify(config));

      // Calculate list of updates to apply
      const steps = this._getUpgradersBetween(config.version, to);
      this._log('info', `Upgrading ${steps.length} steps`);
      for (const transform of steps) {
        // Transforms may be asynchronous
        config = await transform(config);
      }
      return config;
    } catch (error) {
      this._log('error', 'Error while upgrading', error);
      return config;
    }
  }
}

export default ConfigUpgrader;
