import Service from './service';

/**
 * Checks the connection with the Meteor server
 * We see any data package coming from the server as proof it's alive.
 * If we haven't received one for 30s (configurable), we 'ping' the service we monitor
 * If we don't get a 'pong' in 15s (configurable), we assume connection is lost and trigger an event
 */

/**
 * Abstract connection monitor
 */
class ConnectionMonitor extends Service {
  constructor({ heartbeatInterval = 15000, heartbeatTimeout = 15000, heartbeatMaxRetries = 3 }) {
    super({});

    const params = { heartbeatInterval, heartbeatTimeout, heartbeatMaxRetries };
    Object.keys(params).forEach((param) => {
      if (params[param] === undefined) throw new Error(`Please provide a valid ${param}`);
      this[`_${param}`] = params[param];
    });

    this._seenPacket = false;

    this._heartbeatIntervalHandle = null;
    this._heartbeatTimeoutHandle = null;
    this._heartbeatRetries = 0;
  }

  _onStop() {
    this._clearHeartbeatIntervalTimer();
    this._clearHeartbeatTimeoutTimer();
    return super._onStop();
  }

  _onStart() {
    this._startHeartbeatIntervalTimer();
    return super._onStart();
  }

  _startHeartbeatIntervalTimer() {
    this._log('silly', 'Starting interval timer');
    this._heartbeatIntervalHandle = setInterval(
      this._heartbeatIntervalFired.bind(this),
      this._heartbeatInterval,
    );
  }

  _startHeartbeatTimeoutTimer() {
    this._log('silly', 'Starting timeout timer');
    this._heartbeatTimeoutHandle = setTimeout(
      this._heartbeatTimeoutFired.bind(this),
      this._heartbeatTimeout,
    );
  }

  _clearHeartbeatIntervalTimer() {
    if (this._heartbeatIntervalHandle) {
      this._log('silly', 'Clearing interval timer');
      clearInterval(this._heartbeatIntervalHandle);
      this._heartbeatIntervalHandle = null;
    }
  }

  _clearHeartbeatTimeoutTimer() {
    if (this._heartbeatTimeoutHandle) {
      this._log('silly', 'Clearing timeout timer');
      clearTimeout(this._heartbeatTimeoutHandle);
      this._heartbeatTimeoutHandle = null;
    }
  }

  // The heartbeat interval timer is fired when we should send a ping.
  _heartbeatIntervalFired() {
    // don't send ping if we've seen a packet since we last checked,
    // *or* if we have already sent a ping and are awaiting a timeout.
    // That shouldn't happen, but it's possible if
    // `self.heartbeatInterval` is smaller than
    // `self.heartbeatTimeout`.
    this._log('silly', 'Heartbeat interval fired');
    if (!this._seenPacket && !this._heartbeatTimeoutHandle) {
      this._log('silly', 'No data seen for a while, pinging server');
      this._sendPing();
      // Set up timeout, in case a pong doesn't arrive in time.
      this._startHeartbeatTimeoutTimer();
    }
    this._seenPacket = false;
  }

  _sendPing() {
    throw new Error('Connection Monitor must implement sendPing method');
  }

  /**
   * Method to be overwritten in child classes to provide a more descriptive log message
   */
  _logHeartbeatTimeout() {
    this._log('error', 'Heartbeat timeout', { serviceLog: true });
  }

  // The heartbeat timeout timer is fired when we sent a ping, but we
  // timed out waiting for the pong.
  _heartbeatTimeoutFired() {
    this._logHeartbeatTimeout();
    this._heartbeatTimeoutHandle = null;

    if (this._heartbeatRetries < this._heartbeatMaxRetries) {
      this._heartbeatRetries += 1;
      this._log('debug', `Heartbeat timed out, retrying ${this._heartbeatRetries}/${this._heartbeatMaxRetries}`);
    } else {
      this._log('debug', 'Heartbeat failed');
      this.emit('timeout');
    }
  }

  messageReceived() {
    // Tell periodic checkin that we have seen a packet, and thus it
    // does not need to send a ping this cycle.
    this._seenPacket = true;
    this._heartbeatRetries = 0;
    // If we were waiting for a pong, we got it.
    if (this._heartbeatTimeoutHandle) {
      this._log('silly', 'Received pong');
      this._clearHeartbeatTimeoutTimer();
    } else {
      this._log('silly', 'Received pong, resetting heartbeat timer');
    }
  }
}

export default ConnectionMonitor;
