import * as x11 from 'x11';

import createQueue from '../utils/createQueue';
import { execAsync } from '../utils/utils';

export default function createX11Client({ logger }) {
  let X11Client;
  let destroyed = false;

  const commands = createQueue((method, ...params) => {
    // ensure X11 methods are never blocking
    setTimeout(() => {
      logger.debug('Calling method on X11Client', { method, params });
      X11Client[method](...params);
    }, 0);
  });

  x11.createClient((err, display) => {
    if (!err) {
      logger.info('Connected to X11', Object.keys(display));
      X11Client = display.client;
      X11Client.on('error', (clientErr) => {
        logger.error('Error from X11', { err: clientErr });
      });
      X11Client.on('event', (event) => {
        logger.info('Event from X11', { event });
      });
      commands.queued = false;
    } else {
      logger.error('Could not connect to X11', err);
      commands.queued = true;
    }
  });

  async function getHandle(window) {
    if (window && Function.prototype.isPrototypeOf(window.getNativeWindowHandle)) {
      return window.getNativeWindowHandle().readUInt32LE(0);
    }
    if (typeof window === 'string') {
      const { stdout } = await execAsync(`xwininfo -name "${window.replace('"', '\\"')}" -tree -int`);
      const [_all, handle] = stdout.match(/xwininfo: Window id: ([0-9]+) ".*"/i);
      return parseInt(handle, 10);
    }
    throw Error('Could not get window handle');
  }

  return {
    raiseWindow(window) {
      getHandle(window).then((handle) => {
        commands.push('RaiseWindow', handle);
      });
    },
    lowerWindow(window) {
      getHandle(window).then((handle) => {
        commands.push('LowerWindow', handle);
      });
    },
    destroy() {
      if (destroyed) return;
      destroyed = true;
      X11Client.removeAllListeners();
      return new Promise(resolve => X11Client.close(resolve));
    },
  };
}
