/*
eslint-disable no-underscore-dangle
*/
import * as path from 'path';
import * as fs from 'fs-extra';

import CacheManager from '../cacheManager';
import * as files from '../../files';

/**
 * Meteor cache, extends from the base cache
 */
class MeteorCacheManager extends CacheManager {
  /**
   * Call the parent with the required collections
   * @param bridge
   */
  constructor(bridge, collections = ['compositions', 'sinks', 'sources', 'layouts']) {
    super(collections);
    this._bridge = bridge;
  }

  /**
   * If there is already existing data, warm the cache with the data
   */
  warm() {
    return Promise.all(this._collections.map((collection) => {
      if (this._bridge._meteorClient._ddpClient.collections.hasOwnProperty(collection)) {
        return Promise.all(Object.keys(this._bridge._meteorClient._ddpClient.collections[collection])
          .map(id => this._bridge._meteorClient._ddpClient.collections[collection][id])
          .filter(item => !item.device || item.device._id === this._bridge._device._id)
          .map(item => this._bridge._trigger(collection, 'added', item)));
      }
      return Promise.resolve();
    }));
  }

  async save() {
    const file = path.join(files.cachePath, `${this._bridge._device._id}.json`);
    try {
      await fs.writeJson(file, this.data, { spaces: 2 });
    } catch (error) {
      throw new Error('Could not write config to cache');
    }
  }

  /**
   * Add an item in the given collection
   * @param collection name of the collection
   * @param item item to add
   */
  add(collection, item) {
    super.add(collection, item._id, item);
  }

  /**
   * Remove the item from the collection
   * @param collection
   * @param item Can be the item itself, or the item identifier (mongo id)
   */
  remove(collection, item) {
    const id = (typeof item === 'string') ? item : item._id;

    super.remove(collection, id);
  }

  getMeeting() {
    return this._data.$arrays.meetings[0];
  }

  findCompositionOnSink(sink) {
    if (sink.activeCompositionId) {
      return this.compositions[sink.activeCompositionId];
    }
  }

  findDefaultCompositionForSink(sink) {
    if (sink.defaultCompositionId) {
      return this.compositions[sink.defaultCompositionId];
    }
  }

  findActiveSinksForComposition(composition) {
    return this._data.$arrays.sinks.filter(sink => sink.activeCompositionId === composition._id);
  }

  findDefaultSinksForComposition(composition) {
    return this._data.$arrays.sinks.filter(sink => sink.defaultCompositionId === composition._id);
  }

  /**
   * Given a meteor composition by it's role
   * @param role
   * @returns {*}
   */
  findCompositionByRole(role) {
    return this._data.$arrays.compositions.find(composition => composition.role === role);
  }

  /**
   * Given a meteor sink by it's role
   * @param role
   * @returns {*}
   */
  findSinkByRole(role) {
    return this._data.$arrays.sinks.find(sink => sink.role === role);
  }

  /**
   * Given a meteor sink by type
   * @param type
   * @returns {*}
   */
  findSinkByType(type) {
    return this._data.$arrays.sinks.filter(sink => sink.type === type);
  }

  /**
   * Given a meteor sink by types
   * @param type
   * @returns {*}
   */
  findSinkByTypes(types) {
    return this._data.$arrays.sinks.filter(sink => types.includes(sink.type));
  }

  /**
   * Clean-up
   */
  _onDestroy() {
    delete this._bridge;
    return super._onDestroy();
  }
}

export default MeteorCacheManager;
