/**
 * (C) 2015 Copyright BARCO NV
 * All rights reserved. Reproduction in whole or in part is
 * prohibited without the written consent of the copyright owner.
 *
 * Part of Barco WeConnect
 */
import * as DDPClient from 'ddp';
import * as pify from 'pify';
import * as getValue from 'get-value';
import * as Websocket from 'faye-websocket/lib/faye/websocket/api';
import { constants } from 'crypto';
import { proxyConfigToUrl } from '../../utils/network';
import { Service } from '../service';
import DDPConnectionMonitor from './DDPConnectionMonitor';
import serviceRegistry from '../serviceRegistry';
import { waitMaximum } from '../../utils/utils';
import CancelToken from '../cancelToken';
import secretReplacer from '../../utils/secretReplacer';

/**
 * A class that implements the Decorator pattern for the default DDP library.
 * This wraps the DDP client library with extended events, rpc shortcuts and additional socket management functionality
 *
 * For data events, we extend from EventEmitter2, which allows wildcard event names and event namespaces.
 * That way we can listen to all events for one NRC, which comes in handy in the Meteor Bridges.
 *
 * Therefore you can view the meteor client as a data router that routes data from one NRC to the correct meteor bridge,
 * but fully decoupled (using publisher-subscriber pattern)
 */
class MeteorClient extends Service {
  constructor({
    connectionDetails,
  }) {
    super({
      id: false,
      eventOptions: {
        wildcard: true,
        delimiter: ':',
        maxListeners: 500,
      },
    });

    const params = {
      connectionDetails,
    };

    Object.keys(params).forEach((param) => {
      if (params[param] === undefined) throw new Error(`Please provide a valid ${param}`);
      this[`_${param}`] = params[param];
    });

    this._isConnected = false;
    this._hasConnected = false;

    this._autoReconnect = !!this._connectionDetails.autoReconnect;
    // Reconnect after a random timeout between 20 and 60s
    const reconnectionTimeout = Math.ceil(Math.random() * 40000) + this._connectionDetails.autoReconnectTimer;
    this._autoReconnectTimer = Math.max(reconnectionTimeout, 2500);

    this._createDDPClient();
    this._addMethodShortcuts();

    this._log('info', `New MeteorClient created. Reconnection timeout = ${this._autoReconnectTimer} ms. Connection details = ${JSON.stringify(this._connectionDetails, null, ' ')}`);
  }

  // There might bea temporal discrepancy between our isConnected state and the actual state of the socket
  // This might lead into a race condition where a subscription callback is never called
  // Leading to infinetely waiting await statements
  get isConnected() {
    return this._isConnected && this._ddpClient.socket.readyState <= Websocket.OPEN;
  }

  _createDDPClient() {
    this._connectionDetails = {
      port: 80,
      ...this._connectionDetails,
      autoReconnect: false,
      maintainCollections: true,
    };

    if (this._connectionDetails.proxy && this._connectionDetails.proxy.url) {
      try {
        this._connectionDetails.tlsOpts = {
          proxy: {
            origin: proxyConfigToUrl(this._connectionDetails.proxy),
          },
        };
      } catch (error) {
        this._log('error', `Invalid proxy configuration: ${JSON.stringify(this._connectionDetails.proxy, secretReplacer)}`);
      }
    }

    /**
     * In order to make sure client can only connect to TLSv1.3 or better servers
     * we need to enforce it here.
     * Meteor DDP uses faye-websocket. See the documentation here for more info on
     * how to set it up: https://github.com/faye/faye-websocket-node#initialization-options
     * TLS options can be found here https://nodejs.org/api/tls.html#tls_tls_createsecurecontext_options
     * and here https://nodejs.org/api/crypto.html#crypto_openssl_options
     */
    const tls = {
      // Recent node way to only support >= TLSv1.2
      minVersion: 'TLSv1.2',
      // Backwards compatibility: old node way
      secureOptions:
        constants.SSL_OP_CIPHER_SERVER_PREFERENCE
        | constants.SSL_OP_NO_TLSv1
        | constants.SSL_OP_NO_TLSv1_1,
    };
    this._connectionDetails.tlsOpts = this._connectionDetails.tlsOpts || {};
    Object.assign(this._connectionDetails.tlsOpts, { tls });

    // Meteor's implementation does not work very well, we use our own reconnect logic
    this._ddpClient = new DDPClient(this._connectionDetails);

    this._connectionMonitor = new DDPConnectionMonitor({
      ddpClient: this._ddpClient,
    });

    this._ddpClient.on('socket-close', (code, message) => {
      this._log(
        'error',
        `Socket connection close. Error code = ${JSON.stringify(code, null, ' ')} , message = ${JSON.stringify(message, null, ' ')}`,
        { serviceLog: true },
      );

      // The ddp client library only cancels method calls
      // However, since we use async/await we could be waiting infinitely
      // So cancel all subscriptions too
      // This does require us to mess with private variables in ddpclient
      const methodIds = Object.keys(this._ddpClient._pendingMethods); //eslint-disable-line
      const callbackIds = Object.keys(this._ddpClient._callbacks); //eslint-disable-line
      callbackIds
        .filter(id => !methodIds.includes(id))
        .forEach((id) => {
          this._ddpClient._callbacks[id](DDPClient.ERRORS.DISCONNECTED); //eslint-disable-line
          delete this._ddpClient._callbacks[id]; //eslint-disable-line
        });

      if (this.isConnecting || this._isConnected) {
        this._cleanupAndReinitializeIfNeeded();
      }
    });

    this._ddpClient.on('socket-error', (error) => {
      this._log('error', `Socket connection error. Error = ${error}`);
      if (this.isConnecting || this._isConnected) {
        this._cleanupAndReinitializeIfNeeded();
      }
    });
  }

  /**
   *Bind listeners for each relevant collection
   * @private
   */
  _addCollectionObservers() {
    this._meetingObserver = this._ddpClient.observe('meetings');
    this._meetingObserver.added = this._onAdd.bind(this, 'meetings');
    this._meetingObserver.changed = this._onChange.bind(this, 'meetings');
    this._meetingObserver.removed = this._onRemove.bind(this, 'meetings');

    this._deviceObserver = this._ddpClient.observe('devices');
    this._deviceObserver.added = this._onAdd.bind(this, 'devices');
    this._deviceObserver.changed = this._onChange.bind(this, 'devices');
    this._deviceObserver.removed = this._onRemove.bind(this, 'devices');

    this._layoutObserver = this._ddpClient.observe('layouts');
    this._layoutObserver.added = this._onAdd.bind(this, 'layouts');
    this._layoutObserver.changed = this._onChange.bind(this, 'layouts');
    this._layoutObserver.removed = this._onRemove.bind(this, 'layouts');

    this._sinkObserver = this._ddpClient.observe('sinks');
    this._sinkObserver.added = this._onAdd.bind(this, 'sinks');
    this._sinkObserver.changed = this._onChange.bind(this, 'sinks');
    this._sinkObserver.removed = this._onRemove.bind(this, 'sinks');

    this._compositionObserver = this._ddpClient.observe('compositions');
    this._compositionObserver.added = this._onAdd.bind(this, 'compositions');
    this._compositionObserver.changed = this._onChange.bind(this, 'compositions');
    this._compositionObserver.removed = this._onRemove.bind(this, 'compositions');

    this._sourceObserver = this._ddpClient.observe('sources');
    this._sourceObserver.added = this._onAdd.bind(this, 'sources');
    this._sourceObserver.changed = this._onChange.bind(this, 'sources');
    this._sourceObserver.removed = this._onRemove.bind(this, 'sources');

    this._instituteObserver = this._ddpClient.observe('institutes');
    this._instituteObserver.added = this._onAdd.bind(this, 'institutes');
    this._instituteObserver.changed = this._onChange.bind(this, 'institutes');
    this._instituteObserver.removed = this._onRemove.bind(this, 'institutes');

    // RPC are handled differently than other collections, since they are not necessarily device-specific.
    this._rpcObserver = this._ddpClient.observe('rpcs');
    this._rpcObserver.added = (id) => {
      this._log('verbose', `>>>>>>>>>>>>>>> [ADDED] in "rpcs" :  ${JSON.stringify(id, null, ' ')}`);
      const record = this._ddpClient.collections.rpcs[id];
      this.emit('rpc', record);
      this.emit(`rpc:${record.command}`, record);
      if (record.params && record.params.targetNrcId) {
        const { token, cancel } = CancelToken.source();
        this.emit(`${record.params.targetNrcId}:rpc`, {
          ...record,
          cancelDefaultResponse: cancel,
        });

        if (!token.requested) {
          // if cancel method was not called, no bridges responded to the given rpc
          this.markRPCDone({
            rpcId: record._id,
            error: true,
          });
        }

        token.destroy();
      }
    };
  }

  /**
   * We create shortcuts for most methods, so we can call them with a promise-based interface directly as methods.
   * @private
   */
  _addMethodShortcuts() {
    const methods = {
      latestDeviceVersionInfo: 'barco.device.latestVersionInfo',
      updateDevice: 'barco.device.saveByDevice',
      updateTimeStatus: 'barco.device.saveTimeStatus',
      addSource: 'barco.source.addByDevice',
      updateSource: 'barco.source.saveByDevice',
      removeSource: 'barco.source.remove',
      markRPCDone: 'clientrpc.markDone',
      setDeviceStatus: 'barco.devices.setStatus',
      setDeviceUpgradeStatus: 'barco.device.setUpgradeStatus',
      updateSourcePreviewUrl: 'barco.source.setPreviewUrl',
      updateSinkUrl: 'barco.sink.setUrl',
      getRecordingInformation: 'barco.meeting.getRecordingInformation',
      addRecordingFile: 'barco.recordings.addFile',
      updateRecordingFile: 'barco.recordings.updateFile',
      setWatchConnection: 'barco.device.watchConnection',
    };

    Object.keys(methods).forEach((method) => {
      if (!this.hasOwnProperty(method)) {
        this[method] = this.call.bind(this, methods[method]);
      }
    });
  }

  /**
   * Handle meteor socket connection failures
   * @private
   */
  async _cleanupAndReinitializeIfNeeded() {
    this._log('verbose', 'Connection lost with Meteor', { serviceLog: true });
    this._log('info', `

    =================================================================================
    ==                                                                             ==
    ==                   D D P   C O N N E C T I O N   ${this._isConnected ? 'L O S T    ' : 'F A I L E D'}                 ==
    ==                                                                             ==
    =================================================================================

        `);

    if (this._isConnected) {
      this.emit('disconnected');
    }

    this._isConnected = false;
    this.isConnecting = false;

    this._ddpClient.removeAllListeners('connected');
    this._ddpClient.removeAllListeners('failed');

    try {
      await this._connectionMonitor.stop();
    } catch (e) {}

    if (this._autoReconnect && this.isActive) {
      this._log('info', 'Waiting for initialization timer');
      this._scheduleInitialization();
    } else {
      this._log('info', 'Not trying to re-initialize.');
    }
  }

  /**
   * Close the ddp connection
   */
  close() {
    this._log('info', 'Closing connection');
    this._ddpClient.close();
  }

  /**
   * Transforms a meteor call (callback-style) in promise-based callback to better fit the rest of our codebase.
   * This is also a useful utility for the rpc-shortcuts defined in the class constructor.
   *
   * @param method
   * @param params
   * @returns {Promise}
   */
  async call(method, ...params) {
    if (!this.isConnected) {
      throw new Error('Not connected');
    }

    this._log('verbose', `Calling Meteor RPC function "${method}" with params: ${JSON.stringify(params, null, ' ')}`);

    let result;
    try {
      // Add a timeout so if something goes terribly wrong on application level (e.g. infinite loop at server function)
      // We don't get stuck in an await
      const call = pify(this._ddpClient.call.bind(this._ddpClient))(method, params);
      result = await waitMaximum(call, 30 * 1000);
    } catch (error) {
      this._log('error', `Result from Meteor RPC function: "${method}"`, error);
      throw error;
    }

    this._log('verbose', `Result from Meteor RPC function "${method}", result: ${JSON.stringify(result, null, ' ')}`);
    return result;
  }

  /**
   * Generic ddp onAdd listener.
   * This will emit a multitude of events, which come in handy for meteor bridge to target device-related data
   *
   * @param entity
   * @param id
   * @private
   */
  _onAdd(entity, id) {
    this._log('verbose', `>>>>>>>>>>>>>>> [ADDED] in "${entity}" :  ${id}`);
    const record = this._ddpClient.collections[entity][id];
    this._log('debug', `>>>>>>>>>>>>>>> [ADDED] record :  ${JSON.stringify(record, secretReplacer, ' ')}`);

    this.emit('data', entity, 'added', record);
    this.emit(`${entity}:added`, record);
    this.emit(`added:${entity}`, record);

    if (record.device && record.device._id) {
      this.emit(`DeviceData:${record.device._id}:added`, entity, record);
      this.emit(`DeviceData:${record.device._id}:${entity}:added`, record);
      this.emit(`DeviceData:${record.device._id}:added:${entity}`, record);
    }
  }

  /**
   * Generic ddp onChange listener.
   * This will emit a multitude of events, which come in handy for meteor bridge to target device-related data
   *
   * @param entity
   * @param id
   * @param oldFields
   * @param clearedFields
   *
   * @private
   */
  _onChange(entity, id, oldFields, clearedFields, newFields) {
    this._log('verbose', `>>>>>>>>>>>>>>> [CHANGED] in "${entity}" :  ${id}`);

    this._log('verbose', `>>>>>>>>>>>>>>> [CHANGED] old field values:  ${JSON.stringify(oldFields, secretReplacer, ' ')}`);
    this._log('verbose', `>>>>>>>>>>>>>>> [CHANGED] cleared fields:  ${JSON.stringify(clearedFields, secretReplacer, ' ')}`);
    this._log('verbose', `>>>>>>>>>>>>>>> [CHANGED] new fields:  ${JSON.stringify(newFields, secretReplacer, ' ')}`);

    const record = this._ddpClient.collections[entity][id];
    // The line below blows up the logfiles. Only uncomment during debug sessions...
    // this._log('debug', `>>>>>>>>>>>>>>> [CHANGED] total record :  ${JSON.stringify(record, secretReplacer, ' ')}`);

    this.emit('data', entity, 'changed', record);
    this.emit(`changed:${entity}`, record, oldFields, clearedFields, newFields);
    this.emit(`${entity}:changed`, record, oldFields, clearedFields, newFields);

    this.emit(`${record._id}:changed`, record, oldFields, clearedFields, newFields);

    if (record.device && record.device._id) {
      this.emit(`DeviceData:${record.device._id}:changed`, entity, record, oldFields, clearedFields, newFields);
      this.emit(`DeviceData:${record.device._id}:changed:${entity}`, record, oldFields, clearedFields, newFields);
      this.emit(`DeviceData:${record.device._id}:${entity}:changed`, record, oldFields, clearedFields, newFields);
    }
  }

  /**
   * Generic ddp onRemove listener.
   * This will emit a multitude of events, which come in handy for meteor bridge to target device-related data
   *
   * @param entity
   * @param id
   * @param record
   * @private
   */
  _onRemove(entity, id, record) {
    this._log('verbose', `>>>>>>>>>>>>>>> [REMOVED] in "${entity}" :  ${id}`);
    this._log('debug', `>>>>>>>>>>>>>>> [REMOVED] old record: ${JSON.stringify(record, secretReplacer, ' ')}`);

    this.emit('data', entity, 'removed', record);
    this.emit(`${entity}:removed`, record);
    this.emit(`removed:${entity}`, record);

    this.emit(`${entity._id}:removed`, record);

    if (record.device && record.device._id) {
      this.emit(`DeviceData:${record.device._id}:removed`, entity, record);
      this.emit(`DeviceData:${record.device._id}:removed:${entity}`, record);
      this.emit(`DeviceData:${record.device._id}:${entity}:removed`, record);
    }
  }

  _scheduleInitialization() {
    if (!this._initializationTimerHandle) {
      this._log('info', `Scheduling re-initialization in ${this._autoReconnectTimer} ms`);
      this._initializationTimerHandle = setTimeout(() => {
        this._initialize();
      }, this._autoReconnectTimer);
    }
  }

  _clearTimers() {
    this._clearConnectingTimeout();
    this._clearInitializationTimer();
  }

  _clearConnectingTimeout() {
    if (this._connectingTimeoutHandle) {
      clearTimeout(this._connectingTimeoutHandle);
      delete this._connectingTimeoutHandle;
    }
  }

  _clearInitializationTimer() {
    if (this._initializationTimerHandle) {
      clearTimeout(this._initializationTimerHandle);
      delete this._initializationTimerHandle;
    }
  }

  get _loginCredentials() {
    // Preferably with a previous resume token, otherwise a new resume token will be created for each login.
    // These resume tokens are saved in the MongoDB database, therefore we don't want to store more than necessary!

    const credentials = serviceRegistry.get('configManager').getSetting('credentials');

    if (getValue(credentials, 'resume.token')) {
      return {
        resume: credentials.resume.token,
      };
    } else if (credentials.user) {
      return {
        instituteId: credentials.instituteId,
        email: credentials.user.email,
        password: credentials.user.password,
      };
    }
    return {};
  }

  subscribe(name, params, context = {}, attr = '_subscriptionId', timeoutAfter = 30 * 1000) {
    if (this.isConnected) {
      const subscription = new Promise((resolve, reject) => {
        context[attr] = this._ddpClient.subscribe(name, params, (error) => {
          error ? reject(error) : resolve(context[attr]);
        });
      });
      // Add a timeout so if something goes terribly wrong on application level (e.g. infinite loop at server function)
      // We don't get stuck in an await
      return waitMaximum(subscription, timeoutAfter);
    }
    throw new Error('Not connected');
  }

  unsubscribe(id) {
    if (this.isConnected) {
      this._ddpClient.unsubscribe(id);
    } else {
      throw new Error('Not connected');
    }
  }

  async _onStart() {
    this._addCollectionObservers();
    await this._initialize();
    await super._onStart();
  }

  async _onStop() {
    this._clearTimers();

    this._sourceObserver.stop();
    this._rpcObserver.stop();
    this._sinkObserver.stop();
    this._layoutObserver.stop();
    this._compositionObserver.stop();
    this._deviceObserver.stop();
    this._instituteObserver.stop();

    this._unsubscribeToDevices();
    this.close();

    await super._onStop();
  }

  async _initialize() {
    this._log('info', 'Initializing');

    // clean collections which might still be present from old connection attempts
    this._ddpClient.collections = {};
    try {
      await this._connect();
      await this._connectionMonitor.start();
      await this._login();
      try {
        await this.setWatchConnection();
      } catch (error) {
        // XXX: this will fail if we connect new purple with old cloud,
        // ignore for now until we're in a better future
      }
      await this._checkVersion();
      this._subscribeToDevices();
    } catch (error) {
      this._log('error', 'Error during meteor client initialization:', error);
      if (this._isConnected) {
        this._destroyConnection();
      } else {
        this._cleanupAndReinitializeIfNeeded();
      }
    }
  }

  _destroyConnection() {
    if (this._ddpClient && this._ddpClient.socket) {
      this._ddpClient.socket._stream.destroy(); //eslint-disable-line
    }
  }

  /**
   * Connect and authenticate to meteor,
   * handles failures and subscribes to initial data (= which devices to connect to)
   */
  async _connect() {
    this.isConnecting = true;
    this._log('info', 'Trying to connect with Meteor', { serviceLog: true });

    this._clearTimers();

    this._connectingTimeoutHandle = setTimeout(() => {
      this._clearConnectingTimeout();
      this._log('error', 'Connection timeout triggered!', { serviceLog: true });
      this._destroyConnection();
    }, 100000);

    try {
      await pify(this._ddpClient.connect.bind(this._ddpClient))();
    } catch (error) {
      this.isConnecting = false;
      this._clearConnectingTimeout();
      this.emit('connectionError', error);
      this._log('error', 'DDP connection failed.', { ...error, serviceLog: true });
      throw error;
    }

    this._clearConnectingTimeout();
    this.isConnecting = false;
    this._isConnected = true;

    this._log('verbose', 'Connected with Meteor', { serviceLog: true });
    this._log('info', `

    =================================================================================
    ==                                                                             ==
    ==         D D P   C O N N E C T I O N   ${this._hasConnected ? 'R E - ' : '      '}E S T A B L I S H E D           ==
    ==                                                                             ==
    =================================================================================

                `);

    this.emit('connected', this._hasConnected);
    this._hasConnected = true;
  }

  async _login() {
    // Authenticate to Meteor.
    const credentials = this._loginCredentials;

    if (!Object.keys(credentials).length) {
      throw new Error('No valid login credentials set for Meteor');
    }

    this._log('debug', `Trying to log in with credentials: ${JSON.stringify(credentials)}`);

    let result;
    try {
      result = await pify(this._ddpClient.call.bind(this._ddpClient))('login', [credentials]);
    } catch (error) {
      // Could not authenticate
      this._log('error', 'Unable to log into Meteor.', { ...error, serviceLog: true });

      if (credentials.resume) {
        serviceRegistry.get('configManager').remove('credentials.resume');
      }

      throw error;
    }

    // Save resume token if it's been changed
    if (!credentials.resume || credentials.resume !== result.token) {
      delete result.id;
      serviceRegistry.get('configManager').update('credentials.resume', result);
    }

    this.emit('authenticated');
    this._log('info', 'Successfully logged into Meteor.', { serviceLog: true });
  }

  async _checkVersion() {
    // Check server version compatibility
    const versionMonitor = serviceRegistry.get('versionMonitor');
    try {
      // Only upgrade if version is not compatible, since if the upgrade would fail
      // this could lead to MeteorClient failing to start and only recovering after upgrading successfully
      const isCompatible = await versionMonitor.isCurrentVersionCompatible();

      if (!isCompatible) {
        this._log('info', 'Starting upgrade since version is not compatible.');
        await versionMonitor.checkVersionAndUpgradeIfNeeded();
      }
    } catch (error) {
      this._log('error', 'Error while making sure making sure we\'re compatible', error);
      throw error;
    }
  }

  _subscribeToDevices() {
    // Subscribe to initial set of data (= purple RPC's and devices)
    this._log('info', 'Subscribing for NRC related data');

    this.subscribe('barco.devices.mine', [], this, '_deviceSubscription')
      .then(() => {
        this._log('verbose', 'Subscription for "barco.devices.mine" is ready');
        this.emit('deviceSubscriptionReady');
      });

    this._log('info', 'Subscribing for single shot RPC calls to be executed');

    this.subscribe('clientrpc.myRPCs', [], this, '_rpcSubscription')
      .then(() => {
        this._log('verbose', 'Subscribed for "clientrpc.myRPCs"');
      });
  }

  _unsubscribeToDevices() {
    if (this._deviceSubscription) {
      this.unsubscribe(this._deviceSubscription);
    }

    if (this._rpcSubscription) {
      this.unsubscribe(this._rpcSubscription);
    }
  }

  /**
   * Clean up and destroy.
   * This removes all incoming and outgoing observers
   */
  async _onDestroy() {
    if (this._isConnected) {
      // Since the close events come in asynchronous, they could come in
      // after we removed the listeners and thus not triggering a clean-up
      // So make sure we run the clean up
      this._isConnected = false;
      await this._cleanupAndReinitializeIfNeeded();
    }

    this._ddpClient.removeAllListeners();

    this._connectionMonitor.destroy();
    delete this._connectionMonitor;

    delete this._ddpClient;
    delete this._sourceObserver;
    delete this._rpcObserver;
    delete this._sinkObserver;
    delete this._layoutObserver;
    delete this._compositionObserver;
    delete this._deviceObserver;
    delete this._instituteObserver;

    return super._onDestroy();
  }
}

export default MeteorClient;
