import * as chai from 'chai';
import * as Websocket from 'faye-websocket/lib/faye/websocket/api';
import * as DDPClient from 'ddp';
import { EventEmitter } from 'events';
import * as EJSON from 'ddp-ejson';
import { setTimeout } from 'timers';

import MeteorClient from './meteorClient';
import { registerAll } from '../../utils/test-helpers/fillServiceRegistry';

const { assert } = chai;

class FakeSocket extends EventEmitter {
  constructor() {
    super();
    this.readyState = 0;
  }

  async send(data) {
    data = EJSON.parse(data);
    const { id, params } = data;
    await new Promise(resolve => setTimeout(resolve, params[0]));
    this.emit('message', {
      data: EJSON.stringify({
        msg: 'ready',
        subs: [id],
      }),
    });
  }
}

describe('Test DDP Lifecycle', () => {
  let client;
  before(() => {
    registerAll();
    client = new MeteorClient({ connectionDetails: {} });
    client._ddpClient.socket = new FakeSocket();
    client._ddpClient._prepareHandlers();
  });

  after(() => {});

  describe('Test subscriptions', () => {
    it('Should subscribe when the socket is connected', async () => {
      client._isConnected = true;
      client._ddpClient.socket.readyState = Websocket.OPEN;
      await client.subscribe('', [0], {}, '_sub', 100);
    });

    it('Should throw when the socket is not connected', async () => {
      client._ddpClient.socket.readyState = Websocket.CLOSING;
      try {
        await client.subscribe();
        assert.fail('did not throw');
      } catch (error) {
        assert.strictEqual(error.message, 'Not connected');
      }
    });

    it('Should throw when the socket gets disconnected', (done) => {
      client._isConnected = true;
      client._ddpClient.socket.readyState = Websocket.OPEN;
      client.subscribe('', [100], {}, '_sub', 200)
        .then(assert.fail)
        .catch((err) => {
          assert.strictEqual(err, DDPClient.ERRORS.DISCONNECTED);
          done();
        });
      client._isConnected = false;
      client._ddpClient.emit('socket-close');
    });

    it('Should throw if the server does not respond in time', () => {
      client._isConnected = true;
      client._ddpClient.socket.readyState = Websocket.OPEN;
      client.subscribe('', [300], {}, '_sub', 100).then(assert.fail)
        .catch((err) => {
          assert.strictEqual(err.message, 'Subscription timeout');
        });
    });
  });
});
