/*
eslint-disable no-underscore-dangle
*/
import Service, { state as serviceState } from '../service';
import ImmediateScheduler from '../schedulers/immediateScheduler';
import MessengerMixin from '../messengerMixin';
import { sinkType, compositionType } from './constants';

const entities = ['devices', 'meetings', 'sinks', 'compositions', 'layouts', 'sources', 'institutes'];
const actions = ['added', 'changed', 'removed'];

/**
 * The Meteor service extends the base service in a couple of ways:
 *
 * It automatically manages data listeners that are available on the instance.
 *
 * E.g. you can inherit from this class and add a _onSinkChanged method
 * and this will automatically set up and tear down the binding.
 *
 * (//todo replace caching and events with minimongo, as soon as that gets transferred on NPM)
 */
class MeteorService extends Service {
  constructor({
    id,
    meteorClient,
  }) {
    super({ id });
    this._meteorClient = meteorClient;
  }

  get mixins() {
    return [
      ...super.mixins,
      MessengerMixin,
      ImmediateScheduler,
    ];
  }

  get isSynchronous() {
    return false;
  }

  get applyActions() {
    return this.isActive;
  }

  get hasCache() {
    return !!this.cache;
  }

  get supportedCompositionTypes() {
    return Object.values(compositionType);
  }

  get supportedSinkTypes() {
    return Object.values(sinkType);
  }

  get supportedRPCCommands() {
    return [];
  }

  get meteorClientListeners() {
    if (!this._meteorListeners) {
      this._meteorListeners = new Map();

      for (const entity of entities) {
        for (const action of actions) {
          this._meteorListeners.set(`${entity}:${action}`, this._trigger.bind(this, entity, action));
        }
      }

      this._meteorListeners.set(`${this._device._id}:rpc`, async (rpc) => {
        if (this.supportedRPCCommands.includes(rpc.command) && this.applyActions) {
          await this._triggerRPCCommand(rpc);
        }
      });
    }
    return this._meteorListeners;
  }

  async _triggerRPCCommand(rpc) {
    if (rpc.cancelDefaultResponse) {
      rpc.cancelDefaultResponse();
    }

    let result;
    let error;
    try {
      result = await this._onRpcAdded(rpc);
      error = false;
      if (typeof result !== 'object' || result === null) {
        result = { result };
      }
    } catch (err) {
      this._log('error', `Client RPC failed due to ${err.message}`);
      error = true;
      result = { error: err.message };
    } finally {
      this._meteorClient.markRPCDone({
        rpcId: rpc._id,
        result,
        error,
      });
    }
  }

  // XXX: replace with minimongo observers?
  async _trigger(entity, action, record, oldFields, clearedFields, newFields) {
    const method = `_on${entity.charAt(0).toUpperCase() + entity.slice(1, entity.length - 1)}${action.charAt(0).toUpperCase() + action.slice(1, action.length)}`;

    if (
      this.isActive
      && (method in this
      || (
        this.hasCache
        && this.cache._collections.includes(entity)
        && (action === 'added' || action === 'removed')
      ))
    ) {
      if (entity === 'rpcs') {
        if (!this.supportedRPCCommands.includes(record.command)) {
          return true;
        }
      }

      if (this.hasCache && action === 'added' && this.cache[entity].hasOwnProperty(record._id)) {
        return true;
      }

      if (this.hasCache
        && action !== 'added'
        && this.cache._collections.includes(entity)
        && !this.cache[entity].hasOwnProperty(record._id)) {
        return true;
      }

      if (entity === 'compositions' && action === 'added') {
        if (!this.supportedCompositionTypes.includes(record.type)) {
          this._log('verbose', 'Composition type not supported, skipping');
          return true;
        }
      }

      if (entity === 'sinks' && action === 'added') {
        if (!this.supportedSinkTypes.includes(record.type)) {
          this._log('verbose', 'Sink type not supported, skipping');
          return true;
        }

        if (this.hasCache && this.cache.compositions) {
          const composition = this.cache.findCompositionOnSink(record);
          if (!composition || !this.supportedCompositionTypes.includes(composition.type)) {
            this._log('verbose', 'Sink has no supported composition type, skipping');
            return true;
          }
        }
      }

      if (this.hasCache && this.cache.hasOwnProperty(entity)) {
        switch (action) {
        case 'added':
          this.cache.add(entity, record);
          break;
        case 'removed':
          this.cache.remove(entity, record);
          break;
        default:
        }
      }

      if (entity === 'compositions' && action === 'added' && this.hasCache && this.cache.sinks && this._meteorClient._ddpClient.collections.sinks) {
        this._log('verbose', 'Adding sinks that were possible false negatives');

        await Promise.all(Object.values(this._meteorClient._ddpClient.collections.sinks)
          .filter(sink => (this.supportedSinkTypes.includes(sink.type)
                    && sink.defaultCompositionId === record._id
                    && !this.cache.sinks[sink._id]))
          .map(sink => this._trigger('sinks', 'added', sink)));
      }

      return this.scheduler.schedule(method, record, oldFields, clearedFields, newFields);
    }

    return true;
  }

  _installMeteorListeners() {
    // Listen to all device related data,
    // save listeners in a map to later stop listening if this instance gets destroyed
    this._log('verbose', 'Installing meteor listeners');
    this.meteorClientListeners.forEach((listener, event) => {
      this._meteorClient.on(event, listener);
    });
  }

  _uninstallMeteorListeners() {
    this._log('verbose', 'Uninstalling meteor listeners');
    this.meteorClientListeners.forEach((listener, event) => {
      this._meteorClient.off(event, listener);
    });
  }

  set isInFallback(isInFallback) {
    this._isInFallback = isInFallback;
  }

  get isInFallback() {
    return this._isInFallback;
  }

  get meteorConnected() {
    return this._meteorClient.isConnected;
  }

  isSupportedSink(sink) {
    return this.supportedSinkTypes.includes(sink.type);
  }

  async _onStart() {
    this.isInFallback = false;

    this._installMeteorListeners();

    if (this.hasCache) {
      await this.cache.start({
        cancelToken: this.getCancelToken(serviceState.STARTED),
        retryTimer: 2000,
        retries: 30,
      });
    }

    await super._onStart();
  }

  async _onStop() {
    this._uninstallMeteorListeners();

    if (this.hasCache) {
      await this.cache.stop();
    }

    await super._onStop();
  }

  activateFallback() {
    this.isInFallback = true;
  }

  deactivateFallback() {
    this.isInFallback = false;
  }

  _anyChanged(fields, newFields) {
    return fields.reduce((allChanged, field) => allChanged || newFields.hasOwnProperty(field), false);
  }

  _allChanged(fields, newFields) {
    return fields.reduce((allChanged, field) => allChanged && newFields.hasOwnProperty(field), true);
  }

  _onDestroy() {
    delete this._meteorClient;
    delete this._meteorListeners;

    if (this.hasCache) {
      this.cache.destroy();
      delete this.cache;
    }

    return super._onDestroy();
  }
}

export default MeteorService;
