/**
 * (C) 2015 Copyright BARCO NV
 * All rights reserved. Reproduction in whole or in part is
 * prohibited without the written consent of the copyright owner.
 *
 * Part of Barco WeConnect
 */

/*
eslint no-underscore-dangle: ["error", {
  "allow": ["_device", "_id"],
  "allowAfterThis": true,
  "allowAfterSuper": true,
}]
*/
import * as os from 'os';
import { URL } from 'url';
import { serializeError } from 'serialize-error';

import PreviewServer from './previews/server';
import MeteorAgent from '../device-agent/agent';
import MeteorClient from './ddp/meteorClient';
import { getIpAddress } from '../utils/network';
import serviceRegistry from './serviceRegistry';
import { Service, state as serviceState } from './service';
import MessengerMixin from './messengerMixin';
import SynchronousScheduler from './schedulers/synchronousScheduler';
import { rebootSystem } from '../utils/utils';

/**
 This class does the initial set-up and connects to Meteor.
 It then creates a new Meteor Agent for every connected Device

 Once we are connected the EDU proxy works like this:
 - We check the version number, to see if we're compatible with the server (and update if not)
 - We connect to Meteor, authenticate and subscribe for the initial data (done by MeteorClient)
 - For each device we receive from Meteor, we create a Meteor agent.
 - Start the preview server
 - The Meteor Agent creates an NRC API client (cngClient) instance and subscribes for device data from the NRC
 - When all initial data is received AND we are connected to the NRC, we start an initialization phase
 - The initialization phase will force the initial set of meteor data onto the NRC and merge the sources
 - Once initialized, we listen to incoming events from Meteor
  and the NRC (only for sources) and apply the changes accordingly

 Most of the above flow is event-driven and therefore pretty de-coupled.

 If the connection with Meteor is lost, all agents are destroyed. If the connection comes back up, they are re-created.
 If the connection with a CNG is lost,
 the agent is not destroyed and deals with this by re-initializing the device if the connection is restored.

 For more info, we refer to the following flowcharts:
 XXX (todo)
 */

class EduProxy extends Service {
  constructor() {
    super({});

    // Maps device ids to a bridge
    this.agents = {};

    const configManager = serviceRegistry.get('configManager');

    this._meteorListeners = {
      'devices:added': this._createDeviceAgent.bind(this),
      deviceSubscriptionReady: this._onDeviceSubscriptionReady.bind(this),
    };

    configManager.watch('previewServer', () => {
      this._log('info', 'Preview server config changed, scheduling recreation of preview server');
      this.scheduler.schedule(async () => {
        this._log('info', '(Re-)create preview server');
        this._createOrEditPreviewServer();
        if (this.isStarted) {
          this._startPreviewServer();
        }
      });
    });

    configManager.watch('meteor.server', 'device.proxy', () => {
      this._log('info', 'Meteor config changed, scheduling recreation of meteor client');
      this.scheduler.schedule(async () => {
        this._log('info', '(Re-)create meteor client');
        await this._destroyMeteorClient();
        this._createMeteorClient();

        if (this.isStarted) {
          this._startMeteorClient();
        }
      });
    });
  }

  get mixins() {
    return [
      ...super.mixins,
      MessengerMixin,
      SynchronousScheduler,
    ];
  }

  /**
     * Updates the proxy ip
     */
  updatePublicIp() {
    this._log('info', 'Updating public IP"');
    let publicIp = serviceRegistry.get('configManager').getSetting('device.publicIp');

    if (publicIp && publicIp.trim()) {
      publicIp = publicIp.trim();
      this._log('verbose', `Public IP address found in config: "${publicIp}"`);
    } else {
      // use private IP address of this system
      let ip = getIpAddress(this._logger);

      // In case nothing found, use the hostname.
      if (ip) {
        this._log('verbose', `Public IP not found, using private IP address: "${ip}"`);
      } else {
        ip = os.hostname();
        this._log('verbose', `Public IP address not found and private ip not found, so using hostname: "${ip}"`);
      }

      publicIp = ip;
    }
    this._log('info', `Public IP address to be used for outgoing URLs set to "${publicIp}"`);

    serviceRegistry.get('configManager').update('device.publicIp', publicIp);
  }

  async _onStart() {
    const message = this._showMessage({ text: 'connecting to cloud' });

    try {
      const promises = [
        this._startPreviewServer(),
        this._startMeteorClient(),
      ];

      await Promise.all(promises);

      await super._onStart();
    } finally {
      message.hide();
    }
  }

  _createOrEditPreviewServer() {
    const previewServerSettings = serviceRegistry.get('configManager').getSetting('previewServer');

    if (previewServerSettings) {
      let ipAddress;
      if (previewServerSettings.useHostname) {
        ipAddress = os.hostname();
      } else {
        const deviceSettings = serviceRegistry.get('configManager').getSetting('device');
        ipAddress = deviceSettings.publicIp;
      }

      if (!this.previewServer && previewServerSettings) {
        this._log('info', 'Creating preview server');

        const options = {
          port: previewServerSettings.port,
          ipAddress,
        };

        this.previewServer = new PreviewServer(options);
        serviceRegistry.set('previewServer', this.previewServer);
      } else {
        this.previewServer.port = previewServerSettings.port;
        this.previewServer.ipAddress = ipAddress;
      }
    }
  }

  _startPreviewServer() {
    if (this.previewServer) {
      this._log('info', 'Starting preview server');
      return this.previewServer.start({ retryTimer: 5000, cancelToken: this.getCancelToken(serviceState.STARTED) });
    }
    return Promise.resolve();
  }

  _stopPreviewServer() {
    if (this.previewServer) {
      this._log('info', 'Stopping preview server');
      return this.previewServer.stop();
    }
    return Promise.resolve();
  }

  async _destroyPreviewServer() {
    if (this.previewServer) {
      this._log('info', 'Destroying preview server');
      await this.previewServer.destroy();
      delete this.previewServer;
    }
  }

  _createMeteorClient() {
    const meteorConfig = serviceRegistry.get('configManager').getSetting('meteor');
    if (!this.meteorClient && meteorConfig) {
      this._log('info', 'Creating Meteor client');
      // Create the Meteor DDP client and connect if desired

      const proxy = serviceRegistry.get('configManager').getSetting('device.proxy');
      const server = new URL(meteorConfig.server);
      const ssl = server.protocol.includes('https');
      const port = server.port || (ssl ? 443 : 80);

      this.meteorClient = new MeteorClient({
        connectionDetails: {
          autoReconnect: meteorConfig.autoReconnect,
          autoReconnectTimer: meteorConfig.autoReconnectTimer,
          host: server.hostname,
          port,
          ssl,
          proxy,
        },
      });

      serviceRegistry.set('meteorClient', this.meteorClient);
    }
  }

  async _startMeteorClient() {
    if (this.meteorClient) {
      this._log('info', 'Starting Meteor client');

      // For each device assigned to purple, create a agent
      this._uninstallMeteorListeners();
      this._installMeteorListeners();

      // Connect to meteor
      await this.meteorClient.start({ cancelToken: this.getCancelToken(serviceState.STARTED) });
    }
  }

  _installMeteorListeners() {
    if (this.meteorClient) {
      this._log('debug', 'Installing meteor listeners');
      Object.keys(this._meteorListeners).forEach((event) => {
        this.meteorClient.on(event, this._meteorListeners[event]);
      });
    }
  }

  _uninstallMeteorListeners() {
    if (this.meteorClient) {
      this._log('debug', 'Uninstalling meteor listeners');
      Object.keys(this._meteorListeners).forEach((event) => {
        this.meteorClient.removeListener(event, this._meteorListeners[event]);
      });
    }
  }

  _onDeviceSubscriptionReady() {
    Object.keys(this.agents).forEach((deviceId) => {
      if (!this.meteorClient._ddpClient.collections.devices.hasOwnProperty(deviceId)) { // eslint-disable-line
        this.agents[deviceId].destroy();
      }
    });
  }

  _agentFor(device) {
    // Treat devices with different id's but same IP's equal to each other
    // This can happen if we quickly remove & re-add a device
    // where the old agent has not yet been fully destroyed
    return this.agents[device._id]
            || Object.values(this.agents)
              .find(agent => (agent._device && agent._device.privateIp === device.privateIp));
  }

  async _createDeviceAgent(device) {
    const myUserId = serviceRegistry.get('configManager').getSetting('credentials.user.id');

    if (device.managingUserId !== myUserId) {
      // Some publication could query devices we don't manage
      return;
    }

    if (this.meteorClient && this.meteorClient.isActive) {
      this._log('info', `Trying to create agent for device: ${device._id}`);

      const existingAgent = this._agentFor(device);

      if (existingAgent) {
        const existingAgentId = existingAgent._device._id;

        if (existingAgent.isDestroying || existingAgentId !== device._id) {
          // If an old meteor agent is still active
          // or not completely destroyed), destroy it and start the new one afterwards.
          // This is necessary in order to avoid duplicate method calls
          this._log('info', `Waiting for old agent of device: ${existingAgentId} to be destroyed, before starting new one`);

          // We have to remove all existing listeners
          // Otherwise the default destroyed listener might delete the newly created agent
          // Since we've got no guarantees on the order of the 'destroyed' callbacks
          existingAgent.removeAllListeners('destroyed');

          existingAgent.on('destroyed', () => {
            this._log('info', `Agent ${device._id} destroyed... Now we can bootstrap new one`);
            delete this.agents[existingAgentId];
            this._createDeviceAgent(device);
          });

          existingAgent.destroy();
        } else {
          this._log('info', `Restarting old agent for device: ${device._id}`);
          // If this is a re-connect, the device object is another one,
          // So we have to reset it in order for it to keep changing
          existingAgent.device = device;
          try {
            await existingAgent.restart({ retries: 0 }); // Allow only one attempt
          } catch (e) {
            // XXX Temporary workaround for endless loop (exception in MeteorCacheManager warm() function where _ddpClient is undefined)
            this._log('error', `Could not restart existing device agent, going to reboot device after 60 seconds !!! Error = ${JSON.stringify(serializeError(e, { maxDepth: 1 }))}`, { serviceLog: true });
            await new Promise((resolve) => {
              this._timer = setTimeout(resolve, 60 * 1000); // Allow logger to flush + TMX to call Houston
            });
            rebootSystem();
          }
        }
      } else {
        this._log('info', `Actually creating agent for device: ${device._id}`);
        this.agents[device._id] = new MeteorAgent({
          device,
          meteorClient: this.meteorClient,
        });
        this.agents[device._id].on('destroyed', () => {
          this._log('info', `Agent ${device._id} destroyed... cleaning up`);
          delete this.agents[device._id];
        });

        try {
          // Retry timeout should be quite extensive to make it less likely for devices to get stuck in a loop
          // continuously trying to start the agents while overloading the backend.
          // Ideally, some kind of back-off mechanism could be employed here
          this.agents[device._id].start({ retryTimer: 15 * 1000 });
        } catch (e) {}
      }
    }
  }

  _destroyDeviceAgents() {
    this._log('info', 'Destroying device agents');
    return Promise.all(Object.values(this.agents).map(agent => agent.destroy()));
  }

  async _stopMeteorClient() {
    if (this.meteorClient) {
      this._log('info', 'Stopping Meteor Client');
      this._uninstallMeteorListeners();
      await this.meteorClient.stop();
    }
  }

  async _destroyMeteorClient() {
    if (this.meteorClient) {
      this._log('info', 'Destroying meteor client');
      await this.meteorClient.destroy();
      delete this.meteorClient;
    }
  }

  async _onStop() {
    const promises = [
      this._stopMeteorClient(),
      this._stopPreviewServer(),
      this._destroyDeviceAgents(),
    ];
    await Promise.all(promises);
    await super._onStop();
  }

  async _onDestroy() {
    const promises = [
      this._destroyMeteorClient(),
      this._destroyPreviewServer(),
    ];
    await Promise.all(promises);
    await super._onDestroy();
  }
}

let eduProxy;
let eduProxyStopCounter = 1;

// We use a counter to keep track of how many services want edu proxy to be stopped
// That is, each service that wants to start and stop edu proxy has to request a stop and a start
// For every stop call we increase the counter, for every start call we decrease the counter.
// Thus each service must ensure it only calls stop/start once as necessary.
// If during the start the stop counter > 0, it means there's still a service that want to have it stopped
// So we will not start it until the counter reaches 0.
// We initialize the eduProxyStopCounter to 1, since it's stopped by default until we call start for the first time.
function stopEduProxy() {
  eduProxyStopCounter += 1;
  return eduProxy && eduProxy.stop();
}

async function startEduProxy() {
  try {
    eduProxyStopCounter -= 1;
    if (eduProxyStopCounter > 0) return;

    if (!eduProxy) {
      eduProxy = new EduProxy();
      serviceRegistry.set('eduProxy', eduProxy);
    }
    await eduProxy.start();
  } catch (error) {
    stopEduProxy();
  }
}

export default EduProxy;
export {
  startEduProxy,
  stopEduProxy,
};
