/*
eslint no-console: "off"
*/

import { createLogger, format, transports } from 'winston';
import 'winston-daily-rotate-file';
import { SPLAT } from 'triple-beam';
import * as safeStringify from 'safe-stable-stringify';
import * as path from 'path';
import * as isDev from 'electron-is-dev';

import EventEmitter from '../utils/eventEmitter';
import { createDirectory } from '../utils/utils';
import hideLogs from '../hideLogs';

const eduProxyFormat = format.printf(({
  level, message, service, id, timestamp, ...rest
}) => {
  let formatted = `${timestamp} - [${service}${id ? ` (${id})` : ''}] ${level}: ${message}`;
  const splat = rest[SPLAT];
  if (splat !== undefined) {
    formatted += ` ${splat.map(s => safeStringify(s)).join(' ')}`;
  }
  return formatted;
});

const ignoreHiddenLogs = format((info) => {
  if (isDev && hideLogs.includes(info.service)) {
    // Don't use the hideLogs array in production
    // This is for safety in case we accidentally check in a wrong hideLogs array
    return false;
  }
  return info;
})();

const generateFilename = (dir, filename) => {
  const extendedFilename = `${filename}_%DATE%.log`;
  return path.join(dir, extendedFilename);
};

/**
 * Logger service
 *
 * Creates a winston logger and allows to instantiate instances with a certain prefix.
 * Allows dynamically changing logger configuration
 *
 * Also allows to create logger 'instances'. An instance automatically adds a prefix to every log message.
 * E.g. services create their own instance with the service name as prefix, useful for analysing the log file
 */
class Logger extends EventEmitter {
  constructor({
    dir, filename, serviceFilename, level, maxSize, maxFiles,
  }) {
    super();

    this.transports = {
      consoles: [],
      files: [],
    };
    if (isDev) {
      this.transports.consoles.push(
        new transports.Console({
          handleExceptions: true,
        }),
      );
    }

    if (typeof (maxFiles) === 'string' || maxFiles > 0) {
      if (dir) {
        dir = createDirectory(dir);
      }

      if (!dir) {
        dir = path.join(process.cwd(), '..');
      }

      const fileOptions = {
        handleExceptions: true,
        maxFiles,
        // maxSize is pretty inconsistent in winston and should be handled as follows
        // File --> maxsize https://github.com/winstonjs/winston/blob/master/docs/transports.md#file-transport
        // DailyRotateFile --> maxSize https://github.com/winstonjs/winston-daily-rotate-file#readme
        maxSize,
      };

      this.transports.files.push(
        new transports.DailyRotateFile({
          filename: generateFilename(dir, filename),
          ...fileOptions,
        }),
      );

      if (serviceFilename) {
        this.serviceMetadata = {};
        this.transports.files.push(
          new transports.DailyRotateFile({
            format: format.combine(
              this.getServiceFormat(),
              format.timestamp(),
              format.metadata({ fillExcept: ['message', 'timestamp', 'level'], key: 'context' }),
              format.json(),
            ),
            filename: generateFilename(dir, serviceFilename),
            ...fileOptions,
          }),
        );
      }
    }

    // Configure the logger
    // Levels: { error: 0, warn: 1, info: 2, verbose: 3, debug: 4, silly: 5 }
    this._logger = createLogger({
      level,
      defaultMeta: { service: 'ROOT' },
      format: format.combine(
        ignoreHiddenLogs,
        format.timestamp(),
        eduProxyFormat,
      ),
      transports: [...this.transports.consoles, ...this.transports.files],
      exitOnError: false,
    });
  }

  /**
   * Change logger configuration (level, maxSize, maxFiles)
   * @param {*} param0
   */
  configure({ level, maxSize, maxFiles }) {
    this.setLevel(level);

    if (this.transports.files) {
      this.transports.files.forEach((fileTransport) => {
        if (maxSize) {
          fileTransport.maxSize = maxSize;
        }

        if (maxFiles) {
          fileTransport.maxFiles = maxFiles;
        }
      });
    }
  }

  /**
   * Set log level
   * @param {*} level
   */
  setLevel(level) {
    if (level) {
      this._logger.level = level;
    }
  }

  /**
   * Creates an instance of the logger with a certain prefix bound to it.
   * Useful for services
   */
  getInstance(service, id) {
    return this._logger.child({ service, id });
  }

  getServiceFormat() {
    return format((info) => {
      const { serviceLog, ...rest } = info;
      if (!serviceLog) { return false; }
      return {
        ...rest,
        serviceMetadata: this.serviceMetadata,
      };
    })();
  }

  /**
   * Cleanup
   */
  async destroy() {
    if (!this.isDestroying) {
      this.isDestroying = true;
      const close = new Promise(resolve => this._logger.on('finish', resolve));
      this._logger.end();
      await close;
      delete this._logger;
      delete this.transports;
      this.emit('destroyed');
      super.destroy();
    } else {
      await new Promise((resolve) => {
        this.once('destroyed', resolve);
      });
    }
  }
}

export default Logger;
