/**
 * (C) 2015 Copyright BARCO NV
 * All rights reserved. Reproduction in whole or in part is
 * prohibited without the written consent of the copyright owner.
 *
 * Part of Barco WeConnect
 */

import { Socket } from 'net';
import * as fs from 'fs-extra';
import * as fpath from 'path';
import * as _ from 'lodash';
import * as Boom from 'boom';

import PreviewHandler from './handler';
import { Service } from '../service';
import { RESOURCES_DIR } from '../../files';

const noPreviewImage = fs.readFileSync(fpath.join(RESOURCES_DIR, '/icons/sources/no_preview_available.png'));

/**
 * Preview provider base class (Abstract)
 * Subclassed by dynamic (MJPEG) and static (HTTP) providers
 *
 * Contains mainly shared utility methods
 */
class PreviewProvider extends Service {
  /**
     *
     * @param id, identifier for this provider. Useful for logging and debugging purposes
     * @param validateRequest method to validate (secure) incoming requests
     */
  constructor({ id, validateRequest = () => true }) {
    super({ id });

    const params = { id, validateRequest };
    Object.keys(params).forEach((param) => {
      if (params[param] === undefined) throw new Error(`Please provide a valid ${param}`);
      this[`_${param}`] = params[param];
    });

    this._previewNumber = 0;
    this._previewData = {};

    this._log('verbose', 'PreviewProvider created.');
  }

  set previewDataPNG(data) {
    this._previewData.dataPNG = data;
  }

  set previewInitialDataPNG(data) {
    this._previewData.initialDataPNG = data;
  }

  set previewDataJPG(data) {
    this._previewData.dataJPG = data;
  }

  handleIncomingRequest(req, res) {
    if (!this.isActive || (this._validateRequest && !this._validateRequest(req, res))) {
      throw Boom.forbidden();
    }
    return true;
  }

  closeConnections() {}

  /**
     * Tears down and cleans up the provider
     */
  _onDestroy() {
    // Remove listeners
    delete this._validateRequest;
    Object.keys(this._previewData).forEach((key) => { delete this._previewData[key]; });
    return super._onDestroy();
  }
}

/**
 * Static (normal HTTP) preview providers
 */
class StaticPreviewProvider extends PreviewProvider {
  constructor(params) {
    super(params);
    this.previewDataPNG = noPreviewImage;
  }

  /**
     * Preview
     * @param data
     */
  set previewDataPNG(data) {
    if (_.isString(data) && data.length) {
      data = Buffer.from(data, 'base64');
      // XXX: change to Buffer.from when node 6 arrives
    } else {
      data = noPreviewImage;
    }

    super.previewDataPNG = data;

    this.emit('newPreview');

    if (this._previewNumber === 0) {
      // Cache the initial preview, this gets send out when the connection is lost
      this.previewInitialDataPNG = data;
    }

    this._reactivateInitialPreviewTimer();

    this._previewNumber += 1;
  }

  /**
     * Handle an incoming request. This means serving the last thumbail available
     * @param req
     * @param res
     */
  handleIncomingRequest(req, res) {
    this._log('silly', 'Serving last preview for path');
    if (super.handleIncomingRequest(req, res)) {
      this._serveLastPreview(req, res);
    }
  }

  /**
     * Serve the last preview if available, otherwise serve a default 'no preview available' image
     * @param req
     * @param res
     * @private
     */
  _serveLastPreview(req, res) {
    // Send the last preview
    this._log('silly', `Serving preview of ${this._previewData.dataPNG.length} octets`);
    res.writeHead(200, {
      'Content-Type': 'image/png',
      'Content-Length': this._previewData.dataPNG.length,
      'Access-Control-Allow-Origin': '*',
    });
    res.end(this._previewData.dataPNG);
  }

  /**
     * In case we do not receive a new preview after 30s, we probably lost the connection, so reuse the initial preview
     * @private
     */
  _reactivateInitialPreviewTimer() {
    // First make sure to clear old timers if any
    this._stopReuseInitialPreviewTimer();
    if (this.isActive) {
      this._startReuseInitialPreviewTimer();
    }
  }

  /**
     * If an initial preview is available, restore it as the last one.
     * @private
     */
  _reuseInitialPreview() {
    if ((this._previewNumber < 2) || (!this.isActive)) {
      return;
    }

    this._log('verbose', 'Restoring initial preview');

    if (this._previewData.initialDataPNG) {
      this.previewDataPNG = this._previewData.initialDataPNG;
    }
  }

  _startReuseInitialPreviewTimer() {
    this._stopReuseInitialPreviewTimer();
    this._reuseInitialPreviewTimer = setTimeout(() => { this._reuseInitialPreview(); }, 15000);
  }

  _stopReuseInitialPreviewTimer() {
    if (this._reuseInitialPreviewTimer) {
      clearTimeout(this._reuseInitialPreviewTimer);
      delete this._reuseInitialPreviewTimer;
    }
  }

  _onStart() {
    this._startReuseInitialPreviewTimer();
    return super._onStart();
  }

  _onStop() {
    this._stopReuseInitialPreviewTimer();
    return super._onStop();
  }
}

/**
 * Creates a streaming (MJPEG) preview provider.
 * In contrast to the static provider, this class manages a list of clients (called: handlers)
 */
class StreamingPreviewProvider extends PreviewProvider {
  constructor({
    id, validateRequest, mjpegServerIPAddress, mjpegServerPort,
  }) {
    super({ id, validateRequest });

    const params = { mjpegServerIPAddress, mjpegServerPort };
    Object.keys(params).forEach((param) => {
      if (params[param] === undefined) throw new Error(`Please provide a valid ${param}`);
      this[`_${param}`] = params[param];
    });

    this._mjpegSock = undefined;
    this._mjpegSockTargetStateIsConnected = true;
    this._previewHandlerList = [];
    this._msgNbr = 0;
  }

  get numberOfThumnailHandlers() {
    return this._previewHandlerList.length;
  }

  get numberOfSubscribedReceivers() {
    return this._previewHandlerList.length;
  }

  /**
     * Close all client connections
     */
  closeConnections() {
    super.closeConnections();
    this._log('verbose', `About to close ${this.numberOfSubscribedReceivers} connections`);
    this._previewHandlerList.forEach((handler) => {
      handler.removeAllListeners('destroyed');
      handler.close();
    });
    this._previewHandlerList = [];
    this._log('verbose', 'Closed all connections');
  }

  /**
     * Handle an incoming express connection
     * @param req
     * @param res
     */
  handleIncomingRequest(req, res) {
    this._log('debug', 'New incoming request for path , adding a preview handler');
    if (super.handleIncomingRequest(req, res)) {
      this._addPreviewHandler(req, res);
    }
  }

  /**
     * Add a client connection
     * @param req
     * @param res
     * @private
     */
  _addPreviewHandler(req, res) {
    const previewHandlerID = `${this.numberOfThumnailHandlers}`;
    this._log('verbose', `Adding previewHandler ID = ${previewHandlerID}`);

    const handler = new PreviewHandler({ id: previewHandlerID, req, res });
    handler.on('destroyed', () => {
      this._log('debug', `Removing preview handler ID = ${previewHandlerID}`);
      const index = this._previewHandlerList.indexOf(handler);
      if (index > -1) {
        this._previewHandlerList.splice(index, 1);
      }
    });

    handler.start();
    this._previewHandlerList.push(handler);

    this.emit('newSubscriber', this.numberOfThumnailHandlers);
  }

  /**
     * Send the current preview to all clients (handlers)
     * @private
     */
  _sendPreviewPartToAllHandlers() {
    if (!this.isActive) {
      return;
    }

    this._previewHandlerList.forEach((handler) => {
      handler.sendPreviewPart(this._previewData.dataJPG);
    });
  }

  /**
     * handle socket closes/drops from CNG MJPEG server
     * @private
     */
  _cleanupOnSocketClosed() {
    this._log('verbose', 'Cleaning up MJPEG slave socket.');
    // Get rid of the socket
    if (this._mjpegSock) {
      this._log('verbose', `================>>>>>>>>    : About to destroy MJPEG slave socket ${this._mjpegSock._id}`);
      this._mjpegSock.removeAllListeners();
      this._mjpegSock.destroy();
      delete this._mjpegSock;
      this._log('verbose', '================>>>>>>>>    : MJPEG slave socket destroyed');
    }

    if (this._mjpegSockTargetStateIsConnected && this.isActive) {
      // Trigger reconnection attempt after 15 seconds
      this._log('verbose', 'Scheduling MJPEG slave reconnect in 15s.');
      this._startConnectTimer(15000);
    }
  }

  /**
     * Process incoming MJPEG data from cinergy and send to clients
     * @param data
     * @private
     */
  _processReceivedMJpegData(data) {
    if (!this.isActive) {
      return;
    }

    if (!(this._msgNbr % 50)) {
      this.emit('newPreview');
    }

    this._msgNbr += 1;
    this.previewDataJPG = data;
    this._previewNumber += 1;
    this._sendPreviewPartToAllHandlers();
  }

  /**
     * Connect to the MJPEG server of cinergy
     * @private
     */
  _connectToMjpegServer() {
    if (!this.isActive) {
      return;
    }

    if (this._connectTimer) {
      clearTimeout(this._connectTimer);
    }

    this._log('verbose', `Trying to connect to MJPEG server at ${this._mjpegServerIPAddress}:${this._mjpegServerPort}`);
    this._mjpegSock = new Socket();

    const socketId = Math.floor(Math.random() * 7451) + 50;
    this._mjpegSock._id = socketId;

    this._mjpegSock.on('error', (error) => {
      this._log('error', `Mjpeg slave socket ${socketId} error`, error);
    });

    this._mjpegSock.on('close', (error) => {
      this._log('debug', `MJPEG slave socket ${socketId} close`, error);
      this.emit('disconnected');
      this._cleanupOnSocketClosed();
    });

    this._mjpegSock.on('data', (data) => {
      this._log('silly', `MJPEG slave socket ${socketId} data`);
      this._processReceivedMJpegData(data);
    });

    this._mjpegSock.connect(this._mjpegServerPort, this._mjpegServerIPAddress, () => {
      this.emit('connected');
      this._log('debug', `MJPEG slave socket ${socketId} connect`);
      this._log('verbose', `Connected to Mjpeg server at ${this._mjpegServerIPAddress}:${this._mjpegServerPort} with id ${this._mjpegSock._id}`);
    });
  }

  _onStart() {
    this._mjpegSockTargetStateIsConnected = true;
    // Give the server some time to startup before we connect
    this._startConnectTimer();
    return super._onStart();
  }

  _onStop() {
    this._mjpegSockTargetStateIsConnected = false;
    this._stopConnectTimer();
    this._cleanupOnSocketClosed();
    this.closeConnections();
    return super._onStop();
  }

  _startConnectTimer(time = 2000) {
    this._stopConnectTimer();
    this._connectTimer = setTimeout(() => { this._connectToMjpegServer(); }, time);
  }

  _stopConnectTimer() {
    if (this._connectTimer) {
      clearTimeout(this._connectTimer);
      delete this._connectTimer;
    }
  }
}

export {
  PreviewProvider,
  StaticPreviewProvider,
  StreamingPreviewProvider,
};
