/**
 * (C) 2015 Copyright BARCO NV
 * All rights reserved. Reproduction in whole or in part is
 * prohibited without the written consent of the copyright owner.
 *
 * Part of Barco WeConnect
 */
import { URL } from 'url';
import * as Boom from 'boom';

import Server from '../server';

/**
 * A preview server.
 * This class will register paths and route incoming requests to the correct preview providers.
 */
class PreviewServer extends Server {
  constructor({
    ipAddress = '127.0.0.1',
    port,
  }) {
    super({ id: `${ipAddress}:${port}`, port });

    this._ipAddress = ipAddress;
    this._previewProviders = {};
    this._log('verbose', 'New PreviewProviderManager created.');

    // register catch all path
    this.app.get('*', (req, res) => {
      this._handleIncomingRequest(req, res);
    });
  }

  get numberOfThumnailProviders() {
    return Object.keys(this._previewProviders).length;
  }

  set hostname(hostname) {
    this._hostname = hostname;
  }

  set ipAddress(ipAddress) {
    this._ipAddress = ipAddress;
  }

  get hostname() {
    return this._hostname || this._ipAddress;
  }

  /**
   * Route an incoming request to the correct preview provider, if it exists
   * @param req
   * @param res
   * @private
   */
  _handleIncomingRequest(req, res) {
    const path = new URL(`http://127.0.0.1${req.url}`).pathname;
    this._log('silly', `New incoming request for path "${path}"`);
    const provider = this._previewProviders[path];
    if (!provider) {
      this._log('error', `Could not serve request for unknown path "${path}"`);
      throw Boom.notFound();
    }

    try {
      provider.handleIncomingRequest(req, res);
    } catch (error) {
      this._log('error', `Error while handing request for "${path}"`, error);
      throw error;
    }
  }

  /**
   * Add a preview provider for the given path
   * @param path
   * @param provider
   * @returns {*}
   */
  addPreviewProviderIfNeeded(path, provider) {
    if (!path) {
      this._log('error', 'Please provide a valid preview path');
      return undefined;
    }

    if (!provider) {
      this._log('error', 'Please provide a valid preview provider');
      return undefined;
    }

    if (path[0] !== '/') {
      path = `/${path}`;
    }

    if (!this._previewProviders.hasOwnProperty(path)) {
      // If there is no preview provider yet for this source, create one and install the path.
      provider.on('destroyed', () => {
        // Auto delete the previewprovider if it gets destroyed
        this.deletePreviewProvider(path);
      });

      this._previewProviders[path] = provider;
      this._log('verbose', `Added previewProvider for path = ${path}`);
    } else {
      this._log('verbose', `Not adding previewProvider for path = ${path}, path already present`);
    }

    return this._urlForPath(path);
  }

  _urlForPath(path) {
    return `http://${this.hostname}:${this._port}${path}`;
  }

  /**
   * Delete all preview providers
   */
  deleteAllPreviewProviders() {
    this._log('verbose', 'Deleting all preview providers');
    this.deleteAllPreviewProvidersIncluding();
  }

  /**
   * Delete all preview providers that include the given string in their path
   * @param part
   */
  deleteAllPreviewProvidersIncluding(part) {
    if (part) {
      this._log('verbose', `Deleting all preview providers for path including ${part}`);
    }

    Object
      .keys(this._previewProviders)
      .filter(path => !part || path.includes(part))
      .forEach(this.deletePreviewProvider.bind(this));
  }

  /**
   * Stop listening to incoming traffic on the given path.
   * @param path
   */
  deletePreviewProvider(path) {
    if (this._previewProviders.hasOwnProperty(path)) {
      this._log('verbose', `Deleting preview provider with path ${path}`);
      this._previewProviders[path].destroy();
      delete this._previewProviders[path];
    }
  }

  getNumberOfSubscribedReceiversForPath(path) {
    if (!this._previewProviders[path]) {
      return 0;
    }
    return this._previewProviders[path].numberOfSubscribedReceivers;
  }

  async _onStop() {
    // Stop listening and close the server
    this.deleteAllPreviewProviders();
    await super._onStop();
  }
}

export default PreviewServer;
