import { Service } from './service';
import { execParallel, execSynchonousAsync, waitMaximum } from '../utils/utils';

class ProcessManager extends Service {
  constructor({
    id,
    startCommands = [],
    stopCommands = [],
    synchronous = false,
    timeoutMs = 5000,
    maxRetries = 3,
  }) {
    super({
      id,
    });

    this._startCommands = startCommands;
    this._stopCommands = stopCommands;
    this._synchronous = synchronous;
    this._timeoutMs = timeoutMs;
    this._maxRetries = maxRetries;
  }

  get isSynchronous() {
    return this._synchronous;
  }

  async _onStart() {
    await this._execCommands(this._startCommands);
    return super._onStart();
  }

  async _onStop() {
    if (!this.isDestroying) {
      await this._execCommands(this._stopCommands);
    }
    return super._onStop();
  }

  async _execCommands(commands, attempt = 0) {
    this._log('verbose', `Executing commands: ${commands.join(',')}`);
    try {
      const exec = this.isSynchronous ? execSynchonousAsync : execParallel;
      await waitMaximum(exec(commands), this._timeoutMs);
      this._log('verbose', `Successfully executed: ${commands.join(',')}`);
    } catch (error) {
      attempt += 1;
      this._log('error', `Error while executing commands: ${commands.join(',')}`, error);
      if (attempt <= this._maxRetries) {
        this._log('info', `Retrying: attempt ${attempt}`);
        return this._execCommands(commands, attempt);
      } else {
        this._log('error', 'Maximum retries reached, we\'re stuck');
      }
    }
  }

  _onDestroy() {
    delete this._startCommands;
    delete this._stopCommands;
    return super._onDestroy();
  }
}

export default ProcessManager;
