import { Service } from './service';

/**
 * Monitor if a process (via its process id) is still running.
 * Notifies if the process stops running.
 *
 * This works by polling the process, so the notification may have up to 5s delay.
 */
class ProcessMonitor extends Service {
  constructor(pid) {
    super({ id: pid });

    const params = { pid };
    Object.keys(params).forEach((param) => {
      if (params[param] === undefined) throw new Error(`Please provide a valid ${param}`);
      this[`_${param}`] = params[param];
    });
  }

  _onStart() {
    this._startCheckInterval();
    return super._onStart();
  }

  _onStop() {
    this._clearCheckIntervalHandle();
    return super._onStop();
  }

  /**
   * Start the interval check
   * @private
   */
  _startCheckInterval() {
    this._checkIntervalHandle = setInterval(
      this.check.bind(this),
      5000,
    );
  }

  _clearCheckIntervalHandle() {
    if (this._checkIntervalHandle) {
      clearInterval(this._checkIntervalHandle);
      this._checkIntervalHandle = null;
    }
  }

  isRunning(pid) {
    pid = pid || this._pid;

    try {
      return process.kill(pid, 0);
    } catch (error) {
      return error.code === 'EPERM';
    }
  }

  check() {
    this._log('verbose', 'Checking...');
    if (!this.isRunning()) {
      this._log('debug', 'Stopped...');
      this.emit('stopped');
      this.stop();
    } else {
      this._log('debug', 'Still running...');
      this.emit('running');
    }
  }
}

export default ProcessMonitor;
