import { Service, stateDefinitions, state } from './service';

/**
 * A Proxied service is a type of service that has to proxy its listeners to another object.
 * This is particulary useful for API's running over a socket.
 *
 * E.g. In the case of cinergy, You can subscribe for events on the cinergy client.
 * The proxied service will buffer the listeners and subscribe whenever Cinergy is ready to receive event subscribers.
 * This also allows elegant recovery whenever connection with cinergy is lost and recovered.
 */
class ProxiedService extends Service {
  constructor(options = {}) {
    options.eventOptions = {
      ...options.eventOptions,
      newListener: true,
      removeListener: true,
    };
    super(options);

    // In case it's the first listener for an event, start subscribing.
    this.on('newListener', (event) => {
      if (this.listenerCount(event) === 0 && this.readyToSubscribe) {
        if (this.proxyMap.hasOwnProperty(event)) {
          this._subscribeForNotification(this.proxyMap[event], event);
        } else if (this.proxyMap.hasOwnProperty('*')) {
          this._subscribeForNotification(this.proxyMap['*'], event);
        }
      }
    });

    this.on('removeListener', (event) => {
      if (event) {
        if (this.proxyMap.hasOwnProperty[event] && this.listenerCount(event) === 0 && this.readyToSubscribe) {
          this._unsubscribeForNotification(this.proxyMap[event], event);
        } else if (this.proxyMap.hasOwnProperty['*'] && this.listenerCount(event) === 0 && this.readyToSubscribe) {
          this._unsubscribeForNotification(this.proxyMap['*'], event);
        }
      }
      return this;
    });

    // We can only subscribe to events once we are authenticated.
    // However, a client can already add listeners before authenticating.
    // Therefore, initialize subscriptions on authentication for already attached listeners
    this.on(this.readyToSubscribeEvent, () => {
      Object.keys(this.proxyMap)
        .filter(event => this.listenerCount(event) > 0)
        .forEach((event) => {
          this._subscribeForNotification(this.proxyMap[event], event);
        });

      if (this.proxyMap.hasOwnProperty('*')) {
        // this.eventNames()
        Object.keys(this._events || {})
          .filter(event => !this.proxyMap.hasOwnProperty(event) && this.listenerCount(event) > 0)
          .forEach((event) => {
            this._subscribeForNotification(this.proxyMap['*'], event);
          });
      }
    });
  }

  get proxyMap() {
    return {};
  }

  get readyToSubscribeEvent() {
    return stateDefinitions[state.STARTED].event;
  }

  get readyToSubscribe() {
    return this.isStarted;
  }

  _subscribeForNotification() {}

  _unsubscribeForNotification() {}
}

export default ProxiedService;
