// eslint-disable no-await-in-loop
import * as Boom from 'boom';
import * as url from 'url';
import { version, MESSAGE, ACTION } from '@barco/network-test';
import { TestRunner, getDefaultTestSuite } from '@barco/network-test/dist/tests';
import { CommunicationChannel } from '@barco/network-test/dist/utils/messages';

import DebianRepoAccessTest from '../network-tests/DebianRepoAccess';
import SystemVersionTest from '../network-tests/SystemVersionTest';
import { startEduProxy, stopEduProxy } from '../eduProxy';
import serviceRegistry from '../serviceRegistry';
import ServiceMixin from '../serviceMixin';
import { wrapAsync } from '../server';
import { Random } from '../../utils/random';
import decodeToken from '../../utils/jwt';
import { proxyConfigToUrl } from '../../utils/network';
import { sleep } from '../../utils/utils';

function getIpFromRequest(req) {
  return req.headers['x-forwarded-for'] || req.connection.remoteAddress;
}
function verifyIp(req, ip) {
  const requesterIp = getIpFromRequest(req);
  if (ip !== requesterIp) {
    throw Boom.forbidden('Requester IP invalid');
  }
}

const testSuite = getDefaultTestSuite();
testSuite[0].push(DebianRepoAccessTest, SystemVersionTest);

class PurpleTestRunner extends TestRunner {
  constructor(context, code) {
    super(testSuite, context);
    this.code = code;
  }

  async _prepare() {
    this._message = serviceRegistry.get('messenger').show({ text: 'Preparing for network test' });
    await stopEduProxy();
    await sleep(5000);
    if (this._message) {
      this._message.hide();
      delete this._message;
    }
    this._message = serviceRegistry.get('messenger').show({ text: 'Running network test' });
    return super._prepare();
  }

  async _stop() {
    if (this._message) {
      this._message.hide();
      delete this._message;
    }
    this._message = serviceRegistry.get('messenger').show({ text: 'Finishing network test' });
    await super._stop();
    await sleep(2000);
    if (this._message) {
      this._message.hide();
      delete this._message;
    }
    await startEduProxy();
  }
}

/**
 * HTTP endpoints to control the network test
 */
class NetworkTestMixin extends ServiceMixin {
  init() {
    // We basically have 2 steps: A request network test, where a JWT token is checked
    // This returns a code which can then be used for control calls.
    // We chose to just have a 'control' step with an action parameter to allow flexibility
    // for future changes without losing backwards compatibility with old devices.
    // For this reason, the request call also returns the version of the network test package
    this._service.addRoute('/networkTestVersion', 'get', this.version.bind(this));
    this._service.addRoute('/requestNetworkTest', 'post', wrapAsync(this.requestTest.bind(this)));
    this._service.addRoute('/networkTestEventSource', 'get', this.openEventSource.bind(this));
    this._service.addRoute('/controlNetworkTest', 'post', wrapAsync(this.controlTest.bind(this)));
  }

  stop() {
    clearTimeout(this.reconnectTimer);
    return this._cleanup();
  }

  _verifyRequest(req) {
    let { code } = url.parse(req.url, true).query;
    if (!code && req.method === 'POST') {
      code = req.body.code;
    }

    if (!this.activeTest) {
      throw Boom.notFound('No test active, you first need to request a test');
    }

    if (this.activeTest.code !== code) {
      throw Boom.forbidden('Wrong access code');
    }

    try {
      verifyIp(req, this.activeTest.context.master.ipForSlave);
    } catch (error) {
      throw Boom.forbidden('Request comes from an invalid ip');
    }
  }

  async version(req, res) {
    res.json({ version });
  }

  async requestTest(req, res) {
    try {
      this._log('info', 'Requesting network test token');

      // We can only have 1 test at a time
      if (this.activeTest) {
        throw Boom.forbidden('There is already a test in progress');
      }

      // Validate request
      const { token, ...context } = req.body;
      const { action } = await decodeToken(token);
      if (action !== ACTION.REQUEST_NETWORK_TEST) {
        throw Boom.forbidden('Invalid JWT Token action');
      }

      // Get IP and complete context
      const code = Random.id();
      const masterIp = getIpFromRequest(req);
      context.httpServer = this._service;
      context.logger = this._service._logger;
      context.master.ipForSlave = masterIp;
      if (context.master.isInInfra) {
        context.master.privateIp = masterIp;
      } else {
        context.master.publicIp = masterIp;
      }
      context.self.proxy = proxyConfigToUrl(serviceRegistry.get('configManager').getSetting('device.proxy') || context.self.proxy);

      // Create test runner and keep communication channel
      this._log('verbose', 'Created network test request token');

      this.activeTest = new PurpleTestRunner(context, code);
      this.invalidateTimer = setTimeout(() => {
        // If test doesn't get started within a minute, clear the test
        delete this.activeTest;
      }, 60000);
      this.activeTest.on('message', this._send.bind(this));

      res.json({ code, masterIp });
    } catch (error) {
      this._log('error', 'Error while requesting network test', error);
      if (error.isBoom) {
        throw error;
      }
      throw Boom.badRequest();
    }
  }

  async _cleanup() {
    if (this.communicationChannel) {
      this.communicationChannel.close();
      delete this.communicationChannel;
    }

    if (this.activeTest) {
      await this.activeTest.stop();
      delete this.activeTest;
    }

    if (this.messenger) {
      this.messenger.destroy();
      delete this.messenger;
    }
  }

  openEventSource(req, res) {
    this._verifyRequest(req);
    if (this.communicationChannel && this.communicationChannel.outgoingMessage) {
      throw Boom.forbidden('You can only have 1 active communication channel');
    }

    clearTimeout(this.reconnectTimer);

    if (!this.communicationChannel) {
      this.communicationChannel = new CommunicationChannel(res);
    } else {
      this.communicationChannel.outgoingMessage = res;
    }

    req.on('close', async () => {
      this._log('info', 'HTTP Request closed');
      if (this.activeTest && this.activeTest.isRunning) {
        this._log('verbose', 'Test is still running, schedule cleanup in case master reconnects');
        this.reconnectTimer = setTimeout(() => this._cleanup(), 5000);
      } else {
        this._log('info', 'Test is not running, cleanup immediately');
        await this._cleanup();
      }
    });
  }

  // Some tests require other devices to be in a certain state.
  // E.g. they need to add certain endpoints to allow intranet speed test or open certain ports.
  // So a device can only start its test when all other devices are in 'prepared' state.
  // They can only close these endpoints/ports/... after all other devices have completed their tests.
  // That's why you have these steps and a test master which should coordinate these steps.
  async controlTest(req, res) {
    this._verifyRequest(req);

    const { action } = req.body;

    if (
      (action !== ACTION.PREPARE_NETWORK_TEST && !this.activeTest.isActive)
      || (action === ACTION.PREPARE_NETWORK_TEST && this.activeTest.isActive)) {
      throw Boom.badRequest('This action cannot be executed at this moment');
    }

    if (this.invalidateTimer) {
      clearTimeout(this.invalidateTimer);
      delete this.invalidateTimer;
    }

    this._log('verbose', `Network test command received: ${action}`);
    switch (action) {
    case ACTION.PREPARE_NETWORK_TEST:
      this._prepareTest();
      break;
    case ACTION.START_NETWORK_TEST:
      this._requestStart();
      break;
    case ACTION.STOP_NETWORK_TEST:
      this._requestStop();
      break;
    default:
      throw Boom.badRequest('Unknown action');
    }

    res.end();
  }

  async _prepareTest() {
    this._log('debug', 'Request prepare');
    try {
      this._send(MESSAGE.SLAVE_SETUP_STARTED);
      await this.activeTest.prepare();
      this._send(MESSAGE.SLAVE_SETUP_DONE);
    } catch (error) {
      this._send(MESSAGE.SLAVE_SETUP_FAILED);
    }
  }

  async _requestStart() {
    this._log('debug', 'Request start');
    try {
      this._send(MESSAGE.SLAVE_START_STARTED);
      await this.activeTest.start();
      this._send(MESSAGE.SLAVE_START_DONE);
    } catch (error) {
      this._send(MESSAGE.SLAVE_START_FAILED);
    }
  }

  async _requestStop() {
    this._log('debug', 'Request stop');
    try {
      this._send(MESSAGE.SLAVE_STOP_STARTED);
      await this.activeTest.stop();
      this._send(MESSAGE.SLAVE_STOP_DONE);
    } catch (error) {
      this._send(MESSAGE.SLAVE_STOP_FAILED);
    } finally {
      this._end();
    }
  }

  _send(message, params = {}) {
    if (this.communicationChannel) {
      try {
        this._log('verbose', `Sending message: ${message}. With params`, params);
        this.communicationChannel.sendMessage(message, params);
      } catch (error) {
        this._log('warn', `Could not send message ${message}`, error);
      }
    }
  }

  _end() {
    this._log('info', 'Closing the connection');
    this._send([MESSAGE.SLAVE_DONE]);
    if (this.communicationChannel) {
      try {
        this.communicationChannel.close();
      } catch (error) {
        this._log('warn', 'Could not close the connection (could have already been closed)');
      }
    }
  }
}

export default NetworkTestMixin;
