import * as Boom from 'boom';
import * as path from 'path';

import { ROOT_DIR } from '../../files';
import serviceRegistry from '../serviceRegistry';
import { wrapAsync } from '../server';
import platform from '../../utils/platform';
import { readCurrentVersion } from '../../utils/version';
import ServiceMixin from '../serviceMixin';
import decodeToken from '../../utils/jwt';

const proxyVersion = require(path.join(ROOT_DIR, 'package.json')).version; // eslint-disable-line

/**
 * HTTP endpoints to receive the initial configuration.
 * Also provides some other utilities for the pairing tool, like identification messages.
 */
class PairingMixin extends ServiceMixin {
  init() {
    // Accept configurations
    // this.app.post('/postConfig', this.pair.bind(this))
    this._service.addRoute('/postConfig', 'post', wrapAsync(this.pair.bind(this)));

    // Returns system & purple version
    this._service.addRoute('/version', 'get', this.version.bind(this));
  }

  /**
   * Are we already paired or not?
   */
  get isPaired() {
    const meteorConfig = serviceRegistry.get('configManager').getSetting('meteor');
    return !!meteorConfig;
  }

  /**
   * Start listening on the port
   * In case we're not yet paired, show 'ready to pair' message
   */
  start() {
    if (!this.isPaired) {
      this._message = this._service._showMessage({
        text: 'ready for pairing',
      });
      this._service._messenger.stopProgress();
    }
  }

  /**
   * Write configuration file, aka pairing
   *
   * @param {*} req
   * @param {*} res
   */
  async pair(req, res) {
    let config = req.body;
    let serverTime;
    this._log('info', `Received pairing config. Data = ${JSON.stringify(config)}`);

    try {
      this._log('info', 'Trying to decode JWT token');
      ({ config, serverTime } = await decodeToken(config.token));
      this._log('info', `Decoded config data = ${JSON.stringify(config)}`);
    } catch (error) {
      this._log('error', 'Error while decoding config token', error);
      throw Boom.badRequest('Invalid token');
    }

    if (serverTime) {
      // Setting the hardware and/or system clock does not change active timers
      // Since none should be active or stay active (either not paired or disconnecting soon)
      // we don't need to reboot in this case.
      // We will however sync the HW clock with the System clock
      // The hardware clock is only used on boot, the system clock is what software works with
      platform.setHWClock(new Date(serverTime), this.log.bind(this), true);
    }

    try {
      this._log('info', 'Setting config');
      await serviceRegistry.get('configManager').setConfig(config);
      if (this._message) {
        this._message.hide();
        delete this._message;
      }
    } catch (error) {
      this._log('error', 'Could not save edu_proxy config file', error);
      throw Boom.failedDependency('Could not save configuration');
    }

    this._log('info', 'Successfully saved edu_proxy config file.');
    res.json(true);
  }

  /**
   * Returns the system & edu-proxy version.
   * @param {*} req
   * @param {*} res
   */
  version(req, res) {
    const systemVersion = readCurrentVersion(this._service._logger);
    res.json({
      proxy: proxyVersion,
      system: systemVersion,
    });
  }
}

export default PairingMixin;
