import * as pify from 'pify';
import * as express from 'express';
import * as bodyParser from 'body-parser';
import * as cors from 'cors';
import * as Boom from 'boom';
import * as http from 'http';
import * as httpClose from 'http-close';

import { Service } from './service';

// wrapper for our async route handlers
// probably you want to move it to a new file
export const wrapAsync = fn => (req, res, next) => {
  Promise.resolve(fn(req, res, next)).catch((err) => {
    if (!err.isBoom) {
      return next(Boom.badImplementation(err));
    }
    next(err);
  });
};

export function createServer(connectionHandler, port, host) {
  return new Promise((resolve, reject) => {
    const server = http.createServer(connectionHandler);
    // Add http-close hook
    httpClose({ timeout: 2000 }, server);
    // If we don't attach an error handler, it will throw an uncaught exception, which will crash the process
    server.once('error', (err) => { server.close(); reject(err); });
    server.once('listening', () => resolve(server));
    if (host) {
      server.listen(port, host);
    } else {
      server.listen(port);
    }
  });
}

function findRoute(path, stack) {
  let routes = [];
  stack.forEach((layer) => {
    if (!layer) return;
    if (layer && !layer.match(path)) return;
    if (['query', 'expressInit'].includes(layer.name)) return;
    if (layer.name === 'router') {
      routes = routes.concat(findRoute(path.substr(layer.path.length), layer.handle.stack));
    } else if (layer.name === 'bound ') {
      routes.push({ route: layer || null, stack });
    }
  });
  return routes;
}

/**
 * Base REST server
 */
class Server extends Service {
  constructor(options = {}) {
    super(options);
    this._host = options.host;
    this._port = options.port;
    this.app = express();
    const parseJSON = bodyParser.json();
    // Body parser can't handle large bodies
    // So we have to disable it for the testUpload from the network test
    this.app.use((req, res, next) => (req.url !== '/testUpload' ? parseJSON(req, res, next) : next()));
    this.app.use(cors());
  }

  set host(host) {
    if (this._host !== host) {
      this._host = host;
      if (this.isActive) {
        this.restart();
      }
    }
  }

  get host() {
    return this._host;
  }

  set port(port) {
    if (this._port !== port) {
      this._port = port;
      if (this.isActive) {
        this.restart();
      }
    }
  }

  get port() {
    return this._port;
  }

  addRoute(path, method, handler) {
    if (!path.startsWith('/')) {
      path = `/${path}`;
    }
    this.app[method](path, handler);
  }

  removeRoute(path, method) {
    if (!path.startsWith('/')) {
      path = `/${path}`;
    }

    const found = findRoute(path, this.app._router.stack);

    found.forEach((layer) => {
      const { route, stack } = layer;
      if (route) {
        if (!method || JSON.stringify(route.route.methods).toUpperCase().includes(method.toUpperCase())) {
          stack.splice(stack.indexOf(route), 1);
        }
      }
    });

    return true;
  }

  /**
   * Start listening
   */
  async _onStart() {
    this.server = await createServer(this.app, this.port, this.host);
    this._log('info', `Listening at port ${this.port}...`);
    await super._onStart();
  }

  /**
   * Alias for stop (better named for http servers)
   */
  close() {
    return this.stop();
  }

  /**
   * Close express server
   */
  async _onStop() {
    if (this.server) {
      await pify(this.server.close.bind(this.server))();
      delete this.server;
    }
    await super._onStop();
  }

  /**
   * Cleanup
   */
  _onDestroy() {
    delete this.app;
    return super._onDestroy();
  }
}

export default Server;
