import { serializeError } from 'serialize-error';
import EventEmitter from '../utils/eventEmitter';
import Random from '../utils/random';
import { wait } from '../utils/utils';
import serviceRegistry from './serviceRegistry';
import ServiceError from './serviceError';
import CancelError from './cancelError';
import CancelToken from './cancelToken';
import Task from './task';

// State of the service
const state = {
  STOPPING: 0,
  STOPPED: 1,
  STARTING: 2,
  STARTED: 3,
  DESTROYING: 4,
  DESTROYED: 5,
};

// State Lifecycle:
// * => Start node
// ! => End node
//
// |--------- STOPPED* ----------
// |             ^  ^            |
// -> STARTING --|  |--STOPPING  |
//       |  |           ^  ^     |
//       |  |-----------|  |     |
//       |                 |     |
//       |--> STARTED -----|     |
//               |               |
//               v               |
//          DESTROYING <---------|
//               |
//               v
//           DESTROYED!
//
const {
  STOPPING,
  STOPPED,
  STARTING,
  STARTED,
  DESTROYING,
  DESTROYED,
} = state;

const stateDefinitions = {};
stateDefinitions[STOPPING] = {
  label: 'Stopping',
  event: 'stopping',
  intermediate: true,
};
stateDefinitions[STOPPED] = {
  label: 'Stopped',
  event: 'stopped',
  error: 'CAN_NOT_STOP',
  inverse: STARTED,
  intermediate: false,
};
stateDefinitions[STARTING] = {
  label: 'Starting',
  event: 'starting',
  intermediate: true,
};
stateDefinitions[STARTED] = {
  label: 'Started',
  event: 'started',
  error: 'CAN_NOT_START',
  inverse: STOPPED,
  intermediate: false,
};
stateDefinitions[DESTROYING] = {
  label: 'Destroying',
  event: 'destroying',
  intermediate: true,
};
stateDefinitions[DESTROYED] = {
  label: 'Destroyed',
  event: 'destroyed',
  error: 'CAN_NOT_DESTROY',
  inverse: STARTED,
  intermediate: false,
};

const healthState = {
  HEALTHY: 0,
  UNHEALTHY: 1,
};
const {
  HEALTHY,
  UNHEALTHY,
} = healthState;

const healthStateDefinitions = {};
healthStateDefinitions[HEALTHY] = {
  label: 'Healthy',
};
healthStateDefinitions[UNHEALTHY] = {
  label: 'Unhealthy',
};

const ENABLE_WATCHDOGS = false;

/**
 * The base service class.
 * Implements the common code between all services.
 * This allows safely starting, stopping and destroying services with a consistent behavior.
 */
class Service extends EventEmitter {
  constructor({
    id,
    eventOptions = {},
  } = {}) {
    super(eventOptions);

    this._serviceId = id;

    this._logger = serviceRegistry.get('logger').getInstance(this.serviceName, this._serviceId || Random.id());

    this._serviceState = STOPPED;
    this._requestedState = STOPPED;
    this._healthState = HEALTHY;

    this._mixins = Array.from(new Set(this.mixins)).map(Mixin => new Mixin(this));

    this._stateWaiters = {
      [STOPPED]: null,
      [STARTED]: null,
      [DESTROYED]: null,
    };

    this._cancelTS = {
      [STARTED]: null,
    };

    process.nextTick(() => {
      // Since constructors will normally not be asynchronous, this will ensure the init is called right after construct finishes
      this._mixins.forEach(mixin => mixin.init());
    });
  }

  get serviceName() {
    return this.constructor.name;
  }

  get mixins() {
    return [];
  }

  get isFaulty() {
    return this.isUnhealthy;
  }

  get isUnhealthy() {
    return !this.isHealthy;
  }

  get isHealthy() {
    return this._healthState === HEALTHY;
  }

  get canStart() {
    return this.isStopped;
  }

  get canStop() {
    return this.isStarted;
  }

  get isStopped() {
    return this._serviceState === STOPPED;
  }

  get isStopping() {
    return this._serviceState === STOPPING;
  }

  get isStarted() {
    return this._serviceState === STARTED;
  }

  get isStarting() {
    return this._serviceState === STARTING;
  }

  get isDestroying() {
    return this._serviceState === DESTROYING;
  }

  get isDestroyed() {
    return this._serviceState === DESTROYED;
  }

  get isActive() {
    return (this.isStarted || this.isStarting);
  }

  get serviceState() {
    return this._serviceState;
  }

  set serviceState(newState) {
    if (this._serviceState !== newState) {
      const oldState = this._serviceState;
      this._serviceState = newState;
      this._log('info', `Changing state from ${stateDefinitions[oldState].label} to ${stateDefinitions[newState].label} (requested state: ${stateDefinitions[this.requestedState].label})`);
      this.emit('stateChanged', newState, oldState);
      this.emit(stateDefinitions[this._serviceState].event);
      this._transition();
    }
  }

  set requestedState(requestedState) {
    if (this._requestedState !== requestedState) {
      if (stateDefinitions[requestedState].intermediate) {
        throw new Error(`${requestedState} is not a valid requested state`);
      }

      this._log('verbose', `Changing requested state from ${stateDefinitions[this._requestedState].label} to ${stateDefinitions[requestedState].label}`);

      const { inverse } = stateDefinitions[requestedState];
      const cancelTS = this._cancelTS[inverse];
      if (cancelTS) {
        cancelTS.cancel();
        delete this._cancelTS[inverse];
      }

      this._requestedState = requestedState;
      this.emit('requestedStateChanged', requestedState);
      this._transition();
    }
  }

  get requestedState() {
    return this._requestedState;
  }

  get wantsToBeActive() {
    return this.requestedState === STARTED;
  }

  get canChangeState() {
    return this.requestedState !== DESTROYED;
  }

  get canBeDestroyed() {
    return !this.isStarting && !this.isStopping && !this.isDestroyed && !this.isDestroying;
  }

  _transition() {
    if (this.isStarted && this._requestedState === STOPPED) {
      this._log('info', 'We\'re started, but we should be stopped => stop');
      this._stop();
    }

    if (this.isStopped && this._requestedState === STARTED) {
      this._log('info', 'We\'re stopped, but we should be started => start');
      this._start();
    }

    if (this.canBeDestroyed && this._requestedState === DESTROYED) {
      this._log('info', 'We\'re destroyable, but we should be destroyed => stop');
      this._destroy();
    }
  }

  async _awaitEvent(source, event, {
    timeout = 0,
    throwOnTimeout = true,
  } = {}) {
    const promises = [
      new Promise(resolve => source.once(event, resolve)),
    ];
    if (timeout) {
      promises.push(new Promise((resolve, reject) => setTimeout(throwOnTimeout ? reject : resolve, timeout)));
    }
    return Promise.race(promises);
  }

  /**
   * Returns a promise that will resolve when we reach the target state
   * Or rejects when we will not reach the target state (e.g. change in target state requested)
   *
   * E.g.
   * TargetState === STARTED, Curretn State === STARTED => resolves immedeatily
   * TargetState === STARTED, Current State === STARTING => Resolves when we go from STARTING TO STARTED
   *
   * Same as before, but Requested State === STOPPED when we go from STARTING -> STARTED => Rejects
   *
   * @param {*} targetState
   */
  _waitForState(targetState) {
    if (this.serviceState === targetState) {
      return Promise.resolve(true);
    }

    if (!this._stateWaiters[targetState]) {
      this._stateWaiters[targetState] = new Promise((resolve, reject) => {
        const onStateChange = () => {
          if (this.serviceState === targetState
          || (targetState === state.STOPPED && this.serviceState === state.DESTROYED) // XXX: do we want this ?
          ) {
            this.removeListener('stateChanged', onStateChange);
            delete this._stateWaiters[targetState];
            resolve();
          } else if (this.requestedState !== targetState) {
            this.removeListener('stateChanged', onStateChange);
            delete this._stateWaiters[targetState];
            reject(new Error(`Asked to put service "${this.serviceName}" in ${stateDefinitions[this.requestedState].label} state before we could go to the ${stateDefinitions[targetState].label} state`));
          }
        };

        this.on('stateChanged', onStateChange);
      });
    }

    return this._stateWaiters[targetState];
  }

  _startWatchDog() {
    if (!ENABLE_WATCHDOGS) return;

    this._stopWatchDog();

    this._watchdogTimer = setTimeout(async () => {
      this._log('error', 'Watchdog timer triggered');
      await serviceRegistry.get('logger').destroy();
      process.exit(1);
    }, 60 * 1000);
  }

  _stopWatchDog() {
    if (this._watchdogTimer) {
      clearTimeout(this._watchdogTimer);
      delete this._watchdogTimer;
    }
  }

  async _awaitAll(promises) {
    let hasError = false;
    let error;

    const result = await Promise.all(promises.map(promise => promise.catch((e) => {
      if (!hasError) {
        hasError = true;
        error = e;
      }
    })));

    if (hasError) {
      throw error;
    } else {
      return result;
    }
  }

  /**
   * Restart the service, possibly with a timeout in between
   *
   * @param {*} timeout
   */
  async restart({ timeout = 0, ...startOptions } = {}) {
    this._log('info', 'Restarting');

    try {
      await this.stop(true);
    } catch (error) {
      if (error instanceof ServiceError && error.code === 'CAN_NOT_STOP') {
        this._log('debug', error);
      } else {
        throw error;
      }
    }

    if (timeout > 0) {
      await wait(timeout);
    }

    if (this._wantRestart) {
      // Uncomment to simulate meteorCache startup failure (NET0004-9941)
      // Restart nginx to simulate a blue upgrade. This will trigger the
      // meteor cache startup failure. Currently fixed via a reboot of the
      // device as a temporary workaround.
      // if (this._meteorClient) {
      //   this._log('debug', 'DELETING DDP CLIENT !!!!');
      //   this._meteorClient._ddpClient = undefined;
      // }
      await this.start(startOptions);
    } else {
      throw new ServiceError('CAN_NOT_RESTART', 'Service asked to stop while waiting for restart');
    }
  }

  markUnhealthy(fault = 'not fully working') {
    if (this.isHealthy || this.fault !== fault) {
      const wasHealthy = this.isHealthy;
      this.fault = fault;
      this._healthState = UNHEALTHY;
      this.emit('healthStateChanged');
      if (wasHealthy) {
        this.emit('unhealthy');
      }
    }
  }

  markHealthy() {
    if (!this.isHealthy) {
      delete this.fault;
      this._healthState = HEALTHY;
      this.emit('healthStateChanged');
      this.emit('healthy');
    }
  }

  /**
   * Get the cancel token (if available)
   */
  getCancelToken(stateToCancel) {
    if (this._cancelTS[stateToCancel]) {
      return this._cancelTS[stateToCancel].token;
    }
    throw new Error(`No cancel token for state ${stateToCancel}`);
  }

  /**
   * Public method to start the service,
   * Will do some checks to see if we can change the service.
   *
   * Then it will change the requested state, which will trigger the transition
   */
  async start({ retryTimer, retries, cancelToken } = {}) {
    this._log('info', `Asked to start (retry interval = ${retryTimer} ms, max nbr retries = ${retries})`);

    if (!this.canChangeState) {
      const error = `Service can not be started in the current state (${stateDefinitions[this._serviceState].label})`;
      this._log('error', error);
      throw new ServiceError('CAN_NOT_START', error);
    }

    if (this.isStarted) {
      this._log('verbose', 'Service already started');
      return true;
    }

    if (this.isStopping) {
      this._log('verbose', 'Waiting for service to stop before we can re-start service');
    }

    if (this.requestedState === STARTED) {
      this._log('verbose', 'Already requested to start service');
    }

    // Time between retries of the start task
    // The start task will keep retrying until the stop state is requested
    this._retryTimer = retryTimer;
    this._retries = retries;

    if (!this._cancelTS[STARTED]) {
      // Create a cancellation token for the start task
      // Fork from the parent token if one is given\
      if (cancelToken) {
        cancelToken.promise.then(() => {
          // If a cancel token was passed in, and we're requested to be cancelled, then the new desired state would be to stop
          // Otherwise we would cancel the run task, but it would immediately restart the service
          if (this.requestedState === STARTED) {
            this.requestedState = STOPPED;
          }
        });
      }
      this._cancelTS[STARTED] = cancelToken ? cancelToken.fork() : CancelToken.source();
    }

    // This will trigger the transition
    this.requestedState = STARTED;

    return this._waitForState(STARTED);
  }

  /**
  * Small utility method to prevent async actions to log after destruction
  */
  _log(level, ...params) {
    if (this._logger) {
      if (!['silly', 'verbose', 'info', 'warn', 'debug', 'error'].includes(level)) {
        this._logger.log('error', 'Invalid log level', level);
      } else {
        this._logger.log(level, ...params);
      }
    }
  }

  /**
   * This starts the service without the checks in start()
   * Taken out of start to make it more consise and readable
   *
   * This basically creates a task that will keep retry-ing to start the service
   * Until a state change is requested (e.g. stop or destroy)
   * Then it will stop the task if it has not yet been completed by means of a cancellation token
   *  (for the concept of cancellation tokens, see:
   *  https://msdn.microsoft.com/en-us/library/system.threading.cancellationtoken(v=vs.110).aspx)
   * The task 'run' method returns a promise that completes when we can successfully start a task
   *
   * When the task errors, we first call the _onStop() method before we try to _onStart again
   * This makes sure anything created in _onStart is cleaned up first
   *
   */
  async _start() {
    this.serviceState = STARTING;
    this._startWatchDog();
    this.markHealthy();

    let task;
    try {
      task = new Task('Start', async () => {
        try {
          await this._startMixins();
          await this._onStart();
        } catch (error) {
          await this._onStop();
          await this._stopMixins();
          throw error;
        }
      }, {
        logger: this._logger,
        cancelToken: this.getCancelToken(STARTED),
        retryTimer: this._retryTimer,
        retries: this._retries,
      });

      await task.run();

      this._stopWatchDog();
      this.serviceState = STARTED;
    } catch (error) {
      if (error instanceof CancelError) {
        this._log('info', `Start of service was cancelled: ${Math.random()}`);
        if (!this.wantsToBeActive) {
          this._stop();
        } else {
          // this.requestedState = STOPPED;
          this.serviceState = STOPPED;
        }
      } else {
        this._log('error', `Service could not be started in max allowed (= ${this._retries + 1}) attempts. Stopping service. Error = ${JSON.stringify(serializeError(error, { maxDepth: 1 }))})`);
        this.requestedState = STOPPED; // Avoid new Start task to be instantiated (= avoid endless loop)
        this.serviceState = STOPPED;
      }
    } finally {
      task.destroy();
    }
  }

  _startMixins() {
    return Promise.all(this._mixins.map((mixin) => {
      this._log('verbose', `Starting mixin: ${mixin.constructor.name}`);
      return mixin.start();
    }));
  }

  /**
  * on Start hook
  * This is the actual starting logic (without any lifecycle stuff)
  * for the children to implement the start behavior.
  */
  _onStart() {
    return Promise.resolve(true);
  }

  /**
   * Public method to stop the service,
   * Will do some checks to see if we can change the service.
   *
   * Then it will change the requested state, which will trigger the transition
   */
  async stop(fromRestart = false) {
    this._log('info', 'Asked to stop');

    // In case someone tries to stop the service while we're waiting for a restart
    // Don't restart
    this._wantRestart = fromRestart;

    if (!this.canChangeState) {
      const error = `Service can not be stopped in current state (${stateDefinitions[this._serviceState].label})`;
      this._log('error', error);
      throw new ServiceError('CAN_NOT_STOP', error);
    }

    if (this.isStopped) {
      this._log('verbose', 'Service already stopped');
      return true;
    }

    if (this.requestedState === STOPPED) {
      this._log('verbose', 'Already requested to stop service');
    }

    if (this.isStarting) {
      this._log('verbose', 'Waiting for service to start before we can stop service');
    }

    // This will trigger the transition
    this.requestedState = STOPPED;

    return this._waitForState(STOPPED);
  }

  /**
   * The stop functionality
   * Taken out of stop() to make it more consise and readable
   *
   * Basically awaits the _onStop hook (for the children to implement)
   * Also stops the mixins
   */
  async _stop() {
    this.serviceState = STOPPING;
    this._startWatchDog();
    try {
      await this._onStop();
      await this._stopMixins();
    } catch (e) {
      this._log('error', 'Error while stopping service', e);
    }

    this._stopWatchDog();
    this.serviceState = STOPPED;
  }

  _stopMixins() {
    return Promise.all(this._mixins.map((mixin) => {
      this._log('verbose', `Stopping mixin: ${mixin.constructor.name}`);
      return mixin.stop();
    }));
  }

  /**
   * on Stop hook
   * This is the actual stopping logic (without any lifecycle stuff)
   * for the children to implement the stop behavior.
   */
  _onStop() {
    return Promise.resolve(true);
  }

  /**
   * Public method to destroy the service,
   * Will do some checks to see if we can change the service.
   *
   * Then it will change the requested state, which will trigger the transition
   */
  async destroy() {
    this._log('info', 'Asked to destroy');

    if (this.isDestroyed) {
      // Don't log here, since logger is already removed by now and would crash
      return true;
    }

    if (this.requestedState === DESTROYED) {
      this._log('verbose', 'Already requested to destroy service');
    }

    if (this.isStarting) {
      this._log('verbose', 'Waiting for service to fully start before we can destroy service');
    } else if (this.isStopping) {
      this._log('verbose', 'Waiting for service to fully stop before we can destroy service');
    }

    // This will trigger the transition
    this.requestedState = DESTROYED;

    return this._waitForState(DESTROYED);
  }

  /**
   * The destroy functionality
   * Taken out of destroy() to make it more consise and readable
   *
   * Basically awaits the _onStop hook if hte service is still running
   * Then runs the _onDestroy hook (for the children to implement)
   *
   * Also stops the mixins
   */
  async _destroy() {
    const wasActive = this.isActive;
    this.serviceState = DESTROYING;
    this._startWatchDog();
    try {
      if (this._restartTimerHandle) {
        clearTimeout(this._restartTimerHandle);
        delete this._restartTimerHandle;
      }

      if (wasActive) {
        await this._onStop();
      }
    } finally {
      await this._onDestroy();
      this._mixins.forEach(mixin => mixin.destroy());

      this.serviceState = DESTROYED;
      this._stopWatchDog();
      super.destroy();
      delete this._logger;
      delete this._waitForStatePromise;
    }
  }

  /**
   * This is the actual destroy logic (without any lifecycle stuff)
   * for the children to implement the stop behavior.
   */
  _onDestroy() {
    return Promise.resolve(true);
  }
}

export default Service;

export {
  Service,

  state,
  stateDefinitions,

  healthState,
  healthStateDefinitions,
};
