import { EventEmitter } from 'events';

/**
 * Simple map to register and retrieve services, useful for dependency injection
 *
 * Also provides publish-subscribe pattern to listen to registred / removed services
 */

class ServiceRegistry extends Map {
  constructor() {
    super();
    this.listeners = {};
    this._events = new EventEmitter();
    this.on = this._events.on.bind(this._events);
    this.off = this._events.removeListener.bind(this._events);
  }

  onDeleted(key, fn) {
    this._events.on(`delete.${key}`, fn);
  }

  onRegistered(key, fn) {
    this._events.on(`register.${key}`, fn);
  }

  delete(key) {
    if (this.has(key)) {
      const listener = this.listeners[key];
      this.get(key).removeListener('destroying', listener);
      delete this.listeners[key];
      super.delete(key);
      this._events.emit(`delete.${key}`);
    }
  }

  set(key, value) {
    if (this.has(key)) {
      this.delete(key);
    }
    const listener = () => { this.delete(key); };
    this.listeners[key] = listener;
    value.on('destroying', listener);

    super.set(key, value);
    this._events.emit(`register.${key}`);
    this._events.emit('registered', key, value);
  }
}

const registry = new ServiceRegistry();
registry.set('RTSPRegistry', new ServiceRegistry());

export default registry;
export { registry as serviceRegistry };
