import { execSync, spawn } from 'child_process';
import * as spawnargs from 'spawn-args';
import * as pify from 'pify';

import Service from './service';

class SpawningService extends Service {
  constructor({ loop = true, killSignal = 'SIGKILL', ...args }) {
    super(args);
    this._loop = loop;
    this._killSignal = killSignal;
    this._timesRetried = 0;
  }

  get command() {
    throw new Error('Command getter is not implemented!');
  }

  get args() {
    return undefined;
  }

  get env() {
    return undefined;
  }

  _scheduleRetry(timeInMs = 5000) {
    this._log('verbose', `Scheduling retry in ${timeInMs}ms`);
    this._clearRetryTimer();
    this._timesRetried += 1;
    this._retryTimer = setTimeout(() => {
      this._launch();
    }, timeInMs);

    // Every twelve attempts throw a warning
    if (!(this._timesRetried % 12)) {
      const { args } = this;
      this._log('warn', `Retried ${this._timesRetried} times on: "launching process : ${this.command} ${args}"`, {
        serviceLog: true,
      });
    }
  }

  _clearRetryTimer() {
    if (this._retryTimer) {
      this._log('debug', 'Clearing retry timer.');
      clearTimeout(this._retryTimer);
      delete this._retryTimer;
    }
  }

  _launch() {
    this._clearRetryTimer();

    if (this._childProc || !this.isActive) {
      return;
    }

    const { args } = this;

    this._log('info', `Launching process : ${this.command} ${args}`);

    this._childProc = spawn(this.command, spawnargs(args, {
      removequotes: 'always',
    }), {
      cwd: process.cwd(),
      env: this.env,
    });

    const { pid } = this._childProc;

    if (pid) {
      this.markHealthy();
    } else {
      this.markUnhealthy('Could not spawn child process');
    }

    this._childProc.stderr.on('data', (data) => {
      this._log('error', `Error from child process ${pid}: ${data}`);
    });

    this._childProc.on('error', (error) => {
      this._log('error', `process with PID (${pid}) error ${error}`);

      if (typeof error === 'object' && error.message === 'kill EPERM') {
        this._log('warn', `Killing process with PID (${pid}) failed, trying the hard way`);
        this._hardKill();
      } else if (!this.isActive) {
        this._cleanupProc();
      }
    });

    this._childProc.on('close', (code) => {
      // Process has ended, remove it from the list of running processes
      this._cleanupProc();
      if (this.isActive) {
        this.markUnhealthy();
      }

      if (code === 0) {
        if (this._loop) {
          this._log('info', `child process ${pid} ended, relaunching process as relaunch on exit flag is set`);
          if (this.isActive) {
            this._scheduleRetry();
          }
        } else {
          this._log('info', `child process ${pid} ended.`);
        }
      } else {
        this._log('error', `child process ${pid} exit with code ${code}`);
        if (this.isActive) {
          this._scheduleRetry();
        }
      }
    });

    this._log('verbose', `PID of child process = ${pid}`);
  }

  _hardKill() {
    if (this._childProc) {
      const { pid } = this._childProc;
      this._log('warn', `Hard killing ${pid}`);
      try {
        execSync(`sudo pkill -9 -P ${pid}`);
      } catch (e) {
        this._log('error', `Hard kill failed ... ${e}`);
      }
    }

    if (!this.isActive) {
      this._cleanupProc();
    }
  }

  _cleanupProc() {
    this._log('verbose', 'Cleaning up all listeners');
    if (this._childProc) {
      this._childProc.stderr.removeAllListeners();
      this._childProc.stdout.removeAllListeners();
      this._childProc.removeAllListeners();
      delete this._childProc;
    }
    this.emit('_cleanup');
  }

  async _stopProc() {
    if (this._childProc) {
      const cleanupPromise = pify(this.once.bind(this))('_cleanup');
      const { pid } = this._childProc;
      const softKillTimeout = setTimeout(() => this._hardKill(), 5000);

      try {
        this._log('verbose', `Killing process with PID = ${pid}`);
        this._childProc.kill(this._killSignal);
        this._log('verbose', `Successfully killed process with PID = ${pid}`);
      } catch (error) {
        this._log('error', `Trying to stop stream service using ${this._killSignal} threw error: ${error}`);
      }

      await cleanupPromise;
      clearTimeout(softKillTimeout);
      return true;
    }
    this._log('verbose', 'No stream process ID');
  }

  async _onStart() {
    await super._onStart();
    this._launch();
  }

  async _onStop() {
    this._clearRetryTimer();
    this._log('verbose', 'Trying to stop child process...');
    await this._stopProc();
    this._log('verbose', 'Child process stopped');
    await super._onStop();
  }
}

export default SpawningService;
export { SpawningService };
