/**
 * A task will execute a function continuously until
 * - It's asked to stop by means of a cancelled cancellation token
 * - It's configured number of retries ran out
 */

/*
eslint no-await-in-loop: "off"
*/
class Task {
  constructor(name, fns, {
    cancelToken,
    retries = Number.POSITIVE_INFINITY,
    retryTimer = 1000,
    logger,
  }) {
    this._name = name;
    this._fns = Array.isArray(fns) ? fns : [fns];
    this._cancelToken = cancelToken;
    this._retries = retries;
    this._retryTimer = retryTimer;
    this._logger = logger;
  }

  /**
  * Small utility method to prevent async actions to log after destruction
  */
  _log(...params) {
    if (this._logger) {
      this._logger.log(...params);
    }
  }

  set cancelToken(cancelToken) {
    this._cancelToken = cancelToken;
  }

  /**
   * Start the task
   */
  async run() {
    let lastError;

    for (let i = 0; i <= this._retries && !this.isCancelled; i++) {
      try {
        this._log('info', `Attempt nr ${i} at running task: ${this._name}`);

        let result;
        for (const fn of this._fns) {
          this._throwIfCancelled();
          result = await fn();
        }
        this._log('info', `Task: ${this._name} completed`);

        return result;
      } catch (error) {
        lastError = error;

        this._log('error', `Task: ${this._name} error`, error);

        if (this._retryTimer && !this.isCancelled && (i < this._retries)) {
          this._log('error', `Scheduling restart of Task: ${this._name} in ${this._retryTimer}ms`, error);

          await new Promise((resolve) => {
            this._timer = setTimeout(resolve, this._retryTimer);
          });
        }
      }
    }

    this._log('error', `Task: ${this._name} could not be run successfully`);

    this._throwIfCancelled();
    if (lastError) {
      throw lastError;
    } else {
      throw new Error('Task could not run successfully');
    }
  }

  _throwIfCancelled() {
    if (this._cancelToken) {
      this._cancelToken.throwIfRequested();
    }
  }

  get isCancelled() {
    return (this._cancelToken && this._cancelToken.requested);
  }

  _clearTimer() {
    if (this._timer) {
      clearTimeout(this._timer);
      delete this._timer;
    }
  }

  destroy() {
    this._log('info', `Destroying task: ${this._name}`);
    this._retries = -1;
    this._clearTimer();
    delete this._fn;
    delete this._cancelToken;
    delete this._logger;
  }
}

export default Task;
