import { BrowserWindow } from 'electron';
import { enable as enableRemote } from '@electron/remote/main';
import { EventEmitter } from 'events';
import * as path from 'path';
import { getX11Client } from './x11Client';
import { PROGRAM_DIR } from '../../files';
import createQueue from '../../utils/createQueue';

export const IDLE_STRATEGY = {
  DESTROY: 0,
  HIDE: 1,
};

export class MessengerWindow extends EventEmitter {
  constructor({ id, bounds }) {
    super();
    this._id = id;
    this._totalMessagesCounter = 0;
    this._activeMessages = {};

    this._browserWindow = new BrowserWindow({
      ...bounds,
      frame: false,
      backgroundColor: '#F9F9F9',
      width: bounds.width + 1,
      height: bounds.height + 1,
      webPreferences: {
        nodeIntegration: true,
        enableRemoteModule: true,
        // Next line: See https://www.electronjs.org/docs/breaking-changes
        // In Electron 12, contextIsolation will be enabled by default. To restore the
        // previous behavior, contextIsolation: false must be specified in WebPreferences.
        // Enabling the contextIsolation brakes ipcRenderer communication to messenger window.
        contextIsolation: false,
        devTools: false,
      },
    });

    enableRemote(this._browserWindow.webContents);

    setTimeout(() => {
      if (!this._shouldBeVisible) {
        // Timeout necessary because it's not possible directly after creating
        // And there is no event that seems reliable (tried ready-to-show and friends)
        this.hide();
      }
    }, 1000);

    this._browserWindow.loadFile(path.join(PROGRAM_DIR, 'services/ui/renderer/index.html'));
    this._browserWindow.once('closed', () => {
      this.emit('closed');
    });

    this._browserWindowCommunicationBus = createQueue((...params) => {
      try {
        this._browserWindow.webContents.send(...params);
      } catch (ex) {}
    });
    this._browserWindow.webContents.on('did-finish-load', () => {
      this._browserWindowCommunicationBus.queued = false;
    });
  }

  get displayId() {
    return this._id;
  }

  setBounds(bounds) {
    this._browserWindow.setBounds(bounds);
  }

  setContext(context) {
    Object.entries(context).forEach(([key, value]) => {
      this._browserWindowCommunicationBus.push('info', key, value);
    });
  }

  showMessage({ text, time } = {}) {
    this.show();
    const messageId = `m${this._totalMessagesCounter}`;
    this._totalMessagesCounter += 1;
    this._activeMessages[messageId] = 1;
    this._browserWindowCommunicationBus.push('addMessage', messageId, { text });
    this.emit('message-added', messageId);

    let resolvePromise;
    const returnValue = Object.assign(
      new Promise((resolve) => {
        resolvePromise = resolve;
      }),
      {
        messageId,
        hide: () => {
          if (!this._activeMessages[messageId]) return;
          delete this._activeMessages[messageId];
          this._browserWindowCommunicationBus.push('removeMessage', messageId);
          this.emit('message-removed', messageId);
          resolvePromise();
        },
      },
    );

    if (time) {
      const timer = setTimeout(() => {
        returnValue.hide();
      }, time * 1000);
      timer.unref();
    }

    return returnValue;
  }

  get hasActiveMessages() {
    return Object.keys(this._activeMessages).length > 0;
  }

  show() {
    this._shouldBeVisible = true;
    getX11Client().raiseWindow(this._browserWindow);
  }

  hide() {
    this._shouldBeVisible = false;
    // We use X11 here since .blur() seems to only work within the electron windows (so not for cinergy)
    // And .hide() would remove it from the background which we don't want since it introduces more flicker
    getX11Client().lowerWindow(this._browserWindow);
  }

  stopProgress() {
    this._browserWindowCommunicationBus.push('stopProgress');
  }

  startProgress() {
    this._browserWindowCommunicationBus.push('startProgress');
  }

  destroy() {
    // First try to close window "cleanly", then force-close it.
    // To do this, keep track if window is closed
    let closed = false;
    // and keep a timer for when closing takes too long.
    // Set a timer to force-close the window if it takes longer than a second.
    const forceCloseTimer = setTimeout(() => {
      if (!closed) {
        this._browserWindow.destroy();
      }
    }, 1000);
    forceCloseTimer.unref();

    // When the window is closed
    this._browserWindow.once('closed', () => {
      // remember that it is closed
      closed = true;
      this.removeAllListeners();
    });
    // Gently try to close the window.
    this._browserWindow.close();
  }
}

export default function createWindow({ id, bounds, idleStrategy = IDLE_STRATEGY.HIDE }) {
  const messengerWindow = new MessengerWindow({ id, bounds });
  messengerWindow.on('message-removed', () => {
    if (!messengerWindow.hasActiveMessages) {
      switch (idleStrategy) {
      case IDLE_STRATEGY.HIDE:
        messengerWindow.hide();
        break;
      case IDLE_STRATEGY.DESTROY:
        messengerWindow.destroy();
        break;
      default:
      }
    }
  });
  return messengerWindow;
}
