import { screen as screens } from 'electron';

import createWindow, { IDLE_STRATEGY } from './createWindow';
import showMessageOnWindows from './showMessageOnWindows';

export default function createUI({ logger }) {
  // keep track of contextual information that is the same for all displays
  const context = {};
  // keep track of all windows
  const windows = [];
  // allow window tracking: when a window closes, remove references to it
  function createTrackedWindow(display) {
    const window = createWindow({ ...display, idleStrategy: IDLE_STRATEGY.HIDE });
    windows.push(window);
    window.on('closed', () => {
      const windowAt = windows.findIndex(w => w === window);
      if (windowAt !== -1) {
        windows.splice(windowAt, 1);
      }
    });
    return window;
  }

  // get all screens and create a window for each
  screens.getAllDisplays().map(createTrackedWindow);

  // update display windows as displays change
  const screenEventHandlers = {
    'display-added': (_event, display) => {
      const window = createTrackedWindow(display);
      window.setContext(context);
    },
    'display-removed': (_event, display) => {
      const displayAt = windows.findIndex(w => w.displayId === display.id);
      if (displayAt > -1) {
        const [window] = windows.splice(displayAt, 1);
        window.destroy();
      }
    },
    'display-metrics-changed': (_event, display, changedMetrics) => {
      if (changedMetrics.includes('bounds')) {
        const window = windows.find(w => w.displayId === display.id);
        if (window) {
          window.setBounds(display.bounds);
        }
      }
    },
  };

  Object.entries(screenEventHandlers).forEach(([event, handler]) => screens.on(event, handler));

  const contextProps = ['version', 'mac', 'ip', 'hostname', 'version-proxy', 'version-system'];

  return {
    identify(props) {
      contextProps.forEach((prop) => {
        if (props[prop]) {
          context[prop] = props[prop];
        }
      });
      logger.info('Setting context', { context });
      windows.forEach(window => window.setContext(context));
    },
    stopProgress() {
      windows.forEach(window => window.stopProgress());
    },
    startProgress() {
      windows.forEach(window => window.startProgress());
    },
    show(params) {
      logger.info('showing message', params);
      if (params.flash && !params.bounds) {
        // if flash was specified, we want to create custom windows so that they appear on top of other windows
        return showMessageOnWindows({
          windows: screens.getAllDisplays().map(display => createWindow({ ...display, idleStrategy: IDLE_STRATEGY.DESTROY })),
          message: params,
          context,
          showProgress: false,
        });
      }
      if (params.bounds) {
        // if bounds were specified, we have to create a seperate window on those bounds
        return showMessageOnWindows({
          windows: [createWindow({ id: `${Math.floor(Math.random() * 1000)}`, bounds: params.bounds, idleStrategy: IDLE_STRATEGY.DESTROY })],
          message: params,
          context,
          showProgress: false,
        });
      }

      return showMessageOnWindows({
        windows,
        message: params,
      });
    },
    destroy() {
      Object.entries(screenEventHandlers).forEach(([event, handler]) => screens.off(event, handler));
      windows.forEach(window => window.destroy());
    },
  };
}
