import * as _ from 'lodash';
import { URL } from 'url';
import * as fetch from 'node-fetch';
import { exec, execSync, spawn } from 'child_process';
import * as semver from 'semver';
import * as fs from 'fs-extra';
import * as path from 'path';
import * as ProxyAgent from 'proxy-agent';
import { serializeError } from 'serialize-error';
import { promisify } from 'util';

import {
  wait, sleep, waitForChildProcessToFinish, getAvailableDiskSpace, getAvailableBootSpace,
} from '../../utils/utils';
import { proxyConfigToUrl } from '../../utils/network';
import VersionMonitor from './versionMonitor';
import serviceRegistry from '../serviceRegistry';
import MessengerMixin from '../messengerMixin';
import ProcessMonitor from '../processMonitor';
import {
  SCRIPTS_DIR, LOG_DIR, UPGRADE_LOG_BASENAME, CLEANUP_UNUSED_PACKAGES_LOG_FILENAME,
} from '../../files';
import { readCurrentVersion, ensureVersionFile } from '../../utils/version';

// 6GB Minimum space required for upgrade
const MINIMUM_REQUIRED_FREE_DISKSPACE_KB_FOR_UPGRADE = 6000000;

// 100MB space required in /boot to start upgrade
// for debian 11 upgrade around 90mb is required in /boot space
// for 7th gen nuc there is less space available in /boot around 136mb hence 100mb is safe number
const MINIMUM_REQUIRED_FREE_BOOTSPACE_KB_FOR_UPGRADE = 100000;

// Compare to NRC ISO versions
export function compareVersions(a, b) {
  a = a.replace(/-b([0-9]+)/, '-b.$1');
  b = b.replace(/-b([0-9]+)/, '-b.$1');
  return semver.compare(a, b);
}

function waitForUpgradeProcessToFinish(pid) {
  return new Promise((res, rej) => {
    const upgradeMonitor = new ProcessMonitor(pid);
    upgradeMonitor.once('stopped', () => {
      try {
        let result = fs.readFileSync('/tmp/upgradeResult.txt', 'utf8');
        result = result.replace(/[^A-Z ]/g, '');
        if (result === 'UPGRADE SUCCEEDED') {
          res();
        } else {
          rej(new Error('Upgrade failed'));
        }
      } catch (e) {
        rej(new Error('Could not read upgrade result file'));
      }
    });
    upgradeMonitor.start();
  });
}

/**
* Try to extract url from repository
*/
export function parseRepository(repository) {
  if (!_.isString(repository) || !repository.trim().length) {
    throw new Error('Invalid repository url (empty)');
  }

  const parts = repository.split(' ');

  if (parts[0] === 'deb') {
    parts.shift();
  }

  if (parts.length < 3) {
    throw new Error(`Invalid repository: ${repository}`);
  }

  const url = parts[0].replace(/\/$/, '');
  return {
    full: parts.join(' '),
    url,
    parsed_url: new URL(parts[0]),
    release: parts[1],
    branches: parts.slice(1),
  };
}

const sourcesListFile = '/tmp/eduSources.list';
const aptConfigFile = '/tmp/apt.conf';

/**
 * For an entire system (purple & CNG combined).
 * Upgrades happen by setting an apt repository url and doing a dist-upgrade
 */
class SystemVersionMonitor extends VersionMonitor {
  constructor({
    proxy,
    loginUrl,
    checkUrl,
    upgradeStatusUrl,
    offline = false,
    checkInterval = 3600000,
  } = {}) {
    super({
      checkInterval,
    });

    this.loginUrl = loginUrl;
    this.checkUrl = checkUrl;
    this.upgradeStatusUrl = upgradeStatusUrl;
    this.offline = offline;
    this.proxy = proxy;
  }

  get mixins() {
    return [
      ...super.mixins,
      MessengerMixin,
    ];
  }

  /**
   * We read the version info from "/etc/edu_proxy/config/version"
   * We cache the result, since it can not change without a reboot
   * @returns {*}
   */
  get currentVersion() {
    if (!this._currentVersion) {
      ensureVersionFile(this._logger);
      this._updateCurrentVersion();
    }

    return this._currentVersion;
  }

  get proxy() {
    return this._proxy;
  }

  set proxy(proxy) {
    if (proxy) {
      try {
        const proxyUrl = proxyConfigToUrl(proxy);
        if (proxyUrl) {
          this._log('info', `Proxy url: ${proxyUrl}`);
          this._proxy = proxyUrl;
        } else {
          this._log('info', 'Empty proxy config, ignoring...');
        }
      } catch (e) {
        this._log('error', `Invalid proxy configuration: ${JSON.stringify(proxy)}`);
      }
    } else {
      delete this._proxy;
    }

    if (this._proxy) {
      this.agent = new ProxyAgent(this._proxy);
    } else {
      delete this.agent;
    }
  }

  async _onStart() {
    this._monitorExistingUpgrade();
    await super._onStart();
  }

  async _onStop() {
    this._stopHealthCheck();
    await super._onStop();
  }

  _updateCurrentVersion() {
    this._currentVersion = readCurrentVersion(this._logger);
  }

  /**
   * Checks whether the device is ready to upgrade
   * => True when no connected devices are in a meeting and we're at night (between 24 and 05)
   *
   * @returns {boolean}
   * @private
   */
  _readyToUpgrade(result) {
    return (
      super._readyToUpgrade(result)
      && result.hasOwnProperty('directives')
      && (
        (!result.isCompatible && !result.directives.blockMandatoryUpgrades)
        || (result.isCompatible && !result.directives.blockOptionalUpgrades)
      )
    );
  }

  get _loginCredentials() {
    const credentials = serviceRegistry.get('configManager').getSetting('credentials');
    return {
      instituteId: credentials.instituteId,
      email: credentials.user.email,
      password: credentials.user.password,
    };
  }

  get _userId() {
    return serviceRegistry.get('configManager').getSetting('credentials.user.id');
  }

  get _authToken() {
    return serviceRegistry.get('configManager').getSetting('credentials.resume.token');
  }

  get _hasAuthHeaders() {
    return (!!this._userId && !!this._authToken);
  }

  get meteorClient() {
    return serviceRegistry.get('meteorClient');
  }

  get meteorClientConnected() {
    return this.meteorClient && this.meteorClient.isConnected;
  }

  async checkVersionAndUpgradeIfNeeded(force) {
    // Always first cleanup unused packages if needed...
    // eg. During a nightly upgrade multiple upgrade steps can
    // be executed one after another. Such step can include an
    // upgrade to a new debian version. Because of this we
    // need to cleanup unused packages first before each upgrade
    // step else we risk upgrade failure (no space left on device)

    // Only do after a successful upgrade. Sometimes (nvidia!) upgrade
    // fails (watchdog timeout) and succeeds only at second attempt.
    // In such case do not cleanup between the attempts.
    if (this._checkIfCleanupUnusedPackagesNeeded()) {
      try {
        this._log('info', 'About to cleanup unused packages');
        this._removeUnusedPackages();
      } catch (error) {
        this._log('error', `Error during cleanup of unused packages, error = ${JSON.stringify(serializeError(error, { maxDepth: 1 }))}`);
      }
    } else {
      this._log('info', 'No need to cleanup unused packages');
    }
    // Then upgrade to the next version if needed
    this._log('info', 'About to check version info and upgrade if needed');
    await super.checkVersionAndUpgradeIfNeeded(force);
  }

  _scheduleHealthCheck() {
    this._stopHealthCheck();
    this._healthCheckTimer = setTimeout(async () => {
      try {
        await this.healthCheck();
      } catch (error) {}
    }, 2 * 60 * 1000);
  }

  _stopHealthCheck() {
    if (this._healthCheckTimer) {
      clearTimeout(this._healthCheckTimer);
      delete this._healthCheckTimer;
    }
  }

  markHealthy() {
    this._stopHealthCheck();
    super.markHealthy();
  }

  async healthCheck(url, version) {
    if (!url || !version) {
      try {
        const result = await this._fetchLatestVersionInfo();
        url = result.directives.repository.url;
        version = result.latest;
      } catch (error) {}
    }

    try {
      this._log('info', 'Executing health check');
      await this._validateRepositoryUrl(url, version);
      this._log('verbose', 'Health check succeeded');
      this.markHealthy();
    } catch (error) {
      this._log('error', `Health check for ${url} failed: ${error}`);
      this._scheduleHealthCheck();
      this.markUnhealthy(error.message);
      throw error;
    }
  }

  /**
   * Update device upgrade status
   *
   * Prefer to use a DDP connection if it is available, otherwise use the REST API
   *
   * @return {Promise.<*>}
   * @private
   */
  async _updateDeviceUpgradeStatus(status) {
    if (this.meteorClientConnected) {
      return this._updateDeviceUpgradeStatusOverDDP(status);
    } else if (this.upgradeStatusUrl) {
      return this._updateDeviceUpgradeStatusOverRest(status);
    }
    const error = 'Could not find a way to update device status';
    this._log('error', error);
    throw new Error(error);
  }

  async _postRequest(url, data) {
    let response;
    try {
      response = await fetch(url, {
        method: 'POST',
        body: JSON.stringify(data),
        agent: this.agent,
        headers: {
          Authorization: `Bearer ${this._authToken}`,
          'Content-Type': 'application/json',
        },
      });
    } catch (e) {
      this._log('error', 'Error while making a REST call', e);
      throw new Error(e);
    }

    if (!response) {
      throw new Error('Got no response from server');
    }

    let error;
    switch (response.status) {
    case 200:
      try {
        const body = await response.json();
        return typeof body === 'object' ? body : JSON.parse(body);
      } catch (err) {
        error = new Error('Unexpected response body (not a JSON)');
        error.statusCode = 200;
      }
      break;
    case 401:
      error = new Error('Error while logging in to Meteor REST API using resume token');
      error.statusCode = 401;
      break;
    case 403:
      error = new Error('There was a permission problem while calling REST API');
      error.statusCode = 403;
      break;
    default:
      error = new Error(`Unexpected response: ${response.statusCode}`);
      error.statusCode = response.statusCode;
    }

    this._log('error', error.message);
    throw error;
  }

  async _updateDeviceUpgradeStatusOverRest(status) {
    this._log('verbose', 'Updating device status using REST');

    if (!this._hasAuthHeaders) {
      await this._restLogin();
      return this._updateDeviceUpgradeStatus(status);
    }

    return this._postRequest(this.upgradeStatusUrl, { status: +status });
  }

  _updateDeviceUpgradeStatusOverDDP(status = true) {
    this._log('verbose', 'Updating device status using DDP');
    return this.meteorClient.setDeviceUpgradeStatus({ status });
  }

  /**
   * Fetch latest version info from Meteor
   *
   * Prefer to use a DDP connection if it is available, otherwise use the REST API
   *
   * @returns {*}
   * @private
   */
  async _fetchLatestVersionInfo() {
    this._log('info', 'Fetching latest available version info');

    if (this.offline) {
      this._log('info', "We're in offline modus, send mock data");
      return {
        minInclusive: this.currentVersion,
        maxExclusive: semver.inc(this.currentVersion, 'patch'),
        latest: this.currentVersion,
        directives: {
          blockMandatoryUpgrades: true,
          blockOptionalUpgrades: true,
        },
      };
    }

    let result = '';
    if (this.meteorClientConnected) {
      result = await this._fetchLatestVersionInfoOverDDP();
    } else if (this.checkUrl) {
      result = await this._fetchLatestVersionInfoOverRest();
    } else {
      const error = 'Could not find a way to check version info';
      this._log('error', error);
      throw new Error(error);
    }
    try {
      result.directives.repository = parseRepository(result.directives.repo);
    } catch (error) {
      result.directives.repository = {};
    }

    return result;
  }

  _fetchLatestVersionInfoOverDDP() {
    this._log('verbose', 'Checking latest version info using DDP');
    return this.meteorClient.latestDeviceVersionInfo({
      currentDeviceVersion: this.currentVersion,
      currentDeviceDate: new Date(),
    });
  }

  async _restLogin() {
    const loginCredentials = this._loginCredentials;
    this._log('info', 'Trying to log in to Meteor REST API server');

    try {
      const response = await fetch(this.loginUrl, {
        method: 'POST',
        body: JSON.stringify(loginCredentials),
        agent: this.agent,
        headers: { 'Content-Type': 'application/json' },
      });

      if (response.status === 200) {
        const body = await response.json();
        if (!body || !body.token) {
          throw new Error('REST Login did not receive a valid login token');
        }

        this._log('info', 'Logged in to Meteor REST API server.');

        const resume = body;
        delete resume.id;
        if (resume.tokenExpires) {
          resume.tokenExpires = new Date(resume.tokenExpires);
        }
        serviceRegistry.get('configManager').update('credentials.resume', resume);
        return resume.token;
      }
      throw new Error('Could not log in to Meteor REST API server');
    } catch (error) {
      this._log('error', 'Error while logging in to Meteor REST API', error);
      throw error;
    }
  }

  async _fetchLatestVersionInfoOverRest() {
    this._log('verbose', 'Checking latest version info using REST');

    if (!this._hasAuthHeaders) {
      await this._restLogin();
      return this._fetchLatestVersionInfo();
    }

    try {
      return await this._postRequest(this.checkUrl, {
        currentDeviceVersion: this.currentVersion,
        currentDeviceDate: Date.now(),
      });
    } catch (err) {
      if (err.statusCode === 401) {
        serviceRegistry.get('configManager').remove('credentials.resume');
        this._log('error', 'Login error. removing token and trying again with log in');
        await wait(5000);
        return this._fetchLatestVersionInfo();
      }
      throw err;
    }
  }

  _checkIfCleanupUnusedPackagesNeeded() {
    this._log('debug', 'Checking if we need to clean up the unused packages');
    return !fs.existsSync(path.join(LOG_DIR, CLEANUP_UNUSED_PACKAGES_LOG_FILENAME));
  }

  _nextLogFile(dir, basename, maxNbrFilesToKeep = 4) {
    let fileArray;
    // First get all the filenames in directory 'path'
    try {
      fileArray = fs.readdirSync(dir);
    } catch (error) {
      fileArray = [];
    }

    this._log('verbose', `Found following old logfiles for ${basename}: ${fileArray.filter(s => s.includes(basename))}`);

    const re = new RegExp(`^${basename}([0-9]*).log`);
    // Create a list with the numbers of the filenames in the
    // directory eg. "deviceUpgrade19.log" -> [... ,19 , ...]
    const numbers = fileArray
      .map(f => re.exec(f))
      // Only take filenames containing basename
      .filter(r => Array.isArray(r))
      // Get the number part out of the file
      .map(r => r[1]);

    // Sort the array in ascending order
    numbers.sort((a, b) => (a - b));

    // The biggest number is at the end of the array,
    // add one for number to be used for the new to be created file
    const newFileNumber = Number.parseInt((numbers.pop() || 0), 10) + 1;

    let numbersToBeDeleted = [];
    if (numbers.length > maxNbrFilesToKeep) {
      numbersToBeDeleted = numbers.slice(0, numbers.length - maxNbrFilesToKeep);
    }

    this._cleanupOldLogs(dir, basename, numbersToBeDeleted);

    return path.join(LOG_DIR, `${basename}${newFileNumber}.log`);
  }

  _cleanupOldLogs(dir, basename, numbersToBeDeleted) {
    if (numbersToBeDeleted.length) {
      // Now delete old logfiles
      this._log('verbose', `About to delete ${numbersToBeDeleted.length} old logfiles for ${basename}...`);
      numbersToBeDeleted
        .forEach((n) => {
          const fileToBeDeleted = path.join(dir, `${basename}${n}.log`);
          this._log('verbose', `About to delete ${fileToBeDeleted}...`);
          try {
            fs.unlinkSync(fileToBeDeleted);
          } catch (error) {}
        });
    } else {
      this._log('info', `No old logfiles to be deleted for ${basename}...`);
    }
  }

  _copyUpgradeScriptsAndDeleteOldResultFile() {
    // Delete the old upgrade results file
    try {
      fs.unlinkSync('/tmp/upgradeResult.txt');
    } catch (error) {}
    // Use a copy of the scripts to avoid upgrade stops when barco-edu-proxy package itself is upgraded
    this._log('info', 'Creating copy of upgrade script...');

    try {
      fs.unlinkSync('/tmp/scriptLauncher.sh');
    } catch (error) {}
    try {
      fs.unlinkSync('/tmp/upgradeDevice.sh');
    } catch (error) {}

    fs.copySync(path.join(SCRIPTS_DIR, 'scriptLauncher.sh'), '/tmp/scriptLauncher.sh');
    fs.copySync(path.join(SCRIPTS_DIR, 'upgradeDevice.sh'), '/tmp/upgradeDevice.sh');
  }

  // In case we are launched as the result of another upgrade, we have to check the process externally
  // This means finding the PID and using the process monitor to periodically check it.
  async _monitorExistingUpgrade() {
    // if this.isUpgrading === true -> upgrade launched in this process, no need to check external process
    if (this.isUpgrading) return;

    const upgradeProcesses = execSync('ps -ef | grep \'upgradeDevice.sh\\|upgradeAutoRemoveUnusedPackages.sh\'')
      .toString()
      .trim()
      .split('\n')
      .filter(line => line.includes('scriptLauncher.sh'))
      .map(line => line.split(/\s+/))
      .map(([, pid, ppid]) => ({
        pid: Number.parseInt(pid, 10),
        ppid: Number.parseInt(ppid, 10),
      }));
    // You always have the 2 processes from the command itself (ps & grep)
    this.isUpgrading = upgradeProcesses.length > 0;
    if (!this.isUpgrading) return;

    const upgradePid = upgradeProcesses[0].pid;
    this._log('info', `Detected existing upgrade with process: ${upgradePid}`);
    const upgradePromise = waitForUpgradeProcessToFinish(upgradePid);
    try {
      await this._waitForUpgradeToFinish(upgradePromise);
    } catch (e) {}
  }

  /**
   * Removes unused debian packages
   *
   * @private
   */
  async _removeUnusedPackages() {
    if (this.isUpgrading) {
      const error = 'Can not cleanup unused packages, since an other upgrade action is currently in progress';
      this._log('error', error);
      throw new Error(error);
    }

    this.isUpgrading = true;
    this.emit('cleanup_start');

    try {
      // Now launch the device upgrade script
      const upgradeCleanupLogFileName = path.join(LOG_DIR, CLEANUP_UNUSED_PACKAGES_LOG_FILENAME);
      const upgradeCleanupLogFile = fs.openSync(upgradeCleanupLogFileName, 'w');
      const scriptFileName = path.join(SCRIPTS_DIR, 'upgradeAutoRemoveUnusedPackages.sh');

      // If cleanup process takes longer than 15 minutes, kill it.
      this._log('info', 'About to launch the actual cleanup process');
      const cleanupProcess = spawn(path.join(SCRIPTS_DIR, 'scriptLauncher.sh'), [900, 'false', scriptFileName], {
        stdio: ['ignore', upgradeCleanupLogFile, upgradeCleanupLogFile],
      });
      await waitForChildProcessToFinish(cleanupProcess);
      this._log('info', 'Cleanup script finished');
    } catch (error) {
      this._log('error', 'Error during cleanup of unused packages', error);
    } finally {
      this.isUpgrading = false;
    }

    this._log('info', 'Cleanup of unused packages completed');
    this.emit('cleanup_done');
    return true;
  }

  _showUpgradeMessage() {
    this._log('info', 'Display warning message');
    this._messenger.startProgress();
    return this._showMessage({
      text: 'DEVICE UPGRADE IN PROGRESS. PLEASE DO NOT TURN OFF DEVICE',
    });
  }

  async _sufficientDiskSpaceAvailableForUpgrade() {
    const availableDiskSpace = getAvailableDiskSpace();
    this._log('info', `Current available Disk space before Upgrade is ${availableDiskSpace}KB`);
    if (availableDiskSpace < MINIMUM_REQUIRED_FREE_DISKSPACE_KB_FOR_UPGRADE) {
      this._log('info', 'Restarting nrc-logcleaner service since current available Disk space is less than minimum required 6GB for the Device Upgrade');
      await promisify(exec)('sudo systemctl restart nrc-logcleaner.service --wait');
      const diskUsageCleanup = getAvailableDiskSpace();
      if (diskUsageCleanup < MINIMUM_REQUIRED_FREE_DISKSPACE_KB_FOR_UPGRADE) {
        this._log('info', `Available disk space after nrc-logcleaner service restart is ${diskUsageCleanup} is still less then minimum required ${MINIMUM_REQUIRED_FREE_DISKSPACE_KB_FOR_UPGRADE}KB for the Device Upgrade`);
        return false;
      } else {
        this._log('info', `Available disk space after nrc-logcleaner service restart is ${diskUsageCleanup} KB, which is sufficient for upgrade.`);
        return true;
      }
    }
    return true;
  }

  _sufficientBootsSpaceAvailableForUpgrade() {
    const availableBootSpace = getAvailableBootSpace();
    this._log('info', `The current available boot space is ${availableBootSpace}KB`);
    return availableBootSpace > MINIMUM_REQUIRED_FREE_BOOTSPACE_KB_FOR_UPGRADE;
  }

  async _checkDiskSpaceBeforeUpgrade() {
    const sufficientDiskSpaceAvailable = await this._sufficientDiskSpaceAvailableForUpgrade();
    const sufficientBootSpaceAvailable = this._sufficientBootsSpaceAvailableForUpgrade();
    if (!sufficientBootSpaceAvailable) {
      this._log('error', `Aborting Upgrade as boot space is less than minimum required for Device Upgrade. Minimum ${MINIMUM_REQUIRED_FREE_BOOTSPACE_KB_FOR_UPGRADE}KB required`, {
        error: 'Insufficient free disk space in boot for upgrade',
        serviceLog: true,
      });
    }
    if (!sufficientDiskSpaceAvailable) {
      this._log('error', `Aborting Upgrade as current Available Disk space is less than minimum required space for Device Upgrade. Minimum ${MINIMUM_REQUIRED_FREE_DISKSPACE_KB_FOR_UPGRADE}KB required`, {
        error: 'Insufficient free disk space in / dir for upgrade',
        serviceLog: true,
      });
    }
    return sufficientBootSpaceAvailable && sufficientDiskSpaceAvailable;
  }

  /**
   * Does the actual upgrade:
   * 1) Verify repo is reachable
   * 2) Write source file
   * 3) Upgrade using upgrade bash script
   *
   * @param repository
   * @returns {Promise}
   * @private
   */
  async _doUpgrade(directives) {
    try {
      try {
        await this._updateDeviceUpgradeStatus(true);
      } catch (err) {
        this._log('error', 'Failed to set device upgrading status. But still continue the upgrade');
      }
      this._log('info', 'Starting actual upgrade procedure', {
        currentVersion: this.currentVersion,
        upgradeVersion: directives.upgradeTo,
        serviceLog: true,
      });

      // Make sure we can reach the upgrade source
      await this.healthCheck(directives.repository.url, directives.upgradeTo);

      // Prepare upgrade
      await this._createSourcesList(directives.repository.full);
      await this._createAptConfig(directives.repository.parsed_url);
      // Cleanup old logfiles of the device upgrades,
      // so that at most 5 files (= 4 old + the new to be created) exist
      const deviceUpgradeLogFileName = this._nextLogFile(LOG_DIR, UPGRADE_LOG_BASENAME, 4);
      const deviceUpgradeLogFile = fs.openSync(deviceUpgradeLogFileName, 'a');
      this._log('verbose', `Name of logfile to be created for device upgrade: ${deviceUpgradeLogFileName}`);

      // Copy the upgrade scripts to /tmp
      this._copyUpgradeScriptsAndDeleteOldResultFile();

      this._log('info', 'Upgrade preparation done, now starting actual upgrade');

      const message = this._showUpgradeMessage();
      // Make sure message is showed for at least 5s in case of a really fast upgrade
      await sleep(5000);

      // Now launch the device upgrade script.
      // Tell the watchdog timer to reboot if process still running after 2 hours
      const params = [7200, true, '/tmp/upgradeDevice.sh', sourcesListFile, aptConfigFile];
      const upgradeProcess = spawn('/tmp/scriptLauncher.sh', params, {
        detached: true,
        stdio: ['ignore', deviceUpgradeLogFile, deviceUpgradeLogFile],
      });
      upgradeProcess.unref();

      try {
        this._log('info', 'Waiting for upgrade to finish', { serviceLog: true });
        await waitForChildProcessToFinish(upgradeProcess);
        message.hide();
        this._showMessage({
          text: 'finished upgrade',
        });
        this._showMessage({
          text: 'rebooting ...',
        });
        return true;
      } catch (error) {
        message.hide();
        this._messenger.stopProgress();
        await this._showMessage({
          text: 'failed to upgrade device',
          time: 30,
        });
        throw error;
      }
    } finally {
      // if there was an error in the try block, we don't want to
      // accidentally overwrite it with the error in the finally block
      try {
        await this._updateDeviceUpgradeStatus(false);
      } catch (err) {
        this._log('error', 'Failed to set device upgrading status.');
      }
    }
  }

  /**
   * See if the repository is reachable before we try an upgrade, in order to avoid infinite loops
   *
   * Note that the errors thrown here will be sent to blue!
   */
  async _validateRepositoryUrl(url, requiredVersion) {
    this._log('info', `Validating repository url: ${url} for version ${requiredVersion}`);

    url += '/version';

    let response;
    try {
      response = await fetch(url, {
        agent: this.agent,
      });
    } catch (e) {
      throw new Error(`Invalid repository url. ${e}`);
    }

    if (response.status === 200) {
      const body = await response.text();
      this._log('info', `Repository ${url} is reachable.`);
      this._log('info', `Repository version: ${body}.`);

      if (_.isString(body) && body.trim() === requiredVersion) {
        return true;
      }
      throw new Error(`Invalid repository: version mismatch. Requested: ${requiredVersion}, Actual: ${body}`);
    } else if (response) {
      throw new Error(`Inaccessible repository. Statuscode: ${response.statusCode}`);
    } else {
      throw new Error('Invalid repository.');
    }
  }

  /**
   * Creating apt proxy config file
   * @param repository
   * @returns {Promise<T>|Promise}
   * @private
   */
  async _createAptConfig(repository) {
    this._log('info', 'Creating apt config file');

    // The No-Cache and Pipeline-Depth options help us prevent hash sum errors
    // These are mainly because of APT bugs and should be re-evaluated in the future.
    // The pipeline depth is an incompatibility between HTTP 1.0 proxies and apt which assumes HTTP 1.1
    // The no-cache just sends an HTTP header to not return cached results
    // In theory both of these should only matter in the case of an HTTP proxy
    // Since our servers are HTTP 1.1 compliant and there's no caching involved
    // But as long as we run over HTTP, there may be MITM proxies (e.g. reverse proxies by the institute)
    // Or there may be firewall policies like head-of-line blocking intefering with this.
    //
    // So we add the options here, but they should be re-evaluated in the future
    // and only applied to the proxy use case once we have HTTPS proxy support
    let fileContents = `
Debug::Acquire::http "true";
Debug::Acquire::https "true";
Acquire::http::No-Cache "true";
Acquire::http::Pipeline-Depth "0";
`;

    if (this.proxy) {
      this._log('verbose', 'Adding apt proxy config...');
      // HTTPS proxy was giving trouble with apt.
      // So for now we only proxy over HTTP
      // But we should verify some day if this is still a problem today.
      fileContents += `
Acquire::http::Proxy "${this.proxy}";
//  Acquire::https::proxy "${this.proxy}";
//  Acquire::https::${repository.hostname}::Verify-Peer "true";
//  Acquire::https::${repository.hostname}::Verify-Host "true";
//  Acquire::https::${repository.hostname}::CaInfo      "/etc/apt/certs/my-cacert.pem";
//  Acquire::https::${repository.hostname}::SslCert     "/etc/apt/certs/my-cert.pem";
//  Acquire::https::${repository.hostname}::SslKey      "/etc/apt/certs/my-key.pem";
//  Acquire::https::${repository.hostname}::SslForceVersion "SSLv3";
`;
    }

    fileContents = fileContents.trim();

    try {
      this._log('debug', `apt config: ${fileContents}`);
      await fs.writeFile(aptConfigFile, fileContents);
    } catch (error) {
      this._log('error', `Error while writing apt config file ("${aptConfigFile}")`, error);
      throw error;
    }
    this._log('info', `Written apt config file to "${aptConfigFile}"`);
    return aptConfigFile;
  }

  /**
   * Create the apt repository file
   * @param repository
   * @returns {Promise}
   * @private
   */
  async _createSourcesList(repository) {
    this._log('info', `Creating sources list file "${sourcesListFile}" for repository "${repository}"`);

    const fileContents = `deb [trusted=yes] ${repository}`;
    try {
      await fs.writeFile(sourcesListFile, fileContents);
    } catch (error) {
      this._log('error', `Error while writing source file ("${sourcesListFile}")`, error);
      throw error;
    }

    this._log('info', `Written sources list file "${sourcesListFile}" for repository "${repository}`);
    return true;
  }

  /**
   * Compary Cinergy versions
   * Cinergy versions are NOT semver and come in the form of [major].[minor].[patch]-b[build]
   * Left out numbers are filled in with 0
   *
   * Examples:
   * 1.3 === 1.3.0-b28
   * 1.3 === 1.3.0
   * 1.3 < 1.3.1
   * 1.3.0 < 1.3.0-b2
   *
   * 1.2.0-b24 < 1.2.0-b25
   * 1.2.0-b24 < 1.2.1-b24
   *
   * @param a
   * @param b
   * @returns {number}
   * @private
   */
  _compareVersions(a, b) {
    return compareVersions(a, b);
  }
}

export default SystemVersionMonitor;
