import * as semver from 'semver';
import * as _ from 'lodash';

import Service from '../service';

/**
 * A VersionMonitor will check for a new version on a regular interval (or when called manually).
 * It calculates whether the current version is still compatible (>= minimum compatible version)
 */

const VERSIONINFO_LIFETIME = 5 * 60 * 1000;

/**
 * Abstract version monitor
 *
 * The Monitor can be set to check for new versions at regular intervals and operates by these rules:
 *
 * -In case the current version is no longer compatible
 *  and a new version is available, an upgrade is performed immediately
 * -In case the current version is no longer compatible
 *  and the new version is older than current version,
 *  an upgrade is only performed if the version monitor supports downgrades
 * -In case the current version is still compatible and a new version is available,
 *  an upgrade is performed if the _readyToUpgrade function returns true.
 * This is so that non-critical updates can be executed on a time when it's convenient
 */
class VersionMonitor extends Service {
  constructor({
    checkInterval = 3600000,
  }) {
    super({});

    const params = {
      checkInterval,
    };
    Object.keys(params).forEach((param) => {
      if (params[param] === undefined) throw new Error(`Please provide a valid ${param}`);
      this[`_${param}`] = params[param];
    });

    this._checkIntervalHandle = null;
  }

  get currentVersion() {
    throw new Error("Version monitor should implement the 'currentVersion' getter");
  }

  get supportDowngrades() {
    return false;
  }

  /**
   * Fetches the latest available version info and induces whether we are still compatible and can upgrade or not
   * @returns {Promise.<T>}
   */
  async updateVersionInfo() {
    this._log('info', 'Checking version information');
    const versionInfo = await this._fetchLatestVersionInfo();

    if (!_.isObject(versionInfo) || !versionInfo.hasOwnProperty('minInclusive') || !versionInfo.hasOwnProperty('maxExclusive')) {
      const error = 'Invalid versioninfo, no min or max version specified';
      this._log('error', error);
      throw new Error(error);
    }

    const isCompatible = (this._compareVersions(this.currentVersion, versionInfo.minInclusive) >= 0
      && this._compareVersions(this.currentVersion, versionInfo.maxExclusive) === -1);

    const updateAvailable = this._compareVersions(this.currentVersion, versionInfo.latest) === -1
      || (this._compareVersions(this.currentVersion, versionInfo.latest) === 1 && this.supportDowngrades);

    Object.assign(versionInfo, {
      isCompatible,
      updateAvailable,
    });

    // readyToUpgrade might need isCompatible && updateAvailable
    // so dont' add to the assign above.
    versionInfo.readyToUpgrade = this._readyToUpgrade(versionInfo);
    versionInfo.expiresAt = new Date(Date.now() + VERSIONINFO_LIFETIME);

    this._log('info', `

                =================================================================================
                ==                                                                             ==
                ==                          Latest version information                         ==
                ==                                                                             ==
                =================================================================================

                Current Version                         : ${this.currentVersion}
                Minimum compatible version (inclusive)  : ${versionInfo.minInclusive}
                Highest compatible version (exclusive)  : ${versionInfo.maxExclusive}
                Latest available version                : ${versionInfo.latest}
                Is current version compatible           : ${versionInfo.isCompatible}
                Update available                        : ${versionInfo.updateAvailable}
                Ready to upgrade                        : ${versionInfo.readyToUpgrade}
            `);

    this.versionInfo = versionInfo;
    return versionInfo;
  }

  updateVersionInfoIfStale() {
    if (!this.versionInfo) return this.updateVersionInfo();
    if (this.versionInfo.expiresAt < new Date()) {
      return this.updateVersionInfo();
    }
    this._log('verbose', 'No need to update versioninfo, still fresh');
    return Promise.resolve(this.versionInfo);
  }

  /**
   * Fetches latest version info, returns an object that looks like
   * {
   *  minInclusive: Minimum compatible version (inclusive),
   *  maxExclusive: Highest compatible version (exclusive)
   *  latest: Latest available version
   *  directives: Object with instructions to be used by the upgrade function (e.g. debian repository, ...)
   * }
   * @private
   */
  _fetchLatestVersionInfo() {
    throw new Error("Version monitor should implement a '_fetchLatestVersionInfo' method");
  }

  /**
   * Start the interval checker
   */
  async _onStart() {
    this._log('verbose', 'Starting version monitor');
    await this.updateVersionInfo();
    if (!this.isUpgrading) {
      this._startCheckInterval();
    }
    return super._onStart();
  }

  /**
   * Stop checking for new versions
   */
  _onStop() {
    this._clearCheckIntervalHandle();
    return super._onStop();
  }

  /**
   * Execute an upgrade given some instructions (directives)
   * @param directives
   */
  async upgrade(directives) {
    if (this.isUpgrading) {
      const error = 'Can not execute upgrade, since an other upgrade is already in progress';
      this._log('error', error);
      throw new Error(error);
    }

    this.isUpgrading = true;
    this._log('info', 'About to start actual device upgrade');
    this.emit('upgrade_start');

    const upgradePromise = this._doUpgrade(directives);
    await this._waitForUpgradeToFinish(upgradePromise);
    return true;
  }

  async _waitForUpgradeToFinish(upgradePromise) {
    this._clearCheckIntervalHandle();
    try {
      await upgradePromise;
    } catch (error) {
      this._log('error', 'Error during upgrade', error);
      this.emit('upgrade_error', error);
      throw new Error(error);
    } finally {
      this.isUpgrading = false;
      if (this.isActive) {
        await this._startCheckInterval();
      }
    }

    this._log('info', 'Upgrade completed');
    this.emit('upgrade_done');
  }

  /**
   * Check whether we should upgrade and starts the upgrade if we do
   *
   * We upgrade if an update is available AND if it's needed
   * Needed = either if forced by passing the flag or if the cloud says the device is ready to upgrade
   * Ready to upgrade = either we're no longer compatible, or we're still compatible and it's a good moment
   * Good moment = e.g. at night when the device is not being used
   * @private
   */
  async checkVersionAndUpgradeIfNeeded(force = false) {
    await this.updateVersionInfoIfStale();

    // Start an upgrade if there's one available and if diskspace requirements are met
    if (this.versionInfo.updateAvailable && (force || this.versionInfo.readyToUpgrade)) {
      const sufficientDiskSpace = await this._checkDiskSpaceBeforeUpgrade();
      if (!sufficientDiskSpace) {
        throw new Error('Insufficient disk space available for upgrade');
      }
      const directives = {
        ...this.versionInfo.directives,
        upgradeTo: this.versionInfo.latest,
      };
      await this.upgrade(directives);
      return true;
    } else {
      if (this.versionInfo.isCompatible) {
        return true;
      }
      throw new Error("Version is not compatible and there's no update available");
    }
  }

  async isCurrentVersionCompatible() {
    await this.updateVersionInfoIfStale();
    return this.versionInfo.isCompatible;
  }

  /**
   * Forced upgrade
   */
  async checkVersionAndUpgrade() {
    this._log('info', 'Start force upgrade');
    return this.checkVersionAndUpgradeIfNeeded(true);
  }

  /**
   * Checks disk space in / & /boot disk before upgrade
   * @private
   */
  async _checkDiskSpaceBeforeUpgrade() {
    throw new Error("Version monitor should implement a '_checkDiskSpaceBeforeUpgrade' method");
  }

  /**
   * The actual upgrade procedure
   * @param directives
   * @private
   */
  _doUpgrade() {
    throw new Error('Version monitor should implement the upgrade mechanism');
  }

  /**
   * Compare 2 versions
   * @param a
   * @param b
   * @returns -1 if a < b, 0 if a===b, 1 if a > b
   * @private
   */
  _compareVersions(a, b) {
    return semver.compare(a, b);
  }

  /**
   * Whether the device is ready to upgrade
   * @returns {boolean}
   * @private
   */
  _readyToUpgrade() {
    return !this.isUpgrading;
  }

  /**
   * Start the interval check
   * @private
   */
  _startCheckInterval() {
    this._checkIntervalHandle = setInterval(
      this.checkVersionAndUpgradeIfNeeded.bind(this),
      this._checkInterval,
    );
  }

  /**
   * Stop the interval check
   * @private
   */
  _clearCheckIntervalHandle() {
    if (this._checkIntervalHandle) {
      clearInterval(this._checkIntervalHandle);
      this._checkIntervalHandle = null;
    }
  }
}

export default VersionMonitor;
