import * as path from 'path';
import {
  BrowserWindow,
  session as browserSession,
} from 'electron';
import { enable as enableRemote } from '@electron/remote/main';
import * as chromeErrors from 'chrome-network-errors';
import * as isDev from 'electron-is-dev';

import { proxyConfigToUrl } from '../network';
import BrowserConnectionMonitor from './browserConnectionMonitor';

const OK_ERROR_CODE = 0;

function getProxyConfig(proxy) {
  let proxyRules = '';
  const proxyUrl = proxyConfigToUrl(proxy, {
    removeProtocol: true,
    removeTrailingSlash: true,
  });
  if (proxyUrl) {
    proxyRules = `http=${proxyUrl};https=${proxyUrl}`;
  }
  return {
    proxyRules,
    proxyBypassRules: '<local>',
  };
}

async function createBrowserSession(privileged, persistent, name, proxy, logger) {
  logger.log('info', `Creating ${privileged ? '' : 'un'}privileged `
                + `${persistent ? 'persistent' : ''} browser session with name '${name}'`);

  const session = browserSession.fromPartition(`${persistent ? 'persist:' : ''}${name}`);
  session.privileged = privileged;

  session.setPermissionRequestHandler((webContents, permission, callback) => {
    privileged ? callback(true) : callback(permission === 'fullscreen');
  });

  session.on('will-download', (event) => {
    event.preventDefault();
  });

  if ((isDev || process.env.NODE_TLS_REJECT_UNAUTHORIZED === '0') && privileged) {
    session.setCertificateVerifyProc((request, callback) => {
      callback(0);
    });
  }

  await session.setProxy(getProxyConfig(proxy));
  return session;
}

function createBrowserWindow(session, data, bounds, logger, openDevTools = false, visible = true, url = '') {
  const currentWindow = new BrowserWindow({
    frame: false,
    show: false,
    kiosk: false,
    thickFrame: false,
    // our browser windows should not be resizable since this can lead to issues on touch screens
    resizable: false,
    // below 2 properties are required for electron to use the actual full screen bounds!
    // XXX: change if we ever want non-fullscreen windows
    useContentSize: true,
    fullscreen: false,
    webPreferences: {
      experimentalFeatures: session.privileged,
      nodeIntegration: false,
      nodeIntegrationInWorker: false,
      contextIsolation: !session.privileged,
      preload: session.privileged ? path.join(__dirname, 'preload.js') : path.join(__dirname, 'heartbeat.js'),
      session,
      sandbox: !session.privileged,
      devTools: openDevTools,
      ...(url.includes('offwall') && { autoplayPolicy: 'document-user-activation-required' }),
    },
    ...bounds,
  });

  if (session.privileged) {
    enableRemote(currentWindow.webContents);
  }

  currentWindow.data = data;

  currentWindow._connectionMonitor = new BrowserConnectionMonitor({ win: currentWindow });
  currentWindow.bringToFront = () => {
    if (!currentWindow.isDestroyed()) {
      currentWindow.setAlwaysOnTop(true);
    }
  };

  currentWindow.openDevtools = () => {
    currentWindow.webContents.openDevTools();
    currentWindow.webContents.insertCSS('body { cursor: auto }');
  };

  currentWindow.once('closed', () => {
    if (currentWindow._connectionMonitor) {
      currentWindow._connectionMonitor.destroy();
      delete currentWindow._connectionMonitor;
    }
  });

  // Hide mouse cursor or open devtools if required
  currentWindow.once('ready-to-show', () => {
    currentWindow._connectionMonitor.start().catch(err => logger.log('error', err.message));
    if (visible) {
      currentWindow.show();
    }
    if (!isDev) {
      currentWindow.webContents.insertCSS('body { cursor: none }');
    }
    if (openDevTools) {
      currentWindow.openDevTools();
    }
  });

  // Error handling
  currentWindow.webContents.on('did-fail-load', (event, code, description, validatedURL) => {
    // Error code meaning can be found: https://code.google.com/p/chromium/codesearch#chromium/src/net/base/net_error_list.h
    if (code === OK_ERROR_CODE) {
      // See: https://github.com/electron/electron/issues/9647
      logger.log('info', `Ok code error ? while loading ${validatedURL}`);
    } else {
      logger.log('error', `Error while loading ${validatedURL}: ${code} (${chromeErrors[code]}): ${description}`);
    }
  });

  return currentWindow;
}

export {
  getProxyConfig,
  createBrowserSession,
  createBrowserWindow,
};
