import { screen } from 'electron';

function bb(boxes) {
  let bbox = {
    x: 0, y: 0, width: 0, height: 0,
  };

  if (boxes.length) {
    bbox = boxes.reduce((prev, cur) => ({
      x: Math.min(prev.x, cur.x),
      y: Math.min(prev.y, cur.y),
      width: Math.max(prev.width, (cur.x + cur.width)),
      height: Math.max(prev.height, (cur.y + cur.height)),
    }), {
      x: Number.POSITIVE_INFINITY,
      y: Number.POSITIVE_INFINITY,
      width: Number.NEGATIVE_INFINITY,
      height: Number.NEGATIVE_INFINITY,
    });

    bbox.width -= bbox.x;
    bbox.height -= bbox.y;
  }

  return bbox;
}

/**
   * Electron incorporates a scaling factor in its display geometry.
   * This maps a sinks bounds (without scaling factor, coming from the X server)
   * to electron display bounds.
   *
   * Note that this does actually something simpler, since we only have fullscreen windows.
   * This avoids us having to deal with the rounding issues
   * If we do not have full screen windows, we need to do something more advanced.
   *
   * This also assumes displays are all placed next to each other
   */
function translateDisplayBounds({
  x, y, width, height,
}) {
  if (process.env.DISABLE_DISPLAY_MAPPING) {
    // In some environments (e.g. jan dewitte's board room) we split up a screen
    // in multiple virtual screens by faking the xrandr results
    // however, electron would still give us the real coordinates, so we want
    // to disable this mapping.
    // The downside is that this will screw up window positions when there are screens
    // that have a scalingFactor != 1.
    return {
      x, y, width, height,
    };
  }

  const displays = screen.getAllDisplays();

  const includeDisplay = (display) => {
    const l1 = x;
    const l2 = display.bounds.x * display.scaleFactor;
    const r1 = x + width;
    const r2 = l2 + (display.bounds.width * display.scaleFactor);

    return (
      Math.abs(l1 - l2) < 10 // to ignore rounding errors
      || Math.abs(r1 - r2) < 10 // to ignore rounding errors
      || (l1 <= l2 && r1 >= r2) // fully in, e.g. middle of 3 extended displays
    );
  };

  const spanningDisplays = displays.filter(includeDisplay);

  return bb(spanningDisplays.map(d => d.bounds));
}

export {
  bb,
  translateDisplayBounds,
};
