import * as fs from 'fs-extra';

/**
 * We assume a device can crash at any time
 * Therefore, when writing contents to a file, the device might crash in the middle of that
 * For some files, like config files, this might render the device useless.
 *
 * Therefore we first write the contents to a temporary file and then we replace the temporary file
 * This replacing is a rename, which is atomic.
 *
 * All file actions here are synchronous and block the entire event loop.
 * We accept this small performance loss to make this easier to reason about and therefore more bulletproof.
 */

const newFileName = name => (`${name}.new`);
const tmpFileName = name => (`${name}.tmp`);

/**
 * Ensure latest version of file.
 * This means if there is a file with .new extension, we replace that one
 *
 * This might happen because although renaming a file is atomic, you can not rename a file to a file that already exists
 * This means that we first need to delete the file we want to rename the file to (done by moveSync)
 * In between this process the device might crash, leaving us with a file not yet renamed and the other file deleted.
 * This has to be called when a device starts to make sure all files are in place
 * @param {*} file
 * @param {*} logger
 */
function ensureLatest(file, logger) {
  if (logger) {
    logger.log('verbose', `Ensuring latest version of file ${file}`);
  }
  const newFile = newFileName(file);
  if (fs.existsSync(newFile)) {
    if (logger) {
      logger.log('verbose', `There is a newer version of file ${file} => replacing anatomically`);
    }
    fs.moveSync(newFile, file, {
      overwrite: true,
    });
  }
}

/**
 * Replaces a files contents in a safe manner, this means:
 * 1) First ensure the original version is the latest one, this also makes sure there is no other .new file.
 * 2) Write a temporary version
 * 3) After write has ben completed, rename the temporary file to .new
 * 4) Rename the .new file to the original file.
 *
 * All of these steps are necessary!
 *
 * @param {*} file
 * @param {*} contents
 * @param {*} logger
 */
function writeSafely(file, contents = '', logger) {
  if (logger) {
    logger.log('verbose', `Safely writing file ${file} with ${contents.trim().length ? 'new' : 'empty'} contents`);
  }

  ensureLatest(file, logger);
  const tmpFile = tmpFileName(file);
  const newFile = newFileName(file);

  if (logger) {
    logger.log('verbose', `Writing contents to ${tmpFile}`);
  }
  fs.writeFileSync(tmpFile, contents, 'utf-8');
  fs.renameSync(tmpFile, newFile);
  ensureLatest(file, logger);
}

/**
 * Copies a file using the write safely method
 */
function copySafely(from, to, logger) {
  if (logger) {
    logger.log('verbose', `Safely copying from ${from} to ${to}`);
  }

  let contents;
  try {
    contents = fs.readFileSync(from, 'utf-8');
  } catch (error) {
    if (logger) {
      logger.log('verbose', `No source file: ${from}`);
    }
  }

  try {
    writeSafely(to, contents, logger);
  } catch (error) {
    if (logger) {
      logger.log('verbose', `Error while safely writing to ${to}`);
    }
    throw error;
  }
}

/**
 * Try to remove a file, but don't throw an error if it fails
 * @param {*} file
 * @param {*} logger
 */
function tryRemove(file, logger) {
  try {
    if (logger) {
      logger.log('verbose', `Trying to remove ${file}`);
    }
    fs.removeSync(file);
  } catch (e) {
    if (logger) {
      logger.log('verbose', `Could not remove ${file}`);
    }
  }
}

/**
 * Remove a file with all derived versions (temporary and new)
 * that might have been written
 * @param {*} file
 * @param {*} logger
 */
function removeAll(file, logger) {
  tryRemove(file, logger);
  tryRemove(newFileName(file), logger);
  tryRemove(tmpFileName(file), logger);
}

export {
  ensureLatest, copySafely, writeSafely, tryRemove, removeAll,
};
