/**
 * (C) 2015 Copyright BARCO NV
 * All rights reserved. Reproduction in whole or in part is
 * prohibited without the written consent of the copyright owner.
 *
 * Part of Barco WeConnect
 */

import * as os from 'os';
import { URL } from 'url';
import * as _ from 'lodash';
import * as IP from 'ip';

const defaultLogger = {
  log(lvl, msg) {
    console.log('defaultLogger', msg); // eslint-disable-line
  },
};

let cachedInterfaces;
function getInterfaces(logger = defaultLogger, forceRefresh = false) {
  if (!cachedInterfaces || forceRefresh) {
    cachedInterfaces = os.networkInterfaces();
    logger.log('info', `Found network interfaces: ${JSON.stringify(cachedInterfaces, null, 2)}`);
  }
  return cachedInterfaces;
}

let cachedIpAddress;

function getIpAddress(logger = defaultLogger, forceRefresh = false) {
  if (cachedIpAddress && !forceRefresh) {
    return cachedIpAddress;
  }

  if (!logger) {
    logger = defaultLogger;
  }

  const ifaces = getInterfaces(logger, forceRefresh);

  const bestIface = Object.keys(ifaces)
    .map(key => ifaces[key])
    .reduce((flat, iface) => [...flat, ...iface], []) // flatten
    .filter(iface => !iface.internal && iface.family === 'IPv4') // filter ipv4
    .reduce((best, iface) => {
      if (!best || (IP.isPrivate(best) && !IP.isPrivate(iface))) {
        return iface;
      }
      return best;
    }, false);

  if (bestIface) {
    cachedIpAddress = bestIface.address;
    logger.log('verbose', `Settling on IP: ${cachedIpAddress}`);
    return cachedIpAddress;
  }
  logger.log('warn', 'Did not find any valid IP');
}

function normalizeMac(mac = '') {
  return mac.replace(/[^0-9a-z]/gi, '').toUpperCase();
}

function getMacAddress(logger, forceRefresh = false) {
  const ip = getIpAddress(logger, forceRefresh);
  const ifaceFromIp = Object.values(getInterfaces(logger))
    .reduce((acc, el) => [...acc, ...el], [])
    .find(iface => iface.address === ip);

  if (ifaceFromIp) {
    return normalizeMac(ifaceFromIp.mac);
  }
}

function proxyConfigToUrl(proxyConfig, { removeProtocol = false, removeTrailingSlash = true } = {}) {
  if (!proxyConfig || !proxyConfig.url || !proxyConfig.url.trim().length) return;
  proxyConfig.url = proxyConfig.url.trim();

  let proxy;

  try {
    let setProtocol = false;
    if (!proxyConfig.url.startsWith('http')) {
      proxyConfig.url = `http://${proxyConfig.url}`;
      setProtocol = true;
    }

    const url = new URL(proxyConfig.url);

    if (proxyConfig.user) {
      url.username = proxyConfig.user;
    }

    if (proxyConfig.password) {
      url.password = proxyConfig.password;
    }

    if (setProtocol && url.port) {
      switch (Number.parseInt(url.port, 10)) {
      case 443:
        url.protocol = 'https';
        break;
      case 80:
      default:
        url.protocol = 'http';
      }
    }

    if (!url.port) {
      switch (url.protocol) {
      case 'https:':
        url.port = 443;
        break;
      case 'http:':
      default:
        url.port = 80;
      }
    }

    proxy = url.toString();

    if (removeProtocol) {
      proxy = proxy.substr(url.protocol.length + 2);
    }

    if (removeTrailingSlash && proxy[proxy.length - 1] === '/') {
      proxy = proxy.substr(0, proxy.length - 1);
    }

    return proxy;
  } catch (error) {}
}

function waitForNetwork(messenger, logger) {
  if (!getIpAddress(logger, true)) {
    return messenger.show({
      text: 'No IPv4 address assigned. Please solve and reboot.',
      time: 5,
    }).then(() => waitForNetwork(messenger, logger));
  }
}

export {
  waitForNetwork,
  getIpAddress,
  proxyConfigToUrl,
  getMacAddress,
  normalizeMac,
};
