import { Transform } from 'stream';
import * as recorder from 'node-record-lpcm16';
import Service from '../../services/service';
import { CAPTION_TYPE } from './constants';
import RevaiSTTShim from './shims/RevaiSTTShim';
import AzureSTTShim from './shims/AzureShim';

const defaultRecordingConfig = {
  encoding: 'LINEAR16',
  keepSilence: true,
  languageCode: 'en-US',
  recordProgram: 'rec', // Try also "arecord" or "sox"
  sampleRateHertz: 16000,
  silence: 1000,
  streamingLimit: 30000,
  threshold: 0, // Silence threshold
};

// code based on https://github.com/googleapis/nodejs-speech/blob/master/samples/infiniteStreaming.js
class SpeechToTextStream extends Service {
  constructor(id, clientConfig, recordingConfig, sessionConfig, type) {
    super({ id });

    // credentials for the speech-to-text api client
    this.clientConfig = clientConfig;

    // type of the api for the speech-to-text api client
    this.type = type;

    // configuration options for the speech-to-text detection
    this.sessionConfig = sessionConfig;

    // recording configuration options
    this.recordingConfig = {
      ...defaultRecordingConfig,
      ...recordingConfig,
    };

    this.apiClient = this.getApiClient();

    // Used for correcting times after a restart
    this._restartRecognizeCounter = 0;
    this._audioInput = [];
    this._lastAudioInput = [];
    this._resultEndTime = 0;
    this._finalRequestEndTime = 0;
    this._bridgingOffset = 0;
    this._speechCallback = this._speechCallback.bind(this);
    this._restartRecognizing = this._restartRecognizing.bind(this);
    this.destroy = this.destroy.bind(this);
  }

  getApiClient() {
    if (this.type === CAPTION_TYPE.AZURE) {
      return new AzureSTTShim({ clientConfig: this.clientConfig, log: this._log.bind(this) });
    } else if (this.type === CAPTION_TYPE.REVAI) {
      return new RevaiSTTShim({ clientConfig: this.clientConfig, log: this._log.bind(this) });
    }
  }

  /**
   * Starts all children
   */
  async _onStart() {
    this._startRecording();
    this._startRecognizing();
    return super._onStart();
  }

  /**
   * Hides all messages
   */
  async _onStop() {
    this._stopRecording();
    this._stopRecognizing();
    return super._onStop();
  }

  _startRecognizing() {
    this._log('verbose', `Starting speech to text with recording config = ${JSON.stringify(this.recordingConfig, null, 2)}`);
    this._log('verbose', `Starting speech to text with api config = ${JSON.stringify(this.sessionConfig, null, 2)}`);

    this._newStream = true;
    this._audioInput = [];
    this.apiClient.start(
      this.sessionConfig,
      this._speechCallback,
      this.destroy,
      this._restartRecognizing,
    );

    // Restart stream when streamingLimit expires
    if (this.recordingConfig.streamingLimit > 0) {
      this._restartRecognizerTimer = setTimeout(() => {
        this._log('debug', 'Restart timer triggered');
        this._restartRecognizing();
      }, this.recordingConfig.streamingLimit);
    }
  }

  _stopRecognizing() {
    this._log('verbose', 'Stopping recognizer');
    if (this.apiClient) {
      this.apiClient.stop();
    }

    if (this._resultEndTime > 0) {
      this._finalRequestEndTime = this._resultEndTime;
    }
    if (this._restartRecognizerTimer) {
      clearTimeout(this._restartRecognizerTimer);
      delete this._restartRecognizerTimer;
    }
    this._resultEndTime = 0;
    this._lastAudioInput = [];
    this._lastAudioInput = this._audioInput;
  }

  _speechCallback({
    transcript, ts, sequence, recognized,
  }) {
    // Calculate correct time based on offset from audio sent twice
    this._resultEndTime = ts;
    const correctedTime = (this._resultEndTime - this._bridgingOffset) + (this.recordingConfig.streamingLimit * this._restartRecognizeCounter);
    this.emit('result', {
      time: correctedTime, transcript, sequence, recognized,
    });
  }

  _restartRecognizing() {
    this._log('verbose', 'Restarting recognizer');
    this._stopRecognizing();
    this._restartRecognizeCounter += 1;
    this._startRecognizing();
  }

  _startRecording() {
    this._log('verbose', 'Starting recorder');

    this._recorder = recorder.record(this.recordingConfig);

    this._recordStream = this._recorder
      .stream()
      .on('error', err => {
        this._log('error', 'Audio recording error', err);
      });

    this._recordStream.pipe(new Transform({
      transform: (chunk, _encoding, callback) => {
        if (this._newStream && this._lastAudioInput.length !== 0) {
          // Approximate math to calculate time of chunks
          const chunkTime = this.recordingConfig.streamingLimit / this._lastAudioInput.length;
          if (chunkTime !== 0) {
            if (this._bridgingOffset < 0) {
              this._bridgingOffset = 0;
            }
            if (this._bridgingOffset > this._finalRequestEndTime) {
              this._bridgingOffset = this._finalRequestEndTime;
            }
            const chunksFromMS = Math.floor((this._finalRequestEndTime - this._bridgingOffset) / chunkTime);
            this._bridgingOffset = Math.floor((this._lastAudioInput.length - chunksFromMS) * chunkTime);

            for (let i = chunksFromMS; i < this._lastAudioInput.length; i++) {
              this.apiClient.write(this._lastAudioInput[i]);
            }
          }
          this._newStream = false;
        }

        this._audioInput.push(chunk);
        this.apiClient.write(chunk);

        callback();
      },
    }));
  }

  _stopRecording() {
    if (this._recorder) {
      this._recorder.stop();
      delete this._recorder;
    }
    if (this._recordStream) {
      this._recordStream.destroy();
      delete this._recordStream;
    }
  }
}

export default SpeechToTextStream;
export { SpeechToTextStream };
