import {
  AudioConfig,
  AudioInputStream,
  SpeechConfig,
  SpeechRecognizer,
  CancellationReason,
  PropertyId,
  ResultReason,
} from 'microsoft-cognitiveservices-speech-sdk';
import STTShim from './STTShim';

export default class AzureSTTShim extends STTShim {
  constructor({ clientConfig, log }) {
    super({ log });

    const { key, region } = clientConfig.credentials;
    this.clientConfig = SpeechConfig.fromSubscription(key, region);
  }

  start(configuration = {}, onTranscript, onError, _onRestart) {
    this.pushStream = AudioInputStream.createPushStream();
    const audioConfig = AudioConfig.fromStreamInput(this.pushStream);
    this.clientConfig.speechRecognitionLanguage = configuration.language;
    // see https://learn.microsoft.com/en-us/azure/cognitive-services/speech-service/captioning-concepts?pivots=programming-language-csharp
    // recommnended is 3
    this.clientConfig.setProperty(PropertyId.SpeechServiceResponse_StablePartialResultThreshold, configuration.stablePartialResultThreshold || '3');
    // 'default' or 'TrueText'
    this.clientConfig.setProperty(PropertyId.SpeechServiceResponse_PostProcessingOption, configuration.postProcessingOption || 'TrueText');
    this.language = configuration.language;
    this._recognizedLines = [];

    this.recognizer = new SpeechRecognizer(this.clientConfig, audioConfig);

    this.recognizer.recognizing = ((s, e) => {
      const { reason } = e.result;
      if (reason === ResultReason.RecognizingSpeech) {
        this.onData(e, onTranscript, false);
      } else if (reason === ResultReason.NoMatch) {
        this._log('silly', 'NOMATCH: Speech could not be recognized.');
      }
    });

    this.recognizer.recognized = ((s, e) => {
      const { reason } = e.result;
      if (reason === ResultReason.RecognizedSpeech) {
        this.onData(e, onTranscript, true);
      } else if (reason === ResultReason.NoMatch) {
        this._log('silly', 'NOMATCH: Speech could not be recognized.');
      }
    });

    this.recognizer.canceled = (s, e) => {
      let str = `(cancel) Reason: ${CancellationReason[e.reason]}`;
      if (e.reason === CancellationReason.Error) {
        str = `str: ${e.errorDetails}`;
      }
      this._log('info', str);
      onError();
    };

    this.recognizer.startContinuousRecognitionAsync(
      (result) => this._log('debug', JSON.stringify(result)),
      (error) => {
        this._log.error('Something went wrong while starting up the AzureSTTShim', error);
        onError();
      },
    );
  }

  write(chunk) {
    this.pushStream.write(chunk);
  }

  onData(event, callback, recognized) {
    const { text, offset, duration } = event.result;

    // Azure sometimes generates results without text if there is noise present
    // better to filter those updates out
    if (text) {
      // see https://github.com/Azure-Samples/cognitive-services-speech-sdk/blob/master/scenarios/csharp/dotnetcore/captioning/captioning/Program.cs
      const startTime = AzureSTTShim.ticksToMilliSeconds(offset);
      const endTime = startTime + AzureSTTShim.ticksToMilliSeconds(duration);
      if (this._previousEndTime && this._previousEndTime > endTime) {
        // Do nothing.
        return;
      }

      // Record the end timestamp for this result.
      this._previousEndTime = endTime;

      this._log('silly', 'Final recognition result', { transcript: text, ts: startTime });

      callback({
        transcript: text, ts: startTime, sequence: startTime, recognized,
      });
    }
  }

  stop() {
    if (this.recognizer) {
      this.recognizer.stopContinuousRecognitionAsync();
      this.recognizer.dispose();
      delete this.recognizer;
    }
  }

  static ticksToMilliSeconds(ticks) {
    return ticks * 10_000; // ticks are 100ns
  }
}
