import { AudioConfig, RevAiStreamingClient, SessionConfig } from 'revai-node-sdk';
import STTShim from './STTShim';

export default class RevaiSTTShim extends STTShim {
  constructor({ clientConfig, log }) {
    super({ log });
    const { credentials } = clientConfig;
    this.audioConfig = new AudioConfig('audio/x-raw', 'interleaved', 16000, 'S16LE', 1);
    this.apiClient = new RevAiStreamingClient(credentials?.token, this.audioConfig);

    this._previousTS = 0;
  }

  start(configuration = {}, onTranscript, onError, _onRestart) {
    this.sessionConfig = new SessionConfig(configuration);
    this.sessionConfig.startTs = Math.floor(new Date().getTime() / 1000);
    this._recognizeStream = this.apiClient.start(this.sessionConfig);

    this._recognizeStream
      .on('data', (event) => this.onData(event, onTranscript))
      .on('error', onError);
  }

  write(chunk) {
    if (this._recognizeStream) {
      this._recognizeStream.write(chunk);
    }
  }

  stop() {
    if (this._recognizeStream) {
      this._recognizeStream.end();
      delete this._recognizeStream;
    }

    this.apiClient.removeAllListeners();
    this.apiClient.end();
  }

  onData(event, cb) {
    const isFinal = event.type === 'final';

    // only final events contain spaces/punctuation
    const separator = isFinal ? '' : ' ';

    const transcript = event.elements.reduce((acc, { value }) => {
      acc = `${acc}${separator}${value}`;
      return acc;
    }, '');

    // final transcriptions sometimes have a different start ts than the partial results
    // since we use this as a sequence ID, we need to cache and reuse the previous TS
    const sequence = isFinal ? this._previousTS : event.ts * 1000;
    this._previousTS = event.ts * 1000;

    // returned timestamps are expected to be in ms
    const ts = event.end_ts * 1000;

    this._log('silly', 'Final recognition result', { transcript, ts });
    cb({ transcript, ts, sequence });
  }
}
