/**
 * (C) 2015 Copyright BARCO NV
 * All rights reserved. Reproduction in whole or in part is
 * prohibited without the written consent of the copyright owner.
 *
 * Part of Barco WeConnect
 */

/*
eslint no-console: "off"
*/

import { spawn, execSync, exec } from 'child_process';
import * as fs from 'fs-extra';
import * as spawnargs from 'spawn-args';
import { isString } from 'lodash';
import { promisify } from 'util';

const execAsync = promisify(exec);

function noop() { }

function execParallel(commands) {
  const promises = commands.map(cmd => execAsync(cmd));
  return Promise.all(promises);
}

async function execSynchonous(commands) {
  for (const command of commands) {
    await execSync(command); //eslint-disable-line
  }
}

async function execSynchonousAsync(commands) {
  for (const command of commands) {
    await execAsync(command); //eslint-disable-line
  }
}

function getRandomInt(min, max) {
  return Math.floor(Math.random() * ((max - min) + 1)) + min;
}

function runChildProcess({
  tag = '',
  command,
  params = [],
  cwd = process.cwd(),
  env = {},
  logger,
  detached = false,
}) {
  if (isString(params)) {
    params = spawnargs(params, {
      removequotes: 'always',
    });
  }

  if (logger) {
    logger.log('verbose', `${tag}: Starting child process "${command}" with params: ${params.join(' ')}`);
  }

  const options = {
    detached,
    cwd,
    env: {
      ...process.env,
      env,
    },
  };

  if (detached) {
    options.stdio = ['ignore', 'ignore', 'ignore'];
  }

  const proc = spawn(command, params, options);

  if (detached) {
    proc.unref();
  }

  return {
    proc,
    promise: new Promise((resolve, reject) => {
      if (!detached) {
        proc.stdout.on('data', (data) => {
          if (logger) {
            logger.log('verbose', `${tag}: ${data}`);
          }
        });
        proc.stderr.on('data', (data) => {
          if (logger) {
            logger.log('error', `${tag}: ${data}`);
          }
        });
      }

      proc.on('close', (code) => {
        if (code === 0) {
          if (logger) {
            logger.log('info', `${tag}: Process ended.`);
          }
          resolve(`${tag}: Process ended.`);
        } else {
          if (logger) {
            logger.log('error', `${tag}: Process exit with code ${code}.`);
          }
          reject(new Error(`${tag}: Process exit with code ${code}.`));
        }
      });
    }),
  };
}

function waitForChildProcessToFinish(childProcess) {
  return new Promise((resolve, reject) => {
    childProcess.on('error', (error) => reject(error));
    childProcess.on('exit', (code) => ((code === 0)
      ? resolve()
      : reject(new Error(`Exit code ${code}`))));
  });
}

function rebootSystem({ shutdownDelay = 0, startupDelay = 0 } = {}) {
  shutdownDelay = Math.max(0, shutdownDelay);
  startupDelay = Math.max(0, startupDelay);
  if (shutdownDelay) {
    // We use this to spread the load of device restarts
    // Consumers of this method expect nothing to be executed after calling the reboot command
    // Therefore we use the linux sleep method in sync mode instead of some async sleep
    execSync(`sleep ${shutdownDelay}`);
  }
  if (startupDelay) {
    // RTC wake allows us to wait a couple of seconds before starting up
    // This can be useful to ensure USB devices get powered off for some time
    // e.g. in order to reset frozen devices
    // However, we did find that this is a bit flaky and sometimes devices don't restart
    execSync(`sudo rtcwake -m off -s ${startupDelay}`);
  }
  execSync('sudo reboot');
}

function getAvailableDiskSpace() {
  return execSync('df --output=avail / | sed 1d', { encoding: 'utf-8' });
}

function getAvailableBootSpace() {
  return execSync('df --output=avail /boot | sed 1d', { encoding: 'utf-8' });
}

function createDirectory(directory) {
  // Try to create the directory
  try {
    fs.ensureDirSync(directory, '0755');
  } catch (error) {
    console.log('error', `Error creating folder "${directory}"`, error);
  }
  // Now check if it exists
  try {
    const stats = fs.lstatSync(directory);
    if (stats.isDirectory()) {
      console.log('verbose', `Folder "${directory}" exists`);
      return (directory);
    }
  } catch (error) {
    console.log('verbose', `Folder "${directory}" does not exist`);
    return false;
  }
}

async function wait(timeInMS, ...args) {
  return new Promise(function waiter(resolve) {
    const restartTimerHandle = setTimeout(resolve, timeInMS, ...args);
    if (this) {
      this._restartTimerHandle = restartTimerHandle;
    }
  });
}

function safeStringify(ob) {
  const cache = [];
  return JSON.stringify(ob, (key, value) => {
    if (typeof value === 'object' && value !== null) {
      if (cache.includes(value)) {
        // Circular reference found, discard key
        return;
      }
      // Store value in our collection
      cache.push(value);
    }
    return value;
  }, 2);
}

function sleep(timeInMS) {
  return new Promise(resolve => setTimeout(resolve, timeInMS));
}

function timeout(timeInMS) {
  return new Promise((_, reject) => setTimeout(() => reject(new Error(`timeout after ${timeInMS} ms`)), timeInMS));
}

function waitMaximum(promise, timeInMS, error = true) {
  return Promise.race([
    promise,
    error ? timeout(timeInMS) : sleep(timeInMS),
  ]);
}

export {
  noop,
  execAsync,
  execParallel,
  execSynchonous,
  execSynchonousAsync,
  runChildProcess,
  waitForChildProcessToFinish,
  getAvailableDiskSpace,
  getAvailableBootSpace,
  rebootSystem,
  getRandomInt,
  createDirectory,
  wait,
  safeStringify,
  sleep,
  timeout,
  waitMaximum,
};
